(*	XYZ GeoBench (eXperimental geometrY Zurich):		*)
(*	An object oriented, interactive workbench for		*)
(*	experimental geometric computation.					*)
(*														*)
(*	 1999 by Peter Schorn, peter.schorn@acm.org		*)
(*	Culmannstrasse 77, CH-8006 Zrich, Switzerland		*)
(*	This software may be copied freely for educational	*)
(*	and research purposes only. For commercial use,		*)
(*	please contact the author.							*)
(*														*)
(*	Version 5.0.5: 30-Sep-1999							*)

(* Unit written by Michael Vollenbroeker				*)

unit PlaneSweepClosestPair;

interface

uses
	pointVector;

implementation

 {$SETC AclosestPair = TRUE }

const
	below = left;  (* unteres konvexes Objekt steht in y-table (AVL-Baum) in linkem Sohn *)
	above = right;

type

	closestPairEvent = object(obj)                    (* Closest Pair Event *)
			where: point2d;
			procedure processEvent (yTable: dictionary; var shortestConnection: lineSegment; var delta: extended);
			procedure freeAll;
			override;
		end;

	cpInsertEvent = object(closestPairEvent)  (* insertXqueue enthaelt Objekte vom Typ cpInsertEvent *)
			cobj: obj;   (* convex object *)
			deleteEvent: cpDeleteEvent;
			procedure processEvent (yTable: dictionary; var shortestConnection: lineSegment; var delta: extended);
			override;
		end;

	cpDeleteEvent = object(closestPairEvent)  (* deleteXqueue enthaelt Objekte vom Typ cpDeleteEvent *)
			toDelete: reference;   (* in y-table *)
			procedure processEvent (yTable: dictionary; var shortestConnection: lineSegment; var delta: extended);
			override;
		end;

function convexPolygon.HBGLessThan (o: obj): boolean;
(* Diese Methode bestimmt, ob 'self' bezueglich der von Hinrichs/Bartling/Graf definierten Ordnung kleiner als Q ist. *)
(* Die Position der sweepline ist dabei bestimmt durch die x-Koordinate des am weitesten links liegenden Punkts des rechten Polygons. *)
	begin
		HBGLessThan := objectYLessThan(convexPolygon(o));
	end;


function compareConvexObject (cp, cq: obj): relationObj;
	begin
		if cp.HBGLessThan(cq) then
			compareConvexObject := lessThanObj
		else
			compareConvexObject := greaterThanObj;  (* equalObj tritt also nie auf *)
	end;

function lessThancpEvent (a, b: closestPairEvent): boolean;
	begin
		lessThancpEvent := a.where.genericLessThan(b.where);  (* 'true', falls a.where lexikographisch kleiner als b.where *)
	end;


{$IFC NOT THINK_PASCAL }
{$PUSH }
{$PRAGMAC warn_unusedarg off }
{$ENDC }
procedure closestPairEvent.processEvent (yTable: dictionary; var shortestConnection: lineSegment; var delta: extended);
	begin
		unimplementedMethod('closestPairEvent.processEvent');
	end;
{$IFC NOT THINK_PASCAL }
{$POP }
{$ENDC }

procedure checkDelta (cp, cq: obj; var shortestConnection: lineSegment; var delta: extended);
	var
		minP, minQ: point2d;                                            (* cp liegt unter cq *)
		dl: lineSegment;
		newDelta: extended;
		oldAbortEnable: boolean;
	begin
		if cp.verticalDistance(cq) < delta then begin
                  (* Zunaechst wird der vertikale Abstand zwischen dem unteren Punkt des oberen Objekts *)
                  (* und dem oberen Punkt des unteren Objekts ueberprueft. Ist dieser groesser als delta, so *)
                  (* eruebrigt sich die Abstandsberechnung der Objekte, denn man weiss dann, dass ihr Abstand groesser *)
                  (* als delta ist. *)
			new(dl);
			failNil(Handle(dl));
			minP := point2d(newInitObj(cp.getPointType));
			minQ := point2d(newInitObj(cq.getPointType));
			oldAbortEnable := abortEnable;
			abortEnable := false;							(* interrupts fuer Abstandsberechnung werden verboten *)
			newDelta := cp.distanceO(cq, obj(minP), obj(minQ));  (* berechnet Abstand von cp und cq, der in minP und minQ angenommen wird *)
			dl.p := point2d(minP.duplicate);
			dl.q := point2d(minQ.duplicate);
			minP.freeAll;
			minQ.freeAll;
			abortEnable := oldAbortEnable;
{$IFC AnimationEnabled AND AclosestPair }
			if animationFlag[AClosestPairItem] then
				dl.interactiveFlash;                                  (* ueberpruefte Verbindung blitzt auf *)
{$ENDC }
			if newDelta < delta then begin                  (* Es gibt neue kuerzeste Verbindung. *)
				delta := newDelta;
{$IFC AnimationEnabled AND AclosestPair }
				if animationFlag[AClosestPairItem] then
					shortestConnection.interactiveHighlight;                         (* loescht alte kuerzeste Verbindung *)
{$ENDC }
				shortestConnection.freeAll;
				shortestConnection := dl;
{$IFC AnimationEnabled AND AclosestPair }
				if animationFlag[AClosestPairItem] then
					shortestConnection.interactiveHighlight;                           (* zeichnet neue kuerzeste Verbindung *)
{$ENDC }
			end
			else
				dl.freeAll;                                      (* neuer Abstand war nicht kuerzer als alter. Loesche dl. *)
		end
	end;

procedure closestPairEvent.freeAll;
	begin
		where.freeAll;
		free;
	end;

procedure cpInsertEvent.processEvent;
	var
		whereRef: reference;
		direction: directionType;
		aboveRef, belowRef: reference;
	begin
		if not yTable.find(cobj, compareConvexObject, whereRef, direction) then begin  (* findet Einfuegeposition fuer cobj *)
{$IFC AnimationEnabled AND AclosestPair }
			if animationFlag[AClosestPairItem] then begin
				cobj.interactiveHighlight;    (* zeichnet neu eingefuegtes Objekt fett *)
			end;
{$ENDC }
			whereRef := yTable.insert(cobj, whereRef, direction);       (* Einfuegen von cobj in y-table *)
			deleteEvent.toDelete := whereRef;                                             (* Merken der Referenz fuer deleteEvent *)
			aboveRef := yTable.next(whereRef, above);                               (* Referenz auf oberen Nachbarn von cobj *)
			belowRef := yTable.next(whereRef, below);                               (* Referenz auf unteren Nachbarn von cobj *)
			if aboveRef <> nil then
				checkDelta(cobj, yTable.getObject(aboveRef), shortestConnection, delta);
			if (belowRef <> nil) & (delta > 0) then
				checkDelta(yTable.getObject(belowRef), cobj, shortestConnection, delta);
		end;
    (* ELSE-Fall tritt nach Konstruktion von compareConvexObject nie auf *)
	end;

procedure cpDeleteEvent.processEvent;
	var
		aboveRef, belowRef: reference;
	begin
		aboveRef := yTable.next(toDelete, above);
		belowRef := yTable.next(toDelete, below);
		if (aboveRef <> nil) & (belowRef <> nil) then  (* ueberprueft Abstand der Nachbarn des zu loeschenden Objekts *)
			checkDelta(yTable.getObject(belowRef), yTable.getObject(aboveRef), shortestConnection, delta);
		yTable.delete(toDelete);
	end;

(* ------------------------------------------------------------ *)

function obj.closestPair: obj;  (* liefert in Wirklichkeit lineSegment *)
	var
		yTable: avlTree;
		insertXQueue, deleteXQueue: vector;
		shortestConnection: lineSegment;
		current, tail: seqIndex1;
		NumberOfObjects: seqIndex;
		delta: extended;

{$IFC AnimationEnabled }
		aFlag: boolean;
{$ENDC }


{$IFC AnimationEnabled AND AclosestPair }
		deletedObject: obj;
		oldDelta: extended;
{$ENDC }


 {$IFC AnimationEnabled AND AclosestPair }
	procedure drawStatus (p: point2d; delta: extended);
  (* zeichnet sweepline durch p und delta-Streifen links davon *)
		begin
			drawThickVertical(p.getX);
			drawVertical(p.getX - delta);
		end;
{$ENDC }

	procedure initialization;

		procedure insertInQueues (cp: obj);  (* fuegt cp in insert- und deleteQueue ein *)
			var
				insertH: cpInsertEvent;
				deleteH: cpDeleteEvent;
			begin
				new(deleteH);
				failNil(Handle(deleteH));
				deleteH.where := point2d(cp.getRightPoint);
				deleteH.toDelete := nil;  (* wird in InsertEvent gesetzt *)
				deleteXQueue.append(deleteH);
				new(insertH);
				failNil(Handle(insertH));
				insertH.where := point2d(cp.getLeftPoint);
				insertH.cobj := cp;
				insertH.deleteEvent := deleteH;
				insertXQueue.append(insertH);
   {$IFC AnimationEnabled AND AclosestPair }
				if animationFlag[AClosestPairItem] then
					cp.interactiveHighlight;
   {$ENDC }
			end;

		procedure checkQueueEqualElements (queue: vector);
			var
				i: seqIndex;
				ls: lineSegment;
			begin
				for i := 1 to NumberOfObjects - 1 do
					if point2d(closestPairEvent(queue.elements^^[i]).where).equal(point2d(closestPairEvent(queue.elements^^[i + 1]).where)) then begin
         (* Es gibt zwei Objekte mit gleichem linken bzw. rechten Extremalpunkt. *)
						new(ls);
						failNil(Handle(ls));
						ls.p := point2d(closestPairEvent(queue.elements^^[i]).where.duplicate);
						ls.q := point2d(ls.p.duplicate);
						closestPair := ls;
						deleteXQueue.freeAll;
						insertXQueue.freeAll;
						yTable.free;
						exit(closestPair)
					end;
			end;

		procedure buildAndSortTwoXQueues;
			begin
				currentArguments.forAll(insertInQueues);
				insertXQueue.sortAll(lessThancpEvent);
				deleteXQueue.sortAll(lessThancpEvent);
				checkQueueEqualElements(insertXQueue);
				checkQueueEqualElements(deleteXQueue);
			end;

		begin  (* initialization *)
			NumberOfObjects := currentArguments.length;
			currentVectorLength := NumberOfObjects;
			new(yTable);
			failNil(Handle(yTable));
			yTable.init;
			new(insertXQueue);
			failNil(Handle(insertXQueue));
			insertXQueue.init;
			new(deleteXQueue);
			failNil(Handle(deleteXQueue));
			deleteXQueue.init;
			buildAndSortTwoXQueues;
			new(shortestConnection);
			failNil(Handle(shortestConnection));
			shortestConnection.p := point2d(newObj(cpInsertEvent(insertXQueue.elements^^[1]).cobj.getPointType));
			shortestConnection.p.setX(-100);
			shortestConnection.p.setY(-100);
			shortestConnection.q := point2d(shortestConnection.p.duplicate);
  (* unsichtbares Liniensegment shortestConnection wird gezeichnet. Dies ist noetig, *)
  (*weil checkDelta bei der Animation zunaechst die bisherige shortestConnection loescht. *)
{$IFC AnimationEnabled AND AclosestPair }
			if animationFlag[AClosestPairItem] then
				shortestConnection.interactiveHighlight;
{$ENDC }
			delta := 100000000;

{$IFC AnimationEnabled }
			aFlag := animationFlag[AConvexPolygonDistanceItem];
			animationFlag[AConvexPolygonDistanceItem] := false;      (* keine Animation der Abstandsberechnung  *)
{$ENDC }

			closestPairEvent(insertXQueue.elements^^[1]).processEvent(yTable, shortestConnection, delta);
			closestPairEvent(insertXQueue.elements^^[2]).processEvent(yTable, shortestConnection, delta);
			tail := 1;        (* markiert naechstes zu loeschendes Objekt *)
			current := 3;  (* markiert naechstes einzufuegendes Objekt *)
		end;   (* initialization *)



	begin  (* closestPair *)
		initialization;
{$IFC AnimationEnabled AND AclosestPair }
		if animationFlag[AClosestPairItem] then begin
			oldDelta := delta;
			drawStatus(closestPairEvent(insertXQueue.elements^^[2]).where, delta);
		end;
{$ENDC }

		while (current <= NumberOfObjects) and (delta > 0) do begin                (* Sweepen durch Ebene solange noch Objekte in insertQueue *)

{$IFC AnimationEnabled AND AclosestPair }
			if animationFlag[AClosestPairItem] then begin
				drawStatus(closestPairEvent(insertXQueue.elements^^[current - 1]).where, delta);      (* loescht alte sweepline und delta-Streifen *)
				drawStatus(closestPairEvent(insertXQueue.elements^^[current]).where, delta);            (* zeichnet neue sweepline und delta-Streifen *)
			end;
{$ENDC }

			while (closestPairEvent(deleteXQueue.elements^^[tail]).where.getX <= (closestPairEvent(insertXQueue.elements^^[current]).where.getX - delta)) do begin
       (* zunaechst werden deactivation events durchgefuehrt *)

{$IFC AnimationEnabled AND AclosestPair }
				if animationFlag[AClosestPairItem] then begin
					waitForClick(animationFlag[AClosestPairItem]);
					deletedObject := yTable.getObject(cpDeleteEvent(deleteXQueue.elements^^[tail]).toDelete);
				end;
{$ENDC }

{$IFC AnimationEnabled }
				if userAbort then begin
					shortestConnection.freeAll;
					deleteXQueue.freeAll;
					insertXQueue.freeAll;
					yTable.free;
					closestPair := nil;
					exit(closestPair);
				end;
{$ENDC }

				closestPairEvent(deleteXQueue.elements^^[tail]).processEvent(yTable, shortestConnection, delta);    (* DeleteEvent *)
				if delta = 0 then begin
					deleteXQueue.freeAll;
					insertXQueue.freeAll;
					yTable.free;
					closestPair := shortestConnection;
					exit(closestPair)
				end;

{$IFC AnimationEnabled AND AclosestPair }
				if animationFlag[AClosestPairItem] then begin
					if (delta < oldDelta) then begin
						drawStatus(closestPairEvent(insertXQueue.elements^^[current]).where, oldDelta);      (* loescht sweepline und alten delta-Streifen *)
						drawStatus(closestPairEvent(insertXQueue.elements^^[current]).where, delta);           (* zeichnet sweepline und neuen delta-Streifen *)
						oldDelta := delta;
					end;
					deletedObject.interactiveHighlight;  (* zeichnet geloeschtes Objekt duenn *)
				end;
{$ENDC }

				tail := tail + 1;
			end;  (* WHILE-Schleife fuer deactivation events *)


{$IFC AnimationEnabled AND AclosestPair }
			if animationFlag[AClosestPairItem] then
				waitForClick(animationFlag[AClosestPairItem]);
{$ENDC }

{$IFC AnimationEnabled }
			if userAbort then begin
				shortestConnection.freeAll;
				deleteXQueue.freeAll;
				insertXQueue.freeAll;
				yTable.free;
				closestPair := nil;
				exit(closestPair);
			end;
{$ENDC }

			closestPairEvent(insertXQueue.elements^^[current]).processEvent(yTable, shortestConnection, delta);  (* insertEvent *)

{$IFC AnimationEnabled AND AclosestPair }
			if animationFlag[AClosestPairItem] then
				if (delta < oldDelta) then begin
					drawStatus(closestPairEvent(insertXQueue.elements^^[current]).where, oldDelta);      (* loescht sweepline und alten delta-Streifen *)
					drawStatus(closestPairEvent(insertXQueue.elements^^[current]).where, delta);           (* zeichnet sweepline und neuen delta-Streifen *)
					oldDelta := delta;
				end;
{$ENDC }

			current := current + 1;
		end;  (* WHILE-Schleife fuer nichtleere insertQueue *)

{$IFC AnimationEnabled AND AclosestPair }
		if animationFlag[AClosestPairItem] then
			drawStatus(closestPairEvent(insertXQueue.elements^^[current - 1]).where, delta);     (* loescht letzte sweepline und delta-Streifen *)
{$ENDC }


		while (tail <= NumberOfObjects - 2) and (delta > 0) do begin              (* Es sind mehr noch mehr als zwei Objekte in der deleteQueue *)

{$IFC AnimationEnabled AND AclosestPair }
			if animationFlag[AClosestPairItem] then begin
				waitForClick(animationFlag[AClosestPairItem]);
				deletedObject := yTable.getObject(cpDeleteEvent(deleteXQueue.elements^^[tail]).toDelete);
			end;
{$ENDC }
{$IFC AnimationEnabled }
			if userAbort then begin
				shortestConnection.freeAll;
				deleteXQueue.freeAll;
				insertXQueue.freeAll;
				yTable.free;
				closestPair := nil;
				exit(closestPair);
			end;
{$ENDC }

			closestPairEvent(deleteXQueue.elements^^[tail]).processEvent(yTable, shortestConnection, delta);             (* deleteEvent *)

{$IFC AnimationEnabled AND AclosestPair }
			if animationFlag[AClosestPairItem] then begin
				deletedObject.interactiveHighlight;
			end;
{$ENDC }
			tail := tail + 1;
		end;  (* while *)

		deleteXQueue.freeAll;  (* fuer die beiden letzten Objekte in der deleteQueue ist kein deleteEvent mehr erforderlich *)
		insertXQueue.freeAll;
		yTable.free;

{$IFC AnimationEnabled }
		animationFlag[AConvexPolygonDistanceItem] := aFlag;
{$ENDC }

		closestPair := shortestConnection;
	end;  (* closest pair *)

function convexPolygon.distance (Q: convexPolygon; var minP, minQ: point2d): extended;
	var
		dl: lineSegment;
		d: extended;
	begin
		minP.freeAll;
		minQ.freeAll;
		dl := self.minimalDistance(Q, d);
		minP := dl.p;
		minQ := dl.q;
		dl.shallowFree;
		distance := d;
	end;

procedure convexPolygon.findExtremalPoints;

	function cutIn2X (lo, hi: longInt): longInt;
		var                                (* always: d(lo)<>0 , d(hi)<>0 !      *)
			h: longInt;
			found: boolean;
			dlo: extended;
		begin
			found := FALSE;
			dlo := point2d(elements^^[lo]).DxX(point2d(elements^^[((length + lo - 2) mod length) + 1]));
			repeat
				h := (lo + hi) div 2;
				if dlo * point2d(elements^^[h]).DxX(point2d(elements^^[((length + h - 2) mod length) + 1])) < 0 then
                                                                                            (* one point above, one below i-axis  *)
					found := TRUE                                                               (* a point exists because of Lemma 2. *)
				else                                                                              (* both points on same side of i-axis *)
					if dlo * point2d(elements^^[lo]).DxX(point2d(elements^^[h])) <= 0 then
						lo := h
					else
						hi := h;                                                                          (* to handle this case see Lemma 1.   *)
			until found;
			cutIn2X := h;                                                                        (* => d(h)<>0 !                       *)
		end;

  (* ---------------------------------------- *)

	function findExtremumX (lo, hi: longInt): longInt;
		var                                             (* d(lo)<>0 and d(hi)<>0              *)
			h: longInt;
			dlo, dh: extended;
		begin
			dlo := point2d(elements^^[lo]).DxX(point2d(elements^^[((length + lo - 2) mod length) + 1]));
			while (hi - lo) > 1 do begin        (* d(lo)<>0 and d(hi) <> 0 or if      *)
				h := (lo + hi) div 2;                   (* not, then hi-lo <= 1               *)
				dh := point2d(elements^^[h]).DxX(point2d(elements^^[((length + h - 2) mod length) + 1]));
				if dh = 0 then begin
					lo := h;                     (* if d(h)=0 extremun is found, so    *)
					hi := lo;                    (*    set lo=h; hi=lo will stop loop! *)
				end
				else begin                     (* intervall is decreased by half     *)
					if dlo * dh < 0 then
						hi := h
					else
						lo := h;
				end;
			end;                             (* Either lo was set to extremum by   *)
			findExtremumX := lo;                     (* hand (d=0 case) or only two points *)
		end;                               (* are left: then lo is extremum!     *)

  (* ---------------------------------------- *)

	procedure findExtremaX;
		var
			e1, e2, h: longInt;
			d1, dm: extended;
		begin
			if length <= 2 then begin
				e1 := 1;
				e2 := length;
			end
			else begin
				d1 := point2d(elements^^[1]).DxX(point2d(elements^^[length]));
				dm := point2d(elements^^[length]).DxX(point2d(elements^^[length - 1]));
				if d1 * dm > 0 then begin          (*  d(1)<>0 and d(m)<>0           *)
					h := cutIn2X(1, length);              (*  => d(h)<>0                    *)
					e1 := findExtremumX(1, h);         (*  d(1)<>0 and d(h)<>0           *)
					e2 := findExtremumX(h, length);   (*  d(h)<>0 and d(m)<>0           *)
				end
				else if d1 = 0 then begin
					e1 := 1;
					e2 := findExtremumX(2, length);   (* d(2)<>0 and d(m)<>0            *)
				end
				else if dm = 0 then begin
					e1 := length;
					e2 := findExtremumX(1, length - 1);
				end
				else begin                        (* d1*dm < 0 case                  *)
					e1 := length;
					if point2d(elements^^[length - 1]).DxX(point2d(elements^^[length - 2])) * dm <= 0 then
						e2 := length - 1
					else
						e2 := findExtremumX(1, length - 1);                         (* d(p,1)<>0 and d(p,m-1)<>0 *)
				end;
			end;
			if point2d(elements^^[e1]).xLessThan(point2d(elements^^[e2])) then begin     (* figure out, which extremum is      *)
				left := e1;                                                                                                (* the maximum and which the minimum. *)
				right := e2;
			end
			else begin
				left := e2;
				right := e1;
			end;
		end;

(* ---------------------------------------- *)

	function cutIn2Y (lo, hi: longInt): longInt;
		var                                (* always: d(lo)<>0 , d(hi)<>0 !      *)
			h: longInt;
			found: boolean;
			dlo: extended;
		begin
			found := FALSE;
			dlo := point2d(elements^^[lo]).DyX(point2d(elements^^[((length + lo - 2) mod length) + 1]));
			repeat
				h := (lo + hi) div 2;
				if dlo * point2d(elements^^[h]).DyX(point2d(elements^^[((length + h - 2) mod length) + 1])) < 0 then
                                                                                            (* one point above, one below i-axis  *)
					found := TRUE                                                               (* a point exists because of Lemma 2. *)
				else                                                                              (* both points on same side of i-axis *)
					if dlo * point2d(elements^^[lo]).DyX(point2d(elements^^[h])) <= 0 then
						lo := h
					else
						hi := h;                                                                          (* to handle this case see Lemma 1.   *)
			until found;
			cutIn2Y := h;                                                                        (* => d(h)<>0 !                       *)
		end;

  (* ---------------------------------------- *)

	function findExtremumY (lo, hi: longInt): longInt;
		var                                             (* d(lo)<>0 and d(hi)<>0              *)
			h: longInt;
			dlo, dh: extended;
		begin
			dlo := point2d(elements^^[lo]).DyX(point2d(elements^^[((length + lo - 2) mod length) + 1]));
			while (hi - lo) > 1 do begin        (* d(lo)<>0 and d(hi) <> 0 or if      *)
				h := (lo + hi) div 2;                   (* not, then hi-lo <= 1               *)
				dh := point2d(elements^^[h]).DyX(point2d(elements^^[((length + h - 2) mod length) + 1]));
				if dh = 0 then begin
					lo := h;                                    (* if d(h)=0 extremun is found, so    *)
					hi := lo;                                  (*    set lo=h; hi=lo will stop loop! *)
				end
				else begin                              (* intervall is decreased by half     *)
					if dlo * dh < 0 then
						hi := h
					else
						lo := h;
				end;
			end;                                         (* Either lo was set to extremum by   *)
			findExtremumY := lo;                     (* hand (d=0 case) or only two points *)
		end;                                           (* are left: then lo is extremum!     *)

  (* ---------------------------------------- *)

	procedure findExtremaY;
		var
			e1, e2, h: longInt;
			d1, dm: extended;
		begin
			if length <= 2 then begin
				e1 := 1;
				e2 := length;
			end
			else begin
				d1 := point2d(elements^^[1]).DyX(point2d(elements^^[length]));
				dm := point2d(elements^^[length]).DyX(point2d(elements^^[length - 1]));
				if d1 * dm > 0 then begin          (*  d(1)<>0 and d(m)<>0           *)
					h := cutIn2Y(1, length);              (*  => d(h)<>0                    *)
					e1 := findExtremumY(1, h);         (*  d(1)<>0 and d(h)<>0           *)
					e2 := findExtremumY(h, length);   (*  d(h)<>0 and d(m)<>0           *)
				end
				else if d1 = 0 then begin
					e1 := 1;
					e2 := findExtremumY(2, length);   (* d(2)<>0 and d(m)<>0            *)
				end
				else if dm = 0 then begin
					e1 := length;
					e2 := findExtremumY(1, length - 1);
				end
				else begin                                 (* d1*dm < 0 case                  *)
					e1 := length;
					if point2d(elements^^[length - 1]).DyX(point2d(elements^^[length - 2])) * dm <= 0 then
						e2 := length - 1
					else
						e2 := findExtremumY(1, length - 1);                         (* d(p,1)<>0 and d(p,m-1)<>0 *)
				end;
			end;
			if point2d(elements^^[e1]).yLessThan(point2d(elements^^[e2])) then begin     (* figure out, which extremum is      *)
				bottom := e1;                                                                                                (* the maximum and which the minimum. *)
				top := e2;
			end
			else begin
				bottom := e2;
				top := e1;
			end;
		end;

(* ---------------------------------------- *)

	begin
		findExtremaX;
		findExtremaY;
	end;

(* -------------------------------------------------------------------------- *)

end.
