(*	XYZ GeoBench (eXperimental geometrY Zurich):		*)
(*	An object oriented, interactive workbench for		*)
(*	experimental geometric computation.					*)
(*														*)
(*	 1999 by Peter Schorn, peter.schorn@acm.org		*)
(*	Culmannstrasse 77, CH-8006 Zrich, Switzerland		*)
(*	This software may be copied freely for educational	*)
(*	and research purposes only. For commercial use,		*)
(*	please contact the author.							*)
(*														*)
(*	Version 5.0.5: 30-Sep-1999							*)

unit quadTree;


interface

uses
	basicGeometry;


const
	ne = 0;
	nw = 1;
	sw = 2;
	se = 3;
	seP = se + 1;

type
	quadrants = ne..se;		(* the four quadrants of the plane in mathematical succession:	*)
								(* north-east, north-west, south-west, south-east				*)
	quadrantsP = ne..seP;		(* for loop indices *)


	quadTreeNode = object(obj)

			son: array[quadrants] of record			(* the four sons *)
					leafIncluded: boolean;
					succ: quadTreeNode
				end;
		(* possible cases:																															*)
		(*  (son[d].leafIncluded=true) AND (son[d].succ=nil)		-->	object includes whole son rectangle [d]			*)
		(*  (son[d].leafIncluded=true) AND (son[d].succ<>nil)		-->	FORBIDDEN CONSTELLATION						*)
		(*  (son[d].leafIncluded=false) AND (son[d].succ=nil)		-->	whole son rectangle [d] doesn't belong to 		*)
		(*																			object; INITIAL STATE								*)
		(*  (son[d].leafIncluded=false) AND (son[d].succ<>nil)	-->	part of rectangle [d] belongs to object			*)

			procedure init;
			override;
			procedure free;
			override;
			procedure freeAll;
			override;
			function duplicate: obj;
			override;
			function duplicateAll: obj;
			override;
			procedure fileOutput;
			override;
			procedure fileInput;
			override;
			procedure textFileOutput;
			override;
			procedure textFileInput;
			override;
			function getType: objectType;
			override;
			function quadTreeNode.getTextType: Str255;
			override;
			function getInfo: obj;
			override;
			function memberType (o: obj): boolean;
			override;

			procedure freeNode;
(* Frees only the node. 																			*)

			procedure freeNodeAll;
(* Frees only the node and information belonging to the node (used in derived nodes).	*)

			function duplicateNode: obj;
(*	Duplicates only the node. 																		*)

			function duplicateNodeAll: obj;
(*	Duplicates the node and information belonging to the node (used in derived nodes).	*)

			procedure fileInputNode;
(*	Reads in a binary format just the extra information belonging to the node from		*)
(*	'currentStream'. 																				*)

			procedure fileOutputNode;
(*	Writes in a binary format just the extra information belonging to the node to			*)
(* 'currentStream'. The default implementation raises an error condition.				*)

			procedure textFileInputNode;
(*	Reads in a textual format just the extra information belonging to the node from		*)
(*	'currentStream'. The default implementation raises an error condition.				*)

			procedure textFileOutputNode;
(*	Writes in a textual format just the extra information belonging to the node to		*)
(* 'currentStream'. The default implementation raises an error condition.				*)

			function nodeCompare (x: quadTreeNode): relationObj;
(*	Defines a canonical ordering just on the extra information belonging to the node.		*)

			procedure insert (old: obj; inBottom, inLeft, inTop, inRight: extended; actualDepth, depth: integer);
(* The obj covers part of the rectangle (inLeft,inBttom,inTop,inRight). The procedure	*)
(* allocates new nodes according the given rectangle.										*)

			procedure negation;
(* tree is negated starting with "quadTreeNode" *)

			function intersection (qTN2: quadTreeNode; pos1, pos2: boolean): quadTreeNode;
(* returns intersection set of tree starting with "quadTreeNode" and tree "qTN2"; if pos1=true, then the first	*)
(* trees complement is taken as first operand (but tree does not change!); pos2:analog for second tree.			*)

			function union (qTN2: quadTreeNode; pos1, pos2: boolean): quadTreeNode;
(* returns union set of tree starting with "quadTreeNode" and tree "qTN2"; if pos1=true, then the first			*)
(* trees complement is taken as first operand (but tree does not change!); pos2:analog for second tree.			*)

			procedure interactiveOutputNode (inBottom, inLeft, inTop, inRight: integer);

			procedure interactiveOutputClipNode (y1, x1, y2, x2: integer);

			function hitsRect (where: obj; inBottom, inLeft, inTop, inRight: extended): boolean;

		end;  (* quadTreeNode *)

	quadTree = object(obj)
			root: quadTreeNode;			(* the quad trees root *)
			boundingRect: rectangle;		(* starting rectangle for the recursive patrtitioning ("universe", "virtual screen" ) *)

			procedure init;
			override;
			procedure free;
			override;
			procedure freeAll;
			override;
			function duplicate: obj;
			override;
			function duplicateAll: obj;
			override;

			procedure interactiveOutput;
			override;
			procedure interactiveHighlight;
			override;
			procedure windowOutput;
			override;

			procedure fileOutput;
			override;
			procedure fileInput;
			override;
			procedure textFileOutput;
			override;
			procedure textFileInput;
			override;

			procedure describe;
			override;
			procedure execute (item: integer);
			override;

			function getType: objectType;
			override;
			function getInfo: obj;
			override;
			function getTextType: Str255;
			override;
			function memberType (o: obj): boolean;
			override;
			function equal (o: obj): boolean;
			override;
			function quadTree.genericLessThan (o: obj): boolean;
			override;

			function isSelected (where: obj): boolean;
			override;
			procedure translate (dx, dy: extended);
			override;
			procedure rotate (xOrigin, yOrigin, alpha: extended);
			override;
			procedure scale (lambda: extended);
			override;
			procedure reflect (a, b, c: extended);
			override;
			procedure centerOfGravity (var cx, cy: extended);
			override;
			function getBoundingObject: obj;
			override;

(* boolean operations for QuadTrees *)
			function negation: quadTree;
			function intersection (qT2: quadTree; pos1, pos2: boolean): quadTree;
			function union (qT2: quadTree; pos1, pos2: boolean): quadTree;
			function difference (qT2: quadTree): quadTree;

		end; (* quadTree *)


procedure initQuadTree;


implementation

const
	clipLeft = -15000;
	clipBottom = -15000;
	clipRight = 15000;
	clipTop = 15000;

	rectToQuadTreeC = 1;
	rectToQuadTreeDepthC = rectToQuadTreeC + 1;
	circleToQuadTreeC = rectToQuadTreeDepthC + 1;
	circleToQuadTreeDepthC = circleToQuadTreeC + 1;
	vectorToQuadTreeC = circleToQuadTreeDepthC + 1;
	vectorToQuadTreeDepthC = vectorToQuadTreeC + 1;
	quadTreeIntersectionPPC = vectorToQuadTreeDepthC + 1;
	quadTreeIntersectionNPC = quadTreeIntersectionPPC + 1;
	quadTreeIntersectionPNC = quadTreeIntersectionNPC + 1;
	quadTreeIntersectionNNC = quadTreeIntersectionPNC + 1;
	quadTreeUnionPPC = quadTreeIntersectionNNC + 1;
	quadTreeUnionNPC = quadTreeUnionPPC + 1;
	quadTreeUnionPNC = quadTreeUnionNPC + 1;
	quadTreeUnionNNC = quadTreeUnionPNC + 1;
	quadTreeDiffC = quadTreeUnionNNC + 1;
	quadTreeNegationC = quadTreeDiffC + 1;

var
	defaultDepth: extended;

function treeCompare (a, b: quadTreeNode): relationObj;
	var
		t: relationObj;
		dir: quadrantsP;
	begin
		if a = nil then
			if b = nil then
				treeCompare := equalObj
			else
				treeCompare := lessThanObj
		else if b = nil then
			treeCompare := greaterThanObj
		else begin
			t := a.nodeCompare(b);
			dir := ne;
			while (t = equalObj) & (dir <= se) do begin
				t := treeCompare(a.son[dir].succ, b.son[dir].succ);
				dir := dir + 1
			end;
			treeCompare := t
		end
	end;


function quadTree.equal (o: obj): boolean;
	var
		t: quadTree;
	begin
		if (o <> nil) & member(o, quadTree) then begin
			t := quadTree(o);
			equal := boundingRect.equal(t.boundingRect) & (treeCompare(root, t.root) = equalObj);
		end
		else
			equal := false;
	end;

function quadTree.genericLessThan (o: obj): boolean;
	var
		t: quadTree;
	begin
		t := quadTree(o);
		genericLessThan := boundingRect.genericLessThan(t.boundingRect) | boundingRect.equal(t.boundingRect) & (treeCompare(root, t.root) = lessThanObj);
	end;

procedure quadTree.fileOutput;
	begin
		boundingRect.fileOutput;
		writeByte(ord(root = nil));
		if root <> nil then
			root.fileOutputTag;
	end;

procedure quadTree.textFileOutput;
	begin
		boundingRect.textFileOutput;
		writeTextLongInt(ord(root = nil));
		if root <> nil then
			writeTextObj(root);
	end;

function getTreeNode: quadTreeNode;
	var
		n: quadTreeNode;
		b: Byte;
	begin
		readByte(b);
		if b = 0 then begin
			n := quadTreeNode(newObj(fileGetType));
			n.fileInput;
			getTreeNode := n;
		end
		else
			getTreeNode := nil;
	end;

procedure quadTree.fileInput;
	var
		r: rectangle;
	begin
		new(r);
		failNil(Handle(r));
		r.fileInput;
		boundingRect := r;
		root := getTreeNode;
	end;

function textGetTreeNode: quadTreeNode;
	var
		i: longInt;
	begin
		readTextLongInt(i);
		if i = 0 then
			textGetTreeNode := quadTreeNode(readTextObj)
		else
			textGetTreeNode := nil;
	end;

procedure quadTree.textFileInput;
	var
		r: rectangle;
	begin
		new(r);
		failNil(Handle(r));
		r.textFileInput;
		boundingRect := r;
		root := textGetTreeNode;
	end;

procedure quadTreeNode.fileInputNode;
	var
		b: Byte;
		dir: quadrants;
	begin
		for dir := ne to se do begin
			readByte(b);
			son[dir].leafIncluded := b = 1;
		end
	end;

procedure quadTreeNode.fileOutputNode;
	var
		dir: quadrants;
	begin
		for dir := ne to se do
			writeByte(ord(son[dir].leafIncluded))
	end;

procedure quadTreeNode.textFileInputNode;
	var
		i: longInt;
		dir: quadrants;
	begin
		for dir := ne to se do begin
			readTextLongInt(i);
			son[dir].leafIncluded := i = 1;
		end
	end;

procedure quadTreeNode.textFileOutputNode;
	var
		dir: quadrants;
	begin
		for dir := ne to se do
			writeTextLongInt(ord(son[dir].leafIncluded))
	end;

function quadTreeNode.nodeCompare (x: quadTreeNode): relationObj;
	begin
		if (son[ne].leafIncluded = x.son[ne].leafIncluded) & (son[nw].leafIncluded = x.son[nw].leafIncluded) & (son[sw].leafIncluded = x.son[sw].leafIncluded) & (son[se].leafIncluded = x.son[se].leafIncluded) then
			nodeCompare := equalObj
		else
			nodeCompare := unEqualObj
	end;

procedure quadTreeNode.fileOutput;
	var
		dir: quadrants;
	begin
		fileOutputNode;
		for dir := ne to se do begin
			writeByte(ord(son[dir].succ = nil));
			if son[dir].succ <> nil then
				son[dir].succ.fileOutputTag
		end
	end;

procedure quadTreeNode.fileInput;
	var
		dir: quadrants;
	begin
		fileInputNode;
		for dir := ne to se do begin
			son[dir].succ := getTreeNode
		end
	end;

procedure quadTreeNode.textFileOutput;
	var
		dir: quadrants;
	begin
		textFileOutputNode;
		for dir := ne to se do begin
			writeTextLongInt(ord(son[dir].succ = nil));
			if son[dir].succ <> nil then
				writeTextObj(son[dir].succ)
		end
	end;

procedure quadTreeNode.textFileInput;
	var
		dir: quadrants;
	begin
		textFileInputNode;
		for dir := ne to se do
			son[dir].succ := textGetTreeNode
	end;

procedure quadTreeNode.init;
	var
		dir: quadrants;
	begin
		for dir := ne to se do
			with son[dir] do begin
				leafIncluded := false;
				succ := nil
			end
	end;

procedure quadTreeNode.freeNode;
	begin
		shallowFree;
	end;

procedure quadTreeNode.freeNodeAll;
	begin
		shallowFree;
	end;

procedure quadTreeNode.free;
	var
		dir: quadrants;
	begin
		for dir := ne to se do
			if son[dir].succ <> nil then
				son[dir].succ.free;
		freeNode;
	end;

procedure quadTreeNode.freeAll;
	var
		dir: quadrants;
	begin
		for dir := ne to se do
			if son[dir].succ <> nil then
				son[dir].succ.freeAll;
		freeNodeAll;
	end;

function quadTreeNode.duplicateNode: obj;
	begin
		duplicateNode := shallowDuplicate;
	end;

function quadTreeNode.duplicateNodeAll: obj;
	begin
		duplicateNodeAll := shallowDuplicate;
	end;

function quadTreeNode.duplicate: obj;
	var
		r: quadTreeNode;
		dir: quadrants;
	begin
		r := quadTreeNode(duplicateNode);
		for dir := ne to se do
			if son[dir].succ <> nil then
				r.son[dir].succ := quadTreeNode(son[dir].succ.duplicate);
		duplicate := r;
	end;

function quadTreeNode.duplicateAll: obj;
	var
		r: quadTreeNode;
		dir: quadrants;
	begin
		r := quadTreeNode(duplicateNodeAll);
		for dir := ne to se do
			if son[dir].succ <> nil then
				r.son[dir].succ := quadTreeNode(son[dir].succ.duplicateAll);
		duplicateAll := r;
	end;

procedure quadTree.init;
	var
		o: rectangle;
	begin
		root := nil;
		new(o);
		failNil(Handle(o));
		o.init;
		boundingRect := o
	end;

procedure quadTree.free;
	begin
		root.free;
		boundingRect.free;
		shallowFree;
	end;

procedure quadTree.freeAll;
	begin
		root.freeAll;
		boundingRect.free;
		shallowFree;
	end;

function quadTree.duplicate: obj;
	var
		b: quadTree;
	begin
		b := quadTree(shallowDuplicate);
		b.root := quadTreeNode(root.duplicate);
		b.boundingRect := rectangle(boundingRect.duplicate);
		duplicate := b;
	end;

function quadTree.duplicateAll: obj;
	var
		b: quadTree;
	begin
		b := quadTree(shallowDuplicate);
		b.root := quadTreeNode(root.duplicateAll);
		b.boundingRect := rectangle(boundingRect.duplicate);
		duplicateAll := b;
	end;

function quadTreeNode.hitsRect (where: obj; inBottom, inLeft, inTop, inRight: extended): boolean;

	function sonInside (where: obj; thisLeafIncl: boolean; thisSucc: quadTreeNode; bottom, left, top, right: extended): boolean;
		const
			all = 1;
			part = 0;
			no = -1;
		begin (* sonInside *)
			if thisLeafIncl then
				case where.inclRectangle(bottom, left, top, right) of
					all, part: 
						sonInside := true;
					no: 
						sonInside := false;
				end
			else if (thisSucc <> nil) then
				sonInside := thisSucc.hitsRect(where, bottom, left, top, right)
			else
				sonInside := false
		end; (* sonInside *)

		var
			mVertikal, mHorizontal: extended;
	begin(* hitsRect *)
		mVertikal := (inTop + inBottom) / 2;
		mHorizontal := (inLeft + inRight) / 2;
		hitsRect := sonInside(where, son[ne].leafIncluded, son[ne].succ, mVertikal, mHorizontal, inTop, inRight) | sonInside(where, son[nw].leafIncluded, son[nw].succ, mVertikal, inLeft, inTop, mHorizontal) | sonInside(where, son[sw].leafIncluded, son[sw].succ, inBottom, inLeft, mVertikal, mHorizontal) | sonInside(where, son[se].leafIncluded, son[se].succ, inBottom, mHorizontal, mVertikal, inRight)
	end;(* hitsRect *)

function quadTree.isSelected (where: obj): boolean;
	begin
		isSelected := root.hitsRect(where, boundingRect.bottom, boundingRect.left, boundingRect.top, boundingRect.right);
	end;

procedure quadTree.translate (dx, dy: extended);
	begin
		with boundingRect do begin
			left := left + dx;
			right := right + dx;
			bottom := bottom + dy;
			top := top + dy
		end
	end;

{$IFC NOT THINK_PASCAL }
{$PUSH }
{$PRAGMAC warn_unusedarg off }
{$ENDC }
procedure quadTree.rotate (xOrigin, yOrigin, alpha: extended);
	begin
	end;

procedure quadTree.scale (lambda: extended);
	begin
	end;

procedure quadTree.reflect (a, b, c: extended);
	begin
	end;
{$IFC NOT THINK_PASCAL }
{$POP }
{$ENDC }

procedure quadTree.centerOfGravity (var cx, cy: extended);
	begin
		boundingRect.centerOfGravity(cx, cy);
	end;

procedure quadTreeNode.interactiveOutputNode (inBottom, inLeft, inTop, inRight: integer);

	procedure sonOut (thisLeafIncl: boolean; thisSucc: quadTreeNode; bottom, left, top, right: integer);
		begin (* sonOut *)
			if thisLeafIncl then
{$IFC THINK_PASCAL }
				PaintRect(bottom, left, top, right)
{$ELSEC }
				PaintRect2(bottom, left, top, right)
{$ENDC }
			else if thisSucc <> nil then
				thisSucc.interactiveOutputNode(bottom, left, top, right)
		end; (* sonOut *)

		var
			mVertikal, mHorizontal: integer;
	begin
		mVertikal := (inTop + inBottom) div 2;
		mHorizontal := (inLeft + inRight) div 2;
		sonOut(son[ne].leafIncluded, son[ne].succ, mVertikal, mHorizontal, inTop, inRight);
		sonOut(son[nw].leafIncluded, son[nw].succ, mVertikal, inLeft, inTop, mHorizontal);
		sonOut(son[sw].leafIncluded, son[sw].succ, inBottom, inLeft, mVertikal, mHorizontal);
		sonOut(son[se].leafIncluded, son[se].succ, inBottom, mHorizontal, mVertikal, inRight)
	end;

procedure quadTreeNode.interactiveOutputClipNode (y1, x1, y2, x2: integer);

	procedure sonPaint (thisLeafIncl: boolean; thisSucc: quadTreeNode; bottom, left, top, right, bottomS, leftS, topS, rightS: integer);
		begin (* sonPaint *)
			if thisLeafIncl then
{$IFC THINK_PASCAL }
				PaintRect(bottom, left, top, right)
{$ELSEC }
				PaintRect2(bottom, left, top, right)
{$ENDC }
			else if (thisSucc <> nil) then
				thisSucc.interactiveOutputClipNode(bottomS, leftS, topS, rightS)
		end; (* sonPaint *)

		var
			ybottom, xleft, ytop, xright: integer;
			resx1, resx2, resy1, resy2: boolean;
			mVertikal, mHorizontal: integer;
	begin (* interactiveOutputClipNode *)
		if x1 < clippingRectangle.left then begin
			xleft := clippingRectangle.left;
			resx1 := false;
		end
		else
			resx1 := true;
		if x2 > clippingRectangle.right then begin
			xright := clippingRectangle.right;
			resx2 := false;
		end
		else
			resx2 := true;
		if y1 < clippingRectangle.bottom then begin
			ybottom := clippingRectangle.bottom;
			resy1 := false;
		end
		else
			resy1 := true;
		if y2 > clippingRectangle.top then begin
			ytop := clippingRectangle.top;
			resy2 := false;
		end
		else
			resy2 := true;

		if resx1 & resx2 & resy1 & resy2 then begin
			interactiveOutputNode(y1, x1, y2, x2);
		end
		else begin
{$IFC THINK_PASCAL }
			mVertikal := trunc((y1 + y2 + 0.6) / 2);
			mHorizontal := trunc((x1 + x2 + 0.6) / 2);
{$ELSEC }
			mVertikal := truncLong((y1 + y2 + 0.6) / 2);
			mHorizontal := truncLong((x1 + x2 + 0.6) / 2);
{$ENDC }
			sonPaint(son[ne].leafIncluded, son[ne].succ, mVertikal, mHorizontal, ytop, xright, mVertikal, mHorizontal, y2, x2);
			sonPaint(son[nw].leafIncluded, son[nw].succ, mVertikal, xleft, ytop, mHorizontal, mVertikal, x1, y2, mHorizontal);
			sonPaint(son[sw].leafIncluded, son[sw].succ, ybottom, xleft, mVertikal, mHorizontal, y1, x1, mVertikal, mHorizontal);
			sonPaint(son[se].leafIncluded, son[se].succ, ybottom, mHorizontal, mVertikal, xright, y1, mHorizontal, mVertikal, x2)
		end
	end; (* interactiveOutputClipNode *)

procedure quadTree.interactiveOutput; (* uebernimmt quadTree in WeltKoordinaten *)
	var
		old: PenState;
		x1, y1, x2, y2: integer;
	begin
		GetPenState(old);
{$IFC THINK_PASCAL }
		PenPat(black);  		(* ohne Zoom besser black *)
{$ELSEC }
		PenPat(qd.black);  		(* ohne Zoom besser black *)
{$ENDC }
		PenMode(patXor);
		with currentFrame do begin	(* Umwandlung in Windowkoordinaten *)
{$IFC THINK_PASCAL }
			x1 := round(dx + boundingRect.left * mx);
			y1 := round(dy + boundingRect.bottom * my);
			x2 := round(dx + boundingRect.right * mx);
			y2 := round(dy + boundingRect.top * my);
{$ELSEC }
			x1 := roundLong(dx + boundingRect.left * mx);
			y1 := roundLong(dy + boundingRect.bottom * my);
			x2 := roundLong(dx + boundingRect.right * mx);
			y2 := roundLong(dy + boundingRect.top * my);
{$ENDC }
		end;
		ExtendDrawingSheet(x1, y1);
		ExtendDrawingSheet(x2, y2);
		root.interactiveOutputClipNode(y1, x1, y2, x2);
{$IFC THINK_PASCAL }
		PenPat(gray);
{$ELSEC }
		PenPat(qd.gray);
{$ENDC }
		root.interactiveOutputClipNode(y1, x1, y2, x2);
		SetPenState(old);
	end;

procedure quadTree.interactiveHighlight;
	var
		old: PenState;
		x1, y1, x2, y2: integer;
	begin
		GetPenState(old);
{$IFC THINK_PASCAL }
		PenPat(gray);							(* ohne Zoom besser  gray *)
{$ELSEC }
		PenPat(qd.gray);							(* ohne Zoom besser  gray *)
{$ENDC }
		PenMode(patXor);
		with currentFrame do begin		(* Umwandlung in Windowkoordinaten *)
{$IFC THINK_PASCAL }
			x1 := round(dx + boundingRect.left * mx);
			y1 := round(dy + boundingRect.bottom * my);
			x2 := round(dx + boundingRect.right * mx);
			y2 := round(dy + boundingRect.top * my);
{$ELSEC }
			x1 := roundLong(dx + boundingRect.left * mx);
			y1 := roundLong(dy + boundingRect.bottom * my);
			x2 := roundLong(dx + boundingRect.right * mx);
			y2 := roundLong(dy + boundingRect.top * my);
{$ENDC }
		end;
		ExtendDrawingSheet(x1, y1);
		ExtendDrawingSheet(x2, y2);
		root.interactiveOutputClipNode(y1, x1, y2, x2);
		SetPenState(old);
	end;

procedure quadTree.windowOutput;
	var
		old: PenState;
	begin
		GetPenState(old);
{$IFC THINK_PASCAL }
		PenPat(black);								(* ohne Zoom besser black *)
{$ELSEC }
		PenPat(qd.black);								(* ohne Zoom besser black *)
{$ENDC }
		PenMode(patXor);
		root.interactiveOutputClipNode(clipBottom, clipLeft, clipTop, clipRight);
		SetPenState(old);
	end;

function quadTree.memberType (o: obj): boolean;
	begin
		memberType := (o <> nil) & member(o, quadTree);
	end;

function quadTree.getType: objectType;
	begin
		getType := quadTreeType;
	end;

function quadTree.getInfo: obj;
	var
		v: vector;
	begin
		v := newVector(2);
		appendObjInfoObject(v, 'Bounding Rectangle', boundingRect, true);
		appendObjInfoObject(v, 'Root', root, true);
		getInfo := v;
	end;

function quadTreeNode.getType: objectType;
	begin
		getType := quadTreeNodeType;
	end;

function quadTreeNode.memberType (o: obj): boolean;
	begin
		memberType := member(o, quadTreeNode);
	end;

function quadTreeNode.getTextType: Str255;
	begin
		getTextType := 'Quadtree Node';
	end;

function leafName (l: quadrants): Str255;
	begin
		case l of
			ne: 
				leafName := 'NE';
			nw: 
				leafName := 'NW';
			sw: 
				leafName := 'SW';
			se: 
				leafName := 'SE';
		end;
	end;

function quadTreeNode.getInfo: obj;
	var
		v: vector;
		i: quadrants;
	begin
		v := newVector(4 + ord(son[ne].succ <> nil) + ord(son[nw].succ <> nil) + ord(son[sw].succ <> nil) + ord(son[se].succ <> nil) + ord(son[ne].leafIncluded) + ord(son[nw].leafIncluded) + ord(son[sw].leafIncluded) + ord(son[se].leafIncluded));
		for i := ne to se do
			with son[i] do begin
				if leafIncluded then
					appendObjInfoString(v, Concat(leafName(i), '-Included'), 'Yes', true);
				if succ <> nil then
					appendObjInfoObject(v, Concat(leafName(i), '-Node'), succ, true);
			end;
		getInfo := v;
	end;

function quadTree.getTextType: Str255;
	begin
		getTextType := 'Quadtree';
	end;


procedure quadTree.describe;
	begin (* describe *)
		with currentArguments do begin
       (* Erzeugen von Quad-Trees *)
			if (length = 2) then begin
				if currentArgumentsSameType & member(ith(1), rectangle) then begin
					beginSubMenu('Rectangle To QuadTree');
					menuEntry('Maximal Depth', rectToQuadTreeC);
					menuEntry('Chosen Depth...', rectToQuadTreeDepthC);
					endSubMenu
				end
				else begin
					if (member(ith(1), rectangle) & member(ith(2), circle) | member(ith(1), circle) & member(ith(2), rectangle)) then begin
						beginSubMenu('Circle To QuadTree');
						menuEntry('Maximal Depth', circleToQuadTreeC);
						menuEntry('Chosen Depth...', circleToQuadTreeDepthC);
						endSubMenu
					end
					else if member(ith(1), rectangle) & member(ith(2), vector) & vector(ith(2)).memberTypeSet([rectangleType, circleType]) | member(ith(1), vector) & member(ith(2), rectangle) & vector(ith(1)).memberTypeSet([rectangleType, circleType]) then begin
						beginSubMenu('Grouped Objects To QuadTree');
						menuEntry('Maximal Depth', vectorToQuadTreeC);
						menuEntry('Chosen Depth...', vectorToQuadTreeDepthC);
						endSubMenu
					end
				end
			end;

       (* Operationen auf Quad-Trees *)
			if (length = 2) & currentArgumentsSameType & member(ith(1), quadTree) then begin
				beginSubMenu('Intersection');
				menuEntry('Tree1 * Tree2', quadTreeIntersectionPPC);
				menuEntry('NOT Tree1 * Tree2', quadTreeIntersectionNPC);
				menuEntry('Tree1 * NOT Tree2', quadTreeIntersectionPNC);
				menuEntry('NOT Tree1 * NOT Tree2', quadTreeIntersectionNNC);
				endSubMenu;
				beginSubMenu('Union');
				menuEntry('Tree1 + Tree2', quadTreeUnionPPC);
				menuEntry('NOT Tree1 + Tree2', quadTreeUnionNPC);
				menuEntry('Tree1 + NOT Tree2', quadTreeUnionPNC);
				menuEntry('NOT Tree1 + NOT Tree2', quadTreeUnionNNC);
				endSubMenu;
				menuEntry('Difference', quadTreeDiffC);
			end
			else if (length = 1) & member(ith(1), quadTree) then
				menuEntry('Negation', quadTreeNegationC)
		end
	end; (* describe *)


procedure quadTree.execute (item: integer);

	function makeQuadTree (old: obj; boundRect: rectangle; depth: extended): quadTree;
		var
			qT: quadTree;
			actualDepth: integer;
		begin (* makeQuadTree *)
			new(qT);
			failNil(Handle(qT));
			qT.boundingRect := rectangle(boundRect.duplicate);
			new(qT.root);
			failNil(Handle(qT.root));
			qT.root.init;
			if depth = -1 then begin
				displayInfoString('Yeah');
				depth := ln(max(boundRect.top - boundRect.bottom, boundRect.right - boundRect.left)) / ln(2);
			end;
			actualDepth := 0;
{$IFC THINK_PASCAL }
			qT.root.insert(old, boundRect.bottom, boundRect.left, boundRect.top, boundRect.right, actualDepth, trunc(depth) + 1);
{$ELSEC }
			qT.root.insert(old, boundRect.bottom, boundRect.left, boundRect.top, boundRect.right, actualDepth, truncLong(depth) + 1);
{$ENDC }
			makeQuadTree := qT;
		end; (* makeQuadTree *)

	procedure vectorToQuadTree (v: vector; r: rectangle; depth: extended);
		var
			quadTreeVector: homogenVector;
			i: integer;
		begin (*  vectorToQuadTree *)
			quadTreeVector := newHomogenVector(v.length);
			for i := 1 to v.length do
				if member(v.elements^^[i], rectangle) | member(v.elements^^[i], circle) then
					quadTreeVector.appendNonNil(makeQuadTree(v.elements^^[i], r, depth));
			currentResult.appendNonNil(quadTreeVector)
		end; (*  vectorToQuadTree *)

		const
			maxDepth = -1; (* '-1' for flag 'maximal resolution requested' *)
			depthTitle = 'Depth of recursion';
	begin (* quadTree.execute *)
		with currentArguments do
			case item of
				rectToQuadTreeC, circleToQuadTreeC: 
					if member(ith(1), rectangle) then
						currentResult.appendNonNil(makeQuadTree(ith(2), rectangle(ith(1)), maxDepth))
					else
						currentResult.appendNonNil(makeQuadTree(ith(1), rectangle(ith(2)), maxDepth));
				rectToQuadTreeDepthC, circleToQuadTreeDepthC: 
					if getUserParameter1(depthTitle, defaultDepth, defaultDepth, @isNatural) then
						if member(ith(1), rectangle) then
							currentResult.appendNonNil(makeQuadTree(ith(2), rectangle(ith(1)), defaultDepth))
						else
							currentResult.appendNonNil(makeQuadTree(ith(1), rectangle(ith(2)), defaultDepth));
				vectorToQuadTreeC: 
					if member(ith(1), rectangle) then
						vectorToQuadTree(vector(ith(2)), rectangle(ith(1)), maxDepth)
					else
						vectorToQuadTree(vector(ith(1)), rectangle(ith(2)), maxDepth);
				vectorToQuadTreeDepthC: 
					if getUserParameter1(depthTitle, defaultDepth, defaultDepth, @isNatural) then
						if member(ith(1), rectangle) then
							vectorToQuadTree(vector(ith(2)), rectangle(ith(1)), defaultDepth)
						else
							vectorToQuadTree(vector(ith(1)), rectangle(ith(2)), defaultDepth);
				quadTreeIntersectionPPC, quadTreeIntersectionNPC, quadTreeIntersectionPNC, quadTreeIntersectionNNC: 
					currentResult.appendNonNil(quadTree(ith(1)).intersection(quadTree(ith(2)), (item = quadTreeIntersectionPPC) | (item = quadTreeIntersectionPNC), (item = quadTreeIntersectionPPC) | (item = quadTreeIntersectionNPC)));
				quadTreeUnionPPC, quadTreeUnionNPC, quadTreeUnionPNC, quadTreeUnionNNC: 
					currentResult.appendNonNil(quadTree(ith(1)).union(quadTree(ith(2)), (item = quadTreeUnionPPC) | (item = quadTreeUnionPNC), (item = quadTreeUnionPPC) | (item = quadTreeUnionNPC)));
				quadTreeDiffC: 
					currentResult.appendNonNil(quadTree(ith(1)).difference(quadTree(ith(2))));
				quadTreeNegationC: 
					currentResult.appendNonNil(quadTree(ith(1)).negation);
			end;
	end; (* quadTree.execute *)


procedure quadTreeNode.insert (old: obj; inBottom, inLeft, inTop, inRight: extended; actualDepth, depth: integer);

	procedure sonInsert (old: obj; var thisLeafIncl: boolean; var thisSucc: quadTreeNode; bottom, left, top, right: extended; actualDepth, depth: integer);
		const
			all = 1;
			part = 0;
			no = -1;
		begin (* sonInsert *)
			actualDepth := actualDepth + 1;
			case old.inclRectangle(bottom, left, top, right) of
				all: 
					thisLeafIncl := true;
				part: 
					if actualDepth <= depth then begin
						new(thisSucc);
						failNil(Handle(thisSucc));
						thisSucc.init;
						thisSucc.insert(old, bottom, left, top, right, actualDepth, depth);
					end
					else
						thisLeafIncl := true;
				no: 
					;
			end
		end; (* sonInsert *)

		var
			mVertikal, mHorizontal: extended;
	begin (* quadTreeNode.insert  *)
		mVertikal := (inTop + inBottom) / 2;
		mHorizontal := (inLeft + inRight) / 2;
		sonInsert(old, son[ne].leafIncluded, son[ne].succ, mVertikal, mHorizontal, inTop, inRight, actualDepth, depth);
		sonInsert(old, son[nw].leafIncluded, son[nw].succ, mVertikal, inLeft, inTop, mHorizontal, actualDepth, depth);
		sonInsert(old, son[sw].leafIncluded, son[sw].succ, inBottom, inLeft, mVertikal, mHorizontal, actualDepth, depth);
		sonInsert(old, son[se].leafIncluded, son[se].succ, inBottom, mHorizontal, mVertikal, inRight, actualDepth, depth)
	end; (* quadTreeNode.insert  *)


function quadTree.negation: quadTree;
	var
		qTResult: quadTree;
	begin (* quadTree.negation *)
		qTResult := quadTree(duplicateAll);
		qTResult.root.negation;
		negation := qTResult
	end; (* quadTree.negation *)

procedure quadTreeNode.negation;
(* tree is negated starting with "quadTreeNode" *)
	var
		dir: quadrants;
	begin (* quadTreeNode.negation *)
		for dir := ne to se do
			if son[dir].succ = nil then
				son[dir].leafIncluded := not son[dir].leafIncluded
			else
				son[dir].succ.negation
	end; (* quadTreeNode.negation *)


function quadTree.intersection (qT2: quadTree; pos1, pos2: boolean): quadTree;
	var
		qTResult: quadTree;
	begin (* quadTree.intersection *)
		new(qTResult);
		failNil(Handle(qTResult));
		qTResult.boundingRect := rectangle(boundingRect.duplicate);
		qTResult.root := root.intersection(qT2.root, pos1, pos2);
		if qTResult.root = nil then begin
			qTResult.free;
			intersection := nil
		end
		else
			intersection := qTResult
	end; (* quadTree.intersection *)


function quadTreeNode.intersection (qTN2: quadTreeNode; pos1, pos2: boolean): quadTreeNode;
(* returns intersection set of tree starting with "quadTreeNode" and tree "qTN2"; if pos1=true, then the first *)
(* trees complement is taken as first operand (but tree does not change!); pos2:analog for second tree               *)

	procedure sonIntersection (leafIncl1, leafIncl2: boolean; var thisLeafIncl: boolean; succ1, succ2: quadTreeNode; var thisSucc: quadTreeNode; var nrOfFalses: integer; pos1, pos2: boolean);
		begin (* sonIntersection *)
			if not pos1 & (succ1 = nil) then
				leafIncl1 := not leafIncl1; (* only local change (no var-parameter) *)
			if not pos2 & (succ2 = nil) then
				leafIncl2 := not leafIncl2;

       (* "thisSucc" is already "=nil" grace  "qTN.init" *)
			if succ1 = nil then
				if leafIncl1 then
					if succ2 = nil then
						if leafIncl2 then
							thisLeafIncl := true
						else begin
							thisLeafIncl := false;
							nrOfFalses := nrOfFalses + 1
						end
           (* succ2<>nil *)
					else begin
						thisSucc := quadTreeNode(succ2.duplicateAll);
						if not pos2 then
							thisSucc.negation
					end
         (* not leafIncl1 *)
				else begin
					thisLeafIncl := false;
					nrOfFalses := nrOfFalses + 1
				end
       (* succ1<>nil *)
			else if succ2 = nil then
				if leafIncl2 then begin
					thisSucc := quadTreeNode(succ1.duplicateAll);
					if not pos1 then
						thisSucc.negation
				end
				else begin
					thisLeafIncl := false;
					nrOfFalses := nrOfFalses + 1
				end
       (* succ2<>nil *)
			else begin
				thisSucc := succ1.intersection(succ2, pos1, pos2);
				if thisSucc = nil then begin (* intersection set from below returned "nil": only possible with four times "false" in 'leafIncluded's *)
					thisLeafIncl := false;
					nrOfFalses := nrOfFalses + 1
				end
			end
		end; (* sonIntersection *)

		var
			qTN: quadTreeNode;
			nrOfFalses: integer;
			dir: quadrants;
	begin (* quadTreeNode.intersection *)
		new(qTN);
		failNil(Handle(qTN));
		qTN.init;
		nrOfFalses := 0;
		for dir := ne to se do
			sonIntersection(son[dir].leafIncluded, qTN2.son[dir].leafIncluded, qTN.son[dir].leafIncluded, son[dir].succ, qTN2.son[dir].succ, qTN.son[dir].succ, nrOfFalses, pos1, pos2);

    (* all 4 sons "colourless"? *)
		if nrOfFalses = 4 then begin
			qTN.free;
			intersection := nil
		end
		else
			intersection := qTN
	end; (* quadTreeNode.intersection *)


function quadTree.union (qT2: quadTree; pos1, pos2: boolean): quadTree;
	var
		qTResult: quadTree;
		dir: quadrants;
		firstHalf, secondHalf: boolean; (* help constructs because the if-condition would be too long *)
	begin (* quadTree.union *)
		new(qTResult);
		failNil(Handle(qTResult));
		qTResult.boundingRect := rectangle(boundingRect.duplicate);
		qTResult.root := root.union(qT2.root, pos1, pos2);
		if (qTResult.root = nil) then begin (* means all or nothing? *)
			if pos1 then
				firstHalf := (root.son[ne].succ <> nil) | root.son[ne].leafIncluded | (root.son[nw].succ <> nil) | root.son[nw].leafIncluded | (root.son[sw].succ <> nil) | root.son[sw].leafIncluded | (root.son[se].succ <> nil) | root.son[se].leafIncluded
			else
				firstHalf := (root.son[ne].succ <> nil) | not root.son[ne].leafIncluded | (root.son[nw].succ <> nil) | not root.son[nw].leafIncluded | (root.son[sw].succ <> nil) | not root.son[sw].leafIncluded | (root.son[se].succ <> nil) | not root.son[se].leafIncluded;
			if pos2 then
				secondHalf := (qT2.root.son[ne].succ <> nil) | qT2.root.son[ne].leafIncluded | (qT2.root.son[nw].succ <> nil) | qT2.root.son[nw].leafIncluded | (qT2.root.son[sw].succ <> nil) | qT2.root.son[sw].leafIncluded | (qT2.root.son[se].succ <> nil) | qT2.root.son[se].leafIncluded
			else
				secondHalf := (qT2.root.son[ne].succ <> nil) | not qT2.root.son[ne].leafIncluded | (qT2.root.son[nw].succ <> nil) | not qT2.root.son[nw].leafIncluded | (qT2.root.son[sw].succ <> nil) | not qT2.root.son[sw].leafIncluded | (qT2.root.son[se].succ <> nil) | not qT2.root.son[se].leafIncluded;
			if firstHalf | secondHalf then begin
       (* "quadTreeNode" returned "nil" because on highest level there are only full rectangles *)
				new(qTResult.root);
				failNil(Handle(qTResult.root));
				qTResult.root.init;
				for dir := ne to se do
					qTResult.root.son[dir].leafIncluded := true;
				union := qTResult
			end
			else begin
				qTResult.free;
				union := nil
			end
		end
		else
			union := qTResult
	end; (* quadTree.union *)


function quadTreeNode.union (qTN2: quadTreeNode; pos1, pos2: boolean): quadTreeNode;
(* returns union set of tree starting with "quadTreeNode" and tree "qTN2"; if pos1=true, then the first *)
(* trees complement is taken as first operand (but tree does not change!); pos2:analog for second tree    *)

	procedure sonUnion (leafIncl1, leafIncl2: boolean; var thisLeafIncl: boolean; succ1, succ2: quadTreeNode; var thisSucc: quadTreeNode; var nrOfTrues: integer; pos1, pos2: boolean);
		begin (* sonUnion *)
			if not pos1 & (succ1 = nil) then
				leafIncl1 := not leafIncl1; (* only local change (no var-parameter) *)
			if not pos2 & (succ2 = nil) then
				leafIncl2 := not leafIncl2;

       (* "thisSucc" is already "=nil" grace  "qTN.init" *)
			if succ1 = nil then
				if leafIncl1 then begin
					thisLeafIncl := true;
					nrOfTrues := nrOfTrues + 1
				end
         (* not leafIncl1 *)
				else if succ2 = nil then
					if leafIncl2 then begin
						thisLeafIncl := true;
						nrOfTrues := nrOfTrues + 1
					end
					else
						thisLeafIncl := false
         (* succ2<>nil *)
				else begin
					thisSucc := quadTreeNode(succ2.duplicateAll);
					if not pos2 then
						thisSucc.negation
				end
       (* succ1<>nil *)
			else if succ2 = nil then
				if leafIncl2 then begin
					thisLeafIncl := true;
					nrOfTrues := nrOfTrues + 1
				end
				else begin
					thisSucc := quadTreeNode(succ1.duplicateAll);
					if not pos1 then
						thisSucc.negation
				end
       (* succ2<>nil *)
			else begin
				thisSucc := succ1.union(succ2, pos1, pos2);
				if thisSucc = nil then begin (* union set from below returned "nil": only possible with four times "true" in 'leafIncluded's *)
					thisLeafIncl := true;
					nrOfTrues := nrOfTrues + 1
				end
			end;
		end; (* sonUnion *)

		var
			qTN: quadTreeNode;
			nrOfTrues: integer;
			dir: quadrants;
	begin (* quadTreeNode.cut *)
		new(qTN);
		failNil(Handle(qTN));
		qTN.init;
		nrOfTrues := 0;
		for dir := ne to se do
			sonUnion(son[dir].leafIncluded, qTN2.son[dir].leafIncluded, qTN.son[dir].leafIncluded, son[dir].succ, qTN2.son[dir].succ, qTN.son[dir].succ, nrOfTrues, pos1, pos2);

    (* all 4 sons "colourless"? *)
		if nrOfTrues = 4 then begin
			qTN.free;
			union := nil
		end
		else
			union := qTN
	end; (* quadTreeNode.union *)


function quadTree.difference (qT2: quadTree): quadTree;
	begin (* quadTree.difference *)
		difference := intersection(qT2, true, false);
	end; (* quadTree.difference *)

procedure initQuadTree;
	var
		t: quadTree;
		n: quadTreeNode;
	begin
		new(t);
		failNil(Handle(t));
		prototypes[quadTreeType] := t;
		new(n);
		failNil(Handle(n));
		prototypes[quadTreeNodeType] := n;
		defaultDepth := 5;
	end;

end.
