(*	XYZ GeoBench (eXperimental geometrY Zurich):		*)
(*	An object oriented, interactive workbench for		*)
(*	experimental geometric computation.					*)
(*														*)
(*	 1999 by Peter Schorn, peter.schorn@acm.org		*)
(*	Culmannstrasse 77, CH-8006 Zrich, Switzerland		*)
(*	This software may be copied freely for educational	*)
(*	and research purposes only. For commercial use,		*)
(*	please contact the author.							*)
(*														*)
(*	Version 5.0.5: 30-Sep-1999							*)

unit avlTree;

interface

uses
	dictionary;

procedure initFreeListAVL;

procedure extendFreeListAVL (number: longInt);

procedure shrinkFreeListAVL (number: longInt);

function freeListLengthAVL: longInt;

function freeListSizeAVL: longInt; (* number of bytes in the 'freeList' *)

function newElementAVL: avlNodeH;

procedure disposeElementAVL (e: avlNodeH);

implementation

const
	maxDepth = 50; (* maximal depth of any avlTree *)

function avlTree.getObject (where: reference): obj;
	begin
		assert(where <> nil, 'where  nil in avlTree.getObject');
		getObject := avlNodeH(where)^^.value;
	end;

procedure avlTree.setObject (where: reference; newValue: obj);
	begin
		avlNodeH(where)^^.value := newValue;
	end;

procedure traverse (p: avlNodeH);
	begin
		if p <> nil then begin
			traverse(p^^.sons[left]);
			traverse(p^^.sons[right]);
			disposeElementAVL(p);
		end;
	end;

procedure avlTree.reset;
	begin
		traverse(root);
		init;
	end;

procedure avlTree.init;
	begin
		root := nil;
		inherited init;
	end;

procedure avlTree.free;
	begin
		traverse(root);
		shallowFree;
	end;

procedure avlTree.setTreeSize;
	var
		dummy: longInt;

	function traverse (p: avlNodeH): longInt;
		begin
			if p = nil then
				traverse := 0
			else begin
				p^^.treeSize := traverse(p^^.sons[left]) + traverse(p^^.sons[right]) + 1;
				traverse := p^^.treeSize;
			end;
		end;

	begin
		dummy := traverse(root);
	end;

function avlTree.testTreeSize: longInt;
	var
		dummy, violations: longInt;

	function traverse (p: avlNodeH): longInt;
		var
			t: longInt;
		begin
			if p = nil then
				traverse := 0
			else begin
				t := traverse(p^^.sons[left]) + traverse(p^^.sons[right]) + 1;
				if t <> p^^.treeSize then
					violations := violations + 1;
				traverse := t;
			end;
		end;

	begin
		violations := 0;
		dummy := traverse(root);
		testTreeSize := violations;
	end;

procedure avlTree.sequenceToDictionary (s: vector; l: seqIndex1; r: seqIndex0);
	var
		depth: integer;
		tempRoot: avlNodeH;

	procedure convert (s: vector; l: seqIndex1; r: seqIndex0; var depth: integer; var rr: avlNodeH);
		var
			m, leftDepth, rightDepth: integer;
		begin
			if l > r then begin
				depth := 0;
				rr := nil;
			end
			else begin
				m := (l + r) div 2;
				rr := newElementAVL;
				rr^^.value := s.elements^^[m];
				convert(s, l, m - 1, leftDepth, rr^^.sons[left]);
				if rr^^.sons[left] <> nil then
					rr^^.sons[left]^^.father := rr;
				convert(s, m + 1, r, rightDepth, rr^^.sons[right]);
				if rr^^.sons[right] <> nil then
					rr^^.sons[right]^^.father := rr;
				if leftDepth > rightDepth then begin
					rr^^.balance := leftGreaterRight;
					depth := leftDepth + 1;
				end
				else if leftDepth < rightDepth then begin
					rr^^.balance := leftLessRight;
					depth := rightDepth + 1;
				end
				else begin
					rr^^.balance := leftEqualRight;
					depth := leftDepth + 1;
				end;
			end
		end;

	begin
		init;
		convert(s, l, r, depth, tempRoot);
		tempRoot^^.father := nil;
		root := tempRoot;
		numberOfElements := r - l + 1;
		setTreeSize;
	end;

procedure avlTree.textOutput;

	procedure traverse (p: avlNodeH);
		begin
			if p <> nil then begin
				traverse(p^^.sons[left]);
				p^^.value.textOutput;
				writeLn;
				traverse(p^^.sons[right]);
			end;
		end;

	begin
		traverse(root);
	end;

procedure avlTree.interactiveOutput;

	procedure traverse (p: avlNodeH; level: integer);
		var
			i: directionType;
			l, r: integer;

		function height (p: avlNodeH): integer;
			begin
				if p = nil then
					height := 0
				else
{$IFC THINK_PASCAL }
					height := trunc(max(height(p^^.sons[left]), height(p^^.sons[right]))) + 1;
{$ELSEC }
					height := truncLong(max(height(p^^.sons[left]), height(p^^.sons[right]))) + 1;
{$ENDC }
			end;

		begin
			if p <> nil then begin
				traverse(p^^.sons[left], level + 1);
				l := height(p^^.sons[left]);
				r := height(p^^.sons[right]);
				if r - l <> ord(p^^.balance) - 1 then
					write('???');
				p^^.value.interactiveOutput;
				write(':', level : 1, ' ', ord(p^^.balance) - 1 : 2, l : 3, '#', r : 1, ' ', p^^.treeSize : 1);
				for i := left to right do begin
{$IFC THINK_PASCAL }
					write(' ', i, '=');
{$ELSEC }
					write(' ', ord(i), '=');
{$ENDC }
					if p^^.sons[i] = nil then
						write('NIL')
					else
						p^^.sons[i]^^.value.interactiveOutput;
				end;
				write(' father=');
				if p^^.father = nil then
					write('NIL')
				else
					p^^.father^^.value.interactiveOutput;
				writeLn;
				traverse(p^^.sons[right], level + 1);
			end;
		end;

	begin
		traverse(root, 0);
	end;

procedure updateTreeSize (p: avlNodeH; increment: longInt);
	begin
		while p <> nil do begin
			p^^.treeSize := p^^.treeSize + increment;
			p := p^^.father;
		end;
	end;

function getTreeSize (p: avlNodeH): longInt;
	begin
		if p = nil then
			getTreeSize := 0
		else
			getTreeSize := p^^.treeSize;
	end;

function avlTree.insert (x: obj; where: reference; direction: directionType): reference;
	var
		otherDirection: directionType;
		tempRoot: avlNodeH;

	function insertX (x: obj; f: avlNodeH; var son: avlNodeH): avlNodeH;
		var
			h, leftTurn, newLeftTurn: boolean;
			newF: avlNodeH;

		procedure balance (leftGreaterRight, leftLessRight: balanceType; left, right: directionType; var p: avlNodeH; var h: boolean);
			var
				p1, p2: avlNodeH;
				t: longInt;
			begin
				if h then
					if p^^.balance = leftLessRight then begin
						p^^.balance := leftEqualRight;
						h := false;
					end
					else if p^^.balance = leftGreaterRight then begin
						p1 := p^^.sons[left];
						if p1^^.balance = leftGreaterRight then
							with p1^^ do begin
								father := p^^.father;
								p^^.father := p1;
								if sons[right] <> nil then
									sons[right]^^.father := p;
								p^^.sons[left] := sons[right];
								sons[right] := p;
								p^^.balance := leftEqualRight;
								treeSize := p^^.treeSize;
								p^^.treeSize := p^^.treeSize - getTreeSize(p1^^.sons[left]) - 1;
								p := p1;
							end
						else begin
							p2 := p1^^.sons[right];
							with p2^^ do begin
								father := p^^.father;
								p1^^.father := p2;
								p^^.father := p2;
								if sons[left] <> nil then
									sons[left]^^.father := p1;
								if sons[right] <> nil then
									sons[right]^^.father := p;
								p1^^.sons[right] := sons[left];
								sons[left] := p1;
								p^^.sons[left] := sons[right];
								sons[right] := p;
								if balance = leftGreaterRight then
									p^^.balance := leftLessRight
								else
									p^^.balance := leftEqualRight;
								if balance = leftLessRight then
									p1^^.balance := leftGreaterRight
								else
									p1^^.balance := leftEqualRight;
								t := getTreeSize(p^^.sons[left]) + 1;
								treeSize := p^^.treeSize;
							end;
							p1^^.treeSize := p1^^.treeSize - t;
							p^^.treeSize := getTreeSize(p^^.sons[right]) + t;
							p := p2;
						end;
						p^^.balance := leftEqualRight;
						h := false;
					end
					else
						p^^.balance := leftGreaterRight;
			end;

		begin
			son := newElementAVL;
			with son^^ do begin
				value := x;
				balance := leftEqualRight;
				treeSize := 1;
				sons[left] := nil;
				sons[right] := nil;
				father := f;
			end;
			updateTreeSize(f, +1);
			insertX := son;
			leftTurn := (f <> nil) & (f^^.sons[left] = son);
			h := true;
			while (f <> nil) & h do begin
				newF := f^^.father;
				newLeftTurn := (newF <> nil) & (newF^^.sons[left] = f);
				if leftTurn then
					if newF = nil then
						balance(leftGreaterRight, leftLessRight, left, right, tempRoot, h)
					else if newF^^.sons[left] = f then
						balance(leftGreaterRight, leftLessRight, left, right, newF^^.sons[left], h)
					else
						balance(leftGreaterRight, leftLessRight, left, right, newF^^.sons[right], h)
				else if newF = nil then
					balance(leftLessRight, leftGreaterRight, right, left, tempRoot, h)
				else if newF^^.sons[left] = f then
					balance(leftLessRight, leftGreaterRight, right, left, newF^^.sons[left], h)
				else
					balance(leftLessRight, leftGreaterRight, right, left, newF^^.sons[right], h);
				f := newF;
				leftTurn := newLeftTurn;
			end;
		end;

	begin
		tempRoot := root;
		if where = nil then
			insert := reference(insertX(x, nil, tempRoot))
		else if avlNodeH(where)^^.sons[direction] = nil then
			insert := reference(insertX(x, avlNodeH(where), avlNodeH(where)^^.sons[direction]))
		else begin
			where := reference(avlNodeH(where)^^.sons[direction]);
			otherDirection := switchDirection(direction);
			while avlNodeH(where)^^.sons[otherDirection] <> nil do
				where := reference(avlNodeH(where)^^.sons[otherDirection]);
			insert := reference(insertX(x, avlNodeH(where), avlNodeH(where)^^.sons[otherDirection]));
		end;
		root := tempRoot;
		numberOfElements := numberOfElements + 1;
	end;

procedure avlTree.delete (x: reference);
	var
		tempRoot: avlNodeH;

	procedure balance (leftGreaterRight, leftLessRight: balanceType; left, right: directionType; var p: avlNodeH; var h: boolean);
		var
			p1, p2: avlNodeH;
			t: longInt;
		begin
			if p^^.balance = leftLessRight then
				p^^.balance := leftEqualRight
			else if p^^.balance = leftGreaterRight then begin
				p1 := p^^.sons[left];
				if p1^^.balance <> leftLessRight then
					with p1^^ do begin
						father := p^^.father;
						p^^.father := p1;
						if sons[right] <> nil then
							sons[right]^^.father := p;
						p^^.sons[left] := p1^^.sons[right];
						sons[right] := p;
						if balance = leftEqualRight then begin
							p^^.balance := leftGreaterRight;
							balance := leftLessRight;
							h := false;
						end
						else begin
							p^^.balance := leftEqualRight;
							balance := leftEqualRight;
						end;
						treeSize := p^^.treeSize;
						p^^.treeSize := p^^.treeSize - getTreeSize(p1^^.sons[left]) - 1;
						p := p1;
					end
				else begin
					p2 := p1^^.sons[right];
					with p2^^ do begin
						father := p^^.father;
						p1^^.father := p2;
						p^^.father := p2;
						if sons[left] <> nil then
							sons[left]^^.father := p1;
						if sons[right] <> nil then
							sons[right]^^.father := p;
						p1^^.sons[right] := sons[left];
						sons[left] := p1;
						p^^.sons[left] := sons[right];
						sons[right] := p;
						if balance = leftGreaterRight then
							p^^.balance := leftLessRight
						else
							p^^.balance := leftEqualRight;
						if balance = leftLessRight then
							p1^^.balance := leftGreaterRight
						else
							p1^^.balance := leftEqualRight;
						t := getTreeSize(p^^.sons[left]) + 1;
						treeSize := p^^.treeSize;
						p1^^.treeSize := p1^^.treeSize - t;
						p^^.treeSize := getTreeSize(p^^.sons[right]) + t;
						p := p2;
						balance := leftEqualRight;
					end;
				end;
			end
			else begin
				p^^.balance := leftGreaterRight;
				h := false;
			end;
		end;

	procedure delet (var p: avlNodeH);
		var
			q, f, t, newF, r: avlNodeH;
			h, newLeftTurn, leftTurn: boolean;
		begin
			h := true;
			q := p;
			if (q^^.sons[left] <> nil) & (q^^.sons[right] <> nil) then begin
				r := q^^.sons[left];
				leftTurn := r^^.sons[right] = nil;
				while r^^.sons[right] <> nil do
					r := r^^.sons[right];
				with r^^ do begin
					t := sons[left];
					f := father;
					updateTreeSize(r, -1);
					balance := q^^.balance;
					sons := q^^.sons;
					father := q^^.father;
					treeSize := q^^.treeSize;
					if sons[right] <> nil then
						sons[right]^^.father := r;
					p := r;
					if leftTurn then begin
						f := r;
						sons[left] := t
					end
					else begin
						if sons[left] <> nil then
							sons[left]^^.father := r;
						if t <> nil then
							t^^.father := f;
						f^^.sons[right] := t;
					end;
				end;
			end
			else begin
				updateTreeSize(q, -1);
				f := q^^.father;
				if f <> nil then
					leftTurn := f^^.sons[left] = q;
				if q^^.sons[left] = nil then
					p := q^^.sons[right]
				else
					p := q^^.sons[left];
				if p <> nil then
					p^^.father := f;
			end;
			while (f <> nil) & h do begin
				newF := f^^.father;
				if newF <> nil then
					newLeftTurn := newF^^.sons[left] = f;
				if leftTurn then
					if newF = nil then
						balance(leftLessRight, leftGreaterRight, right, left, tempRoot, h)
					else if newF^^.sons[left] = f then
						balance(leftLessRight, leftGreaterRight, right, left, newF^^.sons[left], h)
					else
						balance(leftLessRight, leftGreaterRight, right, left, newF^^.sons[right], h)
				else if newF = nil then
					balance(leftGreaterRight, leftLessRight, left, right, tempRoot, h)
				else if newF^^.sons[left] = f then
					balance(leftGreaterRight, leftLessRight, left, right, newF^^.sons[left], h)
				else
					balance(leftGreaterRight, leftLessRight, left, right, newF^^.sons[right], h);
				f := newF;
				leftTurn := newLeftTurn;
			end;
			disposeElementAVL(q);
		end;

	begin
		if x <> nil then begin
			tempRoot := root;
			if avlNodeH(x)^^.father = nil then
				delet(tempRoot)
			else if avlNodeH(x)^^.father^^.sons[left] = avlNodeH(x) then
				delet(avlNodeH(x)^^.father^^.sons[left])
			else
				delet(avlNodeH(x)^^.father^^.sons[right]);
			root := tempRoot;
			numberOfElements := numberOfElements - 1;
		end;
	end;

function avlTree.find (x: obj; function compare (a, b: obj): relationObj; var where: reference; var direction: directionType): boolean;
	var
		p, f: avlNodeH;
		continue: boolean;
	begin
		p := root;
		f := nil;
		continue := true;
		while (p <> nil) & continue do
			case compare(x, p^^.value) of
				lessThanObj:  begin
					direction := left;
					f := p;
					p := p^^.sons[left];
				end;
				equalObj: 
					continue := false;
				greaterThanObj:  begin
					direction := right;
					f := p;
					p := p^^.sons[right];
				end;
			end;
		if p = nil then
			where := reference(f)
		else
			where := reference(p);
		find := not continue;
	end;

function avlTree.findPosition (x: obj; function compare (a, b: obj): relationObj; var where: reference; var direction: directionType; var position: longInt): boolean;
	var
		p, f: avlNodeH;
		continue: boolean;
	begin
		p := root;
		f := nil;
		continue := true;
		position := 0;
		while (p <> nil) & continue do
			case compare(x, p^^.value) of
				lessThanObj:  begin
					direction := left;
					f := p;
					p := p^^.sons[left];
				end;
				equalObj:  begin
					position := position + getTreeSize(p^^.sons[left]) + 1;
					continue := false;
				end;
				greaterThanObj:  begin
					position := position + getTreeSize(p^^.sons[left]) + 1;
					direction := right;
					f := p;
					p := p^^.sons[right];
				end;
			end;
		if p = nil then
			where := reference(f)
		else
			where := reference(p);
		findPosition := not continue;
	end;

function avlTree.getPosition (x: reference): longInt;
	var
		p, pp: avlNodeH;
		pos: longInt;
	begin
		p := avlNodeH(x);
		pos := getTreeSize(p^^.sons[left]) + 1;
		while p^^.father <> nil do begin
			pp := p^^.father;
			if pp^^.sons[right] = p then
				pos := pos + getTreeSize(pp^^.sons[left]) + 1;
			p := pp;
		end;
		getPosition := pos;
	end;

function avlTree.ith (i: longInt): reference;
	var
		p, l: longInt;
		t: avlNodeH;
	begin
		t := root;
		p := getTreeSize(t);
		assert((1 <= i) & (i <= p), 'Range ok');
		l := getTreeSize(t^^.sons[left]) + 1;
		while l <> i do begin
			if i < l then
				t := t^^.sons[left]
			else begin
				t := t^^.sons[right];
				i := i - l;
			end;
			l := getTreeSize(t^^.sons[left]) + 1;
		end;
		ith := reference(t);
	end;

function avlTree.next (x: reference; direction: directionType): reference;
	var
		otherDirection: directionType;
		directionTurn: boolean;
	begin
		if x = nil then
			next := nil
		else if avlNodeH(x)^^.sons[direction] = nil then begin
			directionTurn := true;
			while (avlNodeH(x)^^.father <> nil) & directionTurn do begin
				directionTurn := avlNodeH(x)^^.father^^.sons[direction] = avlNodeH(x);
				x := reference(avlNodeH(x)^^.father);
			end;
			if directionTurn then
				next := nil
			else
				next := x;
		end
		else begin
			x := reference(avlNodeH(x)^^.sons[direction]);
			otherDirection := switchDirection(direction);
			while avlNodeH(x)^^.sons[otherDirection] <> nil do
				x := reference(avlNodeH(x)^^.sons[otherDirection]);
			next := x;
		end;
	end;

procedure avlTree.swap (p, q: reference);
	procedure setFather (p: avlNodeH);
		var
			i: directionType;
		begin
			with p^^ do
				for i := left to right do
					if sons[i] <> nil then
						sons[i]^^.father := p;
		end;

	procedure update (f, newValue: avlNodeH);
		begin
			if f^^.father = nil then
				root := newValue
			else
				with f^^.father^^ do
					if sons[left] = f then
						sons[left] := newValue
					else
						sons[right] := newValue;
		end;

	procedure exchangeDisjoint (p, q: avlNodeH);
		var
			e: avlNode;
			t: avlNodeH;
		begin
			if p^^.father = q^^.father then
				with p^^.father^^ do begin
					t := sons[left];
					sons[left] := sons[right];
					sons[right] := t;
				end
			else begin
				update(p, q);
				update(q, p);
			end;
			e := p^^;
			with q^^ do begin
				p^^.father := father;
				p^^.balance := balance;
				p^^.treeSize := treeSize;
				p^^.sons := sons;
				setFather(p);
				father := e.father;
				balance := e.balance;
				treeSize := e.treeSize;
				sons := e.sons;
			end;
			setFather(q);
		end;

	procedure exchangeNonDisjoint (newFather, son: avlNodeH; direction, otherDirection: directionType);
		var
			e: avlNode;
		begin
			update(newFather, son);
			e := newFather^^;
			with son^^ do begin
				newFather^^.balance := balance;
				newFather^^.treeSize := treeSize;
				newFather^^.sons := sons;
				setFather(newFather);
				father := e.father;
				balance := e.balance;
				treeSize := e.treeSize;
				sons[direction] := newFather;
				sons[otherDirection] := e.sons[otherDirection];
			end;
			setFather(son);
		end;

	begin
		if avlNodeH(p) = avlNodeH(q)^^.sons[left] then
			exchangeNonDisjoint(avlNodeH(q), avlNodeH(p), left, right)
		else if avlNodeH(p) = avlNodeH(q)^^.sons[right] then
			exchangeNonDisjoint(avlNodeH(q), avlNodeH(p), right, left)
		else if avlNodeH(q) = avlNodeH(p)^^.sons[left] then
			exchangeNonDisjoint(avlNodeH(p), avlNodeH(q), left, right)
		else if avlNodeH(q) = avlNodeH(p)^^.sons[right] then
			exchangeNonDisjoint(avlNodeH(p), avlNodeH(q), right, left)
		else
			exchangeDisjoint(avlNodeH(p), avlNodeH(q));
	end;

function avlTree.extreme (direction: directionType): reference;
	var
		p: avlNodeH;
	begin
		p := root;
		if p = nil then
			extreme := nil
		else begin
			while p^^.sons[direction] <> nil do
				p := p^^.sons[direction];
			extreme := reference(p)
		end;
	end;

procedure avlTree.forAll (procedure whatToDo (x: obj); direction: directionType);

	procedure traverse (p: avlNodeH; procedure whatToDo (x: obj); direction, otherDirection: directionType);
		begin
			if p <> nil then begin
				traverse(p^^.sons[direction], whatToDo, direction, otherDirection);
				whatToDo(p^^.value);
				traverse(p^^.sons[otherDirection], whatToDo, direction, otherDirection);
			end;
		end;

	begin (* avlTree.forAll *)
		traverse(root, whatToDo, direction, switchDirection(direction));
	end;

function avlTree.exists (function property (x: obj): boolean): reference;

	procedure traverse (p: avlNodeH; function property (x: obj): boolean);
		begin
			if p <> nil then begin
				if property(p^^.value) then begin
					exists := reference(p);
					exit(exists);
				end;
				traverse(p^^.sons[left], property);
				traverse(p^^.sons[right], property);
			end;
		end;

	begin (* avlTree.exists *)
		traverse(root, property);
		exists := nil;
	end;

procedure avlTree.rangeForAll (leftBound, rightBound: obj; function compare (a, b: obj): relationObj; procedure whatToDo (x: obj); direction: directionType);

	procedure traverse (p: avlNodeH; leftBound, rightBound: obj; function compare (a, b: obj): relationObj; procedure whatToDo (x: obj); direction, otherDirection: directionType);
		begin
			if p <> nil then
				if (leftBound <> nil) & (compare(p^^.value, leftBound) = lessThanObj) then
					traverse(p^^.sons[right], leftBound, rightBound, compare, whatToDo, direction, otherDirection)
				else if (rightBound <> nil) & (compare(p^^.value, rightBound) = greaterThanObj) then
					traverse(p^^.sons[left], leftBound, rightBound, compare, whatToDo, direction, otherDirection)
				else begin
					traverse(p^^.sons[direction], leftBound, rightBound, compare, whatToDo, direction, otherDirection);
					whatToDo(p^^.value);
					traverse(p^^.sons[otherDirection], leftBound, rightBound, compare, whatToDo, direction, otherDirection);
				end;
		end;

	begin
		traverse(root, leftBound, rightBound, compare, whatToDo, direction, switchDirection(direction));
	end;


(* begin Marco Ciot *)

procedure avlTree.rangeForAllWhile (leftBound, rightBound: obj; function compare (a, b: obj): relationObj; function whatToDo (x: obj): boolean; direction: directionType);
	var
		guard: boolean;
	procedure traverse (p: avlNodeH; leftBound, rightBound: obj; function compare (a, b: obj): relationObj; function whatToDo (x: obj): boolean; direction, otherDirection: directionType);
		begin
			if guard and (p <> nil) then
				if (leftBound <> nil) & (compare(p^^.value, leftBound) = lessThanObj) then
					traverse(p^^.sons[right], leftBound, rightBound, compare, whatToDo, direction, otherDirection)
				else if (rightBound <> nil) & (compare(p^^.value, rightBound) = greaterThanObj) then
					traverse(p^^.sons[left], leftBound, rightBound, compare, whatToDo, direction, otherDirection)
				else begin
					traverse(p^^.sons[direction], leftBound, rightBound, compare, whatToDo, direction, otherDirection);
					if guard then begin
						guard := whatToDo(p^^.value);
						traverse(p^^.sons[otherDirection], leftBound, rightBound, compare, whatToDo, direction, otherDirection);
					end;
				end;
		end;

	begin
		guard := true;
		traverse(root, leftBound, rightBound, compare, whatToDo, direction, switchDirection(direction));
	end;
(* end Marco Ciot *)

function avlTree.compareReference (p, q: reference): relationObj;
	type
		treeDepth = 1..maxDepth;
		treeDepth0 = 0..maxDepth;
		path = record
				leftTurn: array[treeDepth] of boolean;
				n: treeDepth0;
			end;

	var
		pathP, pathQ: path;

	function getPath (x: avlNodeH): path;
		var
			p: path;
		begin
			p.n := 0;
			while x <> root do begin
				p.n := p.n + 1;
				p.leftTurn[p.n] := x^^.father^^.sons[left] = x;
				x := x^^.father;
			end;
			getPath := p;
		end;

	begin
		if p = q then
			compareReference := equalObj
		else begin
			pathP := getPath(avlNodeH(p));
			pathQ := getPath(avlNodeH(q));
			while (pathP.n > 0) & (pathQ.n > 0) & (pathP.leftTurn[pathP.n] = pathQ.leftTurn[pathQ.n]) do begin
				pathP.n := pathP.n - 1;
				pathQ.n := pathQ.n - 1;
			end;
			if pathP.n = 0 then
				if pathQ.leftTurn[pathQ.n] then
					compareReference := greaterThanObj
				else
					compareReference := lessThanObj
			else if pathP.leftTurn[pathP.n] then
				compareReference := lessThanObj
			else
				compareReference := greaterThanObj;
		end;
	end;

	var
		freeList: avlNodeH;
		freeListN: longInt;

procedure initFreeListAVL;
	begin
		freeList := nil;
		freeListN := 0;
	end;

function newElementAVL: avlNodeH;
	var
		t: avlNodeH;
	begin
		if freeList = nil then begin
			t := avlNodeH(newHandleFailNil(sizeOf(avlNode)));
			newElementAVL := t;
		end
		else begin
			newElementAVL := freeList;
			freeList := freeList^^.father;
			freeListN := freeListN - 1;
		end;
	end;

procedure disposeElementAVL (e: avlNodeH);
	begin
		e^^.father := freeList;
		freeList := e;
		freeListN := freeListN + 1;
	end;

function freeListSizeAVL: longInt;
	begin
		freeListSizeAVL := freeListN * (sizeOf(avlNode) + 8);
	end;

function freeListLengthAVL: longInt;
	begin
		freeListLengthAVL := freeListN;
	end;

procedure extendFreeListAVL (number: longInt);
	var
		i: longInt;
		e: avlNodeH;
	begin
		for i := 1 to number do begin
{$IFC AnimationEnabled }
			advanceStopWatchHandsNoMem;
{$ENDC }
			e := avlNodeH(newHandleFailNil(sizeOf(avlNode)));
			e^^.father := freeList;
			freeList := e;
		end;
		freeListN := freeListN + number;
	end;

procedure shrinkFreeListAVL (number: longInt);
	var
		i: longInt;
		e: avlNodeH;
	begin
		if number > freeListN then
			number := freeListN;
		for i := 1 to number do begin
{$IFC AnimationEnabled }
			advanceStopWatchHandsNoMem;
{$ENDC }
			e := freeList;
			freeList := freeList^^.father;
{$IFC THINK_PASCAL }
			disposHandle(Handle(e));
{$ELSEC }
			DisposeHandle(Handle(e));
{$ENDC }
		end;
		freeListN := freeListN - number;
	end;

end.
