(*	XYZ GeoBench (eXperimental geometrY Zurich):		*)
(*	An object oriented, interactive workbench for		*)
(*	experimental geometric computation.					*)
(*														*)
(*	 1999 by Peter Schorn, peter.schorn@acm.org		*)
(*	Culmannstrasse 77, CH-8006 Zrich, Switzerland		*)
(*	This software may be copied freely for educational	*)
(*	and research purposes only. For commercial use,		*)
(*	please contact the author.							*)
(*														*)
(*	Version 5.0.5: 30-Sep-1999							*)

unit boundingBox;

interface

uses
	triangulation3d, rectangleVector, triangulation2d, lineSegmentVector, point3d, triangulation3d, quadTree, twoDTree;

type
	boundingBox2d = object(rectangle)
			function getType: objectType;
			override;
			function getTextType: Str255;
			override;
			function memberType (o: obj): boolean;
			override;
		end;

	cube3d = object(obj)
			xMin, xMax, yMin, yMax, zMin, zMax: real;
			procedure init;
			override;
			function isSelected (where: obj): boolean;
			override;
			procedure describe;
			override;
			procedure execute (item: integer);
			override;
			procedure textFileInput;
			override;
			procedure textFileOutput;
			override;
			procedure fileInput;
			override;
			procedure fileOutput;
			override;
			procedure interactiveOutput;
			override;
			function drawingComplexity (cutOff: longInt): longInt;
			override;
			function getType: objectType;
			override;
			function getTextType: Str255;
			override;
			function equal (o: obj): boolean;
			override;
			function memberType (o: obj): boolean;
			override;
			function getInfo: obj;
			override;
			procedure setInfo (o: obj);
			override;
			procedure randomChange;
			override;
			function getBoundingObject: obj;
			override;
			function genericLessThan (o: obj): boolean;
			override;
			procedure translate (dx, dy: extended);
			override;
			procedure rotate (xOrigin, yOrigin, alpha: extended);
			override;
			procedure scale (lambda: extended);
			override;
			procedure reflect (a, b, c: extended);
			override;

			procedure orderVertices;
			procedure appendCube (cube: cube3d);
		end;

	boundingBox3d = object(cube3d)
			function getType: objectType;
			override;
			function getTextType: Str255;
			override;
			function memberType (o: obj): boolean;
			override;
		end;

procedure initBoundingBoxes;

implementation

function boundingBox2d.getType: objectType;
	begin
		getType := boundingBox2dType;
	end;

function boundingBox2d.getTextType: Str255;
	begin
		getTextType := 'BoundingBox 2d';
	end;

function boundingBox2d.memberType (o: obj): boolean;
	begin
		memberType := member(o, boundingBox2d);
	end;


procedure cube3d.init;
	begin
		xMin := 0.0;
		xMax := 0.0;
		yMin := 0.0;
		yMax := 0.0;
		zMin := 0.0;
		zMax := 0.0;
	end;

function cube3d.genericLessThan (o: obj): boolean;
	var
		c: cube3d;
	begin
		c := cube3d(o);
		genericLessThan := (xMin < c.xMin) | (xMin = c.xMin) & ((xMax < c.xMax) | (xMax = c.xMax) & ((yMin < c.yMin) | (yMin = c.yMin) & ((yMax < c.yMax) | (yMax = c.yMax) & ((zMin < c.zMin) | (zMin = c.zMin) & (zMax < c.zMax)))));
	end;

procedure cube3d.translate (dx, dy: extended);
	begin
		xMin := xMin + dx;
		xMax := xMax + dx;
		yMin := yMin + dy;
		yMax := yMax + dy;
	end;

procedure cube3d.rotate (xOrigin, yOrigin, alpha: extended);
	var
		x, y: extended;
	begin
		x := xMin;
		y := yMin;
		rotatePoint(xOrigin, yOrigin, alpha, x, y);
		xMin := x;
		yMin := y;
		x := xMax;
		y := yMax;
		rotatePoint(xOrigin, yOrigin, alpha, x, y);
		xMax := x;
		yMax := y;
		orderVertices;
	end;

procedure cube3d.scale (lambda: extended);
	begin
		xMin := lambda * xMin;
		xMax := lambda * xMax;
		yMin := lambda * yMin;
		yMax := lambda * yMax;
		orderVertices;
	end;

procedure cube3d.reflect (a, b, c: extended);
	var
		x, y: extended;
	begin
		x := xMin;
		y := yMin;
		reflectPoint(a, b, c, x, y);
		xMin := x;
		yMin := y;
		x := xMax;
		y := yMax;
		reflectPoint(a, b, c, x, y);
		xMax := x;
		yMax := y;
		orderVertices;
	end;

function cube3d.isSelected (where: obj): boolean;
	var
		l: lineSegment;
		sel: boolean;

	function lineIsSelected (x1, y1, z1, x2, y2, z2: real): boolean;
		var
			ix1, ix2, iy1, iy2: integer;
			xs, ys, xe, ye: extended;
			dummy: boolean;
		begin
			dummy := worldToWindow3d(x1, y1, z1, ix1, iy1);
			dummy := worldToWindow3d(x2, y2, z2, ix2, iy2);
			windToWorld(ix1, iy1, xs, ys);
			windToWorld(ix2, iy2, xe, ye);
			l.p.setX(xs);
			l.p.setY(ys);
			l.q.setX(xe);
			l.q.setY(ye);
			lineIsSelected := l.isSelected(where);
		end;

	begin
		l := lineSegment(newInitObj(lineSegmentType));
		sel := lineIsSelected(xMin, yMin, zMin, xMin, yMax, zMin);
		sel := sel | lineIsSelected(xMin, yMax, zMin, xMax, yMax, zMin);
		sel := sel | lineIsSelected(xMax, yMax, zMin, xMax, yMin, zMin);
		sel := sel | lineIsSelected(xMax, yMin, zMin, xMin, yMin, zMin);
		sel := sel | lineIsSelected(xMin, yMin, zMax, xMin, yMax, zMax);
		sel := sel | lineIsSelected(xMin, yMax, zMax, xMax, yMax, zMax);
		sel := sel | lineIsSelected(xMax, yMax, zMax, xMax, yMin, zMax);
		sel := sel | lineIsSelected(xMax, yMin, zMax, xMin, yMin, zMax);
		sel := sel | lineIsSelected(xMin, yMin, zMin, xMin, yMin, zMax);
		sel := sel | lineIsSelected(xMin, yMax, zMin, xMin, yMax, zMax);
		sel := sel | lineIsSelected(xMax, yMax, zMin, xMax, yMax, zMax);
		isSelected := sel | lineIsSelected(xMax, yMin, zMin, xMax, yMin, zMax);
		l.freeAll;
	end;

	const
		gimmeOneC = 1;
		enclosingCubeC = 2;
		boundingBoxC = 3;

procedure cube3d.describe;
	begin
		with list(currentArguments) do begin
			if (length > 0) then begin
{$IFC OptionalCommand }
				menuEntry('Create Cube', gimmeOneC);
{$ENDC }
				menuEntry('Bounding Rectangle', boundingBoxC);
			end;
			if (length = 2) then begin
				if member(ith(1), cube3d) & member(ith(1), cube3d) then
					menuEntry('Bounding Box', enclosingCubeC);
			end;
		end;
	end;

procedure cube3d.execute (item: integer);
	var
		cube: cube3d;
	begin
		with list(currentArguments) do begin
			case item of
				gimmeOneC:  begin
					new(cube);
					failNil(Handle(cube));
					cube.init;
					cube.randomChange;
					currentResult.append(cube);
				end;
				enclosingCubeC:  begin
					cube := cube3d(ith(1).duplicateAll);
					cube.appendCube(cube3d(ith(2)));
					currentResult.append(cube);
				end;
				boundingBoxC: 
					currentResult.appendNonNil(currentArguments.getBoundingObject);
			end;
		end;
	end;

procedure cube3d.textFileInput;
	begin
		readTextReal(xMin);
		readTextReal(xMax);
		readTextReal(yMin);
		readTextReal(yMax);
		readTextReal(zMin);
		readTextReal(zMax);
	end;

procedure cube3d.textFileOutput;
	begin
		writeTextReal(xMin);
		writeTextReal(xMax);
		writeTextReal(yMin);
		writeTextReal(yMax);
		writeTextReal(zMin);
		writeTextReal(zMax);
	end;

procedure cube3d.fileInput;
	begin
		readReal(xMin);
		readReal(xMax);
		readReal(yMin);
		readReal(yMax);
		readReal(zMin);
		readReal(zMax);
	end;

procedure cube3d.fileOutput;
	begin
		writeReal(xMin);
		writeReal(xMax);
		writeReal(yMin);
		writeReal(yMax);
		writeReal(zMin);
		writeReal(zMax);
	end;

procedure cube3d.interactiveOutput;
	begin
		if (not coordSysIsDrawn) then begin
			coordSysIsDrawn := true;
			drawCoordSys;
		end;
		drawLineSegment3d(xMin, yMin, zMin, xMin, yMax, zMin);
		drawLineSegment3d(xMin, yMax, zMin, xMax, yMax, zMin);
		drawLineSegment3d(xMax, yMax, zMin, xMax, yMin, zMin);
		drawLineSegment3d(xMax, yMin, zMin, xMin, yMin, zMin);

		drawLineSegment3d(xMin, yMin, zMax, xMin, yMax, zMax);
		drawLineSegment3d(xMin, yMax, zMax, xMax, yMax, zMax);
		drawLineSegment3d(xMax, yMax, zMax, xMax, yMin, zMax);
		drawLineSegment3d(xMax, yMin, zMax, xMin, yMin, zMax);

		drawLineSegment3d(xMin, yMin, zMin, xMin, yMin, zMax);
		drawLineSegment3d(xMin, yMax, zMin, xMin, yMax, zMax);
		drawLineSegment3d(xMax, yMax, zMin, xMax, yMax, zMax);
		drawLineSegment3d(xMax, yMin, zMin, xMax, yMin, zMax);
	end;

{$IFC NOT THINK_PASCAL }
{$PUSH }
{$PRAGMAC warn_unusedarg off }
{$ENDC }
function cube3d.drawingComplexity (cutOff: longInt): longInt;
	begin
		drawingComplexity := 12;
	end;
{$IFC NOT THINK_PASCAL }
{$POP }
{$ENDC }

function cube3d.getType: objectType;
	begin
		getType := cube3dType;
	end;

function cube3d.getTextType: Str255;
	begin
		getTextType := 'Cube 3d';
	end;

function cube3d.equal (o: obj): boolean;
	begin
		equal := (o <> nil) & member(o, cube3d) & (cube3d(o).xMin = xMin) & (cube3d(o).xMax = xMax) & (cube3d(o).yMin = yMin) & (cube3d(o).yMax = yMax) & (cube3d(o).zMin = zMin) & (cube3d(o).zMax = zMax);
	end;

function cube3d.memberType (o: obj): boolean;
	begin
		memberType := member(o, cube3d);
	end;

function cube3d.getInfo: obj;
	var
		v: vector;
	begin
		v := newVector(6);
		appendObjInfoExtended(v, 'X-Min', xMin, false);
		appendObjInfoExtended(v, 'X-Max', xMax, false);
		appendObjInfoExtended(v, 'Y-Min', yMin, false);
		appendObjInfoExtended(v, 'Y-Max', yMax, false);
		appendObjInfoExtended(v, 'Z-Min', zMin, false);
		appendObjInfoExtended(v, 'Z-Max', zMax, false);
		getInfo := v;
	end;

procedure cube3d.setInfo (o: obj);
	var
		v: vector;
	begin
		v := vector(o);
		xMin := objInfoExtended(v.elements^^[1]).e;
		xMax := objInfoExtended(v.elements^^[2]).e;
		yMin := objInfoExtended(v.elements^^[3]).e;
		yMax := objInfoExtended(v.elements^^[4]).e;
		zMin := objInfoExtended(v.elements^^[5]).e;
		zMax := objInfoExtended(v.elements^^[6]).e;
		orderVertices;
	end;

procedure order (var x, y: real);
	var
		t: real;
	begin
		if x > y then begin
			t := x;
			x := y;
			y := t;
		end;
	end;

procedure cube3d.orderVertices;
	begin
		order(xMin, xMax);
		order(yMin, yMax);
		order(zMin, zMax);
	end;

procedure cube3d.randomChange;
	var
		o, p, q: realPoint;
	begin
		o := realPoint(newObj(realPointType));
		p := realPoint(newObj(realPointType));
		q := realPoint(newObj(realPointType));
		o.randomChange;
		p.randomChange;
		q.randomChange;
		xMin := o.getX;
		yMin := p.getX;
		xMax := o.getY;
		yMax := p.getY;
		zMin := q.getX;
		zMax := q.getY;
		o.free;
		p.free;
		q.free;
		orderVertices;
	end;

function cube3d.getBoundingObject: obj;
	var
		box: boundingBox3d;
	begin
		new(box);
		failNil(Handle(box));
		box.xMin := xMin;
		box.xMax := xMax;
		box.yMin := yMin;
		box.yMax := yMax;
		box.zMin := zMin;
		box.zMax := zMax;
		getBoundingObject := box;
	end;

procedure cube3d.appendCube (cube: cube3d);
	begin
		cube.orderVertices;
		orderVertices;
		if cube.xMin < xMin then
			xMin := cube.xMin;
		if cube.xMax > xMax then
			xMax := cube.xMax;
		if cube.yMin < yMin then
			yMin := cube.yMin;
		if cube.yMax > yMax then
			yMax := cube.yMax;
		if cube.zMin < zMin then
			zMin := cube.zMin;
		if cube.zMax > zMax then
			zMax := cube.zMax;
	end;


function boundingBox3d.getType: objectType;
	begin
		getType := boundingBox3dType;
	end;

function boundingBox3d.getTextType: Str255;
	begin
		getTextType := 'BoundingBox 3d';
	end;

function boundingBox3d.memberType (o: obj): boolean;
	begin
		memberType := member(o, boundingBox3d);
	end;

procedure initBoundingBoxes;
	var
		box2d: boundingBox2d;
		cube: cube3d;
		box3d: boundingBox3d;
	begin
		new(box2d);
		failNil(Handle(box2d));
		prototypes[boundingBox2dType] := box2d;
		new(box3d);
		failNil(Handle(box3d));
		prototypes[boundingBox3dType] := box3d;
		new(cube);
		failNil(Handle(cube));
		prototypes[cube3dType] := cube;
	end;


function sequence.getBoundingObject: obj;
	var
		box: obj;

	procedure appendToBox (o: obj);
		var
			box2: obj;
		begin
			if box = nil then
				box := o.getBoundingObject
			else begin
				box2 := o.getBoundingObject;
				if box2 <> nil then begin
					if box.getType = boundingBox2dType then begin
						if box2.getType = boundingBox2dType then begin
							if boundingBox2d(box2).left < boundingBox2d(box).left then
								boundingBox2d(box).left := boundingBox2d(box2).left;
							if boundingBox2d(box2).right > boundingBox2d(box).right then
								boundingBox2d(box).right := boundingBox2d(box2).right;
							if boundingBox2d(box2).bottom < boundingBox2d(box).bottom then
								boundingBox2d(box).bottom := boundingBox2d(box2).bottom;
							if boundingBox2d(box2).top > boundingBox2d(box).top then
								boundingBox2d(box).top := boundingBox2d(box2).top;
							box2.freeAll;
						end
						else begin  (* box2 is boundingBox3dType *)
							if boundingBox2d(box).left < boundingBox3d(box2).xMin then
								boundingBox3d(box2).xMin := boundingBox2d(box).left;
							if boundingBox2d(box).right > boundingBox3d(box2).xMax then
								boundingBox3d(box2).xMax := boundingBox2d(box).right;
							if boundingBox2d(box).bottom < boundingBox3d(box2).yMin then
								boundingBox3d(box2).yMin := boundingBox2d(box).bottom;
							if boundingBox2d(box).top > boundingBox3d(box2).yMax then
								boundingBox3d(box2).yMax := boundingBox2d(box).top;
							box.freeAll;
							box := box2;
						end;
					end
					else begin	(* box is boundingBox3dType *)
						if box2.getType = boundingBox2dType then begin
							if boundingBox2d(box2).left < boundingBox3d(box).xMin then
								boundingBox3d(box).xMin := boundingBox2d(box2).left;
							if boundingBox2d(box2).right > boundingBox3d(box).xMax then
								boundingBox3d(box).xMax := boundingBox2d(box2).right;
							if boundingBox2d(box2).bottom < boundingBox3d(box).yMin then
								boundingBox3d(box).yMin := boundingBox2d(box2).bottom;
							if boundingBox2d(box2).top > boundingBox3d(box).yMax then
								boundingBox3d(box).yMax := boundingBox2d(box2).top;
							box2.freeAll;
						end
						else begin  (* box2 is boundingBox3dType *)
							if boundingBox3d(box2).xMin < boundingBox3d(box).xMin then
								boundingBox3d(box).xMin := boundingBox3d(box2).xMin;
							if boundingBox3d(box2).xMax > boundingBox3d(box).xMax then
								boundingBox3d(box).xMax := boundingBox3d(box2).xMax;
							if boundingBox3d(box2).yMin < boundingBox3d(box).yMin then
								boundingBox3d(box).yMin := boundingBox3d(box2).yMin;
							if boundingBox3d(box2).yMax > boundingBox3d(box).yMax then
								boundingBox3d(box).yMax := boundingBox3d(box2).yMax;
							if boundingBox3d(box2).zMin < boundingBox3d(box).zMin then
								boundingBox3d(box).zMin := boundingBox3d(box2).zMin;
							if boundingBox3d(box2).zMax > boundingBox3d(box).zMax then
								boundingBox3d(box).zMax := boundingBox3d(box2).zMax;
							box2.freeAll;
						end;
					end;
				end; (* if box2<>nil *)
			end; (* if box=nil *)
		end; (* appendToBox *)

	begin
		box := nil;
		forAll(appendToBox);
		getBoundingObject := box;
	end;

function point2d.getBoundingObject: obj;
	const
		offsetDx = 3;
		offsetDy = 3;
	var
		box: boundingBox2d;
	begin
		new(box);
		failNil(Handle(box));
		with box, currentFrame do begin
			init;
			left := getX - offsetDx / mx;
			right := getX + offsetDx / mx;
			top := getY + offsetDy / my;
			bottom := getY - offsetDy / my;
		end;
		getBoundingObject := box;
	end;

function markedRealPoint.getBoundingObject: obj;
	const
		offsetDx = 6;
		offsetDy = 6;
	var
		box: boundingBox2d;
		currentPort: GrafPtr;
	begin
		new(box);
		failNil(Handle(box));
		GetPort(currentPort);
		with box, currentFrame do begin
			init;
			left := getX - offsetDx / mx;
			right := getX + (offsetDx + StringWidth(mark.value)) / mx;
			top := getY + offsetDy / my;
			bottom := getY - (offsetDy + currentPort^.txSize) / my;
		end;
		getBoundingObject := box;
	end;

function voronoiDiagram.getBoundingObject: obj;
	begin
		getBoundingObject := points.getBoundingObject;
	end;

function quadTree.getBoundingObject: obj;
	begin
		getBoundingObject := boundingRect.getBoundingObject;
	end;

function twoDTree.getBoundingObject: obj;
	var
		rLeft, rBottom, rRight, rTop: extended;
		length: longInt;
		box: boundingBox2d;
	begin
		boundingBox(rLeft, rBottom, rRight, rTop, length);
		new(box);
		failNil(Handle(box));
		with box do begin
			init;
			left := rLeft;
			right := rRight;
			top := rTop;
			bottom := rBottom;
			orderVertices;
		end;
		getBoundingObject := box;
	end;

function simpleUndirectedGraph.getBoundingObject: obj;
	begin
		getBoundingObject := vertices.getBoundingObject;
	end;

function lineSegment.getBoundingObject: obj;
	var
		box: boundingBox2d;
	begin
		new(box);
		failNil(Handle(box));
		with box do begin
			init;
			left := p.getX;
			right := q.getX;
			top := p.getY;
			bottom := q.getY;
			orderVertices;
		end;
		getBoundingObject := box;
	end;

function circle.getBoundingObject: obj;
	var
		box: boundingBox2d;
	begin
		new(box);
		failNil(Handle(box));
		with box do begin
			init;
			left := x - radius;
			right := x + radius;
			top := y - radius;
			bottom := y + radius;
			orderVertices;
		end;
		getBoundingObject := box;
	end;

function rectangle.getBoundingObject: obj;
	var
		box: boundingBox2d;
	begin
		new(box);
		failNil(Handle(box));
		box.left := left;
		box.right := right;
		box.top := top;
		box.bottom := bottom;
		getBoundingObject := box;
	end;

function realPoint3d.getBoundingObject: obj;
	var
		box: boundingBox3d;
	begin
		new(box);
		failNil(Handle(box));
		with box do begin
			init;
			xMin := x;
			xMax := xMin;
			yMin := y;
			yMax := yMin;
			zMin := z;
			zMax := zMin;
			orderVertices;
		end;
		getBoundingObject := box;
	end;

procedure realPoint3d.randomChange;
	begin
		if currentRandomConstraint <> nil then
			if member(currentRandomConstraint, rectangle) then
				with rectangle(currentRandomConstraint) do begin
					x := (right - left) * randomReal + left;
					y := (bottom - top) * randomReal + top;
				end
			else if member(currentRandomConstraint, cube3d) then
				with cube3d(currentRandomConstraint) do begin
					x := (xMax - xMin) * randomReal + xMin;
					y := (yMax - yMin) * randomReal + yMin;
					z := (zMax - zMin) * randomReal + zMin;
				end;
	end;

function triangulation2d.getBoundingObject: obj;
	begin
		getBoundingObject := points.getBoundingObject;
	end;

function triangulation3d.getBoundingObject: obj;
	begin
		getBoundingObject := points.getBoundingObject;
	end;

function pointVector.getBoundingObject: obj;
	var
		box: boundingBox2d;
		i: seqIndex;
		x, y: real;
	begin
		if length = 0 then
			getBoundingObject := nil
		else begin
			box := boundingBox2d(point2d(elements^^[1]).getBoundingObject);
			for i := 2 to length do begin
				x := point2d(elements^^[i]).getX;
				y := point2d(elements^^[i]).getY;
				if x < box.left then
					box.left := x;
				if x > box.right then
					box.right := x;
				if y < box.bottom then
					box.bottom := y;
				if y > box.top then
					box.top := y;
			end;
			getBoundingObject := box;
		end;
	end;

function rectangleVector.getBoundingObject: obj;
	var
		box: boundingBox2d;
		i: seqIndex;
		xMin, xMax, yMin, yMax: real;
	begin
		box := boundingBox2d(point2d(elements^^[1]).getBoundingObject);
		for i := 2 to length do begin
			xMin := rectangle(elements^^[i]).left;
			xMax := rectangle(elements^^[i]).right;
			yMin := rectangle(elements^^[i]).bottom;
			yMax := rectangle(elements^^[i]).top;
			if xMin < box.left then
				box.left := xMin;
			if xMax > box.right then
				box.right := xMax;
			if yMin < box.bottom then
				box.bottom := yMin;
			if yMax > box.top then
				box.top := yMax;
		end;
		getBoundingObject := box;
	end;

function layer.getBoundingObject: obj;
	var
		box: boundingBox3d;
		box2: boundingBox2d;
	begin
		new(box);
		failNil(Handle(box));
		box2 := boundingBox2d(layElems.getBoundingObject);
		box.xMin := box2.left;
		box.xMax := box2.right;
		box.yMin := box2.bottom;
		box.yMax := box2.top;
		box.zMin := z;
		box.zMax := z + thickness;
		box2.freeAll;
		getBoundingObject := box;
	end;


function layerVector.getBoundingObject: obj;
	begin
		getBoundingObject := layers.getBoundingObject;
	end;

end.
