(*	XYZ GeoBench (eXperimental geometrY Zurich):		*)
(*	An object oriented, interactive workbench for		*)
(*	experimental geometric computation.					*)
(*														*)
(*	 1999 by Peter Schorn, peter.schorn@acm.org		*)
(*	Culmannstrasse 77, CH-8006 Zrich, Switzerland		*)
(*	This software may be copied freely for educational	*)
(*	and research purposes only. For commercial use,		*)
(*	please contact the author.							*)
(*														*)
(*	Version 5.0.5: 30-Sep-1999							*)

(* This Unit is added by Joerg Mrkwitschka (JM)			*)
(* Uni-GH Siegen, D-5900 Siegen, Germany				*)
(* Last change: 4-May-93								*)


unit closestPairConvexPolygon;

interface

uses
	pointVector;

implementation

 {$SETC AConvexPolygonDistance = TRUE }
 {$SETC AClosestPairConvexPolygon    = TRUE }

const
	below = left;
	above = right;

(* -------------------------------------------------------------------------- *)
(* --- Define new Object cpcpEvent (C_onvex P_olygon C_losest P_air EVENT) -- *)
(* -------------------------------------------------------------------------- *)

type

	cpcpEvent = object(obj)               (* Convex Polygon Closest Pair EVENT *)
			where: point2d;
			procedure processEvent (yTable: dictionary; var shortestConnection: lineSegment; var delta: extended);
		end;

	cpcpInsertEvent = object(cpcpEvent)
			cPolygon: convexPolygon;
			deleteEvent: cpcpDeleteEvent;
			procedure processEvent (yTable: dictionary; var shortestConnection: lineSegment; var delta: extended);
			override;
		end;

	cpcpDeleteEvent = object(cpcpEvent)
			toDelete: reference;   (* into y-table *)
			procedure processEvent (yTable: dictionary; var shortestConnection: lineSegment; var delta: extended);
			override;
		end;

var
	reallyChecked, notChecked: longInt;
                             (* Counters for closest-pair-algorithm          *)
                             (* reallyChecked: this distance is checked      *)
                             (*                which minimalDistance         *)
                             (* notChecked:  this distance is not checked,   *)
                             (*              because the containers are more *)
                             (*              than delta appart.              *)

(* -------------------------------------------------------------------------- *)
(* -------- Define some functions which are used by cpcpEvent  -------------- *)
(* -------------------------------------------------------------------------- *)

function compareConvexPolygon (cp, cq: convexPolygon): relationObj;
	begin
		if cp.objectYLessThan(cq) then
			compareConvexPolygon := lessThanObj
		else
			compareConvexPolygon := greaterThanObj;
              (* -> note that  'equalObj'  or  'unEqualObj'  never happens ! *)
	end;

(* -------------------------------------------------------------------------- *)

function lessThanCpcpEvent (a, b: cpcpEvent): boolean;
	begin
		lessThanCpcpEvent := a.where.genericLessThan(b.where);
	end;

(* -------------------------------------------------------------------------- *)

procedure checkPolygonDistance (cp, cq: convexPolygon; var shortestConnection: lineSegment; var delta: extended);
	var                         (* cp is below cq !                            *)
		dl: lineSegment;
		newDelta: extended;
	begin                    (* First we check the distance between the      *)
                             (* minimal point of cq (which is the Polygon    *)
                             (* above) and the maximal point of cp in y.     *)
                             (* This distance is the minimum of distance     *)
                             (* between cp and cq. *)
                             (* If even this minimum is greater then delta,  *)
                             (* then we don't have then we don't have to     *)
                             (* calculate the distance,because we know that  *)
                             (* it must be greater than delta.               *)
                             (* This is a kind of a container-method.        *)
		if cq.getVertex(cq.bottom).getY - cp.getVertex(cp.top).getY < delta then begin
			dl := cp.minimalDistance(cq, newDelta);
			reallyChecked := reallyChecked + 1;
      {$IFC AnimationEnabled AND AClosestPairConvexPolygon }
			if animationFlag[AClosestPairItem] then
				dl.interactiveFlash; (* flash the checked Connection *)
      {$ENDC }
			if newDelta < delta then begin (* Check distance between two polygons. *)
                             (* If the distance is smaller, then we have a   *)
				delta := newDelta;   (* new shortest Connection and have to free the *)
                             (* memory.                                      *)
        {$IFC AnimationEnabled AND AClosestPairConvexPolygon }
				if animationFlag[AClosestPairItem] then
					shortestConnection.interactiveHighlight;
        {$ENDC }
           (* Delete old shortest Connection.              *)
				shortestConnection.freeAll;
				shortestConnection := dl;
        {$IFC AnimationEnabled AND AClosestPairConvexPolygon }
				if animationFlag[AClosestPairItem] then
					shortestConnection.interactiveHighlight;
        {$ENDC }
          (* Draw new shortest Connection.                 *)
			end
			else
				dl.freeAll;          (* If the new distance is not smaller then free *)
                             (* the memory of dl.                            *)
		end
		else
			notChecked := notChecked + 1;
	end;

(* -------------------------------------------------------------------------- *)
(* --------- Definition of the cpcpEventMethods  ---------------------------- *)
(* -------------------------------------------------------------------------- *)

{$IFC NOT THINK_PASCAL }
{$PUSH }
{$PRAGMAC warn_unusedarg off }
{$ENDC }
procedure cpcpEvent.processEvent (yTable: dictionary; var shortestConnection: lineSegment; var delta: extended);
	begin
		unimplementedMethod('cpcpEvent.processEvent');
	end;
{$IFC NOT THINK_PASCAL }
{$POP }
{$ENDC }

(* -------------------------------------------------------------------------- *)

procedure cpcpInsertEvent.processEvent;
	var
		whereRef: reference;
		direction: directionType;
		aboveRef, belowRef: reference;
	begin
		if not yTable.find(cPolygon, compareConvexPolygon, whereRef, direction) then begin
      {$IFC AnimationEnabled AND AClosestPairConvexPolygon }
			if animationFlag[AClosestPairItem] then begin
				cPolygon.interactiveHighlight;
			end;
      {$ENDC }
			whereRef := yTable.insert(cPolygon, whereRef, direction); (* insert    *)
			deleteEvent.toDelete := whereRef;          (* set ref. of deleteEvent  *)
			aboveRef := yTable.next(whereRef, above);
			belowRef := yTable.next(whereRef, below);
			if aboveRef <> nil then
				checkPolygonDistance(cPolygon, convexPolygon(yTable.getObject(aboveRef)), shortestConnection, delta);
			if (belowRef <> nil) & (delta > 0) then
				checkPolygonDistance(convexPolygon(yTable.getObject(belowRef)), cPolygon, shortestConnection, delta);
		end;
                             (* The ELSE - case never happens because the    *)
                             (* comparefunction never tells the two Polygons *)
                             (*to be equal !                                 *)
	end;

(* -------------------------------------------------------------------------- *)

procedure cpcpDeleteEvent.processEvent;
	var
		aboveRef, belowRef: reference;
	begin
		aboveRef := yTable.next(toDelete, above);
		belowRef := yTable.next(toDelete, below);
		if (aboveRef <> nil) & (belowRef <> nil) then
			checkPolygonDistance(convexPolygon(yTable.getObject(belowRef)), convexPolygon(yTable.getObject(aboveRef)), shortestConnection, delta);
		yTable.delete(toDelete);
	end;

(* --------------------------------------------------------------------------	*)
(* Define function and procedures that are used by min.-dist. algorithms.  --			*)
(* --------------------------------------------------------------------------	*)

function dPointLine (p: point2d; l: lineSegment; ps: point2d): extended;
	var
		d, w: extended;
		lambda: signType;
	begin
		l.p.projectOnLine(l.q, p, ps, d, w, lambda);
		if w > 0 then
			dPointLine := d
		else
			dPointLine := -d;
	end;

function lambda (p: point2d; l: lineSegment): extended;
	var
		d, ws: extended;
		ll: signType;
	begin
		l.p.projectOnLine(l.q, p, nil, d, ws, ll);
		lambda := ll + 0.5;
	end;

function dPointLineseg (p: point2d; l: lineSegment; ps: point2d): extended;
	var
		dh, lamb: extended;
	begin
		dh := dPointLine(p, l, ps);			(*	ps: projection-point from p onto l.	*)
											(*	sign of d indicates the side.      	*)
		lamb := lambda(ps, l);				(*	calculate line-parameter.          	*)
		if lamb < 0 then begin				(*	decide which point builds minimal  	*)
			ps.setPoint(l.p);				(*	distance; note that ph must lie    	*)
			dPointLineseg := p.distanceX(l.p);
		end									(*	between p1 and p2 to build minimal 	*)
		else if lamb > 1 then begin			(*	distance. Otherwise one of the     	*)
			ps.setPoint(l.q);				(*	endpoints builds minimal distance! 	*)
			dPointLineseg := p.distanceX(l.q);
		end
		else begin
			dPointLineseg := dh;
		end;
	end;

function pointLeftToLine (p: point2d; l: lineSegment): boolean;
	begin
		pointLeftToLine := l.whichSideX(p) <= 0;
	end;

function dLinesegLineseg (l1, l2: lineSegment; var dl: lineSegment; ph: point2d): extended;
	var
		dold, dnew: extended;
	begin
		if l1.intersectLineSegmentX(l2, dl) then begin
			dLinesegLineseg := 0;
			dl.q.setPoint(dl.p);
		end
		else begin
			dold := dPointLineseg(l1.p, l2, dl.q);
			dl.p.setPoint(l1.p);
			dnew := dPointLineseg(l1.q, l2, ph);
			if dnew < dold then begin
				dold := dnew;
				dl.q.setPoint(ph);
				dl.p.setPoint(l1.q);
			end;
			dnew := dPointLineseg(l2.p, l1, ph);
			if dnew < dold then begin
				dold := dnew;
				dl.q.setPoint(ph);
				dl.p.setPoint(l2.p);
			end;
			dnew := dPointLineseg(l2.q, l1, ph);
			if dnew < dold then begin
				dold := dnew;
				dl.q.setPoint(ph);
				dl.p.setPoint(l2.q);
			end;
			dLinesegLineseg := dold;
		end;
	end;

(* -------------------------------------------------------------------------- *)
(* --------- Definition of some convexPolygon methods.  --------------------- *)
(* -------------------------------------------------------------------------- *)

function convexPolygon.objectYLessThan (cq: convexPolygon): boolean;
	var
		cpl, cql, cpr, cqr: point2d;
	begin
		cpl := self.getVertex(self.left);
		cql := cq.getVertex(cq.left);
		cpr := self.getVertex(self.right);
		cqr := cq.getVertex(cq.right);
		if cpl.xLessThan(cql) | cpl.xEqual(cql) then
			if cpr.xLessThan(cql) | cpr.xEqual(cql) then
				objectYLessThan := cpr.yLessThan(cql)
			else
				objectYLessThan := cql.whichSideX(cpl, cpr) < 0
		else if cqr.xLessThan(cpl) | cqr.xEqual(cpl) then
			objectYLessThan := cpl.yLessThan(cqr)
		else
			objectYLessThan := cpl.whichSideX(cql, cqr) > 0;
	end;

(* -------------------------------------------------------------------------- *)

function convexPolygon.simpleMinimalDistance (cq: convexPolygon; var d: extended): lineSegment;
	var
		cp: convexPolygon;
		dl: lineSegment;
		ph: point2d;

   (* --------------------------------------- *)

	procedure comparePointsWithLineSegs (cpp, cqq: convexPolygon; var dl: lineSegment; var d: extended; ph: point2d);
		var
			i, j: longInt;
			dnew: extended;
			p: point2d;
			pointInside: boolean;
			ee: lineSegment;
		begin
			new(ee);
			failNil(Handle(ee));

			i := 1;
			repeat
				p := cpp.getVertex(i);
        {$IFC AnimationEnabled AND AConvexPolygonDistance }
				if animationFlag[AConvexPolygonDistanceItem] then
					p.interactiveHighlight;
        {$ENDC }
				pointInside := true;
				for j := 1 to cqq.length do begin
					ee.p := cqq.getVertex(j);
					ee.q := cqq.getVertex(j + 1);
          {$IFC AnimationEnabled AND AConvexPolygonDistance }
					if animationFlag[AConvexPolygonDistanceItem] then
						ee.interactiveHighlight;
          {$ENDC }
					dnew := abs(dPointLineseg(p, ee, ph));
					if dnew < d then begin
            {$IFC AnimationEnabled AND AConvexPolygonDistance }
						if animationFlag[AConvexPolygonDistanceItem] then
							dl.interactiveHighlight;
            {$ENDC }
						d := dnew;
						dl.p.setPoint(p);
						dl.q.setPoint(ph);
            {$IFC AnimationEnabled AND AConvexPolygonDistance }
						if animationFlag[AConvexPolygonDistanceItem] then
							dl.interactiveHighlight;
            {$ENDC }
					end;
					pointInside := pointInside & pointLeftToLine(p, ee);
                             (* figure out, if one point is left to all      *)
                             (* linesegments of a polygon.                   *)
          {$IFC AnimationEnabled AND AConvexPolygonDistance }
					if animationFlag[AConvexPolygonDistanceItem] then begin
						waitForClick(animationFlag[AConvexPolygonDistanceItem]);
						if userAbort then begin
							simpleMinimalDistance := nil;
							dl.freeAll;
							ph.free;
							dispose(ee);
							exit(simpleMinimalDistance)
						end;
						ee.interactiveHighlight;
					end;
        {$ENDC }
				end;
        {$IFC AnimationEnabled AND AConvexPolygonDistance }
				if animationFlag[AConvexPolygonDistanceItem] then
					p.interactiveHighlight;
        {$ENDC }
				i := i + 1;
			until (i > cpp.length) | pointInside;
			if pointInside then begin        (* If one Point is left to all        *)
                                       (* linesegments then he is inside the *)
				d := 0;                        (* polygon.                           *)
				dl.p.setPoint(p);
				dl.q.setPoint(p);
			end;
			dispose(ee);
		end;

   (* --------------------------------------- *)

	function convexPolygonEdgesIntersect (cpp, cqq: convexPolygon; k: point2d): boolean;
		var
			i, j: longInt;
			ep, eq: lineSegment;
			edgeIntersect: boolean;
		begin
			new(ep);
			failNil(Handle(ep));
			new(eq);
			failNil(Handle(eq));
			edgeIntersect := false;
			i := 1;
			repeat
				ep.p := cpp.getVertex(i);
				ep.q := cpp.getVertex(i + 1);
       {$IFC AnimationEnabled AND AConvexPolygonDistance }
				if animationFlag[AConvexPolygonDistanceItem] then
					ep.interactiveHighlight;
        {$ENDC }
				j := 1;
				repeat
					eq.p := cqq.getVertex(j);
					eq.q := cqq.getVertex(j + 1);
          {$IFC AnimationEnabled AND AConvexPolygonDistance }
					if animationFlag[AConvexPolygonDistanceItem] then
						eq.interactiveHighlight;
          {$ENDC }
					edgeIntersect := edgeIntersect | ep.p.intersectLineSegmentX(ep.q, eq.p, eq.q, k, k);
					j := j + 1;
          {$IFC AnimationEnabled AND AConvexPolygonDistance }
					if animationFlag[AConvexPolygonDistanceItem] then begin
						waitForClick(animationFlag[AConvexPolygonDistanceItem]);
						if userAbort then begin
							simpleMinimalDistance := nil;
							dl.freeAll;
							ph.free;
							dispose(ep);
							dispose(eq);
							exit(simpleMinimalDistance)
						end;
						eq.interactiveHighlight;
					end;
          {$ENDC }
				until edgeIntersect | (j > cqq.length);
				i := i + 1;
        {$IFC AnimationEnabled AND AConvexPolygonDistance }
				if animationFlag[AConvexPolygonDistanceItem] then
					ep.interactiveHighlight;
        {$ENDC }
			until edgeIntersect | (i > cpp.length);
			convexPolygonEdgesIntersect := edgeIntersect;
			dispose(ep);
			dispose(eq);
		end;


  (* ----------------------------------------------- *)

	begin
		cp := self;
		d := 1000000000;
		new(dl);
		failNil(Handle(dl));
    {$IFC AnimationEnabled AND AConvexPolygonDistance }
		if animationFlag[AConvexPolygonDistanceItem] then begin
			cp.interactiveHighlight;
			cq.interactiveHighlight;
		end;
    {$ENDC }
		if (cp.length = 1) & (cq.length = 1) then begin  (* Both are just points *)
			dl.p := point2d(cp.getVertex(1).duplicate);
			dl.q := point2d(cq.getVertex(1).duplicate);
			d := sqrt(dl.p.squaredDistanceX(dl.q));
      {$IFC AnimationEnabled AND AConvexPolygonDistance }
			if animationFlag[AConvexPolygonDistanceItem] then begin
				dl.interactiveHighlight;
				waitForClick(animationFlag[AConvexPolygonDistanceItem]);
				if userAbort then begin
					simpleMinimalDistance := nil;
					dl.freeAll;
					exit(simpleMinimalDistance)
				end;
			end;
      {$ENDC }
		end
		else begin                           (* At least one is line or polygon. *)
			ph := point2d(cp.getVertex(1).duplicate);
			dl.p := point2d(cp.getVertex(1).duplicate);
			dl.p.setX(-100);
			dl.p.setY(-100);
			dl.q := point2d(dl.p.duplicate);
      {$IFC AnimationEnabled AND AConvexPolygonDistance }
			if animationFlag[AConvexPolygonDistanceItem] then
				dl.interactiveHighlight;
      {$ENDC }
			if (cp.length <> 1) then
				comparePointsWithLineSegs(cq, cp, dl, d, ph);
			if (d > 0) & (cq.length <> 1) then
				comparePointsWithLineSegs(cp, cq, dl, d, ph);
			if (d > 0) & (cp.length > 1) & (cq.length > 1) & convexPolygonEdgesIntersect(cp, cq, ph) then begin
				d := 0;
				dl.p.setPoint(ph);
				dl.q.setPoint(ph);
			end;

			ph.free;
		end;

		simpleMinimalDistance := dl;

    {$IFC AnimationEnabled AND AConvexPolygonDistance }
		if animationFlag[AConvexPolygonDistanceItem] then begin
			cp.interactiveHighlight;
			cq.interactiveHighlight;
			dl.interactiveHighlight;
		end;
    {$ENDC }
	end;

(* -------------------------------------------------------------------------- *)

function convexPolygon.minimalDistance (cq: convexPolygon; var d: extended): lineSegment;
	var
		cp: convexPolygon;
		dl, ep, eq, epMinus, epPlus, eqMinus, eqPlus, lh, lhh: lineSegment;
		k, ps, qs, ph: point2d;
		i, j, imin, imax, jmin, jmax: longInt;
		lqk, lpk, dh: extended;
		stop: boolean;

  (* ------------------------------------------------------- *)

  {$IFC AnimationEnabled AND AConvexPolygonDistance }
	procedure drawLongLine (l, ll: lineSegment);
		var
			dx, dy, m, n: extended;
		begin
			dx := l.q.getX - l.p.getX;
			dy := l.q.getY - l.p.getY;
			if dx <> 0 then
				m := 1000 / dx
			else
				m := 1000;
			if dy <> 0 then
				n := 1000 / dy
			else
				n := 1000;
			if m < n then
				m := n;
			ll.p.setX(l.p.getX - m * dx);
			ll.p.setY(l.p.getY - m * dy);
			ll.q.setX(l.p.getX + m * dx);
			ll.q.setY(l.p.getY + m * dy);
			ll.interactiveOutput;
		end;
    {$ENDC }

  (* ------------------------------------------------------- *)

	procedure allocateMemory;
		begin

      (* --------------------- create some empty lineSegments --------------- *)

			new(ep);      (* One of the two edges that are compared in every step. *)
			failNil(Handle(ep));
			new(eq);       (* The other linesegment. *)
			failNil(Handle(eq));
			new(dl);       (* The linesegment that represents the minimum distance, *)
			failNil(Handle(dl));	(* will be filled at the end. *)
			new(lh);       (* Help line, used at several places where two existing *)
			failNil(Handle(lh));	(* points build a line. *)

      (* - create some complete points (with allocated memory for x and y)  - *)

			ps := point2d(self.getVertex(1).duplicate);
                               (* Projection point from ep.p or ep.q onto eq *)
			qs := point2d(self.getVertex(1).duplicate);
                               (* Projection point from eq.p or eq.q onto ep *)
			k := point2d(self.getVertex(1).duplicate);
                               (* Intersection point if ep and eq intersect  *)
			ph := point2d(self.getVertex(1).duplicate);
                               (* Help point, used in several routines       *)

      (* create empty lineSegments which are filled in initialize-procedure  *)

			new(epMinus);
			failNil(Handle(epMinus));
			new(epPlus);
			failNil(Handle(epPlus));
			new(eqMinus);
			failNil(Handle(eqMinus));
			new(eqPlus);
			failNil(Handle(eqPlus));
                   (* epMinus,...,eqPlus border (together with ep resp. eq)  *)
                   (* the 'quarter-plane', which is used to tell whether one *)
                   (* point is definitly inside a polygon or not.            *)

      (* - create a complete line (with allocated memory for the two points) *)

			new(lhh);              (* help line, used to draw long line.           *)
			failNil(Handle(lhh));
			lhh.p := point2d(self.getVertex(1).duplicate);
			lhh.q := point2d(self.getVertex(1).duplicate);
		end;

  (* ------------------------------------------------ *)

	procedure freeMemory;
		begin
			dispose(ep);
			dispose(eq);
			dispose(lh);
			k.freeAll;
			qs.freeAll;
			ps.freeAll;
			ph.freeAll;
			epMinus.freeAll;
			epPlus.freeAll;
			eqMinus.freeAll;
			eqPlus.freeAll;
			lhh.freeAll;
		end;

  (* ---------------------------------------------- *)

	function definitlyInside (p: point2d; l: lineSegment; lMinus, lPlus: lineSegment; var ps: point2d; var dh: extended): boolean;
		var
			leftToLines: boolean;
		begin
			dh := dPointLine(p, l, ps);
      {$IFC AnimationEnabled AND AConvexPolygonDistance }
			if animationFlag[AConvexPolygonDistanceItem] then begin
				lh.p := p;
				lh.q := ps;
				lh.interactiveOutput;
				p.interactiveHighlight;
			end;
      {$ENDC }

			if dh <= 0 then begin
        {$IFC AnimationEnabled AND AConvexPolygonDistance }
				if animationFlag[AConvexPolygonDistanceItem] then begin
					lMinus.interactiveOutput;
					lPlus.interactiveOutput;
					if not l.p.equalPoint(lMinus.q) then begin
						lh.p := lMinus.q;
						lh.q := l.p;
						lh.interactiveOutput;
					end;
					if not l.q.equalPoint(lPlus.p) then begin
						lh.p := l.q;
						lh.q := lPlus.p;
						lh.interactiveOutput;
					end;
					waitForClick(animationFlag[AConvexPolygonDistanceItem]);
					if userAbort then begin
						minimalDistance := nil;
						freeMemory;
						exit(minimalDistance)
					end;
					lMinus.interactiveOutput;
					lPlus.interactiveOutput;
					if not l.p.equalPoint(lMinus.q) then begin
						lh.p := lMinus.q;
						lh.q := l.p;
						lh.interactiveOutput;
					end;
					if not l.q.equalPoint(lPlus.p) then begin
						lh.p := l.q;
						lh.q := lPlus.p;
						lh.interactiveOutput;
					end;
					lh.p := p;
					lh.q := ps;
					lh.interactiveOutput;
					p.interactiveHighlight;
				end;
        {$ENDC }

				if l.p.equalPoint(lMinus.q) then
					leftToLines := pointLeftToLine(p, lMinus)
				else begin
					lh.p := lMinus.q;                        (* construct helpline *)
					lh.q := l.p;
					leftToLines := pointLeftToLine(p, lMinus) & pointLeftToLine(p, lh);
				end;
				if not l.q.equalPoint(lPlus.q) then        (* l is not! generated *)
					if l.q.equalPoint(lPlus.p) then
						leftToLines := leftToLines & pointLeftToLine(p, lPlus)
					else begin
						lh.p := l.q;                           (* construct helpline *)
						lh.q := lPlus.p;
						leftToLines := leftToLines & pointLeftToLine(p, lPlus) & pointLeftToLine(p, lh);
					end;
				definitlyInside := leftToLines;

			end
			else begin
        {$IFC AnimationEnabled AND AConvexPolygonDistance }
				if animationFlag[AConvexPolygonDistanceItem] then begin
					waitForClick(animationFlag[AConvexPolygonDistanceItem]);
					if userAbort then begin
						minimalDistance := nil;
						freeMemory;
						exit(minimalDistance)
					end;
					lh.p := p;
					lh.q := ps;
					lh.interactiveOutput;
					p.interactiveHighlight;
				end;
        {$ENDC }
				definitlyInside := FALSE;
			end;
		end;

  (* ------------------------------------------------ *)

	procedure quit (p: point2d);
		begin
			ps.setX(p.getX);
			ps.setY(p.getY);
			stop := TRUE;
		end;

  (* ------------------------------------------------------- *)

	procedure setImin (h: longInt);
		var
			ii: longInt;
		begin
      {$IFC AnimationEnabled AND AConvexPolygonDistance }
			if animationFlag[AConvexPolygonDistanceItem] then
				for ii := imin to h - 1 do begin
					lh.p := cp.getVertex(ii);
					lh.q := cp.getVertex(ii + 1);
					lh.interactiveHighlight;
				end;
			if animationFlag[AConvexPolygonDistanceItem] then
				drawLongLine(ep, lhh);
      {$ENDC }
			imin := h;
			i := (imin + imax) div 2;
			ep.p := cp.getVertex(i);
			ep.q := cp.getVertex(i + 1);
      {$IFC AnimationEnabled AND AConvexPolygonDistance }
			if animationFlag[AConvexPolygonDistanceItem] then
				drawLongLine(ep, lhh);
      {$ENDC }

		end;

  (* ----------------------------------------------- *)

	procedure setImax (h: longInt);
		var
			ii: longInt;
		begin
      {$IFC AnimationEnabled AND AConvexPolygonDistance }
			if animationFlag[AConvexPolygonDistanceItem] then
				for ii := h to imax - 1 do begin
					lh.p := cp.getVertex(ii);
					lh.q := cp.getVertex(ii + 1);
					lh.interactiveHighlight;
				end;
			if animationFlag[AConvexPolygonDistanceItem] then
				drawLongLine(ep, lhh);
      {$ENDC }
			imax := h;
			i := (imin + imax) div 2;
			ep.p := cp.getVertex(i);
			ep.q := cp.getVertex(i + 1);
      {$IFC AnimationEnabled AND AConvexPolygonDistance }
			if animationFlag[AConvexPolygonDistanceItem] then
				drawLongLine(ep, lhh);
      {$ENDC }
		end;

  (* --------------------------------------------- *)

	procedure setJmin (h: longInt);
		var
			jj: longInt;
		begin
      {$IFC AnimationEnabled AND AConvexPolygonDistance }
			if animationFlag[AConvexPolygonDistanceItem] then
				for jj := jmin to h - 1 do begin
					lh.p := cq.getVertex(jj);
					lh.q := cq.getVertex(jj + 1);
					lh.interactiveHighlight;
				end;
			if animationFlag[AConvexPolygonDistanceItem] then
				drawLongLine(eq, lhh);
      {$ENDC }
			jmin := h;
			j := (jmin + jmax) div 2;
			eq.p := cq.getVertex(j);
			eq.q := cq.getVertex(j + 1);
      {$IFC AnimationEnabled AND AConvexPolygonDistance }
			if animationFlag[AConvexPolygonDistanceItem] then
				drawLongLine(eq, lhh);
      {$ENDC }
		end;

  (* ----------------------------------------------- *)

	procedure setJmax (h: longInt);
		var
			jj: longInt;
		begin
      {$IFC AnimationEnabled AND AConvexPolygonDistance }
			if animationFlag[AConvexPolygonDistanceItem] then
				for jj := h to jmax - 1 do begin
					lh.p := cq.getVertex(jj);
					lh.q := cq.getVertex(jj + 1);
					lh.interactiveHighlight;
				end;
			if animationFlag[AConvexPolygonDistanceItem] then
				drawLongLine(eq, lhh);
      {$ENDC }
			jmax := h;
			j := (jmin + jmax) div 2;
			eq.p := cq.getVertex(j);
			eq.q := cq.getVertex(j + 1);
      {$IFC AnimationEnabled AND AConvexPolygonDistance }
			if animationFlag[AConvexPolygonDistanceItem] then
				drawLongLine(eq, lhh);
      {$ENDC }
		end;

  (* ---------------------------------------------- *)

	procedure initialize;
		var
			ii: longInt;
		begin
			if self.objectYLessThan(cq) then (* Compare the two polygons, and      *)
				cp := self                    (* figure out which one lies below    *)
			else begin                       (* the other.                         *)
				cp := cq;                      (* cp must lie below cq. If not then  *)
				cq := self;                    (* interchange.                       *)
			end;                             (* This has to be done to get the     *)
                                       (* correct quarter of each polygon.   *)

			if cq.getVertex(cq.bottom).xLessThan(cp.getVertex(cp.top)) then begin
                                 (* Now were we know that cp is below cq we  *)
				imin := cp.top;          (* have to figure out, if cp lies to the    *)
                                 (* right or to the left of cq.              *)
				if cp.left < cp.top then
					imax := cp.left + cp.length
				else
					imax := cp.left;
				jmin := cq.bottom;
				if cq.right < cq.bottom then
					jmax := cq.right + cq.length
                             (* It may be possible that the point 1 lies     *)
                             (* between min and max.                         *)
                             (* In order to establich min < max we have to   *)
                             (* add length to max.                           *)
				else
					jmax := cq.right;

        (* ---- Create the borderlines of the quarter that we consider ------ *)

				epMinus.p := point2d(cp.getVertex(cp.top).duplicate);
				epMinus.q := point2d(cp.getVertex(cp.top).duplicate);
				epMinus.p.setX(epMinus.p.getX + 1000);   (* usually +/- 1 is used    *)
                                                 (* to create normal vectors, *)
                                                 (* but here these lines are *)
                                                 (* drawn as long lines.     *)
				epPlus.p := point2d(cp.getVertex(cp.left).duplicate);
				epPlus.q := point2d(cp.getVertex(cp.left).duplicate);
				epPlus.q.setY(epPlus.q.getY - 1000);
				eqMinus.p := point2d(cq.getVertex(cq.bottom).duplicate);
				eqMinus.q := point2d(cq.getVertex(cq.bottom).duplicate);
				eqMinus.p.setX(eqMinus.p.getX - 1000);
				eqPlus.p := point2d(cq.getVertex(cq.right).duplicate);
				eqPlus.q := point2d(cq.getVertex(cq.right).duplicate);
				eqPlus.q.setY(eqPlus.q.getY + 1000);
			end
			else begin
				imin := cp.right;
				if cp.top < cp.right then
					imax := cp.top + cp.length
				else
					imax := cp.top;
				jmin := cq.left;
				if cq.bottom < cq.left then
					jmax := cq.bottom + cq.length
				else
					jmax := cq.bottom;
				epMinus.p := point2d(cp.getVertex(cp.right).duplicate);
				epMinus.q := point2d(cp.getVertex(cp.right).duplicate);
				epMinus.p.setY(epMinus.p.getY - 1000);
				epPlus.p := point2d(cp.getVertex(cp.top).duplicate);
				epPlus.q := point2d(cp.getVertex(cp.top).duplicate);
				epPlus.q.setX(epPlus.q.getX - 1000);
				eqMinus.p := point2d(cq.getVertex(cq.left).duplicate);
				eqMinus.q := point2d(cq.getVertex(cq.left).duplicate);
				eqMinus.p.setY(eqMinus.p.getY + 1000);
				eqPlus.p := point2d(cq.getVertex(cq.bottom).duplicate);
				eqPlus.q := point2d(cq.getVertex(cq.bottom).duplicate);
				eqPlus.q.setX(eqPlus.q.getX + 1000);
			end;


      (* ep and eq are the edges from point i (j) to point i+1 (j+1).        *)
      (* It may be, that the quarter, that we have to consider only consists *)
      (* of a single point. Then we construct a pseudo-edge in a way that we *)
      (* know that this edge works correct. This edge is one of the          *)
      (* borderlines of the quarter we consider (see above)                  *)

			i := (imin + imax) div 2;
			j := (jmin + jmax) div 2;
			ep.p := cp.getVertex(i);
			if imin < imax then
				ep.q := cp.getVertex(i + 1)
			else
				ep.q := epPlus.q;
			eq.p := cq.getVertex(j);
			if jmin < jmax then
				eq.q := cq.getVertex(j + 1)
			else
				eq.q := eqPlus.q;

			stop := FALSE;

      {$IFC AnimationEnabled AND AConvexPolygonDistance }
			if animationFlag[AConvexPolygonDistanceItem] then begin
				ps.setX((cp.getVertex(cp.right).getX + cp.getVertex(cp.left).getX) / 2);
				ps.setY((cp.getVertex(cp.bottom).getY + cp.getVertex(cp.top).getY) / 2);
{$IFC THINK_PASCAL }
				MoveTo(round(ps.getX), round(ps.getY));
{$ELSEC }
				MoveTo(roundLong(ps.getX), roundLong(ps.getY));
{$ENDC }
				DrawChar('P');
				ps.setX((cq.getVertex(cq.right).getX + cq.getVertex(cq.left).getX) / 2);
				ps.setY((cq.getVertex(cq.bottom).getY + cq.getVertex(cq.top).getY) / 2);
{$IFC THINK_PASCAL }
				MoveTo(round(ps.getX), round(ps.getY));
{$ELSEC }
				MoveTo(roundLong(ps.getX), roundLong(ps.getY));
{$ENDC }
				DrawChar('Q');
				cp.interactiveHighlight;
				cq.interactiveHighlight;
				cp.getVertex(imin).interactiveOutput;
				cp.getVertex(imax).interactiveHighlight;
				cq.getVertex(jmin).interactiveOutput;
				cq.getVertex(jmax).interactiveHighlight;
				with cp do
					for ii := imin to imax - 1 do begin
						lh.p := getVertex(ii);
						lh.q := getVertex(ii + 1);
						lh.interactiveHighlight;
					end;
				with cq do
					for ii := jmin to jmax - 1 do begin
						lh.p := getVertex(ii);
						lh.q := getVertex(ii + 1);
						lh.interactiveHighlight;
					end;
			end;
      {$ENDC }
		end;

    (* ------------------------------------------------- *)

	begin
		allocateMemory;
		initialize;

    {$IFC AnimationEnabled AND AConvexPolygonDistance }
		if animationFlag[AConvexPolygonDistanceItem] then begin
			drawLongLine(ep, lhh);
			drawLongLine(eq, lhh);
		end;
    {$ENDC }

		while ((imax - imin > 1) | (jmax - jmin > 1)) & not stop do begin

			if ep.lineIntersectsLine(eq, k) then begin
				lpk := lambda(k, ep);
				lqk := lambda(k, eq);
				if definitlyInside(ep.p, eq, eqMinus, eqPlus, ps, dh) then
					quit(ep.p)
				else if lqk >= 0 then
					if lpk >= 0 then
						if (lpk <= 1) & (lqk <= 1) then
							quit(k)
						else if dh > 0 then
							if imin < i then
								setImin(i)
							else
								setJmin(j)
						else if jmin < j then
							setJmin(j)
						else
							setImin(i)
					else if lambda(ps, eq) <= 0 then
						if imin < i then
							setImax(i)
						else
							setJmax(j)
					else if jmin < j then
						setJmin(j)
					else if definitlyInside(eq.q, ep, epMinus, epPlus, qs, dh) then
						quit(eq.q)
					else if lambda(qs, ep) <= 0 then
						setImax(i)
					else
						setImin(i)
				else if lpk < 0 then
					if dh > 0 then
						if imin < i then
							setImax(i)
						else
							setJmax(j)
					else if jmin < j then
						setJmax(j)
					else
						setImax(i)
				else if lambda(ps, eq) <= 0 then
					if jmin < j then
						setJmax(j)
					else if definitlyInside(eq.p, ep, epMinus, epPlus, qs, dh) then
						quit(eq.p)
					else if lambda(qs, ep) <= 0 then
						setImax(i)
					else
						setImin(i)
				else if imin < i then
					setImin(i)
				else if definitlyInside(ep.q, eq, eqMinus, eqPlus, ps, dh) then
					quit(ep.q)
				else if lambda(ps, eq) <= 0 then
					setJmax(j)
				else
					setJmin(j)
			end
			else if definitlyInside(ep.p, eq, eqMinus, eqPlus, ps, dh) then
				quit(ep.p)
			else if lambda(ps, eq) > 0 then begin
				if jmin < j then
					setJmin(j);
				if imin < i then
					setImin(i);
			end
			else begin
				if jmin < j then
					setJmax(j);
				if imin < i then
					setImax(i);
			end;
		end;

		if stop | definitlyInside(ep.p, eq, eqMinus, eqPlus, ps, dh) | definitlyInside(ep.q, eq, eqMinus, eqPlus, ps, dh) | definitlyInside(eq.p, ep, epMinus, epPlus, ps, dh) | definitlyInside(eq.q, ep, epMinus, epPlus, ps, dh) then begin
			dl.p := point2d(ps.duplicate);
			dl.q := point2d(ps.duplicate);
			d := 0;
      {$IFC AnimationEnabled AND AConvexPolygonDistance }
			if animationFlag[AConvexPolygonDistanceItem] then
				waitForClick(animationFlag[AConvexPolygonDistanceItem]);
      {$ENDC }
		end
		else begin
			dl.q := point2d(cp.getVertex(1).duplicate);
			if ep.q.equalPoint(epPlus.q) then begin  (* ep is equal to pseudo-line *)
				d := dPointLineseg(ep.p, eq, dl.q);    (*=> ep consisted of one point *)
				dl.p := point2d(ep.p.duplicate);
			end
			else if eq.q.equalPoint(eqPlus.q) then begin
				d := dPointLineseg(eq.p, ep, dl.q);   (* eq is equal to pseudo-line *)
				dl.p := point2d(eq.p.duplicate);      (*=> eq consisted of one point *)
			end
			else begin
				dl.p := point2d(cp.getVertex(1).duplicate);
				d := dLinesegLineseg(ep, eq, dl, ph);
			end;
			if d <= 0 then begin
				d := 0;   (* If hesse-distance from one point to a line was negative *)
				dl.q.setPoint(dl.p);(* d is set to 0 and dl consists of equal points. *)
			end;
		end;

    {$IFC AnimationEnabled AND AConvexPolygonDistance }
		if animationFlag[AConvexPolygonDistanceItem] then begin
			drawLongLine(ep, lhh);
			drawLongLine(eq, lhh);
			if d > 0 then
				dl.interactiveHighlight
			else
				dl.p.interactiveFlash;
			waitForClick(animationFlag[AConvexPolygonDistanceItem]);
		end;
    {$ENDC }

		minimalDistance := dl;
		freeMemory;
	end;

(* -------------------------------------------------------------------------- *)

function convexPolygon.simpleClosestPair: lineSegment;
	type
		referenceArrayH = ^referenceArrayP;
		referenceArrayP = ^referenceArray;
		referenceArray = array[seqIndex] of reference;
	var
		i, j: longInt;
		d: extended;
		cp, cq: convexPolygon;
		cpvLength: seqIndex;
		shortestConnection: lineSegment;
		cpList: vector;

  (* ----------------------------------------------------- *)

	procedure checkDistanceSimple (var cpp, cqq: convexPolygon; var shortestConnection: lineSegment; var delta: extended);
		var                                                   (* cp is below cq! *)
			dl: lineSegment;
			newDelta: extended;
		begin
			dl := cpp.simpleMinimalDistance(cqq, newDelta);
      {$IFC AnimationEnabled AND AClosestPairConvexPolygon }
			if animationFlag[AClosestPairItem] then begin
				if userAbort then begin
					shortestConnection.free;
{$IFC THINK_PASCAL }
					disposHandle(Handle(cpList));
{$ELSEC }
					DisposeHandle(Handle(cpList));
{$ENDC }
					simpleClosestPair := nil;
					exit(simpleClosestPair);
				end;
				dl.interactiveFlash;
				waitForClick(animationFlag[AClosestPairItem]);
				if userAbort then begin
					shortestConnection.free;
					cpList.free;
					simpleClosestPair := nil;
					exit(simpleClosestPair);
				end;
			end;
      {$ENDC }
			if newDelta < delta then begin  (* check distance between two polygons. *)
                                      (* If the distance is smaller, then we *)
				delta := newDelta;            (* have a new shortest Connection and  *)
                                      (* we have to free the memory.         *)
        {$IFC AnimationEnabled AND AClosestPairConvexPolygon }
				if animationFlag[AClosestPairItem] then
					shortestConnection.interactiveHighlight;
        {$ENDC }
				shortestConnection.freeAll;
				shortestConnection := dl;
        {$IFC AnimationEnabled AND AClosestPairConvexPolygon }
				if animationFlag[AClosestPairItem] then
					shortestConnection.interactiveHighlight;
        {$ENDC }
			end
			else
				dl.freeAll;                   (* If the new distance is not smaller *)
                                      (* then free memory of dl. *)
		end;

  (* ----------------------------------------------- *)

	procedure insertInList (cpp: convexPolygon);
		begin
			cpList.append(cpp);
      {$IFC AnimationEnabled AND AClosestPairConvexPolygon }
			if animationFlag[AClosestPairItem] then
				cpp.interactiveHighlight;
      {$ENDC }
			i := i + 1;
		end;

  (* ------------------------------------------------------- *)

	begin
		i := 1;
		cpvLength := currentArguments.length;
		cpList := newVector(cpvLength);
		currentArguments.forAll(insertInList);

		d := 1000000000;
		new(shortestConnection);
		failNil(Handle(shortestConnection));
		shortestConnection.p := point2d(convexPolygon(cpList.elements^^[1]).getVertex(1).duplicate);
		shortestConnection.p.setX(-100);
		shortestConnection.p.setY(-100);
		shortestConnection.q := point2d(shortestConnection.p.duplicate);
    {$IFC AnimationEnabled AND AClosestPairConvexPolygon }
		if animationFlag[AClosestPairItem] then
			shortestConnection.interactiveHighlight;
    {$ENDC }
		i := 1;
		while (d > 0) & (i < cpvLength) do begin
			cp := convexPolygon(cpList.elements^^[i]);
      {$IFC AnimationEnabled AND AClosestPairConvexPolygon }
			if animationFlag[AClosestPairItem] then
				cp.interactiveHighlight;
      {$ENDC }
			j := i + 1;
			while (d > 0) & (j <= cpvLength) do begin
				cq := convexPolygon(cpList.elements^^[j]);
        {$IFC AnimationEnabled AND AClosestPairConvexPolygon }
				if animationFlag[AClosestPairItem] then
					cq.interactiveHighlight;
        {$ENDC }
				checkDistanceSimple(cp, cq, shortestConnection, d);
        {$IFC AnimationEnabled AND AClosestPairConvexPolygon }
				if animationFlag[AClosestPairItem] then
					cq.interactiveHighlight;
        {$ENDC }
				j := j + 1;
			end;
      {$IFC AnimationEnabled AND AClosestPairConvexPolygon }
			if animationFlag[AClosestPairItem] then
				cp.interactiveHighlight;
      {$ENDC }
			i := i + 1;
		end;

		cpList.free;
		simpleClosestPair := shortestConnection;
	end;

(* -------------------------------------------------------------------------- *)

function convexPolygon.closestPairConvexPolygons (ask: boolean): lineSegment;
	var
		yTable: dictionary;
		dummy: priorityQueue;
		insertXQueue, deleteXQueue: vector;
		shortestConnection: lineSegment;
		current, tail: seqIndex1;
		cpvLength: seqIndex;
		delta: extended;

    {$IFC AnimationEnabled AND AConvexPolygonDistance }
		aFlag: boolean;
    {$ENDC }

    {$IFC AnimationEnabled AND AClosestPairConvexPolygon }
		deletedPolygon: convexPolygon;
		oldDelta: extended;
    {$ENDC }

  (* ----------------------------------------- *)

  {$IFC AnimationEnabled AND AClosestPairConvexPolygon }
	procedure drawStatus (p: point2d; delta: extended);
		begin
			drawThickVertical(p.getX);
			drawVertical(p.getX - delta);
		end;
  {$ENDC }

  (* ------------------------------------------ *)

	procedure insertInQueues (cp: convexPolygon);
		var
			insertH: cpcpInsertEvent;
			deleteH: cpcpDeleteEvent;
		begin
			cp.findExtremalPoints;
			new(deleteH);
			failNil(Handle(deleteH));
			deleteH.where := cp.getVertex(cp.right);
			deleteH.toDelete := nil;(* will be set when processing the InsertEvent *)
			deleteXQueue.append(deleteH);
			new(insertH);
			failNil(Handle(insertH));
			insertH.where := cp.getVertex(cp.left);
			insertH.cPolygon := cp;
			insertH.deleteEvent := deleteH;
			insertXQueue.append(insertH);
      {$IFC AnimationEnabled AND AClosestPairConvexPolygon }
			if animationFlag[AClosestPairItem] then
				cp.interactiveHighlight;
      {$ENDC }
		end;

  (* ------------------------------------- *)

	procedure buildAndSortTwoXQueues;
		begin
			currentArguments.forAll(insertInQueues);
			insertXQueue.sortAll(lessThanCpcpEvent);
			deleteXQueue.sortAll(lessThanCpcpEvent);
		end;

  (* -------------------------------------- *)

	begin
		cpvLength := currentArguments.length;

		if ask then begin
			if not getXY(0, false, dummy, cpvLength, yTable, false) then begin
				closestPairConvexPolygons := nil;
				exit(closestPairConvexPolygons);
			end;
		end
		else begin
			currentVectorLength := cpvLength;
			yTable := dictionary(newInitObj(avlTreeType));
		end;

		new(insertXQueue);
		failNil(Handle(insertXQueue));
		insertXQueue.init;
		new(deleteXQueue);
		failNil(Handle(deleteXQueue));
		deleteXQueue.init;
		buildAndSortTwoXQueues;

		new(shortestConnection);
		failNil(Handle(shortestConnection));
		shortestConnection.p := point2d(cpcpInsertEvent(insertXQueue.elements^^[1]).cPolygon.getVertex(1).duplicate);
		shortestConnection.p.setX(-100);  (* This is done, to make this line     *)
		shortestConnection.p.setY(-100);  (* invisible to the user.              *)
		shortestConnection.q := point2d(shortestConnection.p.duplicate);
    {$IFC AnimationEnabled AND AClosestPairConvexPolygon }
		if animationFlag[AClosestPairItem] then begin
			shortestConnection.interactiveHighlight;
		end;
    {$ENDC }
		delta := 100000000;

    {$IFC AnimationEnabled AND AConvexPolygonDistance }
		aFlag := animationFlag[AConvexPolygonDistanceItem];
		animationFlag[AConvexPolygonDistanceItem] := false;
             (* Disable animation of minimal distance algorithm in any case. *)
    {$ENDC }

		cpcpEvent(insertXQueue.elements^^[1]).processEvent(yTable, shortestConnection, delta);
		cpcpEvent(insertXQueue.elements^^[2]).processEvent(yTable, shortestConnection, delta);
		tail := 1;
		current := 3;

    {$IFC AnimationEnabled AND AClosestPairConvexPolygon }
		if animationFlag[AClosestPairItem] then begin
			oldDelta := delta;
			drawStatus(cpcpEvent(insertXQueue.elements^^[2]).where, delta);
		end;
    {$ENDC }


		notChecked := 0;
		reallyChecked := 0;

    (* ---------- Sweep through the plane. -------------------- *)

		while (current <= cpvLength) & (delta > 0) do begin

      {$IFC AnimationEnabled AND AClosestPairConvexPolygon }
			if animationFlag[AClosestPairItem] then begin
				drawStatus(cpcpEvent(insertXQueue.elements^^[current - 1]).where, delta);
				drawStatus(cpcpEvent(insertXQueue.elements^^[current]).where, delta);
			end;
      {$ENDC }

			while (cpcpEvent(deleteXQueue.elements^^[tail]).where.getX <= (cpcpEvent(insertXQueue.elements^^[current]).where.getX - delta)) do begin

        {$IFC AnimationEnabled AND AClosestPairConvexPolygon }
				if animationFlag[AClosestPairItem] then begin
					waitForClick(animationFlag[AClosestPairItem]);
					if userAbort then begin
						shortestConnection.free;
						insertXQueue.freeAll;
						deleteXQueue.freeAll;
						yTable.free;
						closestPairConvexPolygons := nil;
						exit(closestPairConvexPolygons);
					end;
					deletedPolygon := convexPolygon(yTable.getObject(cpcpDeleteEvent(deleteXQueue.elements^^[tail]).toDelete));
				end;
        {$ENDC }

				cpcpEvent(deleteXQueue.elements^^[tail]).processEvent(yTable, shortestConnection, delta);

        {$IFC AnimationEnabled AND AClosestPairConvexPolygon }
				if animationFlag[AClosestPairItem] then begin
					if (delta < oldDelta) then begin
						drawStatus(cpcpEvent(insertXQueue.elements^^[current]).where, oldDelta);
						drawStatus(cpcpEvent(insertXQueue.elements^^[current]).where, delta);
						oldDelta := delta;
					end;
					deletedPolygon.interactiveHighlight;
				end;
        {$ENDC }

				tail := tail + 1;
			end;

      {$IFC AnimationEnabled AND AClosestPairConvexPolygon }
			if animationFlag[AClosestPairItem] then begin
				waitForClick(animationFlag[AClosestPairItem]);
				if userAbort then begin
					shortestConnection.free;
					insertXQueue.freeAll;
					deleteXQueue.freeAll;
					yTable.free;
					closestPairConvexPolygons := nil;
					exit(closestPairConvexPolygons);
				end;
			end;
      {$ENDC }

			cpcpEvent(insertXQueue.elements^^[current]).processEvent(yTable, shortestConnection, delta);

      {$IFC AnimationEnabled AND AClosestPairConvexPolygon }
			if animationFlag[AClosestPairItem] then
				if (delta < oldDelta) then begin
					drawStatus(cpcpEvent(insertXQueue.elements^^[current]).where, oldDelta);
					drawStatus(cpcpEvent(insertXQueue.elements^^[current]).where, delta);
					oldDelta := delta;
				end;
      {$ENDC }

			current := current + 1;
		end; (* of sweep *)

    {$IFC AnimationEnabled AND AClosestPairConvexPolygon }
		if animationFlag[AClosestPairItem] then begin
			drawStatus(cpcpEvent(insertXQueue.elements^^[current - 1]).where, delta);
		end;
    {$ENDC }

   (* -------- Delete remaining elements in y-table ------------------- *)

		while (tail <= cpvLength - 2) & (delta > 0) do begin
      {$IFC AnimationEnabled AND AClosestPairConvexPolygon }
			if animationFlag[AClosestPairItem] then begin
				waitForClick(animationFlag[AClosestPairItem]);
				if userAbort then begin
					shortestConnection.free;
					insertXQueue.freeAll;
					deleteXQueue.freeAll;
					yTable.free;
					closestPairConvexPolygons := nil;
					exit(closestPairConvexPolygons);
				end;
				deletedPolygon := convexPolygon(yTable.getObject(cpcpDeleteEvent(deleteXQueue.elements^^[tail]).toDelete));
			end;
     {$ENDC }

			cpcpEvent(deleteXQueue.elements^^[tail]).processEvent(yTable, shortestConnection, delta);

      {$IFC AnimationEnabled AND AClosestPairConvexPolygon }
			if animationFlag[AClosestPairItem] then begin
				deletedPolygon.interactiveHighlight;
			end;
      {$ENDC }
			tail := tail + 1;
		end;

		insertXQueue.freeAll;
		deleteXQueue.freeAll;
		if yTable.infoString <> '' then begin
			displayInfoString(StringOf('Y-Table: ', yTable.infoString, '; Checks: (T=', reallyChecked + notChecked : 1, ', C=', reallyChecked : 1, ')'));
		end;
		yTable.free;

    {$IFC AnimationEnabled AND AConvexPolygonDistance }
		animationFlag[AConvexPolygonDistanceItem] := aFlag;
    {$ENDC }

		closestPairConvexPolygons := shortestConnection;
	end;

(* -------------------------------------------------------------------------- *)

function convexPolygon.getVertex (i: longInt): point2d;
	begin
		getVertex := point2d(elements^^[((length + i - 1) mod length) + 1]);
    (* Changed by JM on 25.6.93.   See Function-Calls in UNIT ConvexPolygon.  There the correct order was established *)

    (* getVertex := point2d(elements^^[length - ((length + i - 1) mod length)]); *)
    (* -> Transform from clockwise to counterclockwise order ! Not used anymore ! *)

	end;

(* -------------------------------------------------------------------------- *)

end.
