(*	XYZ GeoBench (eXperimental geometrY Zurich):		*)
(*	An object oriented, interactive workbench for		*)
(*	experimental geometric computation.					*)
(*														*)
(*	 1999 by Peter Schorn, peter.schorn@acm.org		*)
(*	Culmannstrasse 77, CH-8006 Zrich, Switzerland		*)
(*	This software may be copied freely for educational	*)
(*	and research purposes only. For commercial use,		*)
(*	please contact the author.							*)
(*														*)
(*	Version 5.0.5: 30-Sep-1999							*)

unit closestPairDI;

interface

uses
	pointVector;

implementation

{$SETC AclosestPair = TRUE }

const
	infinity = maxint;

procedure MergeSort (var A, work: pointVector; N: integer);
(* 	    A non-recursive, non-stack version of mergesort based on bottom-up		*)
(*     divide-and-conquer schema.														*)
(*																								*)
(*     Procedure cannot, as a matter of principle, know how long an array		*)
(*     to declare for temporary storage.  Standard solution is to have user		*)
(*     pass in work array.																	*)

	var
		span: integer; (* current length of sorted lists *)
		inA: boolean;
		i: integer;

	procedure Merge (var inarray,	(* array containing lists of length  span  					*)
									outarray: pointVector;	(* array containing lists of length 2*span  upon return	*)
									N, 							(* total number of elements 									*)
									span: integer);
(*       Merge pairs of sorted lists of length  span.  Merge operation is	*)
(*       performed slightly more efficiently than in Figure				*)

		label
			99;
		var
			i,		(* points to first unprocessed element in sublist 1	*)
			j,		(* points to first unprocessed element in sublist 2	*)
			k,		(* points to first free space in  outarray				*)
			t,		(* location past first sublist/start of second sublist	*)
			u:		(* location past second sublist							*)
			integer;
		begin
      (* There are at least two lists to merge. *)
			i := 1;
			j := span + 1;
			k := 1;

{$IFC AnimationEnabled }
			advanceStopWatchHands;
{$ENDC }
			repeat		(* until there is one left over list or no lists left at all	*)
        				(* invariant: k = i at start of merge of two lists			*)
				t := j;		(* set upper bounds for i and j								*)
				u := j + span;
				if u > N then
					u := N + 1; (* second sublist is small *)

        (* perform merge *)
99:
				if point2d(inarray.elements^^[i]).getX <= point2d(inarray.elements^^[j]).getX then begin
					outarray.elements^^[k] := inarray.elements^^[i];
					i := i + 1;
					k := k + 1;
					if i < t then		(* only check cursor that changed	*)
						goto 99
					else
						repeat			(* copy remainder of sublist 2		*)
                          		(* k = j at this point					*)
							outarray.elements^^[k] := inarray.elements^^[k];
							k := k + 1
						until k = u
				end
				else begin
					outarray.elements^^[k] := inarray.elements^^[j];
					j := j + 1;
					k := k + 1;
					if j < u then		(* only check cursor that changed	*)
						goto 99
					else
						repeat			(* copy remainder of sublist 1		*)
							outarray.elements^^[k] := inarray.elements^^[i];
							i := i + 1;
							k := k + 1
						until k = u
				end;
				i := u;
				j := i + span
			until j > N; 			(* until sublist 2 is non-existent		*)

			while i <= N do	(* copy over an unmatched list if one exists *)
				begin
				outarray.elements^^[i] := inarray.elements^^[i];
				i := i + 1
			end
		end; (* Merge *)

	begin (* Mergesort *)
		inA := true;
		span := 1; (* The array consists of N sorted sublists of length one. *)
		while span < N do (* while cannot guarantee array is sorted *)
			begin
			if inA then
				Merge(A, work, N, span)
			else
				Merge(work, A, N, span);
			inA := not inA;
			span := 2 * span
		end;
		if not inA then
			for i := 1 to N do
				A.elements^^[i] := work.elements^^[i]
	end; (* Mergesort *)


	type
		SP = object(obj) (* subproblem description *)
              (* All points within d of the boundaries of the vertical strip			*)
              (* defined by [MinX,MaxX] are in A[start..stop], sorted by their	*)
              (* ordinates.  Because d is monotone nonincreasing, the				*)
              (* implication is only one way, i.e., some of the points in				*)
              (* A[start..stop] may no longer be within d of the boundary. 		*)
				start, stop: integer;
				MinX, MaxX: real;
			end;

procedure ClosestPairDandI (A: pointVector;	(* by value, so as not to destroy the data	*)
								n: integer; var p1, p2: point2d;			(* two closest points							*)
								var d: real									(* their distance apart							*)
								);
  (* Find the closest pair from among n points in O(n log n) time.	*)
  (* p1, p2 and d are global variables. 									*)
	var
		work: pointVector;
		SubProbs: homogenVector; (* array[1..MaxPoints] of SP; *)
		subProblem: SP;
		i, j, k: integer;
{$IFC AnimationEnabled AND AclosestPair }
		flashVector: homogenVector;
{$ENDC }

	procedure Merge (SP1, SP2: SP; var NewSP: SP);
		label
			1;
		var
			i, i1, i2, j, k: integer;
		begin
      (* The dividing line is really a no-man's-land between SP1.MaxX				*)
      (* and SP2.MinX. 																		*)
      (* While respecting the ordering on ordinates, scan both lists, placing		*)
      (* the points in SP1 within d of SP2.MinX and the points in SP2 within d		*)
      (* of SP1.MaxX at the front of the work array.									*)
{$IFC AnimationEnabled }
			if userAbort then begin
				SubProbs.freeAll;
				work.free;
				exit(ClosestPairDandI);
			end;
{$ENDC }
{$IFC AnimationEnabled AND AclosestPair }
			if animationFlag[AClosestPairItem] then begin
				drawThickVertical(SP1.MinX);
				if SP1.MinX <> SP1.MaxX then
					drawThickVertical(SP1.MaxX);
				drawVertical(SP2.MinX);
				if SP2.MinX <> SP2.MaxX then
					drawVertical(SP2.MaxX);
				if (SP1.MinX <> SP1.MaxX) & (SP1.MinX <> SP1.MaxX) then
					waitForClick(animationFlag[AClosestPairItem]);
			end;
{$ENDC }
			i1 := SP1.start;
			i2 := SP2.start;
			k := 0;
			while (i1 <= SP1.stop) & (i2 <= SP2.stop) do
				if point2d(A.elements^^[i1]).getY <= point2d(A.elements^^[i2]).getY then begin
					if point2d(A.elements^^[i1]).getX >= SP2.MinX - d then begin
						k := k + 1;
						work.elements^^[k] := A.elements^^[i1]
					end;
					i1 := i1 + 1
				end
				else begin
					if point2d(A.elements^^[i2]).getX <= SP1.MaxX + d then begin
						k := k + 1;
						work.elements^^[k] := A.elements^^[i2]
					end;
					i2 := i2 + 1
				end;
			while i1 <= SP1.stop do (* Copy any remaining relevant elements. *)
				begin
				if point2d(A.elements^^[i1]).getX >= SP2.MinX - d then begin
					k := k + 1;
					work.elements^^[k] := A.elements^^[i1];
				end;
				i1 := i1 + 1
			end;
			while i2 <= SP2.stop do (* Copy any remaining relevant elements. *)
				begin
				if point2d(A.elements^^[i2]).getX <= SP1.MaxX + d then begin
					k := k + 1;
					work.elements^^[k] := A.elements^^[i2];
				end;
				i2 := i2 + 1
			end;

      (* Scan the work array, updating global variables p1, p2 and d.	*)
      (* This takes constant time per point. 									*)
			i := 1;
			while i < k do begin
				for j := i + 1 to k do begin
{$IFC AnimationEnabled AND AclosestPair }
					if animationFlag[AClosestPairItem] then begin
						flashVector.elements^^[1] := work.elements^^[j];
						flashVector.elements^^[2] := work.elements^^[i];
						flashVector.interactiveFlash;
					end;
{$ENDC }
					if point2d(work.elements^^[j]).getY - point2d(work.elements^^[i]).getY > d then
						goto 1;
					if point2d(work.elements^^[i]).distanceX(point2d(work.elements^^[j])) < d then begin
{$IFC AnimationEnabled AND AclosestPair }
						if animationFlag[AClosestPairItem] & (d <> infinity) then
							p1.drawLine(p2);
{$ENDC }
						p1 := point2d(work.elements^^[i]);
						p2 := point2d(work.elements^^[j]);
						d := p1.distanceX(p2);
{$IFC AnimationEnabled AND AclosestPair }
						if animationFlag[AClosestPairItem] then begin
							p1.drawLine(p2);
							waitForClick(animationFlag[AClosestPairItem]);
						end;
{$ENDC }
					end;
				end;
1:
				i := i + 1
			end;

      (* Merge the two sorted lists, retaining only points within d of the *)
      (* boundaries of the new vertical strip. *)
			i1 := SP1.start;
			i2 := SP2.start;
			k := 0;
			j := n + 1;
			while (i1 <= SP1.stop) & (i2 <= SP2.stop) do
				if point2d(A.elements^^[i1]).getY <= point2d(A.elements^^[i2]).getY then begin
					if (point2d(A.elements^^[i1]).getX <= SP1.MinX + d) | (point2d(A.elements^^[i1]).getX >= SP2.MaxX - d) then begin
						k := k + 1;
						work.elements^^[k] := A.elements^^[i1]
					end;
					i1 := i1 + 1
				end
				else begin
					if (point2d(A.elements^^[i2]).getX <= SP1.MinX + d) | (point2d(A.elements^^[i2]).getX >= SP2.MaxX - d) then begin
						k := k + 1;
						work.elements^^[k] := A.elements^^[i2]
					end;
					i2 := i2 + 1
				end;
			while i1 <= SP1.stop do (* Copy any remaining relevant elements. *)
				begin
				if (point2d(A.elements^^[i1]).getX <= SP1.MinX + d) | (point2d(A.elements^^[i1]).getX >= SP2.MaxX - d) then begin
					k := k + 1;
					work.elements^^[k] := A.elements^^[i1]
				end;
				i1 := i1 + 1
			end;
			while i2 <= SP2.stop do (* Copy any remaining relevant elements. *)
				begin
				if (point2d(A.elements^^[i2]).getX <= SP1.MinX + d) | (point2d(A.elements^^[i2]).getX >= SP2.MaxX - d) then begin
					k := k + 1;
					work.elements^^[k] := A.elements^^[i2]
				end;
				i2 := i2 + 1
			end;
			i1 := SP1.start;
			for i := 1 to k do begin
				A.elements^^[i1] := work.elements^^[i];
				i1 := i1 + 1
			end;
			for i := j to n do begin
				A.elements^^[i1] := work.elements^^[i];
				i1 := i1 + 1
			end;
{$IFC AnimationEnabled AND AclosestPair }
			if animationFlag[AClosestPairItem] then begin
				drawThickVertical(SP1.MinX);
				if SP1.MinX <> SP1.MaxX then
					drawThickVertical(SP1.MaxX);
				drawVertical(SP2.MinX);
				if SP2.MinX <> SP2.MaxX then
					drawVertical(SP2.MaxX);
			end;
{$ENDC }
			NewSP.start := SP1.start;
			NewSP.stop := i1 - 1;
			NewSP.MinX := SP1.MinX;
			NewSP.MaxX := SP2.MaxX;
		end; (* Merge *)

	procedure assignSubProblem (a, b: SP);
		begin
			a.start := b.start;
			a.stop := b.stop;
			a.MinX := b.MinX;
			a.MaxX := b.MaxX;
		end;

	begin (* ClosestPair *)
{$IFC AnimationEnabled AND AclosestPair }
		flashVector := newHomogenVector(2);
		flashVector.length := 2;
{$ENDC }
		work := newPointVector(n);
		MergeSort(A, work, n); (* Sort the points by their abscissae. *)
    (* Define the initial subproblems. *)
		d := infinity;
		SubProbs := newHomogenVector(n);
		for i := 1 to n do begin
			new(subProblem);
			failNil(Handle(subProblem));
			with subProblem do begin
				start := i;
				stop := i;
				MinX := point2d(A.elements^^[i]).getX;
				MaxX := point2d(A.elements^^[i]).getX
			end;
			SubProbs.elements^^[i] := subProblem;
		end;
		SubProbs.length := n;
		i := n;
		while i > 1 do begin
			j := 1;
			k := 1;
			while j + 1 <= i do begin
				Merge(SP(SubProbs.elements^^[j]), SP(SubProbs.elements^^[j + 1]), SP(SubProbs.elements^^[k]));
				j := j + 2;
				k := k + 1
			end;
			if j = i then begin
				assignSubProblem(SP(SubProbs.elements^^[k]), SP(SubProbs.elements^^[j]));
				k := k + 1
			end;
			i := k - 1
		end;
		SubProbs.freeAll;
		work.free;
{$IFC AnimationEnabled AND AclosestPair }
		flashVector.free;
{$ENDC }
	end; (* ClosestPair *)

function pointVector.closestPairDivideAndConquer: lineSegment;
	var
		p1, p2: point2d;
		d: real;
		ls: lineSegment;
		pv: pointVector;
	begin
		pv := pointVector(duplicate);
		ClosestPairDandI(pv, pv.length, p1, p2, d);
		pv.free;
{$IFC AnimationEnabled }
		if userAbort then
			closestPairDivideAndConquer := nil
		else begin
			new(ls);
			failNil(Handle(ls));
			ls.p := point2d(p1.duplicate);
			ls.q := point2d(p2.duplicate);
			closestPairDivideAndConquer := ls;
		end;
{$ELSEC }
		new(ls);
		failNil(Handle(ls));
		ls.p := point2d(p1.duplicate);
		ls.q := point2d(p2.duplicate);
		closestPairDivideAndConquer := ls;
{$ENDC }
	end;


end.
