(*	XYZ GeoBench (eXperimental geometrY Zurich):		*)
(*	An object oriented, interactive workbench for		*)
(*	experimental geometric computation.					*)
(*														*)
(*	 1999 by Peter Schorn, peter.schorn@acm.org		*)
(*	Culmannstrasse 77, CH-8006 Zrich, Switzerland		*)
(*	This software may be copied freely for educational	*)
(*	and research purposes only. For commercial use,		*)
(*	please contact the author.							*)
(*														*)
(*	Version 5.0.5: 30-Sep-1999							*)

unit convexHull;

interface

uses
	pointVector;

implementation

{$SETC AconvexHullScan = TRUE }
{$SETC AconvexHullDI = TRUE }
{$SETC AconvexTangent= TRUE }

function pointVector.convertToConvexPolygon: obj;
	begin
		convertToConvexPolygon := convexHullGrahamScan;
	end;

function sequence.convertToPolygon: obj;
	var
		result: polygon2d;

	procedure doAppend (x: obj);
		begin
			result.append(x.duplicate);
		end;

	begin
		if (length > 0) & sameType & member(ith(1), point2d) then begin
			result := newPolygon2d(length);
			forAll(doAppend);
			convertToPolygon := result;
		end
		else
			convertToPolygon := nil;
	end;

function sequence.convertToConvexPolygon: obj;
	var
		poly: polygon2d;
	begin
		poly := polygon2d(convertToPolygon);
		if poly = nil then
			convertToConvexPolygon := nil
		else begin
			convertToConvexPolygon := poly.convertToConvexPolygon;
			poly.freeAll;
		end;
	end;

function sequence.convertToPolygonPossible: boolean;
	begin
		convertToPolygonPossible := true;
	end;

function pointVector.convertToPolygon: obj;
	var
		p: polygon2d;
		ps: pointVector;
	begin
		new(p);
		failNil(Handle(p));
		ps := pointVector(duplicateAll);
		p.length := ps.length;
		p.elements := ps.elements;
		dispose(ps);
		convertToPolygon := p;
	end;

function polygon2d.convertToPolygon: obj;
	begin
		convertToPolygon := duplicateAll;
	end;

function convexPolygon.convexHullGrahamScan: convexPolygon;
	begin
		convexHullGrahamScan := convexPolygon(duplicateAll);
	end;

function pointVector.convexHullGrahamScan: convexPolygon;
	var
		p: pointVector;
		result: convexPolygon;
		i, hullLength, eliminated: seqIndex0;
	begin
		p := pointVector(duplicate);
		p.convexHull(true, false, hullLength, eliminated);
		if userAbort then
			result := nil
		else begin
			new(result);
			failNil(Handle(result));
			currentVectorLength := hullLength;
			result.init;
			for i := 1 to hullLength do
				result.elements^^[i] := p.elements^^[i].duplicate;
			result.length := hullLength;
		end;
		p.free;
		convexHullGrahamScan := result;
	end;

procedure writePoint (x: obj);
	begin
		x.textFileOutput;
		writeSimpleCR;
	end;

procedure pointVector.textFileOutput;
	begin
		case currentTextFileOutputMethod of
			1:  begin
				writeRawString(textFileOutputId);
				writeSimpleCR;
				forAll(writePoint);
			end;
			2:  begin
				writeTextLongInt(length);
				writeSimpleCR;
				forAll(writePoint);
			end;
			otherwise
				inherited textFileOutput;
		end;
	end;

function pointVector.textFileOutputId: Str255;
	begin
		textFileOutputId := 'PointVector';
	end;

function polyLine.textFileOutputId: Str255;
	begin
		textFileOutputId := 'PolyLine';
	end;

function polygon2d.textFileOutputId: Str255;
	begin
		textFileOutputId := 'Polygon';
	end;

function convexPolygon.textFileOutputId: Str255;
	begin
		textFileOutputId := 'ConvexPolygon';
	end;

function polygon2d.removeRepeatedElements: sequence;
	var
		result: vector;
	begin
		result := vector(inherited removeRepeatedElements);
		if (result.length > 1) & result.elements^^[1].equal(result.elements^^[result.length]) then begin
			result.elements^^[result.length].freeAll;
			result.length := result.length - 1;
		end;
		removeRepeatedElements := result;
	end;

function sequence.createGraph (resultType: objectType): obj;
	var
		result: simpleUndirectedGraph;

	procedure addVertexAndEdge (x: obj);
		var
			l: seqIndex;
		begin
			result.vertices.append(x.duplicateAll);
			l := result.vertices.length;
			if l > 1 then
				result.addNewEdge(l - 1, l);
		end;

	begin
		currentVertices := length;
		currentEdges := length;
		result := simpleUndirectedGraph(newInitObj(resultType));
		if member(result, directedGraph) then begin
			result.vertices.length := length;
			directedGraph(result).initAdjacency;
			result.vertices.length := 0;
		end;
		forAll(addVertexAndEdge);
		createGraph := result;
	end;

function polygon2d.createGraph (resultType: objectType): obj;
	var
		result: simpleUndirectedGraph;
	begin
		result := simpleUndirectedGraph(inherited createGraph(resultType));
		if result.vertices.length > 0 then
			result.addNewEdge(result.vertices.length, 1);
		createGraph := result;
	end;

function classifyAngleSign (angleSign: signType): polygonKind;
	begin
		case angleSign of
			-1: 
				classifyAngleSign := polygonIsConvexMinus;
			0: 
				classifyAngleSign := polygonIsConvexDegenerate;
			+1: 
				classifyAngleSign := polygonIsConvexPlus;
		end;
	end;

function testConvex (function getPoint: point2d): polygonKind;
	var
		first, second, third, saveFirst, saveSecond: point2d;
		angleSenseSign, thisSign: signType;
		currentDirection, thisDirection: relationObj;
		directionChanges: longInt;

	function getDifferentPoint (previous: point2d; var next: point2d): boolean;
		begin
			repeat
				next := getPoint;
			until (next = nil) | not next.equal(previous);
			getDifferentPoint := next <> nil;
		end;

	procedure notConvexIf (assertion: boolean);
		begin
			if assertion then begin
				testConvex := polygonIsNotConvex;
				exit(testConvex);
			end;
		end;

	procedure checkTriple;
		begin
			thisDirection := comparePointXY(second, third);
			if currentDirection = oppositeRelationObj(thisDirection) then
				directionChanges := directionChanges + 1;
			currentDirection := thisDirection;
			thisSign := third.whichSideSign(second, first);
			if thisSign <> 0 then begin
				notConvexIf(angleSenseSign = -thisSign);
				angleSenseSign := thisSign;
			end;
			first := second;
			second := third;
		end;

	begin
		first := getPoint;
		if (first = nil) | not getDifferentPoint(first, second) then begin
			testConvex := polygonIsConvexDegenerate;
			exit(testConvex);
		end;
		saveFirst := first;
		saveSecond := second;
		angleSenseSign := 0;
		directionChanges := 0;
		currentDirection := comparePointXY(first, second);
		while getDifferentPoint(second, third) do
			checkTriple;
		if not second.equal(saveFirst) then begin
			third := saveFirst;
			checkTriple;
		end;
		third := saveSecond;
		checkTriple;
		if directionChanges > 2 then
			if angleSenseSign = 0 then
				testConvex := polygonIsNotConvexDegenerate
			else
				testConvex := polygonIsNotConvex
		else
			testConvex := classifyAngleSign(angleSenseSign);
	end;

function polygon2d.isConvex (var count: seqIndex1): polygonKind;

	function getPoint: point2d;
		begin
			if count > length then
				getPoint := nil
			else begin
				getPoint := point2d(elements^^[count]);
				count := count + 1;
			end;
		end;

	begin
		count := 1;
		isConvex := testConvex(getPoint);
	end;

function convexPolygon.isConvex (var count: seqIndex1): polygonKind;
	var
		angleSign: signType;
	begin
		count := length + 1;
		if length > 2 then begin
			angleSign := point2d(ith(2)).whichSideSign(point2d(ith(1)), point2d(ith(length)));
			if angleSign = 0 then begin
				elements^^[length + 1] := elements^^[1];
				count := 2;
				while (angleSign = 0) & (count <= length) do begin
					angleSign := point2d(ith(count + 1)).whichSideSign(point2d(ith(count)), point2d(ith(count - 1)));
					count := count + 1;
				end;
				count := count - 1;
			end;
			isConvex := classifyAngleSign(angleSign);
		end
		else
			isConvex := polygonIsConvexDegenerate;
	end;

procedure convexHullInternal (var ps: pointVector);
	var
		a, b: convexPolygon;
		upperA, upperB, lowerA, lowerB: seqIndex0;

	procedure hull (l, r: seqIndex; var k: seqIndex0);
		var
			k1, k2, i: seqIndex0;
			m: seqIndex;

		procedure append (r: point2d);
			var
				p, q: point2d;
			begin
				if (k < l) | not r.equalPoint(point2d(ps.elements^^[k])) then
					if k > l then begin
						p := point2d(ps.elements^^[k - 1]);
						q := point2d(ps.elements^^[k]);
						if sign(p.whichSideX(q, r)) = 0 then begin
							if pointLessThanXY(p, q) & pointLessThanXY(q, r) | pointLessThanXY(r, q) & pointLessThanXY(q, p) then
								ps.elements^^[k] := r;
						end
						else begin
							k := k + 1;
							ps.elements^^[k] := r;
						end;
					end
					else begin
						k := k + 1;
						ps.elements^^[k] := r;
					end;
			end;

{$IFC NOT THINK_PASCAL }
{$PRAGMAC global_optimizer on }
{$ENDC }
		begin
			if (r = l) | point2d(ps.elements^^[l]).equalPoint(point2d(ps.elements^^[r])) then
				k := l
			else if r = l + 1 then
				k := r
			else begin
				m := (l + r) div 2;
				hull(l, m, k1);

{$IFC AnimationEnabled AND AconvexHullDI }
				if animationFlag[AConvexHullItem] then begin
					for i := l to k1 do
						a.elements^^[i - l + 1] := ps.elements^^[i];
					a.length := k1 - l + 1;
					a.interactiveOutput;
					waitForClick(animationFlag[AConvexHullItem]);
				end;
{$ENDC }

				hull(m + 1, r, k2);
{$IFC AnimationEnabled AND AconvexHullDI }
				if animationFlag[AConvexHullItem] then begin
					for i := m + 1 to k2 do
						b.elements^^[i - m] := ps.elements^^[i];
					b.length := k2 - m;
					b.interactiveOutput;
					waitForClick(animationFlag[AConvexHullItem]);
				end;
{$ENDC }

				for i := l to k1 do
					a.elements^^[i - l + 1] := ps.elements^^[i];
				a.length := k1 - l + 1;
				for i := m + 1 to k2 do
					b.elements^^[i - m] := ps.elements^^[i];
				b.length := k2 - m;

{$IFC AnimationEnabled AND AconvexHullDI  AND AconvexTangent }
				if animationFlag[AConvexHullItem] & animationFlag[ATangentItem] then begin
					a.interactiveHighlight;
					b.interactiveHighlight;
				end;
{$ENDC }

{$IFC AnimationEnabled }
				advanceStopWatchHands;
{$ENDC }
				a.tangent(b, upperA, upperB, lowerA, lowerB);
				assert(upperA > 0, 'Tangent exists.');
{$IFC AnimationEnabled }
				if userAbort then begin
					b.free;
					a.free;
					exit(convexHullInternal);
				end;
{$ENDC }

{$IFC AnimationEnabled AND AconvexHullDI }
				if animationFlag[AConvexHullItem] then begin
{$IFC AnimationEnabled AND AconvexTangent }
					if animationFlag[ATangentItem] then begin
						point2d(a.elements^^[upperA]).drawLine(point2d(b.elements^^[upperB]));
						point2d(a.elements^^[lowerA]).drawLine(point2d(b.elements^^[lowerB]));
						a.interactiveHighlight;
						b.interactiveHighlight;
					end;
{$ENDC }
					a.interactiveOutput;
					b.interactiveOutput;
				end;
{$ENDC }

				k := l - 1;
				for i := 1 to upperA do
					append(point2d(a.elements^^[i]));
				if upperB > lowerB then
					if lowerB = 1 then begin
						lowerB := b.length + 1;
						b.elements^^[lowerB] := b.elements^^[1];
					end
					else
						assert(false, stringOf('hull ok (', upperB : 1, ':', lowerB : 1, ')'));
				for i := upperB to lowerB do
					append(point2d(b.elements^^[i]));
				if lowerA > 1 then
					for i := lowerA to a.length do
						append(point2d(a.elements^^[i]));
				if (k > l + 1) & (sign(point2d(ps.elements^^[k - 1]).whichSideX(point2d(ps.elements^^[k]), point2d(ps.elements^^[l]))) = 0) then
					k := k - 1;
			end;
		end;

{$IFC NOT THINK_PASCAL }
{$PRAGMAC global_optimizer off }
{$ENDC }

	begin
		currentVectorLength := ps.length;
		new(a);
		failNil(Handle(a));
		a.init;
		currentVectorLength := ps.length;
		new(b);
		failNil(Handle(b));
		b.init;
		if ps.length > 0 then begin
			ps.sortAll(pointLessThanXY);
			hull(1, ps.length, ps.length);
		end;
		b.free;
		a.free;
	end;

function pointVector.convexHullDivideAndConquer: convexPolygon;
	var
		t: pointVector;
		result: convexPolygon;
	begin
		new(result);
		failNil(Handle(result));
		t := pointVector(duplicate);
		result.length := t.length;
		result.elements := t.elements;
		dispose(t);
		convexHullInternal(pointVector(result));
{$IFC AnimationEnabled }
		if userAbort then begin
			convexHullDivideAndConquer := nil;
			result.free;
			exit(convexHullDivideAndConquer);
		end;
{$ENDC }
		convexHullDivideAndConquer := convexPolygon(result.duplicateAll);
		result.free;
	end;

procedure convexHullScan (var p: vector; var n: seqIndex0);
	var
		upper, lower: seqIndex;
		endHull: seqIndex0;
{$IFC AnimationEnabled AND AconvexHullScan }
		hullSegments: list;
		ls: lineSegment;
{$ENDC }

	procedure partition (p: vector; var upper, lower: seqIndex);
		var
			i: seqIndex;
			l, r: point2d;
		begin
			l := point2d(p.elements^^[1]);
			r := point2d(p.elements^^[p.length]);
			upper := 1;
			lower := p.length;
			i := 2;
			while i < lower do begin
{$IFC AnimationEnabled }
				if userAbort then begin
					hullSegments.freeAll;
					exit(convexHullScan);
				end;
{$ENDC }
				if l.whichSideX(r, point2d(p.elements^^[i])) <= 0 then begin
					upper := upper + 1;
					p.swap(upper, i);
					i := i + 1;
				end
				else begin
					lower := lower - 1;
					p.swap(lower, i);
				end;
			end;
		end; (* partition *)

	procedure hull (from: seqIndex1; upto: seqIndex; var endHull: seqIndex0);
		var
			startHull: seqIndex;
{$IFC AnimationEnabled AND AconvexHullScan }
			oldEndHull: seqIndex0;
			oldFrom: seqIndex1;
{$ENDC }
		begin
{$IFC AnimationEnabled AND AconvexHullScan }
			oldEndHull := endHull;
			oldFrom := from;
{$ENDC }
			startHull := endHull + 1;
			while from <= upto do begin
				while (endHull > startHull) & (point2d(p.elements^^[endHull - 1]).whichSideSign(point2d(p.elements^^[endHull]), point2d(p.elements^^[from])) <= 0) do begin
{$IFC AnimationEnabled }
					if userAbort then begin
						hullSegments.freeAll;
						exit(convexHullScan);
					end;
{$ENDC }
{$IFC AnimationEnabled AND AconvexHullScan }
					if animationFlag[AConvexHullItem] then begin
						point2d(p.elements^^[endHull]).interactiveHighlight;
						waitForClick(animationFlag[AConvexHullItem]);
					end;
{$ENDC }
					endHull := endHull - 1;
				end;
{$IFC AnimationEnabled AND AconvexHullScan }
				if animationFlag[AConvexHullItem] & (endHull > 0) & ((endHull <> oldEndHull) | (from <> oldFrom)) then begin
					new(ls);
					failNil(Handle(ls));
					ls.p := point2d(p.elements^^[endHull].duplicate);
					ls.q := point2d(p.elements^^[from].duplicate);
					hullSegments.append(ls);
					ls.interactiveOutput;
					waitForClick(animationFlag[AConvexHullItem]);
					oldEndHull := endHull;
					oldFrom := from;
				end;
{$ENDC }
				if from < upto then begin
					endHull := endHull + 1;
					p.swap(endHull, from);
				end;
				from := from + 1;
			end;
		end; (* hull *)

	begin
		if p.length > 1 then begin
			p.swap(1, p.findExtreme(pointLessThanXY));
			p.swap(p.length, p.findExtreme(pointGreaterThanXY));
{$IFC AnimationEnabled AND AconvexHullScan }
			new(hullSegments);
			failNil(Handle(hullSegments));
			hullSegments.init;
			if animationFlag[AConvexHullItem] then begin
{$IFC THINK_PASCAL }
				PenPat(gray);
				point2d(p.elements^^[1]).drawLine(point2d(p.elements^^[p.length]));
				PenPat(black);
{$ELSEC }
				PenPat(qd.gray);
				point2d(p.elements^^[1]).drawLine(point2d(p.elements^^[p.length]));
				PenPat(qd.black);
{$ENDC }
				waitForClick(animationFlag[AConvexHullItem]);
			end;
{$ENDC }
			if point2d(p.elements^^[1]).xEqual(point2d(p.elements^^[p.length])) then
				if point2d(p.elements^^[1]).yEqual(point2d(p.elements^^[p.length])) then
					n := 1
				else begin
{$IFC AnimationEnabled AND AconvexHullScan }
					if animationFlag[AConvexHullItem] then begin
						for upper := 2 to p.length - 1 do begin
							point2d(p.elements^^[upper]).interactiveHighlight;
							waitForClick(animationFlag[AConvexHullItem]);
						end;
					end;
{$ENDC }
					p.swap(2, p.length);
					n := 2;
				end
			else begin
				partition(p, upper, lower);
{$IFC AnimationEnabled AND AconvexHullScan }
				if animationFlag[AConvexHullItem] then begin
					p.partialInteractiveFlash(1, upper);
					waitForClick(animationFlag[AConvexHullItem]);
				end;
				if animationFlag[AConvexHullItem] then begin
					p.partialInteractiveFlash(lower, p.length);
					waitForClick(animationFlag[AConvexHullItem]);
				end;
{$ENDC }
				endHull := 0;
				p.sort(pointLessThanXY, 1, upper);
				p.swap(upper + 1, p.length);
				hull(1, upper + 1, endHull);
				p.swap(upper + 1, p.length);
				p.sort(pointGreaterThanXY, lower, p.length);
				p.elements^^[p.length + 1] := p.elements^^[1];
				hull(lower, p.length + 1, endHull);
				n := endHull;
			end;
{$IFC AnimationEnabled AND AconvexHullScan }
			hullSegments.interactiveOutput;
			hullSegments.freeAll;
{$ENDC }
		end
		else
			n := p.length;
	end; (* convexHullScan *)

procedure pointVector.convexHull (eliminate, verbose: boolean; var hullLength, eliminated: seqIndex0);
	var
		oldLength: seqIndex0;

	function eliminateRectangle (v: vector): seqIndex0;
		var
			i1, i2, i3, i4, j, m: seqIndex0;
			a1, a2, a3, a4, highX, highY, lowX, lowY, d: extended;
			t: point2d;
{$IFC AnimationEnabled AND AconvexHullScan }
			r: rectangle;
{$ENDC }

		begin
			i1 := 1;
			i2 := 1;
			i3 := 1;
			i4 := 1;
			a1 := point2d(v.elements^^[1]).xPlusYX;
			a2 := point2d(v.elements^^[1]).xMinusYX;
			a3 := a2;
			a4 := a1;
			for j := 2 to v.length do begin
{$IFC AnimationEnabled }
				if userAbort then
					exit(convexHull);
{$ENDC }
				t := point2d(v.elements^^[j]);
				d := t.xPlusYX;
				if d < a1 then begin
					i1 := j;
					a1 := d;
				end
				else if d > a4 then begin
					i4 := j;
					a4 := t.xPlusYX;
				end;
				d := t.xMinusYX;
				if d < a2 then begin
					i2 := j;
					a2 := d;
				end
				else if d > a3 then begin
					i3 := j;
					a3 := t.xMinusYX;
				end;
			end;
			lowX := max(point2d(v.elements^^[i1]).getX, point2d(v.elements^^[i2]).getX);
			lowY := max(point2d(v.elements^^[i1]).getY, point2d(v.elements^^[i3]).getY);
			highX := min(point2d(v.elements^^[i3]).getX, point2d(v.elements^^[i4]).getX);
			highY := min(point2d(v.elements^^[i2]).getY, point2d(v.elements^^[i4]).getY);
{$IFC AnimationEnabled AND AconvexHullScan }
			if animationFlag[AConvexHullItem] then begin
				v.elements^^[i1].interactiveOutput;
				v.elements^^[i2].interactiveOutput;
				v.elements^^[i3].interactiveOutput;
				v.elements^^[i4].interactiveOutput;
				new(r);
				failNil(Handle(r));
				r.init;
				r.left := lowX;
				r.right := highX;
				r.bottom := lowY;
				r.top := highY;
				r.interactiveOutput;
				r.free;
				waitForClick(animationFlag[AConvexHullItem]);
			end;
{$ENDC }
			m := 0;
			j := v.length;
			while j > m do begin
{$IFC AnimationEnabled }
				if userAbort then
					exit(convexHull);
{$ENDC }
				t := point2d(v.elements^^[j]);
				if (lowX < t.getX) & (t.getX < highX) & (lowY < t.getY) & (t.getY < highY) then begin
{$IFC AnimationEnabled AND AconvexHullScan }
					if animationFlag[AConvexHullItem] then begin
						t.interactiveHighlight;
						waitForClick(animationFlag[AConvexHullItem]);
					end;
{$ENDC }
					j := j - 1;
				end
				else begin
					m := m + 1;
					v.swap(m, j);
				end;
			end;
			eliminateRectangle := m;
		end;

	begin
		if eliminate & (length > 3) then begin
			oldLength := length;
			length := eliminateRectangle(vector(self));
			eliminated := oldLength - length;
{$IFC AnimationEnabled AND AConvexHullScan }
			if verbose then
				displayInfoString(stringOf('Eliminated: ', eliminated : 1));
{$ENDC }
			elements^^[oldLength + 1] := elements^^[length + 1];
			convexHullScan(vector(self), hullLength);
{$IFC AnimationEnabled }
			if userAbort then
				exit(convexHull);
{$ENDC }
			elements^^[length + 1] := elements^^[oldLength + 1];
			length := oldLength;
		end
		else begin
			eliminated := 0;
			convexHullScan(vector(self), hullLength);
		end;
	end;

end.
