(*	XYZ GeoBench (eXperimental geometrY Zurich):		*)
(*	An object oriented, interactive workbench for		*)
(*	experimental geometric computation.					*)
(*														*)
(*	 1999 by Peter Schorn, peter.schorn@acm.org		*)
(*	Culmannstrasse 77, CH-8006 Zrich, Switzerland		*)
(*	This software may be copied freely for educational	*)
(*	and research purposes only. For commercial use,		*)
(*	please contact the author.							*)
(*														*)
(*	Version 5.0.5: 30-Sep-1999							*)

unit dictionary;

interface

uses
	sequence;

type
	reference = ^integer; (* any pointer *)
	directionType = (left, right);

	dictionary = object(obj)

			numberOfElements: longInt;

			procedure init;
			override;
			procedure freeAll;
			override;

(* Elements of the dictionary are accessed by means of a 'reference' thereby separating key operations from			*)
(* tree transformations.																												*)

			function getObject (where: reference): obj;
(* Retrieves the object referenced by 'where'. *)

			procedure setObject (where: reference; newValue: obj);
(* Changes the object referenced by 'where' into 'newValue'. *)

			procedure sequenceToDictionary (s: vector; l: seqIndex1; r: seqIndex0);
(* Equivalent to inserting the sorted vector elements between 'l' and 'r' into an empty dictionary. *)

			function find (x: obj; function compare (a, b: obj): relationObj; var where: reference; var direction: directionType): boolean;
(* If 'find' = TRUE then 'where' points to some element 'e' with 'compare(e, x) = equalObj'. If 'find' = FALSE then		*)
(* either 'where' = nil (dictionary empty) or 'x' is a neighbor of 'where' in direction 'direction'. 						*)

			function findPosition (x: obj; function compare (a, b: obj): relationObj; var where: reference; var direction: directionType; var position: longInt): boolean;
(* Same as 'find'. In addition 'position' is the number of elements less than 'x' plus 'eq', *)
(*	where 1  'eq'  number of elements equal to 'x'. *)

			function getPosition (x: reference): longInt;
(* Determines the position of 'x' in the dictionary. *)

			function ith (i: longInt): reference;
(*	Returns the i'th element of the dictionary.	(1  'i'  'numberOfElements') *)

			function member (x: obj; function compare (a, b: obj): relationObj): boolean;
(* Tests whether 'x' is in the dictionary or not. *)

			function insert (x: obj; where: reference; direction: directionType): reference;
(* If 'where' = nil then insert 'x' into the dictionary which must be empty. If 'where'  nil then insert 'x' before		*)
(* ('direction'=left) or after ('direction'=right) 'where'. *)

			function insertIfNew (x: obj; function compare (a, b: obj): relationObj): reference;
(* Insert 'x' into the dictionary only if it is new, otherwise nothing happens.		*)

			function insertObj (x: obj; function compare (a, b: obj): relationObj; var found: boolean): reference;
(* Insert 'x' into the dictionary. 'found' = FALSE iff 'x' is unique. *)

			procedure insertNewObj (x: obj; function compare (a, b: obj): relationObj);
(* Inserts 'x' if it is not a member of the dictionary, otherwise this is an error. *)

			procedure delete (x: reference);
(* Remove 'x' from the dictionary. *)

			procedure swap (p, q: reference);
(* Exchanges the elements 'p' and 'q' in the dictionary. *)

			procedure rangeSwap (p, q: reference);
(* Exchange the elements between 'p' and 'q' in the dictionary ('compare(p, q) = lessThanObj' must hold and is not checked). *)

			function next (x: reference; direction: directionType): reference;
(* Find the predecessor ('direction'=left) or the successor ('direction'=right) of 'x'. *)

			function extreme (direction: directionType): reference;
(* Find the leftmost ('direction'=left) or the rightmost ('direction'=right) value. *)

			function isEmpty: boolean;
(* Determines whether the dictionary is empty. *)

			procedure forAll (procedure whatToDo (x: obj); direction: directionType);
(* Performs 'whatToDo' on all elements 'x' of the dictionary, starting at the location most in direction 'direction'.	*)

			procedure rangeForAll (leftBound, rightBound: obj; function compare (a, b: obj): relationObj; procedure whatToDo (x: obj); direction: directionType);
(* Performs 'whatToDo' on all elements 'x' of the dictionary that are between 'leftBound' and 'rightBound'				*)
(* (including), starting at the location most in direction 'direction'. The value nil serves as -infinity for				*)
(* 'leftBound' and as +infinity for 'rightBound'. 																					*)

(* begin Marco Ciot *)
			procedure rangeForAllWhile (leftBound, rightBound: obj; function compare (a, b: obj): relationObj; function whatToDo (x: obj): boolean; direction: directionType);
(* Like rangeForAll, but stop when whatToDo returns false *)
(* end Marco Ciot *)

			function exists (function property (x: obj): boolean): reference;
(*	Returns a reference to an element which satisfies 'property' if one exists. Otherwise 'nil' is returned.				*)

			procedure reset;
(* Reset the dictionary to the state after initialization. Dependent objects are not removed! *)

			function compareReference (p, q: reference): relationObj;
(* determines the order of the elements referenced by 'p' and 'q' in the dictionary (without key operations!). *)

			function infoString: Str255;
(* Computes information about the dictionary (this only makes sense for a 'dictionaryStatistics'. *)

			function getInfo: obj;
			override;
		end;

	dictionaryStatistics = object(dictionary)
			d: dictionary;		(* Actual dictionary that is used. 					*)
			maxLength,			(* Maximal number of elements in the dictionary.	*)
			insertions,			(* Number of insertions into the dictionary. 		*)
			deletions: longInt;	(* Number of deletions from the dictionary. 		*)
			procedure init;
			override;
(* The global variable 'currentDictionaryType' determines which kind of dictionary is used	*)
(* for 'd' and initializes it. Note that the initialization of 'd' might require the correct setting	*)
(* of additional global variables. *)

			procedure free;
			override;
			procedure freeAll;
			override;
			function getObject (where: reference): obj;
			override;
			procedure setObject (where: reference; newValue: obj);
			override;
			procedure sequenceToDictionary (s: vector; l: seqIndex1; r: seqIndex0);
			override;
			function find (x: obj; function compare (a, b: obj): relationObj; var where: reference; var direction: directionType): boolean;
			override;
			function findPosition (x: obj; function compare (a, b: obj): relationObj; var where: reference; var direction: directionType; var position: longInt): boolean;
			override;
			function getPosition (x: reference): longInt;
			override;
			function ith (i: longInt): reference;
			override;
			function member (x: obj; function compare (a, b: obj): relationObj): boolean;
			override;
			function insert (x: obj; where: reference; direction: directionType): reference;
			override;
			function insertObj (x: obj; function compare (a, b: obj): relationObj; var found: boolean): reference;
			override;
			procedure insertNewObj (x: obj; function compare (a, b: obj): relationObj);
			override;
			procedure delete (x: reference);
			override;
			procedure swap (p, q: reference);
			override;
			procedure rangeSwap (p, q: reference);
			override;
			function next (x: reference; direction: directionType): reference;
			override;
			function extreme (direction: directionType): reference;
			override;
			function isEmpty: boolean;
			override;
			procedure forAll (procedure whatToDo (x: obj); direction: directionType);
			override;
			function exists (function property (x: obj): boolean): reference;
			override;
			procedure reset;
			override; (* do not reset the counters! *)
			procedure rangeForAll (leftBound, rightBound: obj; function compare (a, b: obj): relationObj; procedure whatToDo (x: obj); direction: directionType);
			override;
			function compareReference (p, q: reference): relationObj;
			override;
			function infoString: Str255;
			override;
			procedure countInsertion;
			procedure resetCounters;
		end;

	balanceType = (leftGreaterRight, leftEqualRight, leftLessRight);
	avlNodeH = ^avlNodeP;
	avlNodeP = ^avlNode;
	avlNode = record
			value: obj;
			balance: balanceType;
			treeSize: longInt; (* total size of the subtree rooted at this node, i.e. size(left)+size(right)+1 *)
			sons: array[directionType] of avlNodeH;
			father: avlNodeH;
		end;

	avlTree = object(dictionary)
(* Implementation as a balanced (AVL) tree with optimal insertion / deletion costs. *)
			root: avlNodeH;
(* No global variables are needed for the initialization procedure 'init'. *)
			procedure init;
			override;
			procedure free;
			override;
			procedure interactiveOutput;
			override;
			procedure textOutput;
			override;
			function getObject (where: reference): obj;
			override;
			procedure setObject (where: reference; newValue: obj);
			override;
			procedure sequenceToDictionary (s: vector; l: seqIndex1; r: seqIndex0);
			override;
			function find (x: obj; function compare (a, b: obj): relationObj; var where: reference; var direction: directionType): boolean;
			override;
			function getPosition (x: reference): longInt;
			override;
			function ith (i: longInt): reference;
			override;
			function findPosition (x: obj; function compare (a, b: obj): relationObj; var where: reference; var direction: directionType; var position: longInt): boolean;
			override;
			function insert (x: obj; where: reference; direction: directionType): reference;
			override;
			procedure delete (x: reference);
			override;
			procedure swap (p, q: reference);
			override;
			function next (x: reference; direction: directionType): reference;
			override;
			function extreme (direction: directionType): reference;
			override;
			procedure forAll (procedure whatToDo (x: obj); direction: directionType);
			override;
			function exists (function property (x: obj): boolean): reference;
			override;
			procedure reset;
			override;
			procedure rangeForAll (leftBound, rightBound: obj; function compare (a, b: obj): relationObj; procedure whatToDo (x: obj); direction: directionType);
			override;
			procedure rangeForAllWhile (leftBound, rightBound: obj; function compare (a, b: obj): relationObj; function whatToDo (x: obj): boolean; direction: directionType);
			override;
			function compareReference (p, q: reference): relationObj;
			override;
			procedure setTreeSize;
			function testTreeSize: longInt;
		end;

	sortedListNodeH = ^sortedListNodeP;
	sortedListNodeP = ^sortedListNode;
	sortedListNode = record
			value: obj;
			sons: array[directionType] of sortedListNodeH;
		end;

	sortedList = object(dictionary)
(* Implementation as a sorted list. Insertion / deletion is expensive. *)
			minimum, maximum: sortedListNodeH;
(* No global variables are needed for the initialization procedure 'init'. *)
			procedure init;
			override;
			procedure free;
			override;
			procedure interactiveOutput;
			override;
			procedure textOutput;
			override;
			function getObject (where: reference): obj;
			override;
			procedure setObject (where: reference; newValue: obj);
			override;
			procedure sequenceToDictionary (s: vector; l: seqIndex1; r: seqIndex0);
			override;
			function findPosition (x: obj; function compare (a, b: obj): relationObj; var where: reference; var direction: directionType; var position: longInt): boolean;
			override;
			function getPosition (x: reference): longInt;
			override;
			function ith (i: longInt): reference;
			override;
			function insert (x: obj; where: reference; direction: directionType): reference;
			override;
			procedure delete (x: reference);
			override;
			procedure swap (p, q: reference);
			override;
			function next (x: reference; direction: directionType): reference;
			override;
			function extreme (direction: directionType): reference;
			override;
			procedure forAll (procedure whatToDo (x: obj); direction: directionType);
			override;
			function exists (function property (x: obj): boolean): reference;
			override;
			procedure reset;
			override;
			procedure rangeForAll (leftBound, rightBound: obj; function compare (a, b: obj): relationObj; procedure whatToDo (x: obj); direction: directionType);
			override;
			function compareReference (p, q: reference): relationObj;
			override;
		end;

	refPair = record
			value: obj;			(* the object *)
			ref: seqIndex0;		(* its position in vector if occupied or link field to next empty if free *)
		end;
	refPairArrayH = ^refPairArrayP;
	refPairArrayP = ^refPairArray;
	refPairArray = array[seqIndex] of refPair;
	seqIndex0ArrayH = ^seqIndex0ArrayP;
	seqIndex0ArrayP = ^seqIndex0Array;
	seqIndex0Array = array[seqIndex] of seqIndex0;

	sortedVector = object(dictionary)
(* Implementation as a sorted array. Insertion / deletion is expensive. *)
			freeList: seqIndex;				(* points to first empty element in 'masters', linked by 'ref' field *)
			masters: refPairArrayH;		(* contains objects and their position *)
			vector: seqIndex0ArrayH;	(* sorted *)
			capacity: seqIndex0;
			procedure init;
			override;
(* The global variable 'currentVectorLength' determines how many elements are allocated. *)
			procedure allocateElements (number: seqIndex0);
(* reserves space for 'number' entries *)
			procedure free;
			override;
			procedure interactiveOutput;
			override;
			procedure textOutput;
			override;
			function getObject (where: reference): obj;
			override;
			procedure setObject (where: reference; newValue: obj);
			override;
			procedure sequenceToDictionary (s: vector; l: seqIndex1; r: seqIndex0);
			override;
			function findPosition (x: obj; function compare (a, b: obj): relationObj; var where: reference; var direction: directionType; var position: longInt): boolean;
			override;
			function getPosition (x: reference): longInt;
			override;
			function ith (i: longInt): reference;
			override;
			function insert (x: obj; where: reference; direction: directionType): reference;
			override;
			procedure delete (x: reference);
			override;
			procedure swap (p, q: reference);
			override;
			function next (x: reference; direction: directionType): reference;
			override;
			function extreme (direction: directionType): reference;
			override;
			procedure forAll (procedure whatToDo (x: obj); direction: directionType);
			override;
			function exists (function property (x: obj): boolean): reference;
			override;
			procedure reset;
			override;
			procedure rangeForAll (leftBound, rightBound: obj; function compare (a, b: obj): relationObj; procedure whatToDo (x: obj); direction: directionType);
			override;
			function compareReference (p, q: reference): relationObj;
			override;
			function ithObject (i: seqIndex): obj;
		end;

function switchDirection (d: directionType): directionType;
(* changes left into right and vice versa *)

procedure initDictionary;

implementation

procedure dictionary.reset;
	begin
		unimplementedMethod('dictionary.reset');
	end;

procedure dictionary.init;
	begin
		numberOfElements := 0;
	end;

{$IFC NOT THINK_PASCAL }
{$PUSH }
{$PRAGMAC warn_unusedarg off }
{$ENDC }
function dictionary.getObject (where: reference): obj;
	begin
		unimplementedMethod('dictionary.getObject');
		getObject := nil;
	end;

procedure dictionary.setObject (where: reference; newValue: obj);
	begin
		unimplementedMethod('dictionary.setObject');
	end;

procedure dictionary.sequenceToDictionary (s: vector; l: seqIndex1; r: seqIndex0);
	begin
		unimplementedMethod('dictionary.sequenceToDictionary');
	end;

function dictionary.findPosition (x: obj; function compare (a, b: obj): relationObj; var where: reference; var direction: directionType; var position: longInt): boolean;
	begin
		unimplementedMethod('dictionary.findPosition');
		findPosition := false;
	end;

function dictionary.getPosition (x: reference): longInt;
	begin
		unimplementedMethod('dictionary.getPosition');
		getPosition := 0;
	end;

function dictionary.ith (i: longInt): reference;
	begin
		unimplementedMethod('dictionary.ith');
		ith := nil;
	end;

function dictionary.insert (x: obj; where: reference; direction: directionType): reference;
	begin
		unimplementedMethod('dictionary.insert');
		insert := nil;
	end;

procedure dictionary.delete (x: reference);
	begin
		unimplementedMethod('dictionary.delete');
	end;

procedure dictionary.swap (p, q: reference);
	begin
		unimplementedMethod('dictionary.swap');
	end;

function dictionary.next (x: reference; direction: directionType): reference;
	begin
		unimplementedMethod('dictionary.next');
		next := nil;
	end;

function dictionary.extreme (direction: directionType): reference;
	begin
		unimplementedMethod('dictionary.extreme');
		extreme := nil;
	end;

procedure dictionary.rangeForAll (leftBound, rightBound: obj; function compare (a, b: obj): relationObj; procedure whatToDo (x: obj); direction: directionType);
	begin
		unimplementedMethod('dictionary.rangeForAll');
	end;

procedure dictionary.forAll (procedure whatToDo (x: obj); direction: directionType);
	begin
		unimplementedMethod('dictionary.forAll');
	end;

function dictionary.exists (function property (x: obj): boolean): reference;
	begin
		unimplementedMethod('dictionary.exists');
		exists := nil;
	end;

(* begin Marco Ciot *)
procedure dictionary.rangeForAllWhile (leftBound, rightBound: obj; function compare (a, b: obj): relationObj; function whatToDo (x: obj): boolean; direction: directionType);
	begin
		unimplementedMethod('dictionary.rangeForAllWhile');
	end;
(* end Marco Ciot *)

function dictionary.compareReference (p, q: reference): relationObj;
	begin
		unimplementedMethod('dictionary.compareReference');
		compareReference := lessThanObj;
	end;
{$IFC NOT THINK_PASCAL }
{$POP }
{$ENDC }

function dictionary.find (x: obj; function compare (a, b: obj): relationObj; var where: reference; var direction: directionType): boolean;
	var
		dummy: longInt;
	begin
		find := findPosition(x, compare, where, direction, dummy);
	end;

function dictionary.member (x: obj; function compare (a, b: obj): relationObj): boolean;
	var
		where: reference;
		direction: directionType;
	begin
		member := find(x, compare, where, direction);
	end;

function dictionary.insertObj (x: obj; function compare (a, b: obj): relationObj; var found: boolean): reference;
	var
		where: reference;
		direction: directionType;
	begin
		found := find(x, compare, where, direction);
		if found then
			insertObj := insert(x, where, left)
		else
			insertObj := insert(x, where, direction);
	end;

function dictionary.insertIfNew (x: obj; function compare (a, b: obj): relationObj): reference;
	var
		where: reference;
		direction: directionType;
		found: boolean;
	begin
		found := find(x, compare, where, direction);
		if found then
			insertIfNew := where
		else
			insertIfNew := insert(x, where, direction);
	end;

procedure dictionary.insertNewObj (x: obj; function compare (a, b: obj): relationObj);
	var
		found: boolean;
		where: reference;
	begin
		where := insertObj(x, compare, found);
		assert(not found, 'not found');
	end;

procedure dictionary.rangeSwap (p, q: reference);
	var
		pOld, qOld: reference;
	begin
		if (p <> nil) & (q <> nil) & (p <> q) then begin
			repeat
				swap(p, q);
				pOld := p;
				qOld := q;
				p := next(qOld, right);
				q := next(pOld, left);
			until (p = pOld) & (q = qOld) | (p = q);
		end;
	end;

function dictionary.isEmpty: boolean;
	begin
		isEmpty := numberOfElements = 0;
	end;

function dictionary.infoString: Str255;
	begin
		infoString := '';
	end;

function dictionary.getInfo: obj;
	var
		v: vector;
		i: seqIndex0;
	procedure getInfoP (o: obj);
		begin
			appendObjInfoObject(v, stringOf(i : 2, '-Entry'), o, true);
			i := i + 1;
		end;
	begin
		v := newVector(numberOfElements);
		i := 1;
		forAll(getInfoP, left);
		getInfo := v;
	end;

procedure dictionary.freeAll;
	begin
		forAll(freeAllP, right);
		free;
	end;

procedure dictionaryStatistics.reset;
	begin
		d.reset;
	end;

procedure dictionaryStatistics.resetCounters;
	begin
		maxLength := 0;
		insertions := 0;
		deletions := 0;
	end;

procedure dictionaryStatistics.init;
	begin
		d := dictionary(newInitObj(currentDictionaryType));
		resetCounters;
	end;

procedure dictionaryStatistics.free;
	begin
		d.free;
		shallowFree;
	end;

procedure dictionaryStatistics.freeAll;
	begin
		d.freeAll;
		shallowFree;
	end;

function dictionaryStatistics.getObject (where: reference): obj;
	begin
		getObject := d.getObject(where);
	end;

procedure dictionaryStatistics.setObject (where: reference; newValue: obj);
	begin
		d.setObject(where, newValue);
	end;

procedure dictionaryStatistics.sequenceToDictionary (s: vector; l: seqIndex1; r: seqIndex0);
	begin
		d.sequenceToDictionary(s, l, r);
	end;

function dictionaryStatistics.find (x: obj; function compare (a, b: obj): relationObj; var where: reference; var direction: directionType): boolean;
	begin
		find := d.find(x, compare, where, direction);
	end;

function dictionaryStatistics.findPosition (x: obj; function compare (a, b: obj): relationObj; var where: reference; var direction: directionType; var position: longInt): boolean;
	begin
		findPosition := d.findPosition(x, compare, where, direction, position);
	end;

function dictionaryStatistics.getPosition (x: reference): longInt;
	begin
		getPosition := d.getPosition(x);
	end;

function dictionaryStatistics.ith (i: longInt): reference;
	begin
		ith := d.ith(i);
	end;

function dictionaryStatistics.insert (x: obj; where: reference; direction: directionType): reference;
	begin
		insert := d.insert(x, where, direction);
		countInsertion;
	end;

function dictionaryStatistics.member (x: obj; function compare (a, b: obj): relationObj): boolean;
	begin
		member := d.member(x, compare);
	end;

function dictionaryStatistics.insertObj (x: obj; function compare (a, b: obj): relationObj; var found: boolean): reference;
	begin
		insertObj := d.insertObj(x, compare, found);
		countInsertion;
	end;

procedure dictionaryStatistics.insertNewObj (x: obj; function compare (a, b: obj): relationObj);
	begin
		d.insertNewObj(x, compare);
		countInsertion;
	end;

procedure dictionaryStatistics.delete (x: reference);
	begin
		d.delete(x);
		deletions := deletions + 1;
	end;

procedure dictionaryStatistics.swap (p, q: reference);
	begin
		d.swap(p, q);
	end;

procedure dictionaryStatistics.rangeSwap (p, q: reference);
	begin
		d.rangeSwap(p, q);
	end;

function dictionaryStatistics.next (x: reference; direction: directionType): reference;
	begin
		next := d.next(x, direction);
	end;

function dictionaryStatistics.extreme (direction: directionType): reference;
	begin
		extreme := d.extreme(direction);
	end;

function dictionaryStatistics.isEmpty: boolean;
	begin
		isEmpty := d.isEmpty;
	end;

procedure dictionaryStatistics.forAll (procedure whatToDo (x: obj); direction: directionType);
	begin
		d.forAll(whatToDo, direction);
	end;

function dictionaryStatistics.exists (function property (x: obj): boolean): reference;
	begin
		exists := d.exists(property);
	end;

procedure dictionaryStatistics.rangeForAll (leftBound, rightBound: obj; function compare (a, b: obj): relationObj; procedure whatToDo (x: obj); direction: directionType);
	begin
		d.rangeForAll(leftBound, rightBound, compare, whatToDo, direction);
	end;

function dictionaryStatistics.compareReference (p, q: reference): relationObj;
	begin
		compareReference := d.compareReference(p, q);
	end;

procedure dictionaryStatistics.countInsertion;
	begin
		insertions := insertions + 1;
		if d.numberOfElements > maxLength then
			maxLength := d.numberOfElements;
	end;

function dictionaryStatistics.infoString: Str255;
	begin
		infoString := stringOf('(M=', maxLength : 1, ', I=', insertions : 1, ', D=', deletions : 1, ')');
	end;

function switchDirection (d: directionType): directionType;
	begin
		if d = left then
			switchDirection := right
		else
			switchDirection := left;
	end;

procedure initDictionary;
	var
		s: sortedList;
		a: avlTree;
		v: sortedVector;
	begin
		new(a);
		failNil(Handle(a));
		prototypes[avlTreeType] := a;
		new(s);
		failNil(Handle(s));
		prototypes[sortedListType] := s;
		new(v);
		failNil(Handle(v));
		prototypes[sortedVectorType] := v;
	end;


end.
