(*	XYZ GeoBench (eXperimental geometrY Zurich):		*)
(*	An object oriented, interactive workbench for		*)
(*	experimental geometric computation.					*)
(*														*)
(*	 1999 by Peter Schorn, peter.schorn@acm.org		*)
(*	Culmannstrasse 77, CH-8006 Zrich, Switzerland		*)
(*	This software may be copied freely for educational	*)
(*	and research purposes only. For commercial use,		*)
(*	please contact the author.							*)
(*														*)
(*	Version 5.0.5: 30-Sep-1999							*)

unit exploreGraph;

interface

uses
	triangulation;

function exploreGraph (t: triangulation2d; function flipOk (x: graphEdgeTriangulation): boolean; directed, labeled: boolean): directedGraph;

implementation

{$SETC ATriangulation = TRUE }

const
	maxEdges = 100;
	xOrigin = 50;
	yOrigin = 50;
	dy = 50;
	dx = 10;

type
	randomTable = array[1..maxEdges, 1..maxEdges] of longInt;
	randomTableP = ^randomTable;
	randomTableH = ^randomTableP;

	dictionaryObject = object(obj)
			s: longInt;
			v: seqIndex;
		end;

	numObject = object(obj)
			startVertex, endVertex: seqIndex;
			n: longInt;
		end;

function compareNumObject (a, b: numObject): relationObj;
	begin
		if a.startVertex < b.startVertex then
			compareNumObject := lessThanObj
		else if a.startVertex > b.startVertex then
			compareNumObject := greaterThanObj
		else if a.endVertex < b.endVertex then
			compareNumObject := lessThanObj
		else if a.endVertex > b.endVertex then
			compareNumObject := greaterThanObj
		else
			compareNumObject := equalObj
	end;

function compareDictionaryObject (a, b: dictionaryObject): relationObj;
	begin
		if a.s < b.s then
			compareDictionaryObject := lessThanObj
		else if a.s > b.s then
			compareDictionaryObject := greaterThanObj
		else
			compareDictionaryObject := equalObj;
	end;

function exploreGraph (t: triangulation2d; function flipOk (x: graphEdgeTriangulation): boolean; directed, labeled: boolean): directedGraph;
	var
		d, eDict, numDict: avlTree;
		vertices, edges: list;
		table: randomTableH;
		signature: longInt;
		g: directedGraph;
		dObj: dictionaryObject;
		p: realPoint;

	procedure appendEdge (s, e, startX, endX: seqIndex);
		var
			x: graphEdge;
			y: markedGraphEdge;
			n: numObject;
			direction: directionType;
			where: reference;
		begin
			if labeled then begin
				new(y);
				x := y;
			end
			else
				new(x);
			failNil(Handle(x));
			x.init;
			if labeled then begin
				new(n);
				failNil(Handle(n));
				with n do
					if startX < endX then begin
						startVertex := startX;
						endVertex := endX;
					end
					else begin
						endVertex := startX;
						startVertex := endX;
					end;
				if numDict.find(n, compareNumObject, where, direction) then begin
					markedGraphEdge(x).mark.value := StringOf(n.startVertex : 1, '.', n.endVertex : 1, '=', numObject(numDict.getObject(where)).n : 1);
					n.free;
				end
				else begin
					where := numDict.insert(n, where, direction);
					n.n := numDict.numberOfElements;
					markedGraphEdge(x).mark.value := StringOf(n.startVertex : 1, '.', n.endVertex : 1, '=', n.n : 1);
				end;
			end;
			x.startVertex := s;
			x.endVertex := e;
			if directed then
				edges.append(x)
			else if eDict.find(x, compareUndirectedGraphEdge, where, direction) then begin
				if labeled then
					with markedGraphEdge(eDict.getObject(where)).mark do
						if value < markedGraphEdge(x).mark.value then
							value := concat(value, ',', markedGraphEdge(x).mark.value)
						else
							value := concat(markedGraphEdge(x).mark.value, ',', value);
				x.free;
			end
			else begin
				where := eDict.insert(x, where, direction);
				edges.append(x);
			end;
		end;

	procedure appendVertex;
		begin
			vertices.append(p.duplicate);
		end;

	procedure changeEdge (x: graphEdge);
		begin
			with x do
				if startVertex < endVertex then
					signature := BitXor(signature, table^^[startVertex, endVertex])
				else
					signature := BitXor(signature, table^^[endVertex, startVertex]);
		end;

	procedure explore (parent: seqIndex);
		var
			i, startX, endX: seqIndex;
			x: graphEdgeTriangulation;
			direction: directionType;
			where: reference;
			sigSave: longInt;
		begin
			sigSave := signature;
			p.y := p.y + dy;
			for i := 1 to t.edges.length do begin
				signature := sigSave;
{$IFC AnimationEnabled }
				if userAbort then
					exit(explore);
{$ENDC }
				x := graphEdgeTriangulation(t.edges.elements^^[i]);
				if flipOk(x) then begin
					with x do begin
						startX := startVertex;
						endX := endVertex;
					end;
					changeEdge(x);
{$IFC AnimationEnabled AND ATriangulation }
					if animationFlag[ATriangulationItem] then begin
						x.interactiveOutput;
						x.interactiveFlash;
					end;
{$ENDC }
					t.flip(x);
{$IFC AnimationEnabled AND ATriangulation }
					if animationFlag[ATriangulationItem] then begin
						x.interactiveOutput;
						waitForClick(animationFlag[ATriangulationItem]);
					end;
{$ENDC }
					changeEdge(x);
					dObj.s := signature;
					if d.find(dObj, compareDictionaryObject, where, direction) then
						appendEdge(parent, dictionaryObject(d.getObject(where)).v, startX, endX)
					else begin
						p.x := p.x + dx;
						appendVertex;
						dObj.v := vertices.length;
						where := d.insert(dObj.duplicate, where, direction);
						appendEdge(parent, vertices.length, startX, endX);
						explore(vertices.length);
					end;
{$IFC AnimationEnabled AND ATriangulation }
					if animationFlag[ATriangulationItem] then
						x.interactiveOutput;
{$ENDC }
					t.flip(x);
{$IFC AnimationEnabled AND ATriangulation }
					if animationFlag[ATriangulationItem] then
						x.interactiveOutput;
{$ENDC }
				end;
			end;
			p.y := p.y - dy;
		end;

	procedure initStructures;
		var
			dummy: reference;
			i, j, l: seqIndex;
		begin
			l := t.edges.length;
			assert(l <= maxEdges, 't.edges.length  maxEdges');
			table := randomTableH(newHandleFailNil(SizeOf(randomTable)));
			for i := 1 to l do
				for j := i + 1 to l do
					table^^[i, j] := randomLongInt;
			signature := 0;
			for i := 1 to t.edges.length do
				changeEdge(graphEdge(t.edges.elements^^[i]));
			new(vertices);
			failNil(Handle(vertices));
			vertices.init;
			new(edges);
			failNil(Handle(edges));
			edges.init;
			new(d);
			failNil(Handle(d));
			d.init;
			if labeled then begin
				new(numDict);
				failNil(Handle(numDict));
				numDict.init;
			end;
			if not directed then begin
				new(eDict);
				failNil(Handle(eDict));
				eDict.init;
			end;
			new(p);
			failNil(Handle(p));
			p.x := xOrigin;
			p.y := yOrigin;
			appendVertex;
			new(dObj);
			failNil(Handle(dObj));
			dObj.s := signature;
			dObj.v := 1;
			dummy := d.insert(dObj.duplicate, nil, left);
			currentTriangulation := t;
		end;

	procedure addEdgeToGraph (x: graphEdgeTriangulation);
		begin
			g.edges.append(x);
		end;

	procedure addVertexToGraph (v: realPoint);
		begin
			g.vertices.append(v);
		end;

	procedure produceGraph;
		var
			h: undirectedGraph;
		begin
			if directed then
				new(g)
			else begin
				new(h);
				g := h;
			end;
			failNil(Handle(g));
			currentVertices := vertices.length;
			currentEdges := edges.length;
			g.init;
			vertices.forAll(addVertexToGraph);
			edges.forAll(addEdgeToGraph);
			g.initAdjacency;
		end;

{$IFC AnimationEnabled }
	procedure exitStructuresCommon;
		begin
			d.freeAll;
			if labeled then
				numDict.freeAll;
			if not directed then
				eDict.free;
			p.free;
			dObj.free;
{$IFC THINK_PASCAL }
			disposHandle(Handle(table));
{$ELSEC }
			DisposeHandle(Handle(table));
{$ENDC }
		end;

	procedure exitStructures;
		begin
			edges.free;
			vertices.free;
			if labeled then
				displayInfoString(StringOf('Different Flips = ', numDict.numberOfElements : 1));
			exitStructuresCommon;
		end;
{$ELSEC }
	procedure exitStructures;
		begin
			edges.free;
			vertices.free;
			d.freeAll;
			if labeled then begin
				displayInfoString(StringOf('Different Flips = ', numDict.numberOfElements : 1));
				numDict.freeAll;
			end;
			if not directed then
				eDict.free;
			p.free;
			dObj.free;
			DisposHandle(Handle(table));
		end;
{$ENDC }

	begin
{$IFC AnimationEnabled AND ATriangulation }
		if animationFlag[ATriangulationItem] then
			t.interactiveHighlight;
{$ENDC }
		initStructures;
		explore(1);
{$IFC AnimationEnabled }
		if userAbort then begin
			edges.freeAll;
			vertices.freeAll;
			exitStructuresCommon;
			exploreGraph := nil;
			exit(exploreGraph);
		end;
{$ENDC }
		produceGraph;
		exitStructures;
		exploreGraph := g;
	end;

end.
