(*	XYZ GeoBench (eXperimental geometrY Zurich):		*)
(*	An object oriented, interactive workbench for		*)
(*	experimental geometric computation.					*)
(*														*)
(*	 1999 by Peter Schorn, peter.schorn@acm.org		*)
(*	Culmannstrasse 77, CH-8006 Zrich, Switzerland		*)
(*	This software may be copied freely for educational	*)
(*	and research purposes only. For commercial use,		*)
(*	please contact the author.							*)
(*														*)
(*	Version 5.0.5: 30-Sep-1999							*)

unit floatArithmetic;

interface

uses
	GeoBenchUtility;

{$SETC statisticsVersion = FALSE }
{$SETC IOVersion = FALSE }

const
	maxLenMantissa = 65;
	maxBase = 100;

type
	baseType = 2..maxBase;
	signType = -1..1;
	float = record
			sign: signType; (* 0 represents number 0, base and precision are always defined *)
			base: baseType;
			precision: integer; (* number of significant digits *)
			exponent: integer;
			mantissa: array[1..maxLenMantissa] of 0..maxBase; (* most significant digit in mantissa[1] *)
		end;
	relationFloat = (lessFloat, equalFloat, greaterFloat);

{$IFC statisticsVersion }
	operation = (addition, subtraction, multiplication, division, conversion);
	operationSet = set of operation;
{$ENDC }

{$IFC statisticsVersion }
var
	statisticsOn: boolean;
	epsilon: extended; (* normalized relative error of last arithmetic computation *)
{$ENDC }

procedure setPrecisionFloat (precision: integer; var y: float); (* sets the precision of y *)

procedure intToFloat (x: longInt; base: baseType; precision: integer; var y: float);

function intToFloatF (x: longInt; base: baseType; precision: integer): float;

procedure realToFloat (x: extended; base: baseType; precision: integer; var y: float);

function realToFloatF (x: extended; base: baseType; precision: integer): float;

procedure lineToFloat (l: Str255; base: baseType; precision: integer; var y: float);

function floatToReal (x: float): extended;

function floatToInt (x: float): longInt;

function floatToLine (x: float): Str255;

procedure truncateFloat (x: float; var y: float); (* removes the fractional part of x *)

procedure fractionFloat (x: float; var y: float); (* removes the integral part of x *)

procedure scaleFloat (e: integer; var y: float); (* multiplies y by base^e *)

function getExponentFloat (x: float): integer;

procedure setExponentFloat (e: integer; var x: float);

procedure changeSignFloat (var y: float);

procedure absFloat (var y: float);

procedure nextFloat (var y: float); (* calculate the successor of y *)

procedure machineEpsilon (base: baseType; precision: integer; var y: float);

procedure addFloat (x, y: float; var z: float);

procedure subtractFloat (x, y: float; var z: float);

procedure multiplyFloat (x, y: float; var z: float);

procedure divideFloat (x, y: float; var z: float);

function addF (x, y: float): float; (* for convenience *)

function subF (x, y: float): float;

function mulF (x, y: float): float;

function divF (x, y: float): float;

function absF (x: float): float;

function chsF (x: float): float; (* chsF := -x *)

function sqrF (x: float): float;

function minF (x, y: float): float;

function maxF (x, y: float): float;

function compareFloat (x, y: float): relationFloat;

function equalZeroFloat (x: float): boolean; (* true iff x = 0 *)

{$IFC statisticsVersion }
procedure clearBins; (* initialize error distribution histogram *)

procedure drawBins (ops: operationSet; mark: extended; density, save: boolean); (* draw the histogram *)
{$ENDC }

{$IFC IOVersion }
procedure getInteger (l: Str255; var number: integer);

procedure getLongInt (l: Str255; var number: longInt);

procedure getReal (l: Str255; var number: real);

procedure getExtended (l: Str255; var number: extended);

procedure getBoolean (l: Str255; var bool: boolean);
{$ENDC }

function sign (x: extended): signType;

function min (a, b: extended): extended;

function max (a, b: extended): extended;

implementation

const
	doubleMaxLenMantissa = 2 * maxLenMantissa;

{$IFC statisticsVersion }
const
	maxBin = 150;
	xOrg = 30;
	yOrg = 230;
	magnify = 200;

type
	bin = record
			total: longInt;
			counter: array[-maxBin..maxBin] of longInt;
		end;

var
	bins: array[operation] of bin;
{$ENDC }

{$IFC IOVersion }
procedure getInteger;
	var
		userInput: Str255;
	begin
		write(l, ' [', number : 1, '] = ');
		readLn(userInput);
		if userInput <> '' then
			readString(userInput, number);
	end; (* getInteger *)

procedure getLongInt;
	var
		userInput: Str255;
	begin
		write(l, ' [', number : 1, '] = ');
		readLn(userInput);
		if userInput <> '' then
			readString(userInput, number);
	end; (* getLongInt *)

procedure getReal;
	var
		userInput: Str255;
	begin
		write(l, ' [', number : 5 : 2, '] = ');
		readLn(userInput);
		if userInput <> '' then
			readString(userInput, number);
	end; (* getReal *)

procedure getExtended;
	var
		userInput: Str255;
	begin
		write(l, ' [', number : 5 : 2, '] = ');
		readLn(userInput);
		if userInput <> '' then
			readString(userInput, number);
	end; (* getExtended *)

procedure getBoolean;
	var
		userInput: Str255;
	function translate (x: boolean): Str255;
		begin
			if x then
				translate := 'Y'
			else
				translate := 'N';
		end; (* translate *)
	begin
		repeat
			write(l, ' [', translate(bool), '] (Y/N)? ');
			readLn(userInput);
			if userInput = '' then
				userInput := translate(bool);
		until userInput[1] in ['y', 'Y', 'n', 'N'];
		bool := (userInput[1] = 'y') | (userInput[1] = 'Y');
	end; (* getBoolean *)
{$ENDC }

procedure shiftRight (var z: float);
	var
		i: integer;
	begin
		for i := z.precision - 1 downto 1 do
			z.mantissa[i + 1] := z.mantissa[i];
		z.mantissa[1] := 1;
		z.exponent := z.exponent + 1;
	end; (* shiftRight *)

procedure incrementMantissa (var z: float);
	var
		i: integer;
		carry: boolean;
	begin
		i := z.precision;
		carry := true;
		while carry & (i >= 1) do begin
			z.mantissa[i] := z.mantissa[i] + 1;
			carry := z.mantissa[i] = z.base;
			if carry then
				z.mantissa[i] := 0;
			i := i - 1;
		end;
		if carry then
			shiftRight(z);
	end; (* incrementMantissa *)

function yRemGreater1over2 (y: float; i: integer): boolean;
	var
		d: integer;
	begin
		d := y.base div 2;
		if (i < 1) | (i > y.precision) then
			yRemGreater1over2 := false
		else if odd(y.base) then begin
			while (i <= y.precision) & (y.mantissa[i] = d) do
				i := i + 1;
			yRemGreater1over2 := (i <= y.precision) & (y.mantissa[i] > d);
		end
		else if y.mantissa[i] > d then
			yRemGreater1over2 := true
		else if y.mantissa[i] = d then begin
			i := i + 1;
			while (i <= y.precision) & (y.mantissa[i] = 0) do
				i := i + 1;
			yRemGreater1over2 := i <= y.precision;
		end
		else
			yRemGreater1over2 := false;
	end; (* yRemGreater1over2 *)

function yRemEqual1over2 (y: float; i: integer): boolean;
	begin
		if (i < 1) | (i > y.precision) | odd(y.base) then
			yRemEqual1over2 := false
		else if y.mantissa[i] = y.base div 2 then begin
			i := i + 1;
			while (i <= y.precision) & (y.mantissa[i] = 0) do
				i := i + 1;
			yRemEqual1over2 := i > y.precision;
		end
		else
			yRemEqual1over2 := false;
	end; (* yRemEqual1over2 *)

procedure roundToPrecision (precision: integer; var y: float);
	begin
		if yRemGreater1over2(y, precision + 1) | odd(y.mantissa[precision]) & yRemEqual1over2(y, precision + 1) then begin
			y.precision := precision;
			incrementMantissa(y);
		end
		else
			y.precision := precision;
	end; (* roundToPrecision *)

procedure setPrecisionFloat;
	var
		i: integer;
	begin
		if precision >= maxLenMantissa then
			precision := maxLenMantissa - 1;
		if precision > y.precision then begin
			for i := y.precision + 1 to precision do
				y.mantissa[i] := 0;
			y.precision := precision;
		end
		else if precision < y.precision then
			roundToPrecision(precision, y);
	end; (* setPrecision *)

procedure coercePrecision (var x, y: float);
	begin
		if x.precision < y.precision then
			setPrecisionFloat(y.precision, x)
		else if y.precision < x.precision then
			setPrecisionFloat(x.precision, y);
		assert(x.base = y.base, 'x.base = y.base');
		if x.base <> y.base then begin
			if x.base < y.base then
				x.base := y.base
			else (* y.base < x.base *)
				y.base := x.base;
		end;
	end; (* coercePrecision *)

function power (b, e: integer): extended;

	function notNegativePower (b, e: integer): extended;
		begin
			if e = 0 then
				notNegativePower := 1.0
			else if odd(e) then
				notNegativePower := b * sqr(power(b, e div 2))
			else
				notNegativePower := sqr(power(b, e div 2));
		end; (* notNegativePower *)

	begin
		if e < 0 then
			power := 1.0 / notNegativePower(b, -e)
		else
			power := notNegativePower(b, e);
	end; (* power *)

{$IFC statisticsVersion }
procedure clearBin (var b: bin);
	var
		i: integer;
	begin
		with b do begin
			total := 0;
			for i := -maxBin to maxBin do
				counter[i] := 0;
		end;
	end; (* clearBin *)

procedure clearBins;
	var
		op: operation;
	begin
		for op := addition to conversion do
			clearBin(bins[op]);
	end; (* clearBins *)

procedure drawBins;
	var
		i, xValue, yValue: integer;
		max, total: longInt;
		op: operation;
		sum: bin;
		fileName: Str255;

	procedure writeSet (ops: operationSet);
		begin
			writeDraw('{ ');
			if addition in ops then
				writeDraw('+ ');
			if subtraction in ops then
				writeDraw('- ');
			if multiplication in ops then
				writeDraw('* ');
			if division in ops then
				writeDraw('/ ');
			if conversion in ops then
				writeDraw('rnd()');
			writeDraw(' }');
		end; (* writeSet *)

	begin
		clearBin(sum);
		for op := addition to conversion do
			if op in ops then
				with bins[op] do begin
					sum.total := sum.total + total;
					for i := -maxBin to maxBin do
						sum.counter[i] := sum.counter[i] + counter[i];
				end;
		if density then begin
			for i := -maxBin + 1 to maxBin do
				sum.counter[i] := sum.counter[i] + sum.counter[i - 1];
			max := sum.total;
		end
		else begin
			max := 0;
			for i := -maxBin to maxBin do
				if (sum.counter[i] > max) & (i <> 0) then
					max := sum.counter[i];
		end;
		if max > 0 then begin
			MoveTo(xOrg, yOrg);
			LineTo(xOrg + 2 * maxBin, yOrg);
			xValue := xOrg + maxBin + trunc(mark * maxBin);
			MoveTo(xValue, yOrg);
			LineTo(xValue, yOrg + 8);
			xValue := xOrg + maxBin - trunc(mark * maxBin);
			MoveTo(xValue, yOrg);
			LineTo(xValue, yOrg + 8);
			for i := -maxBin to maxBin do
				if (i <> 0) | density then begin
					yValue := yOrg - trunc(magnify * (sum.counter[i] / max));
					xValue := xOrg + maxBin + i;
					MoveTo(xValue, yValue);
					if density then
						LineTo(xValue, yValue - 1)
					else
						LineTo(xValue, yOrg);
				end;
		end;
		MoveTo(xOrg, yOrg + 20);
		if density then
			writeDraw('Rnd. err. dens. ')
		else if sum.total + sum.counter[0] > 0 then begin
			writeDraw('Rnd. err. dist. [', 100 * (sum.total / (sum.total + sum.counter[0])) : 1 : 2, '%] ');
		end;
		writeSet(ops);
		writeDraw(' N = ', sum.total : 1);
		if not density & ((sum.counter[-maxBin] > 0) | (sum.counter[maxBin] > 0)) then
			writeDraw(' (poss. probl. in floatArithmetic)');
		if save then begin
			write('Filename: (<ret> to discard) ');
			readLn(fileName);
			if fileName <> '' then
				SaveDrawing(fileName);
		end;
	end; (* drawBins *)

procedure calculateEps (op: operation; z: float; exact: extended);
	var
		adr: integer;
	begin
		if exact = 0 then
			epsilon := 0.0
		else
			epsilon := ((floatToReal(z) - exact) / exact) * 2 * power(z.base, z.precision - 1);
		adr := trunc(epsilon * maxBin);
		if adr > maxBin then
			adr := maxBin;
		if adr < -maxBin then
			adr := -maxBin;
		with bins[op] do begin
			total := total + 1;
			counter[adr] := counter[adr] + 1;
		end;
	end; (* calculateEps *)
{$ENDC }

function intToFloatF (x: longInt; base: baseType; precision: integer): float;
	var
		result: float;
	begin
		intToFloat(x, base, precision, result);
		intToFloatF := result;
	end;

function realToFloatF (x: extended; base: baseType; precision: integer): float;
	var
		result: float;
	begin
		realToFloat(x, base, precision, result);
		realToFloatF := result;
	end;

procedure intToFloat;
	var
		i: integer;
		digits: array[1..100] of integer;

{$IFC StatisticsVersion }
		xOrg: extended;
{$ENDC }

	begin
		y.precision := precision;
		y.base := base;

{$IFC StatisticsVersion }
		xOrg := x;
{$ENDC }

		if x = 0 then
			y.sign := 0
		else begin
			if x < 0 then begin
				y.sign := -1;
				x := -x;
			end
			else
				y.sign := 1;
			i := 0;
			repeat
				i := i + 1;
				digits[i] := x mod base;
				x := x div base;
			until x = 0;
			y.exponent := i;
			for i := 1 to y.exponent do
				y.mantissa[i] := digits[y.exponent - i + 1];
			for i := y.exponent + 1 to precision do
				y.mantissa[i] := 0;
			y.precision := y.exponent;
			roundToPrecision(precision, y);
		end;

{$IFC StatisticsVersion }
		if statisticsOn then
			calculateEps(conversion, y, xOrg);
{$ENDC }

	end; (* intToFloat *)

procedure realToFloat;
	var
		sign, offset: integer;

{$IFC StatisticsVersion }
		xOrg: extended;
		statisticsStatus: boolean;
{$ENDC }

	begin
		y.base := base;

{$IFC StatisticsVersion }
		xOrg := x;
{$ENDC }

		if x = 0 then begin
			y.precision := precision;
			y.sign := 0;
		end
		else begin
			if x < 0 then begin
				sign := -1;
				x := -x;
			end
			else
				sign := 1;
{$IFC THINK_PASCAL }
			offset := trunc(ln(2.0e9) / ln(base)) - trunc(ln(x) / ln(base) + 1);
{$ELSEC }
			offset := truncLong(ln(2.0e9) / ln(base)) - truncLong(ln(x) / ln(base) + 1);
{$ENDC }

{$IFC StatisticsVersion }
			statisticsStatus := statisticsOn;
			statisticsOn := false;
{$ENDC }

{$IFC THINK_PASCAL }
			intToFloat(trunc(x * power(base, offset)), base, precision, y);
{$ELSEC }
			intToFloat(truncLong(x * power(base, offset)), base, precision, y);
{$ENDC }

{$IFC StatisticsVersion }
			statisticsOn := statisticsStatus;
{$ENDC }

			scaleFloat(-offset, y);
			if sign = -1 then
				changeSignFloat(y);
		end;

{$IFC StatisticsVersion }
		if statisticsOn then
			calculateEps(conversion, y, xOrg);
{$ENDC }

	end; (* realToFloat *)

procedure lineToFloat;

	const
		endOfLine = '#';

	var
		i, j, exponent: integer;
		expSign: signType;
		ch: char;

	procedure nextCh;
		begin
			ch := l[i];
			if ('a' <= ch) & (ch <= 'z') then
				ch := chr(ord('A') + ord(ch) - ord('a'));
			i := i + 1;
		end; (* nextCh *)

	procedure enterDigits;
		begin
			while ('0' <= ch) & (ch < chr(ord('0') + base)) do begin
				y.mantissa[j] := ord(ch) - ord('0');
				j := j + 1;
				nextCh;
			end;
		end; (* enterDigits *)

	procedure getSign (var sign: signType);
		begin
			nextCh;
			if ch = '-' then begin
				sign := -1;
				nextCh;
			end
			else begin
				sign := 1;
				if ch = '+' then
					nextCh;
			end;
		end; (* getSign *)

	begin
		l := concat(l, endOfLine);
		i := 1;
		y.base := base;
		getSign(y.sign);
		while ch = '0' do
			nextCh;
		j := 1;
		enterDigits;
		y.exponent := j - 1;
		if ch = '.' then begin
			nextCh;
			if j = 1 then
				while ch = '0' do begin
					y.exponent := y.exponent - 1;
					nextCh;
				end;
			enterDigits;
		end;
		if j = 1 then begin
			y.sign := 0;
			y.precision := precision;
		end
		else begin
			while j <= precision do begin
				y.mantissa[j] := 0;
				j := j + 1;
			end;
			y.mantissa[j] := 0;
			y.precision := j;
			if ch = 'E' then begin
				getSign(expSign);
				exponent := 0;
				while ('0' <= ch) & (ch <= '9') do begin
					exponent := 10 * exponent + ord(ch) - ord('0');
					nextCh;
				end;
				y.exponent := y.exponent + expSign * exponent;
			end;
			roundToPrecision(precision, y);
		end;
	end; (* lineToFloat *)

function floatToReal;
	var
		i: integer;
		r: extended;
	begin
		if x.sign = 0 then
			floatToReal := 0.0
		else begin
			r := 0.0;
			for i := 1 to x.precision do
				r := x.base * r + x.mantissa[i];
			floatToReal := x.sign * r * power(x.base, x.exponent - x.precision);
		end;
	end; (* floatToReal *)

function floatToInt;
	var
		r: longInt;
		i: integer;

	function min (x, y: integer): integer;
		begin
			if x < y then
				min := x
			else
				min := y;
		end; (* min *)

	function power (b, e: integer): longInt;
		begin
			if e = 0 then
				power := 1
			else if odd(e) then
				power := b * sqr(power(b, e div 2))
			else
				power := sqr(power(b, e div 2));
		end; (* power *)

	begin
		if x.sign = 0 then
			floatToInt := 0
		else begin
			r := 0;
			for i := 1 to min(x.precision, x.exponent) do
				r := x.base * r + x.mantissa[i];
			if x.exponent > x.precision then
				r := r * power(x.base, x.exponent - x.precision);
			floatToInt := r * x.sign;
		end;
	end; (* floatToInt *)

function floatToLine;
	var
		result: Str255;
		i: integer;

	procedure append (x: Str255);
		begin
			result := concat(result, x);
		end; (* append *)

	function digit (x: integer): Str255;
		begin
			digit := chr(ord('0') + x);
		end; (* digit *)

	function intToLine (x: integer): Str255;

		function posIntToLine (x: integer): Str255;
			begin
				if x < 10 then
					posIntToLine := digit(x)
				else
					posIntToLine := concat(posIntToLine(x div 10), digit(x mod 10));
			end; (* posIntToLine *)

		begin
			if x < 0 then
				intToLine := concat('-', posIntToLine(-x))
			else
				intToLine := posIntToLine(x);
		end; (* intToLine *)

	procedure writeDigit (x: integer);
		begin
			if x < 10 then
				append(digit(x))
			else if x < 16 then
				append(chr(ord('A') - 10 + x))
			else
				append(concat('(', intToLine(x), ')'));
		end; (* writeDigit *)

	begin
		result := '';
		if x.sign = 0 then
			append('0')
		else begin
			if x.sign > 0 then
				append('+0.')
			else
				append('-0.');
			for i := 1 to x.precision do
				writeDigit(x.mantissa[i]);
			append(concat('e', intToLine(x.exponent)));
		end;
		floatToLine := result;
	end; (* floatToLine *)

procedure truncateFloat; (* removes the fractional part of x *)
	var
		i: integer;
	begin
		if (x.sign = 0) | (x.exponent <= 0) then
			y.sign := 0
		else begin
			y := x;
			for i := x.exponent + 1 to x.precision do
				y.mantissa[i] := 0;
		end;
	end; (* truncateFloat *)

procedure fractionFloat; (* removes the integral part of x *)
	var
		i, j: integer;
	begin
		if (x.sign = 0) | (x.exponent >= x.precision) then
			y.sign := 0
		else begin
			y := x;
			if x.exponent > 0 then begin
				i := x.exponent + 1;
				y.exponent := 0;
				while (i <= x.precision) & (x.mantissa[i] = 0) do
					i := i + 1;
				if i > x.precision then
					y.sign := 0
				else
					for j := 1 to x.precision do
						if j + i - 1 <= x.precision then
							y.mantissa[j] := x.mantissa[j + i - 1]
						else
							y.mantissa[j] := 0;
			end;
		end;
	end; (* fractionFloat *)

procedure scaleFloat;
	begin
		y.exponent := y.exponent + e;
	end; (* scaleFloat *)

function getExponentFloat;
	begin
		getExponentFloat := x.exponent;
	end; (* getExponentFloat *)

procedure setExponentFloat;
	begin
		x.exponent := e;
	end; (* setExponentFloat *)

procedure changeSignFloat;
	begin
		y.sign := -y.sign;
	end; (* changeSignFloat *)

procedure absFloat;
	begin
		y.sign := abs(y.sign);
	end; (* absFloat *)

procedure nextFloat;
	begin
		if y.sign = 0 then
			machineEpsilon(y.base, y.precision, y)
		else
			incrementMantissa(y);
	end; (* nextFloat *)

procedure machineEpsilon (base: baseType; precision: integer; var y: float);
	var
		one, two: float;
	begin
		intToFloat(1, base, precision, one);
		intToFloat(2, base, precision, two);
		divideFloat(one, two, y);
		scaleFloat(1 - precision, y);
	end; (* machineEpsilon *)

procedure addFloat;
	var
		i, t, tr, s: integer;
		carry, equal, temp: boolean;

{$IFC statisticsVersion }
		exact: extended;
		op: operation;
{$ENDC }

	procedure exchange (var x, y: float);
		var
			t: float;
		begin
			t := x;
			x := y;
			y := t;
		end; (* exchange *)

	function yRemNotEqualZero (y: float; i: integer): boolean;
		begin
			if i < 1 then
				i := 1;
			while (i <= y.precision) & (y.mantissa[i] = 0) do
				i := i + 1;
			yRemNotEqualZero := i <= y.precision;
		end; (* yRemNotEqualZero *)

	begin
		coercePrecision(x, y);

{$IFC statisticsVersion }
		if statisticsOn then begin
			exact := floatToReal(x) + floatToReal(y);
			if x.sign = y.sign then
				op := addition
			else
				op := subtraction;
		end;
{$ENDC }

		z.precision := x.precision;
		z.base := x.base;
		if x.sign = 0 then
			z := y
		else if y.sign = 0 then
			z := x
		else if x.sign = y.sign then begin
			if x.exponent < y.exponent then
				exchange(x, y);
			t := x.precision - x.exponent + y.exponent;
			tr := t; (* for rounding *)
			z.sign := x.sign;
			z.exponent := x.exponent;
			z.precision := x.precision;
			carry := false;
			for i := x.precision downto 1 do begin
				s := x.mantissa[i] + ord(carry);
				if t >= 1 then
					s := s + y.mantissa[t];
				t := t - 1;
				carry := s >= x.base;
				if carry then
					z.mantissa[i] := s - x.base
				else
					z.mantissa[i] := s;
			end;
			if carry then (* carry = true => tr >= 1 *)
				begin
				y.mantissa[tr] := z.mantissa[z.precision];
				shiftRight(z);
			end
			else
				tr := tr + 1;
			if yRemGreater1over2(y, tr) | odd(z.mantissa[z.precision]) & yRemEqual1over2(y, tr) then
				incrementMantissa(z);
		end
		else begin
			z.sign := x.sign;
			equal := false;
			if x.exponent < y.exponent then begin
				exchange(x, y);
				z.sign := -z.sign;
			end
			else if x.exponent = y.exponent then begin
				i := 1;
				while (i <= x.precision) & (x.mantissa[i] = y.mantissa[i]) do
					i := i + 1;
				if i > x.precision then
					equal := true
				else if x.mantissa[i] < y.mantissa[i] then begin
					exchange(x, y);
					z.sign := -z.sign;
				end;
			end;
			if equal then
				z.sign := 0
			else begin
				setPrecisionFloat(x.precision + 2, x);
				setPrecisionFloat(x.precision + 2, y);
				z.precision := x.precision;
				t := x.precision - x.exponent + y.exponent;
				carry := yRemNotEqualZero(y, t + 1);
				for i := x.precision downto 1 do begin
					s := x.mantissa[i] - ord(carry);
					if t >= 1 then
						s := s - y.mantissa[t];
					t := t - 1;
					carry := s < 0;
					if carry then
						z.mantissa[i] := s + x.base
					else
						z.mantissa[i] := s;
				end; (* assert(carry = false) *)
				t := 1;
				while (t <= x.precision) & (z.mantissa[t] = 0) do
					t := t + 1; (* assert(t<=x.precision) *)
				if t > 1 then begin
					for i := 1 to x.precision do
						if i + t - 1 > x.precision then
							z.mantissa[i] := 0
						else
							z.mantissa[i] := z.mantissa[i + t - 1];
					z.exponent := x.exponent - t + 1;
				end
				else
					z.exponent := x.exponent;
				temp := odd(z.mantissa[z.precision - 2]) | not yRemEqual1over2(y, x.precision - x.exponent + y.exponent + t - 2);
				if yRemGreater1over2(z, z.precision - 1) | yRemEqual1over2(z, z.precision - 1) & temp then begin
					z.precision := z.precision - 2;
					incrementMantissa(z);
				end
				else
					z.precision := z.precision - 2;
			end;
		end;

{$IFC statisticsVersion }
		if statisticsOn then
			calculateEps(op, z, exact);
{$ENDC }

	end; (* addFloat *)

procedure subtractFloat;
	begin
		changeSignFloat(y);
		addFloat(x, y, z);
	end; (* subtractFloat *)

procedure multiplyFloat;
	var
		accu: array[1..doubleMaxLenMantissa] of integer;
		i, j, carry, p: integer;
		t: float;
	begin
		coercePrecision(x, y);
		z.precision := x.precision;
		z.base := x.base;
		if (x.sign = 0) | (y.sign = 0) then
			z.sign := 0
		else begin
			for i := 1 to 2 * x.precision do
				accu[i] := 0;
			for i := x.precision downto 1 do begin
				carry := 0;
				for j := y.precision downto 1 do begin
					p := x.mantissa[i] * y.mantissa[j] + carry + accu[i + j];
					accu[i + j] := p mod x.base;
					carry := p div x.base;
				end;
				accu[i] := carry;
			end;
			j := ord(accu[1] = 0);
			for i := 1 to x.precision do
				z.mantissa[i] := accu[i + j];
			z.exponent := x.exponent + y.exponent - j;
			z.sign := x.sign * y.sign;
			t.precision := x.precision;
			t.base := x.base;
			for i := 1 to t.precision do
				if i + j <= x.precision then
					t.mantissa[i] := accu[i + j + x.precision]
				else
					t.mantissa[i] := 0;
			if yRemGreater1over2(t, 1) | yRemEqual1over2(t, 1) & odd(z.mantissa[z.precision]) then
				incrementMantissa(z);
		end;

{$IFC statisticsVersion }
		if statisticsOn then
			calculateEps(multiplication, z, floatToReal(x) * floatToReal(y));
{$ENDC }

	end; (* multiplyFloat *)

procedure divideFloat;
	var
		i, j, delta, m, p, q, r, s, ym1, ym2, scaleFactor: integer;
		carry: boolean;
		u: array[0..doubleMaxLenMantissa] of integer;

{$IFC statisticsVersion }
		exact: extended;
{$ENDC }

	procedure scaleVector (var x: float; d: integer; var s: integer);
		var
			i: integer;
		begin
			s := 0;
			for i := x.precision downto 1 do begin
				s := s + x.mantissa[i] * d;
				x.mantissa[i] := s mod x.base;
				s := s div x.base;
			end;
		end; (* scaleVector *)

	begin
		coercePrecision(x, y);

{$IFC statisticsVersion }
		if statisticsOn then
			exact := floatToReal(x) / floatToReal(y);
{$ENDC }

		z.precision := x.precision;
		z.base := x.base;
		z.sign := x.sign * y.sign;
		if (y.sign <> 0) & (x.sign <> 0) then
			if x.precision = 1 then begin
				delta := ord(x.mantissa[1] >= y.mantissa[1]);
				p := (x.base * (1 - delta) + delta) * x.mantissa[1];
				z.mantissa[1] := p div y.mantissa[1];
				r := p mod y.mantissa[1];
				z.exponent := x.exponent - y.exponent + delta;
				if (2 * r > y.mantissa[1]) | (2 * r = y.mantissa[1]) & odd(z.mantissa[z.precision]) then
					incrementMantissa(z);
			end
			else begin
				scaleFactor := x.base div (y.mantissa[1] + 1);
				if scaleFactor = 1 then
					r := 0
				else begin
					scaleVector(y, scaleFactor, r);
					scaleVector(x, scaleFactor, r);
				end;
				u[0] := r;
				for i := 1 to x.precision do begin
					u[i] := x.mantissa[i];
					u[i + x.precision] := 0;
				end;
				j := 0;
				delta := 0;
				m := x.precision - 1;
				ym1 := y.mantissa[1];
				ym2 := y.mantissa[2];
				repeat
					if u[j] = ym1 then
						q := x.base - 1
					else
						q := (u[j] * x.base + u[j + 1]) div ym1;
					while ym2 * q > (u[j] * x.base + u[j + 1] - q * ym1) * x.base + u[j + 2] do
						q := q - 1;
					s := 0;
					for i := j + x.precision downto j do begin
						s := u[i] - s;
						if i > j then
							s := s - q * y.mantissa[i - j];
						if s < 0 then
							if (-s) mod x.base = 0 then begin
								u[i] := 0;
								s := (-s) div x.base;
							end
							else begin
								u[i] := x.base - (-s) mod x.base;
								s := (-s) div x.base + 1;
							end
						else begin
							u[i] := s;
							s := 0;
						end;
					end;
					if s > 0 then begin
						q := q - 1;
						carry := false;
						for i := j + x.precision downto j + 1 do begin
							s := ord(carry) + u[i] + y.mantissa[i - j];
							carry := s >= x.base;
							if carry then
								u[i] := s - x.base
							else
								u[i] := s;
						end;
					end;
					j := j + 1;
					if (j = 1) & (q = 0) then
						delta := 1
					else
						z.mantissa[j - delta] := q;
				until j > m + delta;
				z.exponent := x.exponent - y.exponent + 1 - delta;
				carry := false;
				for i := j + x.precision - 1 downto j do begin
					s := ord(carry) + 2 * u[i];
					carry := s >= x.base;
					if carry then
						u[i] := s - x.base
					else
						u[i] := s;
				end;
				if carry then
					incrementMantissa(z)
				else begin
					i := 1;
					while (i <= x.precision) & (u[i + j - 1] = y.mantissa[i]) do
						i := i + 1;
					if (i <= x.precision) & (u[i + j - 1] > y.mantissa[i]) | odd(z.mantissa[z.precision]) & (i > x.precision) then
						incrementMantissa(z);
				end;
			end;

{$IFC statisticsVersion }
		if statisticsOn then
			calculateEps(division, z, exact);
{$ENDC }

	end; (* divideFloat *)

function addF;
	var
		z: float;
	begin
		addFloat(x, y, z);
		addF := z;
	end; (* addF *)

function subF;
	var
		z: float;
	begin
		subtractFloat(x, y, z);
		subF := z;
	end; (* subF *)

function mulF;
	var
		z: float;
	begin
		multiplyFloat(x, y, z);
		mulF := z;
	end; (* mulF *)

function divF;
	var
		z: float;
	begin
		divideFloat(x, y, z);
		divF := z;
	end; (* divF *)

function absF;
	begin
		x.sign := abs(x.sign);
		absF := x;
	end; (* absF *)

function chsF;
	begin
		x.sign := -x.sign;
		chsF := x;
	end; (* chsF *)

function sqrF;
	var
		y: float;
	begin
		multiplyFloat(x, x, y);
		sqrF := y;
	end; (* sqrF *)

function minF;
	begin
		if compareFloat(x, y) = lessFloat then
			minF := x
		else
			minF := y;
	end; (* minF *)

function maxF;
	begin
		if compareFloat(x, y) = greaterFloat then
			maxF := x
		else
			maxF := y;
	end; (* minF *)

function compareFloat;

	function sign: integer;
		var
			i: integer;

		function signum (x, v: integer): integer;
			begin
				if x > 0 then
					signum := v
				else
					signum := -v;
			end; (* signum *)

		begin
			if x.sign = y.sign then
				if x.sign = 0 then
					sign := 0
				else if x.exponent = y.exponent then begin
					i := 1;
					while (i <= x.precision) & (x.mantissa[i] = y.mantissa[i]) do
						i := i + 1;
					if i > x.precision then
						sign := 0
					else
						sign := signum(x.mantissa[i] - y.mantissa[i], x.sign);
				end
				else
					sign := signum(x.exponent - y.exponent, x.sign)
			else
				sign := signum(x.sign - y.sign, 1);
		end; (* sign *)

	begin
		case sign of
			-1: 
				compareFloat := lessFloat;
			0: 
				compareFloat := equalFloat;
			1: 
				compareFloat := greaterFloat;
		end;
	end; (* compareFloat *)

function equalZeroFloat;
	begin
		equalZeroFloat := x.sign = 0;
	end; (* equalZeroFloat *)

function sign (x: extended): signType;
	begin
		if x < 0 then
			sign := -1
		else if x > 0 then
			sign := +1
		else
			sign := 0;
	end;

function min (a, b: extended): extended;
	begin
		if a < b then
			min := a
		else
			min := b;
	end;

function max (a, b: extended): extended;
	begin
		if a > b then
			max := a
		else
			max := b;
	end;

end. (* floatArithmetic *)
