(*	XYZ GeoBench (eXperimental geometrY Zurich):		*)
(*	An object oriented, interactive workbench for		*)
(*	experimental geometric computation.					*)
(*														*)
(*	 1999 by Peter Schorn, peter.schorn@acm.org		*)
(*	Culmannstrasse 77, CH-8006 Zrich, Switzerland		*)
(*	This software may be copied freely for educational	*)
(*	and research purposes only. For commercial use,		*)
(*	please contact the author.							*)
(*														*)
(*	Version 5.0.5: 30-Sep-1999							*)

unit generalExperiment;

interface

uses
	pointVector, GFUtilities;

const
	maxParameters = 16;

type
	parameterIndex = 1..maxParameters;
	parameterIndex0 = 0..maxParameters;
	parameterArray = array[parameterIndex] of extended;
	parameterVector = record
			length: parameterIndex0;			(*	Number of parameters. *)
			parameters: parameterArray;
		end;

procedure doGeneralExperiment (experimentFileName: Str255; experimentFileVolume: integer; experimentTypeName: Str255; function generate (p: parameterVector): obj; function test (o: obj): parameterVector; function compute (o: obj; q: parameterVector; var success: boolean): parameterVector);
(* The result variable success specifies whether the computation was successful.						*)
(*	Only successful computations are counted in the statistics. The procedure doGeneralExperiment	*)
(*	performs the experiment according to the specification in the file experimentFileName on			*)
(*	volume experimentFileVolume. experimentTypeName must agree with the respective parameter	*)
(*	of the description file. *)

procedure doGeneralExperimentAsk (experimentTypeName: Str255; function generate (p: parameterVector): obj; function test (o: obj): parameterVector; function compute (o: obj; q: parameterVector; var success: boolean): parameterVector);
(* Same as doGeneralExperiment but the user is asked for the file name of the file which contains		*)
(*	the experiment description. *)

procedure initGeneralExperiment;

implementation

{$SETC CheckAllocation = True }

const
	maxDictionaryEntries = 50;
	minimumColumnFieldLength = 15;
	headerColumnFieldLength = 36;
	outputDecimalDigits = 8;

type
	rangePtr = ^range;
	range = record
			fromValue, byValue, toValue, currentValue: extended;
			currentCount, toCount: longInt;
			next: rangePtr;
		end;

	parameterDefinitionRecord = record
			parameterName: Str255;
			hasColumn: boolean;
			rangeListHead, currentRange: rangePtr;
			columnWidth: integer;
		(* the column width is set by 'printChangedParametersHeader' and is used by 'printChangedParameters'. *)
		end;

	parameterDefinitionArray = array[parameterIndex] of parameterDefinitionRecord;

	parameterDefinitionVector = record
			length: parameterIndex0;
			parameterDefinitions: parameterDefinitionArray;
		end;

	statisticsDefinitionRecord = record
			sum, sumSquare, minimum, maximum: extended;
			values: realArrayH;
			used: boolean;
		end;

	statisticsDefinitionArray = array[parameterIndex] of statisticsDefinitionRecord;

	resultStatistics = record
			numberOfSuccesses: longInt;
			length: parameterIndex0;
			statistics: statisticsDefinitionArray;
		end;

	resultStatisticsArray = array[seqIndex] of resultStatistics;
	resultStatisticsArrayP = ^resultStatisticsArray;
	resultStatisticsArrayH = ^resultStatisticsArrayP;

	resultDefinitionArray = array[parameterIndex] of record
			parameterName: Str255;
			createHistogram: boolean;
			histogramNumberOfBins: longInt;
			meanColumnWidth, stdDevColumnWidth, meanNoExtColumnWidth, stdDevNoExtColumnWidth, minimumColumnWidth, maximumColumnWidth: integer;
		(* the column widths are set by 'printStatisticsHeader' and is used by 'printStatistics'. *)
		end;

	resultDefinitionVector = record
			successColumnWidth: integer;
			length: parameterIndex0;
			resultDefinitions: resultDefinitionArray;
		end;

	experimentPtr = ^experiment;
	experiment = record
			experimentName, resultFileName: Str255;
			verbose, showArguments, newResultFileName, appendResult: boolean;
			randomConstraint: obj;
			numberOfExperiments, backgroundPriority: longInt;
			generationRandomSeed, computationRandomSeed: extended;
			generation, computation: parameterDefinitionVector;
			generationResult, computationResult: resultDefinitionVector;
			next: experimentPtr;
		end;

	experimentInformation = record
			experimentTypeName, debugFileName: Str255;
			oldBackgroundPriority: longInt;
			oldRandomSeed: extended;
			experimentList: experimentPtr;
		end;

	simpleText = packed file of Byte;

	simpleDictionaryIndex0 = 0..maxDictionaryEntries;
	simpleDictionaryIndex = 1..maxDictionaryEntries;
	simpleDictionary = record
			length: simpleDictionaryIndex0;
			information: array[simpleDictionaryIndex] of record
					name: Str255;
					value: extended;
				end;
		end;

procedure initSimpleDictionary (var d: simpleDictionary);
	begin
		d.length := 0;
	end;

procedure mark (var d: simpleDictionary; var m: simpleDictionaryIndex0);
	begin
		m := d.length;
	end;

procedure release (var d: simpleDictionary; m: simpleDictionaryIndex0);
	begin
		d.length := m;
	end;

function lookup (var d: simpleDictionary; name: Str255; var value: extended): boolean; (* true iff not found *)
	var
		i: simpleDictionaryIndex0;
	begin
		name := capString(name);
		i := d.length;
		while (i > 0) & (d.information[i].name <> name) do
			i := i - 1;
		if i > 0 then
			value := d.information[i].value;
		lookup := i = 0;
	end;

procedure insertNew (var d: simpleDictionary; newName: Str255; newValue: extended);
	var
		oldValue: extended;
	begin
		if d.length < maxDictionaryEntries then
			if lookup(d, newName, oldValue) then begin
				d.length := d.length + 1;
				with d.information[d.length] do begin
					name := capString(newName);
					value := newValue;
				end;
			end
			else
				nonFatalError(StringOf('(', truncate(newName, 10), ', ', oldValue, ') already declared.'))
		else
			nonFatalError(StringOf('Dictionary full. No room for (', truncate(newName, 10), ', ', newValue, ').'));
	end;

	var
		stopCharacters: set of char;

procedure getNextString (var x: Str255);
	begin
		skipWhiteSpace;
		x := ch;
		getCh;
		while not (ch in stopCharacters) & (length(x) < 250) do begin
			x := concat(x, ch);
			getCh;
		end;
		testString(x);
	end;

{$IFC ChecksEnabled AND CheckAllocation }
	var
		alloc: longInt;
{$ENDC }

procedure disposeRanges (var v: parameterDefinitionVector);
	var
		i: parameterIndex0;
		t, d: rangePtr;
	begin
		for i := 1 to v.length do begin
			t := v.parameterDefinitions[i].rangeListHead;
			while t <> nil do begin
				d := t;
				t := t^.next;
				dispose(d);
{$IFC ChecksEnabled AND CheckAllocation }
				alloc := alloc - 1;
{$ENDC }
			end;
		end;
	end;

procedure disposeExperimentInformation (var e: experimentInformation);
	var
		t, d: experimentPtr;
	begin
		t := e.experimentList;
		while t <> nil do begin
			d := t;
			with d^ do begin
				disposeRanges(generation);
				disposeRanges(computation);
				if randomConstraint <> nil then
					randomConstraint.freeAll;
				t := next;
			end;
			dispose(d);
{$IFC ChecksEnabled AND CheckAllocation }
			alloc := alloc - 1;
{$ENDC }
		end;
	end;

function testFiles (experimentFileVolume: integer; e: experimentPtr): boolean;
	var
		directoryExists, fileExists: boolean;
	begin
		testFiles := false;
		while e <> nil do
			with e^ do begin
				testPathName(resultFileName, experimentFileVolume, directoryExists, fileExists);
				if directoryExists then
					if newResultFileName & fileExists & not askUserBoolean(StringOf('File ''', resultFileName, ''' exists. Continue and overwrite?')) then
						exit(testFiles)
					else
				else begin
					nonFatalError(StringOf('Enclosing directory in ''', resultFileName, ''' does not exist.'));
					exit(testFiles);
				end;
				e := next;
			end;
		testFiles := true;
	end;

function readExperimentFile (experimentFileName: Str255; experimentFileVolume: integer; experimentTypeName: Str255; var e: experimentInformation): boolean;

	const
		maxParameters1 = maxParameters - 1;

	var
		s: Str255;
		d: simpleDictionary;
		dictionaryMark: simpleDictionaryIndex0;
		lastExperiment, currentExperiment: experimentPtr;

	procedure cleanUp (errorMessage: Str255);
		begin
			nonFatalError(errorMessage);
			closeReadStream;
			disposeExperimentInformation(e);
{$IFC ChecksEnabled AND CheckAllocation }
			assert(alloc = 0, StringOf('alloc = 0 (', alloc : 1, ')'));
{$ENDC }
			exit(readExperimentFile);
		end;

	procedure conditionalCleanUp (condition: boolean; errorMessage: Str255);
		begin
			if not condition then
				cleanUp(errorMessage);
		end;

	procedure getNoRead (x: Str255);
		begin
			if not (capString(s) = capString(x)) then
				cleanUp(StringOf('Expected ''', x, '''. Got ''', truncate(s, 10), '''.'));
		end;

	procedure get (x: Str255);
		begin
			getNextString(s);
			getNoRead(x);
		end;

	procedure tooManyParameters (partialErrorMessage: Str255);
		begin
			cleanUp(StringOf(' Number of ''', partialErrorMessage, ''' parameters exceeds ', maxParameters1 : 1, ' . '));
		end;

	procedure lookUpString (var d: simpleDictionary; var s: Str255; var x: extended);
		begin
			if lookup(d, s, x) then begin
				IOCheck(false);
				readString(s, x);
				conditionalCleanUp(ioResult <= 0, concat('Illegal extended number, got ''', truncate(s, 10), '''.'));
				IOCheck(true);
			end;
		end;

	procedure readTextExtendedDictionary (var d: simpleDictionary; var x: extended);
		begin
			getNextString(s);
			lookUpString(d, s, x);
		end;

	procedure readDeclarationSection (var d: simpleDictionary; stopString: Str255);
		var
			name: Str255;
			value: extended;
		begin
			getNextString(s);
			while capString(s) <> stopString do begin
				name := s;
				get('=');
				readTextExtendedDictionary(d, value);
				insertNew(d, name, value);
				getNextString(s);
			end;
		end;

	procedure readRangeList (var x: parameterDefinitionRecord);
		var
			t, last: rangePtr;
		begin
			with x do begin
				parameterName := s;
				rangeListHead := nil;
				last := nil;
				getNextString(s);
				if capString(s) = capString('Column') then begin
					hasColumn := true;
					getNextString(s);
				end
				else
					hasColumn := false;
				getNoRead('=');
				repeat
					getNextString(s);
					new(t);
					failNil(Handle(t));
{$IFC ChecksEnabled AND CheckAllocation }
					alloc := alloc + 1;
{$ENDC }
					if last = nil then
						rangeListHead := t
					else
						last^.next := t;
					last := t;
					with t^ do begin
						next := nil;
						if capString(s) = capString('From') then begin
							readTextExtendedDictionary(d, fromValue);
							get('By');
							readTextExtendedDictionary(d, byValue);
							get('To');
							readTextExtendedDictionary(d, toValue);
							conditionalCleanUp((byValue = 0) = (fromValue = toValue), StringOf('(By = 0)  (', fromValue, ' = ', toValue, ').'));
							conditionalCleanUp((fromValue < toValue) <= (byValue > 0), StringOf('Positive ''By'' needed. Got ', byValue, '.'));
							conditionalCleanUp((fromValue > toValue) <= (byValue < 0), StringOf('Negative ''By'' needed. Got ', byValue, '.'));
						end
						else begin
							lookUpString(d, s, fromValue);
							byValue := 0;
							toValue := fromValue;
						end;
					end;
					getNextString(s);
				until s <> ',';
			end;
		end;

	procedure readParameterDefinitionVector (var v: parameterDefinitionVector; stopString, partialErrorMessage: Str255);
		begin
			getNextString(s);
			v.length := 0;
			while capString(s) <> stopString do
				if v.length < maxParameters1 then begin
					v.length := v.length + 1;
					readRangeList(v.parameterDefinitions[v.length]);
				end
				else
					tooManyParameters(partialErrorMessage);
		end;

	procedure readResultDefinitionVector (var v: resultDefinitionVector; stopString1, stopString2, partialErrorMessage, time: Str255);
		var
			capS: Str255;
			t: extended;
		begin
			getNextString(s);
			capS := capString(s);
			v.length := 0;
			while (capS <> stopString1) & (capS <> stopString2) do
				if v.length < maxParameters1 then begin
					v.length := v.length + 1;
					with v.resultDefinitions[v.length] do begin
						parameterName := s;
						getNextString(s);
						capS := capString(s);
						if capS = capString('Histogram') then begin
							createHistogram := true;
							readTextExtendedDictionary(d, t);
							conditionalCleanUp((t > 0) & (trunc(t) = t), StringOf('Positive integer expected. Got ', t, '.'));
{$IFC THINK_PASCAL }
							histogramNumberOfBins := trunc(t);
{$ELSEC }
							histogramNumberOfBins := truncLong(t);
{$ENDC }
							getNextString(s);
							capS := capString(s);
						end
						else begin
							createHistogram := false;
							histogramNumberOfBins := 0;
						end;
					end;
				end
				else
					tooManyParameters(partialErrorMessage);
			v.length := v.length + 1;
			with v.resultDefinitions[v.length] do begin
				parameterName := time;
				createHistogram := false;
				histogramNumberOfBins := 0;
			end;
		end;

	function readRandomConstraint: obj;
		begin
			get('=');
			readRandomConstraint := readTextObj;
			getNextString(s);
		end;

	procedure testRandomSeed (name: Str255; var value: extended);
		begin
			if capString(s) = name then begin
				get('=');
				readTextExtendedDictionary(d, value);
				getNextString(s);
			end
			else
				value := getRandomSeed;
		end;

	function readBackgroundPriority: longInt;
		var
			x: extended;
		begin
			get('=');
			readTextExtendedDictionary(d, x);
{$IFC THINK_PASCAL }
			readBackgroundPriority := trunc(x);
{$ELSEC }
			readBackgroundPriority := truncLong(x);
{$ENDC }
			getNextString(s);
		end;

	procedure readResultFileName (var resultFileName: Str255; var newResultFileName, appendResult: boolean);
		var
			capS: Str255;
		begin
			get('ResultFileName');
			get('=');
			getNextString(s);
			capS := capString(s);
			newResultFileName := capS = capString('New');
			appendResult := capS = capString('Append');
			if newResultFileName | appendResult then
				getNextString(resultFileName)
			else
				resultFileName := s;
		end;

	procedure getBoolean (name: Str255; default: boolean; var value: boolean);
		var
			v: extended;
		begin
			if capString(s) = capString(name) then begin
				get('=');
				readTextExtendedDictionary(d, v);
				conditionalCleanUp((v = 0) | (v = 1), StringOf('0 or 1 expected. Got ', v, '.'));
				value := v = 1;
				getNextString(s);
			end
			else
				value := default;
		end;

	procedure readSingleExperiment (var e: experiment);
		var
			v: extended;
		begin
			getNoRead('ExperimentName');
			get('=');
			getNextString(e.experimentName);

			get('NumberOfExperiments');
			get('=');
			readTextExtendedDictionary(d, v);
			conditionalCleanUp((v >= 0) & (trunc(v) = v), StringOf('Non-negative integer expected. Got ', v, '.'));
{$IFC THINK_PASCAL }
			e.numberOfExperiments := trunc(v);
{$ELSEC }
			e.numberOfExperiments := truncLong(v);
{$ENDC }

			readResultFileName(e.resultFileName, e.newResultFileName, e.appendResult);

			getNextString(s);
			getBoolean('Verbose', false, e.verbose);
			getBoolean('ShowArguments', false, e.showArguments);

			if capString(s) = capString('RandomConstraint') then
				e.randomConstraint := readRandomConstraint
			else
				e.randomConstraint := nil;

			testRandomSeed(capString('GenerationRandomSeed'), e.generationRandomSeed);
			testRandomSeed(capString('ComputationRandomSeed'), e.computationRandomSeed);

			if capString(s) = capString('BackgroundPriority') then
				e.backgroundPriority := readBackgroundPriority
			else
				e.backgroundPriority := getAbortTime;

			if capString(s) = capString('LocalDeclaration') then
				readDeclarationSection(d, capString('Generation'));

			getNoRead('Generation');
			readParameterDefinitionVector(e.generation, capString('GenerationResult'), 'Generation');
			readResultDefinitionVector(e.generationResult, capString('Computation'), '', 'GenerationResult', 'GenerationTime');
			readParameterDefinitionVector(e.computation, capString('ComputationResult'), 'Computation');
			readResultDefinitionVector(e.computationResult, capString('ExperimentEnd'), capString('ExperimentName'), 'ComputationResult', 'ComputationTime');
		end;

	begin (* readExperimentFile *)
		readExperimentFile := false;
		openReadStream(experimentFileName, experimentFileVolume);
		initLookAhead;
		initSimpleDictionary(d);
		insertNew(d, 'false', 0);
		insertNew(d, 'true', 1);
		e.experimentList := nil;
		get('ExperimentBegin');
		get(experimentTypeName);
		e.experimentTypeName := experimentTypeName;
		getNextString(s);
		if capString(s) = capString('ParameterLog') then begin
			get('=');
			getNextString(e.debugFileName);
			getNextString(s);
		end
		else
			e.debugFileName := '';
		if capString(s) = capString('GlobalDeclaration') then
			readDeclarationSection(d, capString('ExperimentName'));
		mark(d, dictionaryMark);
		while capString(s) <> capString('ExperimentEnd') do begin
			release(d, dictionaryMark);
			new(currentExperiment);
			failNil(Handle(currentExperiment));
{$IFC ChecksEnabled AND CheckAllocation }
			alloc := alloc + 1;
{$ENDC }
			with currentExperiment^ do begin
				generation.length := 0;
				computation.length := 0;
				randomConstraint := nil;
				next := nil;
			end;
			if e.experimentList = nil then
				e.experimentList := currentExperiment
			else
				lastExperiment^.next := currentExperiment;
			lastExperiment := currentExperiment;
			readSingleExperiment(currentExperiment^);
		end;
		closeReadStream;
		if testFiles(experimentFileVolume, e.experimentList) then
			readExperimentFile := true
		else
			disposeExperimentInformation(e);
	end;

procedure doGeneralExperimentAsk (experimentTypeName: Str255; function generate (p: parameterVector): obj; function test (o: obj): parameterVector; function compute (o: obj; q: parameterVector; var success: boolean): parameterVector);
	var
		experimentFileName: Str255;
		experimentFileVolume: integer;
	begin
		if SELECTfile(experimentFileName, experimentFileVolume, 'TEXT') then
			doGeneralExperiment(experimentFileName, experimentFileVolume, experimentTypeName, generate, test, compute);
	end;

function safeConcat (a, b: Str255): Str255;
	const
		maxStringLength = 255;
	var
		la, lb: integer;
		result: Str255;
	begin
		la := length(a);
		lb := length(b);
		if la + lb <= maxStringLength then
			safeConcat := concat(a, b)
		else begin
			if la = maxStringLength then
				result := a
			else
				result := concat(a, copy(b, 1, maxStringLength - length(a)));
			result[maxStringLength] := '';
			safeConcat := result;
		end;
	end;

function fixedParameter (var x: parameterDefinitionRecord): boolean;
	begin
		with x, rangeListHead^ do
			fixedParameter := (next = nil) & (fromValue = toValue) & not hasColumn;
	end;

function showParameter (var v: parameterDefinitionVector): Str255;
	var
		i: parameterIndex0;
		result: Str255;
	begin
		result := '';
		for i := v.length downto 1 do
			if not fixedParameter(v.parameterDefinitions[i]) then
				with v.parameterDefinitions[i], currentRange^ do
					result := safeConcat(result, StringOf(' ', truncate(parameterName, 10), '=', currentValue));
		showParameter := copy(result, 2, length(result));
	end;

procedure initCurrentValue (var v: parameterDefinitionVector);
	var
		i: parameterIndex0;
		t: rangePtr;
	begin
		for i := 1 to v.length do
			with v.parameterDefinitions[i] do begin
				currentRange := rangeListHead;
				t := rangeListHead;
				while t <> nil do
					with t^ do begin
						currentValue := fromValue;
						currentCount := 1;
						t := next;
					end;
			end;
	end;

function continue (var x: parameterDefinitionRecord): boolean;
	begin
		with x, x.currentRange^ do
			if currentCount >= toCount then begin
				currentValue := fromValue;
				currentCount := 1;
				currentRange := currentRange^.next;
				if currentRange = nil then begin
					currentRange := rangeListHead;
					continue := true;
				end
				else
					continue := false;
			end
			else begin
				currentValue := currentValue + byValue;
				currentCount := currentCount + 1;
				continue := false;
			end;
	end;

function nextPoint (var v: parameterDefinitionVector): boolean;
	var
		i: parameterIndex0;
	begin
		i := v.length;
		while (i > 0) & continue(v.parameterDefinitions[i]) do
			i := i - 1;
		nextPoint := i = 0;
	end;

function extractParameters (var v: parameterDefinitionVector): parameterVector;
	var
		result: parameterVector;
		i: parameterIndex0;
	begin
		result.length := v.length;
		for i := 1 to v.length do
			result.parameters[i] := v.parameterDefinitions[i].currentRange^.currentValue;
		extractParameters := result;
	end;

function addBlanksRight (s: Str255; l: integer): Str255;
	begin
		s := truncate(s, l);
		while length(s) < l do
			s := concat(s, ' ');
		addBlanksRight := s;
	end;

function addBlanksLeft (l: integer; s: Str255): Str255;
	var
		t: Str255;
	begin
		t := '';
		while length(t) < l do
			t := concat(t, ' ');
		addBlanksLeft := concat(t, s);
	end;

function combine (s1, s2: Str255; minLength: integer; var actualLength: integer): Str255;
	begin
		s1 := concat(s1, s2);
		s1 := addBlanksLeft(minLength - length(s1), s1);
		actualLength := length(s1);
		combine := s1;
	end;

procedure printUnchangedParametersHeader (var v: parameterDefinitionVector);
	var
		i: parameterIndex0;
	begin
		for i := 1 to v.length do
			if fixedParameter(v.parameterDefinitions[i]) then
				with v.parameterDefinitions[i], rangeListHead^ do
					writeRawStringCR(StringOf(addBlanksRight(parameterName, headerColumnFieldLength), ' = ', fromValue : minimumColumnFieldLength : outputDecimalDigits));
	end;

procedure printChangedParametersHeader (var v: parameterDefinitionVector);
	var
		i: parameterIndex0;
	begin
		for i := 1 to v.length do
			if not fixedParameter(v.parameterDefinitions[i]) then
				with v.parameterDefinitions[i] do
					writeRawString(StringOf(' ', combine('', parameterName, minimumColumnFieldLength, columnWidth)));
	end;

procedure printChangedParameters (var v: parameterDefinitionVector);
	var
		i: parameterIndex0;
	begin
		for i := 1 to v.length do
			if not fixedParameter(v.parameterDefinitions[i]) then
				with v.parameterDefinitions[i], currentRange^ do
					writeRawString(StringOf(' ', currentValue : columnWidth : outputDecimalDigits));
	end;

procedure printStatisticsHeader (var r: resultDefinitionVector; name: Str255);
	var
		i: parameterIndex0;
	begin
		writeRawString(StringOf(' ', combine(name, 'SuccessPercentage', minimumColumnFieldLength, r.successColumnWidth)));
		for i := 1 to r.length do
			with r.resultDefinitions[i] do begin
				writeRawString(StringOf(' ', combine('Mean', parameterName, minimumColumnFieldLength, meanColumnWidth)));
				writeRawString(StringOf(' ', combine(' StdDev', parameterName, minimumColumnFieldLength, stdDevColumnWidth)));
				writeRawString(StringOf(' ', combine('MeanNoExt', parameterName, minimumColumnFieldLength, meanNoExtColumnWidth)));
				writeRawString(StringOf(' ', combine('StdDevNoExt', parameterName, minimumColumnFieldLength, stdDevNoExtColumnWidth)));
				writeRawString(StringOf(' ', combine(' Min', parameterName, minimumColumnFieldLength, minimumColumnWidth)));
				writeRawString(StringOf(' ', combine(' Max', parameterName, minimumColumnFieldLength, maximumColumnWidth)));
			end;
	end;

procedure printStatistics (var r: resultStatistics; var v: resultDefinitionVector; numberOfExperiments: extended);
	var
		i: parameterIndex0;
		variance, varianceNoExt, mean, meanNoExt, successes, successPercentage: extended;
	begin
		successes := r.numberOfSuccesses;
		if numberOfExperiments > 0 then
			successPercentage := 100.0 * successes / numberOfExperiments
		else
			successPercentage := 0.0;
		writeRawString(StringOf(' ', successPercentage : v.successColumnWidth : outputDecimalDigits));
		for i := 1 to r.length do
			with r.statistics[i], v.resultDefinitions[i] do begin
				if successes > 0 then
					mean := sum / successes
				else
					mean := 0;
				if successes > 2 then
					meanNoExt := (sum - maximum - minimum) / (successes - 2)
				else
					meanNoExt := 0;
				if successes > 1 then begin
					variance := (sumSquare - successes * sqr(mean)) / (successes - 1);
					if variance < 0 then
						variance := 0;
				end
				else
					variance := 0;
				if successes > 3 then begin
					varianceNoExt := (sumSquare - sqr(maximum) - sqr(minimum) - (successes - 2) * sqr(meanNoExt)) / (successes - 3);
					if varianceNoExt < 0 then
						varianceNoExt := 0;
				end
				else
					varianceNoExt := 0;
				writeRawString(StringOf(' ', mean : meanColumnWidth : outputDecimalDigits));
				writeRawString(StringOf(' ', sqrt(variance) : stdDevColumnWidth : outputDecimalDigits));
				writeRawString(StringOf(' ', meanNoExt : meanNoExtColumnWidth : outputDecimalDigits));
				writeRawString(StringOf(' ', sqrt(varianceNoExt) : stdDevNoExtColumnWidth : outputDecimalDigits));
				writeRawString(StringOf(' ', minimum : minimumColumnWidth : outputDecimalDigits));
				writeRawString(StringOf(' ', maximum : maximumColumnWidth : outputDecimalDigits));
			end;
	end;

procedure initToCount (var v: parameterDefinitionVector);
	const
		epsilon = 1e-3;
	var
		i: parameterIndex0;
		t: rangePtr;
	begin
		for i := 1 to v.length do begin
			t := v.parameterDefinitions[i].rangeListHead;
			while t <> nil do
				with t^ do begin
					if (toValue = fromValue) | (byValue = 0) then
						toCount := 1
					else
{$IFC THINK_PASCAL }
						toCount := trunc(abs((toValue - fromValue) / byValue) + epsilon) + 1;
{$ELSEC }
						toCount := truncLong(abs((toValue - fromValue) / byValue) + epsilon) + 1;
{$ENDC }
					t := next;
				end;
		end;
	end;

function sizeOfStatistics (var v: parameterDefinitionVector): longInt;
	var
		result, sum: longInt;
		i: parameterIndex0;
		t: rangePtr;
	begin
		result := 1;
		for i := 1 to v.length do begin
			t := v.parameterDefinitions[i].rangeListHead;
			sum := 0;
			while t <> nil do
				with t^ do begin
					sum := sum + toCount;
					t := next;
				end;
			result := result * sum;
		end;
		sizeOfStatistics := result;
	end;

procedure initStatistics (var r: resultStatistics; length: parameterIndex0);
	var
		i: parameterIndex0;
	begin
		r.length := length;
		r.numberOfSuccesses := 0;
		for i := 1 to length do
			with r.statistics[i] do begin
				sum := 0;
				sumSquare := 0;
				minimum := 0;
				maximum := 0;
				used := false;
			end;
	end;

procedure initResultStatisticsArray (var v: resultStatisticsArray; size: longInt; length: parameterIndex0);
	var
		i: longInt;
	begin
		for i := 1 to size do
			initStatistics(v[i], length);
	end;

procedure allocateStatistics (var r: resultStatistics; numberOfExperiments: longInt; var rd: resultDefinitionVector);
	var
		i: parameterIndex0;
	begin
		for i := 1 to rd.length do
			with r.statistics[i] do
				if rd.resultDefinitions[i].createHistogram then
					values := realArrayH(dynamicArray(numberOfExperiments, sizeOf(real)))
				else
					values := nil;
	end;

procedure allocateStatisticsArray (var v: resultStatisticsArray; size, numberOfExperiments: longInt; var rd: resultDefinitionVector);
	var
		i: longInt;
	begin
		for i := 1 to size do
			allocateStatistics(v[i], numberOfExperiments, rd);
	end;

procedure deallocateStatistics (var r: resultStatistics);
	var
		i: parameterIndex0;
	begin
		for i := 1 to r.length do
			with r.statistics[i] do
				if values <> nil then
{$IFC THINK_PASCAL }
					disposHandle(Handle(values));
{$ELSEC }
					DisposeHandle(Handle(values));
{$ENDC }
	end;

procedure deallocateStatisticsArray (var v: resultStatisticsArray; size: longInt);
	var
		i: longInt;
	begin
		for i := 1 to size do
			deallocateStatistics(v[i]);
	end;

procedure updateStatistics (var r: resultStatistics; var v: parameterVector);
	var
		i: parameterIndex0;
		t: extended;
	begin
		r.numberOfSuccesses := r.numberOfSuccesses + 1;
		for i := 1 to r.length do
			with r.statistics[i] do begin
				t := v.parameters[i];
				if values <> nil then
					values^^[r.numberOfSuccesses] := t;
				sum := sum + t;
				sumSquare := sumSquare + sqr(t);
				if used then begin
					if t < minimum then
						minimum := t;
					if t > maximum then
						maximum := t;
				end
				else begin
					minimum := t;
					maximum := t;
					used := true;
				end;
			end;
	end;

procedure writeHeaderInformation (var e: experiment; experimentFileName, experimentTypeName: Str255);
	begin
		writeRawStringCR(StringOf(addBlanksRight('Start Date And Time', headerColumnFieldLength), ' = ', getDateAndTimeString));
		writeRawStringCR(StringOf(addBlanksRight('Experiment File Name', headerColumnFieldLength), ' = ', experimentFileName));
		writeRawStringCR(StringOf(addBlanksRight('Result File Name', headerColumnFieldLength), ' = ', e.resultFileName));
		writeRawStringCR(StringOf(addBlanksRight('Experiment Name', headerColumnFieldLength), ' = ', e.experimentName));
		writeRawStringCR(StringOf(addBlanksRight('Experiment Type', headerColumnFieldLength), ' = ', experimentTypeName));
		writeRawStringCR(StringOf(addBlanksRight('Number Of Experiments', headerColumnFieldLength), ' = ', trunc(e.numberOfExperiments) : 1));
		writeRawStringCR(StringOf(addBlanksRight('Generation Random Generator Seed', headerColumnFieldLength), ' = ', trunc(e.generationRandomSeed) : 1));
		writeRawStringCR(StringOf(addBlanksRight('Computation Random Generator Seed', headerColumnFieldLength), ' = ', trunc(e.computationRandomSeed) : 1));
		writeRawStringCR(StringOf(addBlanksRight('Background Priority', headerColumnFieldLength), ' = ', trunc(e.backgroundPriority) : 1));
		if e.randomConstraint <> nil then begin
			writeRawString(StringOf(addBlanksRight('Random Constraint', headerColumnFieldLength), ' = '));
			writeTextObjNoCR(e.randomConstraint);
			writeIndentedCR;
		end;
		writeRawStringCR(StringOf(addBlanksRight('Current Window', headerColumnFieldLength), ' = ', WindowPeek(currentGeoBenchWindow)^.titleHandle^^));
		if (currentArguments.length > 0) & e.showArguments then begin
			writeRawString(StringOf(addBlanksRight('Current Arguments', headerColumnFieldLength), ' = '));
			writeTextObjNoCR(currentArguments);
			writeIndentedCR;
		end;
		writeIndentedCR;
		printUnchangedParametersHeader(e.generation);
		writeIndentedCR;
		printUnchangedParametersHeader(e.computation);
		writeIndentedCR;
		writeIndentedCR;
		printChangedParametersHeader(e.generation);
		printStatisticsHeader(e.generationResult, 'Generation');
		printChangedParametersHeader(e.computation);
		printStatisticsHeader(e.computationResult, 'Computation');
		writeIndentedCR;
		flushStream;
	end;

function makeHistogram (var s: statisticsDefinitionRecord; numberOfValues: longInt; width, height: extended; numberOfBins: longInt): obj;
	var
		result: histogram;
		i: longInt;

	function generateValue: real;
		begin
			i := i + 1;
			generateValue := s.values^^[i];
		end;

	procedure resetGenerator;
		begin
			i := 0;
		end;

	begin (* makeHistogram *)
		new(result);
		failNil(Handle(result));
		result.initFromGenerator(numberOfValues, numberOfBins, width, height, generateValue, resetGenerator);
		makeHistogram := result;
	end;

procedure appendHistograms (var r: resultStatistics; var rd: resultDefinitionVector; s: Str255);
	var
		result: vector;
		mp: markedRealPoint;
		i: parameterIndex0;
	begin
		for i := 1 to rd.length do
			with rd.resultDefinitions[i] do
				if createHistogram then begin
					result := newVector(2);
					result.append(makeHistogram(r.statistics[i], r.numberOfSuccesses, 100, 100, histogramNumberOfBins));
					new(mp);
					failNil(Handle(mp));
					mp.init;
					mp.initLocationMark(result.ith(1), safeConcat(StringOf(truncate(parameterName, 10), ' '), s), true);
					result.append(mp);
					currentResult.append(result);
				end;
	end;

procedure printAllStatistics (var e: experiment; var generationResultStatistics: resultStatistics; var computationResultStatistics: resultStatisticsArray);
	var
		computationResultStatisticsIndex: longInt;
	begin
		computationResultStatisticsIndex := 0;
		initCurrentValue(e.computation);
		appendHistograms(generationResultStatistics, e.generationResult, showParameter(e.generation));
		repeat
			printChangedParameters(e.generation);
			printStatistics(generationResultStatistics, e.generationResult, e.numberOfExperiments);
			printChangedParameters(e.computation);
			computationResultStatisticsIndex := computationResultStatisticsIndex + 1;
			printStatistics(computationResultStatistics[computationResultStatisticsIndex], e.computationResult, e.numberOfExperiments);
			writeIndentedCR;
			appendHistograms(computationResultStatistics[computationResultStatisticsIndex], e.computationResult, safeConcat(showParameter(e.generation), showParameter(e.computation)));
		until nextPoint(e.computation);
		flushStream;
	end;

procedure writeSimpleText (var f: simpleText; s: Str255);
	var
		i: integer;
	begin
		for i := 1 to length(s) do
			write(f, Byte(ord(s[i])));
	end;

procedure dumpParameters (var f: simpleText; typeInfo: Str255; seed: extended; var v: parameterVector);
	var
		i: parameterIndex0;
	begin
		writeSimpleText(f, StringOf(typeInfo, ' ', trunc(seed) : minimumColumnFieldLength));
		for i := 1 to v.length do
			writeSimpleText(f, StringOf(' ', v.parameters[i] : minimumColumnFieldLength : outputDecimalDigits));
		write(f, Byte(ord(CR)));
	end;

procedure doGeneralExperiment (experimentFileName: Str255; experimentFileVolume: integer; experimentTypeName: Str255; function generate (p: parameterVector): obj; function test (o: obj): parameterVector; function compute (o: obj; q: parameterVector; var success: boolean): parameterVector);
	var
		allExperiments: experimentInformation;
		currentExperiment: experimentPtr;
		debugLogFile: simpleText;

	procedure exitDoGeneralExperiment;
		begin
			if allExperiments.debugFileName <> '' then begin
				close(debugLogFile);
				setFileInfo(allExperiments.debugFileName, experimentFileVolume, 'TEXT', possibleCreatorTypes[textCreatorType]);
			end;
			setRandomSeed(allExperiments.oldRandomSeed);
			setAbortTime(allExperiments.oldBackgroundPriority);
			disposeExperimentInformation(allExperiments);
			if userAbort then
				abortFlag := false; (* result should be preserved! *)
			exit(doGeneralExperiment);
		end;

	procedure performSingleExperiment (var e: experiment; experimentTypeName: Str255);
		var
			generationParameters, computationParameters, result: parameterVector;
			i, startTime, globalStartTime, computationResultStatisticsIndex, computationResultStatisticsSize: longInt;
			generatedObject: obj;
			usedTime: extended;
			generationResultStatistics: resultStatistics;
			computationResultStatistics: resultStatisticsArrayH;
			success: boolean;

		procedure cleanUpPerformSingleExperiment;
			begin
				writeIndentedCR;
				writeRawStringCR(StringOf(addBlanksRight('Finish Date And Time', headerColumnFieldLength), ' = ', getDateAndTimeString));
				writeRawStringCR(StringOf(addBlanksRight('TotalUsedTime', headerColumnFieldLength), ' = ', (TickCount - globalStartTime) / 60.0 : 1 : outputDecimalDigits));
				writeIndentedCR;
				writeIndentedCR;
				closeWriteStream;
				if generatedObject <> nil then
					generatedObject.freeAll;
				deallocateStatistics(generationResultStatistics);
				deallocateStatisticsArray(computationResultStatistics^^, computationResultStatisticsSize);
{$IFC THINK_PASCAL }
				disposHandle(Handle(computationResultStatistics));
{$ELSEC }
				DisposeHandle(Handle(computationResultStatistics));
{$ENDC }
			end;

		begin (* performSingleExperiment *)
			globalStartTime := TickCount;
			initToCount(e.generation);
			initToCount(e.computation);
			setAbortTime(e.backgroundPriority);
			generatedObject := nil;
			openWriteStreamTextAppend(e.resultFileName, experimentFileVolume, e.appendResult);
			writeHeaderInformation(e, experimentFileName, experimentTypeName);
			computationResultStatisticsSize := sizeOfStatistics(e.computation);
			computationResultStatistics := resultStatisticsArrayH(dynamicArray(computationResultStatisticsSize, sizeOf(resultStatistics)));
			allocateStatisticsArray(computationResultStatistics^^, computationResultStatisticsSize, e.numberOfExperiments, e.computationResult);
			allocateStatistics(generationResultStatistics, e.numberOfExperiments, e.generationResult);
			initCurrentValue(e.generation);
			repeat
				initStatistics(generationResultStatistics, e.generationResult.length);
				initResultStatisticsArray(computationResultStatistics^^, computationResultStatisticsSize, e.computationResult.length);
				for i := 1 to e.numberOfExperiments do begin
					if generatedObject <> nil then
						generatedObject.freeAll;
					currentRandomConstraint := e.randomConstraint;
					setRandomSeed(e.generationRandomSeed);
					generationParameters := extractParameters(e.generation);
					if allExperiments.debugFileName <> '' then
						dumpParameters(debugLogFile, 'G', e.generationRandomSeed, generationParameters);
					startTime := TickCount;
					generatedObject := generate(generationParameters);
					usedTime := (TickCount - startTime) / 60.0;
					e.generationRandomSeed := getRandomSeed;
					if userAbort then begin
						cleanUpPerformSingleExperiment;
						exitDoGeneralExperiment;
					end;
					if generatedObject <> nil then begin
						result := test(generatedObject);
						if result.length <> e.generationResult.length - 1 then begin
							nonFatalError(StringOf('Error: result.length  e.generationResult.length - 1 (', result.length : 1, '  ', e.generationResult.length - 1 : 1, ' ) '));
							cleanUpPerformSingleExperiment;
							exitDoGeneralExperiment;
						end;
						result.length := result.length + 1;
						result.parameters[result.length] := usedTime;
						updateStatistics(generationResultStatistics, result);
						computationResultStatisticsIndex := 0;
						initCurrentValue(e.computation);
						repeat
							if e.verbose then
								displayInfoString(safeConcat(StringOf(i : 1, ':'), safeConcat(showParameter(e.computation), safeConcat('; ', showParameter(e.generation)))));
							currentRandomConstraint := e.randomConstraint;
							setRandomSeed(e.computationRandomSeed);
							computationParameters := extractParameters(e.computation);
							if allExperiments.debugFileName <> '' then
								dumpParameters(debugLogFile, 'C', e.computationRandomSeed, computationParameters);
							startTime := TickCount;
							result := compute(generatedObject, computationParameters, success);
							usedTime := (TickCount - startTime) / 60.0;
							e.computationRandomSeed := getRandomSeed;
							if userAbort then begin
								cleanUpPerformSingleExperiment;
								exitDoGeneralExperiment;
							end;
							computationResultStatisticsIndex := computationResultStatisticsIndex + 1;
							if success then begin
								if result.length <> e.computationResult.length - 1 then begin
									nonFatalError(StringOf('result.length  e.computationResult.length - 1 (', result.length : 1, '  ', e.computationResult.length - 1 : 1, ' ) '));
									cleanUpPerformSingleExperiment;
									exitDoGeneralExperiment;
								end;
								result.length := result.length + 1;
								result.parameters[result.length] := usedTime;
								updateStatistics(computationResultStatistics^^[computationResultStatisticsIndex], result);
							end;
						until nextPoint(e.computation);
					end;
				end;
				printAllStatistics(e, generationResultStatistics, computationResultStatistics^^);
			until nextPoint(e.generation);
			cleanUpPerformSingleExperiment;
		end;

	begin (* doGeneralExperiment *)
		if readExperimentFile(experimentFileName, experimentFileVolume, experimentTypeName, allExperiments) then begin
			allExperiments.oldRandomSeed := getRandomSeed;
			allExperiments.oldBackgroundPriority := getAbortTime;
			currentExperiment := allExperiments.experimentList;
			if allExperiments.debugFileName <> '' then
				rewrite(debugLogFile, allExperiments.debugFileName);
			while currentExperiment <> nil do begin
				performSingleExperiment(currentExperiment^, allExperiments.experimentTypeName);
				currentExperiment := currentExperiment^.next;
			end;
			exitDoGeneralExperiment;
		end;
	end;

procedure initGeneralExperiment;
	begin
		stopCharacters := [' ', TAB, CR, ')', '{', EOFChar, '%', '#', '=', ','];
{$IFC ChecksEnabled AND CheckAllocation }
		alloc := 0;
{$ENDC }
	end;

end.
