(*	XYZ GeoBench (eXperimental geometrY Zurich):		*)
(*	An object oriented, interactive workbench for		*)
(*	experimental geometric computation.					*)
(*														*)
(*	 1999 by Peter Schorn, peter.schorn@acm.org		*)
(*	Culmannstrasse 77, CH-8006 Zrich, Switzerland		*)
(*	This software may be copied freely for educational	*)
(*	and research purposes only. For commercial use,		*)
(*	please contact the author.							*)
(*														*)
(*	Version 5.0.5: 30-Sep-1999							*)

unit graphics;

interface

uses
	GeoBenchUtility, streamIO, threeDVision;

const
	transX = 200;	(* translation for 3d-Graphics *)
	transY = 200;

type
	projDescPtr = ^projDescT;
	projDescT = record
			visibility: boolean;				(* visibility (hidden lines...) enabled?				*)
			userTrafo: trafoMatrT;		(* user defined rotations and translations			*)
			userTScal: point3dT;			(* user defined scaling									*)
			projTrafo: trafoMatrT;		(* projection											*)
			viewTrafo: trafoMatrT;		(* viewing transformation							*)
			eyePoint: point3dT;			(* eyepoint in world coordinates						*)
			proPoint: point3dT;			(* where is the projection plane?					*)
			lightPoint: point3dT;			(* where is light source 								*)
			projZ: boolean;					(* central projection enabled?						*)
			distP: extended;				(* between eyept and proj. plane						*)
			eyeProHasChanged: boolean;	(* signs whether eye or projectionpoint changed	*)
			lightHasChanged: boolean;		(* signs whether light changed						*)
			visHasChanged: boolean;		(* signs whether visibility changed					*)
			proHasChanged: boolean;		(* signs whether projection type has changed		*)
		end;
	wFrame = record
			proj3d: boolean;				(* ONLY true during layer.interactiveOutput !		*)
			proD: ^projDescT;				(* projection descriptor								*)
			dx, dy, mx, my: extended;	(* world to window transformation					*)
			z: extended;						(* z-Value (set by layer)								*)
		end;
	clipper = record
			left, right, bottom, top: integer;
		end;

function worldToWindow (x, y: extended; var ix, iy: integer): boolean;
		(* transforms word coordinates (x, y) to window coordinates (ix, iy) using 'currentFrame' 		*)
		(* true <==> inside clipping region 																			*)

function worldToWindow3d (x, y, z: extended; var ix, iy: integer): boolean;
		(* transforms world coords (x,y,z) to window coords (ix,iy) using 'currentFrame' 				*)
		(* true <==> inside clipping region 																			*)

procedure windToWorld (ix, iy: integer; var x, y: extended);
		(* transforms window coordinates (ix, iy) to world coordinates (x, y) using 'currentFrame' 		*)

procedure createFrame (oldFrame: wFrame; lx, ly, x1, y1, x2, y2: extended; var newFrame: wFrame);
		(* creates newFrame, allocates memory for it															*)
 		(* oldFrame = transformation valid in this window 														*)
		(* lx = length in x-direction of the window 																*)
		(* ly = length in y-direction of the window 																*)
		(* x1, y1, x2, y2 = coordinates of the rectangle which should be mapped on the given one 		*)
		(* newFrame = new transformation to obtain the necessary change in size 							*)

procedure copyFrame (oldFrame: wFrame; var newFrame: wFrame);
		(* copies oldFrame and allocates memory for newFrame													*)

procedure deleteframe (var f: wFrame);
		(* deallocates memory	of f																					*)

procedure writeframe (var v: wFrame);

procedure readFrame (var v: wFrame);

procedure oldReadFrame (var v: wFrame);

procedure drawOval (x, y, pointRadius: integer);

procedure drawPoint (x, y: extended; pointRadius: integer);

procedure drawPoint3d (x, y, z: extended; pointRadius: integer);

procedure drawLineSegment (x1, y1, x2, y2: extended);

procedure drawLineSegment3d (x1, y1, z1, x2, y2, z2: extended);

procedure drawThickLineSegment (x1, y1, x2, y2: extended);

procedure drawThickLineSegment3d (x1, y1, z1, x2, y2, z2: extended);

procedure drawVertical (x: extended);

procedure drawVertical3d (x, z: extended);

procedure drawThickVertical (x: extended);

procedure drawThickVertical3d (x, z: extended);

procedure drawCircle (x, y, r: extended; startAngle, arcAngle: integer);

procedure drawCircle3d (x, y, z, r: extended; startAngle, arcAngle: integer);

procedure drawRectangle2d (left, bottom, right, top: extended);

procedure drawCoordSys;

procedure rotatePoint (xOrigin, yOrigin, alpha: extended; var x, y: extended);
		(* rotate the point ('x', 'y') by 'alpha' around ('xOrigin', 'yOrigin') 									*)

procedure reflectPoint (a, b, c: extended; var x, y: extended);
		(* reflect the point ('x', 'y') at the line a*x + b*y = c, a*a+b*b<>0									*)

procedure moveSweepX (xOld, xNew: extended);

procedure moveSweepY (yOld, yNew: extended);

procedure ExtendDrawingSheet (x, y: integer);
(* called internally by worldToWindow and the drawing procedures of this unit. *)
(* Used only in special cases where the ToolBox Routines are called directly *)

procedure initGraphics;

const
oneDegree = 0.017453292; (*  / 180 *)

var
standardFrame, currentFrame: wFrame;			(* the current world coordinate system *)
clippingRectangle: clipper;
colorAvailable, coordSysIsDrawn: boolean;
{$IFC THINK_PASCAL }
sheetExtendProc: ProcPtr;
{$ELSEC }
type	sheetExtendProcT = procedure  (x, y: integer);
var		sheetExtendProc: sheetExtendProcT;
{$ENDC }

implementation

const
	clipLeft = -15000; 		(* coordinates of standard clipping rectangle *)
	clipBottom = -15000;
	clipRight = 15000;
	clipTop = 15000;

{$IFC THINK_PASCAL }
procedure execute (x, y: integer; p: ProcPtr);
inline
	$205f, $4e90; (* cf. transSkel *)

procedure ExtendDrawingSheet (x, y: integer);
	begin
		execute(x, y, sheetExtendProc);
	end;

{$ELSEC }
procedure ExtendDrawingSheet (x, y: integer);
	begin
		sheetExtendProc(x, y);
	end;
{$ENDC }

function worldToWindow (x, y: extended; var ix, iy: integer): boolean;
	var
		tx, ty: extended;
	begin
		with currentFrame do begin
			tx := dx + x * mx;
			ty := dy + y * my;
		end;
		if (clippingRectangle.left <= tx) & (tx <= clippingRectangle.right) & (clippingRectangle.bottom <= ty) & (ty <= clippingRectangle.top) then begin
			worldToWindow := true;
{$IFC THINK_PASCAL }
			ix := round(tx);
			iy := round(ty);
			execute(ix, iy, sheetExtendProc);
{$ELSEC }
			ix := roundLong(tx);
			iy := roundLong(ty);
			sheetExtendProc(ix, iy);
{$ENDC }
		end
		else
			worldToWindow := false;
	end;

function worldToWindow3d (x, y, z: extended; var ix, iy: integer): boolean;
	var
		p, pdot: point3dT;
		dist: extended;
	begin
		p[3] := z;
		with currentFrame do begin
			p[1] := x;
			p[2] := y;
			p[4] := 1.0;
			mvm(proD^.viewTrafo, p, pdot);
			p := pdot;
			if proD^.projZ then begin
				mvm(proD^.viewTrafo, proD^.eyePoint, pdot);
				dist := -pdot[3];
				p[1] := p[1] * dist / (p[3] + dist);
				p[2] := p[2] * dist / (p[3] + dist);
			end;
			worldToWindow3d := worldToWindow(p[1] + transX, p[2] + transY, ix, iy);
		end;
	end;

procedure windToWorld (ix, iy: integer; var x, y: extended);
	begin
		with currentFrame do begin
			x := (ix - dx) / mx;
			y := (iy - dy) / my;
		end;
	end;

procedure createFrame (oldFrame: wFrame; lx, ly, x1, y1, x2, y2: extended; var newFrame: wFrame);
	begin
		copyFrame(oldFrame, newFrame);
		with newFrame do begin
			mx := lx / (x2 - x1);
			dx := -x1 * mx + mx * oldFrame.dx;
			mx := mx * oldFrame.mx;
			my := ly / (y2 - y1);
			dy := -y1 * my + my * oldFrame.dy;
			my := my * oldFrame.my;
		end;
	end;

procedure copyFrame (oldFrame: wFrame; var newFrame: wFrame);
	begin
		newFrame := oldFrame;
		new(newFrame.proD);
		failNil(Handle(newFrame.proD));
		newFrame.proD^ := oldFrame.proD^;
(* inserted by Bert Leukens *)
		with newFrame.proD^ do begin
			eyeProHasChanged := true;
			lightHasChanged := true;
			visHasChanged := true;
			proHasChanged := true
		end;
	end;

procedure deleteframe (var f: wFrame);
	begin
		if f.proD <> nil then
			dispose(f.proD);
	end;

procedure writeframe (var v: wFrame);
	var
		i, j: integer;
	begin
		with v, proD^ do begin
			writeByte(0);
			writeByte(ord(visibility));
			for i := 1 to 4 do
				for j := 1 to 4 do
					writeExtended(userTrafo[i, j]);
			for i := 1 to 4 do
				writeExtended(userTScal[i]);
			for i := 1 to 4 do
				for j := 1 to 4 do
					writeExtended(projTrafo[i, j]);
			for i := 1 to 4 do
				for j := 1 to 4 do
					writeExtended(viewTrafo[i, j]);
			for i := 1 to 4 do
				writeExtended(eyePoint[i]);
			for i := 1 to 4 do
				writeExtended(proPoint[i]);
			for i := 1 to 4 do
				writeExtended(lightPoint[i]);
			writeByte(ord(projZ));
			writeExtended(distP);
			writeExtended(dx);
			writeExtended(dy);
			writeExtended(mx);
			writeExtended(my);
			writeExtended(z);
		end;
	end;

procedure readFrame (var v: wFrame);
	var
		ch: Byte;
		i, j: integer;
	begin
		with v do begin
			readByte(ch); (* proj3d flag, but proj3d always false *)
			proj3d := false;
			new(proD);
			failNil(Handle(proD));
			readByte(ch);
			with proD^ do begin
				visibility := ch = 1;
				for i := 1 to 4 do
					for j := 1 to 4 do
						readExtended(userTrafo[i, j]);
				for i := 1 to 4 do
					readExtended(userTScal[i]);
				for i := 1 to 4 do
					for j := 1 to 4 do
						readExtended(projTrafo[i, j]);
				for i := 1 to 4 do
					for j := 1 to 4 do
						readExtended(viewTrafo[i, j]);
				for i := 1 to 4 do
					readExtended(eyePoint[i]);
				for i := 1 to 4 do
					readExtended(proPoint[i]);
				if currentVersion >= version7 then
					for i := 1 to 4 do
						readExtended(lightPoint[i])
				else begin
 (* changed by Bert Leukens *)
					lightPoint[1] := 1000.0;
					lightPoint[2] := 3000.0;
					lightPoint[3] := 5000.0;
					lightPoint[4] := 1.0;
				end;
				readByte(ch);
				projZ := ch = 1;
				readExtended(distP);
				readExtended(dx);
				readExtended(dy);
				readExtended(mx);
				readExtended(my);
				readExtended(z);
(* changed by Bert Leukens *)
				eyeProHasChanged := true;
				lightHasChanged := true;
				visHasChanged := true;
				proHasChanged := true;
			end;
		end;
	end;

procedure oldReadFrame (var v: wFrame);
	begin
		copyFrame(standardFrame, v);
		with v do begin
			readExtended(dx);
			readExtended(dy);
			readExtended(mx);
			readExtended(my);
		end;
	end;

procedure drawOval (x, y, pointRadius: integer);
	begin
{$IFC THINK_PASCAL }
		PaintOval(y - pointRadius, x - pointRadius, y + pointRadius, x + pointRadius);
{$ELSEC }
		PaintOval2(y - pointRadius, x - pointRadius, y + pointRadius, x + pointRadius);
{$ENDC }
	end;

procedure drawPoint (x, y: extended; pointRadius: integer);
	var
		ix, iy: integer;
	begin
		if currentFrame.proj3d then
			drawPoint3d(x, y, currentFrame.z, pointRadius)
		else if worldToWindow(x, y, ix, iy) then
			drawOval(ix, iy, pointRadius);
	end;

procedure drawPoint3d (x, y, z: extended; pointRadius: integer);
	var
		ix, iy: integer;
	begin
		if worldToWindow3d(x, y, z, ix, iy) then
			drawOval(ix, iy, pointRadius);
	end;

procedure drawLineSegment (x1, y1, x2, y2: extended);
	type
		edgeCrossing = (left, right, bottom, top);
		outCode = set of edgeCrossing;
	var
		x, y, tx1, ty1, tx2, ty2: extended;
		c, c1, c2: outCode;
		penPos: Point;

	procedure code (x, y: extended; var c: outCode);
		begin
			c := [];
			if x < clippingRectangle.left then
				c := [left]
			else if x > clippingRectangle.right then
				c := [right];
			if y < clippingRectangle.bottom then
				c := c + [bottom]
			else if y > clippingRectangle.top then
				c := c + [top];
		end;

	begin
		if currentFrame.proj3d then
			drawLineSegment3d(x1, y1, currentFrame.z, x2, y2, currentFrame.z)
		else begin
			with currentFrame do begin
				tx1 := dx + x1 * mx;
				ty1 := dy + y1 * my;
				tx2 := dx + x2 * mx;
				ty2 := dy + y2 * my;
			end;
			code(tx1, ty1, c1);
			code(tx2, ty2, c2);
			while (c1 <> []) | (c2 <> []) do begin
				if c1 * c2 <> [] then
					exit(drawLineSegment);
				c := c1;
				if c = [] then
					c := c2;
				if left in c then begin
					y := ty1 + (ty2 - ty1) * (clippingRectangle.left - tx1) / (tx2 - tx1);
					x := clippingRectangle.left
				end
				else if right in c then begin
					y := ty1 + (ty2 - ty1) * (clippingRectangle.right - tx1) / (tx2 - tx1);
					x := clippingRectangle.right;
				end
				else if bottom in c then begin
					x := tx1 + (tx2 - tx1) * (clippingRectangle.bottom - ty1) / (ty2 - ty1);
					y := clippingRectangle.bottom;
				end
				else if top in c then begin
					x := tx1 + (tx2 - tx1) * (clippingRectangle.top - ty1) / (ty2 - ty1);
					y := clippingRectangle.top;
				end;
				if c = c1 then begin
					tx1 := x;
					ty1 := y;
					code(x, y, c1);
				end
				else begin
					tx2 := x;
					ty2 := y;
					code(x, y, c2);
				end;
			end;
{$IFC THINK_PASCAL }
			execute(round(tx1), round(ty1), sheetExtendProc);
			execute(round(tx2), round(ty2), sheetExtendProc);
{$ELSEC }
			sheetExtendProc(roundLong(tx1), roundLong(ty1));
			sheetExtendProc(roundLong(tx2), roundLong(ty2));
{$ENDC }
			GetPen(penPos);
			if (penPos.h <> round(tx1)) | (penPos.v <> round(ty1)) then
{$IFC THINK_PASCAL }
				MoveTo(round(tx1), round(ty1));
			LineTo(round(tx2), round(ty2));
{$ELSEC }
				MoveTo(roundLong(tx1), roundLong(ty1));
			LineTo(roundLong(tx2), roundLong(ty2));
{$ENDC }
		end;
	end;

procedure drawLineSegment3d (x1, y1, z1, x2, y2, z2: extended);

	type
		edgeCrossing = (top, right, bottom, left);
		outCode = set of edgeCrossing;

	const
		eps = 10E-8;

	var
		p, q, s, clip, pdot, qdot: point3dT;
		ix, iy: integer;
		dist: extended;
		c, c1, c2: outCode;
		t: extended;
		dummy: boolean;

	procedure code (var p: point3dT; var c: outCode);
		var
			k: extended;
		begin
			k := 1.0 + p[3] / dist;
			c := [];
			if p[2] < k * clip[1] then
				c := [top]
			else if p[2] > k * clip[3] then
				c := [bottom];
			if p[1] > k * clip[2] then
				c := c + [right]
			else if p[1] < k * clip[4] then
				c := c + [left];
		end;

	begin
		with currentFrame do begin
			p[1] := x1;
			p[2] := y1;
			p[3] := z1;
			p[4] := 1.0;
			mvm(proD^.viewTrafo, p, pdot);
			q[1] := x2;
			q[2] := y2;
			q[3] := z2;
			q[4] := 1.0;
			mvm(proD^.viewTrafo, q, qdot);
			if not proD^.projZ then begin
				x1 := pdot[1];
				y1 := pdot[2];
				x2 := qdot[1];
				y2 := qdot[2];
				dummy := proj3d;
				proj3d := false;
				drawLineSegment(x1 + transX, y1 + transY, x2 + transX, y2 + transY);
				proj3d := dummy;
			end
			else begin
				mvm(proD^.viewTrafo, proD^.eyePoint, p);
				dist := -p[3] - eps;
				with clippingRectangle do begin
					windToWorld(left, bottom, clip[4], clip[1]);
					clip[4] := clip[4] - transX;
					clip[1] := clip[1] - transY;
					windToWorld(right, top, clip[2], clip[3]);
					clip[2] := clip[2] - transX;
					clip[3] := clip[3] - transY;
				end;
				p := pdot;
				q := qdot;
				code(p, c1);
				code(q, c2);
				while (c1 <> []) | (c2 <> []) do begin
					if (c1 * c2) <> [] then
						exit(drawLineSegment3d);
					c := c1;
					if c = [] then
						c := c2;
					if top in c then begin
						t := (p[2] * dist - clip[1] * (dist + p[3])) / (dist * (p[2] - q[2]) + clip[1] * (q[3] - p[3]));
						s[1] := p[1] + t * (q[1] - p[1]);
						s[3] := p[3] + t * (q[3] - p[3]);
						s[2] := (1.0 + s[3] / dist) * clip[1];
					end
					else if bottom in c then begin
						t := (p[2] * dist - clip[3] * (dist + p[3])) / (dist * (p[2] - q[2]) + clip[3] * (q[3] - p[3]));
						s[1] := p[1] + t * (q[1] - p[1]);
						s[3] := p[3] + t * (q[3] - p[3]);
						s[2] := (1.0 + s[3] / dist) * clip[3];
					end
					else if right in c then begin
						t := (p[1] * dist - clip[2] * (dist + p[3])) / (dist * (p[1] - q[1]) + clip[2] * (q[3] - p[3]));
						s[2] := p[2] + t * (q[2] - p[2]);
						s[3] := p[3] + t * (q[3] - p[3]);
						s[1] := (1.0 + s[3] / dist) * clip[2];
					end
					else if left in c then begin
						t := (p[1] * dist - clip[4] * (dist + p[3])) / (dist * (p[1] - q[1]) + clip[4] * (q[3] - p[3]));
						s[2] := p[2] + t * (q[2] - p[2]);
						s[3] := p[3] + t * (q[3] - p[3]);
						s[1] := (1.0 + s[3] / dist) * clip[4];
					end;
					if c = c1 then begin
						p := s;
						code(p, c1);
					end
					else begin
						q := s;
						code(q, c2);
					end;
				end;
				p[3] := (p[3] + dist);
				q[3] := (q[3] + dist);
				if (p[3] > eps) & (q[3] > eps) then begin
					p[3] := dist / p[3];
					q[3] := dist / q[3];
					p[1] := p[1] * p[3];
					p[2] := p[2] * p[3];
{$IFC THINK_PASCAL }
					ix := round(dx + (p[1] + transX) * mx);
					iy := round(dy + (p[2] + transY) * my);
					execute(ix, iy, sheetExtendProc);
					MoveTo(ix, iy);
					q[1] := q[1] * q[3];
					q[2] := q[2] * q[3];
					ix := round(dx + (q[1] + transX) * mx);
					iy := round(dy + (q[2] + transY) * my);
					execute(ix, iy, sheetExtendProc);
{$ELSEC }
					ix := roundLong(dx + (p[1] + transX) * mx);
					iy := roundLong(dy + (p[2] + transY) * my);
					sheetExtendProc(ix, iy);
					MoveTo(ix, iy);
					q[1] := q[1] * q[3];
					q[2] := q[2] * q[3];
					ix := roundLong(dx + (q[1] + transX) * mx);
					iy := roundLong(dy + (q[2] + transY) * my);
					sheetExtendProc(ix, iy);
{$ENDC }
					LineTo(ix, iy);
				end;
			end;
		end;
	end;

procedure drawThickLineSegment (x1, y1, x2, y2: extended);
	begin
		PenSize(3, 3);
		if currentFrame.proj3d then
			drawLineSegment3d(x1, y1, currentFrame.z, x2, y2, currentFrame.z)
		else
			drawLineSegment(x1, y1, x2, y2);
		PenSize(1, 1);
	end;

procedure drawThickLineSegment3d (x1, y1, z1, x2, y2, z2: extended);
	begin
		PenSize(3, 3);
		drawLineSegment3d(x1, y1, z1, x2, y2, z2);
		PenSize(1, 1);
	end;

procedure drawVertical (x: extended);
	begin
		if currentFrame.proj3d then
			drawVertical3d(x, currentFrame.z)
		else
			drawLineSegment(x, clippingRectangle.bottom, x, clippingRectangle.top);
	end;

procedure drawVertical3d (x, z: extended);
	begin
		drawLineSegment3d(x, clippingRectangle.bottom, z, x, clippingRectangle.top, z);
	end;

procedure drawThickVertical (x: extended);
	begin
		PenSize(3, 3);
		if currentFrame.proj3d then
			drawThickVertical3d(x, currentFrame.z)
		else
			drawVertical(x);
		PenSize(1, 1);
	end;

procedure drawThickVertical3d (x, z: extended);
	begin
		PenSize(3, 3);
		drawVertical3d(x, z);
		PenSize(1, 1);
	end;

procedure drawCoordSys;
	var
		oldPen, newPen: PenState;
	begin
		GetPenState(oldPen);										(* set paint mode 				*)
		newPen := oldPen;
		newPen.pnMode := patCopy;
		SetPenState(newPen);
		PenSize(1, 1);
		drawLineSegment3d(0, 0, 0, 150, 0, 0);				(* draw x-axis with arrow	*)
		drawLineSegment3d(150, 0, 0, 145, -3, 0);
		drawLineSegment3d(150, 0, 0, 145, 3, 0);
		drawLineSegment3d(160, -10, 0, 176, 10, 0);		(* label it						*)
		drawLineSegment3d(160, 10, 0, 176, -10, 0);
		drawLineSegment3d(0, 0, 0, 0, 150, 0);				(* draw y-axis with arrow	*)
		drawLineSegment3d(0, 150, 0, -3, 145, 0);
		drawLineSegment3d(0, 150, 0, 3, 145, 0);
		drawLineSegment3d(-10, 168, 0, 0, 168, 0);			(* label it						*)
		drawLineSegment3d(0, 168, 0, 10, 176, 0);
		drawLineSegment3d(0, 168, 0, 10, 160, 0);
		drawLineSegment3d(0, 0, 0, 0, 0, 150);				(* draw z-axis with arrow	*)
		drawLineSegment3d(0, 0, 150, -3, 0, 145);
		drawLineSegment3d(0, 0, 150, 3, 0, 145);
		drawLineSegment3d(-10, 0, 160, -10, 0, 176);		(* label it						*)
		drawLineSegment3d(-10, 0, 176, 10, 0, 160);
		drawLineSegment3d(10, 0, 160, 10, 0, 176);
		SetPenState(oldPen);										(* reset paint mode 			*)
	end;

procedure drawCircle (x, y, r: extended; startAngle, arcAngle: integer);
	var
		x1, y1, x2, y2: integer;
	begin
		if currentFrame.proj3d then
			drawCircle3d(x, y, currentFrame.z, r, startAngle, arcAngle)
		else begin
			if worldToWindow(x - r, y - r, x1, y1) & worldToWindow(x + r, y + r, x2, y2) then
{$IFC THINK_PASCAL }
				FrameArc(y1, x1, y2, x2, startAngle, arcAngle);
{$ELSEC }
				FrameArc2(y1, x1, y2, x2, startAngle, arcAngle);
{$ENDC }
		end;
	end;

procedure drawCircle3d (x, y, z, r: extended; startAngle, arcAngle: integer);
	(* not yet implemented *)
	var
		x1, y1, x2, y2: integer;
	begin
		if worldToWindow3d(x - r, y - r, z, x1, y1) & worldToWindow3d(x + r, y + r, z, x2, y2) then
{$IFC THINK_PASCAL }
			FrameArc(y1, x1, y2, x2, startAngle, arcAngle);
{$ELSEC }
			FrameArc2(y1, x1, y2, x2, startAngle, arcAngle);
{$ENDC }
	end;

procedure drawRectangle2d (left, bottom, right, top: extended);
	begin
		drawLineSegment(left, bottom, right, bottom);
		drawLineSegment(right, bottom, right, top);
		drawLineSegment(right, top, left, top);
		drawLineSegment(left, top, left, bottom);
	end;

procedure rotatePoint (xOrigin, yOrigin, alpha: extended; var x, y: extended);
	var
		cosAlpha, sinAlpha, xt, yt: extended;
	begin
		cosAlpha := cos(alpha);
		sinAlpha := sin(alpha);
		xt := x - xOrigin;
		yt := y - yOrigin;
		x := yt * sinAlpha + xt * cosAlpha + xOrigin;
		y := yt * cosAlpha - xt * sinAlpha + yOrigin;
	end;

procedure reflectPoint (a, b, c: extended; var x, y: extended);
	var
		xNew, d: extended;
	begin
		d := sqr(a) + sqr(b);
		xNew := 2 * (b * (b * x - a * y) + a * c) / d - x;
		y := 2 * (a * (a * y - b * x) + b * c) / d - y;
		x := xNew;
	end;

procedure moveSweep (startPix, endPix: integer; procedure drawIt (j: integer));
	var
		i: integer;
		now, drawnLines, linesPerSecond: longInt;
	begin
		PenMode(patOr);
		drawnLines := 0;
		now := TickCount;
{$IFC THINK_PASCAL }
		linesPerSecond := trunc(120 * 60 / movieWait);
{$ELSEC }
		linesPerSecond := truncLong(120 * 60 / movieWait);
{$ENDC }
		if linesPerSecond < 60 then
			linesPerSecond := 60;
		with currentFrame do
			for i := startPix to endPix do begin
				drawnLines := drawnLines + 60;
				drawIt(i);
				while linesPerSecond * (TickCount - now) < drawnLines do
					;
			end;
		PenMode(patXor);
{$IFC THINK_PASCAL }
		PenPat(black);
{$ELSEC }
		PenPat(qd.black);
{$ENDC }
	end;

procedure drawItX (i: integer);
	begin
{$IFC THINK_PASCAL }
		PaintRect(0, i, 2000, i + 1);
{$ELSEC }
		PaintRect2(0, i, 2000, i + 1);
{$ENDC }
	end;

procedure moveSweepX (xOld, xNew: extended);
	begin
{$IFC THINK_PASCAL }
		PenPat(ltGray);
		with currentFrame do
			moveSweep(round(mx * xOld + dx), round(mx * xNew + dx), drawItX);
{$ELSEC }
		PenPat(qd.ltGray);
		with currentFrame do
			moveSweep(roundLong(mx * xOld + dx), roundLong(mx * xNew + dx), drawItX);
{$ENDC }
	end;

procedure drawItY (i: integer);
	begin
{$IFC THINK_PASCAL }
		PaintRect(i, 0, i + 1, 2000);
{$ELSEC }
		PaintRect2(i, 0, i + 1, 2000);
{$ENDC }
	end;

procedure moveSweepY (yOld, yNew: extended);
	begin
{$IFC THINK_PASCAL }
		PenPat(gray);
		with currentFrame do
			moveSweep(round(my * yOld + dy), round(my * yNew + dy), drawItY);
{$ELSEC }
		PenPat(qd.gray);
		with currentFrame do
			moveSweep(roundLong(my * yOld + dy), roundLong(my * yNew + dy), drawItY);
{$ENDC }
	end;

{$IFC NOT THINK_PASCAL }
{$PUSH }
{$PRAGMAC warn_unusedarg off }
{$ENDC }
procedure doNothing (x, y: integer);
	begin
	end;
{$IFC NOT THINK_PASCAL }
{$POP }
{$ENDC }

procedure initGraphics;

	function ColorQDAvailable: boolean;
		const
			QD32Trap = $AB03;
			UnImplTrap = $A89F;
		var
			theWorld: SysEnvRec;
			error: integer;
		begin
			error := SysEnvirons(2, theWorld);
			ColorQDAvailable := theWorld.hasColorQD & (NGetTrapAddress(QD32Trap, ToolTrap) <> NGetTrapAddress(UnImplTrap, ToolTrap))
		end;

	begin
		with standardFrame do begin
			proj3d := false;
			new(proD);
			failNil(Handle(proD));
			with proD^ do begin
				initViewingTransformation(eyePoint, proPoint, lightPoint, userTrafo, userTScal, projTrafo, viewTrafo);
				projZ := false;
				distP := sqrt(sqr(eyePoint[1] - proPoint[1]) + sqr(eyePoint[2] - proPoint[2]) + sqr(eyePoint[3] - proPoint[3]));
(* changed by Bert Leukens *)
				eyeProHasChanged := true;
				lightHasChanged := true;
				visHasChanged := true;
				proHasChanged := true;
			end;
			dx := 0;
			dy := 0;
			mx := 1;
			my := 1;
			z := 0;
		end;
		currentFrame := standardFrame;
		clippingRectangle.left := clipLeft;				(* coordinates of standard clipping rectangle *)
		clippingRectangle.bottom := clipBottom;
		clippingRectangle.right := clipRight;
		clippingRectangle.top := clipTop;
		sheetExtendProc := @doNothing;
		colorAvailable := ColorQDAvailable;
	end;

end.
