(*	XYZ GeoBench (eXperimental geometrY Zurich):		*)
(*	An object oriented, interactive workbench for		*)
(*	experimental geometric computation.					*)
(*														*)
(*	 1999 by Peter Schorn, peter.schorn@acm.org		*)
(*	Culmannstrasse 77, CH-8006 Zrich, Switzerland		*)
(*	This software may be copied freely for educational	*)
(*	and research purposes only. For commercial use,		*)
(*	please contact the author.							*)
(*														*)
(*	Version 5.0.5: 30-Sep-1999							*)

(*	Written by Richard Dittrich, Aug-Sep 93				*)

unit gridPointEdge;

interface

uses
	triangulation;

const
	EXT_MAX = 1e300;

type
	gridPoint = object(idRealPoint)
			nX, nY: extended;      (* normalized grid coordinates *)
			used: boolean;        (* flag indicating wether a point is used by the triangulation process *)
			procedure init;
			override;
			procedure setNX (newNX: extended);    (* set normalized x - coordinate *)
			procedure setNY (newNY: extended);     (* set normalized y - coordinate *)
			function getNX: extended;   (*    return normalized x - coordinate *)
			function getNY: extended;   (*    return normalized y - coordinate *)
			procedure setId (newId: seqIndex);  (* set  number of gridpoint within grid *)
			function getRow: integer; (* the row of the grid the point is in *)
			function getCol: integer;  (* the column of the grid the point is in *)
			function isUsed: boolean;  (* get status of used - flag *)
			procedure setUsed;            (* set used - flag TRUE *)
			function distanceX (p: point2d): extended; (*  distance using normalized coordinates *)
			override;
			function rowDistance (m: integer): extended; (* perpendecular distance to  row m using normalized coordinates *)
			function colDistance (n: integer): extended;   (*  perpendecular distance to column n using normalized coordinates *)
(* comment on m,n : integer : negative rows and negative columns are allowed *)
			function equal (o: obj): boolean;  (* equal if same vertexNr *)
			override;
			function whichSideX (p, q: point2d): extended;
			override;
			procedure gridPoint.circleCenterX (p, q: point2d; var cx, cy: extended);
			override;
			function getInfo: obj;
			override;
			procedure setInfo (o: obj);
			override;
			function getTextType: Str255;
			override;
		end;

	gridEdge = object(graphEdgeTriangulation)
			p1, p2: gridPoint;   (* the corresponding pointers into the grid *)
			q, b, qinv: extended; (* line through vertices : y = q * x + b, qinv := 1/q *)
			octant: integer; (* octant the line starting from startVertex  to endVertex lies in *)
			procedure init;
			override;
			procedure initialize (newP1, newP2: gridPoint; startV, endV: seqIndex);
(* sets the values startVertex, endVertex, p1,p2,q,b,octant *)

			function rightOf (p: gridPoint): boolean;   (* determines whether p lies right of the edge p1,p2 *)
			function cosAngle (p: gridPoint): extended; (* computes cosinus of p1,p,p2 *)
			function isDual (gE: gridEdge): boolean;  (* dual if p1 == gE.p2 and p2 == gE.p1 *)
			procedure nextCell (var j, i: integer); (* given the gridcell j,i, computes the next cell j,i  *)
			procedure prevCell (var j, i: integer); (* given the gridcell j,i, computes the previous cell j,i  *)
			function getInfo: obj;
			override;
			function getTextType: Str255;
			override;
		end;

implementation

(* ---------------------------------- BEGIN TYPE GRIDPOINT ----------------------------------- *)

procedure gridPoint.init;
	begin
		used := false;
	end;

function gridPoint.getInfo: obj;
	var
		v: vector;
	begin
		v := vector(inherited getInfo);
		appendObjInfoExtended(v, 'Norm-X', nX, false);
		appendObjInfoExtended(v, 'Norm-Y', nY, false);
		appendObjInfoString(v, 'Used', StringOf(used), false);
		getInfo := v;
	end;

procedure gridPoint.setInfo (o: obj);
	var
		newUsed: Str255;
	begin
		inherited setInfo(o);
		with vector(o) do begin
			nX := objInfoExtended(ith(4)).e;
			nY := objInfoExtended(ith(5)).e;
			newUsed := capString(objInfoString(ith(6)).s);
		end;
		if (newUsed = 'F') | (newUsed = 'FALSE') then
			used := false
		else if (newUsed = 'T') | (newUsed = 'TRUE') then
			used := true;
	end;

function gridPoint.getTextType: Str255;
	begin
		getTextType := 'Grid Point';
	end;

procedure gridPoint.setNX (newNX: extended);
	begin
		nX := newNX;
	end;

procedure gridPoint.setNY (newNY: extended);
	begin
		nY := newNY;
	end;

function gridPoint.getNX: extended;
	begin
		getNX := nX;
	end;

function gridPoint.getNY: extended;
	begin
		getNY := nY;
	end;

procedure gridPoint.setId (newId: seqIndex);
	begin
		id := newId;
	end;

function gridPoint.getRow: integer;
	begin
{$IFC THINK_PASCAL }
		getRow := trunc(ny) + 1;
{$ELSEC }
		getRow := truncLong(nY) + 1;
{$ENDC }
	end;

function gridPoint.getCol: integer;
	begin
{$IFC THINK_PASCAL }
		getCol := trunc(nx) + 1;
{$ELSEC }
		getCol := truncLong(nX) + 1;
{$ENDC }
	end;

function gridPoint.isUsed: boolean;
	begin
		isUsed := used;
	end;

procedure gridPoint.setUsed;
	begin
		used := TRUE;
	end;

function gridPoint.distanceX (p: point2d): extended;
	begin
		distanceX := sqrt(sqr(gridPoint(p).nX - nX) + sqr(gridPoint(p).nY - nY));
	end;

function gridPoint.rowDistance (m: integer): extended;
	begin
		if m >= nY then    (* m is above *)
			rowDistance := m - nY - 1  (* subtract line below cell m *)
		else                      (* m is  below *)
			rowDistance := nY - m  (* subtract line above cell m *)
	end;

function gridPoint.colDistance (n: integer): extended;
	begin
		if n >= nX then    (* n is right *)
			colDistance := n - nX - 1  (* subtract line left of cell n *)
		else                      (* n is left *)
			colDistance := nX - n (* subtract line right of cell n *)
	end;

function gridPoint.equal (o: obj): boolean;
	begin
		equal := (o <> nil) & member(o, gridPoint) & (gridPoint(o).id = id);
	end;

function gridPoint.whichSideX (p, q: point2d): extended;
	var
		pg, qg: gridPoint;
	begin
		pg := gridPoint(p);
		qg := gridPoint(q);
		whichSideX := (pg.nX - nX) * (pg.nY - qg.nY) - (pg.nX - qg.nX) * (pg.nY - nY);
	end;

procedure gridPoint.circleCenterX (p, q: point2d; var cx, cy: extended);
	const
		epsilon = 1E-6;
	var
		d, s1, s2, s3: extended;
		pg, qg: gridPoint;
	begin
		pg := gridPoint(p);
		qg := gridPoint(q);
		d := 2 * whichSideX(pg, qg);
		if d = 0.0 then
			d := epsilon;
		s1 := sqr(pg.nX) + sqr(pg.nY);
		s2 := sqr(nX) + sqr(nY);
		s3 := sqr(qg.nX) + sqr(qg.nY);
		cx := (s1 * (nY - qg.nY) + s2 * (qg.nY - pg.nY) + s3 * (pg.nY - nY)) / d;
		cy := (s1 * (qg.nX - nX) + s2 * (pg.nX - qg.nX) + s3 * (nX - pg.nX)) / d;
	end;

(* ---------------------------------- END TYPE GRIDPOINT ----------------------------------- *)

(* ---------------------------------- BEGIN TYPE GRIDEDGE  ----------------------------------- *)
	const
		qeps = 1e-8;    (* if  x2 - x1 of gridedge is less than qeps then q := EXT_MAX, qinv := 0 *)

function gridEdge.getInfo: obj;
	var
		v: vector;
	begin
		v := vector(inherited getInfo);
		appendObjInfoExtended(v, 'Q', q, false);
		appendObjInfoExtended(v, 'Q-Inv', qinv, false);
		appendObjInfoExtended(v, 'B', b, false);
		appendObjInfoExtended(v, 'Octant', octant, false);
		getInfo := v;
	end;

function gridEdge.getTextType: Str255;
	begin
		getTextType := 'Grid Edge';
	end;

procedure gridEdge.init;
	begin
		p1 := nil;
		p2 := nil;
		inherited init;
	end;

procedure gridEdge.initialize (newP1, newP2: gridPoint; startV, endV: seqIndex);

	procedure getLineValues;
		var
			x1, y1, x2, y2, rx, ry: extended;
		begin
			x1 := p1.getNX;
			y1 := p1.getNY;
			x2 := p2.getNX;
			y2 := p2.getNY;
			assert((x1 <> x2) | (y1 <> y2), 'gridedge is no point');
			rx := x2 - x1;
			ry := y2 - y1;
(* ------------ first get octant -------- *)
			if (rx >= 0) & (ry >= 0) then
				if rx >= ry then
					octant := 1
				else
					octant := 2
(* ---------------------- *)
			else if (rx < 0) & (ry >= 0) then
				if -rx < ry then
					octant := 3
				else
					octant := 4
(* ----------------------- *)
			else if (rx < 0) & (ry < 0) then
				if rx <= ry then
					octant := 5
				else
					octant := 6
(* ----------------------- *)
			else if (rx >= 0) & (ry < 0) then
				if rx < -ry then
					octant := 7
				else
					octant := 8;
(* ----------  second get q,b, qinv ------------- *)
			if abs(rx) < qeps then begin
				q := EXT_MAX;
				qinv := 0;
				b := EXT_MAX;
			end
			else begin
				q := ry / rx;
				b := (x2 * y1 - x1 * y2) / rx;
				if (abs(q) < qeps) then
					qinv := EXT_MAX
				else
					qinv := 1.0 / q;
			end;
		end;

	begin
		p1 := newP1;
		p2 := newP2;
		startVertex := startV;
		endVertex := endV;
		getLineValues;
	end;


function gridEdge.rightOf (p: gridPoint): boolean;
	begin
		rightOf := (p1.nX - p.nX) * (p1.nY - p2.nY) - (p1.nX - p2.nX) * (p1.nY - p.nY) > 0;
	end;

function gridEdge.cosAngle (p: gridPoint): extended;
	var
		r1X, r1Y, r2X, r2Y: extended;
	begin
		r1X := p.nX - p1.nX;
		r1Y := p.nY - p1.nY;
		r2X := p.nX - p2.nX;
		r2Y := p.nY - p2.nY;
		cosAngle := (r1X * r2X + r1Y * r2Y) / sqrt((sqr(r1X) + sqr(r1Y)) * (sqr(r2X) + sqr(r2Y)));
 (* because of tolerance divisor is greater 0 *)
	end;

function gridEdge.isDual (gE: gridEdge): boolean;
	begin
		isDual := p1.equal(gE.p2) & p2.equal(gE.p1);
	end;

procedure gridEdge.nextCell (var j, i: integer);
	var
		jold, iold: integer;
	begin
		jold := j;
		iold := i;
		case octant of
			1, 8:  begin
{$IFC THINK_PASCAL }
				j := trunc(q * i + b) + 1;
{$ELSEC }
				j := truncLong(q * i + b) + 1;
{$ENDC }
				if j = jold then
					i := succ(i);
			end;
			4, 5:  begin
{$IFC THINK_PASCAL }
				j := trunc(q * (i - 1) + b) + 1;
{$ELSEC }
				j := truncLong(q * (i - 1) + b) + 1;
{$ENDC }
				if j = jold then
					i := pred(i);
			end;
			2, 3:  begin
				if q = EXT_MAX then
					j := succ(j)
				else begin
{$IFC THINK_PASCAL }
					i := trunc((j - b) * qinv) + 1;
{$ELSEC }
					i := truncLong((j - b) * qinv) + 1;
{$ENDC }
					if i = iold then
						j := succ(j);
				end;
			end;
			6, 7:  begin
				if q = EXT_MAX then
					j := pred(j)
				else begin
{$IFC THINK_PASCAL }
					i := trunc((j - 1 - b) * qinv) + 1;
{$ELSEC }
					i := truncLong((j - 1 - b) * qinv) + 1;
{$ENDC }
					if i = iold then
						j := pred(j);
				end;
			end;
		end; (* case *)
	end; (* nextCell *)

procedure gridEdge.prevCell (var j, i: integer);
	var
		s_octant: integer;
	begin
		s_octant := octant;
(* invert the direction *)
		if octant <= 4 then
			octant := octant + 4
		else
			octant := octant - 4;
		nextCell(j, i);
		octant := s_octant;
	end; (* prevCell *)

(* ---------------------------------- END TYPE GRIDEDGE  ----------------------------------- *)

end.
