(*	XYZ GeoBench (eXperimental geometrY Zurich):		*)
(*	An object oriented, interactive workbench for		*)
(*	experimental geometric computation.					*)
(*														*)
(*	 1999 by Peter Schorn, peter.schorn@acm.org		*)
(*	Culmannstrasse 77, CH-8006 Zrich, Switzerland		*)
(*	This software may be copied freely for educational	*)
(*	and research purposes only. For commercial use,		*)
(*	please contact the author.							*)
(*														*)
(*	Version 5.0.5: 30-Sep-1999							*)

(*	Written by Richard Dittrich, Aug-Sep 93				*)

unit gridTriangulation;

interface

uses
	gridPointEdge;

implementation

{$SETC AGridTriangulation = TRUE		}
{$SETC Debug = FALSE					}
	(*	If Debug is TRUE, each point forming a new triangle is checked for being the one		*)
	(* with the maximum angle. This is done with getMaxAnglePointNaive, see below. 		*)
{$SETC testInvariants=TRUE			}
	(*	If true, the triangles of the resulting triangulation are checked whether they			*)
	(*	consist of exactly three points.																*)

{$IFC Debug }
var
	debug: boolean;
{$ENDC }

(* ---------------------------------- BEGIN TYPE GRID ----------------------------------- *)

type
	grid = object(obj)
			rows, cols: seqIndex0;					(* number of rows and columns												*)
			xMin, yMin, xMax, yMax: extended;	(* left lower corner and right upper corner of grid						*)
			size: extended;								(* size of a gridcell																*)
			g: homogenVector;							(* m homogen vectors store n homogen vectors  which store a list	*)
													(* each list stores the points of the respective gridcell					*)
			nrPoints: seqIndex0;                            (* number of points, which are currently in grid							*)

			procedure initialize (pV: pointVector; tolerance, sizeFactor: extended);
(* pV            :   points to be stored in the grid for further processing. *)
(*tolerance  : if  the distance of two gridpoints within a gridcell lies below tolerance *)
(*                   one of the points is discarded. *)
(* sizeFactor : a factor to adjust the gridsize *)

			procedure interactiveOutput; (* showing the grid *)
			override;
			procedure free;
			override;
			procedure freeNotUsedPoints;       (* in case of user abort, free all points not belonging to current triangulation *)
			procedure interactiveOutputCell (m, n: seqIndex0);  (* show cell in row m, column n with points *)
			function getList (row, col: seqIndex0): list;  (* get the list of gridcell with row and col *)
			procedure remove (gP: gridPoint);    (* remove point from grid structure *)
			function getMiddlePoint: gridPoint;   (* gets a point somewhere in the middle of the grid *)
(* ---------------------------------------------------------------------------------- *)
			function getClosestPoint (p: gridPoint): gridPoint;  (* get's to a gridpoint p the closest point *)
			function getClosestPointNaive (p: gridPoint): gridPoint;  (* get's to a gridpoint p the closest point *)
			procedure checkClosestPoint (p, closestP: gridPoint); (* check's a closest point with naive method *)
(* --------------------------------------------------------------------------------- *)
			function getMaxAnglePoint (gE: gridEdge; var p: gridPoint): boolean;
(* get's the point p right of the gridedge with the maximum angle by searching right of the gridedge.  *)
(*    If not point was found false  is returned and p is assigned to nil *)
			function getMaxAnglePointNaive (gE: gridEdge; var p: gridPoint): boolean;
(* get's the point p right of the gridedge with the maximum angle by checking each point in the grid.  *)
(*    If not point was found false  is returned and p is assigned to nil *)
			procedure checkMaxAnglePoint (gE: gridEdge; p3: gridPoint; found: boolean);
(* Checks a point found or not found with getMaxAnglePoint with getMaxAnglePointNaive. *)
(* If a difference is encountered,  a text to standard output is written (writeln) and assert(false,txt) will be called. *)
(* ----------------------------------------------------------------------------------------- *)

			function triangulation (p1, p2: gridPoint): triangulation2d;
(* makes a triangulation using getMaxAnglePoint *)
(* p1,p2 is the startedge, where p1 should be some point in the middle, and p2 must be the closest point to p1 *)
(* (see  getMiddlePoint, getClosestPoint)        *)

		end;

procedure grid.free;
	var
		m, n: seqIndex;
		hV: homogenVector;

	begin
(* first free lists *)
		for m := 1 to rows do begin
			hV := homogenVector(g.elements^^[m]);
			for n := 1 to cols do begin
				hV.elements^^[n].free;
			end;
			hV.free;
		end;
		g.free;
		shallowFree;
	end;

procedure grid.freeNotUsedPoints;
	var
		m, n: seqIndex;

	procedure freeNotUsedPoint (x: obj);
		begin
			if not gridPoint(x).isUsed then
				x.free;
		end;

	begin
		for m := 1 to rows do
			for n := 1 to cols do
				list(homogenVector(g.elements^^[m]).elements^^[n]).forAll(freeNotUsedPoint);
	end;

procedure grid.interactiveOutputCell (m, n: seqIndex0);
	procedure drawPoint (x: obj); (* whatToDo *)
		begin
			gridPoint(x).interactiveOutput;
		end;

	begin
		if (m <= rows) & (m >= 1) & (n <= cols) & (n >= 1) then
			getList(m, n).forAll(drawPoint);
	end;

procedure grid.interactiveOutput;
	var
		i: seqIndex0;
		rP1, rP2: realPoint;
	begin
(* initialize rP1,rP2 *)
		new(rP1);
		failNil(Handle(rP1));
		new(rP2);
		failNil(Handle(rP2));
		rP1.init;
		rP2.init;
(* first draw gridlines *)
		for i := 0 to rows do begin
			rP1.setXY(xMin, yMin + i * size);
			rP2.setXY(xMin + cols * size, yMin + i * size);
			rP1.drawLine(rP2);
(* waitForClick(animationFlag[ATriangulationItem]); *)
		end;
		for i := 0 to cols do begin
			rP1.setXY(xMin + i * size, yMin);
			rP2.setXY(xMin + i * size, yMin + rows * size);
			rP1.drawLine(rP2);
(* waitForClick(animationFlag[ATriangulationItem]); *)
		end;
(* second show points *)
(* for i := 1 to rows do *)
(* for j := 1 to cols do *)
(* interactiveOutputCell(i, j); *)
		rP1.free;
		rP2.free;
	end;

procedure grid.remove (gP: gridPoint);
	var
		l: list;
		e: elementH;
	begin
		l := getList(gP.getRow, gP.getCol);
		if gP.equal(l.first^^.value) then
			l.pop
		else begin
			e := l.first;
			while not gP.equal(e^^.next^^.value) do
				e := e^^.next;
			l.delete(e);
		end;
		nrPoints := pred(nrPoints);
	end;   (* remove *)

procedure grid.initialize (pV: pointVector; tolerance, sizeFactor: extended);
	var
		i: seqIndex0;
		hV: homogenVector;   (* temporary  *)
		gP: gridPoint;       (* gridPoint to be stored *)
		hL: list;

	function lessThanX (p, q: obj): boolean;
		begin
			lessThanX := point2d(p).xLessThan(point2d(q));
		end;

	function lessThanY (p, q: obj): boolean;
		begin
			lessThanY := point2d(p).yLessThan(point2d(q));
		end;

	function greaterThanX (p, q: obj): boolean;
		begin
			greaterThanX := lessThanX(q, p);
		end;

	function greaterThanY (p, q: obj): boolean;
		begin
			greaterThanY := lessThanY(q, p);
		end;

	function withinTolerance (x: obj): boolean; (* a property of x *)
  (* global are tolerance of grid.initialize and gP of grid.initialize *)
		begin
			withinTolerance := (abs(gridPoint(x).getNX - gP.getNX) < tolerance) & (abs(gridPoint(x).getNY - gP.getNY) < tolerance);
		end;

	begin
		assert(pV.length > 2, 'pointVector has more than two points');
		nrPoints := 0;
		xMin := point2d(pV.elements^^[pV.findExtreme(lessThanX)]).getX - tolerance;
		yMin := point2d(pV.elements^^[pV.findExtreme(lessThanY)]).getY - tolerance;
		xMax := point2d(pV.elements^^[pV.findExtreme(greaterThanX)]).getX + tolerance;
		yMax := point2d(pV.elements^^[pV.findExtreme(greaterThanY)]).getY + tolerance;
		size := sizeFactor * sqrt(((xMax - xMin) * (yMax - yMin)) / (pV.length - 1)); (* size of a gridcell *)
{$IFC THINK_PASCAL }
		rows := trunc((yMax - yMin) / size) + 1; (* number of rows *)
		cols := trunc((xMax - xMin) / size) + 1; (* number of columns *)
{$ELSEC }
		rows := truncLong((yMax - yMin) / size) + 1; (* number of rows *)
		cols := truncLong((xMax - xMin) / size) + 1; (* number of columns *)
{$ENDC }
(* ----------------- BEGIN initializing the matrix of lists -------------- *)

		g := newHomogenVector(rows);
		for i := 1 to rows do begin
			new(hV);
			failNil(Handle(hV));
			g.elements^^[i] := hV;
			hV.preset(cols, listType);
		end;
		g.length := rows;

(* ------------------ END initializing the matrix of lists ------------------ *)
(* ------------------ BEGIN putting points into the grid ------------------- *)

		for i := 1 to pV.length do begin
			new(gP);
			failNil(Handle(gP));
			gP.init;
			gP.setNX((point2d(pV.elements^^[i]).getX - xMin) / size);
			gP.setNY((point2d(pV.elements^^[i]).getY - yMin) / size);
			gP.setX(point2d(pV.elements^^[i]).getX);
			gP.setY(point2d(pV.elements^^[i]).getY);
			gP.setId(i);
			hL := getList(gP.getRow, gP.getCol);
			if 0 = hL.exists(withinTolerance) then begin (* if point doesn't exist,  append *)
				hL.append(gP);
				nrPoints := succ(nrPoints);
			end
			else
				gP.free;
		end;
(* ------------------ END putting points into the grid ------------------- *)
		assert(nrPoints > 2, 'Grid has more than two points');
	end;

function grid.getList (row, col: seqIndex0): list;
	begin
		getList := list(homogenVector(g.elements^^[row]).elements^^[col]);   (* select list in row,col *)
	end;

function grid.getMiddlePoint: gridPoint;
	var
		m, n: seqIndex0;    (* startcell of search *)
		i, j: integer;          (* indices of other cells to be searched *)
		p: gridPoint;          (*  found point to be returned *)

	function getPoint (m, n: integer; var p: gridPoint): boolean;
		begin
			if (m > rows) | (n > rows) | (m < 1) | (n < 1) then
				getPoint := false
			else if getList(m, n).length = 0 then
				getPoint := false
			else begin
				p := gridPoint(getList(m, n).ith(1));
				getPoint := true;
			end;
		end;

		label
			1;
	begin
		m := rows div 2;
		n := cols div 2;
		if getPoint(m, n, p) then   (* if in cell m,n ok *)
			goto 1;
		i := 0;
		while true do begin   (* point will be found, because there are more than two points in the grid *)
			i := succ(i);                (* and the search is exhaustive *)
        (* search top row *)
			for j := n - i to n + i do
				if getPoint(m + i, j, p) then
					goto 1;
       (* search bottom row *)
			for j := n - i to n + i do
				if getPoint(m - i, j, p) then
					goto 1;
       (* search left column  *)
			for j := m - i + 1 to m + i - 1 do
				if getPoint(j, n - i, p) then
					goto 1;
       (* search right column  *)
			for j := m - i + 1 to m + i - 1 do
				if getPoint(j, n + i, p) then
					goto 1;
		end;  (* while *)
1:
		getMiddlePoint := p;
	end;

(* ------------------------------- BEGIN CLOSESTPOINT ------------------------------- *)

function grid.getClosestPoint (p: gridPoint): gridPoint;
	var
		m, n: seqIndex0; (* cell of p, startcell for search *)
		i: integer;        (* indices of other cells to be searched *)
		cP: gridPoint;       (* temporary for the current closest gridPoint *)
		dMin: extended;  (* distance of the current closest gridPoint *)
		perpDist: extended; (* perpendecular distance to gridcells *)
		top, bot, lft, rgt: boolean; (* four searchdirections : top, bottom, left, right *)
		countDir: integer; (* counting how many searchdirections are left *)

	procedure searchCloserInCell (row, col: integer; p: gridPoint; var dMin: extended; var cP: gridPoint);
(* find in cell row,col the point cP closer to p than Distance dMin *)
(* if point is found set dMin to this distance of p to cP, else do nothing *)

		procedure distanceToP (x: obj);
(* global are p, cP, dMin *)
			var
				dist: extended;
			begin
				if not p.equal(x) then begin	(* don't take distance to p itsself into account *)
					dist := p.distanceX(point2d(x));
					if dist < dMin then begin
						dMin := dist;
						cP := gridPoint(x);
					end; (* if dist *)
				end; (* if not *)
			end;

		begin
			if (row >= 1) & (row <= rows) & (col >= 1) & (col <= cols) then   (* check all points in list for better one *)
				getList(row, col).forAll(distanceToP);
		end;

	procedure searchCloserInRow (row, col1, col2: integer; p: gridPoint; var dMin: extended; var cP: gridPoint);
		var
			i: integer;
		begin
			for i := col1 to col2 do
				searchCloserInCell(row, i, p, dMin, cP);
		end;

	procedure searchCloserInCol (col, row1, row2: integer; p: gridPoint; var dMin: extended; var cP: gridPoint);
		var
			i: integer;
		begin
			for i := row1 to row2 do
				searchCloserInCell(i, col, p, dMin, cP);
		end;

		label
			1;

	begin
		m := p.getRow;
		n := p.getCol;
(* search the cell of p *)
		dMin := EXT_MAX;   (* until now the best distance *)
		searchCloserInCell(m, n, p, dMin, cP);  (* no point found : dmin == EXT_MAX , cP undefined *)
(* all searchdirections are valid *)
		top := true;
		bot := true;
		lft := true;
		rgt := true;
		countDir := 4;
		i := 0;
		while true do begin   (* point will be found, because there are more than two points in the grid *)
			i := succ(i);                (* and the search is exhaustive *)
(* ------------------- BEGIN TOP ------------------------------ *)
			if top then begin  (* search top row *)
				perpDist := p.rowDistance(m + i);
				if perpDist < dMin then  (* the cells in row still must be searched for closer points *)
					searchCloserInRow(m + i, n - i, n + i, p, dMin, cP)   (* if better point, dMin and cP will be updated *)
				else begin  (* mark the searchdirection as invalid *)
					top := false;
					if countDir = 0 then  (* search is finished *)
						goto 1
					else
						countDir := pred(countDir);
				end;  (* if perpDist < dMin *)
			end;  (* if top *)
(* ------------------- END TOP ------------------------------ *)
(* ------------------- BEGIN BOT ------------------------------ *)
			if bot then begin  (* search bottom row *)
				perpDist := p.rowDistance(m - i);
				if perpDist < dMin then  (* the cells in row still must be searched for closer points *)
					searchCloserInRow(m - i, n - i, n + i, p, dMin, cP)   (* if better point, dMin and cP will be updated *)
				else begin  (* mark the searchdirection as invalid *)
					bot := false;
					if countDir = 0 then  (* search is finished *)
						goto 1
					else
						countDir := pred(countDir);
				end;  (* if perpDist < dMin *)
			end;  (* if bot *)
(* ------------------- END BOT ------------------------------ *)
(* ------------------- BEGIN LEFT ------------------------------ *)
			if lft then begin  (* search left column *)
				perpDist := p.colDistance(n - i);
				if perpDist < dMin then  (* the cells in column still must be searched for closer points *)
					searchCloserInCol(n - i, m - i + 1, m + i - 1, p, dMin, cP)   (* if better point, dMin and cP will be updated *)
				else begin  (* mark the searchdirection as invalid *)
					lft := false;
					if countDir = 0 then  (* search is finished *)
						goto 1
					else
						countDir := pred(countDir);
				end;  (* if perpDist < dMin *)
			end;  (* if lft *)
(* ------------------- END LEFT ------------------------------ *)
(* ------------------- BEGIN RIGHT ------------------------------ *)
			if rgt then begin  (* search right column *)
				perpDist := p.colDistance(n + i);
				if perpDist < dMin then  (* the cells in column still must be searched for closer points *)
					searchCloserInCol(n + i, m - i + 1, m + i - 1, p, dMin, cP)   (* if better point, dMin and cP will be updated *)
				else begin  (* mark the searchdirection as invalid *)
					rgt := false;
					if countDir = 0 then  (* search is finished *)
						goto 1
					else
						countDir := pred(countDir);
				end;  (* if perpDist < dMin *)
			end;  (* if rgt *)
(* ------------------- END RIGHT ------------------------------ *)
			if countDir = 0 then
				goto 1;
		end;  (* while *)
1:
		getClosestPoint := cP;
	end;  (* getClosestPoint *)

function grid.getClosestPointNaive (p: gridPoint): gridPoint;
	var
		minDist: extended;  (* current minimum distance *)
		m, n: seqIndex0;
		cP: gridPoint;        (* current closest point *)

	procedure distanceToP (x: obj);  (* computes distance to global point p *)
(* global are p,gE,minCos *)
		var
			hDist: extended;
		begin
			if not p.equal(gridPoint(x)) then begin
				hDist := p.distanceX(gridPoint(x));
				if hDist < minDist then begin
					minDist := hDist;
					cP := gridPoint(x);
				end;
			end;
		end;

	begin
		minDist := EXT_MAX;
		cP := nil;
		for m := 1 to rows do
			for n := 1 to cols do
				getList(m, n).forAll(distanceToP);

		getClosestPointNaive := cP;
	end; (* getClosestPointNaive *)

procedure grid.checkClosestPoint (p, closestP: gridPoint);
	var
		cPNaive: gridPoint;
		distNaive, dist: extended;

	begin
		cPNaive := getClosestPointNaive(p);
		distNaive := cPNaive.distanceX(p);
		dist := closestP.distanceX(p);
		if abs(dist - distNaive) > 1e-6 then begin
{$IFC Debug }
			writeln('closest point naive distance to p different to closest point distance to p');
			writeln('p.getnX : ', p.getnX : 10 : 2);
			writeln('p.getnY : ', p.getnY : 10 : 2);
			writeln('closestP.getnX : ', closestP.getnX : 10 : 2);
			writeln('closestP.getnY : ', closestP.getnY : 10 : 2);
			writeln('cPNaive.getnX : ', cPNaive.getnX : 10 : 2);
			writeln('cPNaive.getnY : ', cPNaive.getnY : 10 : 2);
			debug := true;
			closestP := getClosestPoint(p);
{$ENDC }
			assert(false, 'checkClosestPoint');
		end;
	end;  (* checkClosestPoint *)

(* ------------------------------- END CLOSESTPOINT ------------------------------- *)

(* ------------------------------- BEGIN MAXANGLEPOINT ------------------------------- *)

function grid.getMaxAnglePointNaive (gE: gridEdge; var p: gridPoint): boolean;
	var
		minCos: extended;  (* current minimal cosinus *)
		m, n: seqIndex0;   (* current index of point with minimal cosinus,  m,n : current row, col *)

	procedure cosIfRight (x: obj);  (* computes cosinus if point lies right *)
(* global are p,gE,minCos *)
		var
			hCos: extended;
		begin
			if not gE.p1.equal(gridPoint(x)) & not gE.p2.equal(gridPoint(x)) & gE.rightOf(gridPoint(x)) then begin
				hCos := gE.cosAngle(gridPoint(x));
				if hCos < minCos then begin
					minCos := hCos;
					p := gridPoint(x);
				end;
			end;
		end;

	begin
		minCos := EXT_MAX;
		p := nil;
		for m := 1 to rows do
			for n := 1 to cols do
				getList(m, n).forAll(cosIfRight);

		getMaxAnglePointNaive := p <> nil;
	end;  (* getMaxAnglePointNaive *)

(* ------------   BEGIN getMaxAnglePoint  ----------------------------------- *)

function grid.getMaxAnglePoint (gE: gridEdge; var p: gridPoint): boolean;
	var
		found: boolean;       (* if a point with minimal cosinus is found *)
		minCos: extended;   (* the value of the current best minimal cosinus *)
		i: integer;
(* values of the current search area *)
		m1, n1, m2, n2, m3, n3, dm, dn: integer;
	label
		1;

(* GLOBAL VARIABLES FOR THE SEARCHPROCEDURES ARE  : found, minCos,p, *)
(*                                                                                        m1 , n1 , m2 , n2 , m3 , n3,dm,dn *)

	procedure m3n3 (gE: gridEdge; i: integer);
		begin
			case gE.octant of
				1, 2:  begin
					m3 := gE.p1.getRow - i;
					n3 := gE.p2.getCol + i;
				end;
				3, 4:  begin
					m3 := gE.p2.getRow + i;
					n3 := gE.p1.getCol + i;
				end;
				5, 6:  begin
					m3 := gE.p1.getRow + i;
					n3 := gE.p2.getCol - i;
				end;
				7, 8:  begin
					m3 := gE.p2.getRow - i;
					n3 := gE.p1.getCol - i;
				end;
			end; (* case *)
		end;

	procedure search (gE: gridEdge; m, n: integer);

		procedure cosIfRight (x: obj);  (* computes cosinus if point lies right *)
(* global are p,gE,minCos *)
			var
				hCos: extended;
			begin
				if not gE.p1.equal(gridPoint(x)) & not gE.p2.equal(gridPoint(x)) & gE.rightOf(gridPoint(x)) then begin
					hCos := gE.cosAngle(gridPoint(x));
					if hCos < minCos then begin
						minCos := hCos;
						p := gridPoint(x);
						found := true;
					end;
				end;
			end;

		begin
{$IFC Debug }
			if debug then
				writeln(m, n);
{$ENDC }
			if (m >= 1) & (m <= rows) & (n >= 1) & (n <= cols) then
				getList(m, n).forAll(cosIfRight);
		end; (* search *)

	procedure searchtriangle (gE: gridEdge);
(* searchtriangle searches the triangle (in the first octant ) : *)
(* (j1,i1) (j2,i2) (j1,i2) , where j1,i1 : cells of gE.p1;j2,i2 : cells of gE.p2 *)
		var
			i1, j1, i2, j2, i, j, jold, iold, ih, di: integer;  (* help variables *)
			y, dy, x, dx: extended;

		begin
{$IFC Debug }
			if debug then
				writeln('begin searchtriangle');
{$ENDC }
			m3n3(gE, 0);  (* calculate third point *)
			i1 := gE.p1.getCol;
			j1 := gE.p1.getRow;
			i2 := gE.p2.getCol;
			j2 := gE.p2.getRow;
			i := i1;
			j := j1;
			jold := j;
			iold := i;
			search(gE, j, i);
			if (i <> i2) | (j <> j2) then (* not the same cell *)
				if gE.octant in [1, 8, 4, 5] then begin
					if gE.octant in [1, 8] then begin
						y := gE.q * i + gE.b; (* first intersection with gridline *)
						di := 1;
						dy := gE.q;
					end
					else begin (* octant 4,5 *)
						y := gE.q * (i - 1) + gE.b;
						di := -1;
						dy := -gE.q;
					end;
					while (i <> i2) | (j <> j2) do begin
{$IFC THINK_PASCAL }
						j := trunc(y) + 1;
{$ELSEC }
						j := truncLong(y) + 1;
{$ENDC }
						if j <> jold then begin (* first cell *)
							search(gE, j, i);
                 (* ------------- begin triangle ------------ *)
							if (gE.octant in [8, 4]) & (i <> i1) then begin
								ih := i - di;
								repeat
									search(gE, j, ih);
									ih := ih - di;
								until ih + di = i1;
							end;
                 (* ------------- end triangle -------------- *)
						end;  (* first cell *)
						if (i <> i2) | (j <> j2) then begin (* second cell *)
							i := i + di;
							search(gE, j, i);
                 (* ------------- begin triangle ------------ *)
							if (gE.octant in [5, 1]) & (j <> jold) then begin
								ih := i;
								repeat
									search(gE, jold, ih);
									ih := ih + di;
								until ih - di = i2;
							end;
                 (* ------------- end triangle -------------- *)
						end;  (* second cell *)
						jold := j;
						y := y + dy;
					end;  (* while *)
				end
				else begin (* octant in [2,3,6,7] *)
					if gE.octant in [2, 3] then begin
						di := 1;
						dx := gE.qinv;
						if gE.q = EXT_MAX then
							x := gE.p1.getNX
						else
							x := (j - gE.b) * gE.qinv;
					end
					else begin (* octant in [6,7] *)
						di := -1;
						dx := -gE.qinv;
						if gE.q = EXT_MAX then
							x := gE.p1.getNX
						else
							x := (j - 1 - gE.b) * gE.qinv;
					end;
					while (i <> i2) | (j <> j2) do begin
{$IFC THINK_PASCAL }
						i := trunc(x) + 1;
{$ELSEC }
						i := truncLong(x) + 1;
{$ENDC }
						if i <> iold then (* first cell *)
							search(gE, j, i);
                 (* ------------- begin triangle ------------ *)
						if (gE.octant in [2, 6]) & (i <> i2) then begin
							ih := i + di;
							repeat
								search(gE, j, ih);
								ih := ih + di;
							until ih - di = i2;
						end;
                 (* ------------- end triangle -------------- *)
						if (i <> i2) | (j <> j2) then begin (* second cell *)
							j := j + di;
							search(gE, j, i);
                 (* ------------- begin triangle ------------ *)
							if (gE.octant in [3, 7]) & (i <> i1) then begin
								ih := i + di;
								repeat
									search(gE, j, ih);
									ih := ih + di;
								until ih - di = i1;
							end;
                 (* ------------- end triangle -------------- *)
						end;  (* second cell *)
						iold := i;
						x := x + dx;
					end;  (* while *)
				end;  (* octant in 2,3,6,7 *)
{$IFC Debug }
			if debug then
				writeln('end searchtriangle');
{$ENDC }
		end; (* searchtriangle *)

	function boxsearched (gE: gridEdge): boolean;
		var
			xm, ym, r: extended; (* middle and radius of a circle *)
			lln, llm, run, rum: integer; (* left lower and right upper corner of minmaxbox of circle *)
			h: boolean;

		procedure circle3p (gE: gridEdge; p: gridPoint; var xm, ym, r: extended);
			begin
				p.circleCenterX(gE.p1, gE.p2, xm, ym);
				r := sqrt(sqr(gE.p1.nX - xm) + sqr(gE.p1.nY - ym));
			end;

		begin
			circle3p(gE, p, xm, ym, r);
      (* ---------- prevent integer overflow *)
			if r > 5000 then
				r := 5000;
			if (xm > 5000) then
				xm := 5000;
			if (ym > 5000) then
				ym := 5000;
			if (xm < -5000) then
				xm := -5000;
			if (ym < -5000) then
				ym := -5000;

{$IFC THINK_PASCAL }
(* cell  left lower corner *)
			lln := trunc(xm - r) + 1;
			llm := trunc(ym - r) + 1;
(* right upper corner *)
			run := trunc(xm + r) + 1;
			rum := trunc(ym + r) + 1;
{$ELSEC }
(* cell  left lower corner *)
			lln := truncLong(xm - r) + 1;
			llm := truncLong(ym - r) + 1;
(* right upper corner *)
			run := truncLong(xm + r) + 1;
			rum := truncLong(ym + r) + 1;
{$ENDC }
			h := true;
			case gE.octant of
				1, 2:  begin
					if llm < m3 then
						h := false
					else if run > n3 then
						h := false;
					if gE.octant = 1 then begin
						if lln < n1 then
							h := false;
					end
					else if rum > m2 then
						h := false;
				end;
(* -------------- *)
				3, 4:  begin
					if rum > m3 then
						h := false
					else if run > n3 then
						h := false;
					if gE.octant = 3 then begin
						if llm < m1 then
							h := false;
					end
					else if lln < n2 then
						h := false;
				end;
(* ---------------- *)
				5, 6:  begin
					if rum > m3 then
						h := false
					else if lln < n3 then
						h := false;
					if gE.octant = 5 then begin
						if run > n1 then
							h := false;
					end
					else if llm < m2 then
						h := false;
				end;
(* --------------------- *)
				7, 8:  begin
					if llm < m3 then
						h := false
					else if lln < n3 then
						h := false;
					if gE.octant = 7 then begin
						if rum > m1 then
							h := false;
					end
					else if run > n2 then
						h := false;
				end;
			end; (* case *)
			boxsearched := h;
		end; (* boxsearched *)

	function last (gE: gridEdge): boolean;
		var
			lst: boolean;
		begin
			case gE.octant of
				1: 
					lst := (n1 <= 1) & (m3 <= 1) & (n3 >= cols);
				2: 
					lst := (m2 >= rows) & (m3 <= 1) & (n3 >= cols);
				3: 
					lst := (m1 <= 1) & (m3 >= rows) & (n3 >= cols);
				4: 
					lst := (n2 <= 1) & (m3 >= rows) & (n3 >= cols);
				5: 
					lst := (n1 >= cols) & (m3 >= rows) & (n3 <= 1);
				6: 
					lst := (m2 <= 1) & (m3 >= rows) & (n3 <= 1);
				7: 
					lst := (m1 >= rows) & (m3 <= 1) & (n3 <= 1);
				8: 
					lst := (n2 >= cols) & (m3 <= 1) & (n3 <= 1);
			end; (* case *)
			last := lst;
		end; (* last *)


	procedure searchStripe (gE: gridEdge; i: integer);
		var
			h, dmStep, dnStep: integer;
		begin
{$IFC Debug }
			if debug then
				writeln('begin searchStripe');
{$ENDC }
			m3n3(gE, i);   (* calculate third point of trapez *)
			case gE.octant of
				1, 5:  begin
(* left column *)
					h := gE.p1.getCol - dn * i;
					repeat
						gE.prevCell(m1, n1);
					until n1 = h;
					h := m1;
					while (h <> m3 + dm) do begin
						search(gE, h, n1);
						h := h + dm;
					end;
(* right column *)
					repeat
						gE.nextCell(m2, n2);
						search(gE, m2, n2);
					until n2 = n3;
(* correct the trapez if necessary *)
					dmStep := dm;
					gE.nextCell(m2, n2);
					if (n2 = n3) then begin (* one cell to correct *)
						search(gE, m2, n2);
						dmStep := dmStep + dm;
					end
					else
						gE.prevCell(m2, n2);  (* step back *)
					h := m2 + dmStep;
					while h <> m3 + dm do begin
						search(gE, h, n3);
						h := h + dm;
					end;
(* row below *)
					h := n1 + dn;
					while h <> n3 do begin
						search(gE, m3, h);
						h := h + dn;
					end;
				end;
(* --------------------------------------- *)
				4, 8:  begin
(* left column *)
					h := gE.p2.getCol - dn * i;
					repeat
						gE.nextCell(m2, n2);
					until n2 = h;
					h := m2;
					while (h <> m3 + dm) do begin
						search(gE, h, n2);
						h := h + dm;
					end;
(* right column *)
					repeat
						gE.prevCell(m1, n1);
						search(gE, m1, n1);
					until n1 = n3;
(* correct the trapez if necessary *)
					dmStep := dm;
					gE.prevCell(m1, n1);
					if (n1 = n3) then begin (* one cell to correct *)
						search(gE, m1, n1);
						dmStep := dmStep + dm;
					end
					else
						gE.nextCell(m1, n1);  (* step back *)
					h := m1 + dmStep;
					while h <> m3 + dm do begin
						search(gE, h, n3);
						h := h + dm;
					end;
(* row below *)
					h := n2 + dn;
					while h <> n3 do begin
						search(gE, m3, h);
						h := h + dn;
					end;
				end;
(* ----------------------------------------- *)
				2, 6:  begin
(* row above  *)
					h := gE.p2.getRow - dm * i;
					repeat
						gE.nextCell(m2, n2);
					until m2 = h;
					h := n2;
					while (h <> n3 + dn) do begin
						search(gE, m2, h);
						h := h + dn;
					end;
(* row below *)
					repeat
						gE.prevCell(m1, n1);
						search(gE, m1, n1);
					until m1 = m3;
(* correct the trapez if necessary *)
					dnStep := dn;
					gE.prevCell(m1, n1);
					if (m1 = m3) then begin (* one cell to correct *)
						search(gE, m1, n1);
						dnStep := dnStep + dn;
					end
					else
						gE.nextCell(m1, n1);  (* step back *)
					h := n1 + dnStep;
					while h <> n3 + dn do begin
						search(gE, m3, h);
						h := h + dn;
					end;
(* right column *)
					h := m2 + dm;
					while h <> m3 do begin
						search(gE, h, n3);
						h := h + dm;
					end;
				end;
(* ----------------------------------------- *)
				3, 7:  begin
(* row below *)
					h := gE.p1.getRow - dm * i;
					repeat
						gE.prevCell(m1, n1);
					until m1 = h;
					h := n1;
					while (h <> n3 + dn) do begin
						search(gE, m1, h);
						h := h + dn;
					end;
(* row above *)
					repeat
						gE.nextCell(m2, n2);
						search(gE, m2, n2);
					until m2 = m3;
(* correct the trapez if necessary *)
					dnStep := dn;
					gE.nextCell(m2, n2);
					if (m2 = m3) then begin (* one cell to correct *)
						search(gE, m2, n2);
						dnStep := dnStep + dm;
					end
					else
						gE.prevCell(m2, n2);  (* step back *)
					h := n2 + dnStep;
					while h <> n3 + dn do begin
						search(gE, m3, h);
						h := h + dn;
					end;
(* right column  *)
					h := m1 + dm;
					while h <> m3 do begin
						search(gE, h, n3);
						h := h + dm;
					end;
				end;
(* --------------------------------------- *)

			end; (* case *)
{$IFC Debug }
			if debug then
				writeln('end searchStripe');
{$ENDC }
		end;  (* searchStripe *)

	begin
		found := false;
		minCos := EXT_MAX;
(* set the start variables of the searcharea *)
		m1 := gE.p1.getRow;
		n1 := gE.p1.getCol;
		m2 := gE.p2.getRow;
		n2 := gE.p2.getCol;
		if gE.octant in [1, 2, 7, 8] then
			dm := -1
		else
			dm := 1;
		if gE.octant in [1, 2, 3, 4] then
			dn := 1
		else
			dn := -1;
(* start search *)
		searchtriangle(gE);   (* first searcharea : triangle *)
		i := 1;
(* while search isn't finished *)
		while true do begin
      (* first exit : point is found and the minmaxbox of circle is searched *)
			if found then
				if boxsearched(gE) then
					goto 1;
(* continue search *)
			searchStripe(gE, i);     (* search next stripe around triangle *)
(* second exit :  all cells right of gE are searched, so stop *)
			if last(gE) then
				goto 1;
			i := succ(i);
		end;  (* while *)
1:
		getMaxAnglePoint := found;
	end; (* getMaxAnglePoint *)

(* ------------   END getMaxAnglePoint  ----------------------------------- *)

procedure grid.checkMaxAnglePoint (gE: gridEdge; p3: gridPoint; found: boolean);
	var
		p3Naive: gridPoint;
		foundNaive: boolean;
		angleNaive, angle: extended;

	procedure DataOutandHalt;
		begin
{$IFC Debug }
			writeln('gE.octant :  ', gE.octant);
			writeln('gE.p1.getnX : ', gE.p1.getnX : 10 : 2);
			writeln('gE.p1.getnY : ', gE.p1.getnY : 10 : 2);
			writeln('gE.p2.getnX : ', gE.p2.getnX : 10 : 2);
			writeln('gE.p2.getnY : ', gE.p2.getnY : 10 : 2);
			debug := true;
			dummy := getMaxAnglePoint(gE, h);
{$ENDC }
			assert(false, 'checkMaxAnglePoint');
		end;

	begin
		foundNaive := getMaxAnglePointNaive(gE, p3Naive);
		if foundNaive then begin
			angleNaive := gE.cosAngle(p3Naive);
			if not found then begin
{$IFC Debug }
				writeln('Naive point found, point not found');
				writeln('p3Naive.nX : ', p3Naive.nX : 10 : 2);
				writeln('p3Naive.nY : ', p3Naive.nY : 10 : 2);
				writeln('p3Naive.getRow : ', p3Naive.getRow);
				writeln('p3Naive.getCol : ', p3Naive.getCol);
{$ENDC }
				DataOutandHalt;
			end
			else begin (* both points found *)
				angle := gE.cosAngle(p3);
				if abs(angle - angleNaive) > 1e-6 then begin
{$IFC Debug }
					writeln('p3 angle is not equal p3Naive angle: ');
					writeln('p3.getnX : ', p3.getnX : 10 : 2);
					writeln('p3.getnY : ', p3.getnY : 10 : 2);
					writeln('p3Naive.nX : ', p3Naive.nX : 10 : 2);
					writeln('p3Naive.nY : ', p3Naive.nY : 10 : 2);
					p3Naive.drawline(gE.p1);
{$ENDC }
					DataOutandHalt;
				end;
			end;
		end;
(* ---- ok else, but I'm starting to get sick on all these if then begin end else begin end ----- *)

		if not foundNaive & found then begin
{$IFC Debug }
			writeln('naive point NOT found : ');
			writeln('maxAnglePoint point found');
{$ENDC }
			DataOutandHalt;
		end;
	end;  (* checkMaxAnglePoint *)

(* ------------------------------- END MAXANGLEPOINT  ------------------------------- *)


(* ----------------- BEGIN TRIANGULATION --------------------------------- *)

function NewGridEdge (newP1, newP2: gridPoint; startV, endV: seqIndex): gridEdge;
	var
		gE: gridEdge;
	begin
		new(gE);
		failNil(Handle(gE));
		gE.init;
		gE.initialize(newP1, newP2, startV, endV);
		NewGridEdge := gE;
	end;

function grid.triangulation (p1, p2: gridPoint): triangulation2d;
(* shortenings : HEL := half edge list, HE := half edge *)
	var
		t: triangulation2d;
		l: doubleList;   (* list for HE's *)
		hP12, P12, P21, P32, P13: gridEdge;
		p3: gridPoint;
		found, first: boolean;
		p3Id: seqIndex;

	procedure addNewEdge (var gE: gridEdge);
(* new edge  to t : triangulation2d  *)
		begin
			t.edges.append(gE);
			gE.id := t.edges.length;
		end;

	procedure append (var gE: gridEdge);
(* a new edge is appended into the list l and into the edge list of the triangulation t *)
(* global variables : t,l *)
		begin
			addNewEdge(gE);  (* appending into t *)
			l.append(gE);        (* appending into l *)
		end;

	procedure moveFront2Rear;
		begin
			l.append(l.first^^.value);
			l.delete(l.first);
		end;

	procedure deleteFirst;
		begin
			l.delete(l.first);
		end;

	procedure deleteLast;
		begin
			l.delete(l.last);
		end;

	procedure appendPoint (var gP: gridPoint);
(* appending into triangulation t and setting usedflag *)
		begin
			t.points.append(gP);
			gP.setUsed;
		end;

	begin
(* --------------------- BEGIN initialize triangulation 2D ------------------------------- *)
		new(t);
		failNil(Handle(t));
		t.initialize(nrPoints + 3, 3 * nrPoints + 3, 2 * nrPoints + 1);
		currentTriangulation := t;
(* --------------------- END initialize triangulation 2D ------------------------------- *)
		new(l);
		failNil(Handle(l));
		l.init;

		P12 := NewGridEdge(p1, p2, 1, 2);
		P21 := NewGridEdge(p2, p1, 2, 1);
		appendPoint(p1);
		appendPoint(p2);

		append(P21);    (* P21 into list and triangulation *)
		l.append(P12);  (* P12 only into list,but P12 has the same id, same triangles as P21. *)
		P12.id := P21.id;
		first := true;    (* to handle the special case of two gridEdges P21,P12, but only one graphEdgeTriangulation *)
		while l.first <> nil do begin
			hP12 := gridEdge(l.last^^.value);   (* get current edge *)
			found := getMaxAnglePoint(hP12, p3);
{$IFC Debug }
			checkMaxAnglePoint(hP12, p3, found);
{$ENDC }

			if found then begin
				p3Id := t.points.length + 1; (* if p3 is not used this would be the new id *)
  (* two new edges, touch : is an edge alread in use ? *)
				P32 := NewGridEdge(p3, hP12.p2, p3Id, hP12.endVertex);
				P13 := NewGridEdge(hP12.p1, p3, hP12.startVertex, p3Id);

				if p3.isUsed then begin  (* touch means : is edge already in triangulation ? *)
(* ---------------- RIGHT TOUCH ---------------------- *)
					if P13.isDual(gridEdge(l.first^^.value)) then begin
{$IFC AnimationEnabled AND AGridTriangulation }
						if animationFlag[ATriangulationItem] then begin
							P32.p1.drawLine(P32.p2);
							displayInfoString('Right Touch');
							waitForClick(animationFlag[ATriangulationItem]);
						end;
{$ENDC }

						remove(P13.p1);   (* remove point from grid *)
    (* P13 is  already in use, we need the triangleA,triangleB,startVertex,endVertex info *)
						P13.free;
						P13 := gridEdge(l.first^^.value);       (* P13 is now dual P31 !! *)
						P32.startVertex := P13.startVertex;

						deleteFirst;
						deleteLast;
						append(P32);
						t.addNewTriangle(hP12, P32, P13);
					end
(* ---------------- LEFT TOUCH ---------------------- *)
					else if P32.isDual(gridEdge(l.last^^.pred^^.value)) then begin
{$IFC AnimationEnabled AND AGridTriangulation }
						if animationFlag[ATriangulationItem] then begin
							P13.p1.drawLine(p3);
							displayInfoString('Left Touch');
							waitForClick(animationFlag[ATriangulationItem]);
						end;
{$ENDC }

						remove(P32.p2);    (* remove point from grid *)

 (* P32 is already in use, we need the triangleA, triangleB ,startVertex,endVertex info *)
						P32.free;
						P32 := gridEdge(l.last^^.pred^^.value);       (* P32 is now dual P23 !! *)
						P13.endVertex := P32.endVertex;
						deleteLast;
						deleteLast;
						append(P13);
						if first then begin
							first := false;
							P12.free;
							t.addNewTriangle(P21, P32, P13);
						end
						else
							t.addNewTriangle(hP12, P32, P13);
					end
(* ---------------- NO TOUCH ---------------------- *)
					else begin
						P32.free;
						P13.free;
						moveFront2Rear;
					end;
				end
(* ---------------- p3 IS NOT USED ------------------ *)
				else begin
{$IFC AnimationEnabled AND AGridTriangulation }
					if animationFlag[ATriangulationItem] then begin
						P32.p1.drawLine(P32.p2);
						P13.p1.drawLine(P13.p2);
						displayInfoString('No Touch');
						waitForClick(animationFlag[ATriangulationItem]);
					end;
{$ENDC }
					appendPoint(p3);
					deleteLast;
					append(P32);
					append(P13);
					if first then begin
						first := false;
						P12.free;
						t.addNewTriangle(P21, P32, P13);
					end
					else
						t.addNewTriangle(hP12, P32, P13);
				end
			end
(* ---------------- NO POINT WAS FOUND --------------- *)
			else begin
				if l.first <> nil then begin
					deleteLast;
					if first then begin
						first := false;
						P12.free;
					end;
					if l.first <> nil then
						moveFront2Rear;
				end;
			end;  (* else if getMaxAnglePoint *)
{$IFC AnimationEnabled }
			if userAbort then begin
				l.free;
				freeNotUsedPoints;
				t.freeAll;
				triangulation := nil;
				exit(triangulation);
			end;
{$ENDC }
		end;  (* while *)
		l.free;
		triangulation := t;
	end; (* triangulation *)

(* ---------------------------------- END TYPE GRID ----------------------------------- *)

{$IFC testInvariants }
function triangulationInvariant (t: triangulation): triangulation;
	var
		v: homogenVector;
		i: seqIndex0;

	function vOk (v1, v2, v3, v4, v5, v6: longInt): boolean;
		begin
			vOk := (v1 = v2) & (v2 < v3) & (v3 = v4) & (v4 < v5) & (v5 = v6);
		end;

	function triangleOk (tr: triangle): boolean;
		begin
			with graphEdge(t.edges.ith(tr.edgeA)) do begin
				int(v.elements^^[1]).value := startVertex;
				int(v.elements^^[2]).value := endVertex;
			end;
			with graphEdge(t.edges.ith(tr.edgeB)) do begin
				int(v.elements^^[3]).value := startVertex;
				int(v.elements^^[4]).value := endVertex;
			end;
			with graphEdge(t.edges.ith(tr.edgeC)) do begin
				int(v.elements^^[5]).value := startVertex;
				int(v.elements^^[6]).value := endVertex;
			end;
			v.sortAll(intLessThan);
			triangleOk := vOk(int(v.ith(1)).value, int(v.ith(2)).value, int(v.ith(3)).value, int(v.ith(4)).value, int(v.ith(5)).value, int(v.ith(6)).value);
		end;

	begin
		if t <> nil then begin
			new(v);
			failNil(Handle(v));
			v.preset(6, intType);
			for i := 1 to t.triangles.length do
				assert(triangleOk(triangle(t.triangles.ith(i))), StringOf('Triangle ', i : 1, ' ok.'));
			v.freeAll;
		end;
		triangulationInvariant := t;
	end;
{$ENDC }


function pointVector.gridTriangulation (tolerance, gridSizeFactor: extended): obj;
	var
		grd: grid;
		p1, p2: gridPoint;
{$IFC AnimationEnabled AND AGridTriangulation }
		save: PenState;
{$ENDC }
	begin
{$IFC Debug }
		debug := false;
{$ENDC }
{$IFC AnimationEnabled AND AGridTriangulation }
		GetPenState(save);
{$ENDC }
		new(grd);
		failNil(Handle(grd));
		grd.init;
		grd.initialize(self, tolerance, gridSizeFactor);

{$IFC AnimationEnabled AND AGridTriangulation }
		if animationFlag[ATriangulationItem] then
			grd.interactiveOutput;
{$ENDC }

(* to get a start edge for the triangulation : get some point p1 in the middle and the closest point p2 to p1 *)
		p1 := grd.getMiddlePoint;
		p2 := grd.getClosestPoint(p1);

{$IFC Debug }
		grd.checkClosestPoint(p1, p2);
{$ENDC }

{$IFC AnimationEnabled AND AGridTriangulation }
		if animationFlag[ATriangulationItem] then begin
			PenSize(3, 3);
			p1.drawLine(p2);
		end;
{$ENDC }

{$IFC testInvariants }
		gridTriangulation := triangulationInvariant(grd.triangulation(p1, p2));
{$ElSEC }
		gridTriangulation := grd.triangulation(p1, p2);
{$ENDC }
		grd.free;
{$IFC AnimationEnabled AND AGridTriangulation }
		SetPenState(save);
{$ENDC }
	end;
(* ----------------- END TRIANGULATION --------------------------------- *)
end.
