(*	XYZ GeoBench (eXperimental geometrY Zurich):		*)
(*	An object oriented, interactive workbench for		*)
(*	experimental geometric computation.					*)
(*														*)
(*	 1999 by Peter Schorn, peter.schorn@acm.org		*)
(*	Culmannstrasse 77, CH-8006 Zrich, Switzerland		*)
(*	This software may be copied freely for educational	*)
(*	and research purposes only. For commercial use,		*)
(*	please contact the author.							*)
(*														*)
(*	Version 5.0.5: 30-Sep-1999							*)

unit infiniteUndirectedGraph;

interface

uses
	graph;

type
	infiniteVertex = object(obj)
			id: longInt;										(*	id of the vertex.																*)
			neighbors: avlTree;							(* 'of infiniteEdge', sorted by 'value'.										*)
			verticesReference: reference;				(*	where the vertex is stored in the 'vertices' dictionary.				*)
			verticesByDegreeReference: reference;	(* where the vertex is stored in the 'verticesByDegree' dictionary.	*)
			procedure init;
			override;
			procedure free;
			override;
			procedure textOutput;
			override;										(* for testing only																*)
			procedure destroy;
		end;

	infiniteEdge = object(int)						(* The value denotes the number of the adjacent vertex.					*)
			procedure free;
			override;
			procedure destroy;
		end;

	infiniteUndirectedGraph = object(obj)
(*	Defines an undirected graph where the vertices are (potentially) the integers (longInt).		*)
(*	In the following denotes n the current number of vertices present in the graph.					*)
			vertices: avlTree;				(*	of 'infiniteVertex', sorted by 'id'.								*)
			verticesByDegree: avlTree;	(*	of 'infiniteVertex', sorted by degree.							*)
			procedure init;
			override;
			procedure free;
			override;
			procedure textOutput;		(* for testing only *)
			override;

			procedure addVertex (id: longInt);
(*	Adds the vertex with number 'id' to the graph.	  If the vertex is already present, the			*)
(*	procedure does nothing . Cost O(log n).																	*)

			procedure addEdge (startVertex, endVertex: longInt);
(*	Adds an edge between the vertices with number 'startVertex' and number 'endVertex'.		*)
(*	If the edge is already present, the procedure does nothing.	Cost O(log n).						*)

			procedure deleteVertex (id: longInt);
(*	Removes the vertex with number 'id'. If the vertex is not present, the procedure does			*)
(*	nothing. Cost O( (k +1) log n), where k is the number of vertices adjacent to 'id'.				*)

			procedure deleteVertexByVerticesReference (where: reference; id: longInt);
(*	Same as above, but the vertex to be removed is additionally specified by its reference in the	*)
(*	'vertices' dictionary. Cost O(k log n), where k is the number of vertices adjacent to 'id'.		*)

			procedure deleteEdge (startVertex, endVertex: longInt);
(*	Removes the edge between vertices with number 'startVertex' and number 'endvertex'.		*)
(*	If the edge is not present, the procedure does nothing. Cost O(log n).								*)

			function degree (id: longInt): longInt;
(*	Computes the degree of the vertex with number 'id'. If the vertex is not present, the			*)
(*	degree is -1.	 Cost O(log n).																				*)

			function countEdges: longInt;
(*	Computes the number of edges in the graph. Cost O(n).												*)

			function countEdgesSpecial (function property (startVertex, endVertex: longInt): boolean): longInt;
(*	Computes the number of edges in the graph which satisfy 'property'. Cost O(n+m).				*)

			function neighbors (id, which: longInt): longInt;
(*	Returns the number of the neighbor of the vertex with number 'id' which has ordinal number	*)
(*	'which'. If no such neighbor exists, the function returns -1. Cost O(log n).						*)

			function isConnected (a, b: longInt): boolean;
(*	Determines whether the vertices with number 'a' and 'b' are connected. Cost O(log n).			*)

			procedure fromSimpleUndirectedGraph (g: simpleUndirectedGraph);
(*	Initializes the structure from the given graph 'g'.	Must be called after 'init'.						*)
(*	Cost O((n + m) log n), where n is the number of vertices in 'g' and m is the number of edges	*)
(*	in 'g'.																											*)

			procedure fromSimpleUndirectedGraphSpecial (g: simpleUndirectedGraph; function edgeOk (ge: graphEdge): boolean; k: seqIndex0);
(*	Initializes the structure from the given graph 'g'.	Must be called after 'init'.						*)
(*	Cost O((n + m) log n), where n is the number of vertices in 'g' and m is the number of edges	*)
(*	in 'g'.	Vertices which are connected by an edge which does not satisfy 'edgeOk' are				*)
(*	connected	 to a clique of size k+2. The vertices of the graph get id's from 1 to					*)
(*	g.vertices.length, the vertices of the clique are numbered from g.vertices.length+1 to			*)
(*	g.vertices.length+k+2.																						*)

			function toUndirectedGraph (newVertices: vector; maxNumberOfVertices: longInt): undirectedGraph;
(*	Computes an isomorphic 'undirectedGraph' (only the first 'maxNumberOfVertices' vertices	*)
(*	are used) from the	given 'infiniteUndirectedGraph'. The elements of the vector					*)
(*	'newVertices' are copied as new vertices of the resulting graph. The user must supply at		*)
(*	least as many vertices as min('maxNumberOfVertices', 'self.vertices.numberOfElements').	*)

			function hennebergSequence (oldVertices: homogenVector; var reduceDegreeThreeVertex: seqIndex0): boolean;
(*	Compute a Henneberg 2-sequence for the graph 'self'. 'oldVertices' is a vector of				*)
(*	'markedRealPoint' which will be marked by the algorithm with the appropriate action for the	*)
(*	corresponding vertex. 'reduceDegreeThreeVertex' returns the number of degree three			*)
(*	reductions necessary. The result is true, iff a Henneberg 2-sequence was found. The			*)
(*	algorithm is randomized and might fail to find a sequence, even if it exists. The graph is		*)
(*	altered in the course of the algorithm.																	*)

			function infiniteUndirectedGraph.hennebergKReduction (k: seqIndex0): homogenVector;
(*	Removes all degree 'k' vertices from the graph until all vertices have degree more than 'k'	*)
(*	or the graph is empty.	The result contains the vertices to be deleted in descending order.			*)

			procedure reset;
(*	Resets the graph into an empty graph.																	*)

(*	private methods follow *)
			function lookup (id: longInt): seqIndex;
			procedure associate (v, n: longInt);
			procedure disAssociate (v, n: longInt);
			procedure updateDegree (v: infiniteVertex);
			function insertVertex (id: longInt): infiniteVertex;
		end;

procedure initFastUndirectedGraph (vertices, edges: longInt);

procedure exitFastUndirectedGraph;

function findHennebergSequenceMaxTrials (g: simpleUndirectedGraph; maxTrials: longInt): homogenVector;

function computeHennebergKReduction (g: simpleUndirectedGraph; k: seqIndex0): undirectedGraph;

function computeHennebergKReductionSpecial (g: simpleUndirectedGraph; k: seqIndex0; function edgeOk (ge: graphEdge): boolean): homogenVector;

implementation

{$SETC AHennebergKReduction = true }

var
	globalV: infiniteVertex;
	globalE: infiniteEdge;
	vertexList, edgeList: list;
{$IFC AnimationEnabled AND AHennebergKReduction }
	animationVertices: vector;
{$ENDC }

procedure makeMRP (oldVertices: homogenVector; count, vertexNumber, startVertexNumber, endVertexNumber: longInt);
	begin
		with markedRealPoint(oldVertices.ith(vertexNumber)) do
			if startVertexNumber = 0 then
				mark.value := StringOf(count : 1, '=', vertexNumber : 1)
			else
				mark.value := StringOf(count : 1, '=', vertexNumber : 1, ' (', startVertexNumber : 1, ',', endVertexNumber : 1, ')');
	end;

function infiniteUndirectedGraph.hennebergSequence (oldVertices: homogenVector; var reduceDegreeThreeVertex: seqIndex0): boolean;
	var
		continue: boolean;
		v: infiniteVertex;
		e: array[1..3, 1..2] of longInt;
		el: 0..3;
		a, b, c, count: longInt;

	procedure check (a, b: longInt);
		begin
			if not isConnected(a, b) then begin
				el := el + 1;
				e[el, 1] := a;
				e[el, 2] := b;
			end;
		end;

	begin (* infiniteUndirectedGraph.hennebergSequence *)
		reduceDegreeThreeVertex := 0;
		continue := vertices.numberOfElements > 0;
		count := 0;
		while continue do begin
			count := count + 1;
			v := infiniteVertex(verticesByDegree.getObject(verticesByDegree.extreme(left)));
			case v.neighbors.numberOfElements of
				0, 2:  begin
					makeMRP(oldVertices, count, v.id, 0, 0);
					deleteVertexByVerticesReference(v.verticesReference, v.id);
				end;
				3:  begin
					el := 0;
					a := infiniteEdge(v.neighbors.getObject(v.neighbors.ith(1))).value;
					b := infiniteEdge(v.neighbors.getObject(v.neighbors.ith(2))).value;
					c := infiniteEdge(v.neighbors.getObject(v.neighbors.ith(3))).value;
					check(a, b);
					check(b, c);
					check(c, a);
					if el > 0 then begin
{$IFC THINK_PASCAL }
						el := trunc(randomReal * el) + 1;
{$ELSEC }
						el := truncLong(randomReal * el) + 1;
{$ENDC }
						addEdge(e[el, 1], e[el, 2]);
						makeMRP(oldVertices, count, v.id, e[el, 1], e[el, 2]);
						reduceDegreeThreeVertex := reduceDegreeThreeVertex + 1;
						deleteVertexByVerticesReference(v.verticesReference, v.id);
					end
					else
						continue := false;
				end;
				otherwise
					continue := false;
			end;
		end;
		if vertices.numberOfElements = 2 then begin
			hennebergSequence := true;
			with infiniteVertex(vertices.getObject(vertices.ith(1))) do
				makeMRP(oldVertices, count, id, 0, 0);
			with infiniteVertex(vertices.getObject(vertices.ith(2))) do
				makeMRP(oldVertices, count, id, 0, 0);
		end
		else
			hennebergSequence := false;
		reset;
	end;

function infiniteUndirectedGraph.hennebergKReduction (k: seqIndex0): homogenVector;
	var
		continue: boolean;
		v: infiniteVertex;
		r: list;
		rv: homogenVector;
{$IFC AnimationEnabled AND AHennebergKReduction }
		pv: obj;

	procedure doDeleteEdge (x: infiniteEdge);
		begin
			if member(pv, point2d) then
				point2d(pv).drawLine(point2d(animationVertices.ith(x.value)))
			else if member(pv, point3d) then
				point3d(pv).drawLine(point3d(animationVertices.ith(x.value)));
		end;
{$ENDC }

	begin (* infiniteUndirectedGraph.hennebergKReduction *)
		continue := true;
		r := newEmptyList;
		while continue & (vertices.numberOfElements > 0) do begin
			v := infiniteVertex(verticesByDegree.getObject(verticesByDegree.extreme(left)));
			if v.neighbors.numberOfElements <= k then begin
{$IFC AnimationEnabled AND AHennebergKReduction }
				if (animationFlag[ATriangulationItem]) & (1 <= v.id) & (v.id <= animationVertices.length) then begin
					displayInfoString(StringOf('Vertex = ', v.id : 1, '. Degree = ', v.neighbors.numberOfElements : 1, '.'));
					pv := animationVertices.ith(v.id);
					pv.interactiveFlash;
					waitForClick(animationFlag[ATriangulationItem]);
					pv.interactiveOutput;
					v.neighbors.forAll(doDeleteEdge, left);
				end;
{$ENDC }
				r.appendInt(v.id);
				deleteVertexByVerticesReference(v.verticesReference, v.id);
			end
			else
				continue := false;
		end;
		new(rv);
		failNil(Handle(rv));
		rv.initFromList(r);
		r.free;
		rv.sortAll(intGreaterThan);
		hennebergKReduction := rv;
	end;

function getInitVertex (id: longInt): infiniteVertex;
	var
		v: infiniteVertex;
	begin
		if vertexList.length = 0 then begin
			new(v);
			failNil(Handle(v));
			v.init;
		end
		else begin
			v := infiniteVertex(vertexList.first^^.value);
			vertexList.pop;
			v.neighbors.init;
		end;
		v.verticesByDegreeReference := nil;
		v.id := id;
		getInitVertex := v;
	end;

procedure infiniteVertex.init;
	var
		d: avlTree;
	begin
		new(d);
		failNil(Handle(d));
		d.init;
		neighbors := d;
	end;

procedure doEdgeFree (e: infiniteEdge);
	begin
		e.free;
	end;

procedure infiniteVertex.free;
	begin
		neighbors.forAll(doEdgeFree, left);
		neighbors.reset;
		vertexList.append(self);
	end;

procedure infiniteVertex.destroy;
	begin
		neighbors.free;
		shallowFree;
	end;

procedure infiniteVertex.textOutput;
	begin
		write(' ', neighbors.numberOfElements : 1);
	end;

procedure infiniteUndirectedGraph.textOutput;
	var
		k: longInt;

	procedure outputVertex (v: infiniteVertex);

		procedure outputNeighbors (e: infiniteEdge);
			begin
				write(' [', lookup(e.value) : 1, ',', e.value : 1, ']');
			end;

		begin (* outputVertex *)
			write('[', k : 1, ',', v.id : 1, '] ->');
			v.neighbors.forAll(outputNeighbors, left);
			writeLn;
			k := k + 1;
		end;

	begin (* infiniteUndirectedGraph *)
		k := 1;
		vertices.forAll(outputVertex, left);
	end;

function getInitEdge (id: longInt): infiniteEdge;
	var
		e: infiniteEdge;
	begin
		if edgeList.length = 0 then begin
			new(e);
			failNil(Handle(e));
		end
		else begin
			e := infiniteEdge(edgeList.first^^.value);
			edgeList.pop;
		end;
		e.value := id;
		getInitEdge := e;
	end;

procedure infiniteEdge.destroy;
	begin
		shallowFree;
	end;

function compareVertices (a, b: infiniteVertex): relationObj;
	begin
		if a.id < b.id then
			compareVertices := lessThanObj
		else if a.id > b.id then
			compareVertices := greaterThanObj
		else
			compareVertices := equalObj;
	end;

function compareVerticesByDegree (a, b: infiniteVertex): relationObj;
	begin
		if a.neighbors.numberOfElements < b.neighbors.numberOfElements then
			compareVerticesByDegree := lessThanObj
		else if a.neighbors.numberOfElements > b.neighbors.numberOfElements then
			compareVerticesByDegree := greaterThanObj
		else
			compareVerticesByDegree := equalObj;
	end;

procedure infiniteEdge.free;
	begin
		edgeList.append(self);
	end;

procedure infiniteUndirectedGraph.init;
	var
		d: avlTree;
	begin
		new(d);
		failNil(Handle(d));
		d.init;
		vertices := d;
		verticesByDegree := avlTree(d.duplicate);
	end;

procedure infiniteUndirectedGraph.free;
	begin
		vertices.freeAll;
		verticesByDegree.free;
		shallowFree;
	end;

procedure doVertexFree (v: infiniteVertex);
	begin
		v.free;
	end;

procedure infiniteUndirectedGraph.reset;
	begin
		vertices.forAll(doVertexFree, right);
		vertices.reset;
		verticesByDegree.reset;
	end;

procedure infiniteUndirectedGraph.updateDegree (v: infiniteVertex);
	var
		found: boolean;
	begin
		if v.verticesByDegreeReference <> nil then
			verticesByDegree.delete(v.verticesByDegreeReference);
		v.verticesByDegreeReference := verticesByDegree.insertObj(v, compareVerticesByDegree, found);
	end;

procedure infiniteUndirectedGraph.addVertex (id: longInt);
	var
		v: infiniteVertex;
	begin
		v := insertVertex(id);
	end;

function infiniteUndirectedGraph.insertVertex (id: longInt): infiniteVertex;
	var
		where: reference;
		direction: directionType;
		found: boolean;
		v: infiniteVertex;
	begin
		globalV.id := id;
		found := vertices.find(globalV, compareVertices, where, direction);
		if found then
			insertVertex := infiniteVertex(vertices.getObject(where))
		else begin
			v := getInitVertex(id);
			v.verticesReference := vertices.insert(v, where, direction);
			updateDegree(v);
			insertVertex := v;
		end;
	end;

function infiniteUndirectedGraph.degree (id: longInt): longInt;
	var
		where: reference;
		direction: directionType;
		found: boolean;
	begin
		globalV.id := id;
		found := vertices.find(globalV, compareVertices, where, direction);
		if found then
			degree := infiniteVertex(vertices.getObject(where)).neighbors.numberOfElements
		else
			degree := -1;
	end;

function infiniteUndirectedGraph.neighbors (id, which: longInt): longInt;
	var
		where: reference;
		direction: directionType;
		found: boolean;
	begin
		neighbors := -1;
		if 1 <= which then begin
			globalV.id := id;
			found := vertices.find(globalV, compareVertices, where, direction);
			if found then
				with infiniteVertex(vertices.getObject(where)).neighbors do
					if which <= numberOfElements then
						neighbors := int(getObject(ith(which))).value;
		end;
	end;

procedure infiniteUndirectedGraph.associate (v, n: longInt);
	var
		where: reference;
		direction: directionType;
		found: boolean;
		u: infiniteVertex;
	begin
		u := insertVertex(v);
		globalE.value := n;
		found := u.neighbors.find(globalE, intCompare, where, direction);
		if not found then begin
			where := u.neighbors.insert(getInitEdge(n), where, direction);
			updateDegree(u);
		end;
	end;

procedure infiniteUndirectedGraph.addEdge (startVertex, endVertex: longInt);
	begin
		associate(startVertex, endVertex);
		associate(endVertex, startVertex);
	end;

procedure infiniteUndirectedGraph.disAssociate (v, n: longInt);
	var
		where: reference;
		direction: directionType;
		found: boolean;
		u: infiniteVertex;
	begin
		globalV.id := v;
		globalE.value := n;
		found := vertices.find(globalV, compareVertices, where, direction);
		if found then begin
			u := infiniteVertex(vertices.getObject(where));
			found := u.neighbors.find(globalE, intCompare, where, direction);
			if found then begin
				u.neighbors.getObject(where).free;
				u.neighbors.delete(where);
				updateDegree(u);
			end;
		end;
	end;

procedure infiniteUndirectedGraph.deleteEdge (startVertex, endVertex: longInt);
	begin
		disAssociate(startVertex, endVertex);
		disAssociate(endVertex, startVertex);
	end;

procedure infiniteUndirectedGraph.deleteVertex (id: longInt);
	var
		where: reference;
		direction: directionType;
		found: boolean;
	begin
		globalV.id := id;
		found := vertices.find(globalV, compareVertices, where, direction);
		if found then
			deleteVertexByVerticesReference(where, id);
	end;

procedure infiniteUndirectedGraph.deleteVertexByVerticesReference (where: reference; id: longInt);

	procedure doDisAssociate (e: infiniteEdge);
		begin
			disAssociate(e.value, id);
		end;

	begin (* infiniteUndirectedGraph.deleteVertexByVerticesReference *)
		with infiniteVertex(vertices.getObject(where)) do begin
			neighbors.forAll(doDisAssociate, left);
			verticesByDegree.delete(verticesByDegreeReference);
			free;
		end;
		vertices.delete(where);
	end;

function infiniteUndirectedGraph.isConnected (a, b: longInt): boolean;
	var
		where: reference;
		direction: directionType;
		found: boolean;
	begin
		globalV.id := a;
		found := vertices.find(globalV, compareVertices, where, direction);
		if found then begin
			globalE.value := b;
			isConnected := infiniteVertex(vertices.getObject(where)).neighbors.find(globalE, intCompare, where, direction);
		end
		else
			isConnected := false;
	end;

procedure infiniteUndirectedGraph.fromSimpleUndirectedGraph (g: simpleUndirectedGraph);
	var
		i: longInt;

	procedure doAddEdge (e: graphEdge);
		begin
			addEdge(e.startVertex, e.endVertex);
		end;

	begin (* infiniteUndirectedGraph.fromSimpleUndirectedGraph *)
		for i := 1 to g.vertices.length do
			addVertex(i);
		g.edges.forAll(doAddEdge);
	end;

procedure infiniteUndirectedGraph.fromSimpleUndirectedGraphSpecial (g: simpleUndirectedGraph; function edgeOk (ge: graphEdge): boolean; k: seqIndex0);
	var
		i, j, n: longInt;

	procedure doAddEdge (e: graphEdge);
		var
			i: longInt;
		begin
			addEdge(e.startVertex, e.endVertex);
			if not edgeOk(e) then
				for i := 1 to k + 1 do begin
					addEdge(e.startVertex, i + n);
					addEdge(e.endVertex, i + n);
				end;
		end;

	begin (* infiniteUndirectedGraph.fromSimpleUndirectedGraphSpecial *)
		n := g.vertices.length;
		for i := 1 to n do
			addVertex(i);
		for i := 1 to k + 2 do
			addVertex(n + i);
		g.edges.forAll(doAddEdge);
		for i := 1 to k + 1 do
			for j := i + 1 to k + 2 do
				addEdge(i + n, j + n);
	end;

function infiniteUndirectedGraph.countEdges: longInt;
	var
		count: longInt;

	procedure addNeighbors (v: infiniteVertex);
		begin
			count := count + v.neighbors.numberOfElements;
		end;

	begin (* infiniteUndirectedGraph.countEdges *)
		count := 0;
		vertices.forAll(addNeighbors, left);
		countEdges := count div 2;
	end;

function infiniteUndirectedGraph.countEdgesSpecial (function property (startVertex, endVertex: longInt): boolean): longInt;
	var
		count: longInt;

	procedure addNeighbors (v: infiniteVertex);

		procedure doCount (w: infiniteEdge);
			begin
				if (v.id < w.value) & property(v.id, w.value) then
					count := count + 1;
			end;

		begin (* addNeighbors *)
			v.neighbors.forAll(doCount, left);
		end;

	begin (* infiniteUndirectedGraph.countEdgesSpecial *)
		count := 0;
		vertices.forAll(addNeighbors, left);
		countEdgesSpecial := count;
	end;

function infiniteUndirectedGraph.lookup (id: longInt): seqIndex;
	var
		where: reference;
		direction: directionType;
		position: longInt;
	begin
		globalV.id := id;
		assert(vertices.findPosition(globalV, compareVertices, where, direction, position), 'Vertex exists');
		lookup := position;
	end;

function makeGraphEdge (startVertex, endVertex: seqIndex): graphEdge;
	var
		ge: graphEdge;
	begin
		new(ge);
		failNil(Handle(ge));
		ge.startVertex := startVertex;
		ge.endVertex := endVertex;
		makeGraphEdge := ge;
	end;

function infiniteUndirectedGraph.toUndirectedGraph (newVertices: vector; maxNumberOfVertices: longInt): undirectedGraph;
	var
		result: undirectedGraph;
		usedH: booleanArrayH;
		i, n, n1, k: seqIndex0;

	function goodEdge (s, e: longInt): boolean;
		begin
			goodEdge := (s <= n1) & (e <= n1);
		end;

	function goodEdgeLookup (s, e: longInt): boolean;
		begin
			goodEdgeLookup := (lookup(s) <= n1) & (lookup(e) <= n1);
		end;

	procedure doAddEdge (v: infiniteVertex);

		procedure doAddGraphEdge (e: infiniteEdge);
			begin
				if v.id < e.value then
					result.edges.append(makeGraphEdge(lookup(v.id), lookup(e.value)))
			end;

		begin (* doAddEdge *)
			v.neighbors.forAll(doAddGraphEdge, left);
		end;

	procedure doAddEdgeSpecial (v: infiniteVertex);

		procedure doAddGraphEdge (e: infiniteEdge);
			var
				startV, endV: longInt;
			begin
				if v.id < e.value then begin
					startV := lookup(v.id);
					endV := lookup(e.value);
					if goodEdge(startV, endV) then
						result.edges.append(makeGraphEdge(startV, endV));
				end;
			end;

		begin (* doAddEdgeSpecial *)
			v.neighbors.forAll(doAddGraphEdge, left);
		end;

	procedure doAddVertex (v: infiniteVertex);
		var
			nv: seqIndex;
		begin
			if result.vertices.length < n1 then begin
				if v.id <= n then begin
					assert(not usedH^^[v.id], StringOf(v.id : 1, ' used only once.'));
					nv := v.id;
				end
				else begin
					while usedH^^[k] do
						k := k + 1;
					nv := k;
				end;
				result.vertices.append(newVertices.ith(nv).duplicateAll);
				usedH^^[nv] := true;
			end;
		end;

	begin (* infiniteUndirectedGraph.toUndirectedGraph *)
		n := newVertices.length;
		if maxNumberOfVertices < vertices.numberOfElements then
			n1 := maxNumberOfVertices
		else
			n1 := vertices.numberOfElements;
		assert(n1 <= n, 'Enough vertices supplied');
		usedH := booleanArrayH(dynamicArray(n, sizeOf(boolean)));
		for i := 1 to n do
			usedH^^[i] := false;
		new(result);
		failNil(Handle(result));
		result.vertices := newVector(n1);
		k := 1;
		vertices.forAll(doAddVertex, left);
{$IFC THINK_PASCAL }
		disposHandle(Handle(usedH));
{$ELSEC }
		DisposeHandle(Handle(usedH));
{$ENDC }
		if n1 < vertices.numberOfElements then begin
			result.edges := newHomogenVector(countEdgesSpecial(goodEdgeLookup));
			vertices.forAll(doAddEdgeSpecial, left);
		end
		else begin
			result.edges := newHomogenVector(countEdges);
			vertices.forAll(doAddEdge, left);
		end;
		result.adjacency := newHomogenVector(n1);
		result.initAdjacency;
		toUndirectedGraph := result;
	end;

procedure initFastUndirectedGraph (vertices, edges: longInt);
	var
		v: infiniteVertex;
		i: longInt;
		e: infiniteEdge;
	begin
		vertexList := newEmptyList;
		for i := 1 to vertices do begin
			new(v);
			failNil(Handle(v));
			v.init;
			vertexList.append(v);
		end;
		edgeList := newEmptyList;
		for i := 1 to 2 * edges do begin
			new(e);
			failNil(Handle(e));
			edgeList.append(e);
		end;
		globalV := getInitVertex(0);
		globalE := getInitEdge(0);
	end;

procedure doVertexDestroy (v: infiniteVertex);
	begin
		v.destroy;
	end;

procedure doEdgeDestroy (e: infiniteEdge);
	begin
		e.destroy;
	end;

procedure exitFastUndirectedGraph;
	begin
		globalV.free;
		globalE.free;
		vertexList.forAll(doVertexDestroy);
		vertexList.free;
		edgeList.forAll(doEdgeDestroy);
		edgeList.free;
	end;

{$IFC AnimationEnabled AND AHennebergKReduction }
procedure prepareAnimation (g: simpleUndirectedGraph);
	begin
		animationVertices := g.vertices;
		if animationFlag[ATriangulationItem] then
			g.interactiveHighlight;
	end;
{$ENDC }

function computeHennebergKReductionSpecial (g: simpleUndirectedGraph; k: seqIndex0; function edgeOk (ge: graphEdge): boolean): homogenVector;
	var
		f: infiniteUndirectedGraph;
	begin
		new(f);
		failNil(Handle(f));
		f.init;
		f.fromSimpleUndirectedGraphSpecial(g, edgeOk, k);
{$IFC AnimationEnabled AND AHennebergKReduction }
		prepareAnimation(g);
{$ENDC }
		computeHennebergKReductionSpecial := f.hennebergKReduction(k);
		f.free;
	end;

function computeHennebergKReduction (g: simpleUndirectedGraph; k: seqIndex0): undirectedGraph;
	var
		f: infiniteUndirectedGraph;
		r: homogenVector;
	begin
		new(f);
		failNil(Handle(f));
		f.init;
		f.fromSimpleUndirectedGraph(g);
{$IFC AnimationEnabled AND AHennebergKReduction }
		prepareAnimation(g);
{$ENDC }
		r := f.hennebergKReduction(k);
		r.freeAll;
		computeHennebergKReduction := f.toUndirectedGraph(g.vertices, g.vertices.length);
		f.free;
	end;

function findHennebergSequenceMaxTrials (g: simpleUndirectedGraph; maxTrials: longInt): homogenVector;
	var
		f: infiniteUndirectedGraph;
		result: homogenVector;
		reduceThreeVertex: seqIndex0;
		trials: longInt;
		i: seqIndex0;
		cx, cy: extended;
		done: boolean;
	begin
		new(f);
		failNil(Handle(f));
		f.init;
		new(result);
		failNil(Handle(result));
		result.elements := nil;
		result.preset(g.vertices.length, markedRealPointType);
		for i := 1 to g.vertices.length do begin
			g.vertices.ith(i).centerOfGravity(cx, cy);
			markedRealPoint(result.ith(i)).setXY(cx, cy);
		end;

		f.fromSimpleUndirectedGraph(g);
		done := f.hennebergSequence(result, reduceThreeVertex);
		trials := 1;
		while (trials <= maxTrials) & not done do begin
{$IFC AnimationEnabled }
			if userAbort then
				leave;
{$ENDC }
			f.fromSimpleUndirectedGraph(g);
			done := f.hennebergSequence(result, reduceThreeVertex);
			trials := trials + 1;
		end;
		f.free;
		if done then begin
			displayInfoString(StringOf('Degree 3 Reductions = ', reduceThreeVertex : 1, ' Trials = ', trials : 1));
			findHennebergSequenceMaxTrials := result;
		end
		else begin
			result.freeAll;
			findHennebergSequenceMaxTrials := nil;
		end;
	end;

end.
