(*	XYZ GeoBench (eXperimental geometrY Zurich):		*)
(*	An object oriented, interactive workbench for		*)
(*	experimental geometric computation.					*)
(*														*)
(*	 1999 by Peter Schorn, peter.schorn@acm.org		*)
(*	Culmannstrasse 77, CH-8006 Zrich, Switzerland		*)
(*	This software may be copied freely for educational	*)
(*	and research purposes only. For commercial use,		*)
(*	please contact the author.							*)
(*														*)
(*	Version 5.0.5: 30-Sep-1999							*)

(*	By Ernst Hinz, Mnster.								*)

unit kBoxTree; (* Stand : 01.02.1993 *)

(* Datenstruktur : BinarykBoxTree :                                         *)
(* Der " BinarykBoxTree " ist ein Binaerbaum. In jedem " BinarykBoxNode "   *)
(* Knoten werden folgende Objekte gespeichert :                             *)
(*  Left      : Zeiger auf den linken Sohn.                                 *)
(*  Right     : Zeiger auf den rechten Sohn.                                *)
(*  Box       : Enthaelt die Grenzen eines Hyperrechtecks.                  *)
(*  Line      : Teilt obiges Hyperrechteck in zwei disjunkte Teilrechtecke, *)
(*              wenn der Knoten kein Blatt ist.                             *)
(*  ListElt   : Enthaelt eine Liste von Punkten, die in obigen Hyperrecht-  *)
(*              eck liegen, wenn der Knoten ein Blatt ist.                  *)
(*  Internal  : Legt fest, ob der Knoten ein Innerer oder ein Blatt ist.    *)
(*  Complete  : Ist True, falls Anzahl Punkte in ListElt groesser als       *)
(*              MaxPoints ist.                                              *)
(* Der BinarykBoxTree liefert also eine hierarchische Unterteilung des      *)
(* k - dim. Raumes. Sowohl die Wurzel als auch die Vereinigung der          *)
(* Blaetter enthalten jeweils den gesamten Raum. Der Baum kann zu einer     *)
(* linearen Liste degenerieren.                                             *)

(* Datenstruktur : BetakBoxTree :                                           *)
(* Der " BetakBoxTree " ist ein dreigliedriger Baum. In jedem "BetakBoxNode *)
(* Knoten werden ausser den in dem BinarykBoxNode schon vorhandenen Objekten *)
(* folgende zusaetzliche Objekte gespeichert :                              *)
(*  UP         : Zeiger auf einen Vorgaenger im BinarykBoxTree.             *)
(*  Component  : Summe aller Knoten des Teilbaumes, desses Wurzel dieser    *)
(*               Knoten ist.                                                *)
(*  TwinNode   : Zeiger auf den zugehoerigen Binaerknoten (Zwilling).       *)
(* Der BetakBoxTree liefert also eine hierarchische Unterteilung des        *)
(* k - dim. Raumes. Die Vereinigung der Blaetter enthalten jeweils den      *)
(* gesamten Raum. Durch balancieren wird eine logarithmische Laufzeit       *)
(* fuer das Einfuegen und Suchen erreicht.                                  *)
(* Zu der Datenstruktur BetakBoxTree gehoert das Objekt " BinNode",welches  *)
(* einen Binaerbaum bildet. Dieser Baum wird benoetigt, um den BetakBoxTree *)
(* balancieren zu koennen. In BinNode sind gespeichert :                    *)
(* Left        : Zeiger auf den linken Sohn.                                *)
(* Right       : Zeiger auf den linken Sohn.                                *)
(* Parent      : Zeiger auf den Vater.                                      *)
(* BetaTwinNode: Zeiger auf den zugehoerigen BetakBoxNode (Zwilling).       *)
(* Mark     : True false markiert, sonst false ( Wichtig wegen balancieren). *)
(* Index    : Hoehe des Knoten im BetakBoxTree ( Wichtig wegen balancieren). *)
(* Lab      : Labelvektor ( Wichtig wegen balancieren).                     *)
(*                                                                          *)
(*                            Ebenen der Objekte                            *)
(*         BinarykBoxTree(100)  c  BetakBoxTree(200)                        *)
(*                  I                     I                                 *)
(*         BinarykBoxNode(110)  c  BetakBoxNode(210) <---> BinNode(310)     *)
(*        /         I         \                                             *)
(* Box(111)     Line(112)     ListElt(113)                                  *)

interface

uses
	basicGeometry;

const
	RootBoxMin = 0;      (* Minimale i.- Koordinate *)
	RootBoxMax = 10000;  (* Maximale i.- Koordinate *)
	LabelMax = 15;       (* Laenge des Label - Vektor fuer BinNode *)
type
	NextNode = (NodeLeft, NodeRight, NodeUp, NodeNil);
	LabelVector = array[0..LabelMax] of integer; (* Label - Vektor BinNode *)

	BoxObj = object(obj)                                   (* Objekt - Nummer 111 *)
			TopLeft, BottomRight: dDimPoint;
			function Include (p: dDimPoint): boolean;
(* 111.01: Ueberprueft, ob der Punkt p innerhalb des Rechteck (TopLeft,     *)
(*         BottomRight) liegt oder nicht.                                   *)
		end;

	Line = object(obj)                                  (* Objekt - Nummer 112 *)
			FirstPoint, LastPoint: dDimPoint;
			procedure Create;
(* 112.01: Stellt Speicherplatz fuer die beiden Punkte zur Verfuegung.      *)
			procedure freeAll;
			override;
(* 112.02: Gibt Speicherplatz frei.                                         *)
			procedure MakeDividingPoints (TL, BR: dDimPoint; i: integer; Limes: extended);
(* 112.03: Weist die Parameter auf FirstPoint und LastPoint zu.             *)
			function FindNextNode (p: dDimPoint): NextNode;
(* 112.04: Findet den Sohn, in dessen Box der Punkt p liegt. Nur anwenden,  *)
(*         wenn BoxObj.InClude(p) (111.01) true ist.                           *)
		end;

	ListElt = object(obj)                               (* Objekt - Nummer 113 *)
			Next: ListElt;
			Element: dDimPoint;
			procedure freeAll;
			override;
(* 113.01: Gibt den Speicherplatz der Objekte ListElt frei bzgl eines Knoten *)
			procedure InitList (p: dDimPoint);
(* 113.02: Initialisiert das erste Listenelement mit p.                     *)
			function FindMax (i: integer): extended;
(* 113.03: Findet den maximalen Wert einer Liste von Punkten bzgl. der      *)
(*         Koordinate i.                                                    *)
			function FindMin (i: integer): extended;
(* 113.04: Findet den minimalen Wert einer Liste von Punkten bzgl. der      *)
(*         Koordinate i.                                                    *)
		end;

	BinarykBoxNode = object(obj)                        (* Objekt - Nummer 110 *)
			Left, Right: BinarykBoxNode;
			kBox: BoxObj;
			DividingLine: Line;
			Head: ListElt;
			Internal: boolean;
			Complete: boolean;
			procedure freeAll;
			override;
(* 110.01: Gibt den Speicherplatz aller Objekte BinarykBoxNode frei, die    *)
(*         Nachkommen von self sind, einschliesslich aller UnterObjekte.    *)
			procedure FreeNode;
(* 110.02: Gibt den Speicherplatz des Objekt BinarykBoxNode frei,           *)
(*         einschliesslich aller UnterObjekte.                              *)
			procedure InitNode (var TL, BR: dDimPoint);
(* 110.03: Initialisiert den Knoten (Einschliesslich Grenzen des Rechteck). *)
			procedure Insert (p: dDimPoint);
(* 110.04: Fuegt einen Punkt in den BinarykBoxNode ein.                     *)
			procedure InsertTwoLeaves;
(* 110.05: Erzeugt zwei Soehne und markiert den Knoten als inneren Knoten.  *)
			procedure MakeDividingLine;
(* 110.06: Erzeugt eine Trennlinie fuer die Box, wenn mehr als MaxPoints in *)
(*         ListElt gespeichert sind.                                        *)
			procedure MovePoints;
(* 110.07: Bewegt die Punkte aus dem Vaterknoten zu den Soehnen,            *)
(*         entsprechend ihrer Lage in den Boxen.                            *)
			procedure Split;
(* 110.08: Berechnet die neue DividingLine (110.06),erzeugt zwei neue Knoten *)
(*         (110.05) und bewegt die Punkte von self zu den Soehnen(110.07).  *)
			function Search (p: dDimPoint): BinarykBoxNode;
(* 110.09: Liefert als Ergebnis einen Zeiger auf das Blatt, dessen Box den  *)
(*         Punkt p enthaelt.                                                *)
		end;

	BinarykBoxTree = object(obj)                        (* Objekt - Nummer 100 *)
			Root: BinarykBoxNode;
			procedure freeAll;
			override;
(* 100.01: Gibt den Speicherplatz des Objekt BinarykBoxTree frei,           *)
(*         einschliesslich aller UnterObjekte.                              *)
			procedure init;
			override;
(* 100.02: Initialisiert den BinarykBoxTree.                                *)
			procedure Insert (p: dDimPoint);
(* 100.03: Fuegt einen Punkt in den BinarykBoxTree ein und erzeugt eventuell *)
(*         zwei neue Knoten. Delta ist dabei die min. Laenge der neuen Box. *)
			procedure MakeRoot;
(* 100.04: Legt Speicherplatz fuer die Wurzel an und initialisiert die max. *)
(*         Grenzen der Root k-Box.                                          *)
			function Search (p: dDimPoint): BinarykBoxNode;
(* 100.05: Liefert als Ergebnis einen Zeiger auf das Blatt, dessen Box den  *)
(*         Punkt p enthaelt. Verwendet dabei BinarykBoxNode.Search(110.09). *)
		end;

	BetakBoxNode = object(BinarykBoxNode)               (* Objekt - Nummer 210 *)
			Up: BetakBoxNode;
			TwinNode: BinNode;
			Component: integer;
			procedure Balance (var Node: BetakBoxNode);
(* 210.01: Balanciert den BetakBoxTree ab den Knoten self neu aus und haengt *)
(*         den balancierten Teilbaum an Node wieder an.                     *)
			procedure Balance2 (var Node: BetakBoxNode; NodeType: NextNode);
(* 210.02: Wie 210.01 (Anderer Balancierungs - Algorithmus)                     *)
			procedure CentroidTest (Son: BetakBoxNode; var Node: BetakBoxNode);
(* 210.03: Ueberprueft, ob self ein Beta - Centroid gegenueber Son ist.     *)
(*         Falls self kein Beta - Centroid ist und Node = nil, dann wird    *)
(*         CentroidTest := self.                                            *)
			procedure ComputeNewComponents;
(* 210.04: Berechnet fuer jeden Knoten des Teilbaum (Wurzel ist self) die   *)
(*         neue Componentenzahl.                                            *)
			procedure freeAll;
			override;
(* 210.05: Gibt den Speicherplatz aller Objekte BetakBoxNode frei, die Nach- *)
(*         kommen von self sind, einschliesslich aller UnterObjekte.        *)
			procedure FreeNode;
			override;
(* 210.06: Gibt den Speicherplatz des Objekt BetakBoxNode frei,             *)
(*         einschliesslich aller UnterObjekte.                              *)
			procedure IncComponents (p: dDimPoint; var Node: BetakBoxNode);
(* 210.07: Erhoeht die Komponentenzahl fuer all die inneren Knoten, die auf *)
(*         der Suche nach dem Blatt, das den Punkt p enthaelt, durchlaufen  *)
(*         werden um zwei. Wird dabei ein Centroid gefunden, der kein Beta - *)
(*         Centroid mehr ist, so wird dieser ausgegeben (immer der erste),  *)
(*         sonst Node := nil.                                               *)
			procedure InitNode (var TL, BR: dDimPoint);
			override;
(* 210.08: Initialisiert den Knoten (Einschliesslich Grenzen des Rechteck). *)
			procedure InsertTwoLeaves;
			override;
(* 210.09: Erzeugt zwei Soehne und markiert den Knoten als inneren Knoten.  *)
			procedure MarkBinNode;
(* 210.10: Markiert von allen Nachkommen von self die Zwillingsknoten des   *)
(*         parallelen Binaerbaums.                                          *)
			procedure MovePoints;
			override;
(* 210.11: Bewegt die Punkte aus dem Vaterknoten zu den Soehnen,            *)
(*         entsprechend ihrer Lage in den Boxen                             *)
			procedure Split;
			override;
(* 210.12: Berechnet die neue DividingLine (110.06),erzeugt zwei neue Knoten *)
(*         (210.09) und bewegt die Punkte von self zu den Soehnen (210.11). *)
			function FindMaxComponent: BetakBoxNode;
(* 210.13: Findet den Sohn von self mit der groessten Komponentenzahl.      *)
			function FindNextNode (p: dDimPoint): BetakBoxNode;
(* 210.14: Findet den Nachfolgerknoten im BetakBoxTree, der auf dem Wurzel  *)
(*         Blatt Weg von p liegt.                                           *)
			function FindNewRoot (HalfComponent: integer): BetakBoxNode;
(* 210.15: Findet den Nachkommen von self, dessen Soehne Komponenten haben, *)
(*         die kleiner oder gleich der Komponente von self sind.            *)
			function PointerToFather (var Node, Father: BetakBoxNode; p: dDimPoint): BetakBoxNode;
(* 210.16: Liefert als Ergebnis einen Zeiger auf den Knoten, dessen Sohn    *)
(*         kein Beta - Centroid ist.                                        *)
			function PointerToTwin (Node: NextNode): BinNode;
(* 210.17: Liefert einen Zeiger fuer den Sohn Node auf den Zwillingsknoten  *)
(*         im Binaerbaum.                                                   *)
			function Search (p: dDimPoint): BinarykBoxNode;
			override;
(* 210.18: Liefert als Ergebnis einen Zeiger auf das Blatt, dessen Box den  *)
(*         Punkt p enthaelt.                                                *)
		end;

	BetakBoxTree = object(BinarykBoxTree)               (* Objekt - Nummer 200 *)
			procedure freeAll;
			override;
(* 200.01 : Gibt den Speicherplatz des Objekt BetakBoxTree frei,            *)
(*          einschliesslich aller UnterObjekte.                             *)
			procedure init;
			override;
(* 200.02 : Initialisiert den BetakBoxTree                                  *)
			procedure Insert (p: dDimPoint);
			override;
(* 200.03 : Fuegt einen Punkt in den BetakBoxTree ein und erzeugt eventuell *)
(*          zwei neue Knoten. Delta ist dabei die min. Laenge einer k - Box. *)
(*          Entsteht ein NoCentroid, so wird der Teilbaum balanciert.       *)
			procedure Insert2 (p: dDimPoint);
(* 200.04: Wie 200.03. Aber anderer Balancier - Algorithmus.                *)
			procedure MakeRoot;
			override;
(* 200.05 : Legt Speicherplatz fuer die Wurzel an und initialisiert die max. *)
(*          Grenzen der Root k-Box.                                         *)
			function Search (p: dDimPoint): BinarykBoxNode;
			override;
(* 200.06 : Liefert als Ergebnis einen Zeiger auf das Blatt, dessen Box den *)
(*          Punkt p enthaelt. Verwendet dabei BetakBoxNode.Search.(210.18)  *)
		end;

	BinNode = object(obj)                               (* Objekt - Nummer 310 *)
			Left, Right, Parent: BinNode;
			BetaTwinNode: BetakBoxNode;
			Mark: boolean;
			Index: integer;
			Lab: LabelVector;
			procedure Balance (NodeType: NextNode);
(* 310.01: Berechnet fuer den Teilbaum  von NodeType Labels und Indices.Dann *)
(*         balanciert sie den Teilbaum mittels dieser Labels und Indices aus *)
(*         und haengt die Zeiger im BetakBoxTree entsprechend um.           *)
			procedure Balance2 (var Node: BetakBoxNode; NodeType: NextNode);
(* 310.02: Berechnet fuer den Teilbaum self eine 1/2 Zerlegung.             *)
			procedure ChangeNextPointer (NodeType: NextNode; var Node: BinNode);
(* 310.03: Findet den naechsten Knoten, der im BetakBoxTree an den Zwillings *)
(*         Knoten von Node angehaengt werden muss. Ruft dann ChangePointers *)
(*         auf.                                                             *)
			procedure ChangePointers (NodeType: NextNode; Node: BinNode);
(* 310.04: Haengt Zwillingsknoten von self an den Zwillingsknoten von Node. *)
(*         Ruft dann evtl. ChangeNextPointer auf.                           *)
			procedure ComputeLabel;
(* 310.05: Berechnet Label des Knoten self.                                 *)
			procedure ComputeWeight;
(* 310.06: Berechnet Gewichte der Knoten des Binaerbaum mit Wurzel self.    *)
			procedure FindHighestIndex (var HighestNode: BinNode; var GreatestIndex: integer);
(* 310.07: Findet groessten Index in Teilbaum von self. HighestNode zeigt   *)
(*         auf den zugehoerigen Knoten.                                     *)
			procedure InitBinNode (Node: BetakBoxNode);
(* 310.08: Initialisiert den BinNode (einschliesslich Verbindung mit        *)
(*         Zwillingsknoten " BetakBoxNode ".                                *)
			procedure Insert;
(* 310.09: Fuegt Zeiger zu den zwei Soehnen und dem Vater ein.              *)
			procedure MakeIndex;
(* 310.10: Weist self.Index einen Wert zu, der durch ComputeIndex (310.12)  *)
(*         berechnet wird.                                                  *)
			procedure MakeLabels;
(* 310.11: Berechnet Labels und Indices fuer alle markierten Knoten.        *)
			function ComputeIndex (Value: integer): integer;
(* 310.12: Berechnet den Index von self.                                    *)
			function FindBinRoot: BinNode;
(* 310.13: Findet den hoechsten markierten Knoten von self aus.             *)
			function FindHalfNode (Weight: integer): BinNode;
(* 310.14: Findet den Knoten im Baum mit Gewicht >= Weight.                 *)
			function FindNextNode: BinNode;
(* 310.15: Findet den Nachfolgerknoten mit hoechstem Gewicht. Ist nil, falls *)
(*         keiner vorhanden oder markiert ist.                              *)
		end;

var
	RootTopLeft, RootBottomRight: dDimPoint;(* maximale Eckpunkte von RootBox *)
	k: integer; (* Dimension von dDimPoint *)
function Interval (a, b, x: real): boolean;
(* 111.01.1 : Ueberprueft, ob gilt a <= x < b                               *)

function PowerOf (Value: integer): real;
(* 110.04.2 : Berechnet (2* Value + 2) Hoch Value                           *)

implementation

(************************  OBJECT Box 111 ********************************** *)
(* 111.01.1 **************************************************************** *)
function Interval (a, b, x: real): boolean;
	begin
		if (a <= x) and (x < b) then
			Interval := true
		else
			Interval := false
	end;

(* 111.01 ****************************************************************** *)
function BoxObj.Include (p: dDimPoint): boolean;
	var
		i: dimensionRange;
		Inside: boolean;
	begin
		i := 1;
		Inside := true;
		repeat
			Inside := Interval(TopLeft.coordinates[i], BottomRight.coordinates[i], p.coordinates[i]);(* 111.01.1 *)
			i := i + 1;
		until (i > k) or not Inside;
		Include := Inside
	end;

(************************  OBJECT Line 112 ********************************* *)
(* 112.01 ****************************************************************** *)
procedure Line.Create;
	begin
		new(FirstPoint);
		failNil(Handle(FirstPoint));
		new(LastPoint);
		failNil(Handle(LastPoint));
	end;

(* 112.02 ****************************************************************** *)
procedure Line.freeAll;
	begin
		FirstPoint.free;
		LastPoint.free;
		self.free
	end;

(* 112.03 ****************************************************************** *)
procedure Line.MakeDividingPoints (TL, BR: dDimPoint; i: integer; Limes: extended);
	var
		j: integer;
		p1, p2: realPoint;
	begin
		for j := 1 to k do
			if i = j then begin
				FirstPoint.coordinates[j] := Limes;
				LastPoint.coordinates[j] := Limes;
			end
			else begin
				FirstPoint.coordinates[j] := TL.coordinates[j];
				LastPoint.coordinates[j] := BR.coordinates[j];
			end;
		FirstPoint.dimension := k;
		LastPoint.dimension := k;
		if animationFlag[AClosestPairItem] then begin
			p1 := FirstPoint.toRealPoint;
			p2 := LastPoint.toRealPoint;
			p1.drawLine(p2);
			p1.free;
			p2.free
		end;
	end;

(* 112.04 ****************************************************************** *)
function Line.FindNextNode (p: dDimPoint): NextNode;
	var
		i: integer;
	begin
		i := 0;
		repeat
			i := i + 1;
		until (FirstPoint.coordinates[i] = LastPoint.coordinates[i]) or (i = k);
		if p.coordinates[i] < FirstPoint.coordinates[i] then
			FindNextNode := NodeLeft
		else
			FindNextNode := NodeRight
	end;

(************************  OBJECT ListElt 113 ****************************** *)
(* 113.01 ****************************************************************** *)
procedure ListElt.freeAll;
	begin
		if Next <> nil then
			Next.freeAll;
		self.free                                                          (* obj *)
	end;

(* 113.02 ****************************************************************** *)
procedure ListElt.InitList (p: dDimPoint);
	begin
		Next := nil;
		Element := p
	end;

(* 113.03 ****************************************************************** *)
function ListElt.FindMax (i: integer): extended;
	var
		Help: ListElt;
		Value, Maximum: real;
	begin
		Help := self;
		Maximum := Help.Element.coordinates[i];
		while Help <> nil do begin
			Value := Help.Element.coordinates[i];
			if Value > Maximum then
				Maximum := Value;
			Help := Help.Next;
		end;
		FindMax := Maximum
	end;

(* 113.04 ****************************************************************** *)
function ListElt.FindMin (i: integer): extended;
	var
		Help: ListElt;
		Value, Minimum: real;
	begin
		Help := self;
		Minimum := Help.Element.coordinates[i];
		while Help <> nil do begin
			Value := Help.Element.coordinates[i];
			if Value < Minimum then
				Minimum := Value;
			Help := Help.Next;
		end;
		FindMin := Minimum
	end;

(************************  OBJECT BinarykBoxNode   110 ********************* *)
(* 110.01 ****************************************************************** *)
procedure BinarykBoxNode.freeAll;
	begin
		if Left <> nil then
			Left.freeAll;                                                  (* 110.01 *)
		if Right <> nil then
			Right.freeAll;                                                 (* 110.01 *)
		self.FreeNode;                                                  (* 110.02 *)
	end;

(* 110.02 ****************************************************************** *)
procedure BinarykBoxNode.FreeNode;
	begin
		kBox.free;                                                         (* obj *)
		if Internal then
			DividingLine.freeAll                                            (* 112.02 *)
		else
			Head.freeAll;                                                   (* 113.01 *)
		self.free;                                                          (* obj *)
	end;

(* 110.03 ****************************************************************** *)
procedure BinarykBoxNode.InitNode (var TL, BR: dDimPoint);
	begin
		Left := nil;
		Right := nil;
		Head := nil;
		new(kBox);
		failNil(Handle(kBox));
		kBox.TopLeft := TL;
		kBox.BottomRight := BR;
		Internal := false;
		Complete := false;
	end;

(* 110.04.1 **************************************************************** *)
procedure PointInsert (var m: ListElt; n: ListElt; var Counter: real);
	begin
		if m = nil then begin
			m := n;
			Counter := Counter + 1
		end
		else begin
			PointInsert(m.Next, n, Counter);                             (* 110.04.1 *)
			Counter := Counter + 1
		end;
	end;

(* 110.04.2 **************************************************************** *)
function PowerOf (Value: integer): real;
	var
		i, Base: integer;
		Res: real;
	begin
		Base := 2 * Value + 2;
		Res := 1;
		if Value > 0 then
			for i := 1 to Value do
				Res := Res * Base;
		PowerOf := Res
	end;

(* 110.04 ****************************************************************** *)
procedure BinarykBoxNode.Insert (p: dDimPoint);
	var
		Point: ListElt;
		Counter: real;
	begin
		Counter := 0;
		new(Point);
		failNil(Handle(Point));
		Point.InitList(p);                                              (* 113.02 *)
		PointInsert(Head, Point, Counter);                            (* 110.04.1 *)
		if Counter > PowerOf(k) then                               (* 110.04.2 *)
			Complete := true;
	end;

(* 110.05 ****************************************************************** *)
procedure BinarykBoxNode.InsertTwoLeaves;
	var
		LeftNode, RightNode: BinarykBoxNode;
	begin
		new(LeftNode);
		failNil(Handle(LeftNode));
		new(RightNode);
		failNil(Handle(RightNode));
		LeftNode.InitNode(kBox.TopLeft, DividingLine.LastPoint);        (* 110.03 *)
		RightNode.InitNode(DividingLine.FirstPoint, kBox.BottomRight);  (* 110.03 *)
		Internal := true;
		Left := LeftNode;
		Right := RightNode;
	end;

(* 110.06 ****************************************************************** *)
procedure BinarykBoxNode.MakeDividingLine;
	var
		i, coord: integer;
		Minimum, Maximum, Min, Lim, Limes: extended;
	begin
		new(DividingLine);
		failNil(Handle(DividingLine));
		DividingLine.Create;                                            (* 112.01 *)
		i := 0;
		Limes := 0;
		repeat
			i := i + 1;
			Maximum := Head.FindMax(i);                                    (* 113.03 *)
			Minimum := Head.FindMin(i);                                    (* 113.04 *)
			Lim := Maximum - Minimum;
			if Lim > Limes then begin
				Min := Minimum;
				Limes := Lim;
				coord := i
			end;
		until i = k;
		Limes := Min + Limes / 2;
		DividingLine.MakeDividingPoints(kBox.TopLeft, kBox.BottomRight, coord, Limes);(* 112.03 *)
	end;

(* 110.07 ****************************************************************** *)
procedure BinarykBoxNode.MovePoints;
	var
		HelpPtr: ListElt;
	begin
		HelpPtr := Head;
		while HelpPtr <> nil do begin
			case DividingLine.FindNextNode(HelpPtr.Element) of             (* 112.04 *)
				NodeLeft: 
					Left.Insert(HelpPtr.Element);                                (* 110.04 *)
				NodeRight: 
					Right.Insert(HelpPtr.Element);                               (* 110.04 *)
			end;
			HelpPtr := HelpPtr.Next
		end;
		if Left.Complete or Right.Complete then
			unimplementedMethod('to many equal points');
		Head.freeAll;                                                   (* 113.01 *)
	end;

(* 110.08 ****************************************************************** *)
procedure BinarykBoxNode.Split;
	begin
		self.MakeDividingLine;                                   (* 110.06 *)
		self.InsertTwoLeaves;                                           (* 110.05 *)
		self.MovePoints;                                                (* 110.07 *)
	end;

(* 110.09 ****************************************************************** *)
function BinarykBoxNode.Search (p: dDimPoint): BinarykBoxNode;
	begin
		if Internal then
			case DividingLine.FindNextNode(p) of                           (* 112.04 *)
				NodeLeft: 
					Search := Left.Search(p);                                    (* 110.09 *)
				NodeRight: 
					Search := Right.Search(p)                                    (* 110.09 *)
			end
		else
			Search := self
	end;

(************************  OBJECT BinarykBoxTree 100 *********************** *)
(* 100.01 ****************************************************************** *)
procedure BinarykBoxTree.freeAll;
	begin
		Root.freeAll;                                                   (* 110.01 *)
		RootTopLeft.free;                                                  (* obj *)
		RootBottomRight.free;                                              (* obj *)
		self.free                                                          (* obj *)
	end;

(* 100.02 ****************************************************************** *)
procedure BinarykBoxTree.init;
	begin
		Root := nil;
	end;

(* 100.03 ****************************************************************** *)
procedure BinarykBoxTree.Insert (p: dDimPoint);
	var
		SearchedNode: BinarykBoxNode;
	begin
		SearchedNode := self.Search(p);                                 (* 100.05 *)
		if SearchedNode = nil then begin
			self.MakeRoot;                                                 (* 100.04 *)
			Root.InitNode(RootTopLeft, RootBottomRight);                   (* 110.03 *)
			Root.Insert(p)                                                 (* 110.04 *)
		end
		else begin
			SearchedNode.Insert(p);                                        (* 110.04 *)
			if SearchedNode.Complete then
				SearchedNode.Split;                                    (* 110.08 *)
		end;
	end;

(* 100.04 ****************************************************************** *)
procedure BinarykBoxTree.MakeRoot;
	var
		i: integer;
	begin
		new(Root);
		failNil(Handle(Root));
		new(RootTopLeft);
		failNil(Handle(RootTopLeft));
		new(RootBottomRight);
		failNil(Handle(RootBottomRight));
		RootTopLeft.dimension := k;
		RootBottomRight.dimension := k;
		for i := 1 to k do begin
			RootTopLeft.coordinates[i] := RootBoxMin;
			RootBottomRight.coordinates[i] := RootBoxMax;
		end;
	end;

(* 100.05 ****************************************************************** *)
function BinarykBoxTree.Search (p: dDimPoint): BinarykBoxNode;
	begin
		if Root = nil then
			Search := nil
		else
			Search := Root.Search(p);                                      (* 110.09 *)
	end;

(************************  OBJECT BetakBoxNode   210 *********************** *)
(* 210.01 ****************************************************************** *)
procedure BetakBoxNode.Balance (var Node: BetakBoxNode);
	var
		HalfComponent: integer;
		NewRoot: BetakBoxNode;
		NodeType: NextNode;
	begin
{$IFC THINK_PASCAL }
		HalfComponent := trunc(0.5 * Component);
{$ELSEC }
		HalfComponent := truncLong(0.5 * Component);
{$ENDC }
		NewRoot := self.FindNewRoot(HalfComponent);                     (* 210.15 *)
		self.MarkBinNode;                                               (* 210.10 *)
		for NodeType := NodeLeft to NodeUp do begin
			NewRoot.TwinNode.Balance(NodeType);                            (* 310.01 *)
			NewRoot.TwinNode.Mark := true
		end;
		NewRoot.TwinNode.Mark := false;
		if Node <> nil then begin
			if BetakBoxNode(Node.Left) = self then
				BetakBoxNode(Node.Left) := NewRoot
			else if BetakBoxNode(Node.Right) = self then
				BetakBoxNode(Node.Right) := NewRoot
			else
				Node.Up := NewRoot;
		end
		else
			Node := NewRoot;
		NewRoot.ComputeNewComponents;                                   (* 210.04 *)
	end;

(* 210.02 ****************************************************************** *)
procedure BetakBoxNode.Balance2 (var Node: BetakBoxNode; NodeType: NextNode);
	var
		BinRoot: BinNode;
		HelpNode: BetakBoxNode;
	begin
		self.MarkBinNode;                                               (* 210.10 *)
		BinRoot := TwinNode.FindBinRoot;                                (* 310.13 *)
		BinRoot.ComputeWeight;                                          (* 310.06 *)
		BinRoot.Balance2(Node, NodeType);                               (* 310.02 *)
		case NodeType of
			NodeLeft: 
				HelpNode := BetakBoxNode(Node.Left);
			NodeRight: 
				HelpNode := BetakBoxNode(Node.Right);
			NodeUp: 
				HelpNode := Node.Up;
			NodeNil: 
				HelpNode := Node;
		end;
		HelpNode.ComputeNewComponents                                   (* 210.04 *)
	end;

(* 210.03 ****************************************************************** *)
procedure BetakBoxNode.CentroidTest (Son: BetakBoxNode; var Node: BetakBoxNode);
	begin
		if (Node = nil) and Son.Internal then
			if 3 * Component > 4 * (Son.Component + 2) then
				Node := Node
			else
				Node := self
		else
			Node := Node
	end;
(* 210.04 ****************************************************************** *)
procedure BetakBoxNode.ComputeNewComponents;
	begin
		Component := 1;
		if BetakBoxNode(Left) <> nil then
			BetakBoxNode(Left).ComputeNewComponents;                       (* 210.04 *)
		if BetakBoxNode(Right) <> nil then
			BetakBoxNode(Right).ComputeNewComponents;                      (* 210.04 *)
		if Up <> nil then
			Up.ComputeNewComponents;                                       (* 210.04 *)
		if BetakBoxNode(Left) <> nil then
			Component := Component + BetakBoxNode(Left).Component;
		if BetakBoxNode(Right) <> nil then
			Component := Component + BetakBoxNode(Right).Component;
		if Up <> nil then
			Component := Component + Up.Component;
	end;

(* 210.05 ****************************************************************** *)
procedure BetakBoxNode.freeAll;
	begin
		if BetakBoxNode(Left) <> nil then
			BetakBoxNode(Left).freeAll;                                    (* 210.05 *)
		if BetakBoxNode(Right) <> nil then
			BetakBoxNode(Right).freeAll;                                   (* 210.05 *)
		if Up <> nil then
			Up.freeAll;                                                    (* 210.05 *)
		self.FreeNode;                                                  (* 210.06 *)
	end;

(* 210.06 ****************************************************************** *)
procedure BetakBoxNode.FreeNode;
	begin
		kBox.free;                                                         (* obj *)
		TwinNode.free;                                                     (* obj *)
		if Internal then
			DividingLine.freeAll                                           (* 112.02 *)
		else
			Head.freeAll;                                                  (* 113.01 *)
		self.free;                                                         (* obj *)
	end;

(* 210.07 ****************************************************************** *)
procedure BetakBoxNode.IncComponents (p: dDimPoint; var Node: BetakBoxNode);
	var
		NextNode: BetakBoxNode;
	begin
		if Internal then begin
			Component := Component + 2;
			NextNode := self.FindNextNode(p);                              (* 210.14 *)
			self.CentroidTest(NextNode, Node);                             (* 210.03 *)
			NextNode.IncComponents(p, Node)                                (* 210.07 *)
		end
	end;

(* 210.08 ****************************************************************** *)
procedure BetakBoxNode.InitNode (var TL, BR: dDimPoint);
	begin
		BetakBoxNode(Left) := nil;
		BetakBoxNode(Right) := nil;
		Up := nil;
		Head := nil;
		new(kBox);
		failNil(Handle(kBox));
		kBox.TopLeft := TL;
		kBox.BottomRight := BR;
		new(TwinNode);
		failNil(Handle(TwinNode));
		TwinNode.InitBinNode(self);                                     (* 310.08 *)
		Internal := false;
		Complete := false;
		Component := 1;
	end;

(* 210.09 ****************************************************************** *)
procedure BetakBoxNode.InsertTwoLeaves;
	var
		LeftNode, RightNode: BetakBoxNode;
	begin
		new(LeftNode);
		failNil(Handle(LeftNode));
		new(RightNode);
		failNil(Handle(RightNode));
		LeftNode.InitNode(kBox.TopLeft, DividingLine.LastPoint);        (* 210.08 *)
		RightNode.InitNode(DividingLine.FirstPoint, kBox.BottomRight);  (* 210.08 *)
		Internal := true;
		BetakBoxNode(Left) := LeftNode;
		BetakBoxNode(Right) := RightNode;
		TwinNode.Insert;                                                (* 310.09 *)
	end;

(* 210.10******************************************************************* *)
procedure BetakBoxNode.MarkBinNode;
	begin
		if BetakBoxNode(Left) <> nil then
			BetakBoxNode(Left).MarkBinNode;                                (* 210.10 *)
		if BetakBoxNode(Right) <> nil then
			BetakBoxNode(Right).MarkBinNode;                               (* 210.10 *)
		if Up <> nil then
			Up.MarkBinNode;                                                (* 210.10 *)
		TwinNode.Mark := true;
		BetakBoxNode(Left) := nil;
		BetakBoxNode(Right) := nil;
		Up := nil;
	end;

(* 210.11 ****************************************************************** *)
procedure BetakBoxNode.MovePoints;
	var
		HelpPtr: ListElt;
	begin
		HelpPtr := Head;
		while HelpPtr <> nil do begin
			case DividingLine.FindNextNode(HelpPtr.Element) of             (* 112.04 *)
				NodeLeft: 
					BetakBoxNode(Left).Insert(HelpPtr.Element);                  (* 110.04 *)
				NodeRight: 
					BetakBoxNode(Right).Insert(HelpPtr.Element);                 (* 110.04 *)
			end;
			HelpPtr := HelpPtr.Next
		end;
		if BetakBoxNode(Left).Complete or BetakBoxNode(Right).Complete then
			unimplementedMethod('to many equal points');
		Head.freeAll;                                                   (* 113.01 *)
	end;

(* 210.12 ****************************************************************** *)
procedure BetakBoxNode.Split;
	begin
		self.MakeDividingLine;                                   (* 110.06 *)
		self.InsertTwoLeaves;                                           (* 210.09 *)
		self.MovePoints;                                                (* 210.11 *)
	end;

(* 210.13 ****************************************************************** *)
function BetakBoxNode.FindMaxComponent: BetakBoxNode;
	var
		Maximum: integer;
		Node: BetakBoxNode;
	begin
		Maximum := 0;
		if BetakBoxNode(Left) <> nil then begin
			Maximum := BetakBoxNode(Left).Component;
			Node := BetakBoxNode(Left)
		end;
		if BetakBoxNode(Right) <> nil then
			if Maximum < BetakBoxNode(Right).Component then begin
				Maximum := BetakBoxNode(Right).Component;
				Node := BetakBoxNode(Right)
			end;
		if Up <> nil then
			if Maximum < Up.Component then begin
				Maximum := Up.Component;
				Node := Up
			end;
		FindMaxComponent := Node
	end;

(* 210.14 ****************************************************************** *)
function BetakBoxNode.FindNextNode (p: dDimPoint): BetakBoxNode;
	begin
		if kBox.Include(p) then                                         (* 111.01 *)
			case DividingLine.FindNextNode(p) of                           (* 112.04 *)
				NodeLeft: 
					FindNextNode := BetakBoxNode(Left);
				NodeRight: 
					FindNextNode := BetakBoxNode(Right)
			end
		else
			FindNextNode := Up
	end;

(* 210.15 ****************************************************************** *)
function BetakBoxNode.FindNewRoot (HalfComponent: integer): BetakBoxNode;
	var
		Node: BetakBoxNode;
	begin
		Node := self.FindMaxComponent;                                  (* 210.13 *)
		if Node.Component <= HalfComponent then
			FindNewRoot := self
		else
			FindNewRoot := Node.FindNewRoot(HalfComponent)                 (* 210.15 *)
	end;

(* 210.16 ****************************************************************** *)
function BetakBoxNode.PointerToFather (var Node, Father: BetakBoxNode; p: dDimPoint): BetakBoxNode;
	var
		NextNode: BetakBoxNode;
	begin
		if Component = Node.Component then
			PointerToFather := Father
		else begin
			NextNode := self.FindNextNode(p);                              (* 210.14 *)
			Father := self;
			PointerToFather := NextNode.PointerToFather(Node, Father, p);  (* 210.16 *)
		end
	end;

(* 210.17 ****************************************************************** *)
function BetakBoxNode.PointerToTwin (Node: NextNode): BinNode;
	begin
		if Node = NodeLeft then
			PointerToTwin := BetakBoxNode(Left).TwinNode
		else
			PointerToTwin := BetakBoxNode(Right).TwinNode
	end;

(* 210.18 ****************************************************************** *)
function BetakBoxNode.Search (p: dDimPoint): BinarykBoxNode;
	var
		NextNode: BetakBoxNode;
	begin
		if Internal then begin
			NextNode := self.FindNextNode(p);                              (* 210.14 *)
			Search := BetakBoxNode(NextNode.Search(p))                     (* 210.18 *)
		end
		else
			Search := self;
	end;
(************************  OBJECT BetakBoxTree 200 ************************* *)
(* 200.01 ****************************************************************** *)
procedure BetakBoxTree.freeAll;
	begin
		BetakBoxNode(Root).freeAll;                                     (* 210.05 *)
		RootTopLeft.free;                                                  (* obj *)
		RootBottomRight.free;                                              (* obj *)
		self.free                                                          (* obj *)
	end;

(* 200.02 ****************************************************************** *)
procedure BetakBoxTree.init;
	begin
		BetakBoxNode(Root) := nil;
	end;

(* 200.03 ****************************************************************** *)
procedure BetakBoxTree.Insert (p: dDimPoint);
	var
		SearchedNode, NoCentroid, Father: BetakBoxNode;
	begin
		SearchedNode := BetakBoxNode(self.Search(p));                   (* 200.06 *)
		if SearchedNode = nil then begin
			self.MakeRoot;                                                 (* 200.05 *)
			BetakBoxNode(Root).InitNode(RootTopLeft, RootBottomRight);     (* 210.08 *)
			BetakBoxNode(Root).Insert(p)                                   (* 110.04 *)
		end
		else begin
			SearchedNode.Insert(p);                                        (* 110.04 *)
			if SearchedNode.Complete then begin
				SearchedNode.Split;                                    (* 210.12 *)
				NoCentroid := nil;
				BetakBoxNode(Root).IncComponents(p, NoCentroid);              (* 210.07 *)
				if NoCentroid <> nil then begin
					if animationFlag[AClosestPairItem] then begin
						displayInfoString(StringOf('weight : tree = ', BetakBoxNode(Root).Component : 3, ' unbalanced subtree = ', NoCentroid.Component : 3));
						waitForClick(animationFlag[AClosestPairItem]);
						displayInfoString(StringOf('                                                                   '));
					end;
					Father := nil;
					Father := BetakBoxNode(Root).PointerToFather(NoCentroid, Father, p);(* 210.16 *)
					if Father <> nil then
						NoCentroid.Balance(Father)                                  (* 210.01 *)
					else begin
						NoCentroid.Balance(Father);                                 (* 210.01 *)
						Root := Father
					end;
				end
			end;
		end;
	end;

(* 200.04 ****************************************************************** *)
procedure BetakBoxTree.Insert2 (p: dDimPoint);
	var
		SearchedNode, NoCentroid, Father: BetakBoxNode;
		NodeType: NextNode;
	begin
		SearchedNode := BetakBoxNode(self.Search(p));                   (* 200.06 *)
		if SearchedNode = nil then begin
			self.MakeRoot;                                                 (* 200.05 *)
			BetakBoxNode(Root).InitNode(RootTopLeft, RootBottomRight);     (* 210.08 *)
			BetakBoxNode(Root).Insert(p)                                   (* 110.04 *)
		end
		else begin
			SearchedNode.Insert(p);                                        (* 110.04 *)
			if SearchedNode.Complete then begin
				SearchedNode.Split;                                    (* 210.12 *)
				NoCentroid := nil;
				BetakBoxNode(Root).IncComponents(p, NoCentroid);              (* 210.07 *)
				if NoCentroid <> nil then begin
					if animationFlag[AClosestPairItem] then begin
						displayInfoString(StringOf('weight : tree = ', BetakBoxNode(Root).Component : 3, ' unbalanced subtree = ', NoCentroid.Component : 3));
						waitForClick(animationFlag[AClosestPairItem]);
						displayInfoString(StringOf('                                                                   '));
					end;
					Father := nil;
					Father := BetakBoxNode(Root).PointerToFather(NoCentroid, Father, p);(* 210.16 *)
					if Father <> nil then begin
						if Father.Left = NoCentroid then
							NodeType := NodeLeft
						else if Father.Right = NoCentroid then
							NodeType := NodeRight
						else
							NodeType := NodeUp;
						NoCentroid.Balance2(Father, NodeType);                      (* 210.02 *)
					end
					else begin
						NoCentroid.Balance2(Father, NodeNil);                       (* 210.02 *)
						Root := Father
					end;
				end
			end;
		end;
	end;

(* 200.05 ****************************************************************** *)
procedure BetakBoxTree.MakeRoot;
	var
		i: integer;
	begin
		new(BetakBoxNode(Root));
		failNil(Handle(Root));
		new(RootTopLeft);
		failNil(Handle(RootTopLeft));
		new(RootBottomRight);
		failNil(Handle(RootBottomRight));
		RootTopLeft.dimension := k;
		RootBottomRight.dimension := k;
		for i := 1 to k do begin
			RootTopLeft.coordinates[i] := RootBoxMin;
			RootBottomRight.coordinates[i] := RootBoxMax
		end;
	end;

(* 200.06 ****************************************************************** *)
function BetakBoxTree.Search (p: dDimPoint): BinarykBoxNode;
	begin
		if BetakBoxNode(Root) = nil then
			Search := nil
		else
			Search := BetakBoxNode(Root).Search(p);                        (* 210.18 *)
	end;

(***********************  OBJECT BinNode 310 ******************************* *)
(* 310.01 ****************************************************************** *)
procedure BinNode.Balance (NodeType: NextNode);
	var
		BinRoot, HighestNode: BinNode;
		HighestIndex: integer;
	begin
		Mark := false;
		if NodeType = NodeUp then
			BinRoot := self.FindBinRoot;                                   (* 310.13 *)
		if NodeType = NodeLeft then
			BinRoot := Left;
		if NodeType = NodeRight then
			BinRoot := Right;
		BinRoot.MakeLabels;                                             (* 310.11 *)
		HighestNode := nil;
		HighestIndex := -1;
		BinRoot.FindHighestIndex(HighestNode, HighestIndex);            (* 310.07 *)
		HighestNode.ChangePointers(NodeType, self)                      (* 310.04 *)
	end;

(* 310.02 ****************************************************************** *)
procedure BinNode.Balance2 (var Node: BetakBoxNode; NodeType: NextNode);
	var
		HalfWeight: integer;
		HalfNode, BinRoot: BinNode;
	begin
{$IFC THINK_PASCAL }
		HalfWeight := trunc(0.5 * Index) + 1;
{$ELSEC }
		HalfWeight := truncLong(0.5 * Index) + 1;
{$ENDC }
		HalfNode := self.FindHalfNode(HalfWeight);                      (* 310.14 *)
		HalfNode.Mark := false;
		case NodeType of
			NodeLeft: 
				BetakBoxNode(Node.Left) := HalfNode.BetaTwinNode;
			NodeRight: 
				BetakBoxNode(Node.Right) := HalfNode.BetaTwinNode;
			NodeUp: 
				Node.Up := HalfNode.BetaTwinNode;
			NodeNil: 
				Node := HalfNode.BetaTwinNode;
		end;
		if HalfNode.Left <> nil then
			if HalfNode.Left.Mark then begin
				HalfNode.Left.ComputeWeight;                                  (* 310.06 *)
				HalfNode.Left.Balance2(HalfNode.BetaTwinNode, NodeLeft);      (* 310.02 *)
			end;
		if HalfNode.Right <> nil then
			if HalfNode.Right.Mark then begin
				HalfNode.Right.ComputeWeight;                                 (* 310.06 *)
				HalfNode.Right.Balance2(HalfNode.BetaTwinNode, NodeRight);    (* 310.02 *)
			end;
		if HalfNode.Parent <> nil then
			if HalfNode.Parent.Mark then begin
				BinRoot := HalfNode.Parent.FindBinRoot;                       (* 310.13 *)
				BinRoot.ComputeWeight;                                        (* 310.06 *)
				BinRoot.Balance2(HalfNode.BetaTwinNode, NodeUp);              (* 310.02 *)
			end;
	end;

(* 310.03 ****************************************************************** *)
procedure BinNode.ChangeNextPointer (NodeType: NextNode; var Node: BinNode);
	var
		HighestNode: BinNode;
		HighestIndex: integer;
	begin
		HighestNode := nil;
		HighestIndex := -1;
		self.FindHighestIndex(HighestNode, HighestIndex);               (* 310.07 *)
		if HighestNode <> nil then
			HighestNode.ChangePointers(NodeType, Node)                     (* 310.04 *)
	end;

(* 310.04 ****************************************************************** *)
procedure BinNode.ChangePointers (NodeType: NextNode; Node: BinNode);
	var
		BinRoot: BinNode;
	begin
		case NodeType of
			NodeLeft: 
				BetakBoxNode(Node.BetaTwinNode.Left) := self.BetaTwinNode;
			NodeRight: 
				BetakBoxNode(Node.BetaTwinNode.Right) := self.BetaTwinNode;
			NodeUp: 
				Node.BetaTwinNode.Up := self.BetaTwinNode;
		end;
		Mark := false;
		if Left <> nil then
			if Left.Mark then
				Left.ChangeNextPointer(NodeLeft, self);                      (* 310.03 *)
		if Right <> nil then
			if Right.Mark then
				Right.ChangeNextPointer(NodeRight, self);                    (* 310.03 *)
		if Parent <> nil then
			if Parent.Mark then begin
				BinRoot := FindBinRoot;
				BinRoot.ChangeNextPointer(NodeUp, self);                     (* 310.03 *)
			end;
	end;

(* 310.05 ****************************************************************** *)
procedure BinNode.ComputeLabel;
	var
		i, Value, uebertrag: integer;
	begin
		if Index > 0 then
			for i := 0 to Index - 1 do
				Lab[i] := 0;
		Lab[Index] := 1;
		uebertrag := 0;
		for i := Index + 1 to LabelMax do begin
			Value := 0;
			if Left <> nil then
				if Left.Mark then
					Value := Left.Lab[i];
			if Right <> nil then
				if Right.Mark then
					Value := Value + Right.Lab[i];
			Value := Value + uebertrag;
			Lab[i] := Value mod 2;
			uebertrag := Value div 2;
		end;
	end;

(* 310.06 ****************************************************************** *)
procedure BinNode.ComputeWeight;
	begin
		Index := 1;
		if Left <> nil then
			if Left.Mark then
				Left.ComputeWeight;                                           (* 310.06 *)
		if Right <> nil then
			if Right.Mark then
				Right.ComputeWeight;                                          (* 310.06 *)
		if Left <> nil then
			if Left.Mark then
				Index := Index + Left.Index;
		if Right <> nil then
			if Right.Mark then
				Index := Index + Right.Index;
	end;

(* 310.07 ****************************************************************** *)
procedure BinNode.FindHighestIndex (var HighestNode: BinNode; var GreatestIndex: integer);
	begin
		if Mark then begin
			if Index > GreatestIndex then begin
				HighestNode := self;
				GreatestIndex := Index;
			end;
			if Left <> nil then
				if Left.Mark then
					Left.FindHighestIndex(HighestNode, GreatestIndex);           (* 310.07 *)
			if Right <> nil then
				if Right.Mark then
					Right.FindHighestIndex(HighestNode, GreatestIndex);          (* 310.07 *)
		end;
	end;

(* 310.08 ****************************************************************** *)
procedure BinNode.InitBinNode (Node: BetakBoxNode);
	begin
		Parent := nil;
		Left := nil;
		Right := nil;
		BetaTwinNode := Node;
		Mark := false;
	end;

(* 310.09 ****************************************************************** *)
procedure BinNode.Insert;
	begin
		Left := BetaTwinNode.PointerToTwin(NodeLeft);                   (* 210.17 *)
		Right := BetaTwinNode.PointerToTwin(NodeRight);                 (* 210.17 *)
		Left.Parent := self;
		Right.Parent := self;
	end;

(* 310.10 ****************************************************************** *)
procedure BinNode.MakeIndex;
	var
		l, r: integer;
	begin
		if Left <> nil then
			if Left.Mark then
				l := 2
			else
				l := 1
		else
			l := 1;
		if Right <> nil then
			if Right.Mark then
				r := 8
			else
				r := 4
		else
			r := 4;
		Index := self.ComputeIndex(l + r);                              (* 310.12 *)
	end;

(* 310.11 ****************************************************************** *)
procedure BinNode.MakeLabels;
	begin
		if Left <> nil then
			if Left.Mark then
				Left.MakeLabels;                                              (* 310.11 *)
		if Right <> nil then
			if Right.Mark then
				Right.MakeLabels;                                             (* 310.11 *)
		self.MakeIndex;                                                 (* 310.10 *)
		self.ComputeLabel;                                              (* 310.05 *)
	end;

(* 310.12 ****************************************************************** *)
function BinNode.ComputeIndex (Value: integer): integer;
	var
		i, j: integer;
		ok: boolean;
	begin
		i := -1;
		ok := false;
		case Value of
			10: 
				repeat (* Left und Right ungleich nil und markiert *)
					i := i + 1;
					if (Left.Lab[i] = 0) and (Right.Lab[i] = 0) then begin
						j := i;
						repeat
							j := j + 1;
							if Left.Lab[j] * Right.Lab[j] = 0 then
								ok := true
							else
								ok := false;
						until (j = LabelMax) or not ok
					end;
				until (i = LabelMax) or ok;
			9: 
				repeat (* Nur Right ist ungleich nil und markiert *)
					i := i + 1;
					if Right.Lab[i] = 0 then begin
						j := i;
						repeat
							j := j + 1;
							if Right.Lab[j] = 0 then
								ok := true
							else
								ok := false;
						until (j = LabelMax) or not ok
					end;
				until (i = LabelMax) or ok;
			6: 
				repeat (* Nur Left ist ungleich nil und markiert *)
					i := i + 1;
					if Left.Lab[i] = 0 then begin
						j := i;
						repeat
							j := j + 1;
							if Left.Lab[j] = 0 then
								ok := true
							else
								ok := false;
						until (j = LabelMax) or not ok
					end;
				until (i = LabelMax) or ok;
			5: 
				i := 0;
		end; (* case *)
		ComputeIndex := i;
	end;

(* 310.13 ****************************************************************** *)
function BinNode.FindBinRoot: BinNode;
	begin
		if Parent <> nil then
			if Parent.Mark then
				FindBinRoot := Parent.FindBinRoot                             (* 310.13 *)
			else
				FindBinRoot := self
		else
			FindBinRoot := self;
	end;

(* 310.14 ****************************************************************** *)
function BinNode.FindHalfNode (Weight: integer): BinNode;
	var
		Node: BinNode;
	begin
		Node := self.FindNextNode;                                      (* 310.15 *)
		if Node <> nil then
			if Node.Index >= Weight then
				FindHalfNode := Node.FindHalfNode(Weight)                     (* 310.14 *)
			else
				FindHalfNode := self
		else
			FindHalfNode := self
	end;

(* 310.15 ****************************************************************** *)
function BinNode.FindNextNode: BinNode;
	var
		Node: BinNode;
		Value: integer;
	begin
		Node := nil;
		Value := 0;
		if Left <> nil then
			if Left.Mark then begin
				Value := Left.Index;
				Node := Left
			end;
		if Right <> nil then
			if Right.Mark and (Right.Index > Value) then begin
				Value := Right.Index;
				Node := Right;
			end;
		FindNextNode := Node;
	end;

end.
