(*	XYZ GeoBench (eXperimental geometrY Zurich):		*)
(*	An object oriented, interactive workbench for		*)
(*	experimental geometric computation.					*)
(*														*)
(*	 1999 by Peter Schorn, peter.schorn@acm.org		*)
(*	Culmannstrasse 77, CH-8006 Zrich, Switzerland		*)
(*	This software may be copied freely for educational	*)
(*	and research purposes only. For commercial use,		*)
(*	please contact the author.							*)
(*														*)
(*	Version 5.0.5: 30-Sep-1999							*)

unit lineSegmentVector;

interface

uses
{$IFC ScriptEnabled }
	incremental;
{$ELSEC }
	basicGeometry;
{$ENDC }

type
	neighborKind = (leftNeighbors, rightNeighbors, allNeighbors);

	lineSegmentVector = object(homogenVector)
(* a collection of line segments *)

			function equal (o: obj): boolean;
			override;
			procedure describe;
			override;
			procedure execute (item: integer);
			override;
			function memberType (o: obj): boolean;
			override;
			function getType: objectType;
			override;
			function getTextType: Str255;
			override;
			function convertToType (t: pointTypeRange): obj;
			override;

{$IFC VorlesungGeoAlg }
			function lowerEnvelopeNew: obj; (* actually a polyLine *)
{$ENDC }

(* Intersection routines *)
			function simpleFirstIntersect: obj;
(* Computes the first intersection using the trivial method. The result can be a 'point2d' object or	*)
(* a 'lineSegment' object or nil if no intersection exists. *)

			function planeSweepFirstIntersect (ask: boolean): obj;
(* Computes the first intersection using a plane sweep. The result can be a 'point2d' object or		*)
(* a 'lineSegment' object or nil if no intersection exists. *)

			function simpleClosestPair: obj;								(* point or lineSegment *)
(* added by FB *)

			function planeSweepClosestPair (ask: boolean): obj;		(* point or lineSegment *)
(* added by FB *)

			function simpleAllIntersect: vector;
(* Computes all pairwise intersections with the trivial algorithm. The result is a vector of points	*)
(* and line segments. *)

			function planeSweepAllIntersect (ask: boolean): vector;
(* Computes all intersections using a plane sweep algorithm. The result is a vector of points and		*)
(* line segments. If 'ask = TRUE' the user is asked to choose how the x-queue and the y-table			*)
(* should be implemented. *)

{$IFC ScriptEnabled }
			function planeSweepAllIntersectInc (var bloc: incAllIntersect; finish: boolean): vector;
(* M_SCM, 02.02.94 *)
{$ENDC }

			function hvPlaneSweepAllIntersect (ask: boolean): vector;
(* Like 'planeSweepAllIntersect' but all line segments must be either horizontal or vertical. *)

(* Projection routines *)
			function projectOnX: lineSegmentVector;
(* Projects all line segments on the x-axis and produces a new 'lineSegmentVector'. *)

			function projectOnY: lineSegmentVector;
(* Projects all line segments on the y-axis and produces a new 'lineSegmentVector'. *)

			function projectOnXY: lineSegmentVector;
(* Projects all line segments on the axis that results in a shorter segment and produces a			*)
(* new 'lineSegmentVector'. *)

(* Miscellaneous *)
			function isHorizontalVertical: boolean;
(* Tests whether all line segments are either horizontal or vertical. *)

			function eliminateZeroLengthSegments: lineSegmentVector;
(* Eliminates all line segments where the start point and the end point coincide and produces a	*)
(* new 'lineSegmentVector'. *)

			function clipRect (rect: rectangle): lineSegmentVector;
(* Clips lineSegmentVector 'self' by rectangle 'rect'; the algorithm follows the idea by Cohen & Sutherland. *)

			function lowerEnvelope: obj; (* actually a polyLine *)

			procedure finishNearestNeighbors (kind: neighborKind);

			procedure nilQ;
		end;

procedure initLineSegmentVector;


implementation

{$SETC AclipLSVbyRectangle = TRUE }
{$SETC ALineSegmentIntersection = TRUE }

const
{$IFC VorlesungGeoAlg }
	LowerEnvelopeNewC = -1;
{$ENDC }
	PSClosestPairDynamicC = 0;
	SimpleAllIntersectC = 1;
	PSClosestPairC = 2; (* added by FB *)
	PSClosestPairAskC = 3; (* added by FB *)
	ClippingC = 4;
	PSAllIntersectC = 5;
	PSAllIntersectAskC = 6;
	ProjectOnXC = 7;
	ProjectOnYC = 8;
	HvPlaneSweepAllIntersectC = 9;
	HvPlaneSweepAllIntersectAskC = 10;
	SimpleFirstIntersectC = 11;
	PSFirstIntersectC = 12;
	PSFirstIntersectAskC = 13;
	ProjectOnXYC = 14;
	EliminateZeroLengthSegmentsC = 15;
	LowerEnvelopeC = 16;
	SimpleClosestPairC = 17; (* added by FB *)
	SimplePolygLineC = 18;
	PSAllIntersectDynamicC = 19;
	HvPlaneSweepAllIntersectDynamicC = 20;
	PSFirstIntersectDynamicC = 21;
	LowerEnvelopeDynamicC = 22;

type
	destroyAction = (doNothing, doFree, doDispose);

function doComputeClosestPair (lv: lineSegmentVector): obj;
	begin
		doComputeClosestPair := lv.planeSweepClosestPair(false);
	end;

function doComputeAllIntersect (lv: lineSegmentVector): obj;
	begin
		doComputeAllIntersect := lv.planeSweepAllIntersect(false);
	end;

function doComputeFirstIntersect (lv: lineSegmentVector): obj;
	begin
		doComputeFirstIntersect := lv.planeSweepFirstIntersect(false);
	end;

function doComputeHVIntersect (lv: lineSegmentVector): obj;
	begin
		doComputeHVIntersect := lv.hvPlaneSweepAllIntersect(false);
	end;

function doComputeLowerEnvelope (lv: lineSegmentVector): obj;
	begin
		doComputeLowerEnvelope := lv.lowerEnvelope;
	end;

function lineSegmentVector.convertToType (t: pointTypeRange): obj;
	var
		v: lineSegmentVector;
		i: seqIndex0;
	begin
		v := lineSegmentVector(duplicate);
		for i := 1 to length do
			v.elements^^[i] := elements^^[i].convertToType(t);
		convertToType := v;
	end;

procedure lineSegmentVector.nilQ;
	var
		i: seqIndex0;
	begin
		for i := 1 to length do
			lineSegment(elements^^[i]).q := nil;
	end;

procedure lineSegmentVector.finishNearestNeighbors (kind: neighborKind);
	var
		i: seqIndex0;
	begin
		if kind = leftNeighbors then begin
			elements^^[1].free;
			swap(1, length);
			length := length - 1;
		end
		else if kind = rightNeighbors then begin
			elements^^[length].free;
			length := length - 1;
		end;
		for i := 1 to length do
			with lineSegment(elements^^[i]) do
				q := point2d(q.duplicate);
	end;

procedure makeLineSegmentVector (var lv: lineSegmentVector; var destroy: destroyAction);
	var
		s: vector;
	begin
		if currentArguments.length > 1 then begin
			new(lv);
			failNil(Handle(lv));
			lv.initFromList(currentArguments);
			destroy := doFree;
		end
		else if member(currentArguments.ith(1), lineSegmentVector) then begin
			lv := lineSegmentVector(currentArguments.ith(1));
			destroy := doNothing;
		end
		else begin
			new(lv);
			failNil(Handle(lv));
			s := vector(currentArguments.ith(1));
			lv.length := s.length;
			lv.elements := s.elements;
			destroy := doDispose;
		end;
	end;

function makeSimplePolyLine (lv: lineSegmentVector): obj;
	var
		result: vector;
		i: seqIndex0;
	begin
		currentVectorLength := 2 * lv.length;
		result := vector(newInitObj(polyLineType));
		for i := 1 to lv.length do begin
			with lineSegment(lv.ith(i)) do begin
				result.append(p.duplicate);
				result.append(q.duplicate);
			end;
		end;
		result.sortAll(pointLessThanXY);
		makeSimplePolyLine := result;
	end;

procedure lineSegmentVector.describe;

	function isLineSegmentVector (v: vector): boolean;
		begin
			isLineSegmentVector := (v.length > 1) & (member(v, lineSegmentVector) | member(v.elements^^[1], lineSegment));
		end;

	function testType (o: obj; t: objectType): boolean;
		begin
			if member(o, vector) then
				testType := lineSegment(vector(o).elements^^[1]).p.getType = t
			else
				testType := lineSegment(o).p.getType = t;
		end;

	function testHorizontalVertical: boolean;
		var
			lv: lineSegmentVector;
			destroy: destroyAction;
		begin
			makeLineSegmentVector(lv, destroy);
			testHorizontalVertical := lv.isHorizontalVertical;
			if destroy = doFree then
				lv.free
			else if destroy = doDispose then
				dispose(lv);
		end;

	begin
		with currentArguments do
			if (length = 1) & currentArguments1IsHomogenVector & isLineSegmentVector(vector(ith(1))) | (length > 1) & currentArgumentsSameType & member(first^^.value, lineSegment) then begin
				beginSubMenu('All Intersect');
				menuEntry('Simple', SimpleAllIntersectC);
				menuEntry('Plane Sweep', PSAllIntersectC);
				menuEntry('Plane Sweep Ask...', PSAllIntersectAskC);
				menuEntry('Plane Sweep Dynamic', PSAllIntersectDynamicC);
				if testHorizontalVertical then begin
					menuEntry('Horizontal Vertical', HvPlaneSweepAllIntersectC);
					menuEntry('Horizontal Vertical Ask...', HvPlaneSweepAllIntersectAskC);
					menuEntry('Horizontal Vertical Dynamic', HvPlaneSweepAllIntersectDynamicC);
				end;
				endSubMenu;
				beginSubMenu('First Intersect');
				menuEntry('Simple', SimpleFirstIntersectC);
				menuEntry('Plane Sweep', PSFirstIntersectC);
				menuEntry('Plane Sweep Ask...', PSFirstIntersectAskC);
				menuEntry('Plane Sweep Dynamic', PSFirstIntersectDynamicC);
				endSubMenu;
				beginSubMenu('Closest Pair');(* added by FB *)
				menuEntry('Simple', SimpleClosestPairC);
				menuEntry('Plane Sweep', PSClosestPairC);
				menuEntry('Plane Sweep Ask...', PSClosestPairAskC);
				menuEntry('Plane Sweep Dynamic', PSClosestPairDynamicC);
				endSubMenu;
				beginSubMenu('Make');
				menuEntry('Horizontal', ProjectOnXC);
				menuEntry('Vertical', ProjectOnYC);
				menuEntry('Horizontal Vertical', ProjectOnXYC);
				menuEntry('Simple Polyline', SimplePolygLineC);
				endSubMenu;
				menuEntry('Eliminate Zero Segments', EliminateZeroLengthSegmentsC);
				menuEntry('Lower Envelope', LowerEnvelopeC);
				menuEntry('Lower Envelope Dynamic', LowerEnvelopeDynamicC);
{$IFC VorlesungGeoAlg }
				menuEntry('Lower Envelope New', LowerEnvelopeNewC);
{$ENDC }
			end
			else begin
				if (length = 2) & (member(first^^.value, rectangle) & currentArguments2IsHomogenVector & isLineSegmentVector(vector(ith(2))) | member(last^^.value, rectangle) & currentArguments1IsHomogenVector & isLineSegmentVector(vector(ith(1)))) then
					menuEntry('Clip With Rectangle', ClippingC);
			end;
	end;

procedure makeClipLineSegmentVector (lineSegmentVectorCandidat: vector; var lv: lineSegmentVector; var destroy: destroyAction);
(* spezielles "makeLineSegmentVector" fuer Aufgabe 3 von "Von0Auf100" *)
	var
		s: vector;
	begin (* makeClipLineSegmentVector *)
		if member(lineSegmentVectorCandidat, lineSegmentVector) then begin
			lv := lineSegmentVector(lineSegmentVectorCandidat);
			destroy := doNothing;
		end
		else begin
			new(lv);
			failNil(Handle(lv));
			s := vector(lineSegmentVectorCandidat);
			lv.length := s.length;
			lv.elements := s.elements;
			destroy := doDispose;
		end;
	end;  (* makeClipLineSegmentVector *)

procedure lineSegmentVector.execute (item: integer);
	var
		lv: lineSegmentVector;
		destroy: destroyAction;

	begin
		if item = ClippingC then
			if currentArguments2IsHomogenVector then begin
				makeClipLineSegmentVector(vector(currentArguments.ith(2)), lv, destroy);
				currentResult.appendNonNil(lv.clipRect(rectangle(currentArguments.ith(1))));
			end
			else begin
				makeClipLineSegmentVector(vector(currentArguments.ith(1)), lv, destroy);
				currentResult.appendNonNil(lv.clipRect(rectangle(currentArguments.ith(2))));
			end
		else begin
			makeLineSegmentVector(lv, destroy);
			case item of
				PSClosestPairDynamicC: 
					lv.dynamicExecutionHighlightShowType(doComputeClosestPair);
				PSAllIntersectDynamicC: 
					lv.dynamicExecutionHighlightShowType(doComputeAllIntersect);
				HvPlaneSweepAllIntersectDynamicC: 
					lv.dynamicExecutionHighlightShowType(doComputeHVIntersect);
				PSFirstIntersectDynamicC: 
					lv.dynamicExecutionHighlightShowType(doComputeFirstIntersect);
				LowerEnvelopeDynamicC: 
					lv.dynamicExecutionHighlightShowType(doComputeLowerEnvelope);
				SimpleAllIntersectC: 
					currentResult.appendNonNil(lv.simpleAllIntersect);
				SimpleFirstIntersectC: 
					currentResult.appendNonNil(lv.simpleFirstIntersect);
				PSAllIntersectC, PSAllIntersectAskC: 
					currentResult.appendNonNil(lv.planeSweepAllIntersect(item = PSAllIntersectAskC));
				PSFirstIntersectC, PSFirstIntersectAskC: 
					currentResult.appendNonNil(lv.planeSweepFirstIntersect(item = PSFirstIntersectAskC));
				HvPlaneSweepAllIntersectC, HvPlaneSweepAllIntersectAskC: 
					currentResult.appendNonNil(lv.hvPlaneSweepAllIntersect(item = HvPlaneSweepAllIntersectAskC));
				ProjectOnXC: 
					currentResult.appendNonNil(lv.projectOnX);
				ProjectOnYC: 
					currentResult.appendNonNil(lv.projectOnY);
				ProjectOnXYC: 
					currentResult.appendNonNil(lv.projectOnXY);
				EliminateZeroLengthSegmentsC: 
					currentResult.appendNonNil(lv.eliminateZeroLengthSegments);
				SimplePolygLineC: 
					currentResult.appendNonNil(makeSimplePolyLine(lv));
				LowerEnvelopeC: 
					currentResult.appendNonNil(lv.lowerEnvelope);
{$IFC VorlesungGeoAlg }
				LowerEnvelopeNewC: 
					currentResult.appendNonNil(lv.lowerEnvelopeNew);
{$ENDC }
				SimpleClosestPairC: 
					currentResult.appendNonNil(lv.simpleClosestPair); (* added by FB *)
				PSClosestPairC, PSClosestPairAskC: 
					currentResult.appendNonNil(lv.planeSweepClosestPair(item = PSClosestPairAskC)); (* added by FB *)
			end;
		end;
		if destroy = doFree then
			lv.free
		else if destroy = doDispose then
			dispose(lv);
	end;

(* Added by "Von 0 auf 100" *)
function lineSegmentVector.clipRect (rect: rectangle): lineSegmentVector;
(* clips lineSegmentVector 'self' by rectangle 'rect'; the algorithm follows the idea by Cohen&Sutherland *)
	type
		position = (left, right, bottom, top);
		SegmentCode = set of position;
	var
		resLSV: lineSegmentVector; (* collects the visible parts of the line segments *)
{$IFC AnimationEnabled AND AclipLSVbyRectangle }
		leftLS, rightLS, bottomLS, topLS: lineSegment; (* segmentation of the plain *)
{$ENDC }


	procedure code (x, y: extended; var c: SegmentCode);
		begin
			c := [];
			if x < rect.left then
				c := [left]
			else if x > rect.right then
				c := [right];
			if y < rect.bottom then
				c := c + [bottom]
			else if y > rect.top then
				c := c + [top]
		end;


{$IFC AnimationEnabled AND AclipLSVbyRectangle }
	procedure showNineSegments (var leftLS, rightLS, bottomLS, topLS: lineSegment; rect: rectangle);
   (* the plain is divided in nine segments by the bounding lines of the rectangle 'rect' *)
		const
			minLeft = -15000;
			maxRight = 15000;
			minBottom = -15000;
			maxTop = 15000;
		begin (* showNineSegments *)
			new(leftLS);
			failNil(Handle(leftLS));
			leftLS.init;
			with leftLS do begin
				p.setX(rect.left);
				q.setX(rect.left);
				p.setY(minBottom);
				q.setY(maxTop)
			end;
			new(rightLS);
			failNil(Handle(rightLS));
			rightLS.init;
			with rightLS do begin
				p.setX(rect.right);
				q.setX(rect.right);
				p.setY(minBottom);
				q.setY(maxTop)
			end;
			new(bottomLS);
			failNil(Handle(bottomLS));
			bottomLS.init;
			with bottomLS do begin
				p.setY(rect.bottom);
				q.setY(rect.bottom);
				p.setX(minLeft);
				q.setX(maxRight)
			end;
			new(topLS);
			failNil(Handle(topLS));
			topLS.init;
			with topLS do begin
				p.setY(rect.top);
				q.setY(rect.top);
				p.setX(minLeft);
				q.setX(maxRight)
			end;
			leftLS.interactiveOutput;
			rightLS.interactiveOutput;
			bottomLS.interactiveOutput;
			topLS.interactiveOutput;
		end; (* showNineSegments *)


	procedure hideNineSegments (leftLS, rightLS, bottomLS, topLS: lineSegment);
   (* remove display of plane segmentation *)
		begin (* hideNineSegments *)
			leftLS.freeAll;
			rightLS.freeAll;
			bottomLS.freeAll;
			topLS.freeAll
		end; (* hideNineSegments *)


	procedure clearAnimation;
		var
			i: integer;
		begin (* clearAnimation *)
			hideNineSegments(leftLS, rightLS, bottomLS, topLS);
			for i := 1 to length do (* restore selection highlighting in input window *)
				lineSegment(elements^^[i]).interactiveHighlight
		end; (* clearAnimation *)
{$ENDC }


	procedure subClip (var x, y: extended; x_c, y_c: extended; var c: SegmentCode; c_c: SegmentCode);


		procedure code_horizontal (x: extended; var c: SegmentCode);
     (* exclusively used for horizontal tests, this procedure is faster than the general procedure 'code' *)
			begin(* code_horizontal *)
				c := [];
				if x < rect.left then
					c := [left]
				else if x > rect.right then
					c := [right]
			end; (* code_horizontal *)


		procedure code_vertical (y: extended; var c: SegmentCode);
     (* exclusively used for vertical tests, this procedure is faster than the general procedure 'code' *)
			begin(* code_vertical *)
				c := [];
				if y < rect.bottom then
					c := c + [bottom]
				else if y > rect.top then
					c := c + [top]
			end;(* code_vertical *)


{$IFC AnimationEnabled AND AclipLSVbyRectangle }
			var
				clippedLS: lineSegment; (* shortened line segment (recursively) *)
{$ENDC }

		begin (* subClip *)

{$IFC AnimationEnabled AND AclipLSVbyRectangle }
			if animationFlag[AClipWithRectangleItem] then begin
				new(clippedLS);
				failNil(Handle(clippedLS));
				clippedLS.init;
				with clippedLS do begin
					p.setX(x);
					p.setY(y);
					q.setX(x_c);
					q.setY(y_c)
				end
			end;
{$ENDC }

			while (c <> []) do begin

{$IFC AnimationEnabled }
				if userAbort then begin
					clearAnimation;
					resLSV.freeAll;
					clipRect := nil;
					exit(clipRect)
				end;
{$ENDC }

				if c * c_c <> [] then begin								(* nothing is visible *)

{$IFC AnimationEnabled AND AclipLSVbyRectangle }
					if animationFlag[AClipWithRectangleItem] then begin
						clippedLS.interactiveFlash;
						waitForClick(animationFlag[AClipWithRectangleItem]);
						clippedLS.freeAll
					end;
{$ENDC }

					exit(subClip)
				end
				else begin													(* it is possible that a part is visible *)
					if (left in c) then begin
						y := y + (rect.left - x) * ((y_c - y) / (x_c - x));
						x := rect.left;
						code_vertical(y, c)
					end
					else if (right in c) then begin
						y := y + (rect.right - x) * ((y_c - y) / (x_c - x));
						x := rect.right;
						code_vertical(y, c)
					end
					else if (top in c) then begin
						x := x + (rect.top - y) * ((x_c - x) / (y_c - y));
						y := rect.top;
						code_horizontal(x, c)
					end
					else begin (* at this point: bottom IS in c *)
						x := x + (rect.bottom - y) * ((x_c - x) / (y_c - y));
						y := rect.bottom;
						code_horizontal(x, c)
					end
				end;

{$IFC AnimationEnabled AND AclipLSVbyRectangle }
				if animationFlag[AClipWithRectangleItem] then begin
					with clippedLS do begin
						p.setX(x);
						p.setY(y)
					end;
					clippedLS.interactiveHighlight;
					waitForClick(animationFlag[AClipWithRectangleItem]);
					clippedLS.interactiveHighlight
				end;
{$ENDC }

			end;

{$IFC AnimationEnabled AND AclipLSVbyRectangle }
			if animationFlag[AClipWithRectangleItem] then begin
				clippedLS.freeAll
			end
{$ENDC }

		end; (* subClip *)


		var
			xp, yp, xq, yq: extended; (* coordinates of endpoints of a line segment (PQ) *)
			cp, cq: SegmentCode;
			ls: lineSegment; (* clipped line segment *)
			i: integer;

	begin (* clipRect *)
		new(resLSV);
		failNil(Handle(resLSV));
		currentVectorLength := length;
		resLSV.init;

{$IFC AnimationEnabled AND AclipLSVbyRectangle }
		if animationFlag[AClipWithRectangleItem] then begin
			for i := 1 to length do (* remove selection highlighting in input window *)
				lineSegment(elements^^[i]).interactiveHighlight;
			showNineSegments(leftLS, rightLS, bottomLS, topLS, rect)
		end;
{$ENDC }

		for i := 1 to length do begin (* deal with each input line segment *)

{$IFC AnimationEnabled AND AclipLSVbyRectangle }
			if animationFlag[AClipWithRectangleItem] then begin
				lineSegment(elements^^[i]).interactiveHighlight;
				waitForClick(animationFlag[AClipWithRectangleItem]);
				lineSegment(elements^^[i]).interactiveHighlight
			end;
{$ENDC }

			xp := lineSegment(elements^^[i]).p.getX;
			yp := lineSegment(elements^^[i]).p.getY;
			xq := lineSegment(elements^^[i]).q.getX;
			yq := lineSegment(elements^^[i]).q.getY;
			code(xp, yp, cp);
			code(xq, yq, cq);
			subClip(xp, yp, xq, yq, cp, cq);
			if (cp = []) then begin (* cp<>[] would guarantee that nothing were visible *)
				if cq <> [] then              (* not contented until cq=[] *)
					subClip(xq, yq, xp, yp, cq, cp);
				new(ls);
				failNil(Handle(ls));
				ls.init;
				ls.p.setX(xp);
				ls.p.setY(yp);
				ls.q.setX(xq);
				ls.q.setY(yq);
				resLSV.append(ls)
			end
		end;

{$IFC AnimationEnabled AND AclipLSVbyRectangle }
		if animationFlag[AClipWithRectangleItem] then
			clearAnimation;
{$ENDC }

		clipRect := resLSV
	end; (* clipRect *)

procedure initLineSegmentVector;
	var
		x: lineSegmentVector;
	begin
		new(x);
		failNil(Handle(x));
		prototypes[lineSegmentVectorType] := x;
	end;

end.
