(*	XYZ GeoBench (eXperimental geometrY Zurich):		*)
(*	An object oriented, interactive workbench for		*)
(*	experimental geometric computation.					*)
(*														*)
(*	 1999 by Peter Schorn, peter.schorn@acm.org		*)
(*	Culmannstrasse 77, CH-8006 Zrich, Switzerland		*)
(*	This software may be copied freely for educational	*)
(*	and research purposes only. For commercial use,		*)
(*	please contact the author.							*)
(*														*)
(*	Version 5.0.5: 30-Sep-1999							*)

unit nearestNeighborBoxes;
(* This unit was updated last at 26. August 1991.				*)
(*  by peter v. viebahn, universitt-gesamthochschule siegen	*)

{$SETC AANN = TRUE }

interface

uses
	pointVector;

const
	maxNeighbors = 100;
	maxSuccessors = powerOfTwo;   (* defined in unit sequence *)
	minSizeForAnimation = 300;
	maxPointsForAnimation = 1000;

type
	point2dArray = array[markIndex] of point2d;
	extArray = array[markIndex] of extended;
	realPointArray = array[markIndex] of realPoint;
	boxArray = array[markIndex] of obj;
(*     succArray declared in unit sequence  *)

	box = object(obj)
			corners: point2dArray;
			middle: point2d;
			size: extended;
			numOfPoints: seqIndex;
			estimate: extended;
			xTree, yTree: binTreeWithList;
			neighbors: priorityQueue;
			attractors: doubleList;
			procedure freeAll;
			override;
			procedure specialPrepareForFreeing;
			function getNumOfPoints: seqIndex;
			function getSize: extended;
{$IFC AnimationEnabled AND AANN }
			procedure drawNeighborRelation (b: box);
{$ENDC }
			procedure determineBordersOfBoxSimple (xL: listMuchLinked);
			procedure determineBordersOfBox (xL, yL: listMuchLinked; succNum: markIndex0; oldCornerLB, oldCornerRT, oldMiddle: point2d);
			procedure difMinMax (b: box; var min, max: extended);
			procedure addSuccAsNeighborAndAttractor (neighborBox: box; dMin: extended; animationIngenious: boolean);
			procedure createFromNewLists (xL, yL: listMuchLinked; neighborStructure: priorityQueue; succNum: markIndex0; oldCornerLB, oldCornerRT, oldMiddle: point2d);
			procedure createFromOldBox (oldBox: box; neighborStructure: priorityQueue; succNum: markIndex0);
			procedure deleteNeighborsWithLargeDistance (animationIngenious: boolean);
			procedure split (var newXLists, newYLists: succArray; var maxSucc: markIndex);
			procedure updateAllANE (var succBoxes: boxArray; animationIngenious: boolean);
		end;


	neighborId = object(obj)	(* identificates a neighbor of a box b and their minimal distance *)
			dMin: extended;
			neighbor: box;
			accordingAttractor: doubleElementH;	(* If b1 neighbor from b, than b attractor from b1; *)
													(* accordingAttractor is a pointer to b in b1.attractors. *)
			procedure putToSuccessors (var minDMax: extArray; var succ: boxArray; animationIngenious: boolean);
		end;

	attractorId = object(obj)	(* identificates an attractor of a box b *)
			attractor: box;
			accordingNeighbor: reference;		(* If b1 attractor from b, than b neighbor from b1; *)
												(* accordingNeighbor is the location of b in b1.neighbors. *)
		end;

implementation

function inverseCompareBoxesBySize (k, l: extended): relationObj;
	begin
		if k < l then
			inverseCompareBoxesBySize := greaterThanObj
		else if k > l then
			inverseCompareBoxesBySize := lessThanObj
		else
			inverseCompareBoxesBySize := equalObj
	end;

function boxGreaterThanBySize (b, c: box): relationObj;
	begin
		boxGreaterThanBySize := inverseCompareBoxesBySize(b.getSize, c.getSize)
(* A box b is smaller than a box c iff b.size is greater than c.size. *)
	end;

function inverseCompareNeighborIdByDistance (k, l: extended): relationObj;
	begin
		if k < l then
			inverseCompareNeighborIdByDistance := greaterThanObj
		else if k > l then
			inverseCompareNeighborIdByDistance := lessThanObj
		else
			inverseCompareNeighborIdByDistance := equalObj
	end;

function neighborIdGreaterThanByDistance (m, n: neighborId): relationObj;
	begin
		neighborIdGreaterThanByDistance := inverseCompareNeighborIdByDistance(m.dMin, n.dMin)
(* A neighborId m is smaller than a neighborId n iff m.dMin is greater than n.dMin. *)
	end;

procedure box.freeAll;
	var
		i: markIndex;
	begin
		if numOfPoints = 1 then begin
			if corners[1] <> nil then
				corners[1].freeAll;
			xTree.freeAll;
			yTree.freeAll
		end
		else begin
			for i := 1 to 4 do
				corners[i].freeAll;
			middle.freeAll
	(* xTree and yTree continue in maxSucc. *)
		end;
		priorityQueue(neighbors).freeAll;
		attractors.freeAll;
		shallowFree
	end;   { freeAll }

procedure box.specialPrepareForFreeing;
	var
		im: markIndex;
	begin
(* It is supposed that numOfPoints>1 and all points are the same ones. *)
		for im := 1 to 4 do
			corners[im].freeAll;
		middle.freeAll;
(*    corners[1] := point2d(treeElementH(listMuchLinked(xTree.list).first)^^.el); *)
		corners[1] := nil;
(* Down to the first elements all elements in the list are deleted and with them their ancestors too. *)
		xTree.specialPrepareForFreeing;
		yTree.specialPrepareForFreeing;
		numOfPoints := 1
	end;   { specialPrepareForFreeing }

function box.getNumOfPoints: seqIndex;
	begin
		getNumOfPoints := numOfPoints
	end;   { getNumOfPoints }

function box.getSize: extended;
	begin
		getSize := size
	end;   { getSize }

{$IFC AnimationEnabled AND AANN }

procedure box.drawNeighborRelation (b: box);
	var
		start, neighbor, mark: point2d;
		pType: objectType;
	begin
(* b is a neighbor of self. *)
(* If the middlepoint of self is smaller than b's one, draw the neighbor-relation above the middlepoints; *)
(* otherwise draw it below the middlepoints. *)
(* The direction of the box-neighbor-relation is typified by a thick point on the line near the neighbor. *)
		pType := point2d(treeElementH(listMuchLinked(xTree.list).first)^^.el).getType;
		start := point2d(newInitObj(pType));
		neighbor := point2d(start.duplicate);
		mark := point2d(start.duplicate);
		if middle.xLessThan(b.middle) then begin
			start.setX(middle.getX - 3);
			neighbor.setX(point2d(b.middle).getX + 3);
			start.setY(middle.getY + 3);
			neighbor.setY(point2d(b.middle).getY + 3)
		end
		else begin
			start.setX(middle.getX + 3);
			neighbor.setX(point2d(b.middle).getX - 3);
			start.setY(middle.getY - 3);
			neighbor.setY(point2d(b.middle).getY - 3)
		end;
		start.drawLine(neighbor);
		mark.setX(0.15 * start.getX + 0.85 * neighbor.getX);
		mark.setY(0.15 * start.getY + 0.85 * neighbor.getY);
		mark.interactiveOutput;
		start.free;
		neighbor.free;
		mark.free
	end;   { drawNeighborRelation }

{$ENDC }

procedure box.determineBordersOfBoxSimple (xL: listMuchLinked);
	begin
(* Self conatins only one point. *)
		corners[1] := point2d(xL.first^^.el.duplicate);
		corners[3] := corners[1];
(* All corners are the same point - the box itself. *)
		middle := corners[1];
		size := 0
	end; {  determineBordersOfBoxSimple }

procedure box.determineBordersOfBox (xL, yL: listMuchLinked; succNum: markIndex0; oldCornerLB, oldCornerRT, oldMiddle: point2d);
	var
		begX, endX, midthX, begY, endY, midthY, xDif, yDif: extended;
		rp: point2d;
		pType: objectType;
	begin
(* Self contains more than one point. *)
(* Destinate the real dimension of self. *)
		begX := point2d(xL.first^^.el).getX;
		endX := point2d(xL.last^^.el).getX;
		xDif := endX - begX;
		begY := point2d(yL.first^^.el).getY;
		endY := point2d(yL.last^^.el).getY;
		yDif := endY - begY;
(* Shrink the box so that it fits into the quarter of the ancestor box and *)
(* is the smallest cubical box, which contains the given points. *)
		if xDif > yDif then begin
			size := xDif;
			midthX := (begX + endX) / 2;
			midthY := (begY + endY) / 2;
			begY := midthY - size / 2;
			endY := begY + size;
			if succNum <> 0 then
		(* succNum indicates the quarter of the old box, whose borders hasn't to be overstepped by the successor-box. *)
				if (succNum = 1) | (succNum = 2) then begin
					if endY > oldMiddle.getY then begin
						begY := oldMiddle.getY - size;
						endY := oldMiddle.getY
					end
					else if begY < oldCornerRT.getY then begin
						begY := oldCornerRT.getY;
						endY := begY + size
					end;
					midthY := (begY + endY) / 2
				end
				else begin (* succNum=3 or succNum=4 *)
					if begY <= oldMiddle.getY then begin
						begY := oldMiddle.getY + 1;
						endY := begY + size
					end
					else if endY > oldCornerLB.getY then begin
						endY := oldCornerLB.getY;
						begY := endY - size
					end;
					midthY := (begY + endY) / 2
				end
		end
		else begin
			size := yDif;
			midthY := (begY + endY) / 2;
			midthX := (begX + endX) / 2;
			begX := midthX - size / 2;
			endX := begX + size;
			if succNum <> 0 then
				if (succNum = 1) | (succNum = 3) then begin
					if endX > oldMiddle.getX then begin
						begX := oldMiddle.getX - size;
						endX := oldMiddle.getX
					end
					else if begX < oldCornerLB.getX then begin
						begX := oldCornerLB.getX;
						endX := begX + size
					end;
					midthX := (begX + endX) / 2
				end
				else (* succNum=2 or succNum=4 *)
					begin
					if begX <= oldMiddle.getX then begin
						begX := oldMiddle.getX + 1;
						endX := begX + size;
					end
					else if endX > oldCornerRT.getX then begin
						endX := oldCornerRT.getX;
						begX := endX - size
					end;
					midthX := (begX + endX) / 2
				end
		end;
		pType := point2d(treeElementH(xL.first)^^.el).getType;
		rp := point2d(newInitObj(pType));
		rp.setX(begX);
		rp.setY(endY);
		corners[1] := rp;
		rp := point2d(newInitObj(pType));
		rp.setX(begX);
		rp.setY(begY);
		corners[2] := rp;
		rp := point2d(newInitObj(pType));
		rp.setX(endX);
		rp.setY(begY);
		corners[3] := rp;
		rp := point2d(newInitObj(pType));
		rp.setX(endX);
		rp.setY(endY);
		corners[4] := rp;
		rp := point2d(newInitObj(pType));
		rp.setX(midthX);
		rp.setY(midthY);
		middle := rp
	end;   { determineBordersOfBox }

procedure box.createFromNewLists (xL, yL: listMuchLinked; neighborStructure: priorityQueue; succNum: markIndex0; oldCornerLB, oldCornerRT, oldMiddle: point2d);
	begin
		numOfPoints := xL.getLength;
		if numOfPoints = 1 then
			self.determineBordersOfBoxSimple(xL)
		else
			self.determineBordersOfBox(xL, yL, succNum, oldCornerLB, oldCornerRT, oldMiddle);
		xTree := xL.embeddInBinaryTree;
		yTree := yL.embeddInBinaryTree;
		neighbors := neighborStructure;
		attractors := doubleList(newInitObj(doubleListType));
{$IFC AnimationEnabled AND AANN }
		if animationFlag[ANearestNeighborsItem] then
	(* The successor-box is drawn. *)
			corners[1].drawRectangle(corners[3]);
{$ENDC }
	end;   { createFromNewLists }

procedure box.createFromOldBox (oldBox: box; neighborStructure: priorityQueue; succNum: markIndex0);
	begin
		numOfPoints := listMuchLinked(binTreeWithList(oldBox.xTree).list).getLength;
		if numOfPoints = 1 then
			self.determineBordersOfBoxSimple(binTreeWithList(oldBox.xTree).list)
		else
			self.determineBordersOfBox(binTreeWithList(oldBox.xTree).list, binTreeWithList(oldBox.yTree).list, succNum, oldBox.corners[1], oldBox.corners[3], oldBox.middle);
		xTree := oldBox.xTree;
		yTree := oldBox.yTree;	(* The trees of the old box continue in the successor-box. *)
		neighbors := neighborStructure;
		attractors := doubleList(newInitObj(doubleListType));
{$IFC AnimationEnabled AND AANN }
		if animationFlag[ANearestNeighborsItem] then begin
	(* Redraw the cross of the old box. *)
			point2d(oldBox.corners[1]).drawCrossOfRectangle(point2d(oldBox.corners[3]), point2d(oldBox.middle));
	(* Draw the successor-box. *)
			corners[1].drawRectangle(corners[3]);
			waitForClick(animationFlag[ANearestNeighborsItem])
		end;
{$ENDC }
	end;   {createFromOldBox}

procedure box.difMinMax (b: box; var min, max: extended);
	var
		numVertices, numVerticesB, i, j, selfCorner, boxCorner: markIndex;
		dist: extended;
		selfP, boxP: point2d;
	begin
(* Compute the minimal and maximal distances between the boxes self and b, independent of *)
(* the points within these boxes. *)
		if numOfPoints = 1 then
			numVertices := 1
		else begin
			if size = 0 then
				numVertices := 1  (* Duplicate points *)
			else
				numVertices := 4
		end;
		if b.getNumOfPoints = 1 then
			numVerticesB := 1
		else begin
			if b.getSize = 0 then
				numVerticesB := 1  (* Duplicate points *)
			else
				numVerticesB := 4
		end;
		if (numVertices = 1) & (numVerticesB = 1) then begin
			min := sqrt(point2d(corners[1]).squaredDistanceX(b.corners[1]));
			max := min
		end
		else begin
(* First the distances between the corners are computed. *)
			min := 10000;
			max := 0;
			for i := 1 to numVertices do
				for j := 1 to numVerticesB do begin
					dist := sqrt(point2d(corners[i]).squaredDistanceX(b.corners[j]));
					if dist > max then
						max := dist;
					if dist < min then begin
						min := dist;
						selfCorner := i;
						boxCorner := j
					end;
				end;
(* In some cases the minimal distance is not correctly computed, it must be refined. *)
			selfP := point2d(corners[selfCorner]);
			boxP := point2d(b.corners[boxCorner]);
(* For the following constructions it is neccessary that the boxes don't intersect and are orthogonal. *)
			case selfCorner of
				1: 
					case boxCorner of
						1:
			(* One of the two boxes consists of one point. *)
							if numVertices = 1 then   (* self consists of one point. *)
								begin
								if (selfP.getY < boxP.getY) then
									min := boxP.getX - selfP.getX
								else if (selfP.getX > boxP.getX) then
									min := selfP.getY - boxP.getY
							end
							else   (* b consists of one point. *)
								begin
								if (selfP.getX < boxP.getX) then
									min := boxP.getY - selfP.getY
								else if (selfP.getY > boxP.getY) then
									min := selfP.getX - boxP.getX
							end;
						2: 
							if numVertices > 1 then begin
								if selfP.getX <> boxP.getX then
									min := boxP.getY - selfP.getY
							end
							else begin
								if selfP.getX > boxP.getX then
									min := boxP.getY - selfP.getY
								else if selfP.getY > boxP.getY then
									min := boxP.getX - selfP.getX
							end;
						3: 
							if selfP.getY > boxP.getY then
								min := selfP.getX - boxP.getX
							else if selfP.getX < boxP.getX then
								min := boxP.getY - selfP.getY;
						4: 
							if numVertices > 1 then begin
								if selfP.getY <> boxP.getY then
									min := selfP.getX - boxP.getX
							end
							else begin
								if selfP.getX < boxP.getX then
									min := selfP.getY - boxP.getY
								else if selfP.getY < boxP.getY then
									min := selfP.getX - boxP.getX
							end
					end;
				2: 
					case boxCorner of
						1: 
							if numVerticesB > 1 then begin
								if selfP.getX <> boxP.getX then
									min := selfP.getY - boxP.getY
							end
							else begin
								if selfP.getX < boxP.getX then
									min := selfP.getY - boxP.getY
								else if selfP.getY < boxP.getY then
									min := selfP.getX - boxP.getX
							end;
	(* 2 is not possible. *)
						3: 
							if selfP.getY <> boxP.getY then
								min := selfP.getX - boxP.getX;
						4: 
							if selfP.getY < boxP.getY then
								min := selfP.getX - boxP.getX
							else if selfP.getX < boxP.getX then
								min := selfP.getY - boxP.getY
					end;
				3: 
					case boxCorner of
						1: 
							if selfP.getY < boxP.getY then
								min := boxP.getX - selfP.getX
							else if selfP.getX > boxP.getX then
								min := selfP.getY - boxP.getY;
						2: 
							if selfP.getY <> boxP.getY then
								min := boxP.getX - selfP.getX;
	(* 3 is not possible. *)
						4: 
							if selfP.getX <> boxP.getX then
								min := selfP.getY - boxP.getY
					end;
				4: 
					case boxCorner of
						1: 
							if numVerticesB > 1 then begin
								if selfP.getY <> boxP.getY then
									min := boxP.getX - selfP.getX
							end
							else begin
								if selfP.getX > boxP.getX then
									min := boxP.getY - selfP.getY
								else if selfP.getY > boxP.getY then
									min := boxP.getX - selfP.getX
							end;
						2: 
							if selfP.getX > boxP.getX then
								min := boxP.getY - selfP.getY
							else if selfP.getY > boxP.getY then
								min := boxP.getX - selfP.getX;
						3: 
							if selfP.getX <> boxP.getX then
								min := boxP.getY - selfP.getY;
	(* 4 is not possible. *)
					end
			end(* case *)
		end   (* else *)
	end;   {difMinMax }

procedure box.addSuccAsNeighborAndAttractor (neighborBox: box; dMin: extended; animationIngenious: boolean);
	var
		nId: neighborId;
		aId: attractorId;
	begin
(* Add neighborBox to self.neighbors and add self to neighborBox.attractors. *)
		new(aId);
		failNil(Handle(aId));
		new(nId);
		failNil(Handle(nId));
		nId.dMin := dMin;
		nId.neighbor := neighborBox;
		nId.accordingAttractor := doubleList(neighborBox.attractors).appendWithPointer(aId);
		aId.attractor := self;
		aId.accordingNeighbor := priorityQueue(neighbors).insert(nId, neighborIdGreaterThanByDistance);
{$IFC AnimationEnabled AND AANN }
		if animationFlag[ANearestNeighborsItem] & animationIngenious then
	(* Draw the neighbor-relation [self - neighborBox] . *)
			self.drawNeighborRelation(neighborBox);
{$ENDC }
	end;   { addSuccAsNeighborAndAttractor }

procedure box.deleteNeighborsWithLargeDistance (animationIngenious: boolean);
	var
		neighbId: neighborId;
		neighborBox: box;
		attractorLocation: doubleElementH;
		neighborAttr: doubleList;
		nDMin: extended;
	begin
(* If a neighbor's distance is greater than self.estimate, the neighbor must be deleted. *)
		repeat
			neighbId := neighborId(neighbors.getObject(neighbors.minimum));
			nDMin := neighbId.dMin;
			if (nDMin > self.estimate) then begin
				neighborBox := neighbId.neighbor;
	(* Delete neighborBox from neighbors and delete self from attractors(neighborBox). *)
				attractorLocation := neighbId.accordingAttractor;
				neighborAttr := neighborBox.attractors;
				attractorId(attractorLocation^^.value).free;
				neighborAttr.delete(attractorLocation);
				neighbId.free;
				neighbors.delete((neighbors.minimum), neighborIdGreaterThanByDistance);
(* First element in the heap is deleted. *)
{$IFC AnimationEnabled AND AANN }
				if animationFlag[ANearestNeighborsItem] & animationIngenious then
		(* Redraw the neighbor-relation. *)
					self.drawNeighborRelation(neighborBox);
{$ENDC }
			end
		until not (nDMin > estimate) | (priorityQueue(neighbors).isEmpty)
	end;   { deleteNeighborsWithLargeDistance }

procedure box.split (var newXLists, newYLists: succArray; var maxSucc: markIndex);
	var
		px, py: treeElementH;  						 (* first element in x/y-list that doesn't belong to its smaller half *)
		xLeftSmaller, yLeftSmaller: boolean;  	 (* true if the smaller half is the left one *)
		numSuccPoints: array[1..maxSuccessors] of seqIndex0;   (* number of points from self, which fall into the ith successor *)
		im, unDetSucc: markIndex;   		(*  im / successor whose number of points will not be determinded directly. *)
		sumPointsInMinHalfs: seqIndex0; 	  	(* sum of numSuccPoints[i], i <> unDetSucc *)
		list, xList, yList: listMuchLinked;
		midthX, midthY: real;

	procedure getgXAndgY;
(* Traverse xList and yList to find the smaller half of each one. *)
(* px / py will point to the first element that doesn't belong to the smaller half of xList/yList. *)
		var
			xLeft, xRight, yLeft, yRight: treeElementH;
			pxLeft, pxRight, pyLeft, pyRight: point2d;
			emptyHalf: markIndex0;
			im: markIndex;
		begin
			emptyHalf := 0;
			xLeft := xList.first;
			xRight := xList.last;
			yLeft := yList.first;
			yRight := yList.last;
	(* first and last are different because self contains more than one point. *)
			pxLeft := point2d(xLeft^^.el);
			pxRight := point2d(xRight^^.el);
			pyLeft := point2d(yLeft^^.el);
			pyRight := point2d(yRight^^.el);
	(* pxLeft is the first element in xList, pxRight the last element in xList and analogous for yList. *)

	(* Look after empty halfs: *)
			if (pxLeft.getX >= midthX) & (pxRight.getX >= midthX) then begin
	(* The smaller x-half is empty *)
		(* cases 1b/2b *)
				px := xLeft;
				xLeftSmaller := true;
				emptyHalf := 1;
				numSuccPoints[1] := 0;
				numSuccPoints[3] := 0
			end
			else if (pxLeft.getX < midthX) & (pxRight.getX < midthX) then begin
	(* The smaller x-half is empty *)
		(* cases 4b/3b *)
				px := xRight;
				xLeftSmaller := false;
				emptyHalf := 2;
				numSuccPoints[2] := 0;
				numSuccPoints[4] := 0
			end
			else if (pyLeft.getY >= midthY) & (pyRight.getY >= midthY) then begin
	(* The smaller y-half is empty *)
		(* cases 1a/4a *)
				py := yLeft;
				yLeftSmaller := true;
				emptyHalf := 4;
				numSuccPoints[1] := 0;
				numSuccPoints[2] := 0
			end
			else if (pyLeft.getY < midthY) & (pyRight.getY < midthY) then begin
	(* The smaller y-half is empty *)
		(* cases 2a/3a *)
				py := yRight;
				yLeftSmaller := false;
				emptyHalf := 3;
				numSuccPoints[3] := 0;
				numSuccPoints[4] := 0
			end;

(* Traverse xList: *)
			if (emptyHalf <> 1) & (emptyHalf <> 2) then begin
				repeat
					if emptyHalf = 4 then begin
		(* all points.y >= midthY, cases 1a, 4a *)
						numSuccPoints[3] := numSuccPoints[3] + 1;
						numSuccPoints[4] := numSuccPoints[4] + 1
					end
					else if emptyHalf = 3 then begin
		(* all points.y < midthY, cases 2a, 3a *)
						numSuccPoints[1] := numSuccPoints[1] + 1;
						numSuccPoints[2] := numSuccPoints[2] + 1
					end
					else   (* emptyHalf = 0 *)
						begin
						if pxLeft.getY >= midthY then
							numSuccPoints[3] := numSuccPoints[3] + 1
						else
							numSuccPoints[1] := numSuccPoints[1] + 1;
						if pxRight.getY >= midthY then
							numSuccPoints[4] := numSuccPoints[4] + 1
						else
							numSuccPoints[2] := numSuccPoints[2] + 1
					end;
					xLeft := xLeft^^.right;
					xRight := xRight^^.left;
					pxLeft := point2d(xLeft^^.el);
					pxRight := point2d(xRight^^.el);
				until (pxLeft.getX >= midthX) | (pxRight.getX < midthX);

				if pxLeft.getX >= midthX then begin
		(* cases 1/2 *)
					px := xLeft;
					xLeftSmaller := true
				end
				else begin
		(* cases 3/4 *)
					px := xRight;
					xLeftSmaller := false
				end
			end;

(* Traverse yList: *)

(* Certain fields must be counted new because in the greater half the counting was breaked. *)
			if xLeftSmaller then begin
			(* cases 1/2 *)
				numSuccPoints[2] := 0;
				numSuccPoints[4] := 0
			end
			else begin
			(* cases 3/4 *)
				numSuccPoints[1] := 0;
				numSuccPoints[3] := 0
			end;

			if (emptyHalf <> 3) & (emptyHalf <> 4) then begin
				repeat
					if emptyHalf = 1 then begin
		(* all points.x >= midthX *)
						numSuccPoints[2] := numSuccPoints[2] + 1;
						numSuccPoints[4] := numSuccPoints[4] + 1
					end
					else if emptyHalf = 2 then begin
		(* all points.x < midthX *)
						numSuccPoints[1] := numSuccPoints[1] + 1;
						numSuccPoints[3] := numSuccPoints[3] + 1
					end
					else   (* emptyHalf = 0 *)
						begin
						if pyLeft.getX >= midthX then begin
							if xLeftSmaller then   (* In the other case numSuccPoints[2] was counted exactly. *)
				(* cases 1/2 *)
								numSuccPoints[2] := numSuccPoints[2] + 1
						end
						else if not xLeftSmaller then   (* In the other case numSuccPoints[1] was counted exactly. *)
			(* cases 3/4 *)
							numSuccPoints[1] := numSuccPoints[1] + 1;
						if pyRight.getX >= midthX then begin
							if xLeftSmaller then
				(* cases 1/2 *)
								numSuccPoints[4] := numSuccPoints[4] + 1
						end
						else if not xLeftSmaller then
			(* cases 3/4 *)
							numSuccPoints[3] := numSuccPoints[3] + 1
					end;
					yLeft := yLeft^^.right;
					yRight := yRight^^.left;
					pyLeft := point2d(yLeft^^.el);
					pyRight := point2d(yRight^^.el);
				until (pyLeft.getY >= midthY) | (pyRight.getY < midthY);

				if pyLeft.getY >= midthY then begin
					py := yLeft;
					yLeftSmaller := true
				end
				else begin
					py := yRight;
					yLeftSmaller := false
				end
			end;

			case emptyHalf of   (* unDetSucc wasn't counted exactly. *)
				0: 
					if xLeftSmaller then   (* cases 1/2 *)
						if yLeftSmaller then
							unDetSucc := 4
						else
							unDetSucc := 2
					else begin
						if yLeftSmaller then   (* cases 3/4 *)
							unDetSucc := 3
						else
							unDetSucc := 1
					end;
				1: 
					if yLeftSmaller then   (* cases 1b/2b *)
						unDetSucc := 4
					else
						unDetSucc := 2;
				2: 
					if yLeftSmaller then   (* cases 4b/3b *)
						unDetSucc := 3
					else
						unDetSucc := 1;
				3: 
					if xLeftSmaller then    (* cases 2a/3a *)
						unDetSucc := 2
					else
						unDetSucc := 1;
				4: 
					if xLeftSmaller then   (* cases 1a/4a *)
						unDetSucc := 4
					else
						unDetSucc := 3
			end;
			sumPointsInMinHalfs := 0;
			for im := 1 to maxSuccessors do
				if im <> unDetSucc then
					sumPointsInMinHalfs := sumPointsInMinHalfs + numSuccPoints[im];
			numSuccPoints[unDetSucc] := self.numOfPoints - sumPointsInMinHalfs;
		end;   { getgXAndgY }

	procedure markAndTraverseToTop (m: markIndex0; tr: treeElementH);
		var
			start: treeElementH;
		begin
			tr^^.mark := m;
			start := tr^^.second;
			while (tr^^.ancestor <> nil) & (treeElementH(tr^^.ancestor)^^.mark = 0) do begin
				tr := treeElementH(tr^^.ancestor);
				tr^^.mark := 1;
			end;
(* The ancestors of an element in one list are marked now; its ancestors in the other list must be marked too. *)
			tr := start;
			tr^^.mark := m;
			while (tr^^.ancestor <> nil) & (treeElementH(tr^^.ancestor)^^.mark = 0) do begin
				tr := treeElementH(tr^^.ancestor);
				tr^^.mark := 1;
			end
		end;   { markAndTraverseToTop }

	function destinateYHalf (t: treeElementH; xHalf: markIndex): markIndex;
		var
			rp: point2d;
		begin
			rp := point2d(t^^.el);
			if rp.getY >= midthY then
				if xHalf = 1 then
					destinateYHalf := 3
				else
					destinateYHalf := 4
			else begin
				if xHalf = 1 then
					destinateYHalf := 1
				else
					destinateYHalf := 2
			end
		end;   { destinateYHalf }

	function destinateXHalf (t: treeElementH; yHalf: markIndex): markIndex;
		var
			rp: point2d;
		begin
			rp := point2d(t^^.el);
			if rp.getX >= midthX then
				if yHalf = 1 then
					destinateXHalf := 2
				else
					destinateXHalf := 4
			else begin
				if yHalf = 1 then
					destinateXHalf := 1
				else
					destinateXHalf := 3
			end
		end;   { destinateXHalf }

	procedure markPointsSeparately;
		var
			m: markIndex0;
		begin
(* Retraverse the smaller half of the x-list. *)
(* Retraverse the smaller half of the y-list, but there treat only those points, which have not been treated *)
(* during traversing the x-list. *)
			if xLeftSmaller then begin
				while px <> xList.first do begin
					px := px^^.left;
					m := destinateYHalf(px, 1);
					markAndTraverseToTop(m, px)
				end;
				if yLeftSmaller then
					while py <> yList.first do    (* case 1 *)
						begin
						py := py^^.left;
						if py^^.mark = 0 then
							markAndTraverseToTop(2, py)
					end
				else  (* if not yLeftSmaller *)
					while py <> yList.last do   (* case 2 *)
						begin
						py := py^^.right;
						if py^^.mark = 0 then
							markAndTraverseToTop(4, py)
					end
			end
			else   (*if not xLeftSmaller *)
				begin
				while px <> xList.last do begin
					px := px^^.right;
					m := destinateYHalf(px, 2);
					markAndTraverseToTop(m, px)
				end;
				if yLeftSmaller then
					while py <> yList.first do   (* case 4 *)
						begin
						py := py^^.left;
						if py^^.mark = 0 then
							markAndTraverseToTop(1, py)
					end
				else  (* if not yLeftSmaller *)
					while py <> yList.last do   (* case 3 *)
						begin
						py := py^^.right;
						if py^^.mark = 0 then
							markAndTraverseToTop(3, py)
					end
			end
		end;   { markPointsSeparately }

	procedure markPointsNormal;
		var
			trList: treeElementH;
			is: seqIndex;
			firstHalf: seqIndex0;
			m: markIndex0;
		begin
(* Only the x-list is traversed (but completely). *)
			trList := xList.first;
			firstHalf := numSuccPoints[1] + numSuccPoints[3];
			if (maxSucc = 1) | (maxSucc = 3) then begin
				if firstHalf <> 0 then
					for is := 1 to firstHalf do begin
						m := destinateYHalf(trList, 1);
						if m <> maxSucc then
							markAndTraverseToTop(m, trList);
						trList := trList^^.right
					end;
				for is := firstHalf + 1 to numOfPoints do begin
					m := destinateYHalf(trList, 2);
					markAndTraverseToTop(m, trList);
					trList := trList^^.right
				end
			end
			else  (* maxSucc=2 or maxSucc=4 *)
				begin
				if firstHalf <> 0 then
					for is := 1 to firstHalf do begin
						m := destinateYHalf(trList, 1);
						markAndTraverseToTop(m, trList);
						trList := trList^^.right
					end;
		(* Second half of x: *)
				for is := firstHalf + 1 to numOfPoints do begin
					m := destinateYHalf(trList, 2);
					if m <> maxSucc then
						markAndTraverseToTop(m, trList);
					trList := trList^^.right
				end
			end
		end;   { markPointsNormal }

	procedure markPointsOfBMinusBMaxInTrees;
		var
			im: markIndex;
		begin
			maxSucc := unDetSucc;
			if (4 * sumPointsInMinHalfs) < self.numOfPoints then
				markPointsSeparately
			else begin
				for im := 1 to maxSuccessors do
					if numSuccPoints[im] > numSuccPoints[maxSucc] then
						maxSucc := im;
				if maxSucc = unDetSucc then
					markPointsSeparately
(*    markPointsNormal *)
				else
					markPointsNormal
			end
		end;   { markPointsOfBMinusBMaxInTrees }

	begin  { split }
		midthX := middle.getX;
		midthY := middle.getY;
		xList := listMuchLinked(xTree.list);
		yList := listMuchLinked(yTree.list);
		for im := 1 to maxSuccessors do
			numSuccPoints[im] := 0;
	(* The old lists must be split up into the successor-lists. *)
		getgXAndgY;   (* Get the smaller halfs of self wrt. x- and y-direction and the correct number of their points *)
		markPointsOfBMinusBMaxInTrees;
		for im := 1 to maxSuccessors do begin
			list := listMuchLinked(newInitObj(listMuchLinkedType));
			newXLists[im] := list;
			newYLists[im] := listMuchLinked(list.duplicate)
		end;
		xTree.traverse(newXLists);
		yTree.traverse(newYLists);
	end;   { split }

procedure box.updateAllANE (var succBoxes: boxArray; animationIngenious: boolean);
	var
		minDMax: extArray;
		dMin, dMax: extended;
		i, j: markIndex;

	procedure handleOldNeighbors (x: obj);
		var
			oldNeighborId: neighborId;
		begin
			oldNeighborId := neighborId(x);
{$IFC AnimationEnabled AND AANN }
			if animationFlag[ANearestNeighborsItem] & animationIngenious then begin
	(* Redraw the neighbor-relation [self - its neighbor]. *)
				self.drawNeighborRelation(oldNeighborId.neighbor);
				waitForClick(animationFlag[ANearestNeighborsItem])
			end;
{$ENDC }
	(* x becomes neighbor of every successor. *)
			oldNeighborId.putToSuccessors(minDMax, succBoxes, animationIngenious)
		end;    { handleOldNeighbors }

	procedure handleOldAttractors (x: obj);
		var
			attr: attractorId;
			accNeighb: neighborId;
			attrBox: box;
{$IFC AnimationEnabled AND AANN }
			oldNeighbor: box;
{$ENDC }
			miniDMax: extended;
			im: markIndex0;
			ref: reference;
		begin
			attr := attractorId(x);
			attrBox := box(attr.attractor);
	(* Delete self from attrBox.neighbors. *)
			ref := attr.accordingNeighbor;
			accNeighb := neighborId(priorityQueue(attrBox.neighbors).getObject(ref));
{$IFC AnimationEnabled AND AANN }
			if animationFlag[ANearestNeighborsItem] & animationIngenious then begin
		(* Redraw neighbor. *)
				oldNeighbor := box(neighborId(priorityQueue(attrBox.neighbors).getObject(ref)).neighbor);
				attrBox.drawNeighborRelation(oldNeighbor);
				waitForClick(animationFlag[ANearestNeighborsItem])
			end;
{$ENDC }
			accNeighb.free;
			priorityQueue(attrBox.neighbors).delete(ref, neighborIdGreaterThanByDistance);
			miniDMax := 10000;
	(* Add all successors to attrBox.neighbors. *)
			for im := 1 to maxSuccessors do
				if succBoxes[im] <> nil then begin
					attrBox.difMinMax(box(succBoxes[im]), dMin, dMax);
					if dMax < miniDMax then
						miniDMax := dMax;
					attrBox.addSuccAsNeighborAndAttractor(box(succBoxes[im]), dMin, animationIngenious)
				end;
	(* Update estimate (attrBox). *)
			if attrBox.getNumOfPoints = 1 then
				if miniDMax < attrBox.estimate then
					attrBox.estimate := miniDMax;
{$IFC AnimationEnabled AND AANN }
			if animationFlag[ANearestNeighborsItem] & animationIngenious then
				waitForClick(animationFlag[ANearestNeighborsItem]);
{$ENDC }
		end;   (* handleOldAttractors *)

	procedure updateAttractors (x: obj);
		begin
			box(attractorId(x).attractor).deleteNeighborsWithLargeDistance(animationIngenious)
		end;   { updateAttractors }

	begin   { updateAllANE }
(* The attractors, neighbors and estimates are updated where necessary. *)
		for i := 1 to maxSuccessors do
			minDMax[i] := 10000;
	(* The minimum of dMax(box i, neighbor-box j) becomes the estimate of box i. *)

	(* handleOldNeighbors *)
	(* Traverse neighbors of self; they become neighbors of the successors. *)
		if not (priorityQueue(self.neighbors).isEmpty) then
			priorityQueue(self.neighbors).forAll(handleOldNeighbors);

	(* chainNewSuccessors *)
	(* Every successor becomes neighbor of the other successors. *)
		for i := 1 to maxSuccessors do
			if succBoxes[i] <> nil then begin
				if i < maxSuccessors then
					for j := i + 1 to maxSuccessors do
						if succBoxes[j] <> nil then begin
							box(succBoxes[i]).difMinMax(box(succBoxes[j]), dMin, dMax);
			(* update neighbors of box(succBoxes[i]) *)
							if dMax < minDMax[i] then
								minDMax[i] := dMax;
							box(succBoxes[i]).addSuccAsNeighborAndAttractor(box(succBoxes[j]), dMin, animationIngenious);
			(* update neighbors of box(succBoxes[j]) *)
							if dMax < minDMax[j] then
								minDMax[j] := dMax;
							box(succBoxes[j]).addSuccAsNeighborAndAttractor(box(succBoxes[i]), dMin, animationIngenious);
						end;
	(* Update estimate[box(succBoxes[i])) *)
				if box(succBoxes[i]).getNumOfPoints = 1 then
					box(succBoxes[i]).estimate := minDMax[i]
				else
					box(succBoxes[i]).estimate := sqrt(2) * (box(succBoxes[i]).getSize)
			end;
{$IFC AnimationEnabled AND AANN }
		if animationFlag[ANearestNeighborsItem] & animationIngenious then
			waitForClick(animationFlag[ANearestNeighborsItem]);
{$ENDC }

	(* handleOldAttractors *)
		doubleList(self.attractors).forAll(handleOldAttractors);
{$IFC AnimationEnabled AND AANN }
		if animationFlag[ANearestNeighborsItem] & animationIngenious then
			waitForClick(animationFlag[ANearestNeighborsItem]);
{$ENDC }

	(* update neighbors of successors and attractors *)
		for i := 1 to maxSuccessors do
			if succBoxes[i] <> nil then
				box(succBoxes[i]).deleteNeighborsWithLargeDistance(animationIngenious);
		doubleList(self.attractors).forAll(updateAttractors);

{$IFC AnimationEnabled AND AANN }
		if animationFlag[ANearestNeighborsItem] then begin
			waitForClick(animationFlag[ANearestNeighborsItem]);
	(* Delete the borders of the old box (=currentBox). *)
	(* The cross has yet been deleted during building the maxSucc-box. *)
			PenSize(2, 2);
			point2d(self.corners[1]).drawRectangle(point2d(self.corners[3]));
			PenSize(1, 1)
		end;
{$ENDC }
	end;   { updateAllANE }

procedure neighborId.putToSuccessors (var minDMax: extArray; var succ: boxArray; animationIngenious: boolean);
	var
		i: markIndex;
		neighborAttr: doubleList;
		distMin, distMax: extended;
	begin
(* self.neighbor becomes neighbor of the i-th successor succ[i] box too. *)
		neighborAttr := neighbor.attractors;
		for i := 1 to maxSuccessors do
			if succ[i] <> nil then begin
				box(succ[i]).difMinMax(neighbor, distMin, distMax);
				if distMax < minDMax[i] then
					minDMax[i] := distMax;
				box(succ[i]).addSuccAsNeighborAndAttractor(neighbor, distMin, animationIngenious);
			end;
		attractorId(accordingAttractor^^.value).free;
		doubleList(neighborAttr).delete(accordingAttractor);
{$IFC AnimationEnabled AND AANN }
		if animationFlag[ANearestNeighborsItem] & animationIngenious then
			waitForClick(animationFlag[ANearestNeighborsItem])
{$ENDC }
	end;   { putToSuccessors }

function pointVector.nearestNeighborBoxes (ask: boolean): lineSegmentVector;
	var
		nearestNeighbors: lineSegmentVector;		(* vector of line segments - the result *)
		xList, yList: listMuchLinked;
		startBox, currentBox, succBox: box;
		allBoxes, neighborStructure: priorityQueue;
		succBoxes: boxArray;
		succXLists, succYLists: succArray;
		maxSucc: markIndex;
		i: markIndex;
		nLength: seqIndex;   (* nLength = length of neigbor-structure *)
		animationIngenious: boolean;
		ref: reference;
		x2QS: priorityQueueStatistics;  (* Explanation see in main-program. *)
		x2Type: priorityQueueTypeRange;
		statisticsX2: boolean;

	procedure getNeighbors (x: obj);
		var
			p: point2d;
			is: seqIndex;
			neighborBox: box;
		begin
(* For box x a list is created, in which all neighbors q are inserted as lineSegments from x to q. *)
			currentBox := box(x);
			if currentBox.getNumOfPoints > 1 then begin
	(* Possible if there are double points, then size=0 and the box hasn't yet been split. *)
	(* Every point gets itself as neighbor. *)
				for is := 1 to currentBox.getNumOfPoints do begin
					p := point2d(treeElementH(listMuchLinked(binTreeWithList(currentBox.xTree).list).first)^^.el);
					nearestNeighbors.append(p.createDirectedLineSegment(p))
				end;
				currentBox.specialPrepareForFreeing
			end
			else begin
				with priorityQueue(currentBox.neighbors) do
					neighborBox := neighborId(getObject(minimum)).neighbor;
				nearestNeighbors.append(point2d(currentBox.middle).createDirectedLineSegment(point2d(neighborBox.middle)))
			end
		end;   { getNeighbors }

	procedure getNeighborStructure;
		begin
(* The type of the neighbor structure is determined and an object created - the default type is a avl-tree. *)
			currentVectorLength := nLength;
			if ask then
				if statisticsX2 then begin
					currentPriorityQueueType := x2Type;
					new(x2QS);
					failNil(Handle(x2QS));
					x2QS.init;
					neighborStructure := x2QS
				end
				else
					neighborStructure := priorityQueue(newInitObj(x2Type))
			else (* ask = false *)
				begin
				currentDictionaryType := avlTreeType;
				neighborStructure := priorityQueue(newInitObj(dictPriorityQueueType))
			end
		end;   { getNeighborStructure }

	begin   { nearestNeighborBoxes }
(* self is from type pointVector and contains the start points. *)
(* instance variables: length: seqIndex0 *)
(*							elements:objArrayH *)
(* *)
(* Macintosh-coordinates are supposed [ position (0,0) is at left,top]. *)
(* *)
		currentVectorLength := length;
		nearestNeighbors := lineSegmentVector(newInitObj(lineSegmentVectorType));
		if length > 1 then begin
			nLength := length;    (* No box can have more boxes as neighbors than the number of points in self (=length). *)
			if nLength > maxNeighbors then
				nLength := maxNeighbors;
			if ask then begin   (* Build window to choose alternative structures. *)
				if not getXX(length, allBoxes, nLength, neighborStructure, statisticsX2, x2Type) then begin
					nearestNeighbors.free;
					nearestNeighborBoxes := nil;
					exit(nearestNeighborBoxes)   (* End of function nearestNeighborBoxes. *)
				end
		(* statisticsX2 and x2Type must be known because if ask=true then for every box its neighbor structure *)
		(* must be formed in the future [before createFromNewLists or createFromOldBox] out of the function getXX. *)
			end
			else begin
				allBoxes := priorityQueue(newInitObj(heapPriorityQueueType));
				currentDictionaryType := avlTreeType;
				neighborStructure := priorityQueue(newInitObj(dictPriorityQueueType))
			end;
			self.buildXAndYLists(xList, yList);
	(* xList / yList are now sorted by x / y -coordinate and the equivalent elements are chained. *)
			new(startBox);
			failNil(Handle(startBox));
			startBox.createFromNewLists(xList, yList, neighborStructure, 0, startBox.middle, startBox.middle, startBox.middle);
	(* Parameter 0 because there is not an ancestor-box, whose borders must not be overstepped by building the new boxes. *)
			animationIngenious := (round(startBox.size) > minSizeForAnimation) and (length <= maxPointsForAnimation);
	(* The neighbor relations are only drawn, if it is possible to follow up them. *)
			currentVectorLength := length;
			ref := allBoxes.insert(startBox, boxGreaterThanBySize);
	(* ref is never used *)
			while box(allBoxes.getObject(allBoxes.minimum)).getSize > 0 do begin
{$IFC AnimationEnabled }
				advanceStopWatchHands;
{$ENDC }
				currentBox := box(allBoxes.getObject(allBoxes.minimum));
{$IFC AnimationEnabled AND AANN }
				if animationFlag[ANearestNeighborsItem] then begin
			(* Draw the borders of the current box bold; draw its cross. *)
					point2d(currentBox.corners[1]).drawRectangle(point2d(currentBox.corners[3]));
					point2d(currentBox.corners[1]).drawCrossOfRectangle(point2d(currentBox.corners[3]), currentBox.middle);
					PenSize(2, 2);
					point2d(currentBox.corners[1]).drawRectangle(point2d(currentBox.corners[3]));
					PenSize(1, 1);
					waitForClick(animationFlag[ANearestNeighborsItem])
				end;
{$ENDC }
				currentBox.split(succXLists, succYLists, maxSucc);
		(* The points of the ith successor are stored in succXLists[i] and succYLists[i]. *)
		(* Initialize successor-boxes. *)
				for i := 1 to maxSuccessors do
					if succXLists[i].first <> nil then begin
						getNeighborStructure;
						new(succBox);
						failNil(Handle(succBox));
						succBox.createFromNewLists(succXLists[i], succYLists[i], neighborStructure, i, currentBox.corners[1], currentBox.corners[3], currentBox.middle);
						succBoxes[i] := succBox;
					end
					else begin
						succBoxes[i] := nil;
						succXLists[i].free;
						succYLists[i].free;
					end;
		(* The successor with maximal number of points is created directly from the old box. *)
				getNeighborStructure;
				new(succBox);
				failNil(Handle(succBox));
				succBox.createFromOldBox(currentBox, neighborStructure, maxSucc);
				succBoxes[maxSucc] := succBox;
		(* Now update the neighbor - and attractor - sets and the estimates. *)
				currentBox.updateAllANE(succBoxes, animationIngenious);

		(* Now the new successors can be inserted into allBoxes; the currentBox can be deleted from allBoxes. *)
				currentBox.freeAll;
				allBoxes.delete((allBoxes.minimum), boxGreaterThanBySize);
				for i := 1 to maxSuccessors do
					if succBoxes[i] <> nil then
						ref := allBoxes.insert(box(succBoxes[i]), boxGreaterThanBySize);

{$IFC AnimationEnabled AND AANN }
				if animationFlag[ANearestNeighborsItem] & animationIngenious then
					waitForClick(animationFlag[ANearestNeighborsItem])
{$ENDC }

			end;   (* while box.size>0 *)

	(* Now every box b consistsof one point! *)
	(* Create the result of function nearestNeighborBoxes. *)
			priorityQueue(allBoxes).forAll(getNeighbors);
			displayInfoString2('All Bx', allBoxes.infoString, 'Nbor lst Bx', priorityQueue(box(allBoxes.getObject(allBoxes.minimum)).neighbors).infoString);
			allBoxes.freeAll;   (* Every box consists of one point. *)

		end;  (* length>1 *)

		nearestNeighborBoxes := nearestNeighbors;
	end;  { nearestNeighborBoxes }

end.
