(*	XYZ GeoBench (eXperimental geometrY Zurich):		*)
(*	An object oriented, interactive workbench for		*)
(*	experimental geometric computation.					*)
(*														*)
(*	 1999 by Peter Schorn, peter.schorn@acm.org		*)
(*	Culmannstrasse 77, CH-8006 Zrich, Switzerland		*)
(*	This software may be copied freely for educational	*)
(*	and research purposes only. For commercial use,		*)
(*	please contact the author.							*)
(*														*)
(*	Version 5.0.5: 30-Sep-1999							*)

unit nearestNeighborInSector;

{$SETC AnearestNeighborInSector = TRUE }

interface

uses
	pointVector;

procedure initNearestNeighborInSector;

procedure exitNearestNeighborInSector;

implementation

type
	nnX = object(obj)	(* abstract class *)
			where: point2d;	(* intersection point of boundaries *)
			procedure process (xQueue: priorityQueue; yTable: dictionary);
			procedure free;
			override;
		end;

	nnXBend = object(nnX)
			rayReference: reference;
			newPartner, newOwner: point2d;
			procedure process (xQueue: priorityQueue; yTable: dictionary);
			override;
		end;

	nnXBoundaryIntersection = object(nnX)
			lower: reference;	(* into yTable,  lower boundary which particiaptes in intersection	 *)
			upper: reference;		(* into yTable, upper boundary which particiaptes in intersection *)
		end;

	nnXRayRay = object(nnXBoundaryIntersection) (* lower is ray, upper is ray *)
			procedure process (xQueue: priorityQueue; yTable: dictionary);
			override;
		end;

	nnXRayBisector = object(nnXBoundaryIntersection) (* lower is ray, upper is bisector *)
			procedure process (xQueue: priorityQueue; yTable: dictionary);
			override;
		end;

	nnXBisectorRay = object(nnXBoundaryIntersection) (* lower is bisector, upper is ray *)
			procedure process (xQueue: priorityQueue; yTable: dictionary);
			override;
		end;

	nnXBisectorBisector = object(nnXBoundaryIntersection) (* lower is bisector, upper is bisector *)
			procedure process (xQueue: priorityQueue; yTable: dictionary);
			override;
		end;


	boundary = (rayBoundary, bisectorBoundary);

	nnY = object(obj)	(* abstract class *)
			owner: point2d;		(* owner of interval bounded by this element and the successor of this element *)
			intersectUpper: reference;	(* maybe nil, into xQueue, where this element intersects with successor	 *)
			function getBoundaryType: boundary;	(* type of boundary *)
			function evaluate (current: point2d): extended;	(* value of boundary at 'p.getX' *)
			function intersect (neighbor: nnY; intersection: point2d): boolean;	(* check for intersection with 'neighbor' *)
{$IFC AnimationEnabled AND AnearestNeighborInSector }
			procedure display (current: point2d);
{$ENDC }
		end;

	nnYRay = object(nnY)
			lightSource: point2d;	(* where the ray emanates *)
			cosSlope: extended;	(* cos of slope of ray *)
			sinSlope: extended;	(* sin of slope of ray *)
			bend: reference; (* maybe nil, into xQueue to nnXBend event for this ray *)
			function getBoundaryType: boundary;
			override;
			function evaluate (current: point2d): extended;
			override;
			function intersect (neighbor: nnY; intersection: point2d): boolean;
			override;
{$IFC AnimationEnabled AND AnearestNeighborInSector }
			procedure display (current: point2d);
			override;
{$ENDC }
		end;

	nnYBisector = object(nnY)
			partner: point2d;		(* bisector of 'partner' and 'owner' *)
			function getBoundaryType: boundary;
			override;
			function evaluate (current: point2d): extended;
			override;
			function intersect (neighbor: nnY; intersection: point2d): boolean;
			override;
{$IFC AnimationEnabled AND AnearestNeighborInSector }
			procedure display (current: point2d);
			override;
{$ENDC }
		end;

var
	nnXPrototypes: array[boundary, boundary] of nnXBoundaryIntersection;

{$IFC AnimationEnabled AND AnearestNeighborInSector }

const
	dx = 35;

{$IFC NOT THINK_PASCAL }
{$PUSH }
{$PRAGMAC warn_unusedarg off }
{$ENDC }
procedure nnY.display (current: point2d);
	begin
		unimplementedMethod('nnY.display');
	end;
{$IFC NOT THINK_PASCAL }
{$POP }
{$ENDC }

procedure nnYRay.display (current: point2d);
	var
		a, b: point2d;
	begin
		a := point2d(current.duplicate);
		a.setX(max(current.getX - dx, lightSource.getX));
		b := point2d(current.duplicate);
		b.setX(current.getX + dx);
		drawLineSegment(a.getX, a.rayEvalX(lightSource, cosSlope, sinSlope), b.getX, b.rayEvalX(lightSource, cosSlope, sinSlope));
		b.free;
		a.free;
	end;

procedure nnYBisector.display (current: point2d);
	var
		a, b: point2d;
	begin
		a := point2d(current.duplicate);
		a.setX(current.getX - dx);
		b := point2d(current.duplicate);
		b.setX(current.getX + dx);
		drawLineSegment(a.getX, a.bisectorEvalX(partner, owner), b.getX, b.bisectorEvalX(partner, owner));
		b.free;
		a.free;
	end;

procedure showYTable (yTable: dictionary; current: point2d);
	procedure nnYDisplay (x: nnY);
		begin
			x.display(current);
		end;
	begin
		yTable.forAll(nnYDisplay, left);
	end;

{$ENDC }

procedure nnX.free;
	begin
		where.free;
		shallowFree;
	end;

function compareY (p: point2d; q: nnY): relationObj;
	begin
		compareY := relationObj(1 + sign(p.getY - q.evaluate(p)));
	end;

function compareX (a, b: nnX): relationObj;
	begin
		compareX := comparePointXY(a.where, b.where);
	end;

{$IFC NOT THINK_PASCAL }
{$PUSH }
{$PRAGMAC warn_unusedarg off }
{$ENDC }
procedure nnX.process (xQueue: priorityQueue; yTable: dictionary);
	begin
		unimplementedMethod('nnX.process');
	end;
{$IFC NOT THINK_PASCAL }
{$POP }
{$ENDC }

procedure removeIntersection (r: reference; xQueue: priorityQueue; yTable: dictionary);
	var
		intersection: reference;
	begin
		if r <> nil then begin
			intersection := nnY(yTable.getObject(r)).intersectUpper;
			if intersection <> nil then begin
{$IFC AnimationEnabled AND AnearestNeighborInSector }
				if animationFlag[ANearestNeighborsItem] then begin
					nnX(xQueue.getObject(intersection)).where.interactiveFlash;
					nnX(xQueue.getObject(intersection)).where.interactiveOutput;
				end;
{$ENDC }
				xQueue.getObject(intersection).free;
				xQueue.delete(intersection, compareX);
				nnY(yTable.getObject(r)).intersectUpper := nil;
			end;
		end;
	end;

procedure checkIntersection (lower, upper: reference; xQueue: priorityQueue; yTable: dictionary; current: point2d);
	var
		t: point2d;
		event: nnXBoundaryIntersection;
		lowerY, upperY: nnY;
	begin
		if (lower <> nil) & (upper <> nil) then begin
			lowerY := nnY(yTable.getObject(lower));
			upperY := nnY(yTable.getObject(upper));
			assert(lowerY.intersectUpper = nil, 'lowerY.intersectUpper=nil');
			t := point2d(current.duplicate);
			if lowerY.intersect(upperY, t) & not pointLessThanXY(t, current) then begin
{$IFC AnimationEnabled AND AnearestNeighborInSector }
				if animationFlag[ANearestNeighborsItem] then begin
					t.interactiveOutput;
					t.interactiveFlash;
				end;
{$ENDC }
				event := nnXBoundaryIntersection(nnXPrototypes[lowerY.getBoundaryType, upperY.getBoundaryType].duplicate);
				event.where := t;
				event.lower := lower;
				event.upper := upper;
				lowerY.intersectUpper := xQueue.insert(event, compareX);
			end
			else
				t.free;
		end;
	end;

procedure removeBend (r: reference; xQueue: priorityQueue);
	begin
		if r <> nil then begin
{$IFC AnimationEnabled AND AnearestNeighborInSector }
			if animationFlag[ANearestNeighborsItem] then begin
				nnX(xQueue.getObject(r)).where.interactiveFlash;
				nnX(xQueue.getObject(r)).where.interactiveHighlight;
			end;
{$ENDC }
			xQueue.getObject(r).free;
			xQueue.delete(r, compareX);
		end;
	end;

procedure scheduleBend (thisRay: nnYRay; bend: nnXBend; xQueue: priorityQueue);
	begin
		if thisRay.bend = nil then begin
			thisRay.bend := xQueue.insert(bend, compareX);
{$IFC AnimationEnabled AND AnearestNeighborInSector }
			if animationFlag[ANearestNeighborsItem] then begin
				bend.where.interactiveFlash;
				bend.where.interactiveHighlight;
			end;
{$ENDC }
		end
		else if pointLessThanXY(bend.where, nnXBend(xQueue.getObject(thisRay.bend)).where) then begin
			removeBend(thisRay.bend, xQueue);
{$IFC AnimationEnabled AND AnearestNeighborInSector }
			if animationFlag[ANearestNeighborsItem] then begin
				bend.where.interactiveFlash;
				bend.where.interactiveHighlight;
			end;
{$ENDC }
			thisRay.bend := xQueue.insert(bend, compareX);
		end
		else
			bend.free;
	end;

procedure testBendEvent (thisRay: nnYRay; rayReference: reference; current, newPartner, newOwner: point2d; xQueue: priorityQueue);
	var
		t: point2d;
		bend: nnXBend;
	begin
		t := point2d(current.duplicate);
		if thisRay.lightSource.intersectRayBisectorX(thisRay.cosSlope, thisRay.sinSlope, newPartner, newOwner, t) & not pointLessThanXY(t, current) then begin
			new(bend);
			failNil(Handle(bend));
			bend.where := t;
			bend.rayReference := rayReference;
			assert(newPartner.getY <= newOwner.getY, 'newPartner.getY<=newOwner.getY');
			bend.newPartner := newPartner;
			bend.newOwner := newOwner;
			scheduleBend(thisRay, bend, xQueue);
		end
		else
			t.free;
	end;

procedure nnXBend.process (xQueue: priorityQueue; yTable: dictionary);
	var
		oldRay: nnYRay;
		newBisector: nnYBisector;
		predRayReference: reference;
	begin
		oldRay := nnYRay(yTable.getObject(rayReference));
		new(newBisector);
		failNil(Handle(newBisector));
		newBisector.partner := newPartner;
		newBisector.owner := newOwner;
		newBisector.intersectUpper := nil;
		removeIntersection(rayReference, xQueue, yTable);
		predRayReference := yTable.next(rayReference, left);
		yTable.setObject(rayReference, newBisector);
		oldRay.free;
		removeIntersection(predRayReference, xQueue, yTable);
		checkIntersection(predRayReference, rayReference, xQueue, yTable, where);
		checkIntersection(rayReference, yTable.next(rayReference, right), xQueue, yTable, where);
	end;

procedure nnXRayRay.process (xQueue: priorityQueue; yTable: dictionary);
	var
		lowerRay, upperRay: nnYRay;
		intersection: point2d;
		predUpper: reference;
		bend: nnXBend;
		intersectsLower, intersectsUpper: boolean;
	begin
		lowerRay := nnYRay(yTable.getObject(lower));
		upperRay := nnYRay(yTable.getObject(upper));
		intersection := point2d(where.duplicate);
		new(bend);
		failNil(Handle(bend));
		bend.where := intersection;
		bend.newPartner := lowerRay.lightSource;
		bend.newOwner := upperRay.lightSource;
		intersectsLower := lowerRay.lightSource.intersectRayBisectorX(lowerRay.cosSlope, lowerRay.sinSlope, lowerRay.lightSource, upperRay.lightSource, intersection);
		if intersectsLower & not pointLessThanXY(intersection, where) then begin (* lower ray survives *)
			removeIntersection(upper, xQueue, yTable);
			lowerRay.intersectUpper := nil;
			bend.rayReference := lower;
			scheduleBend(lowerRay, bend, xQueue);
			lowerRay.owner := upperRay.owner;
			removeBend(upperRay.bend, xQueue);
			upperRay.free;
			yTable.delete(upper);
			checkIntersection(lower, yTable.next(lower, right), xQueue, yTable, where);
		end
		else begin
			intersectsUpper := upperRay.lightSource.intersectRayBisectorX(upperRay.cosSlope, upperRay.sinSlope, lowerRay.lightSource, upperRay.lightSource, intersection);
			if intersectsUpper & not pointLessThanXY(intersection, where) then begin (* upper ray survives *)
				bend.rayReference := upper;
				scheduleBend(upperRay, bend, xQueue);
				removeBend(lowerRay.bend, xQueue);
				lowerRay.free;
				yTable.delete(lower);
				predUpper := yTable.next(upper, left);
				removeIntersection(predUpper, xQueue, yTable);
				checkIntersection(predUpper, upper, xQueue, yTable, where);
			end
			else begin
				bend.free;
				assert(not (intersectsUpper & intersectsLower), 'not (intersectsUpper&intersectsLower)');
				if intersectsUpper then begin
					lowerRay.owner := upperRay.owner;
					lowerRay.intersectUpper := nil;
					removeIntersection(upper, xQueue, yTable);
					removeBend(upperRay.bend, xQueue);
					upperRay.free;
					yTable.delete(upper);
					checkIntersection(lower, yTable.next(lower, right), xQueue, yTable, where);
				end
				else begin
					removeBend(lowerRay.bend, xQueue);
					lowerRay.free;
					yTable.delete(lower);
					predUpper := yTable.next(upper, left);
					removeIntersection(predUpper, xQueue, yTable);
					checkIntersection(predUpper, upper, xQueue, yTable, where);
				end;
			end;
		end
	end;

procedure nnXRayBisector.process (xQueue: priorityQueue; yTable: dictionary);
	var
		lowerRay: nnYRay;
		upperBisector: nnYBisector;
	begin
		lowerRay := nnYRay(yTable.getObject(lower));
		upperBisector := nnYBisector(yTable.getObject(upper));
		lowerRay.owner := upperBisector.owner;
		lowerRay.intersectUpper := nil;
		testBendEvent(lowerRay, lower, where, lowerRay.lightSource, upperBisector.owner, xQueue);
		removeIntersection(upper, xQueue, yTable);
		upperBisector.free;
		yTable.delete(upper);
		checkIntersection(lower, yTable.next(lower, right), xQueue, yTable, where);
	end;

procedure nnXBisectorRay.process (xQueue: priorityQueue; yTable: dictionary);
	var
		lowerBisector: nnYBisector;
		upperRay: nnYRay;
		predUpper: reference;
	begin
		lowerBisector := nnYBisector(yTable.getObject(lower));
		upperRay := nnYRay(yTable.getObject(upper));
		testBendEvent(upperRay, upper, where, lowerBisector.partner, upperRay.lightSource, xQueue);
		lowerBisector.free;
		yTable.delete(lower);
		predUpper := yTable.next(upper, left);
		removeIntersection(predUpper, xQueue, yTable);
		checkIntersection(predUpper, upper, xQueue, yTable, where);
	end;

procedure nnXBisectorBisector.process (xQueue: priorityQueue; yTable: dictionary);
	var
		lowerBisector, upperBisector: nnYBisector;
		predLower: reference;
	begin
		lowerBisector := nnYBisector(yTable.getObject(lower));
		upperBisector := nnYBisector(yTable.getObject(upper));
		predLower := yTable.next(lower, left);
		removeIntersection(upper, xQueue, yTable);
		removeIntersection(predLower, xQueue, yTable);
		assert(lowerBisector.owner = upperBisector.partner, 'lowerBisector.owner=upperBisector.partner');
		upperBisector.partner := lowerBisector.partner;
		lowerBisector.free;
		yTable.delete(lower);
		checkIntersection(predLower, upper, xQueue, yTable, where);
		checkIntersection(upper, yTable.next(upper, right), xQueue, yTable, where);
	end;

function nnY.getBoundaryType: boundary;
	begin
		unimplementedMethod('nnY.getBoundaryType');
		getBoundaryType := rayBoundary;
	end;

{$IFC NOT THINK_PASCAL }
{$PUSH }
{$PRAGMAC warn_unusedarg off }
{$ENDC }
function nnY.evaluate (current: point2d): extended;
	begin
		unimplementedMethod('nnY.evaluate');
		evaluate := 0;
	end;

function nnY.intersect (neighbor: nnY; intersection: point2d): boolean;
	begin
		unimplementedMethod('nnY.intersect');
		intersect := false;
	end;
{$IFC NOT THINK_PASCAL }
{$POP }
{$ENDC }

function nnYRay.getBoundaryType: boundary;
	begin
		getBoundaryType := rayBoundary;
	end;

function nnYRay.evaluate (current: point2d): extended;
	begin
		evaluate := current.rayEvalX(lightSource, cosSlope, sinSlope);
	end;

function nnYRay.intersect (neighbor: nnY; intersection: point2d): boolean;
	var
		nr: nnYRay;
		nb: nnYBisector;
	begin
		if member(neighbor, nnYRay) then begin
			nr := nnYRay(neighbor);
			intersect := lightSource.intersectRayRayX(cosSlope, sinSlope, nr.lightSource, nr.cosSlope, nr.sinSlope, intersection);
		end
		else begin
			nb := nnYBisector(neighbor);
			intersect := lightSource.intersectRayBisectorX(cosSlope, sinSlope, nb.partner, nb.owner, intersection);
		end;
	end;

function nnYBisector.getBoundaryType: boundary;
	begin
		getBoundaryType := bisectorBoundary;
	end;

function nnYBisector.evaluate (current: point2d): extended;
	begin
		evaluate := current.bisectorEvalX(partner, owner);
	end;

function nnYBisector.intersect (neighbor: nnY; intersection: point2d): boolean;
	var
		nr: nnYRay;
		nb: nnYBisector;
	begin
		if member(neighbor, nnYRay) then begin
			nr := nnYRay(neighbor);
			intersect := nr.lightSource.intersectRayBisectorX(nr.cosSlope, nr.sinSlope, partner, owner, intersection);
		end
		else begin
			nb := nnYBisector(neighbor);
			intersect := partner.intersectBisectorBisectorX(owner, nb.partner, nb.owner, intersection);
		end;
	end;

procedure processNewPoint (newPoint: point2d; xQueue: priorityQueue; yTable: dictionary; nearestNeighbors: lineSegmentVector; cosLower, sinLower, cosUpper, sinUpper: extended);
	var
		yReference: reference;
		direction: directionType;
		found: boolean;
		lowerBoundary: reference;
		ls: lineSegment;
		lowerRay, upperRay: nnYRay;
		lowerRef, upperRef, predLowerRef: reference;
		neighbor: point2d;
	begin
		found := yTable.find(newPoint, compareY, yReference, direction);
		assert(not found, 'not found'); (* do not handle degeneracies for the time being *)
		if yReference = nil then
			neighbor := nil
		else begin
			if direction = left then
				lowerBoundary := yTable.next(yReference, left)
			else
				lowerBoundary := yReference;
			if lowerBoundary = nil then
				neighbor := nil
			else begin
				neighbor := nnY(yTable.getObject(lowerBoundary)).owner;
				if neighbor <> nil then begin
					new(ls);
					failNil(Handle(ls));
					ls.p := point2d(neighbor.duplicate);
					ls.q := point2d(newPoint.duplicate);
					nearestNeighbors.append(ls);
{$IFC AnimationEnabled AND AnearestNeighborInSector }
					if animationFlag[ANearestNeighborsItem] then begin
						ls.interactiveFlash;
						ls.interactiveHighlight;
					end;
{$ENDC }
				end;
			end;
		end;

		new(lowerRay);
		failNil(Handle(lowerRay));
		with lowerRay do begin
			owner := newPoint;
			intersectUpper := nil;
			lightSource := newPoint;
			cosSlope := cosLower;
			sinSlope := sinLower;
			bend := nil;
		end;
		new(upperRay);
		failNil(Handle(upperRay));
		with upperRay do begin
			owner := neighbor;
			intersectUpper := nil;
			lightSource := newPoint;
			cosSlope := cosUpper;
			sinSlope := sinUpper;
			bend := nil;
		end;
		if direction = left then begin
			upperRef := yTable.insert(upperRay, yReference, left);
			lowerRef := yTable.insert(lowerRay, upperRef, left);
		end
		else begin
			lowerRef := yTable.insert(lowerRay, yReference, right);
			upperRef := yTable.insert(upperRay, lowerRef, right);
		end;
		if neighbor <> nil then begin
			testBendEvent(lowerRay, lowerRef, newPoint, neighbor, newPoint, xQueue);
			testBendEvent(upperRay, upperRef, newPoint, newPoint, neighbor, xQueue);
		end;
		predLowerRef := yTable.next(lowerRef, left);
		removeIntersection(predLowerRef, xQueue, yTable);
		checkIntersection(predLowerRef, lowerRef, xQueue, yTable, newPoint);
		checkIntersection(upperRef, yTable.next(upperRef, right), xQueue, yTable, newPoint);
	end;

function pointVector.nearestNeighborInSector (ask: boolean; cosLower, sinLower, cosUpper, sinUpper: extended): lineSegmentVector;
	var
		yTable: dictionary;
		xQueue: priorityQueue;
		nearestNeighbors: lineSegmentVector;
		currentIndex: seqIndex1;
		currentPoint: point2d;
		minimum: reference;
		xEvent: nnX;
	begin
		currentVectorLength := length;
		new(nearestNeighbors);
		failNil(Handle(nearestNeighbors));
		nearestNeighbors.init;
		if length > 1 then begin

			if ask then begin
				if not getXY(2 * length, false, xQueue, 2 * length, yTable, false) then begin
					nearestNeighbors.free;
					nearestNeighborInSector := nil;
					exit(nearestNeighborInSector);
				end;
			end
			else begin
				currentVectorLength := 2 * length;
				xQueue := priorityQueue(newInitObj(heapPriorityQueueType));
				yTable := dictionary(newInitObj(avlTreeType));
			end;

			sortAll(pointLessThanXY);
			currentIndex := 1;
			repeat
{$IFC AnimationEnabled }
				if userAbort then begin
					nearestNeighbors.freeAll;
					xQueue.freeAll;
					yTable.freeAll;
					nearestNeighborInSector := nil;
					exit(nearestNeighborInSector);
				end;
{$ENDC }
				currentPoint := point2d(elements^^[currentIndex]);
				minimum := xQueue.minimum;
				if (minimum = nil) | pointLessThanXY(currentPoint, nnX(xQueue.getObject(minimum)).where) then begin
{$IFC AnimationEnabled AND AnearestNeighborInSector }
					if animationFlag[ANearestNeighborsItem] then begin
						drawThickVertical(currentPoint.getX);
						currentPoint.interactiveFlash;
						showYTable(yTable, currentPoint);
						waitForClick(animationFlag[ANearestNeighborsItem]);
						showYTable(yTable, currentPoint);
					end;
{$ENDC }
					processNewPoint(currentPoint, xQueue, yTable, nearestNeighbors, cosLower, sinLower, cosUpper, sinUpper);
{$IFC AnimationEnabled AND AnearestNeighborInSector }
					if animationFlag[ANearestNeighborsItem] then begin
						showYTable(yTable, currentPoint);
						waitForClick(animationFlag[ANearestNeighborsItem]);
						drawThickVertical(currentPoint.getX);
						showYTable(yTable, currentPoint);
					end;
{$ENDC }
					currentIndex := currentIndex + 1;
				end
				else begin
					xEvent := nnX(xQueue.getObject(minimum));
					xQueue.delete(minimum, compareX);
{$IFC AnimationEnabled AND AnearestNeighborInSector }
					if animationFlag[ANearestNeighborsItem] then begin
						drawThickVertical(xEvent.where.getX);
						xEvent.where.interactiveFlash;
						showYTable(yTable, xEvent.where);
						waitForClick(animationFlag[ANearestNeighborsItem]);
						showYTable(yTable, xEvent.where);
					end;
{$ENDC }
					xEvent.process(xQueue, yTable);
{$IFC AnimationEnabled AND AnearestNeighborInSector }
					if animationFlag[ANearestNeighborsItem] then begin
						showYTable(yTable, xEvent.where);
						waitForClick(animationFlag[ANearestNeighborsItem]);
						drawThickVertical(xEvent.where.getX);
						showYTable(yTable, xEvent.where);
					end;
{$ENDC }
					xEvent.free;
				end;
			until currentIndex > length;
			displayInfoString2('X', xQueue.infoString, 'Y', yTable.infoString);
			xQueue.freeAll;
			yTable.freeAll;
		end;
		if nearestNeighbors.length = 0 then begin
			nearestNeighbors.free;
			nearestNeighborInSector := nil;
		end
		else
			nearestNeighborInSector := nearestNeighbors;
	end;

procedure initNearestNeighborInSector;
	var
		rayRay: nnXRayRay;
		rayBisector: nnXRayBisector;
		bisectorRay: nnXBisectorRay;
		bisectorBisector: nnXBisectorBisector;
	begin
		new(rayRay);
		failNil(Handle(rayRay));
		nnXPrototypes[rayBoundary, rayBoundary] := rayRay;
		new(rayBisector);
		failNil(Handle(rayBisector));
		nnXPrototypes[rayBoundary, bisectorBoundary] := rayBisector;
		new(bisectorRay);
		failNil(Handle(bisectorRay));
		nnXPrototypes[bisectorBoundary, rayBoundary] := bisectorRay;
		new(bisectorBisector);
		failNil(Handle(bisectorBisector));
		nnXPrototypes[bisectorBoundary, bisectorBoundary] := bisectorBisector;
	end;

procedure exitNearestNeighborInSector;
	begin
		dispose(nnXPrototypes[rayBoundary, rayBoundary]);
		dispose(nnXPrototypes[rayBoundary, bisectorBoundary]);
		dispose(nnXPrototypes[bisectorBoundary, rayBoundary]);
		dispose(nnXPrototypes[bisectorBoundary, bisectorBoundary]);
	end;

end.
