(*	XYZ GeoBench (eXperimental geometrY Zurich):		*)
(*	An object oriented, interactive workbench for		*)
(*	experimental geometric computation.					*)
(*														*)
(*	 1999 by Peter Schorn, peter.schorn@acm.org		*)
(*	Culmannstrasse 77, CH-8006 Zrich, Switzerland		*)
(*	This software may be copied freely for educational	*)
(*	and research purposes only. For commercial use,		*)
(*	please contact the author.							*)
(*														*)
(*	Version 5.0.5: 30-Sep-1999							*)

unit planeSweepAllIntersect;

{$SETC ALineSegmentIntersection = TRUE }
{$SETC TestSignChange = TRUE }
{$SETC ConsistencyCheck = TRUE }

interface

uses
	lineSegmentVector;

type
	psAllIntersectEventType = (startEvent, intersectionEvent, endEvent);

	lsXelement = object(obj) (* abstract class for x-queue elements *)
			where: point2d;
			procedure processEvent (xQueue: priorityQueue; yTable, intersections: dictionary); (* process the event *)
			function getEventType: psAllIntersectEventType;
			procedure checkIntersection (lower, upper: reference; yTable: dictionary; xQueue: priorityQueue; intersections: dictionary);
		end;

function compareEvent (a, b: lsXelement): relationObj;

type
lsYelement = object(obj)
		l: lineSegment;
		endL: reference; (* into x-queue to its endEvent *)
		intersectLower, intersectUpper: reference; (* into x-queue to intersetionEvent with lower resp. upper neighbor *)
		procedure textOutput;
		override;
	end;

lsStartEvent = object(lsXelement)
		l: lineSegment;
		endL: reference; (* into x-queue to its endEvent *)
		procedure processEvent (xQueue: priorityQueue; yTable, intersections: dictionary);
		override;
		function getEventType: psAllIntersectEventType;
		override;
	end;

lsIntersectionEvent = object(lsXelement)
		lower, upper: reference; (* into y-table to intersecting line segments *)
		lowerSegment, upperSegment: lineSegment;
		intersectionObject: obj; (* 'nil' for pure intersection events, otherwise the right end point of the intersection segment *)
		procedure free;
		override;
		procedure processEvent (xQueue: priorityQueue; yTable, intersections: dictionary);
		override;
		function getEventType: psAllIntersectEventType;
		override;
	end;

lsEndEvent = object(lsXelement)
		remove: reference; (* into y-table to line segment to be removed *)
		procedure processEvent (xQueue: priorityQueue; yTable, intersections: dictionary);
		override;
		function getEventType: psAllIntersectEventType;
		override;
	end;

lineSegmentIntersection = object(obj)
		r, s: lineSegment;
		intersection: obj; (* point or lineSegment *)
	end;

implementation

{$IFC ConsistencyCheck AND ChecksEnabled }
var
	falseIntersections, consistencyFix: longInt;
{$ENDC }

function compareIntersection (a, b: lineSegmentIntersection): relationObj;
	begin
		if longInt(a.r) < longInt(b.r) then
			compareIntersection := lessThanObj
		else if longInt(a.r) > longInt(b.r) then
			compareIntersection := greaterThanObj
		else if longInt(a.s) < longInt(b.s) then
			compareIntersection := lessThanObj
		else if longInt(a.s) > longInt(b.s) then
			compareIntersection := greaterThanObj
		else
			compareIntersection := equalObj;
	end;

function compareEvent (a, b: lsXelement): relationObj;

	type
		epsRecord = record
				x, y: 0..1;
			end;

	function compareType (x, y: psAllIntersectEventType): relationObj;
		begin
			if x < y then
				compareType := lessThanObj
			else if x > y then
				compareType := greaterThanObj
			else
				compareType := equalObj;
		end;

	function comparePointXYEps (a: longIntPoint; aTrue: epsRecord; b: longIntPoint; bTrue: epsRecord): relationObj;
		begin
			a.x := 2 * a.x + aTrue.x;
			a.y := 2 * a.y + aTrue.y;
			b.x := 2 * b.x + bTrue.x;
			b.y := 2 * b.y + bTrue.y;
			comparePointXYEps := comparePointXY(a, b);
			a.x := a.x div 2;
			a.y := a.y div 2;
			b.x := b.x div 2;
			b.y := b.y div 2;
		end;

	function truePoint (x: lsXelement): epsRecord;

		function intersectTest (p, q, r, s: longIntPoint): epsRecord;
			var
				prq, psq, d: longInt;
				e: epsRecord;

			function ws (p, q, r: longIntPoint): longInt;
				begin
					ws := (p.x - q.x) * (q.y - r.y) - (p.y - q.y) * (q.x - r.x);
				end;

			begin
				prq := ws(p, r, q);
				psq := ws(p, s, q);
				if prq <> psq then begin
					d := prq - psq;
					e.x := ord(((prq * s.x - psq * r.x) mod d) <> 0);
					e.y := ord(((prq * s.y - psq * r.y) mod d) <> 0);
				end
				else begin
					e.x := 0;
					e.y := 0;
				end;
				intersectTest := e;
			end;

		begin
			if (x.getEventType = startEvent) | (x.getEventType = endEvent) then begin
				truePoint.x := 0;
				truePoint.y := 0;
			end
			else
				with lsIntersectionEvent(x) do
					truePoint := intersectTest(longIntPoint(lowerSegment.p), longIntPoint(lowerSegment.q), longIntPoint(upperSegment.p), longIntPoint(upperSegment.q));
		end;

	begin
		if member(a.where, longIntPoint) then
			case comparePointXYEps(longIntPoint(a.where), truePoint(a), longIntPoint(b.where), truePoint(b)) of
				lessThanObj: 
					compareEvent := lessThanObj;
				equalObj: 
					compareEvent := compareType(a.getEventType, b.getEventType);
				greaterThanObj: 
					compareEvent := greaterThanObj;
			end
		else
			case comparePointXY(a.where, b.where) of
				lessThanObj: 
					compareEvent := lessThanObj;
				equalObj: 
					compareEvent := compareType(a.getEventType, b.getEventType);
				greaterThanObj: 
					compareEvent := greaterThanObj;
			end
	end;

function compareLineSegment (a, b: lsYelement): relationObj;
	begin
		with b.l do
			case a.l.p.whichSideSign(p, q) of
				-1: 
					compareLineSegment := greaterThanObj;
				0: 
					case a.l.q.whichSideSign(p, q) of
						-1: 
							compareLineSegment := lessThanObj;
						0: 
							compareLineSegment := comparePointXY(a.l.p, p);
						+1: 
							compareLineSegment := greaterThanObj;
					end;
				+1: 
					compareLineSegment := lessThanObj;
			end;
	end;

procedure clearIntersection (pred, succ: reference; yTable: dictionary; xQueue: priorityQueue);
	var
		neighborsIntersect: reference;
		intersectionEvent: lsIntersectionEvent;
	begin
		if (pred <> nil) & (succ <> nil) & (lsYelement(yTable.getObject(succ)).intersectLower <> nil) then begin
			neighborsIntersect := lsYelement(yTable.getObject(succ)).intersectLower;
{$IFC AnimationEnabled AND ALineSegmentIntersection }
			if animationFlag[ALineSegmentIntersectionItem] then begin
				lsXelement(xQueue.getObject(neighborsIntersect)).where.interactiveHighlight;
				lsXelement(xQueue.getObject(neighborsIntersect)).where.interactiveOutput;
			end;
{$ENDC }
			intersectionEvent := lsIntersectionEvent(xQueue.getObject(neighborsIntersect));
			intersectionEvent.free;
			xQueue.delete(neighborsIntersect, compareEvent);
			assert(neighborsIntersect = lsYelement(yTable.getObject(pred)).intersectUpper, 'lower.upper = upper.lower');
			lsYelement(yTable.getObject(succ)).intersectLower := nil;
			lsYelement(yTable.getObject(pred)).intersectUpper := nil;
		end;
	end;

procedure lsYelement.textOutput;
	begin
		l.textOutput;
		write(longInt(endL) : 1, ':', longInt(intersectLower) : 1, ':', longInt(intersectUpper) : 1);
	end;

{$IFC NOT THINK_PASCAL }
{$PUSH }
{$PRAGMAC warn_unusedarg off }
{$ENDC }
procedure lsXelement.processEvent (xQueue: priorityQueue; yTable, intersections: dictionary);
	begin
		unimplementedMethod('lsXelement.processEvent');
	end;
{$IFC NOT THINK_PASCAL }
{$POP }
{$ENDC }

function lsXelement.getEventType: psAllIntersectEventType;
	begin
		unimplementedMethod('lsXelement.getEventType');
		getEventType := startEvent;
	end;

procedure makeIntersection (var newIntersection: lineSegmentIntersection; lower, upper: lineSegment; i: obj);
	begin
		new(newIntersection);
		failNil(Handle(newIntersection));
		newIntersection.init;
		if longInt(lower) < longInt(upper) then begin
			newIntersection.r := lower;
			newIntersection.s := upper;
		end
		else begin
			newIntersection.r := upper;
			newIntersection.s := lower;
		end;
		newIntersection.intersection := i;
	end;

procedure lsXelement.checkIntersection (lower, upper: reference; yTable: dictionary; xQueue: priorityQueue; intersections: dictionary);
	var
		intersection: lsIntersectionEvent;
		l, u, r: lineSegment;
		intersect: lineSegmentIntersection;
		ref: reference;
{$IFC AnimationEnabled AND ALineSegmentIntersection }
		v: homogenVector;
{$ENDC }
	begin
		if (lower <> nil) & (upper <> nil) then begin
			l := lsYelement(yTable.getObject(lower)).l;
			u := lsYelement(yTable.getObject(upper)).l;
			r := lineSegment(l.duplicate);
{$IFC AnimationEnabled AND ALineSegmentIntersection }
			if animationFlag[ALineSegmentIntersectionItem] then begin
				v := newHomogenVector(3);
				v.append(l);
				v.append(u);
			end;
{$ENDC }
			if l.intersectLineSegmentX(u, r) then begin
				makeIntersection(intersect, l, u, nil);
				if not intersections.member(intersect, compareIntersection) then begin
					new(intersection);
					failNil(Handle(intersection));
					intersection.where := point2d(r.p.duplicate);
					intersection.lower := lower;
					intersection.upper := upper;
					intersection.lowerSegment := l;
					intersection.upperSegment := u;
					with r do
						if p.equalPoint(q) then
							intersection.intersectionObject := p.duplicate
						else
							intersection.intersectionObject := r.duplicate;
{$IFC AnimationEnabled AND ALineSegmentIntersection }
					if animationFlag[ALineSegmentIntersectionItem] then begin
						v.append(intersection.intersectionObject);
						v.interactiveFlash;
						v.free;
						intersection.where.interactiveHighlight;
						intersection.where.interactiveOutput;
						waitForClick(animationFlag[ALineSegmentIntersectionItem]);
					end;
{$ENDC }
					ref := xQueue.insert(intersection, compareEvent);
					assert((lsYelement(yTable.getObject(lower)).intersectUpper = nil) & (lsYelement(yTable.getObject(upper)).intersectLower = nil), 'lower.upper=nil & upper.lower=nil');
					lsYelement(yTable.getObject(lower)).intersectUpper := ref;
					lsYelement(yTable.getObject(upper)).intersectLower := ref;
				end;
				intersect.free;
			end
{$IFC AnimationEnabled AND ALineSegmentIntersection }
			else if animationFlag[ALineSegmentIntersectionItem] then begin
				v.interactiveFlash;
				waitForClick(animationFlag[ALineSegmentIntersectionItem]);
				v.free;
			end;
{$ELSEC }
			;
{$ENDC }
			r.free;
		end;
	end;

procedure lsStartEvent.processEvent (xQueue: priorityQueue; yTable, intersections: dictionary);
	var
		whereRef, pred, succ: reference;
		yEl: lsYelement;
		found: boolean;

{$IFC AnimationEnabled AND ALineSegmentIntersection AND TestSignChange AND ChecksEnabled }
		mp: markedRealPoint;

	function dictionarySignChange (yTable: dictionary; yEl: lsYelement): integer;
		var
			cursor: reference;
			previousRelation, newRelation: relationObj;
			signChange: integer;
		begin
			cursor := yTable.extreme(left);
			previousRelation := compareLineSegment(yEl, lsYelement(yTable.getObject(cursor)));
			signChange := 0;
			cursor := yTable.next(cursor, right);
			repeat
				newRelation := compareLineSegment(yEl, lsYelement(yTable.getObject(cursor)));
				if newRelation <> previousRelation then begin
					previousRelation := newRelation;
					signChange := signChange + 1;
				end;
				cursor := yTable.next(cursor, right);
			until cursor = nil;
			dictionarySignChange := signChange;
		end;
{$ENDC }

{$IFC ConsistencyCheck AND ChecksEnabled }
		var
			sucSuc, predPred: reference;

	function intersectY (a, b: lsYelement): boolean;
		var
			ll: lineSegment;
		begin
			ll := lineSegment(a.l.duplicate);
			intersectY := a.l.intersectLineSegmentX(b.l, ll);
			ll.free;
		end;

	function upwardInconsistent (r: reference; yTable: dictionary): boolean;
		var
			sucR, sucSucR: reference;
			a, ap, app: lsYelement;
		begin
			a := lsYelement(yTable.getObject(r));
			sucR := yTable.next(r, right);
			if sucR <> nil then
				ap := lsYelement(yTable.getObject(sucR));
			sucSucR := yTable.next(sucR, right);
			if sucSucR <> nil then
				app := lsYelement(yTable.getObject(sucSucR));
			upwardInconsistent := (sucR <> nil) & (sucSucR <> nil) & (compareLineSegment(a, ap) = lessThanObj) & (compareLineSegment(a, app) = greaterThanObj) & intersectY(a, ap) & not intersectY(a, app);
		end;

	function downwardInconsistent (r: reference; yTable: dictionary): boolean;
		var
			predR, predPredR: reference;
			a, am, amm: lsYelement;
		begin
			a := lsYelement(yTable.getObject(r));
			predR := yTable.next(r, left);
			if predR <> nil then
				am := lsYelement(yTable.getObject(predR));
			predPredR := yTable.next(predR, right);
			if predPredR <> nil then
				amm := lsYelement(yTable.getObject(predPredR));
			downwardInconsistent := (predR <> nil) & (predPredR <> nil) & (compareLineSegment(a, am) = greaterThanObj) & (compareLineSegment(a, amm) = lessThanObj) & intersectY(a, am) & not intersectY(a, amm);
		end;
{$ENDC }

	begin
		new(yEl);
		failNil(Handle(yEl));
		yEl.l := l;
		yEl.endL := endL;
		yEl.intersectLower := nil;
		yEl.intersectUpper := nil;
{$IFC AnimationEnabled AND ALineSegmentIntersection }
		if animationFlag[ALineSegmentIntersectionItem] then begin
{$IFC TestSignChange AND ChecksEnabled }
			if yTable.numberOfElements > 1 then begin
				new(mp);
				failNil(Handle(mp));
				mp.init;
				mp.x := l.p.getX;
				mp.y := l.p.getY;
				mp.mark.value := stringOf(' ->', dictionarySignChange(yTable, yEl) : 1);
			end
			else
				mp := nil;
{$ENDC }
			l.interactiveFlash;
			l.interactiveHighlight;
{$IFC TestSignChange AND ChecksEnabled }
			if mp <> nil then
				mp.interactiveOutput;
{$ENDC }
			waitForClick(animationFlag[ALineSegmentIntersectionItem]);
{$IFC TestSignChange AND ChecksEnabled }
			if mp <> nil then begin
				mp.interactiveOutput;
				mp.free;
			end;
{$ENDC }
			l.interactiveHighlight;
		end;
{$ENDC }
{$IFC TestSignChange AND ChecksEnabled }
		assert((yTable.numberOfElements <= 1) | (dictionarySignChange(yTable, yEl) <= 2), 'signChange <= 2');
{$ENDC }
		whereRef := yTable.insertObj(yEl, compareLineSegment, found);
{$IFC ConsistencyCheck AND ChecksEnabled }
		if upwardInconsistent(whereRef, yTable) then begin
			sucSuc := yTable.next(yTable.next(whereRef, right), right);
			yTable.delete(whereRef);
			whereRef := yTable.insert(yEl, sucSuc, right);
			consistencyFix := consistencyFix + 1;
		end;
		if downwardInconsistent(whereRef, yTable) then begin
			predPred := yTable.next(yTable.next(whereRef, left), left);
			yTable.delete(whereRef);
			whereRef := yTable.insert(yEl, predPred, left);
			consistencyFix := consistencyFix + 1;
		end;
		assert(not upwardInconsistent(whereRef, yTable), 'upward consistency');
		assert(not downwardInconsistent(whereRef, yTable), 'downward consistency');
		if upwardInconsistent(whereRef, yTable) | downwardInconsistent(whereRef, yTable) then
			l.interactiveFlash;
{$ENDC }
		lsEndEvent(xQueue.getObject(endL)).remove := whereRef;
		pred := yTable.next(whereRef, left);
		succ := yTable.next(whereRef, right);
		clearIntersection(pred, succ, yTable, xQueue);
		checkIntersection(pred, whereRef, yTable, xQueue, intersections);
		checkIntersection(whereRef, succ, yTable, xQueue, intersections);
	end;

function lsStartEvent.getEventType: psAllIntersectEventType;
	begin
		getEventType := startEvent;
	end;

procedure lsIntersectionEvent.free;
	begin
		where.free;
		intersectionObject.free;
		shallowFree;
	end;

procedure lsIntersectionEvent.processEvent (xQueue: priorityQueue; yTable, intersections: dictionary);
	var
		intersection: lineSegmentIntersection;
		pred, succ: reference;
		lowerY, upperY: lsYelement;
{$IFC ConsistencyCheck AND ChecksEnabled }
		lowerFirst, lowerSecond, upperFirst, upperSecond: signType;
{$ENDC }
{$IFC AnimationEnabled AND ALineSegmentIntersection }
		segments: list;
{$ENDC }
	begin
		lowerY := lsYelement(yTable.getObject(lower));
		upperY := lsYelement(yTable.getObject(upper));
		makeIntersection(intersection, lowerY.l, upperY.l, intersectionObject.duplicate);
		lowerY.intersectUpper := nil;
		upperY.intersectLower := nil;
		intersections.insertNewObj(intersection, compareIntersection);
{$IFC AnimationEnabled AND ALineSegmentIntersection }
		if animationFlag[ALineSegmentIntersectionItem] then begin
			new(segments);
			failNil(Handle(segments));
			segments.init;
			segments.append(intersection.r);
			segments.append(intersection.s);
			segments.append(intersection.intersection);
			segments.interactiveFlash;
			segments.interactiveHighlight;
			waitForClick(animationFlag[ALineSegmentIntersectionItem]);
			segments.interactiveHighlight;
			segments.free;
		end;
{$ENDC }
{$IFC ConsistencyCheck AND ChecksEnabled }
		lowerFirst := lowerY.l.p.whichSideSign(upperY.l.p, upperY.l.q);
		lowerSecond := lowerY.l.q.whichSideSign(upperY.l.p, upperY.l.q);
		upperFirst := upperY.l.p.whichSideSign(lowerY.l.p, lowerY.l.q);
		upperSecond := upperY.l.q.whichSideSign(lowerY.l.p, lowerY.l.q);
		if (lowerFirst > 0) & (upperFirst < 0) & ((lowerSecond <= 0) & (upperSecond >= 0)) | (lowerFirst = 0) & (lowerSecond <= 0) | (upperFirst = 0) & (upperSecond >= 0) then begin
			pred := yTable.next(lower, left);
			succ := yTable.next(upper, right);
			clearIntersection(pred, lower, yTable, xQueue);
			clearIntersection(upper, succ, yTable, xQueue);
			yTable.swap(lower, upper);
			checkIntersection(pred, upper, yTable, xQueue, intersections);
			checkIntersection(lower, succ, yTable, xQueue, intersections);
		end
		else
			falseIntersections := falseIntersections + 1;
{$ELSEC }
		pred := yTable.next(lower, left);
		succ := yTable.next(upper, right);
		clearIntersection(pred, lower, yTable, xQueue);
		clearIntersection(upper, succ, yTable, xQueue);
		yTable.swap(lower, upper);
		checkIntersection(pred, upper, yTable, xQueue, intersections);
		checkIntersection(lower, succ, yTable, xQueue, intersections);
{$ENDC }
	end;

function lsIntersectionEvent.getEventType: psAllIntersectEventType;
	begin
		getEventType := intersectionEvent;
	end;

procedure lsEndEvent.processEvent (xQueue: priorityQueue; yTable, intersections: dictionary);
	var
		pred, succ: reference;
{$IFC AnimationEnabled AND ALineSegmentIntersection }
		l: lineSegment;
{$ENDC }
	begin
{$IFC AnimationEnabled AND ALineSegmentIntersection }
		if animationFlag[ALineSegmentIntersectionItem] then begin
			l := lsYelement(yTable.getObject(remove)).l;
			l.interactiveFlash;
			l.interactiveHighlight;
			waitForClick(animationFlag[ALineSegmentIntersectionItem]);
			l.interactiveHighlight;
			l.interactiveOutput;
		end;
{$ENDC }
		pred := yTable.next(remove, left);
		succ := yTable.next(remove, right);
		assert(((pred = nil) | (lsYelement(yTable.getObject(pred)).intersectUpper = nil)) & ((succ = nil) | (lsYelement(yTable.getObject(succ)).intersectLower = nil)), '(pred=nil | pred.upper=nil) & (succ=nil | succ.lower=nil');
		yTable.getObject(remove).free;
		yTable.delete(remove);
		checkIntersection(pred, succ, yTable, xQueue, intersections);
	end;

function lsEndEvent.getEventType: psAllIntersectEventType;
	begin
		getEventType := endEvent;
	end;

function lineSegmentVector.planeSweepAllIntersect (ask: boolean): vector;
	var
		xQueue: priorityQueue;
		yTable, intersections: dictionary;
		minRef: reference;
		minObj: lsXelement;
{$IFC AnimationEnabled AND ALineSegmentIntersection }
		order: list;
{$ENDC }

	procedure initXqueue;
		var
			sEvent: lsStartEvent;
			eEvent: lsEndEvent;
			dummy: reference;
			l: lineSegment;
			n: seqIndex0;
			i: seqIndex1;
			heap: heapPriorityQueue;
		begin
			n := length;
			i := 1;
			while i <= n do
				with lineSegment(elements^^[i]) do
					if p.equalPoint(q) then begin
						swap(i, n);
						n := n - 1;
					end
					else
						i := i + 1;
			if member(xQueue, heapPriorityQueue) then begin
				heap := heapPriorityQueue(xQueue);
				for i := 1 to n do begin
					l := lineSegment(elements^^[i]);
					orderPointsGreaterThanXY(l.p, l.q);
					new(eEvent);
					failNil(Handle(eEvent));
					eEvent.where := l.q;
					eEvent.remove := nil; (* will be set at insertion time *)
					new(sEvent);
					failNil(Handle(sEvent));
					sEvent.where := l.p;
					sEvent.l := l;
					sEvent.endL := heap.simpleAppend(eEvent);
					dummy := heap.simpleAppend(sEvent);
				end;
				heap.heapify(compareEvent);
			end
			else
				for i := 1 to n do begin
					l := lineSegment(elements^^[i]);
					orderPointsGreaterThanXY(l.p, l.q);
					new(eEvent);
					failNil(Handle(eEvent));
					eEvent.where := l.q;
					eEvent.remove := nil; (* will be set at insertion time *)
					new(sEvent);
					failNil(Handle(sEvent));
					sEvent.where := l.p;
					sEvent.l := l;
					sEvent.endL := xQueue.insert(eEvent, compareEvent);
					dummy := xQueue.insert(sEvent, compareEvent);
				end;
		end;

	function extractSolution (d: dictionary): vector;
		var
			v: vector;
		procedure addElement (x: lineSegmentIntersection);
			begin
				v.append(x.intersection);
			end;
		begin
			if d.isEmpty then
				extractSolution := nil
			else begin
				v := newVector(d.numberOfElements);
				d.forAll(addElement, left);
				extractSolution := v;
			end;
		end;

{$IFC AnimationEnabled }
	procedure deleteIntersection (x: lineSegmentIntersection);
		begin
			x.intersection.free;
		end;
{$ENDC }

{$IFC AnimationEnabled AND ALineSegmentIntersection }
	function displayOrder (yTable: dictionary): list;
		var
			rank: integer;
			result: list;

		procedure show (y: lsYelement);
			var
				mp: markedRealPoint;
			begin
				rank := rank + 1;
				new(mp);
				failNil(Handle(mp));
				mp.init;
				with y.l.p do begin
					mp.x := getX;
					mp.y := getY;
				end;
				mp.mark.value := stringOf(rank : 1);
				result.append(mp);
			end;

		begin
			rank := 0;
			result := newEmptyList;
			yTable.forAll(show, left);
			displayOrder := result;
		end;
{$ENDC }

	function nonZero (t: Str255; n: longInt): Str255;
		begin
			if n <> 0 then
				nonZero := stringOf(t, n : 1)
			else
				nonZero := '';
		end;

	function separate (t: Str255): Str255;
		begin
			if t = '' then
				separate := ''
			else
				separate := Concat(t, ' ');
		end;

	begin
		if ask then begin
			if not getXY(3 * length, false, xQueue, length, yTable, false) then begin
				planeSweepAllIntersect := nil;
				exit(planeSweepAllIntersect);
			end;
		end
		else begin
			currentVectorLength := 3 * length;
			xQueue := priorityQueue(newInitObj(heapPriorityQueueType));
			yTable := dictionary(newInitObj(avlTreeType));
		end;
		intersections := dictionary(newInitObj(avlTreeType));
{$IFC AnimationEnabled AND ALineSegmentIntersection }
		if animationFlag[ALineSegmentIntersectionItem] then
			interactiveHighlight;
{$ENDC }
		initXqueue;
{$IFC ConsistencyCheck AND ChecksEnabled }
		falseIntersections := 0;
		consistencyFix := 0;
{$ENDC }
		while not xQueue.isEmpty do begin
{$IFC AnimationEnabled }
			if userAbort then begin
				intersections.forAll(deleteIntersection, left);
				intersections.freeAll;
				xQueue.freeAll;
				yTable.freeAll;
				planeSweepAllIntersect := nil;
				exit(planeSweepAllIntersect);
			end;
{$ENDC }
			minRef := xQueue.minimum;
			minObj := lsXelement(xQueue.getObject(minRef));
			xQueue.delete(minRef, compareEvent);
{$IFC AnimationEnabled AND ALineSegmentIntersection }
			if animationFlag[ALineSegmentIntersectionItem] then begin
				drawThickVertical(minObj.where.getX);
				order := displayOrder(yTable);
				order.interactiveOutput;
			end
			else
				order := nil;
{$ENDC }
			minObj.processEvent(xQueue, yTable, intersections);
{$IFC AnimationEnabled AND ALineSegmentIntersection }
			if animationFlag[ALineSegmentIntersectionItem] then begin
				drawThickVertical(minObj.where.getX);
				order.interactiveOutput;
				order.freeAll;
			end
			else if order <> nil then
				order.freeAll;
{$ENDC }
			minObj.free;
		end;
		planeSweepAllIntersect := extractSolution(intersections);
		displayInfoString2('X', xQueue.infoString, 'Y', yTable.infoString);
{$IFC ConsistencyCheck AND ChecksEnabled }
		displayInfoString(Concat(separate(stringOf(nonZero('False X=', falseIntersections), nonZero(' Fix=', consistencyFix))), infoString));
		if infoString = ' ' then
			infoString := '';
{$ENDC }
		intersections.freeAll;
		xQueue.free;
		yTable.free;
	end;

function lineSegmentVector.memberType (o: obj): boolean;
	begin
		memberType := member(o, lineSegmentVector) & memberBaseType(o);
	end;

function lineSegmentVector.equal (o: obj): boolean;
	begin
		equal := (o <> nil) & member(o, lineSegmentVector) & inherited equal(o);
	end;

function lineSegmentVector.getType: objectType;
	begin
		getType := lineSegmentVectorType;
	end;

function lineSegmentVector.getTextType: Str255;
	begin
		getTextType := concat('Line Segment Vector', getTextBaseType);
	end;

function lineSegmentVector.simpleAllIntersect: vector;
	var
		i, j: seqIndex;
		result: vector;
		intersections: list;
		l, r: lineSegment;
	begin
{$IFC AnimationEnabled AND ALineSegmentIntersection }
		if animationFlag[ALineSegmentIntersectionItem] then
			interactiveHighlight;
{$ENDC }
		intersections := newEmptyList;
		l := lineSegment(elements^^[1].duplicate);
		for i := 1 to length do begin
			r := lineSegment(elements^^[i]);
{$IFC AnimationEnabled AND ALineSegmentIntersection }
			if animationFlag[ALineSegmentIntersectionItem] then
				r.interactiveHighlight;
{$ENDC }
			for j := i + 1 to length do begin

{$IFC AnimationEnabled }
				if userAbort then begin
					l.free;
					intersections.freeAll;
					simpleAllIntersect := nil;
					exit(simpleAllIntersect);
				end;
{$ENDC }

{$IFC AnimationEnabled AND ALineSegmentIntersection }
				if animationFlag[ALineSegmentIntersectionItem] then begin
					elements^^[j].interactiveHighlight;
					waitForClick(animationFlag[ALineSegmentIntersectionItem]);
				end;
{$ENDC }
				if lineSegment(elements^^[j]).intersectLineSegmentX(r, l) then
					with l do
						if p.equalPoint(q) then begin
{$IFC AnimationEnabled AND ALineSegmentIntersection }
							if animationFlag[ALineSegmentIntersectionItem] then begin
								p.interactiveHighlight;
								waitForClick(animationFlag[ALineSegmentIntersectionItem]);
							end;
{$ENDC }
							intersections.append(p.duplicate)
						end
						else begin
{$IFC AnimationEnabled AND ALineSegmentIntersection }
							if animationFlag[ALineSegmentIntersectionItem] then begin
								l.interactiveHighlight;
								waitForClick(animationFlag[ALineSegmentIntersectionItem]);
							end;
{$ENDC }
							intersections.append(l.duplicate);
						end;
{$IFC AnimationEnabled AND ALineSegmentIntersection }
				if animationFlag[ALineSegmentIntersectionItem] then
					elements^^[j].interactiveHighlight;
{$ENDC }
			end;
{$IFC AnimationEnabled AND ALineSegmentIntersection }
			if animationFlag[ALineSegmentIntersectionItem] then
				r.interactiveHighlight;
{$ENDC }
		end;
		new(result);
		failNil(Handle(result));
		result.initFromList(intersections);
{$IFC AnimationEnabled AND ALineSegmentIntersection }
		if animationFlag[ALineSegmentIntersectionItem] then
			result.interactiveHighlight;
{$ENDC }
		l.free;
		intersections.free;
		if result.length > 0 then
			simpleAllIntersect := result
		else begin
			result.free;
			simpleAllIntersect := nil;
		end;
	end;

function lineSegmentVector.eliminateZeroLengthSegments: lineSegmentVector;
	var
		oldLength: seqIndex0;
		n: seqIndex0;
		i: seqIndex1;
	begin
		i := 1;
		n := length;
		while i <= n do begin
{$IFC AnimationEnabled }
			if userAbort then begin
				eliminateZeroLengthSegments := nil;
				exit(eliminateZeroLengthSegments);
			end;
{$ENDC }
			with lineSegment(elements^^[i]) do
				if p.equalPoint(q) then begin
					swap(i, n);
					n := n - 1;
				end
				else
					i := i + 1;
		end;
		if length = n then
			eliminateZeroLengthSegments := nil
		else begin
			oldLength := length;
			length := n;
			eliminateZeroLengthSegments := lineSegmentVector(duplicateAll);
			length := oldLength;
		end;
	end;

end.
