(*	XYZ GeoBench (eXperimental geometrY Zurich):		*)
(*	An object oriented, interactive workbench for		*)
(*	experimental geometric computation.					*)
(*														*)
(*	 1999 by Peter Schorn, peter.schorn@acm.org		*)
(*	Culmannstrasse 77, CH-8006 Zrich, Switzerland		*)
(*	This software may be copied freely for educational	*)
(*	and research purposes only. For commercial use,		*)
(*	please contact the author.							*)
(*														*)
(*	Version 5.0.5: 30-Sep-1999							*)

unit pointVector;

interface

uses
	testGraph, unionFind;

const
	maxPolys = 15;
	maxPolysM1 = maxPolys - 1;
	givenByNumber = 0;
	givenByProbability = 1;
	pointIsOutSide = 0;
	pointIsOnBoundary = 1;
	pointIsInside = 2;

type
	polRange = 0..maxPolysM1;
	polSet = set of polRange;
	clusterObjectType = (circleObject, convexPolygonObject);

	pointVector = object(homogenVector)
			function equal (o: obj): boolean;
			override;
			procedure describe;
			override;
			procedure execute (item: integer);
			override;
			function memberType (o: obj): boolean;
			override;
			function getType: objectType;
			override;
			function getTextType: Str255;
			override;
			function getBoundingObject: obj;
			override;
			function textFileOutputId: Str255;
			override;
			function convertToType (t: pointTypeRange): obj;
			override;
			function convertToConvexPolygon: obj;
			override;
			function convertToPolygon: obj;
			override;
			procedure textFileOutput;
			override;
			function angle (i, j, k: seqIndex): real;
(* begin Marco Ciot *)
(* RNG = Relative Neighborhood Graph *)
(* GG = Gabriel Graph *)

	(* simple algorithms: *)
			function ggSimple: modifiableUndirectedGraph;
			function rngFalse: modifiableUndirectedGraph;

	(* efficient algorithms: *)
			function ggMatulaSokal: modifiableUndirectedGraph;
			function rngSupowit: modifiableUndirectedGraph;

(* end Marco Ciot *)

{$IFC Tutorial }
			function gravityCenter: point2d;
{$ENDC }

{$IFC OptionalCommand }
			function distanceDistribution (resolution: longInt; width, height: extended): histogram;
{$ENDC }

(* Convex hull *)
			procedure convexHull (eliminate, verbose: boolean; var hullLength, eliminated: seqIndex0);
(* Computes in place the convex hull of the given points using the Graham scan. The elements			*)
(* from 1 to 'hullLength' constitute a convex polygon  whereas the elements from 'hullLength+1' to	*)
(* 'self.length' are the inner points. If 'eliminate = true', a heuristic is used which finds a			*)
(*	rectangle guaranteed to be inside the hull and removes the points prior to scan. 'eliminated'		*)
(*	gives the	number of points eliminated by this process.													*)

			function convexHullGrahamScan: convexPolygon;
(*	Computes the convex hull of the given points using the method 'convexHull'.							*)

			function convexHullDivideAndConquer: convexPolygon;
(* Computes the convex hull using the divide and conquer algorithm. *)


(* Closest pair *)
			function closestPairHeuristic: lineSegment;
(* Computes the closest pair using the heuristic method which sweeps only in x direction. The		*)
(* resulting 'lineSegment' is formed by the closest pair and might have zero length. The order of	*)
(* the points might change. *)

			function closestPairPoints (ask: boolean): lineSegment;
(* Computes the closest pair using the plane sweep method. The resulting 'lineSegment' is formed	*)
(* by the closest pair and might have zero length. The order of the points might change. *)

(* inserted by JM *)
			function closestPairModified (ask: boolean): lineSegment;
(* Computes the closest pair using the plane sweep method. The resulting 'lineSegment' is formed	*)
(* by the closest pair and might have zero length. The order of the points might change. *)
(* This is a modified version of the standard plane-sweep-algorithm. It checks less Points then the standard one *)
(* and was developed by Frank Bartling (Uni-Gh-Siegen).   Implemented by JM.  *)

			procedure closestPairInitialGuessing (stepsize: stepIndex; var squaredDelta: extended; var cl, cr: point2d);
(* Computes the closest pair among the neightbours in the x-Queue.  This routine can be used to get a fast and small *)
(* estimation of the closest pair, so that the delta-surrounding for the regular plane-sweep is fairly small  (This     *)
(* leeds to an O(n)-algorithm in the average case).              *)

(*  --- *)

			function closestPairProbabilistic (ask: boolean): lineSegment;
(* Computes the closest pair using Rabin's probabilistic algorithm. The resulting 'lineSegment' is	*)
(* formed by the closest pair and might have zero length. The order of the points might change.		*)
(* 'ask = TRUE' asks the user for the size of the hash table, otherwise the size is determined by		*)
(* the algorithm. This is useful for demonstration purposes where giving a large hash table causes	*)
(* few or no collisions.																							*)

			function closestPairNewHeuristic: lineSegment;
(* Computes the closest pair using the heuristic method which sweeps simultaneously in x and y		*)
(* direction. The resulting 'lineSegment' is formed by the closest pair and might have zero length.	*)
(* The order of the points might change. *)

			function closestPairDivideAndConquer: lineSegment;
(* Computes the closest pair using the divide and conquer method. Program adapated from 			*)
(* 'Algorithms from P to NP, Volume I - Design & Efficiency', by B. Moret and H. Shapiro.			*)

			function closestPairN2: lineSegment;
(* Computes the closest pair using the trivial quadratic method. The resulting 'lineSegment' is		*)
(* formed by the closest pair and might have zero length. The order of the points does not change.	*)


(* Nearest neighbors *)
			function nearestNeighborSweep (kind: neighborKind; ask: boolean): lineSegmentVector;
(* Computes all nearest neighbors using the plane sweep method according to 'kind'. The result is a	*)
(* homogen vector of line segments where the start point is a given point and the end point  is the	*)
(* start points nearest neighbor to the (left, right, all). A point without a nearest neighbor is		*)
(* eliminated. The order of the points might change.															*)

			function nearestNeighborProjection (kind: neighborKind; useY: boolean): lineSegmentVector;
(* Same as 'nearestNeighborSweep' but the projection method is used. If 'useY' = true, then the		*)
(*	projection method is used simultaneously in both directions.												*)

			function nearestNeighborInSector (ask: boolean; cosLower, sinLower, cosUpper, sinUpper: extended): lineSegmentVector;
(* Computes all nearest neighbors to the left in the sector that is bounded by the rays with slope	*)
(* 'cosLower / sinLower' and 'cosUpper / sinUpper' using the plane sweep method. The result is a	*)
(* homogen vector of line segments where the start point is a given point and the end point is the	*)
(* start points nearest neighbor to the left in the specified sector. The left most point gets itself as	*)
(* nearest neighbor to the left. The order of the points might change.										*)


(* inserted by PV *)

			function nearestNeighborBoxes (ask: boolean): lineSegmentVector;
(* Computes all nearest neighbors (not only one of them) using Vaidyas method with boxes. The result is	*)
(* a homogen vector of lists. Every list contains the nearest neighbors of a given point as line segments.	*)
(* The start point is the given point and the end points are the start points nearest neighbors. The order	*)
(*	of the points might change.																							*)

(* *)

(* Voronoi Diagram and DelaunayTriangulation *)
			function voronoiDiagramSweepLine (ask: boolean): voronoiDiagram;
(* Computes the Voronoi diagram using the plane sweep method. The order of the points does not		*)
(* change. *)

			function voronoiDiagramSweepLineSeidel (ask: boolean): voronoiDiagram;
(*	Computes the Voronoi diagram using the 'topological' plane sweep method. The order of the		*)
(*	points does not change. By Silke Schweifer, Universitt Mnster, 27.3.93.							*)

			function gridTriangulation (tolerance, gridSizeFactor: extended): obj;
(*	Computes the Delaunay triangulation using a uniform grid, forming the triangles in a circular		*)
(*	fashion.																											*)
(*	tolerance :	if maximum distance of two points in a gridcell  is within tolerance one point is		*)
(*					discarded.																						*)
(*	gridSizeFactor:  a factor which changes the size of one gridcell.  The bigger a gridcell the			*)
(*						more points have to  be checked while searching,  the less cells must be			*)
(*						visited to find an optimal point for the Delaunay triangle.							*)
(*	By Richard Dittrich, Aug-Sep 93.																			*)


(* Euclidean minimum spanning tree *)
			function EMST (k: seqIndex; var uf: unionFindSimple): spanningTree;
(* Computes a Euclidean minimum spanning tree of the point set. The order of the points does not	*)
(* change. If k=1, an ordinary EMST is computed. If k>1, a forest with k components is computed	*)
(*	and 'uf' contains the corresponding union find structure.													*)

			function cluster (k: seqIndex; kind: clusterObjectType): homogenVector;
(*	Computes a vector of 'k' objects of type 'kind' which overlap the given points.						*)

			function computeSmallestSizeCluster: homogenVector;
(*	Finds the smallest 'k' > 2 which yields a clustering (with the method 'pointVector.cluster') with	*)
(*	'k' non-intersecting circles.																					*)

			function computeSmallestAreaCluster (intersectionAllowed: boolean; maxComponents: seqIndex): homogenVector;
(*	Finds the clustering (with the method 'pointVector.cluster') which has the smallest area and		*)
(*	which consists of non-intersecting circles iff 'intersectionAllowed' = true.							*)
(*	'k'  'maxComponents'.																							*)

(* Traveling salesman heuristics *)
			function TSPEMST: polygon2d;
(* Computes a tour of the traveling salesman through the given points with the Euclidean minimum	*)
(* spanning tree heuristic. The order of the points does not change. 										*)

			function TSPNN: polygon2d;
(* Computes a tour of the traveling salesman through the given points with the nearest neighbor		*)
(* heuristic. The order of the points does not change. 														*)

			function TSPConvexHull: polygon2d;
(* Computes a tour of the traveling salesman through the given points with the convex hull			*)
(* heuristic. The order of the points does not change. 														*)

(* Traveling salesman optimal *)
			function TSPSimple: polygon2d;
(* Computes the optimal traveling salesman tour through the given points with depth first search.	*)
(* All (n-1)! / 2 permutations of the given points are considered.	The order of the	points does not	*)
(* change.																											*)
(* Inserted by Adrian Bruengger.																				*)

{$IFC false }
			function TSPSimpleBranch (initTour: boolean): polygon2d;
(* Computes the optimal traveling salesman tour through the given points with depth first search.	*)
(* If initTour is set to 'true', an inital tour (convex hull heuristic) is generated. A branch occurs	*)
(* whenever the considered subtour is longer than the current shortest tour.							*)
(* The order of the points does not change.																		*)

			function TSPConsConvHullOrder: polygon2d;
(* Computes the optimal traveling salesman tour through the given points with depth first search.	*)
(* The convex Hull heuristic is used as an initial value. The order of the points does not change.		*)
(* The points on the convex hull are visited in the same order as they appear on the convex hull. 	*)

			function TSPNoInters: polygon2d;
(* Computes the optimal traveling salesman tour through the given points with depth first search.	*)
(* The convex Hull heuristic is used as an initial value. The order of the points does not change.		*)
(* Only subtours without intersections are generated.														*)

			function TSP2NNBranchAndBound: polygon2d;
(* Computes the optimal traveling salesman tour through the given points with a branch and bound	*)
(* method. As an estimated (conservativ!) distance for the rest of a current tour, it is supposed	*)
(* that each remaining point is connected with his two nearest neighbors.								*)
{$ENDC }

			function TSP1TreeBranchAndBound: polygon2d;
(* Computes the optimal traveling salesman tour through the given points using the subgradient 		*)
(* optimization for the 1-Tree relaxation. In a first step, an initial lower bound is computed.			*)
(* Based on that value, the length of the tour is estimated and for some edges their inclusion or		*)
(* exclusion in the optimal tour is proven. On the other edges, a branch and bound algorithm is		*)
(* performed. If no tour with length < estimation is found, the process starts again with an			*)
(* increased estimation.																							*)
(* Inserted by Adrian Bruengger.																				*)

{$IFC false }
			function TSPDiagram: QDPicture;
{$ENDC }

(* Minimal area disk *)
			function minimalDisk (verbose: boolean): circle;
(* Computes the smallest circle which contains the given points using a randomized algorithm.		*)
(* The order of the points might change. 																		*)

			function simpleMinimalDisk: circle;
(* Computes an approximation to the smallest circle which contains the given points using a simple	*)
(*	heuristic algorithm.  The order of the points might change. 												*)

			function containedInCircle (c: circle): boolean;
(* Tests whether all the given points lie in circle 'c'. The order of the points does not change. 		*)


(* Projection *)
			function projectOnX: pointVector;
(* Projects the points on the x-axis and produces a new 'pointVector'. The order of the points		*)
(* does not change. 																								*)

			function projectOnY: pointVector;
(* Projects the points on the y-axis and produces a new 'pointVector'. The order of the points does	*)
(* not change. 																										*)


(* Test data generation *)
			function createGrid (yCoordinates: pointVector): pointVector;
(* The 'pointVector' 'self' provides the x-coordinates, the 'pointVector' 'yCoordinates' the			*)
(* y-coordinates of a grid of points of size 'self.length * yCoordinates.length'. The order of the		*)
(* points does not change. *)

			function createLineSegmentVector (q: pointVector): lineSegmentVector;
(* Creates a vector of line segments where 'self' provides the starting points and 'q' the end		*)
(* points. Both must have the same number of elements. The order of the points does not change.	*)

			function createRectangleVector (q: pointVector): homogenVector; (* should be 'rectangleVector' *)
(* Creates a vector of rectangles where 'self' provides the lower left corners and 'q' the upper	*)
(* right corners. Both must have the same number of elements. The order of the points does not		*)
(*	change.																											*)

			function createRandomGraph (isDirectedGraph, isConnectedGraph: boolean; edgesGivenMethod: integer; edgeProbability: extended; numberOfEdges: seqIndex0): simpleUndirectedGraph;
(*	Creates a random graph where the points of the point vector are the vertices of the graph. If 	*)
(*	'IsDirectedGraph = true' then the graph will be directed otherwise undirected. If					*)
(*	'edgesGivenMethod = givenByNumber' then the number of edges of the resulting graph	will be	*)
(*	'numberOfEdges'. If 'edgesGivenMethod = givenByProbability' then each edge in the graph is		*)
(*	present with probability 'edgeProbability'.	If 'isConnectedGraph = true' then the resulting graph	*)
(*	is connected.																										*)


(* Conversions *)
			function makePolyLine: polyLine;
(* Produces a poly line from the given points. The order of the points does not change. 				*)

			function makeSimplePolygon: polygon2d;
(* Produces a simple polygon using a modification of Graham's scan. 										*)

			function makeStarShapedPolygon: polygon2d;
(* Produces a star-shaped polygon. Its kernel contains the center of gravity of the first three		*)
(*	points of the input pointVector 'self'. A rotation sweep method is used.								*)


			function shortestPath (pStart, pEnd: point2d): polyLine;
(*	Computes the shortest path from 'pStart' to 'pEnd'. The points in 'self' define vertical obstacles		*)
(*	with the definition that 'self.elements^^[2*i-1]' is the lower end-point and 'self.elements^^[2*i]'	*)
(*	is the upper end-point of the i-th vertical line segment (1  i  'self.length / 2'. The result is the		*)
(*	shortest path connecting the points without hitting the obstacles.											*)

			function orderTypeOne: pointVector;
(*	Computes a vector of markedRealPoint where each point q is the corresponding point from 'self'		*)
(*	marked with the number of points which lie to the right of the directed line segment from the first	*)
(*	point of 'self' to q.																									*)

(* inserted by PV *)

			function makeIdPoints: pointVector;
(* Self is duplicated and to every element of self its position in the array is added. *)

			procedure buildXAndYLists (var xList, yList: listMuchLinked);
(* xList / yList are  sorted by x / y -coordinate and the equivalent elements are chained. *)

(* *)
			function initNearestNeighbors (kind: neighborKind): lineSegmentVector;

		end;

	polyLine = object(pointVector)
			procedure interactiveOutput;
			override;
			procedure windowInput (px, py: integer);
			override;
			function dragVertex (where: obj; mouse: Point; highlight: boolean): boolean;
			override;
			procedure interactiveHighlight;
			override;
			function isSelected (where: obj): boolean;
			override;
			function equal (o: obj): boolean;
			override;
			function memberType (o: obj): boolean;
			override;
			function getType: objectType;
			override;
			function getTextType: Str255;
			override;
			procedure describe;
			override;
			procedure execute (item: integer);
			override;
			function closestPairPoints (ask: boolean): lineSegment;
			override;
			function closestPairHeuristic: lineSegment;
			override;
			function closestPairProbabilistic (ask: boolean): lineSegment;
			override;
			function closestPairNewHeuristic: lineSegment;
			override;
			function closestPairModified (ask: boolean): lineSegment;
			override;
			function nearestNeighborSweep (kind: neighborKind; ask: boolean): lineSegmentVector;
			override;
			function nearestNeighborBoxes (ask: boolean): lineSegmentVector;
			override;
			function nearestNeighborProjection (kind: neighborKind; useY: boolean): lineSegmentVector;
			override;
			function nearestNeighborInSector (ask: boolean; cosLower, sinLower, cosUpper, sinUpper: extended): lineSegmentVector;
			override;
			function minimalDisk (verbose: boolean): circle;
			override;
			function simpleMinimalDisk: circle;
			override;
			function pointCover (n: longInt): obj;
			override;
			function randomPointCover (n: longInt): obj;
			override;
			function textFileOutputId: Str255;
			override;
			function perimeter: extended;
			override;
			function drawingComplexity (cutOff: longInt): longInt;
			override;

			procedure drawAdjacentEdges (i: seqIndex1; highlight: boolean);

			function toLineSegmentVector (segmentType: objectType): lineSegmentVector;
	(* Computes a vector of line segments that corresponds to the edges of the poly line or polygon. *)

			function convexHullLinear: convexPolygon;
	(*	Computes the convex hull of the simple (non-intersecting) poly line with a linear time algorithm. *)

		end;

	polygonKind = (polygonIsConvexMinus, polygonIsConvexDegenerate, polygonIsConvexPlus, polygonIsNotConvex, polygonIsNotConvexDegenerate);

	polygon2d = object(polyLine)
			function isSelected (where: obj): boolean;
			override;
			procedure drawAdjacentEdges (i: seqIndex1; highlight: boolean);
			override;
			function equal (o: obj): boolean;
			override;
			function memberType (o: obj): boolean;
			override;
			function getType: objectType;
			override;
			function getTextType: Str255;
			override;
			procedure describe;
			override;
			procedure execute (item: integer);
			override;
			function toLineSegmentVector (segmentType: objectType): lineSegmentVector;
			override;
			function perimeter: extended; (* the sum of the length of all edges *)
			override;
			function removeRepeatedElements: sequence;
			override;
			function createGraph (resultType: objectType): obj;
			override;
			function textFileOutputId: Str255;
			override;
			function area: extended;
			override;
			function convertToPolygon: obj;
			override;

{$IFC VorlesungGeoAlg }
			function pointInPolygonNew (q: point2d): integer;
{$ENDC }

			function windingNumber (r: point2d; var onBoundary: boolean): longInt;
(* Computes the winding number of 'r' around the polygon. 'onBoundary = TRUE' iff the point 'r'		*)
(* lies on the polygon's boundary. *)

			function pointInPolygon (q: point2d): integer;

			function TSPOptimize (whichN: longInt; ask: boolean): polygon2d;
(* Tries to shorten a given traveling salesman tour. The parameters 'whichN' and 'ask' determine	*)
(* what kind of optimization is tried. The reader is referred to the source code for their precise		*)
(* meaning. 																											*)

			function clip (ls: lineSegment): polygon2d;
(* Clips 'self' on the given lineSegment. Everything on the right side (the line segment 'ls' itself	*)
(* excluded) of the directed line segment 'ls' is assumed to be visible. Algorithm: Sutherland-Hodgman. *)

			function decompose (clipper: lineSegment): vector;
(*	Computes the partition induced by intersecting the polygon with the 'clipper' line segment. The result is	*)
(*	a vector containing two vectors of polygons describing the set of polygons on each side of 'clipper' whose	*)
(*	union is 'self'.																												*)

			procedure intersection (pols: sequence; function zoneQ (s: polSet): boolean; var result: list; ask: boolean; pointsInPolygon: pointVector);
(* Computes boolean operations on polygons using the sweepline algorithm of Nievergelt and Preparata	*)
(* and adds them to the list 'result'. 'pols' is a sequence of polygons enumerated from 1 to pols.length.	*)
(* The function 'zoneQ' determines which areas belong to the result. E.g. to get the union of all			*)
(* polygons use 'zoneQ := polSet  []'. To get the intersection, use 'zoneQ := polSet = [1..pols.length]'.	*)
(* In particular, a single polygon can be decomposed into simple parts by using the function				*)
(* 'zoneQ := polSet  []' (menu entry 'Decompose'). To divide a polygon into its 'simple hull' and zero	*)
(* or more simple polygons inside, use 'zoneQ := polSet = []' (menu entry 'Simplify').						*)
(*	If pointsInPolygon  nil, then for each markedRealPoint p in pointsInPolygon its mark is set	*)
(*	to either In (point has odd winding number), Out (point has even windingnumber) or On (point	*)
(*	lies on the perimeter of the polygons).																			*)

{$IFC ScriptEnabled }
			procedure intersectionInc (var bloc: incPolIntersect; finish: boolean; pols: sequence; function zoneQ (s: polSet): boolean; pointsInPolygon: pointVector);
(* M_SCM, 15.02.94 *)
{$ENDC }

			function simpleSelfIntersect: vector;
(* Computes all intersections of edges of a polygon that do not occur at vertices using the trivial algorithm. *)

			function selfIntersect: vector;
(* Computes all intersections of edges of a polygon that do not occur at vertices using a boundary		*)
(* traversal algorithm. This algorithm is efficient for the class of polygons that have a left turn in		*)
(* each vertex. *)

			function isMonotone: boolean;
(*	Determines whether the polygon is monotone.	*)

			function isConvex (var count: seqIndex1): polygonKind;
(* Classifies the polygon. The variable 'count' is the index of the last point not yet processed.				*)

			function triangulate2d: obj; (* should be triangulation2d *)
(* Triangulates the simple polygon with a plane sweep algorithm. 													*)

			function lastVertexNumber (vertexNumber: seqIndex): seqIndex;
(* liefert den Index der Vorgaengerecke von Ecke mit vertexnumber *)
(* Eingefuegt von Michael Vollenbroeker *)

			function nextVertexNumber (vertexNumber: seqIndex): seqIndex;
(* liefert den Index der Nachfolgerecke von Ecke mit vertexnumber *)
(* Eingefuegt von Michael Vollenbroeker *)

			function middleVertex (vertIndex1, vertIndex2: seqIndex): seqIndex;
(* liefert den Index der mittleren Ecke zwischen den Ecken mit den Nummern vertindex1 und vertindex2 *)
(* Eingefuegt von Michael Vollenbroeker *)

			function numberOfVertices (m, gm: seqIndex): seqIndex0;
(* liefert die Anzahl der Ecken, die im Polygon zwischen den Ecken mit den Nummern m und gm liegen *)
(* Eingefuegt von Michael Vollenbroeker *)

		end;

	convexPolygon = object(polygon2d)
			bottom, top, left, right: seqIndex; (* <----------- inserted by JM -------------------- *)
			procedure establishConvexity;
			procedure windowInput (px, py: integer);
			override;
			function dragVertex (where: obj; mouse: Point; highlight: boolean): boolean;
			override;
			procedure setInfo (o: obj);
			override;
			function equal (o: obj): boolean;
			override;
			procedure describe;
			override;
			procedure execute (item: integer);
			override;
			function getType: objectType;
			override;
			function getTextType: Str255;
			override;
			function memberType (o: obj): boolean;
			override;
			procedure randomChange;
			override;
			procedure convexHull (eliminate, verbose: boolean; var hullLength, eliminated: seqIndex0);
			override;
			function convexHullDivideAndConquer: convexPolygon;
			override;
			function isConvex (var count: seqIndex1): polygonKind;
			override;
			function textFileOutputId: Str255;
			override;
			function convexHullGrahamScan: convexPolygon;
			override;

			function intersect (x: convexPolygon; verbose: boolean): convexPolygon;
(* Computes the intersection of the two convex polygons using the boundary traversal method. 	*)

			procedure tangent (c: convexPolygon; var upperP, upperQ, lowerP, lowerQ: seqIndex0);
(* Computes the two common outer tangents of the two convex polygons. The lower tangent is	*)
(* given by index 'lowerP' in 'self' and 'lowerQ' in 'c' and the upper tangent is given by index	*)
(* 'upperP' in 'self' and 'upperQ' in 'c'.	 The result variables are '0' if no tangents exist.			*)

			procedure diameter (var iMax, jMax, iMin, jMin: seqIndex0);
(* Computes the indices 'iMax' and 'jMax' of the two points determining the diameter (width) of the convex polygon		*)
(*	and the indices 'iMin' and 'jMin' determining the 'height' of the convex polygon. 												*)

			function inside (p: point2d): boolean;
(* Tests whether point 'p' is inside the convex polygon or not. The boundary belongs per definition to the inside. *)

(* ------------------------------------- inserted by JM --------------------------------------------------------------- *)

			procedure findExtremalPoints;
(* Finds the points with minimal and maximal x- and y-coordinates of CONVEX polygons by using a binary search. *)
(* Their indexes ara stored in the variables bottom, top, left and right which belong to the internal state of self. *)
(* The base idea was developed by Dobkin & Souvaine (1991). This version, by Joerg Mrkwitschka, *)
(* is modified  to increase speed (binary search cuts 1/2 not just 1/3 every step) and to handle every *)
(* configuration correct (not every projection-function is unimodal - some are bimodal or just monotone). *)
(* The implementation combines many cases so that the code became more complex (faster) and easier *)
(* to understand. The time complexity is O(log m) instead of O(m)  for the brute force version *)

			function simpleMinimalDistance (cq: convexPolygon; var d: extended): lineSegment;
(* This is a simple minimal distance algorithm. It uses the convexity just to determine if one is inside the other. *)
(* The time complexity is O (m^2). *)

			function minimalDistance (cq: convexPolygon; var d: extended): lineSegment;
(* Computes the minimal Distance between two CONVEX polygons using a binary search. *)
(* The base idea was developed by Chin & Wang (1983).  This implementation, by Joerg Mrkwitschka, *)
(* considers all (more then just 2) cases, detects intersections(!) and inclusions(!) and shrinks the  *)
(* the size of the relevant start-edges more and easier(!) . *)
(* The time complexity is O(log m) (instead of O (m^2) for the straight forward method)  *)
(* Note that  bottom, top, left and right must have been calculated first !!!  (using findExtrema) *)
(* If the polygons intersect or overlap d is set to 0 and  the minimalDistance is a linesegment *)
(* consisting of to equal Points.  This point is not nessesarily inside both polygons nor it is the *)
(* intersectionpoint!  It is the point where the algorithm could tell that an intersection or inclusion *)
(* must  be somewhere!   If the polygons do not intersect or overlap minimalDistance consists  *)
(* of the two points which represent the minimal distance! *)

			function simpleClosestPair: lineSegment;
(* This is a simple closest pair algorithm. Together with the simple minimal distance algorithm the time complexity is O(m^2 * n^2).  *)

			function closestPairConvexPolygons (ask: boolean): lineSegment;
(* Computes the closest-pair of a given number of CONVEX polygons. The algoritm was developed by Frank Bartling (1991) . *)
(* This implementation is a more general version of the one for linesegments by Frank Barkling. *)
(* It works for points, linesegment and convex polygons as long as they are of the type convexPolygon. *)
(* The time complexity is O(n*log(m*n)  ), but only with the used log(m)-minimal-distance-algorithm! *)

			function getVertex (i: longInt): point2d;
(* Gives the vertex with index i. ( i is taken modulo length, in a way that this function can be called with numbers *)
(* between -length+1 ...  +infinity -- especially called with 0 it gives back vertex[length]!! )  *)
(* Note that the order is changed from clockwise to counterclockwise ! *)
(* !!!!!!!!!!!! This is important, otherwise the algorithms above don't work !!!!!!!!!!!!!!!!  *)

			function objectYLessThan (cq: convexPolygon): boolean;
(* This method compares the pseudofunction of the two polygons self and cq at s. The point s , where the polygons are *)
(* compared, is the minimal point in x of the right polygon (The maximal point of the left on would be ok too). *)
(* The pseudofunction is a line through the points vertex[left] and vertex[right] if a vertical line at s cuts the polygon.  *)
(* If the vertival line lies to the left of the polygon then the pseudo-function is vertex[left].y   *)
(* If the vertival line lies to the right of the polygon then the pseudo-function is vertex[right].y  *)

(* ------------------------------------------------------------------------------------------------------------------------------------- *)

			function counterclockwise: boolean;
(* ueberprueft, ob 'self' entgegen dem Uhrzeigersinn gespeichert ist *)
(* Eingefuegt von Michael Vollenbroeker 6.11.92 *)

			function verticalDistance (o: obj): extended;
			override;
(* Berechnet den vertikalen Abstand der Polygone 'self' und'o'.  *)
(*  Eingefuegt von Michael Vollenbroeker 14.1.93 *)

			function getPointType: objectType;
			override;
(* liefert den Punkttyp von 'self'.  Eingefuegt von Michael Vollenbroeker 12.2.93 *)

			function distanceO (o: obj; var minP, minQ: obj): extended;
			override;
(* berechnet den Abstand von 'self' und 'o', der in minP und minQ angenommen wird. *)
(* Eingefuegt von Michael Vollenbroeker 12.2.93 *)

			function getLeftPoint: obj;
			override;
(* liefert den am weitesten links liegenden Punkt von 'self' .  *)
(* Eingefuegt von Michael Vollenbroeker 12.2.93 *)

			function getRightPoint: obj;
			override;
(* liefert den am weitesten rechts liegenden Punkt von 'self' .  *)
(* Eingefuegt von Michael Vollenbroeker 12.2.93 *)

			function distance (Q: convexPolygon; var minP, minQ: point2d): extended;
(* Berechnet den minimalen Abstand von 'self' und Q in der euklidischen Metrik. Die Punkte, in denen *)
(* dieser Abstand angenommen wird, sind minP und minQ. Vor dem Aufruf von distance muessen die *)
(* Instanzvariablen top und bottom bestimmt werden. *)
(* Eingefuegt von Michael Vollenbroeker 16.11.92 *)

			function HBGLessThan (o: obj): boolean;
			override;
(* Diese Methode bestimmt, ob 'self' bezueglich der von Hinrichs/Bartling/Graf definierten Ordnung kleiner als 'o' ist. *)
(* Dies ist genau dann der Fall, wenn Qquer(self) an der Position der sweepline (linker Punkt des rechten Polygons) unter *)
(* Qquer(o) liegt. Ist P ein konvexes Polygon, so bezeichnet Qquer(P) das Liniensegment zwischen den Ecken mit den *)
(* Indizes 'left' und 'right', verlaengert um einen horizontalen Strahl nach rechts in der Ecke mit dem Index 'right'. *)
(* Eingefuegt von Michael Vollenbroeker 3.12.92 *)

		end;

	histogram = object(polygon2d)
			procedure interactiveOutput;
			override;
			procedure interactiveHighlight;
			override;
			function dragVertex (where: obj; mouse: Point; highlight: boolean): boolean;
			override;
			function equal (o: obj): boolean;
			override;
			function memberType (o: obj): boolean;
			override;
			function getType: objectType;
			override;
			function getTextType: Str255;
			override;

			procedure initFromGenerator (numberOfValues, numberOfBins: longInt; width, height: extended; function generateValue: real; procedure resetGenerator);
(*	Computes a histogram. There are numberOfValues many values which are produced by successive calls to generateValue.	*)
(*	The generator is initialized by resetGenerator. The parameter numberOfBins determines the number of columns				*)
(*	to be used. The parameters width and height determine the total width and height of the diagram.								*)
		end;

	voronoiEdgeT = (bisector, ray, segment);
	voronoiEdge = object(obj)
			eType: voronoiEdgeT;
			p1, p2, v1, v2: realPoint;	(* for easier implementation: p2 is greater than p1 *)
			procedure interactiveOutput;
			override;
			procedure textOutput;
			override;
			procedure textFileInput;
			override;
			procedure textFileOutput;
			override;
			procedure fileInput;
			override;
			procedure fileOutput;
			override;
			function getType: objectType;
			override;
			function getTextType: Str255;
			override;
			function getInfo: obj;
			override;
			function equal (o: obj): boolean;
			override;
			function genericLessThan (o: obj): boolean;
			override;
			function memberType (o: obj): boolean;
			override;
			function drawingComplexity (cutOff: longInt): longInt;
			override;
			function bisecX (y: real): real;
(* calculates the x-coordinate of the bisector of p1 and p2 at y-position *)
			function bisecY (x: real): real;
(* calculates the y-coordinate of the bisector of p1 and p2 at x-position *)
			function definingPoints (p, q: realPoint): boolean;
(* returns TRUE iff p and q are the defining points of the edge *)
			procedure toPoints (var p, q: realPoint; maxCoord: real);
(* gets the endpoints of the edge *)
			procedure commonPoints (e: voronoiEdge; var singlePoint1, doublePoint, singlePoint2: realPoint; var ok: boolean);
(* determines the point 'doublePoints', which is the common point of the defining points of edges 'self' and 'e' 	*)
(* 'singlePoint1' is the other point of edge 'self', 'singlePoint2'  is the other point of edge 'e'						*)
			function otherPointOfEdge (p: realPoint): realPoint;
(* returns the other of the two defining points of edge than p *)
		end;

	voronoiDiagram = object(obj)
			points: pointVector;
			edges: list;
			procedure free;
			override;
			procedure freeAll;
			override;
			procedure describe;
			override;
			procedure execute (item: integer);
			override;
			function isSelected (where: obj): boolean;
			override;
			procedure translate (dx, dy: extended);
			override;
			procedure rotate (xOrigin, yOrigin, alpha: extended);
			override;
			procedure scale (lambda: extended);
			override;
			procedure reflect (a, b, c: extended);
			override;
			procedure interactiveOutput;
			override;
			procedure interactiveHighlight;
			override;
			procedure textFileInput;
			override;
			procedure textFileOutput;
			override;
			procedure fileInput;
			override;
			procedure fileOutput;
			override;
			function duplicate: obj;
			override;
			function duplicateAll: obj;
			override;
			function memberType (o: obj): boolean;
			override;
			function getInfo: obj;
			override;
			function getType: objectType;
			override;
			function getTextType: Str255;
			override;
			function genericLessThan (o: obj): boolean;
			override;
			procedure centerOfGravity (var cx, cy: extended);
			override;
			function equal (o: obj): boolean;
			override;
			function getBoundingObject: obj;
			override;
			function pointCover (n: longInt): obj;
			override;
			function drawingComplexity (cutOff: longInt): longInt;
			override;

			function getFiniteEdges: lineSegmentVector;
	(* Computes the finite Voronoi edges. *)

			function getVertices: pointVector;
	(* Computes the Voronoi vertices. *)

			function delaunayTriangulation: simpleUndirectedGraph;
	(* Computes the Delaunay triangulation, the dual to the Voronoi diagram. *)

			function EMST (k: seqIndex; var uf: unionFindSimple): spanningTree;
	(* Computes a Euclidean minimum spanning tree from the Voronoi Diagram. *)

(* begin Marco Ciot *)
			function modifiableDT: modifiableUndirectedGraph;
	(* The same as delaunayTriangulation returning the extended graph type that may be further modified. *)

			function gabrielGraph: modifiableUndirectedGraph;
	(* Matula, Sokal *)

			function RNGfalse: modifiableUndirectedGraph;
	(* produces an almost RNG using the technique of Matula, Sokal *)

(* end Marco Ciot *)

			function nearestNeighborsVoronoi: lineSegmentVector;
	(* Like 'pointVector.nearestNeighborSweep(allNeighbors)', but the information is computed using the Voronoi diagram. *)

			function polygonArea: homogenVector;
	(*	Computes for each point the size of its Voronoi-polygon (-1, if the polygon is unbounded) as a vector of 'realNumber'. *)

		end;

var
	optimizeFirst, optimizeLast, stdSize, lowerAngle, upperAngle, defaultRadius: extended; (* length of the hash table of closestPairProbabilistic *)

function createRealPointVector (l: seqIndex): pointVector;

function newPointVector (length: longInt): pointVector;

function newPolygon2d (length: longInt): polygon2d;

procedure initPointVector;

procedure drawClippedPolyLine (poly: pointVector; closed: boolean);

implementation

{$SETC CompareANN=FALSE }

var
	numberOfClusters: extended;

function makeRandomSpanningTree (ps: pointVector): spanningTree;
	var
		result: spanningTree;
		e: graphEdge;
		i: seqIndex;
	begin
		if ps.length > 1 then begin
			new(result);
			failNil(Handle(result));
			result.vertices := vector(ps.duplicateAll);
			result.edges := newHomogenVector(ps.length - 1);
			for i := 1 to ps.length - 1 do begin
				if userAbort then begin
					result.freeAll;
					makeRandomSpanningTree := nil;
					exit(makeRandomSpanningTree);
				end;
				new(e);
				failNil(Handle(e));
				e.startVertex := i;
				e.endVertex := randomRange(i + 1, ps.length + 1);
				result.edges.append(e);
			end;
			makeRandomSpanningTree := result;
		end
		else
			makeRandomSpanningTree := nil;
	end;

	const
		initialStdSize = 107; (* size of the hashTable for closestPairProbabilistic *)

function pointVector.convertToType (t: pointTypeRange): obj;
	var
		v: pointVector;
		i: seqIndex0;
	begin
		v := pointVector(duplicate);
		for i := 1 to length do
			v.elements^^[i] := elements^^[i].convertToType(t);
		convertToType := v;
	end;

function newPointVector (length: longInt): pointVector;
	var
		pv: pointVector;
	begin
		currentVectorLength := length;
		new(pv);
		failNil(Handle(pv));
		pv.init;
		newPointVector := pv;
	end;

function newPolygon2d (length: longInt): polygon2d;
	var
		pv: polygon2d;
	begin
		currentVectorLength := length;
		new(pv);
		failNil(Handle(pv));
		pv.init;
		newPolygon2d := pv;
	end;

function createRealPointVector (l: seqIndex): pointVector;
	var
		pv: pointVector;
	begin
		new(pv);
		failNil(Handle(pv));
		pv.elements := nil;
		pv.preset(l, realPointType);
		createRealPointVector := pv;
	end;

procedure drawClippedPolyLinePart (poly: pointVector; left, right: seqIndex0);
	type
		point2dExtT = record			(* for efficiency reasons *)
				p: record
						x, y: extended;
					end;
				vis: array[1..4] of boolean;
			end;
	var
		po: PolyHandle;
		neverMoved: boolean;
		clip: point3dT;
		pk: point2dExtT;
		pp, sp, fp: array[1..4] of point2dExtT;
		firstExist: array[1..4] of boolean;
		i: seqIndex0;
		tx, ty: extended;

	procedure intSec (var e: integer; var p, q, s: point2dExtT);
		var
			t: extended;
		begin
			if (e = 3) | (e = 1) then begin
				assert(q.p.y <> p.p.y, 'q.p.y  p.p.y');
				t := (clip[e] - p.p.y) / (q.p.y - p.p.y);
				if (t < 0.0) | (t > 1.0) then
					SysBeep(2);
				s.p.y := clip[e];
				s.p.x := p.p.x + t * (q.p.x - p.p.x);
			end
			else begin
				assert(q.p.x <> p.p.x, 'q.p.x  p.p.x');
				t := (clip[e] - p.p.x) / (q.p.x - p.p.x);
				if (t < 0.0) | (t > 1.0) then
					SysBeep(2);
				s.p.x := clip[e];
				s.p.y := p.p.y + t * (q.p.y - p.p.y);
			end;
			s.vis[1] := s.p.y >= clip[1];
			s.vis[2] := s.p.x <= clip[2];
			s.vis[3] := s.p.y <= clip[3];
			s.vis[4] := s.p.x >= clip[4];
		end;

	procedure prDrawLine (p: point2dExtT);
		var
			ix, iy: integer;
		begin
{$IFC THINK_PASCAL }
			ix := round(p.p.x);
			iy := round(p.p.y);
{$ELSEC }
			ix := roundLong(p.p.x);
			iy := roundLong(p.p.y);
{$ENDC }
			if neverMoved then begin
				MoveTo(ix, iy);
				neverMoved := false;
			end
			else
				LineTo(ix, iy);
			ExtendDrawingSheet(ix, iy);
		end;

	procedure prLineTo (e: integer; var p: point2dExtT);
		var
			iSp: point2dExtT;
		begin
			pp[e] := p;
			if firstExist[e] then begin
				if sp[e].vis[e] <> pp[e].vis[e] then begin
					intSec(e, sp[e], pp[e], iSp);
					if e <> 4 then
						prLineTo(e + 1, iSp)
					else
						prDrawLine(iSp);
				end;
			end
			else begin
				fp[e] := p;
				firstExist[e] := true;
			end;
			sp[e] := pp[e];
			if sp[e].vis[e] then begin
				if e <> 4 then
					prLineTo(e + 1, sp[e])
				else
					prDrawLine(sp[e]);
			end;
		end;

	procedure last (e: integer);
		var
			iSp: point2dExtT;
		begin
			if firstExist[e] then begin
				if sp[e].vis[e] <> fp[e].vis[e] then begin
					intSec(e, sp[e], fp[e], iSp);
					if e <> 4 then
						prLineTo(e + 1, iSp)
					else
						prDrawLine(iSp);
				end;
				if e <> 4 then
					last(e + 1);
			end;
		end;

	begin	(*DrawClippedPolyLine *)
		po := OpenPoly;
		neverMoved := true;
		with clippingRectangle do begin
			clip[4] := left;
			clip[3] := top;
			clip[2] := right;
			clip[1] := bottom;
		end;
		for i := 1 to 4 do
			firstExist[i] := false;
		with poly do begin
			for i := left to right do begin
				with currentFrame do begin
					tx := dx + point2d(elements^^[i]).getX * mx;
					ty := dy + point2d(elements^^[i]).getY * my;
				end;
				pk.p.x := tx;
				pk.p.y := ty;
				pk.vis[1] := ty >= clip[1];
				pk.vis[2] := tx <= clip[2];
				pk.vis[3] := ty <= clip[3];
				pk.vis[4] := tx >= clip[4];
				prLineTo(1, pk);
			end;
		end;
		last(1);
		ClosePoly;
		FramePoly(po);
		KillPoly(po);
	end;

{$IFC AnimationEnabled }
procedure showTrials (trials: longInt);
	begin
		displayInfoString(StringOf('Trials = ', trials : 1));
	end;
{$ENDC }

procedure findRandomObject (function generate: obj; function test (var x: obj; var proof: obj): boolean);
	var
{$IFC AnimationEnabled }
		trials: longInt;
{$ENDC }
		x, proof: obj;
	begin
{$IFC AnimationEnabled }
		trials := 0;
{$ENDC }
		repeat
{$IFC AnimationEnabled }
			trials := trials + 1;
			if trials mod 50 = 0 then
				showTrials(trials);
{$ENDC }
			x := generate;
{$IFC AnimationEnabled }
			if userAbort then begin
				showTrials(trials);
				exit(findRandomObject);
			end;
{$ENDC }
			proof := nil;
			if test(x, proof) then begin
				currentResult.appendNonNil(x);
				currentResult.appendNonNil(proof);
{$IFC AnimationEnabled }
				showTrials(trials);
{$ENDC }
				exit(findRandomObject);
			end
			else begin
				if x <> nil then
					x.freeAll;
				if proof <> nil then
					proof.freeAll;
			end;
		until false;
	end;

procedure findHennebergCounterexample (ps: pointVector);
	const
		maxIterations = 10;

	function generateGraph: obj;
		begin
			generateGraph := ps.createRandomGraph(false, true, givenByNumber, 0, 2 * ps.length - 3);
		end;

	function testGraph (var g: simpleUndirectedGraph; var proof: homogenVector): boolean;
		var
			h1, h2: homogenVector;
			reduceDegreeThreeVertex: seqIndex0;
			trials: longInt;
		begin
			h1 := g.hennebergSequenceDeterministic(reduceDegreeThreeVertex);
			if h1.length < ps.length then begin
				h2 := g.hennebergSequenceMaxTrials(maxIterations, trials, reduceDegreeThreeVertex);
				if h2.length = ps.length then begin
					proof := newHomogenVector(2);
					proof.append(h1);
					proof.append(h2);
					testGraph := true;
				end
				else begin
					h1.freeAll;
					h2.freeAll;
					testGraph := false;
				end;
			end
			else begin
				h1.freeAll;
				testGraph := false;
			end;
		end;

	begin (* findHennebergCounterexample *)
		findRandomObject(generateGraph, testGraph);
	end;

procedure findTriangulableGraph (ps: pointVector);

	function generateGraph: obj;
		var
			g: simpleUndirectedGraph;
			u: undirectedGraph;
		begin
			g := ps.createRandomGraph(false, true, givenByNumber, 0, 2 * ps.length - 3);
			if g = nil then
				generateGraph := nil
			else begin
				new(u);
				failNil(Handle(u));
				u.initializeFromSimpleUndirectedGraph(g);
				g.freeAll;
				generateGraph := u;
			end;
		end;

{$IFC NOT THINK_PASCAL }
{$PUSH }
{$PRAGMAC warn_unusedarg off }
{$ENDC }
	function testGraph (var u: undirectedGraph; var proof: obj): boolean;
		var
			startEdge: seqIndex0;
		begin
			testGraph := u.maxTriangulationOrder(startEdge) = ps.length;
		end;
{$IFC NOT THINK_PASCAL }
{$POP }
{$ENDC }

	begin	(* findTriangulableGraph *)
		if ps.length > 1 then
			findRandomObject(generateGraph, testGraph);
	end;

procedure findTriangulableGraph2 (ps: pointVector);
	const
		maxIterations = 10;

	function generateGraph: obj;
		begin
			generateGraph := ps.createRandomGraph(false, true, givenByNumber, 0, 2 * ps.length - 3);
		end;

	function testGraph (var g: simpleUndirectedGraph; var proof: homogenVector): boolean;
		var
			reduceDegreeThreeVertex: seqIndex0;
			trials: longInt;
		begin
			proof := g.hennebergSequenceMaxTrials(maxIterations, trials, reduceDegreeThreeVertex);
			testGraph := (proof.length = ps.length) & (reduceDegreeThreeVertex = 0);
		end;

	begin	(* findTriangulableGraph *)
		if ps.length > 1 then
			findRandomObject(generateGraph, testGraph);
	end;

procedure findRigidNotTriangulableGraph (ps: pointVector);

	function generateGraph: obj;
		begin
			generateGraph := ps.createRandomGraph(false, true, givenByNumber, 0, 2 * ps.length - 3);
		end;

	function testGraph (var g: simpleUndirectedGraph; var proof: homogenVector): boolean;
		var
			startEdge, reduceDegreeThreeVertex: seqIndex0;
			u: undirectedGraph;
		begin
			proof := g.hennebergSequence(reduceDegreeThreeVertex);
			if (proof.length = ps.length) & (reduceDegreeThreeVertex > 0) then begin
				new(u);
				failNil(Handle(u));
				u.initializeFromSimpleUndirectedGraph(g);
				g.freeAll;
				g := u;
				testGraph := u.maxTriangulationOrder(startEdge) < ps.length;
			end
			else
				testGraph := false;
		end;

	begin (* findRigidNotTriangulableGraph	*)
		if ps.length > 5 then
			findRandomObject(generateGraph, testGraph);
	end;

procedure findRigidGraph (ps: pointVector);

	function generateGraph: obj;
		begin
			generateGraph := ps.createRandomGraph(false, true, givenByNumber, 0, 2 * ps.length - 3);
		end;

	function testGraph (var g: simpleUndirectedGraph; var proof: homogenVector): boolean;
		var
			reduceDegreeThreeVertex: seqIndex0;
		begin
			proof := g.hennebergSequence(reduceDegreeThreeVertex);
			testGraph := proof.length = ps.length;
		end;

	begin (* findRigidGraph *)
		if ps.length > 1 then
			findRandomObject(generateGraph, testGraph);
	end;

procedure findDefaultFormulaGraph (ps: pointVector);
	var
		dummy: boolean;
		errorPosition: integer;

	function generateGraph: obj;
		var
			r1: simpleUndirectedGraph;
			r2: undirectedGraph;
		begin
			r1 := ps.createRandomGraph(false, true, givenByNumber, 0, 2 * ps.length - 3);
			new(r2);
			failNil(Handle(r2));
			r2.initializeFromSimpleUndirectedGraph(r1);
			r1.freeAll;
			generateGraph := r2;
		end;

{$IFC NOT THINK_PASCAL }
{$PUSH }
{$PRAGMAC warn_unusedarg off }
{$ENDC }
	function testGraph (var g: undirectedGraph; var proof: homogenVector): boolean;
		var
			errorPosition: integer;
		begin
			testGraph := isGraphOk(g, defaultFormula, true, errorPosition);
		end;
{$IFC NOT THINK_PASCAL }
{$POP }
{$ENDC }

	begin (* findDefaultFormulaGraph *)
		dummy := isGraphOk(nil, defaultFormula, false, errorPosition);
		if not showErrorInDefaultFormula(errorPosition) & (ps.length > 1) then
			findRandomObject(generateGraph, testGraph);
	end;


procedure drawClippedPolyLine (poly: pointVector; closed: boolean);
	const
		maxPartSize = 8100;

	function min (a, b: longInt): longInt;
		begin
			if a < b then
				min := a
			else
				min := b;
		end;

		var
			n, l, r: longInt;
	begin
		n := poly.length;
		if closed then begin
			n := n + 1;
			poly.elements^^[n] := poly.elements^^[1];
		end;
		l := 1;
		r := min(maxPartSize + 1, n);
		while l < r do begin
			drawClippedPolyLinePart(poly, l, r);
			l := l + maxPartSize;
			r := min(r + maxPartSize, n);
		end;
	end;

	const
{$IFC false }
		TSPSimpleBranchInitC = 78;
		TSPConsConvHullC = 79;
		TSPNoIntersC = 80;
		TSPBranchAndBoundC = 81;
		TSPDiagrC = 82;
{$ENDC }
		MakePolygonC = 1;
		EMSTC = 2;
		AllNNSweepAskC = 3;
		VoronoiDiagramC = 4;
		VoronoiDiagramAskC = 5;
		ClosestPairModifiedAskC = 6;
		ClosestPairC = 7;
		ConvexHullC = 8;
		SalesmanEMSTC = 9;
		SalesmanNNC = 10;
		SalesmanConvexHullC = 11;
		ConvexHullDIC = 12;
		RightANNSweepC = 14;
		ClosestPairAskC = 15;
		ClosestPairProbabilisticC = 16;
		ConvexHullOnlyC = 17;
		ClosestPairProbabilisticAskC = 18;
		ClosestPairHeuristicC = 19;
		ClosestPairSlowC = 20;
		VoronoiDiagramDynamicC = 21;
		VoronoiDiagramSeidelDynamicC = 22;
		ClosestPairDynamicC = 23;
		XProjectC = 24;
		YProjectC = 25;
		CreateLineSegmentVectorC = 26;
		MakeSimplePolygonC = 27;
		CreateGridC = 28;
		ClosestPairNewHeuristicC = 29;
		LeftANNNewHeuristicC = 30;
		QuickSortByXYC = 31;
		ConvexHullEliminateC = 32;
		HeapSortByXYC = 33;
		NearestNeighborInSectorC = 34;
		NearestNeighborInSectorAskC = 35;
		MiniDiskC = 37;
		SimpleMiniDiskC = 38;
		InCircleC = 39;
		MakePolyLineC = 40;
		MakeStarShapedPolygonC = 41;
		ClosestPairDivideAndConquerC = 42;
		LeftANNSweepC = 43;
		LeftANNSweepAskC = 44;
(* Inserted by PV *)
		ANNBoxesC = 45;
		ANNBoxesAskC = 46;
(* *)
		RightANNSweepAskC = 47;
		AllANNSweepC = 48;
		LeftANNHeuristicC = 49;
		LeftANNHeuristicXYC = 50;
		RightANNHeuristicC = 51;
		RightANNHeuristicXYC = 52;
		AllANNHeuristicC = 53;
		AllANNHeuristicXYC = 13;
		ClosestPairModifiedC = 54;
		CreateRectangleVectorC = 55;
		MakeCirclesC = 56;
		QuickSortByYXC = 57;
		HeapSortByYXC = 58;
		ShortestPathC = 59;
		AlignedPointsXC = 60;
		AlignedPointsYC = 61;
		DirectedGraphC = 62;
		RandomGraphC = 63;
		EmptyDirectedGraphC = 64;
		EmptyGraphC = 65;
		HennebergCounterExampleC = 66;
		TriangulableGraphC = 67;
		RigidGraphC = 68;
		RigidNotTriangulableGraphC = 69;
		RigidSimpleGraphC = 70;
		FormulaGraphC = 71;
		VoronoiDiagramSeidelC = 72;
		VoronoiDiagramSeidelAskC = 73;
		OrderTypeC = 74;
		StarBushelC = 75;
		TSPSimpleC = 76; (* inserted by ab *)
		TSP1TreeC = 77; (* inserted by ab *)
		EMSTDynamicC = 78;
		EMSTVoronoiC = 79;
		EMSTVoronoiDynamicC = 80;
		TSP1TreeDynamicC = 81;
		LeftPlaneSweepDynamicC = 82;
		RightPlaneSweepDynamicC = 83;
		AllPlaneSweepDynamicC = 84;
		ANNBoxesDynamicC = 85;
		NearestNeighborInSectorDynamicC = 86;
		ConvexHullDynamicC = 87;
		MiniDiskDynamicC = 88;
{$IFC OptionalCommand }
		DistanceDistributionC = MiniDiskDynamicC + 1;
		DistanceDistributionDynamicC = DistanceDistributionC + 1;
		DistanceDistributionDynamicTwoVectorsC = DistanceDistributionDynamicC + 1;
		DistanceDistributionNoAskC = DistanceDistributionDynamicTwoVectorsC + 1;
		VoronoiAndTSPDynamicC = DistanceDistributionNoAskC + 1;
		MakeRandomSpanningTreeC = VoronoiAndTSPDynamicC + 1;
		PreviousCommand = MakeRandomSpanningTreeC;
{$ELSEC }
		PreviousCommand = MiniDiskDynamicC;
{$ENDC }
{$IFC CompareANN }
		CompareNearestNeighborAlgorithmsC = PreviousCommand + 1;
		PreviousCommand2 = CompareNearestNeighborAlgorithmsC;
{$ELSEC }
		CompareNearestNeighborAlgorithmsC = InCircleC;
		PreviousCommand2 = PreviousCommand;
{$ENDC }
		DelaunayDynamicC = PreviousCommand2 + 1;
		DelaunayC = DelaunayDynamicC + 1;
		MiniDiskApproximateDynamicC = DelaunayC + 1;
		MiniDiskRandomizedApproximateDynamicC = MiniDiskApproximateDynamicC + 1;
		DelaunayGridC = MiniDiskRandomizedApproximateDynamicC + 1;
		DelaunayGridDynamicC = DelaunayGridC + 1;
		ClusterC = DelaunayGridDynamicC + 1;
		SmallestSizeClusterC = ClusterC + 1;
		SmallestAreaClusterC = SmallestSizeClusterC + 1;
		SmallestAreaIntersectClusterC = SmallestAreaClusterC + 1;
		SmallestAreaIntersectClusterAskC = SmallestAreaIntersectClusterC + 1;
		SmallestAreaClusterAskC = SmallestAreaIntersectClusterAskC + 1;
		ConvexPolygonClusterC = SmallestAreaClusterAskC + 1;
		FullDirectedGraphC = ConvexPolygonClusterC + 1;
		FullGraphC = FullDirectedGraphC + 1;

(* begin Marco Ciot *)
		RNGC = FullGraphC + 1;
		GGsimpleC = RNGC + 1;
		GGmatulaSokalC = GGsimpleC + 1;
		RNGsupowitC = GGmatulaSokalC + 1;
		RNGfalseC = RNGsupowitC + 1;
(* end Marco Ciot *)

function testType (o: obj; t: objectType): boolean;
	begin
		testType := member(o, vector) & (vector(o).elements^^[1].getType = t) | (o.getType = t);
	end;

function isVerticalSegment (o: obj): boolean;
	begin
		isVerticalSegment := member(o, lineSegment) & (lineSegment(o).p.getX = lineSegment(o).q.getX);
	end;

function isPointVector (v: vector): boolean;
	begin
		isPointVector := (v.length > 1) & (member(v, pointVector) | member(v.elements^^[1], point2d));
	end;

procedure pointVector.describe;
	var
		isRealPoints: boolean;
		currentLength: seqIndex0;
	begin
		with currentArguments do
			if (length = 1) & currentArguments1IsHomogenVector & isPointVector(vector(ith(1))) | (length > 1) & currentArgumentsSameType & member(ith(1), point2d) then begin
				isRealPoints := testType(ith(1), realPointType) | testType(ith(1), markedRealPointType);
				if length = 1 then
					currentLength := vector(ith(1)).length
				else
					currentLength := length;
{$IFC Tutorial }
				menuEntry('Center Of Gravity', 0);
{$ENDC }
				beginSubMenu('Closest Pair');
				menuEntry('Simple', ClosestPairSlowC);
				menuEntry('Plane Sweep', ClosestPairC);
				menuEntry('Plane Sweep Ask...', ClosestPairAskC);
(* inserted by JM *)
				menuEntry('Plane Sweep Modified', ClosestPairModifiedC);
				menuEntry('Plane Sweep Modified Ask...', ClosestPairModifiedAskC);
(*   *)
				menuEntry('Projection On X', ClosestPairHeuristicC);
				menuEntry('Projection On XY', ClosestPairNewHeuristicC);
				menuEntry('Divide And Conquer', ClosestPairDivideAndConquerC);
				if isRealPoints then begin
					menuEntry('Probabilistic', ClosestPairProbabilisticC);
					menuEntry('Probabilistic Ask...', ClosestPairProbabilisticAskC);
				end;
				menuEntry('Dynamic', ClosestPairDynamicC);
				endSubMenu;
				beginSubMenu('Nearest Neighbors');
				menuEntry('Left Plane Sweep', LeftANNSweepC);
				menuEntry('Left Plane Sweep Ask...', LeftANNSweepAskC);
				menuEntry('Left Projection On X', LeftANNHeuristicC);
				menuEntry('Left Projection On XY', LeftANNHeuristicXYC);
				menuEntry('Left Plane Sweep Dynamic', LeftPlaneSweepDynamicC);
				menuEntry('Right Plane Sweep', RightANNSweepC);
				menuEntry('Right Plane Sweep Ask...', RightANNSweepAskC);
				menuEntry('Right Projection On X', RightANNHeuristicC);
				menuEntry('Right Projection On XY', RightANNHeuristicXYC);
				menuEntry('Right Plane Sweep Dynamic', RightPlaneSweepDynamicC);
				menuEntry('All Plane Sweep', AllANNSweepC);
				menuEntry('All Plane Sweep Ask...', AllNNSweepAskC);
				menuEntry('All Projection On X', AllANNHeuristicC);
				menuEntry('All Projection On XY', AllANNHeuristicXYC);
				menuEntry('All Plane Sweep Dynamic', AllPlaneSweepDynamicC);
				if isRealPoints then begin
					menuEntry('All Boxes', ANNBoxesC);
					menuEntry('All Boxes Ask...', ANNBoxesAskC);
					menuEntry('All Boxes Dynamic', ANNBoxesDynamicC);
					menuEntry('Left In Sector', NearestNeighborInSectorC);
					menuEntry('Left In Sector Ask...', NearestNeighborInSectorAskC);
					menuEntry('Left In Sector Dynamic', NearestNeighborInSectorDynamicC);
				end;
				endSubMenu;
				if isRealPoints then begin
					beginSubMenu('Voronoi Diagram');
					menuEntry('Plane Sweep Fortune', VoronoiDiagramC);
					menuEntry('Plane Sweep Fortune Ask...', VoronoiDiagramAskC);
					menuEntry('Dynamic Fortune', VoronoiDiagramDynamicC);
					menuEntry('Plane Sweep Seidel', VoronoiDiagramSeidelC);
					menuEntry('Plane Sweep Seidel Ask...', VoronoiDiagramSeidelAskC);
					menuEntry('Dynamic Seidel', VoronoiDiagramSeidelDynamicC);
					menuEntry('Delaunay Triangulation', DelaunayC);
					menuEntry('Delaunay Triangulation Dynamic', DelaunayDynamicC);
					menuEntry('Delaunay Triangulation Grid', DelaunayGridC);
					menuEntry('Delaunay Triangulation Grid Dynamic', DelaunayGridDynamicC);
					endSubMenu;
				end;
(* begin Marco Ciot *)
				beginSubMenu('Relative Neighborhood');
				menuEntry('Gabriel Graph Simple', GGsimpleC);
				menuEntry('Relative Neighborhood Graph False', RNGfalseC);
				menuEntry('Gabriel Graph Matula Sokal', GGmatulaSokalC);
				menuEntry('Relative Neighborhood Graph Supowit', RNGsupowitC);
				endSubMenu;
(* end Marco Ciot *)
				if isRealPoints then begin
					beginSubMenu('Euclidean Minimum Spanning Tree');
					menuEntry('Nearest Neighbor In Octant', EMSTC);
					menuEntry('Nearest Neighbor In Octant Dynamic', EMSTDynamicC);
					menuEntry('Voronoi Diagram', EMSTVoronoiC);
					menuEntry('Voronoi Diagram Dynamic', EMSTVoronoiDynamicC);
					endSubMenu;
					beginSubMenu('Clustering');
					menuEntry('Smallest Area', SmallestAreaClusterC);
					menuEntry('Smallest Area Intersect', SmallestAreaIntersectClusterC);
					menuEntry('Smallest Area...', SmallestAreaClusterAskC);
					menuEntry('Smallest Area Intersect...', SmallestAreaIntersectClusterAskC);
					menuEntry('Smallest Size', SmallestSizeClusterC);
					menuEntry('Circle Components...', ClusterC);
					menuEntry('Convex Polygon Components...', ConvexPolygonClusterC);
					endSubMenu;
				end;
				beginSubMenu('Traveling Salesman Heuristics');
				if isRealPoints then
					menuEntry('Euclidean Minimum Spanning Tree', SalesmanEMSTC);
				menuEntry('Nearest Neighbor', SalesmanNNC);
				menuEntry('Convex Hull', SalesmanConvexHullC);
				endSubMenu;
				if isRealPoints & (currentLength > 2) then begin
					beginSubMenu('Traveling Salesman Optimal');
					menuEntry('Simple', TSPSimpleC);
{$IFC false }
					menuEntry('Simple Branch, Heuristic Initial Tour', TSPSimpleBranchInitC);
					menuEntry('Conserving Convex Hull Order', TSPConsConvHullC);
					menuEntry('No Intersecting Tours', TSPNoIntersC);
					menuEntry('2 NN Branch and Bound', TSPBranchAndBoundC);
{$ENDC }
					menuEntry('1 Tree Branch and Bound', TSP1TreeC);
					menuEntry('1 Tree Dynamic', TSP1TreeDynamicC);
{$IFC OptionalCommand }
					menuEntry('Voronoi + TSP Dynamic', VoronoiAndTSPDynamicC);
{$ENDC }
{$IFC false }
					menuEntry('TSP Diagram', TSPDiagrC);
{$ENDC }
					endSubMenu;
				end;
				if not member(ith(1), convexPolygon) then begin
					beginSubMenu('Convex Hull');
					menuEntry('No Eliminate', ConvexHullOnlyC);
					menuEntry('Eliminate', ConvexHullEliminateC);
					menuEntry('Hull And Internal', ConvexHullC);
					menuEntry('Divide And Conquer', ConvexHullDIC);
					menuEntry('Dynamic', ConvexHullDynamicC);
					endSubMenu;
					beginSubMenu('Sort');
					menuEntry('By XY Quicksort', QuickSortByXYC);
					menuEntry('By YX Quicksort', QuickSortByYXC);
					menuEntry('By XY Heapsort', HeapSortByXYC);
					menuEntry('By YX Heapsort', HeapSortByYXC);
					endSubMenu;
					beginSubMenu('Make');
					menuEntry('Polyline', MakePolyLineC);
					menuEntry('Polygon', MakePolygonC);
					menuEntry('Simple Polygon', MakeSimplePolygonC);
					menuEntry('Star-Shaped Polygon', MakeStarShapedPolygonC);
					menuEntry('Aligned Points X', AlignedPointsXC);
					menuEntry('Aligned Points Y', AlignedPointsYC);
					menuEntry('Circles...', MakeCirclesC);
					menuEntry('Directed Graph', DirectedGraphC);
					menuEntry('Random Graph...', RandomGraphC);
					menuEntry('Empty Graph', EmptyGraphC);
					menuEntry('Empty Directed Graph', EmptyDirectedGraphC);
					menuEntry('Full Graph', FullGraphC);
					menuEntry('Full Directed Graph', FullDirectedGraphC);
{$IFC OptionalCommand }
					menuEntry('Random Spanning Tree', MakeRandomSpanningTreeC);
{$ENDC }
					endSubMenu;
					beginSubMenu('Find');
{$IFC OptionalCommand }
					menuEntry('Henneberg Counterexample', HennebergCounterExampleC);
{$ENDC }
					menuEntry('Random Triangulable Graph', TriangulableGraphC);
					menuEntry('Random Rigid Graph', RigidGraphC);
					menuEntry('Random Rigid Simple Graph', RigidSimpleGraphC);
					menuEntry('Random Rigid Not Triangulable Graph', RigidNotTriangulableGraphC);
					menuEntry('Graph Satisfying Formula', FormulaGraphC);
					endSubMenu;
				end;
				beginSubMenu('Project');
				menuEntry('On X', XProjectC);
				menuEntry('On Y', YProjectC);
				endSubMenu;
				beginSubMenu('Minimal Area Disk');
				menuEntry('Randomized', MiniDiskC);
				menuEntry('Randomized Dynamic', MiniDiskDynamicC);
				menuEntry('Approximate', SimpleMiniDiskC);
				menuEntry('Approximate Dynamic', MiniDiskApproximateDynamicC);
				menuEntry('Dynamic', MiniDiskRandomizedApproximateDynamicC);
				endSubMenu;
				menuEntry('Order Type', OrderTypeC);
				menuEntry('Create Star Bushel', StarBushelC);
{$IFC OptionalCommand }
				if currentLength > 2 then begin
					beginSubMenu('Distance Distribution');
					menuEntry('Defaults', DistanceDistributionNoAskC);
					menuEntry('Ask...', DistanceDistributionC);
					menuEntry('Defaults Dynamic', DistanceDistributionDynamicC);
					endSubMenu;
				end;
{$ENDC }
			end
			else if (length = 2) & currentArguments1IsHomogenVector & currentArguments2IsHomogenVector & isPointVector(vector(ith(1))) & isPointVector(vector(ith(2))) then begin
				menuEntry('Create Grid', CreateGridC);
{$IFC OptionalCommand }
				menuEntry('Distance Distribution Dynamic', DistanceDistributionDynamicTwoVectorsC);
{$ENDC }
				if vector(ith(1)).length = vector(ith(2)).length then begin
					menuEntry('Create Line Segment Vector', CreateLineSegmentVectorC);
					menuEntry('Create Rectangle Vector', CreateRectangleVectorC);
				end;
			end
			else if (length = 2) & (currentArguments1IsHomogenVector & isPointVector(vector(ith(1))) & member(ith(2), circle) | currentArguments2IsHomogenVector & isPointVector(vector(ith(2))) & member(ith(1), circle)) then
				menuEntry('In Circle', InCircleC)
			else if (length = 3) & member(ith(1), vector) & (vector(ith(1)).length > 0) & vector(ith(1)).sameProperty(isVerticalSegment) & member(ith(2), point2d) & member(ith(3), point2d) then
				menuEntry('Shortest Path', ShortestPathC)
{$IFC CompareANN }
			else if (length = 2) & (currentArguments1IsHomogenVector & isPointVector(vector(ith(1))) & member(ith(2), rectangle) | currentArguments2IsHomogenVector & isPointVector(vector(ith(2))) & member(ith(1), rectangle)) then
				menuEntry('Compare Nearest Neighbor Algorithms', CompareNearestNeighborAlgorithmsC)
{$ENDC }
	end;

function computeClosestPairDynamic (ps: pointVector): obj;
	begin
		computeClosestPairDynamic := ps.closestPairPoints(false);
	end;

function computeVoronoiDiagramDynamic (ps: pointVector): obj;
	begin
		computeVoronoiDiagramDynamic := ps.voronoiDiagramSweepLine(false);
	end;

function computeVoronoiDiagramSeidelDynamic (ps: pointVector): obj;
	begin
		computeVoronoiDiagramSeidelDynamic := ps.voronoiDiagramSweepLineSeidel(false);
	end;

function computeEMSTNNDynamic (ps: pointVector): obj;
	var
		uf: unionFindSimple;
	begin
		computeEMSTNNDynamic := ps.EMST(1, uf);
	end;

function computeEMSTVDDynamic (ps: pointVector): obj;
	var
		vd: voronoiDiagram;
		uf: unionFindSimple;
	begin
		vd := ps.voronoiDiagramSweepLineSeidel(false);
		computeEMSTVDDynamic := vd.EMST(1, uf);
		vd.freeAll;
	end;

function computeTSPDynamic (ps: pointVector): obj;
	begin
		computeTSPDynamic := ps.TSP1TreeBranchAndBound;
	end;

function computeDelaunayDynamic (ps: pointVector): obj;
	var
		vd: voronoiDiagram;
	begin
		vd := ps.voronoiDiagramSweepLineSeidel(false);
		computeDelaunayDynamic := vd.delaunayTriangulation;
		vd.freeAll;
	end;

function computeDelaunayGridDynamic (ps: pointVector): obj;
	begin
		computeDelaunayGridDynamic := ps.gridTriangulation(1E-3, 1.0);
	end;

function computeVoronoiAndTSPDynamic (ps: pointVector): obj;
	var
		v: vector;
	begin
		v := newVector(2);
		v.append(ps.voronoiDiagramSweepLineSeidel(false));
		v.append(ps.TSP1TreeBranchAndBound);
		computeVoronoiAndTSPDynamic := v;
	end;

function computeLeftNearestNeighbor (ps: pointVector): obj;
	begin
		computeLeftNearestNeighbor := ps.nearestNeighborSweep(leftNeighbors, false);
	end;

function computeRightNearestNeighbor (ps: pointVector): obj;
	begin
		computeRightNearestNeighbor := ps.nearestNeighborSweep(rightNeighbors, false);
	end;

function computeAllNearestNeighbor (ps: pointVector): obj;
	begin
		computeAllNearestNeighbor := ps.nearestNeighborSweep(allNeighbors, false);
	end;

function computeAllNearestNeighborBoxes (ps: pointVector): obj;
	begin
		computeAllNearestNeighborBoxes := ps.nearestNeighborBoxes(false);
	end;

function computeAllNearestNeighborInSector (ps: pointVector): obj;
	begin
		computeAllNearestNeighborInSector := ps.nearestNeighborInSector(false, cos(pi / 6), -sin(pi / 6), cos(pi / 6), sin(pi / 6));
	end;

function computeConvexHull (ps: pointVector): obj;
	begin
		computeConvexHull := ps.convexHullGrahamScan;
	end;

function computeAllMiniDisks (randomized: boolean; ps: pointVector): obj;
	var
		v: vector;
		c: circle;
	begin
		v := newVector(2);
		if randomized then
			c := ps.minimalDisk(false)
		else
			c := ps.simpleMinimalDisk;
		v.append(c);
		v.append(c.getRadius);
		computeAllMiniDisks := v;
	end;

function computeMiniDisk (ps: pointVector): obj;
	begin
		computeMiniDisk := computeAllMiniDisks(true, ps);
	end;

function computeSimpleMiniDisk (ps: pointVector): obj;
	begin
		computeSimpleMiniDisk := computeAllMiniDisks(false, ps);
	end;

function computeBothMiniDisks (ps: pointVector): obj;
	var
		v: vector;
		c: circle;
	begin
		v := newVector(4);
		c := ps.minimalDisk(false);
		v.append(c);
		v.append(c.getRadius);
		c := ps.simpleMinimalDisk;
		v.append(c);
		v.append(c.getRadius);
		computeBothMiniDisks := v;
	end;

function compareVertical (a, b: lineSegment): boolean;
	begin
		compareVertical := pointLessThanXY(a.p, b.p);
	end;

	const
		isDirectedGraphDefault = false;
		edgesGivenMethodDefault = givenByNumber;
		numberOfEdgesDefault = 20;
		edgeProbabilityDefault = 0.5;
		isConnectedGraphDefault = true;
		stdUseFormula = false;

	var
		isDirectedGraph, isConnectedGraph, useFormula: boolean;
		edgesGivenMethod: integer;
		numberOfEdges, edgeProbability: extended;
		psl: longInt;

{$IFC OptionalCommand }
	var
		defaultNumberOfBins, defaultWidth, defaultHeight: extended;
		globalPS: pointVector;

function doDistanceDistribution (ps: pointVector): obj;
	begin
{$IFC THINK_PASCAL }
		doDistanceDistribution := ps.distanceDistribution(trunc(defaultNumberOfBins), defaultWidth, defaultHeight);
{$ELSEC }
		doDistanceDistribution := ps.distanceDistribution(truncLong(defaultNumberOfBins), defaultWidth, defaultHeight);
{$ENDC }
	end;

{$IFC NOT THINK_PASCAL }
{$PUSH }
{$PRAGMAC warn_unusedarg off }
{$ENDC }
function doVectorDistanceDistribution (s: sequence): obj;
	begin
{$IFC THINK_PASCAL }
		doVectorDistanceDistribution := globalPS.distanceDistribution(trunc(defaultNumberOfBins), defaultWidth, defaultHeight);
{$ELSEC }
		doVectorDistanceDistribution := globalPS.distanceDistribution(truncLong(defaultNumberOfBins), defaultWidth, defaultHeight);
{$ENDC }
	end;
{$IFC NOT THINK_PASCAL }
{$POP }
{$ENDC }

procedure doAppend (x: obj);
	begin
		globalPS.append(x);
	end;

procedure vectorDistanceDistribution (v1, v2: pointVector);
	begin
		globalPS := newPointVector(v1.length + v2.length);
		v1.forAll(doAppend);
		v2.forAll(doAppend);
		currentArguments.dynamicExecutionNoHighlightShowType(doVectorDistanceDistribution);
		globalPS.free;
	end;
{$ENDC }

function isVisibleNumberOfEdges: boolean;
	begin
		isVisibleNumberOfEdges := edgesGivenMethod = givenByNumber;
	end;

function isVisibleProbability: boolean;
	begin
		isVisibleProbability := edgesGivenMethod = givenByProbability;
	end;

function edgeProbabilityOk (x: extended): boolean;
	begin
		edgeProbabilityOk := (0 <= x) & (x <= 1);
	end;

function defaultNumberOfEdgesOk (x: extended): boolean;
	begin
		defaultNumberOfEdgesOk := (x = trunc(x)) & (isConnectedGraph & (x >= psl - 1) | not isConnectedGraph & (x >= 0)) & ((x <= (psl - 1) * psl div 2) | isDirectedGraph & (x <= (psl - 1) * psl));
	end;

function createStarBushel (ps: pointVector): lineSegmentVector;
	var
		result: lineSegmentVector;
		l: lineSegment;
		origin: point2d;
		i: seqIndex;
	begin
		if ps.length > 1 then begin
			currentVectorLength := ps.length - 1;
			new(result);
			failNil(Handle(result));
			result.init;
			origin := point2d(ps.ith(1));
			for i := 2 to ps.length do begin
				new(l);
				failNil(Handle(l));
				l.p := point2d(origin.duplicate);
				l.q := point2d(ps.elements^^[i].duplicate);
				result.append(l);
			end;
			createStarBushel := result;
		end
		else
			createStarBushel := nil;
	end;

{$IFC CompareANN }
function compareANN (r: rectangle; gpv: pointVector): pointVector;
	var
		r1, r2: linesegmentVector;
		pv: pointVector;
		c: longInt;
	begin
		pv := pointVector(gpv.duplicateAll);
		c := 0;
		while not userAbort do begin
			c := c + 1;
			if c mod 20 = 0 then
				displayInfoString(StringOf(c : 1));
			currentRandomConstraint := r;
			pv.randomChange;
			r1 := pv.nearestNeighborSweep(leftNeighbors, false);
			r2 := pv.nearestNeighborProjection(leftNeighbors, false);
			if (r1 <> nil) & (r2 <> nil) & not r1.equal(r2) then begin
				r1.freeAll;
				r2.freeAll;
				displayInfoString(StringOf(c : 1));
				compareANN := pv;
				exit(compareANN);
			end;
			if r1 <> nil then
				r1.freeAll;
			if r2 <> nil then
				r2.freeAll;
		end;
		pv.freeAll;
		displayInfoString(StringOf(c : 1));
		compareANN := nil;
	end;
{$ENDC }

function pointVector.createRandomGraph (isDirectedGraph, isConnectedGraph: boolean; edgesGivenMethod: integer; edgeProbability: extended; numberOfEdges: seqIndex0): simpleUndirectedGraph;
	var
		result: simpleUndirectedGraph;
		dResult: directedGraph;
	begin
		if isDirectedGraph then begin
			new(dResult);
			result := dResult;
		end
		else
			new(result);
		failNil(Handle(result));
		result.vertices := vector(duplicateAll);

		if edgesGivenMethod = givenByNumber then begin
			if isDirectedGraph then
				createRandomGraphByEdgeNumber(numberOfEdges, isDirectedGraph, isConnectedGraph, result)
			else
				createRandomGraphByEdgeNumber(numberOfEdges, isDirectedGraph, isConnectedGraph, result);
		end
		else if edgesGivenMethod = givenByProbability then
			createRandomGraphByProbability(edgeProbability, isDirectedGraph, isConnectedGraph, result);
{$IFC AnimationEnabled }
		if userAbort then begin
			createRandomGraph := nil;
			exit(createRandomGraph);
		end;
{$ENDC }
		if isDirectedGraph then
			with directedGraph(result) do begin
				adjacency := newHomogenVector(length);
				initAdjacency;
			end;
		createRandomGraph := result;
	end;

function defaultFormulaVisible: boolean;
	begin
		defaultFormulaVisible := useFormula;
	end;

function callCreateFullGraph (ps: pointVector): simpleUndirectedGraph;
	begin
		callCreateFullGraph := ps.createRandomGraph(false, false, givenByNumber, 0, ps.length * (ps.length - 1) div 2);
	end;

function callCreateFullDirectedGraph (ps: pointVector): simpleUndirectedGraph;
	begin
		callCreateFullDirectedGraph := ps.createRandomGraph(true, false, givenByNumber, 0, ps.length * (ps.length - 1));
	end;

procedure callCreateRandomGraph (ps: pointVector);
	var
		h: vector;
		maxEdges: longInt;
		errorPosition: integer;
		dummy: boolean;

	function generateGraph: obj;
		var
			r1: simpleUndirectedGraph;
			r2: undirectedGraph;
		begin
{$IFC THINK_PASCAL }
			r1 := ps.createRandomGraph(isDirectedGraph, isConnectedGraph, edgesGivenMethod, edgeProbability, trunc(numberOfEdges));
{$ELSEC }
			r1 := ps.createRandomGraph(isDirectedGraph, isConnectedGraph, edgesGivenMethod, edgeProbability, truncLong(numberOfEdges));
{$ENDC }
			if r1 = nil then
				generateGraph := nil
			else if isDirectedGraph then
				generateGraph := r1
			else begin
				new(r2);
				failNil(Handle(r2));
				r2.initializeFromSimpleUndirectedGraph(r1);
				r1.freeAll;
				generateGraph := r2;
			end;
		end;

{$IFC NOT THINK_PASCAL }
{$PUSH }
{$PRAGMAC warn_unusedarg off }
{$ENDC }
	function testGraph (var g: directedGraph; var proof: homogenVector): boolean;
		var
			errorPosition: integer;
		begin
			testGraph := isGraphOk(g, defaultFormula, true, errorPosition);
		end;
{$IFC NOT THINK_PASCAL }
{$POP }
{$ENDC }

	begin
		psl := ps.length;
		maxEdges := (psl - 1) * psl;
		if not isDirectedGraph then
			maxEdges := maxEdges div 2;
		if numberOfEdges > maxEdges then
			numberOfEdges := maxEdges;

		h := newVector(9);
		h.append(getBooleanHandler(7, @isDirectedGraph, isDirectedGraphDefault, nil));
		h.append(getBooleanHandler(13, @isConnectedGraph, isConnectedGraphDefault, nil));
		h.append(getEnumerationHandler(4, @edgesGivenMethod, edgesGivenMethodDefault, nil, givenByProbability));
		h.append(getExtendedHandler(8, @edgeProbability, edgeProbabilityDefault, @isVisibleProbability, @edgeProbabilityOk));
		h.append(getExtendedHandler(9, @numberOfEdges, numberOfEdgesDefault, @isVisibleNumberOfEdges, @defaultNumberOfEdgesOk));
		h.append(getStaticTextHandler(10, '', false, @isVisibleNumberOfEdges));
		h.append(getStaticTextHandler(11, '', false, @isVisibleProbability));
		h.append(getBooleanHandler(14, @useFormula, stdUseFormula, nil));
		h.append(getStringHandler(15, @defaultFormula, stdDefaultFormula, @defaultFormulaVisible));

		if processDialogBox(130, 3, 12, 0, alwaysTrue, h) then
			if useFormula then begin
				dummy := isGraphOk(nil, defaultFormula, false, errorPosition);
				if not showErrorInDefaultFormula(errorPosition) then
					findRandomObject(generateGraph, testGraph);
			end
			else
{$IFC THINK_PASCAL }
				currentResult.appendNonNil(ps.createRandomGraph(isDirectedGraph, isConnectedGraph, edgesGivenMethod, edgeProbability, trunc(numberOfEdges)));
{$ELSEC }
				currentResult.appendNonNil(ps.createRandomGraph(isDirectedGraph, isConnectedGraph, edgesGivenMethod, edgeProbability, truncLong(numberOfEdges)));
{$ENDC }
	end;

function angleOk (x: extended): boolean;
	begin
		angleOk := (-90 < x) & (x < 90);
	end;

procedure pointVector.execute (item: integer);

	const
{$IFC OptionalCommand }
		createTwoPointVectors = [CreateGridC, CreateLineSegmentVectorC, CreateRectangleVectorC, DistanceDistributionDynamicTwoVectorsC];
{$ELSEC }
		createTwoPointVectors = [CreateGridC, CreateLineSegmentVectorC, CreateRectangleVectorC];
{$ENDC }

	var
		ps, ps2: pointVector;
		uf: unionFindSimple;

	function callShortestPath (v: vector; p1, p2: point2d): polyLine;
		var
			points: pointVector;
			i: seqIndex0;
		begin
			points := newPointVector(2 * v.length);
			v.sortAll(compareVertical);
			for i := 1 to v.length do
				with lineSegment(v.elements^^[i]) do begin
					if p.getY < q.getY then begin
						points.elements^^[2 * i - 1] := p;
						points.elements^^[2 * i] := q;
					end
					else begin
						points.elements^^[2 * i - 1] := q;
						points.elements^^[2 * i] := p;
					end;
				end;
			points.length := 2 * v.length;
			callShortestPath := points.shortestPath(p1, p2);
			points.free;
		end;

	procedure callConvexHull (ps: pointVector);
{$IFC AnimationEnabled }
		label
			9999;
{$ENDC }
		var
			hullLength, innerLength, i: seqIndex0;
			innerPoints: pointVector;
			hull: convexPolygon;
			t: Handle;
		begin
			if ps.getType <> pointVectorType then begin
				t := Handle(ps.elements);
				if HandToHand(t) <> noErr then
					failNil(nil);
			end;
			ps.convexHull(true, true, hullLength, innerLength); (* 'innerLength' is used as dummy *)
{$IFC AnimationEnabled }
			if userAbort then
				goto 9999;
{$ENDC }
			innerLength := ps.length - hullLength;
			if innerLength > 0 then begin
				innerPoints := newPointVector(innerLength);
				innerPoints.length := innerLength;
				for i := 1 to innerLength do
					innerPoints.elements^^[i] := ps.elements^^[hullLength + i].duplicate;
			end
			else
				innerPoints := nil;
			currentVectorLength := hullLength;
			new(hull);
			failNil(Handle(hull));
			hull.init;
			hull.length := hullLength;
			for i := 1 to hullLength do
				hull.elements^^[i] := ps.elements^^[i].duplicate;
			currentResult.append(hull);
			currentResult.appendNonNil(innerPoints);
{$IFC AnimationEnabled }
9999:
{$ENDC }
			if ps.getType <> pointVectorType then begin
{$IFC THINK_PASCAL }
				disposHandle(Handle(ps.elements));
{$ELSEC }
				DisposeHandle(Handle(ps.elements));
{$ENDC }
				ps.elements := objArrayH(t);
			end;
		end;

	procedure callConvexHullOnly (eliminate: boolean; ps: pointVector);
{$IFC AnimationEnabled }
		label
			9999;
{$ENDC }
		var
			hullLength, i: seqIndex0;
			hull: convexPolygon;
			t: Handle;
		begin
			if ps.getType <> pointVectorType then begin
				t := Handle(ps.elements);
				if HandToHand(t) <> noErr then
					failNil(nil);
			end;
			ps.convexHull(eliminate, eliminate, hullLength, i);(* 'i' is used as dummy *)
{$IFC AnimationEnabled }
			if userAbort then
				goto 9999;
{$ENDC }
			currentVectorLength := hullLength;
			new(hull);
			failNil(Handle(hull));
			hull.init;
			hull.length := hullLength;
			for i := 1 to hullLength do
				hull.elements^^[i] := ps.elements^^[i].duplicate;
			currentResult.append(hull);
{$IFC AnimationEnabled }
9999:
{$ENDC }
			if ps.getType <> pointVectorType then begin
{$IFC THINK_PASCAL }
				disposHandle(Handle(ps.elements));
{$ELSEC }
				DisposeHandle(Handle(ps.elements));
{$ENDC }
				ps.elements := objArrayH(t);
			end;
		end;

	procedure callTravelingSalesman (p: polygon2d);
		begin
			if p <> nil then begin
				currentResult.append(p);
				currentResult.append(p.perimeterMP);
			end;
		end;

	function toPointVector (x: vector): pointVector;
		var
			pv: pointVector;
		begin
			new(pv);
			failNil(Handle(pv));
			pv.shallowInitializeFromObject(x);
			toPointVector := pv;
		end;

	function callSortByXY (p: pointVector; quickSort, isXY: boolean): pointVector;
		var
			r: pointVector;
		begin
			r := pointVector(p.duplicateAll);
			if quickSort then
				if isXY then
					r.sortAll(pointLessThanXY)
				else
					r.sortAll(pointLessThanYX)
			else if isXY then
				r.heapSort(pointGreaterThanXY)
			else
				r.heapSort(pointGreaterThanYX);
			callSortByXY := r;
		end;

	procedure callNearestNeighborInSector (ps: pointVector);
		var
			cosLowerAngle, sinLowerAngle, cosUpperAngle, sinUpperAngle: extended;

		procedure cosSin (a: extended; var cosAngle, sinAngle: extended);
			begin
				a := a * oneDegree;
				cosAngle := cos(a);
				sinAngle := sin(a);
			end;

		begin
			if getUserParameter2('Lower Angle (in degree)', 'Upper Angle (in degree)', lowerAngle, upperAngle, lowerAngle, upperAngle, @angleOk, @angleOk) then begin
				cosSin(lowerAngle, cosLowerAngle, sinLowerAngle);
				cosSin(upperAngle, cosUpperAngle, sinUpperAngle);
				currentResult.appendNonNil(ps.nearestNeighborInSector(true, cosLowerAngle, sinLowerAngle, cosUpperAngle, sinUpperAngle));
			end;
		end;

	function callContainedInCircle (p: pointVector; c: circle): markedRealPoint;
		var
			mp: markedRealPoint;
		begin
			new(mp);
			failNil(Handle(mp));
			mp.init;
			mp.mark.value := stringOf(p.containedInCircle(c));
			mp.x := c.x;
			mp.y := c.y;
			callContainedInCircle := mp;
		end;

	procedure callMiniDisk (c: circle);
		begin
			if c <> nil then begin
				currentResult.append(c);
				currentResult.append(c.getRadius);
			end;
		end;

	function callAlign (p: pointVector; isXDirection: boolean): pointVector;
		var
			result: pointVector;
			first: point2d;
			i: seqIndex;
			v: extended;
		begin
			result := pointVector(p.duplicateAll);
			first := point2d(p.elements^^[1]);
			if isXDirection then begin
				v := first.getX;
				for i := 2 to result.length do
					with point2d(result.elements^^[i]) do
						setX(v);
			end
			else begin
				v := first.getY;
				for i := 2 to result.length do
					with point2d(result.elements^^[i]) do
						setY(v);
			end;
			callAlign := result;
		end;

	function callMakeCircles (p: pointVector): homogenVector;
		var
			result: homogenVector;
			realRadius: real;

		procedure makeCircle (o: point2d);
			var
				c: circle;
			begin
				new(c);
				failNil(Handle(c));
				with c do begin
					x := o.getX;
					y := o.getY;
					radius := realRadius;
				end;
				result.append(c);
			end;

		begin
			if getUserParameter1('Radius', defaultRadius, defaultRadius, @isPositive) then begin
				realRadius := defaultRadius;
				result := newHomogenVector(p.length);
				p.forAll(makeCircle);
				callMakeCircles := result;
			end
			else
				callMakeCircles := nil;
		end;

	begin
		if item = ShortestPathC then
			with currentArguments do
				currentResult.appendNonNil(callShortestPath(vector(ith(1)), point2d(ith(2)), point2d(ith(3))))
		else begin
			if currentArguments.length > 1 then
				if item in createTwoPointVectors then begin
					ps := toPointVector(vector(currentArguments.ith(1)));
					ps2 := toPointVector(vector(currentArguments.ith(2)));
				end
				else if (item = InCircleC) | (item = CompareNearestNeighborAlgorithmsC) then
					if member(currentArguments.ith(1), vector) then
						ps := toPointVector(vector(currentArguments.ith(1)))
					else
						ps := toPointVector(vector(currentArguments.ith(2)))
				else begin
					new(ps);
					failNil(Handle(ps));
					ps.initFromList(currentArguments);
				end
			else if member(currentArguments.ith(1), pointVector) then
				ps := pointVector(currentArguments.ith(1))
			else
				ps := toPointVector(vector(currentArguments.ith(1)));
			case item of
{$IFC Tutorial }
				0: 
					currentResult.appendNonNil(ps.gravityCenter);
{$ENDC }
{$IFC OptionalCommand }
				MakeRandomSpanningTreeC: 
					currentResult.appendNonNil(makeRandomSpanningTree(ps));
				DistanceDistributionNoAskC: 
					currentResult.appendNonNil(doDistanceDistribution(ps));
				DistanceDistributionC: 
					if getUserParameter3('Number of Bins', defaultNumberOfBins, defaultNumberOfBins, @isNatural, 'Width', defaultWidth, defaultWidth, @isNatural, 'Height', defaultHeight, defaultHeight, @isNatural) then
						currentResult.appendNonNil(doDistanceDistribution(ps));
				DistanceDistributionDynamicC: 
					ps.dynamicExecutionNoHighlightShowType(doDistanceDistribution);
				DistanceDistributionDynamicTwoVectorsC: 
					vectorDistanceDistribution(ps, ps2);
				VoronoiAndTSPDynamicC: 
					ps.dynamicExecutionHighlightShowType(computeVoronoiAndTSPDynamic);
{$ENDC }
				ClosestPairC, ClosestPairAskC: 
					currentResult.appendNonNil(ps.closestPairPoints(item = ClosestPairAskC));
(* inserted by JM *)
				ClosestPairModifiedC, ClosestPairModifiedAskC: 
					currentResult.appendNonNil(ps.closestPairModified(item = ClosestPairModifiedAskC));
(*   *)
				VoronoiDiagramSeidelDynamicC: 
					ps.dynamicExecutionHighlightShowType(computeVoronoiDiagramSeidelDynamic);
				VoronoiDiagramDynamicC: 
					ps.dynamicExecutionHighlightShowType(computeVoronoiDiagramDynamic);
				DelaunayDynamicC: 
					ps.dynamicExecutionNoHighlightShowType(computeDelaunayDynamic);
				DelaunayC: 
					currentResult.appendNonNil(computeDelaunayDynamic(ps));
				DelaunayGridC: 
					currentResult.appendNonNil(ps.gridTriangulation(1E-3, 1.0));
				DelaunayGridDynamicC: 
					ps.dynamicExecutionHighlightShowType(computeDelaunayGridDynamic);
(* begin Marco Ciot *)
				GGsimpleC: 
					currentResult.appendNonNil(ps.ggSimple);
				RNGfalseC: 
					currentResult.appendNonNil(ps.rngFalse);
				GGmatulaSokalC: 
					currentResult.appendNonNil(ps.ggMatulaSokal);
				RNGsupowitC: 
					currentResult.appendNonNil(ps.rngSupowit);
(* end Marco Ciot *)
				ClosestPairDynamicC: 
					ps.dynamicExecutionHighlightShowType(computeClosestPairDynamic);
				EMSTDynamicC: 
					ps.dynamicExecutionNoHighlightShowType(computeEMSTNNDynamic);
				EMSTVoronoiDynamicC: 
					ps.dynamicExecutionNoHighlightShowType(computeEMSTVDDynamic);
				TSP1TreeDynamicC: 
					ps.dynamicExecutionNoHighlightShowType(computeTSPDynamic);
				EMSTVoronoiC: 
					currentResult.appendNonNil(computeEMSTVDDynamic(ps));
				StarBushelC: 
					currentResult.appendNonNil(createStarBushel(ps));
				OrderTypeC: 
					currentResult.appendNonNil(ps.orderTypeOne);
				FormulaGraphC: 
					if getUserString1('Formula', defaultFormula) then
						findDefaultFormulaGraph(ps);
				TriangulableGraphC: 
					findTriangulableGraph(ps);
				RigidGraphC: 
					findRigidGraph(ps);
				RigidSimpleGraphC: 
					findTriangulableGraph2(ps);
				RigidNotTriangulableGraphC: 
					findRigidNotTriangulableGraph(ps);
				HennebergCounterExampleC: 
					findHennebergCounterexample(ps);
				ClosestPairHeuristicC: 
					currentResult.appendNonNil(ps.closestPairHeuristic);
				ClosestPairNewHeuristicC: 
					currentResult.appendNonNil(ps.closestPairNewHeuristic);
				ClosestPairDivideAndConquerC: 
					currentResult.appendNonNil(ps.closestPairDivideAndConquer);
				ClosestPairProbabilisticC, ClosestPairProbabilisticAskC: 
					currentResult.appendNonNil(ps.closestPairProbabilistic(item = ClosestPairProbabilisticAskC));
				ClosestPairSlowC: 
					currentResult.appendNonNil(ps.closestPairN2);
				ConvexHullC: 
					callConvexHull(ps);
				ConvexHullOnlyC, ConvexHullEliminateC: 
					callConvexHullOnly(item = ConvexHullEliminateC, ps);
				ConvexHullDIC: 
					currentResult.appendNonNil(ps.convexHullDivideAndConquer);
				ConvexHullDynamicC: 
					ps.dynamicExecutionNoHighlightShowType(computeConvexHull);
				VoronoiDiagramC, VoronoiDiagramAskC: 
					currentResult.appendNonNil(ps.voronoiDiagramSweepLine(item = VoronoiDiagramAskC));
				VoronoiDiagramSeidelC, VoronoiDiagramSeidelAskC: 
					currentResult.appendNonNil(ps.voronoiDiagramSweepLineSeidel(item = VoronoiDiagramSeidelAskC));
				EmptyGraphC, EmptyDirectedGraphC: 
					currentResult.appendNonNil(ps.createRandomGraph(item = EmptyDirectedGraphC, false, givenByNumber, 0, 0));
				RandomGraphC: 
					callCreateRandomGraph(ps);
				FullGraphC: 
					currentResult.append(callCreateFullGraph(ps));
				FullDirectedGraphC: 
					currentResult.append(callCreateFullDirectedGraph(ps));
				DirectedGraphC: 
					currentResult.append(ps.createGraph(directedGraphType));
				MakePolygonC: 
					currentResult.append(ps.convertToPolygon);
				MakePolyLineC: 
					currentResult.append(ps.makePolyLine);
				MakeSimplePolygonC: 
					currentResult.appendNonNil(ps.makeSimplePolygon);
				MakeStarShapedPolygonC: 
					currentResult.appendNonNil(ps.makeStarShapedPolygon);
				MakeCirclesC: 
					currentResult.appendNonNil(callMakeCircles(ps));
				AlignedPointsXC, AlignedPointsYC: 
					currentResult.append(callAlign(ps, item = AlignedPointsXC));
				EMSTC: 
					currentResult.appendNonNil(ps.EMST(1, uf));
				SmallestSizeClusterC: 
					currentResult.appendNonNil(ps.computeSmallestSizeCluster);
				SmallestAreaClusterC, SmallestAreaIntersectClusterC: 
					currentResult.appendNonNil(ps.computeSmallestAreaCluster(item = SmallestAreaIntersectClusterC, ps.length div 2));
				ClusterC, ConvexPolygonClusterC: 
					if getUserParameter1('Number of Components', numberOfClusters, numberOfClusters, @isNatural) then
						if item = ClusterC then
{$IFC THINK_PASCAL }
							currentResult.appendNonNil(ps.cluster(trunc(numberOfClusters), circleObject))
						else
							currentResult.appendNonNil(ps.cluster(trunc(numberOfClusters), convexPolygonObject));
{$ELSEC }
							currentResult.appendNonNil(ps.cluster(truncLong(numberOfClusters), circleObject))
						else
							currentResult.appendNonNil(ps.cluster(truncLong(numberOfClusters), convexPolygonObject));
{$ENDC }
				SmallestAreaIntersectClusterAskC, SmallestAreaClusterAskC: 
					if getUserParameter1('Maximum Number of Components', numberOfClusters, numberOfClusters, @isNatural) then
{$IFC THINK_PASCAL }
						currentResult.appendNonNil(ps.computeSmallestAreaCluster(item = SmallestAreaIntersectClusterAskC, trunc(numberOfClusters)));
{$ELSEC }
						currentResult.appendNonNil(ps.computeSmallestAreaCluster(item = SmallestAreaIntersectClusterAskC, truncLong(numberOfClusters)));
{$ENDC }
				SalesmanEMSTC: 
					callTravelingSalesman(ps.TSPEMST);
				SalesmanNNC: 
					callTravelingSalesman(ps.TSPNN);
				SalesmanConvexHullC: 
					callTravelingSalesman(ps.TSPConvexHull);

				TSP1TreeC: 
					callTravelingSalesman(ps.TSP1TreeBranchAndBound);
				TSPSimpleC: 
					callTravelingSalesman(ps.TSPSimple);

{$IFC false }
				TSPSimpleBranchInitC: 
					callTravelingSalesman(ps.TSPSimpleBranch(true));
				TSPConsConvHullC: 
					callTravelingSalesman(ps.TSPConsConvHullOrder);
				TSPNoIntersC: 
					callTravelingSalesman(ps.TSPNoInters);
				TSPBranchAndBoundC: 
					callTravelingSalesman(ps.TSP2NNBranchAndBound);
				TSPDiagrC: 
					currentResult.appendNonNil(ps.TSPDiagram);
{$ENDC }

				LeftANNHeuristicC, LeftANNHeuristicXYC: 
					currentResult.appendNonNil(ps.nearestNeighborProjection(leftNeighbors, item = LeftANNHeuristicXYC));
				RightANNHeuristicC, RightANNHeuristicXYC: 
					currentResult.appendNonNil(ps.nearestNeighborProjection(rightNeighbors, item = RightANNHeuristicXYC));
				AllANNHeuristicC, AllANNHeuristicXYC: 
					currentResult.appendNonNil(ps.nearestNeighborProjection(allNeighbors, item = AllANNHeuristicXYC));
				XProjectC: 
					currentResult.append(ps.projectOnX);
				YProjectC: 
					currentResult.append(ps.projectOnY);
				CreateGridC: 
					currentResult.appendNonNil(ps.createGrid(ps2));
				CreateLineSegmentVectorC: 
					currentResult.appendNonNil(ps.createLineSegmentVector(ps2));
				CreateRectangleVectorC: 
					currentResult.appendNonNil(ps.createRectangleVector(ps2));
				QuickSortByXYC, HeapSortByXYC, QuickSortByYXC, HeapSortByYXC: 
					currentResult.append(callSortByXY(ps, (item = QuickSortByXYC) | (item = QuickSortByYXC), (item = QuickSortByXYC) | (item = HeapSortByXYC)));
				NearestNeighborInSectorC: 
					currentResult.appendNonNil(ps.nearestNeighborInSector(false, cos(pi / 6), -sin(pi / 6), cos(pi / 6), sin(pi / 6)));
				NearestNeighborInSectorDynamicC: 
					ps.dynamicExecutionNoHighlightShowType(computeAllNearestNeighborInSector);
				NearestNeighborInSectorAskC: 
					callNearestNeighborInSector(ps);
				MiniDiskC: 
					callMiniDisk(ps.minimalDisk(true));
				MiniDiskDynamicC: 
					ps.dynamicExecutionNoHighlightShowType(computeMiniDisk);
				MiniDiskApproximateDynamicC: 
					ps.dynamicExecutionNoHighlightShowType(computeSimpleMiniDisk);
				SimpleMiniDiskC: 
					callMiniDisk(ps.simpleMinimalDisk);
				MiniDiskRandomizedApproximateDynamicC: 
					ps.dynamicExecutionNoHighlightShowType(computeBothMiniDisks);
				InCircleC: 
					if member(currentArguments.ith(1), circle) then
						currentResult.append(callContainedInCircle(ps, circle(currentArguments.ith(1))))
					else
						currentResult.append(callContainedInCircle(ps, circle(currentArguments.ith(2))));
				LeftANNSweepC, LeftANNSweepAskC: 
					currentResult.appendNonNil(ps.nearestNeighborSweep(leftNeighbors, item = LeftANNSweepAskC));
				LeftPlaneSweepDynamicC: 
					ps.dynamicExecutionNoHighlightShowType(computeLeftNearestNeighbor);
				RightANNSweepC, RightANNSweepAskC: 
					currentResult.appendNonNil(ps.nearestNeighborSweep(rightNeighbors, item = RightANNSweepAskC));
				RightPlaneSweepDynamicC: 
					ps.dynamicExecutionNoHighlightShowType(computeRightNearestNeighbor);
				AllANNSweepC, AllNNSweepAskC: 
					currentResult.appendNonNil(ps.nearestNeighborSweep(allNeighbors, item = AllNNSweepAskC));
				AllPlaneSweepDynamicC: 
					ps.dynamicExecutionNoHighlightShowType(computeAllNearestNeighbor);
(* Inserted by PV *)
				ANNBoxesC, ANNBoxesAskC: 
					currentResult.appendNonNil(ps.nearestNeighborBoxes(item = ANNBoxesAskC));
(* *)
				ANNBoxesDynamicC: 
					ps.dynamicExecutionNoHighlightShowType(computeAllNearestNeighborBoxes);
{$IFC CompareANN }
				CompareNearestNeighborAlgorithmsC: 
					with currentArguments do
						if member(ith(1), rectangle) then
							currentResult.appendNonNil(compareANN(rectangle(ith(1)), ps))
						else
							currentResult.appendNonNil(compareANN(rectangle(ith(2)), ps));
{$ENDC }
			end;
			if currentArguments.length > 1 then
				if item in createTwoPointVectors then begin
					dispose(ps);
					dispose(ps2);
				end
				else if (item = InCircleC) | (item = CompareNearestNeighborAlgorithmsC) then
					dispose(ps)
				else
					ps.free
			else if not member(currentArguments.ith(1), pointVector) then
				dispose(ps);
		end;
	end;

procedure initPointVector;
	var
		x: pointVector;
		y: convexPolygon;
		z: polygon2d;
		v: voronoiDiagram;
		w: polyLine;
		ve: voronoiEdge;
		h: histogram;
	begin
		stdSize := initialStdSize;
		optimizeFirst := 3;
		optimizeLast := 1;
		lowerAngle := -30;
		upperAngle := +30;
		defaultRadius := 50;
		numberOfClusters := 2;
		isDirectedGraph := isDirectedGraphDefault;
		edgesGivenMethod := edgesGivenMethodDefault;
		numberOfEdges := numberOfEdgesDefault;
		edgeProbability := edgeProbabilityDefault;
		isConnectedGraph := isConnectedGraphDefault;
		useFormula := stdUseFormula;
{$IFC OptionalCommand }
		defaultNumberOfBins := 10;
		defaultWidth := 100;
		defaultHeight := 100;
{$ENDC }
		new(x);
		failNil(Handle(x));
		prototypes[pointVectorType] := x;
		new(y);
		failNil(Handle(y));
		prototypes[convexPolygonType] := y;
		new(z);
		failNil(Handle(z));
		prototypes[polygon2dType] := z;
		new(v);
		failNil(Handle(v));
		prototypes[voronoiDiagramType] := v;
		new(w);
		failNil(Handle(w));
		prototypes[polyLineType] := w;
		new(ve);
		failNil(Handle(ve));
		ve.init;
		prototypes[voronoiEdgeType] := ve;
		new(h);
		failNil(Handle(h));
		prototypes[histogramType] := h;
	end;

end.
