(*	XYZ GeoBench (eXperimental geometrY Zurich):		*)
(*	An object oriented, interactive workbench for		*)
(*	experimental geometric computation.					*)
(*														*)
(*	 1999 by Peter Schorn, peter.schorn@acm.org		*)
(*	Culmannstrasse 77, CH-8006 Zrich, Switzerland		*)
(*	This software may be copied freely for educational	*)
(*	and research purposes only. For commercial use,		*)
(*	please contact the author.							*)
(*														*)
(*	Version 5.0.5: 30-Sep-1999							*)

unit pointVector2;

{$SETC ApolyLine = TRUE }

interface

uses
	pointVector;

function intersectionFree (h: homogenVector): boolean;

function nonZero (h: homogenVector): boolean;

implementation
{$SETC ATravelingSalesman = TRUE }
const
	infinity = 1000000;

(* inserted by PV *)

function pointVector.makeIdPoints: pointVector;
	var
		r: pointVector;
		ip: idRealPoint;
		i: seqIndex0;
	begin
(* This method was earlier defined in VoronoiSL; it became a seperate function, because it is used in VoronoiDac and ANN too. *)
(* Self is duplicated and to every element of self its position in the array is added. *)
		r := newPointVector(length);
		r.length := length;
		for i := 1 to length do begin
			new(ip);
			failNil(Handle(ip));
			with realPoint(elements^^[i]) do begin
				ip.x := x;
				ip.y := y;
			end;
			ip.id := i;
			r.elements^^[i] := ip;
		end;
		makeIdPoints := r;
	end;   { makeIdPoints }

procedure pointVector.buildXAndYLists (var xList, yList: listMuchLinked);
	var
		xP: pointVector;
		tempVector: homogenVector;
		temp: handleAsObject;
		is: seqIndex;
		k: longInt;
	begin
(* Make two sorted vectors enable to chain: *)
(* self is sorted by y- and its duplicate xPV by x- coordinate. *)
		self.sortAll(pointLessThanYX);        	(* Self sorted by y-coordinate. *)
		xP := self.makeIdPoints;				(* Every point its index is added. *)
		xP.sortAll(pointLessThanXY);			(* xPV sorted by x-coordinate. *)
		xList := listMuchLinked(newInitObj(listMuchLinkedType));
		yList := listMuchLinked(newInitObj(listMuchLinkedType));
		tempVector := newHomogenVector(length);
(* Build first list: *)
(* The points of xPV are chained to a list, elements[i] in temp get a pointer to list[i]. *)
		for is := 1 to length do begin
			new(temp);
			failNil(Handle(temp));
			k := longint(idRealPoint(xP.elements^^[is]).id);
			temp.tr := treeElementH(xList.appendWithPointerBack(self.elements^^[k]));
			tempVector.elements^^[k] := temp
		end;
		xP.freeAll;
(* Build second list and link with first list: *)
(* The points of self are chained to a list. *)
(* Every list element points to its equivalent in the second list. *)
		for is := 1 to length do begin
			yList.appendAndLinkWithOtherList(self.elements^^[is], handleAsObject(tempVector.elements^^[is]).tr);
			tempVector.elements^^[is].free
		end;
		tempVector.free;
  (* Now only the points of self exist. *)
	end;  { buildXAndYLists }

(* *)

function pointVector.EMST (k: seqIndex; var uf: unionFindSimple): spanningTree;
	type
		octantIndex = 0..7;
	var
		g: simpleUndirectedGraph;
		i: seqIndex0;
		j: seqIndex01;
		e: graphEdge;
		nearestNeighbors: array[octantIndex] of seqIndex;
		distances: array[octantIndex] of real;
		oi: octantIndex;
		p, q: realPoint;

	procedure octant (o: octantIndex);
		var
			d: real;
		begin
			d := p.squaredDistanceX(q);
			if d < distances[o] then begin
				distances[o] := d;
				nearestNeighbors[o] := j;
			end;
		end;

	function edgeLessThan (e1, e2: graphEdge): boolean;
		begin
			edgeLessThan := realPoint(g.vertices.elements^^[e1.startVertex]).squaredDistanceX(realPoint(g.vertices.elements^^[e1.endVertex])) < realPoint(g.vertices.elements^^[e2.startVertex]).squaredDistanceX(realPoint(g.vertices.elements^^[e2.endVertex]));
		end;

	begin
		currentVertices := length;
		currentEdges := 8 * length;
		new(g);
		failNil(Handle(g));
		g.init;
		for i := 1 to length do
			g.vertices.elements^^[i] := elements^^[i].duplicateAll;
		g.vertices.length := length;
		for i := 1 to length do begin
			for oi := 0 to 7 do
				distances[oi] := infinity;
			p := realPoint(elements^^[i]);
			for j := i + 1 to length do begin
{$IFC AnimationEnabled }
				if userAbort then begin
					g.freeAll;
					EMST := nil;
					exit(EMST);
				end;
{$ENDC }
				q := realPoint(elements^^[j]);
				octant(4 * ord(q.y > p.y) + 2 * ord(q.x > p.x) + ord(q.y - p.y < q.x - p.x));
			end;
			for oi := 0 to 7 do
				if distances[oi] <> infinity then begin
					new(e);
					failNil(Handle(e));
					e.init;
					e.startVertex := i;
					e.endVertex := nearestNeighbors[oi];
					g.addEdge(e);
				end;
		end;
		g.edges.sortAll(edgeLessThan);
		EMST := g.minimumSpanningTree(k, uf);
		g.freeAll;
	end;

function pointVector.cluster (k: seqIndex; kind: clusterObjectType): homogenVector;
	var
		uf: unionFindSimple;
		l, result: homogenVector;
		i: seqIndex0;
		ll: list;
		pv: pointVector;
		e: spanningTree;
{$IFC AnimationEnabled }
		j: seqIndex0;
{$ENDC }
	begin
		result := newHomogenVector(k);
		if k = 1 then
			result.append(minimalDisk(false))
		else begin
			e := EMST(k, uf);
{$IFC AnimationEnabled }
			if e = nil then begin (* i.e. userAbort *)
				result.freeAll;
				cluster := nil;
				exit(cluster);
			end;
{$ENDC }
			uf.compressAll;
			new(l);
			failNil(Handle(l));
			l.preset(length, listType);
			for i := 1 to length do
				list(l.elements^^[uf.uf^^[i]]).append(elements^^[i]);
			uf.free;
			for i := 1 to length do begin
{$IFC AnimationEnabled }
				if userAbort then begin
					for j := i to length do
						l.elements^^[j].free;
					l.free;
					e.freeAll;
					result.freeAll;
					cluster := nil;
					exit(cluster);
				end;
{$ENDC }
				ll := list(l.elements^^[i]);
				if ll.length > 0 then begin
					new(pv);
					failNil(Handle(pv));
					pv.initFromList(ll);
					case kind of
						circleObject: 
							result.appendNonNil(pv.minimalDisk(false));
						convexPolygonObject: 
							result.appendNonNil(pv.convexHullGrahamScan);
					end;
					pv.free;
				end;
				ll.free;
			end;
			l.free;
			e.freeAll;
		end;
		cluster := result;
	end;

function intersectionFree (h: homogenVector): boolean;
	var
		i, j: seqIndex0;
		c1, c2: circle;
	begin
		intersectionFree := false;
		if h = nil then
			exit(intersectionFree);
		for i := 1 to h.length - 1 do begin
			c1 := circle(h.elements^^[i]);
			for j := i + 1 to h.length do begin
{$IFC AnimationEnabled }
				if userAbort then
					exit(intersectionFree);
{$ENDC }
				c2 := circle(h.elements^^[j]);
				if sqr(c1.x - c2.x) + sqr(c1.y - c2.y) < sqr(c1.radius + c2.radius) then
					exit(intersectionFree);
			end;
		end;
		intersectionFree := true;
	end;

function nonZero (h: homogenVector): boolean;
	var
		i: seqIndex0;
	begin
		nonZero := false;
		if h = nil then
			exit(nonZero);
		for i := 1 to h.length do begin
{$IFC AnimationEnabled }
			if userAbort then
				exit(nonZero);
{$ENDC }
			if circle(h.elements^^[i]).radius = 0 then
				exit(nonZero);
		end;
		nonZero := true;
	end;

function pointVector.computeSmallestSizeCluster: homogenVector;
	var
		i: seqIndex0;
		r: homogenVector;
	begin
		for i := 2 to length div 2 do begin
			r := cluster(i, circleObject);
{$IFC AnimationEnabled }
			if userAbort then begin
				if r <> nil then
					r.freeAll;
				computeSmallestSizeCluster := nil;
				exit(computeSmallestSizeCluster);
			end;
{$ENDC }
			if nonZero(r) & intersectionFree(r) then begin
				computeSmallestSizeCluster := r;
				exit(computeSmallestSizeCluster);
			end
			else
				r.freeAll;
		end;
		computeSmallestSizeCluster := nil;
	end;

function pointVector.computeSmallestAreaCluster (intersectionAllowed: boolean; maxComponents: seqIndex): homogenVector;
	var
		i: seqIndex0;
		r, bestCluster: homogenVector;
		bestArea, newArea: extended;
	begin
		bestCluster := nil;
		if maxComponents > length div 2 then
			maxComponents := length div 2;
		for i := 2 to maxComponents do begin
			r := cluster(i, circleObject);
{$IFC AnimationEnabled }
			if userAbort then begin
				if bestCluster <> nil then
					bestCluster.freeAll;
				if r <> nil then
					r.freeAll;
				computeSmallestAreaCluster := nil;
				exit(computeSmallestAreaCluster);
			end;
{$ENDC }
			if nonZero(r) & (intersectionAllowed | intersectionFree(r)) then begin
				newArea := r.area;
				if bestCluster = nil then begin
					bestArea := newArea;
					bestCluster := r;
				end
				else if newArea < bestArea then begin
					bestCluster.freeAll;
					bestArea := newArea;
					bestCluster := r;
				end
				else
					r.freeAll;
			end
			else
				r.freeAll;
		end;
		computeSmallestAreaCluster := bestCluster;
	end;

function pointVector.TSPEMST: polygon2d;
	var
		s: spanningTree;
		r: vector;
		result: polygon2d;
		uf: unionFindSimple;
	begin
		s := EMST(1, uf);
{$IFC AnimationEnabled }
		if userAbort then begin
			TSPEMST := nil;
			exit(TSPEMST);
		end;
{$ENDC }
{$IFC AnimationEnabled AND ATravelingSalesman }
		if animationFlag[ATravelingSalesmanItem] then begin
			if not animationFlag[AMinimumSpanningTreeItem] then
				s.drawEdges;
			PenSize(3, 3);
			s.drawEdges;
			PenSize(1, 1);
		end;
{$ENDC }
		r := s.TSPEMST;
{$IFC AnimationEnabled }
		if userAbort then begin
			s.freeAll;
			TSPEMST := nil;
			exit(TSPEMST);
		end;
{$ENDC }
		new(result);
		failNil(Handle(result));
		result.length := r.length;
		result.elements := r.elements;
		dispose(r);
		s.freeAll;
		TSPEMST := result;
	end;

function pointVector.TSPNN: polygon2d;
	var
		i: seqIndex0;
		where: seqIndex;
		result: polygon2d;
		d: real;
		p: point2d;
		visited: booleanArrayH;
	begin
		result := newPolygon2d(length);
		result.append(elements^^[1].duplicate);
		visited := booleanArrayH(dynamicArray(length, sizeOf(boolean)));
		visited^^[1] := true;
		for i := 2 to length do
			visited^^[i] := false;
		where := 1;
		while result.length <> length do begin
			p := point2d(elements^^[where]);
			d := infinity;
			for i := 1 to length do begin
{$IFC AnimationEnabled }
				if userAbort then begin
					result.freeAll;
{$IFC THINK_PASCAL }
					disposHandle(Handle(visited));
{$ELSEC }
					DisposeHandle(Handle(visited));
{$ENDC }
					TSPNN := nil;
					exit(TSPNN);
				end;
{$ENDC }
				if not visited^^[i] & (p.squaredDistanceX(point2d(elements^^[i])) < d) then begin
					where := i;
					d := p.squaredDistanceX(point2d(elements^^[i]));
				end;
			end;
			result.append(elements^^[where].duplicate);
			visited^^[where] := true;
{$IFC AnimationEnabled AND ATravelingSalesman }
			if animationFlag[ATravelingSalesmanItem] then begin
				p.drawLine(point2d(elements^^[where]));
				waitForClick(animationFlag[ATravelingSalesmanItem]);
			end;
{$ENDC }
		end;
{$IFC AnimationEnabled AND ATravelingSalesman }
		if animationFlag[ATravelingSalesmanItem] then begin
			point2d(result.elements^^[length]).drawLine(point2d(elements^^[1]));
			waitForClick(animationFlag[ATravelingSalesmanItem]);
		end;
{$ENDC }
{$IFC THINK_PASCAL }
		disposHandle(Handle(visited));
{$ELSEC }
		DisposeHandle(Handle(visited));
{$ENDC }
		TSPNN := result;
	end;

function pointVector.angle (i, j, k: seqIndex): real;
	var
		jDist, kDist: real;
	begin
		jDist := point2d(elements^^[i]).distanceX(point2d(elements^^[j]));
		kDist := point2d(elements^^[i]).distanceX(point2d(elements^^[k]));
		if ((jDist = 0) | (kDist = 0)) then
			angle := infinity
		else
			angle := (point2d(elements^^[j]).squaredDistanceX(point2d(elements^^[k])) - point2d(elements^^[i]).squaredDistanceX(point2d(elements^^[j])) - point2d(elements^^[i]).squaredDistanceX(point2d(elements^^[k]))) / jDist / kDist;
	end;

function pointVector.TSPConvexHull: polygon2d; (* by Martin Mller *)
{$IFC AnimationEnabled }
	label
		9999, 999;
{$ENDC }
	var
		nuHull, nuRest: seqIndex0;
		nb, maxPt, i, k: seqIndex;
		a, d: real;
		p: point2d;
		DeltaDist, PolygonDist: realArrayH;
		NextNb: integerArrayH;
		maxAngle: real;
		saveObj: obj;
		t: Handle;
	begin
		if getType <> pointVectorType then begin
			t := Handle(elements);
			if HandToHand(t) <> noErr then
				failNil(nil);
		end;
		convexHull(true, false, nuHull, nuRest); (* 'nuRest' is used as dummy *)
{$IFC AnimationEnabled }
		if userAbort then
			goto 999;
{$ENDC }
		nuRest := length - nuHull;
		PolygonDist := realArrayH(dynamicArray(length, sizeOf(real)));
		DeltaDist := realArrayH(dynamicArray(nuRest, sizeOf(real)));
		NextNb := integerArrayH(dynamicArray(nuRest, sizeOf(integer)));
{$IFC AnimationEnabled AND ATravelingSalesman }
		if animationFlag[ATravelingSalesmanItem] then begin
			for i := 1 to nuHull do
				point2d(elements^^[i]).drawLine(point2d(elements^^[i mod nuHull + 1]));
			waitForClick(animationFlag[ATravelingSalesmanItem]);
		end;
{$ENDC }
		while (nuRest > 0) do begin
			for i := 1 to nuHull do
				PolygonDist^^[i] := point2d(elements^^[i]).distanceX(point2d(elements^^[i mod nuHull + 1]));
			for i := 1 to nuRest do (* check all remaining points *)
				begin
				DeltaDist^^[i] := infinity;
				p := point2d(elements^^[nuHull + i]); (* i-th element of rest *)
				for k := 1 to nuHull do (* check if [k,k+1] is best edge to insert p *)
					begin
{$IFC AnimationEnabled }
					if userAbort then
						goto 9999;
{$ENDC }
					d := p.distanceX(point2d(elements^^[k])) + p.distanceX(point2d(elements^^[k mod nuHull + 1])) - PolygonDist^^[k];
					if (d < DeltaDist^^[i]) then begin
						DeltaDist^^[i] := d;
						NextNb^^[i] := k;
					end;
				end;
			end;
			maxAngle := -infinity; (* now compute best point according to maximum angle heuristic *)
			for i := 1 to nuRest do begin
{$IFC AnimationEnabled }
				if userAbort then
					goto 9999;
{$ENDC }
				a := angle(nuHull + i, NextNb^^[i], NextNb^^[i] mod nuHull + 1);
				if (a > maxAngle) then begin
					maxAngle := a;
					maxPt := nuHull + i;
				end;
			end;
			nb := NextNb^^[maxPt - nuHull];
{$IFC AnimationEnabled AND ATravelingSalesman }
			if animationFlag[ATravelingSalesmanItem] then begin
				point2d(elements^^[nb]).drawLine(point2d(elements^^[nb mod nuHull + 1])); (* XOR-delete old line *)
				point2d(elements^^[nb]).drawLine(point2d(elements^^[maxPt]));
				point2d(elements^^[nb mod nuHull + 1]).drawLine(point2d(elements^^[maxPt]));
				waitForClick(animationFlag[ATravelingSalesmanItem]);
			end;
{$ENDC }
(* now insert maxPt after nb := NextNb^^[maxPt] : maxPt=nb+1, old successor of nb := nb+2 or 1 *)
			saveObj := elements^^[maxPt];
			elements^^[maxPt] := elements^^[nuHull + 1]; (* free space at nuHull+1: store element at location of maxPt *)
			for i := nuHull + 1 downto nb + 2 do
				elements^^[i] := elements^^[i - 1];
			elements^^[nb + 1] := saveObj;
			nuHull := nuHull + 1;
			nuRest := nuRest - 1;
		end;
{$IFC AnimationEnabled }
9999:
{$ENDC }
{$IFC THINK_PASCAL }
		disposHandle(Handle(DeltaDist));
		disposHandle(Handle(PolygonDist));
		disposHandle(Handle(NextNb));
{$ELSEC }
		DisposeHandle(Handle(DeltaDist));
		DisposeHandle(Handle(PolygonDist));
		DisposeHandle(Handle(NextNb));
{$ENDC }
{$IFC AnimationEnabled }
999:
		if userAbort then
			TSPConvexHull := nil
		else
			TSPConvexHull := polygon2d(convertToPolygon);
{$ELSEC }
		TSPConvexHull := polygon2d(convertToPolygon);
{$ENDC }
		if getType <> pointVectorType then begin
{$IFC THINK_PASCAL }
			disposHandle(Handle(elements));
{$ELSEC }
			DisposeHandle(Handle(elements));
{$ENDC }
			elements := objArrayH(t);
		end;
	end;

function pointVector.projectOnX: pointVector;
	var
		ps: pointVector;
		i: seqIndex;
		y: extended;
	begin
		ps := pointVector(duplicateAll);
		if length > 0 then begin
			y := point2d(elements^^[1]).getY;
			for i := 1 to length do
				point2d(ps.elements^^[i]).setY(y);
		end;
		projectOnX := ps;
	end;

function pointVector.projectOnY: pointVector;
	var
		ps: pointVector;
		i: seqIndex;
		x: extended;
	begin
		ps := pointVector(duplicateAll);
		if length > 0 then begin
			x := point2d(elements^^[1]).getX;
			for i := 1 to length do
				point2d(ps.elements^^[i]).setX(x);
		end;
		projectOnY := ps;
	end;

function pointVector.createGrid (yCoordinates: pointVector): pointVector;
	var
		result: pointVector;
		p: point2d;
		i, j: seqIndex0;
	begin
		result := newPointVector(length * yCoordinates.length);
		for i := 1 to length do
			for j := 1 to yCoordinates.length do begin
{$IFC AnimationEnabled }
				if userAbort then begin
					result.freeAll;
					createGrid := nil;
					exit(createGrid);
				end;
{$ENDC }
				p := point2d(elements^^[i].duplicate);
				p.setY(point2d(yCoordinates.elements^^[j]).getY);
				if (result.length = 0) | not p.equal(result.elements^^[result.length]) then
					result.append(p);
			end;
		createGrid := result;
	end;

function pointVector.createLineSegmentVector (q: pointVector): lineSegmentVector;
	var
		lv: lineSegmentVector;
		l: lineSegment;
		i: seqIndex0;
		pointType: pointTypeRange;
	begin
		currentVectorLength := length;
		new(lv);
		failNil(Handle(lv));
		lv.init;
		if length > 0 then begin
			pointType := elements^^[1].getType;
			for i := 1 to length do begin
{$IFC AnimationEnabled }
				if userAbort then begin
					lv.freeAll;
					createLineSegmentVector := nil;
					exit(createLineSegmentVector);
				end;
{$ENDC }
				new(l);
				failNil(Handle(l));
				l.p := point2d(elements^^[i].duplicate);
				l.q := point2d(q.elements^^[i].convertToType(pointType));
				lv.append(l);
			end;
		end;
		createLineSegmentVector := lv;
	end;

function pointVector.createRectangleVector (q: pointVector): homogenVector;
	var
		rv: homogenVector;
		r: rectangle;
		i: seqIndex0;
	begin
		currentVectorLength := length;
		rv := homogenVector(newInitObj(rectangleVectorType));
		if length > 0 then begin
			for i := 1 to length do begin
{$IFC AnimationEnabled }
				if userAbort then begin
					rv.freeAll;
					createRectangleVector := nil;
					exit(createRectangleVector);
				end;
{$ENDC }
				new(r);
				failNil(Handle(r));
				r.initFromOppositePoints(point2d(elements^^[i]), point2d(q.elements^^[i]));
				rv.append(r);
			end;
		end;
		createRectangleVector := rv;
	end;

function pointVector.makePolyLine: polyLine;
	var
		p: polyLine;
		ps: pointVector;
	begin
		new(p);
		failNil(Handle(p));
		ps := pointVector(duplicateAll);
		p.length := ps.length;
		p.elements := ps.elements;
		dispose(ps);
		makePolyLine := p;
	end;

function pointVector.makeSimplePolygon: polygon2d;
	var
		i, j: seqIndex;
		p, q: point2d;
		result: polygon2d;
	begin
		result := polygon2d(convertToPolygon);
		if length > 2 then begin
			result.swap(1, result.findExtreme(pointLessThanXY));
			result.swap(result.length, result.findExtreme(pointGreaterThanXY));
			p := point2d(result.elements^^[1]);
			q := point2d(result.elements^^[result.length]);
			i := 2;
			j := result.length;
			while i < j do begin
{$IFC AnimationEnabled }
				if userAbort then begin
					result.freeAll;
					makeSimplePolygon := nil;
					exit(makeSimplePolygon);
				end;
{$ENDC }
				if point2d(result.elements^^[i]).whichSideX(p, q) <= 0 then
					i := i + 1
				else begin
					j := j - 1;
					result.swap(i, j);
				end;
			end;
			result.sort(pointLessThanXY, 1, i - 1);
			result.sort(pointGreaterThanXY, i, result.length);
		end;
		makeSimplePolygon := result;
	end;

function pointVector.makeStarShapedPolygon: polygon2d;
	var
		i, j: seqIndex;
		result: polygon2d;
		center: point2d;

	function angleLessThan (p, q: point2d): boolean;
		var
			w: extended;
		begin
			if p = q then
				angleLessThan := false
			else begin
				w := center.whichSideX(p, q);
				if w = 0 then
					angleLessThan := center.distanceX(p) < center.distanceX(q)
				else
					angleLessThan := w < 0;
			end;
		end;

	begin
		result := polygon2d(convertToPolygon);
		if length > 2 then begin
			center := point2d(elements^^[1].duplicate);
			center.setX((point2d(elements^^[1]).getX + point2d(elements^^[2]).getX + point2d(elements^^[3]).getX) / 3);
			center.setY((point2d(elements^^[1]).getY + point2d(elements^^[2]).getY + point2d(elements^^[3]).getY) / 3);
			i := 1;
			j := result.length + 1;
			while i < j do begin
{$IFC AnimationEnabled }
				if userAbort then begin
					center.free;
					result.freeAll;
					makeStarShapedPolygon := nil;
					exit(makeStarShapedPolygon);
				end;
{$ENDC }
				if pointLessThanXY(point2d(result.elements^^[i]), center) then
					i := i + 1
				else begin
					j := j - 1;
					result.swap(i, j);
				end;
			end;
			result.sort(angleLessThan, 1, i - 1);
			result.sort(angleLessThan, i, result.length);
			center.free;
		end;
		makeStarShapedPolygon := result;
	end;

function pointVector.getType: objectType;
	begin
		getType := pointVectorType;
	end;

function pointVector.getTextType: Str255;
	begin
		getTextType := concat('Point Vector', getTextBaseType);
	end;

function pointVector.memberType (o: obj): boolean;
	begin
		memberType := member(o, pointVector) & memberBaseType(o);
	end;

function pointVector.equal (o: obj): boolean;
	begin
		equal := (o <> nil) & member(o, pointVector) & inherited equal(o);
	end;

	type
		orderTypeEvent = object(obj)
				transformed, aboveOrigin: boolean;
				p: point2d;
				index: seqIndex;
			end;

	var
		origin: point2d;

function orderTypeEventLessThan (q, r: orderTypeEvent): boolean;
	begin
		if q.aboveOrigin = r.aboveOrigin then
			if q.transformed = r.transformed then
				orderTypeEventLessThan := q.p.whichSideX(origin, r.p) > 0
			else
				orderTypeEventLessThan := q.p.whichSideX(origin, r.p) < 0
		else
			orderTypeEventLessThan := q.aboveOrigin;
	end;

function pointVector.orderTypeOne: pointVector;
	var
		queue: homogenVector;
		result: pointVector;
		event: orderTypeEvent;
		p: point2d;
		i: seqIndex;
		count, value: seqIndex0;
		mp: markedRealPoint;
	begin
		if length > 2 then begin
			origin := point2d(ith(1));
			result := newPointVector(length - 1);
			queue := newHomogenVector(2 * length - 2);
			count := 0;
			for i := 2 to length do begin
				p := point2d(elements^^[i]);
				new(mp);
				failNil(Handle(mp));
				mp.init;
				mp.setXY(p.getX, p.getY);
				result.append(mp);
				new(event);
				failNil(Handle(event));
				event.transformed := false;
				event.p := p;
				event.index := i - 1;
				if origin.yLessThan(p) then begin
					count := count + 1;
					event.aboveOrigin := true;
				end
				else
					event.aboveOrigin := false;
				queue.append(event.duplicate);
				event.transformed := true;
				event.aboveOrigin := not event.aboveOrigin;
				queue.append(event);
			end;
			queue.sortAll(orderTypeEventLessThan);
			for i := 1 to length - 1 do
				with orderTypeEvent(queue.elements^^[i]) do begin
					if transformed then begin
						value := count;
						count := count + 1;
					end
					else begin
						count := count - 1;
						value := count;
					end;
					with markedRealPoint(result.elements^^[index]) do
						mark.value := StringOf(value : 1);
				end;
			queue.freeAll;
			orderTypeOne := result;
		end
		else
			orderTypeOne := nil;
	end;

end.
