(*	XYZ GeoBench (eXperimental geometrY Zurich):		*)
(*	An object oriented, interactive workbench for		*)
(*	experimental geometric computation.					*)
(*														*)
(*	 1999 by Peter Schorn, peter.schorn@acm.org		*)
(*	Culmannstrasse 77, CH-8006 Zrich, Switzerland		*)
(*	This software may be copied freely for educational	*)
(*	and research purposes only. For commercial use,		*)
(*	please contact the author.							*)
(*														*)
(*	Version 5.0.5: 30-Sep-1999							*)

unit polygonAreaExperiment;

interface

uses
	triangulatePlaneSweep, pointVector2;

procedure polygonAreaExperiment (r: obj);

function clusteringExperiment (r: obj): pointVector;

procedure initPolygonAreaExperiment;

implementation

const
	stdResultFileName = 'Polgyon Area Result';
	stdNumberOfExperiments = 10;
	stdMinVertices = 3;
	stdMaxVertices = 100;
	stdStepSize = 1;
	stdIsStarShaped = false;

var
	resultFileName: Str255;
	minVertices, maxVertices, stepSize: extended;
	isStarShaped: boolean;

function polygon2d.triangulate2d: obj;
	var
		t2d: triangulation2d;
		polyInfo: triangulationPolygonInfoDirectory;
		sweepLine: triangulationSweepLine;
	begin
		new(t2d);
		failNil(Handle(t2d));
		new(polyInfo);
		failNil(Handle(polyInfo));
		new(sweepLine);
		failNil(Handle(sweepLine));
		sweepLine.init;
		polyInfo.initialize(t2d, length, 1, 0);
		polyInfo.appendPolygon(self);
		polyInfo.sort(1, t2d.points.length);
		with sweepLine do begin
			initialize(polyInfo, t2d);
			startSweep;
			free;
		end;
		polyInfo.free;
		triangulate2d := t2d
	end;

function polygon2d.area: extended;
	var
		s: extended;
		i: seqIndex0;
		pi, pi1: point2d;
	begin
		if length > 2 then begin
			pi1 := point2d(elements^^[1]);
			elements^^[length + 1] := pi1;
			s := 0;
			for i := 1 to length do begin
				pi := pi1;
				pi1 := point2d(elements^^[i + 1]);
				s := s + pi.getX * pi1.getY - pi.getY * pi1.getX;
			end;
			area := abs(s) / 2.0;
		end
		else
			area := 0;
	end;

function doExperiment (pv: pointVector; isStarShaped: boolean): extended;
	var
		p: polygon2d;
	begin
		pv.randomChange;
		if isStarShaped then
			p := pv.makeStarShapedPolygon
		else
			p := pv.makeSimplePolygon;
		if p = nil then
			doExperiment := 0.0
		else begin
			doExperiment := p.area;
			p.freeAll;
		end;
	end;

procedure polygonAreaExperiment (r: obj);
	var
		h: vector;
		pv: pointVector;
		f: text;
		currentNumberOfVertices: seqIndex;
		experiments: longInt;
		countSum, countSqrSum, mean, variance, area, boundingArea: extended;
		continue: boolean;
	begin
		h := newVector(6);
		h.append(getExtendedHandler(4, @maxVertices, stdMaxVertices, nil, @isNatural));
		h.append(getExtendedHandler(5, @stepSize, stdStepSize, nil, @isNonZeroInteger));
		h.append(getExtendedHandler(6, @numberOfExperiments, stdNumberOfExperiments, nil, @isNatural));
		h.append(getStringHandler(14, @resultFileName, stdResultFileName, nil));
		h.append(getExtendedHandler(3, @minVertices, stdMinVertices, nil, @isNatural));
		h.append(getBooleanHandler(16, @isStarShaped, stdIsStarShaped, nil));
		if processDialogBox(136, 9, 8, 10, alwaysTrue, h) then begin
{$IFC THINK_PASCAL }
			pv := createRealPointVector(trunc(maxVertices));
{$ELSEC }
			pv := createRealPointVector(truncLong(maxVertices));
{$ENDC }
			currentRandomConstraint := r;
			boundingArea := r.area;
			rewrite(f, resultFileName);
{$IFC THINK_PASCAL }
			writeLn(f, 'Type Of Bounding Object = ', r.getType);
{$ELSEC }
			writeLn(f, 'Type Of Bounding Object = ', ord(r.getType));
{$ENDC }
			writeLn(f, 'Minimum Number Of Vertices = ', trunc(minVertices) : 1);
			writeLn(f, 'Maximum Number Of Vertices = ', trunc(maxVertices) : 1);
			writeLn(f, 'Step Size = ', trunc(stepSize) : 1);
			writeLn(f, 'Number Of Experiments Per Size = ', trunc(numberOfExperiments) : 1);
			writeLn(f, 'Is Star Shaped = ', isStarShaped);
			writeLn(f);

			if stepSize > 0 then
{$IFC THINK_PASCAL }
				currentNumberOfVertices := trunc(minVertices)
			else
				currentNumberOfVertices := trunc(maxVertices);
{$ELSEC }
				currentNumberOfVertices := truncLong(minVertices)
			else
				currentNumberOfVertices := truncLong(maxVertices);
{$ENDC }
			continue := true;
			while continue & ((stepSize > 0) & (currentNumberOfVertices <= maxVertices) | (stepSize < 0) & (currentNumberOfVertices >= minVertices)) do begin
				write(f, currentNumberOfVertices : 6, ' ');
				displayInfoString(StringOf('Number Of Vertices = ', currentNumberOfVertices : 1));
				pv.length := currentNumberOfVertices;
				countSum := 0;
				countSqrSum := 0;
				variance := 0;
				experiments := 0;
				while experiments < trunc(numberOfExperiments) do begin
{$IFC AnimationEnabled }
					if userAbort then begin
						continue := false;
						leave;
					end;
{$ENDC }
					if experiments mod 50 = 0 then
						displayInfoString(StringOf('Number Of Vertices = ', currentNumberOfVertices : 1, ' Experiment = ', experiments : 1));
					area := doExperiment(pv, isStarShaped) / boundingArea;
					countSum := countSum + area;
					countSqrSum := countSqrSum + sqr(area);
					experiments := experiments + 1;
				end;
				if experiments = 0 then
					mean := 0
				else
					mean := countSum / experiments;
				if experiments > 1 then
					variance := (countSqrSum - experiments * sqr(mean)) / (experiments - 1)
				else
					variance := 0;
				writeLn(f, mean : 10 : 6, ' ', variance : 10 : 6, ' ', experiments : 1);
{$IFC THINK_PASCAL }
				currentNumberOfVertices := currentNumberOfVertices + trunc(stepSize);
{$ELSEC }
				currentNumberOfVertices := currentNumberOfVertices + truncLong(stepSize);
{$ENDC }
			end;
			close(f);
			setFileInfo(resultFileName, 0, 'TEXT', possibleCreatorTypes[textCreatorType]);
{$IFC THINK_PASCAL }
			pv.length := trunc(maxVertices);
{$ELSEC }
			pv.length := truncLong(maxVertices);
{$ENDC }
			pv.freeAll;
		end;
	end;

function counterExample (r: homogenVector): boolean;
	var
		i: convexPolygon;
		oldAnimationFlag: animationVector;
		c1, c2: convexPolygon;
	begin
		oldAnimationFlag := animationFlag;
		setAllAnimationFlag(false);
		c1 := convexPolygon(r.ith(1));
		c2 := convexPolygon(r.ith(2));
		if (c1.length > 1) & (c2.length > 1) then begin
			i := c1.intersect(c2, false);
			if i = nil then
				counterExample := false
			else begin
				counterExample := true;
				i.freeAll;
			end;
		end
		else
			counterExample := false;
		animationFlag := oldAnimationFlag;
	end;

function clusteringExperiment (r: obj): pointVector;
	var
		pv: pointVector;
		result: homogenVector;
		i, j: seqIndex0;
		oldAnimationFlag: animationVector;

	procedure show;
		begin
			displayInfoString(StringOf('Points = ', i : 1, '. Experiments = ', j : 1));
		end;

	begin
		if getUserParameter3('Minimum Number Of Points', minVertices, minVertices, @isNatural, 'Maximum Number Of Points', maxVertices, maxVertices, @isNatural, 'Number Of Experiments', numberOfExperiments, numberOfExperiments, @isNatural) then begin
{$IFC THINK_PASCAL }
			for i := trunc(minVertices) to trunc(maxVertices) do begin
{$ELSEC }
			for i := truncLong(minVertices) to truncLong(maxVertices) do begin
{$ENDC }
				new(pv);
				failNil(Handle(pv));
				pv.preset(i, realPointType);
{$IFC THINK_PASCAL }
				for j := 1 to trunc(numberOfExperiments) do begin
{$ELSEC }
				for j := 1 to truncLong(numberOfExperiments) do begin
{$ENDC }
{$IFC AnimationEnabled }
					if userAbort then begin
						pv.freeAll;
						clusteringExperiment := nil;
						exit(clusteringExperiment);
					end;
{$ENDC }
					if (j - 1) mod 50 = 0 then
						show;
					currentRandomConstraint := r;
					pv.randomChange;
					oldAnimationFlag := animationFlag;
					setAllAnimationFlag(false);
					result := pv.cluster(2, convexPolygonObject);
					animationFlag := oldAnimationFlag;
{$IFC AnimationEnabled }
					if result = nil then begin
						pv.freeAll;
						clusteringExperiment := nil;
						exit(clusteringExperiment);
					end;
{$ENDC }
{$IFC AnimationEnabled }
					if animationFlag[AMinimumSpanningTreeItem] then begin
						pv.interactiveOutput;
						result.interactiveOutput;
						result.interactiveHighlight;
						waitForClick(animationFlag[AMinimumSpanningTreeItem]);
						result.interactiveHighlight;
						result.interactiveOutput;
						pv.interactiveOutput;
					end;
{$ENDC }
					if counterExample(result) then begin
						show;
						clusteringExperiment := pv;
						result.freeAll;
						exit(clusteringExperiment);
					end
					else
						result.freeAll;
				end;
				pv.freeAll;
			end;
			clusteringExperiment := nil;
		end
		else
			clusteringExperiment := nil;
	end;

procedure initPolygonAreaExperiment;
	begin
		resultFileName := stdResultFileName;
		minVertices := stdMinVertices;
		maxVertices := stdMaxVertices;
		stepSize := stdStepSize;
		isStarShaped := stdIsStarShaped;
	end;

end.
