(*	XYZ GeoBench (eXperimental geometrY Zurich):		*)
(*	An object oriented, interactive workbench for		*)
(*	experimental geometric computation.					*)
(*														*)
(*	 1999 by Peter Schorn, peter.schorn@acm.org		*)
(*	Culmannstrasse 77, CH-8006 Zrich, Switzerland		*)
(*	This software may be copied freely for educational	*)
(*	and research purposes only. For commercial use,		*)
(*	please contact the author.							*)
(*														*)
(*	Version 5.0.5: 30-Sep-1999							*)

unit polygonIntersect;

(* Copyright 1990 by Christoph Ammann  ammann@inf.ethz.ch	*)
(* Informatik, ETH, CH-8092 Zrich, Switzerland				*)
(* Version 2.00: 19-Dec-90									*)

interface

{$SETC APolygonIntersection = TRUE }
{$SETC BelongingsCheck = TRUE }
{$SETC ProduceResults = TRUE }

uses
	pointVector;

type
	piEventType = (piStart, piPointQuery, piIntersection, piBend);
	edge = longInt; (* edgeIndex + maxPolys * polygonIndex *)
	piXelement = object(obj)
			where: point2d;
			procedure processEvent (xQueue: priorityQueue; yTable, intersections: dictionary; function areaQ (s: polSet): boolean);
			function getEventType: piEventType;
		end;

	piPointQueryEvent = object(piXelement)
			procedure processEvent (xQueue: priorityQueue; yTable, intersections: dictionary; function areaQ (s: polSet): boolean);
			override;
			function getEventType: piEventType;
			override;
		end;

	piStartEvent = object(piXelement)
			nextLower, nextUpper: point2d;
			refUpper, refLower: reference; {into xQueue to adjacent bend or end events}
			edgeUpper, edgeLower: edge;
			procedure processEvent (xQueue: priorityQueue; yTable, intersections: dictionary; function areaQ (s: polSet): boolean);
			override;
			function getEventType: piEventType;
			override;
		end;

	piBendEvent = object(piXelement)
			next: point2d;
			ref: reference; {into xQueue to adjacent bend or end event}
			ed: edge;
			change: reference; {into yTable to piYelement to be bended}
			procedure processEvent (xQueue: priorityQueue; yTable, intersections: dictionary; function areaQ (s: polSet): boolean);
			override;
			function getEventType: piEventType;
			override;
		end;

	piIntersectionEvent = object(piXelement)
			lower, upper: reference; (* into yTable to intersecting piYelements *)
			lowerYel, upperYel: piYelement;
			procedure processEvent (xQueue: priorityQueue; yTable, intersections: dictionary; function areaQ (s: polSet): boolean);
			override;
			procedure free;
			override;
			function getEventType: piEventType;
			override;
		end;

	piYelement = object(obj)
			p, q: point2d; {p.lessThanXY(q) always true}
			ed: edge;
			intersectLower, intersectUpper: reference; (* into x-queue to intersetionEvent with lower resp. upper neighbor *)
			belongings: polSet;
			areaQbelongings: boolean;
{$IFC ProduceResults }
			buddy: piYelement;
			buddySign: signType;
			res: list;
			procedure freeAll;
(* emergency procedure, only called when yTable is not empty at the end. *)
			override;
{$ENDC }
			procedure interactiveOutput;
			override;
		end;

{$IFC ProduceResults }
var
	resList: list;
{$ENDC }
implementation

type

	edgeIntersection = object(obj)
			r, s: edge;
		end;


procedure piYelement.interactiveOutput;
	begin
		p.drawLine(q);
	end;

function piXelement.getEventType: piEventType;
	begin
		unimplementedMethod('piXelement.getEventType');
		getEventType := piStart;
	end;

function piPointQueryEvent.getEventType: piEventType;
	begin
		getEventType := piPointQuery;
	end;

function piIntersectionEvent.getEventType: piEventType;
	begin
		getEventType := piIntersection;
	end;

function piBendEvent.getEventType: piEventType;
	begin
		getEventType := piBend;
	end;

function piStartEvent.getEventType: piEventType;
	begin
		getEventType := piStart;
	end;

function symmDiff (a, b: polSet): polSet;
	begin
		symmDiff := (a - b) + (b - a);
	end;

function compareIntersection (a, b: edgeIntersection): relationObj;
	begin
		if a.r > b.r then
			compareIntersection := greaterThanObj
		else if a.r = b.r then
			if a.s > b.s then
				compareIntersection := greaterThanObj
			else if a.s = b.s then
				compareIntersection := equalObj
			else
				compareIntersection := lessThanObj
		else
			compareIntersection := lessThanObj;
	end;

function compareYelement (a, b: piYelement): relationObj;
	begin
		case a.p.whichSideSign(b.p, b.q) of
			-1: 
				compareYelement := greaterThanObj;
			0: 
				compareYelement := relationObj(1 - a.q.whichSideSign(b.p, b.q));
			1: 
				compareYelement := lessThanObj;
		end;
	end;

function compareEvent (a, b: piXelement): relationObj;
	var
		aw, bw: point2d;
(* cf. lineSegmentVector.planeSweepAllIntersect: stolen there *)

	type
		epsRecord = record
				x, y: 0..1;
			end;

	function comparePointXYEps (a: longIntPoint; aTrue: epsRecord; b: longIntPoint; bTrue: epsRecord): relationObj;
		begin
			a.x := 2 * a.x + aTrue.x;
			a.y := 2 * a.y + aTrue.y;
			b.x := 2 * b.x + bTrue.x;
			b.y := 2 * b.y + bTrue.y;
			comparePointXYEps := comparePointXY(a, b);
			a.x := a.x div 2;
			a.y := a.y div 2;
			b.x := b.x div 2;
			b.y := b.y div 2;
		end;

	function truePoint (x: piXelement): epsRecord;

		function intersectTest (p, q, r, s: longIntPoint): epsRecord;
			var
				prq, psq, d: longInt;
				e: epsRecord;

			function ws (p, q, r: longIntPoint): longInt;
				begin
					ws := (p.x - q.x) * (q.y - r.y) - (p.y - q.y) * (q.x - r.x);
				end;

			begin
				prq := ws(p, r, q);
				psq := ws(p, s, q);
				if prq <> psq then begin
					d := prq - psq;
					e.x := ord(((prq * s.x - psq * r.x) mod d) <> 0);
					e.y := ord(((prq * s.y - psq * r.y) mod d) <> 0);
				end
				else begin
					e.x := 0;
					e.y := 0;
				end;
				intersectTest := e;
			end;

		begin
			if not member(x, piIntersectionEvent) then begin
				truePoint.x := 0;
				truePoint.y := 0;
			end
			else
				with piIntersectionEvent(x) do
					truePoint := intersectTest(longIntPoint(lowerYel.p), longIntPoint(lowerYel.q), longIntPoint(upperYel.p), longIntPoint(upperYel.q));
		end;

	begin
		aw := a.where;
		bw := b.where;
		if member(a.where, longIntPoint) then
			case comparePointXYEps(longIntPoint(a.where), truePoint(a), longIntPoint(b.where), truePoint(b)) of
				lessThanObj: 
					compareEvent := lessThanObj;
				equalObj: 
					compareEvent := equalObj;
				greaterThanObj: 
					compareEvent := greaterThanObj;
			end
		else if aw.xLessThan(bw) then
			compareEvent := lessThanObj
		else if bw.xLessThan(aw) then
			compareEvent := greaterThanObj
		else if aw.yLessThan(bw) then
			compareEvent := lessThanObj
		else if bw.yLessThan(aw) then
			compareEvent := greaterThanObj
		else
			compareEvent := relationObj(sign(ord(a.getEventType) - ord(b.getEventType)) + 1);
	end; {compareEvent}

procedure clearIntersection (pred, succ: reference; yTable: dictionary; xQueue: priorityQueue);
	var
		neighborsIntersect: reference;
		intersectionEvent: piIntersectionEvent;
	begin
		if (pred <> nil) & (succ <> nil) & (piYelement(yTable.getObject(succ)).intersectLower <> nil) then begin
			neighborsIntersect := piYelement(yTable.getObject(succ)).intersectLower;
			intersectionEvent := piIntersectionEvent(xQueue.getObject(neighborsIntersect));
			intersectionEvent.free;
			xQueue.delete(neighborsIntersect, compareEvent);
			assert(neighborsIntersect = piYelement(yTable.getObject(pred)).intersectUpper, 'lower.upper = upper.lower');
			piYelement(yTable.getObject(succ)).intersectLower := nil;
			piYelement(yTable.getObject(pred)).intersectUpper := nil;
		end;
	end;

{$IFC NOT THINK_PASCAL }
{$PUSH }
{$PRAGMAC warn_unusedarg off }
{$ENDC }
procedure piXelement.processEvent (xQueue: priorityQueue; yTable, intersections: dictionary; function areaQ (s: polSet): boolean);
	begin
		unimplementedMethod('piXelement.processEvent');
	end;
{$IFC NOT THINK_PASCAL }
{$POP }
{$ENDC }

procedure makeIntersection (var intersection: edgeIntersection; var e1, e2: edge);
	begin
		new(intersection);
		failNil(Handle(intersection));
		if e1 < e2 then begin
			intersection.r := e1;
			intersection.s := e2;
		end
		else begin
			intersection.r := e2;
			intersection.s := e1;
		end;
	end;

procedure checkIntersection (lower, upper: reference; yTable: dictionary; xQueue: priorityQueue; intersections: dictionary);
	var
		l, u: piYelement;
		iEvent: piIntersectionEvent;
		int1, int2: point2d;
		ref: reference;
		intersection: edgeIntersection;
	begin
		if (lower <> nil) & (upper <> nil) then begin
			l := piYelement(yTable.getObject(lower));
			u := piYelement(yTable.getObject(upper));
			int1 := point2d(l.p.duplicate);
			int2 := point2d(l.p.duplicate);

{$IFC AnimationEnabled AND APolygonIntersection }
			if animationFlag[APolygonIntersectionItem] then begin
				PenSize(3, 3);
				l.p.drawLine(l.q);
				u.p.drawLine(u.q);
				PenSize(1, 1);
				waitForClick(animationFlag[APolygonIntersectionItem]);
			end;
{$ENDC }

	(*	s.whichSideX(p,q) - r.whichSideX(p,q) = s.crossProductX(r,q,p) *)
	(*	s.whichsideX(p,q) = s.crossProductX(p,q,p) *)

			if (u.q.crossProductX(u.p, l.q, l.p) >= 0) & l.p.intersectLineSegmentX(l.q, u.p, u.q, int1, int2) then begin
				makeIntersection(intersection, l.ed, u.ed);
				if not intersections.member(intersection, compareIntersection) then begin
					new(iEvent);
					failNil(Handle(iEvent));
					if l.q = u.q then
						iEvent.where := point2d(int2.duplicate)
					else
						iEvent.where := point2d(int1.duplicate);
					iEvent.lower := lower;
					iEvent.upper := upper;
					iEvent.lowerYel := l;
					iEvent.upperYel := u;
{$IFC AnimationEnabled AND APolygonIntersection }
					if animationFlag[APolygonIntersectionItem] then begin
						int1.interactiveHighlight;
						waitForClick(animationFlag[APolygonIntersectionItem]);
						int1.interactiveHighlight;
					end;
{$ENDC }
					ref := xQueue.insert(iEvent, compareEvent);
					assert((l.intersectUpper = nil) & (u.intersectLower = nil), 'lower.upper=nil & upper.lower=nil');
					l.intersectUpper := ref;
					u.intersectLower := ref;
				end;
				intersection.free;
			end;

{$IFC AnimationEnabled AND APolygonIntersection }
			if animationFlag[APolygonIntersectionItem] then begin
				PenSize(3, 3);
				l.p.drawLine(l.q);
				u.p.drawLine(u.q);
				PenSize(1, 1);
			end;
{$ENDC }

			int1.free;
			int2.free;
		end;
	end;

{$IFC ProduceResults }
procedure finishPoly (res: list);
	var
		p: polygon2d;
		i: seqIndex0;

	procedure copyPoint (pt: point2d);
		var
			this, last: point2d;
		begin
			while i > 0 do begin
				this := point2d(p.elements^^[i]);
				if i > 1 then
					last := point2d(p.elements^^[i - 1])
				else
					last := point2d(res.last^^.value);
				if this.whichSideX(last, pt) <> 0 then
					leave;
				this.free;
				i := i - 1;
			end;
			i := i + 1;
			p.elements^^[i] := pt;
		end;

	begin
(* eliminate degeneracies in output polygon *)
		p := newPolygon2d(res.length + 1);
		i := 0;
		res.forAll(copyPoint);
		copyPoint(point2d(p.elements^^[1]));
		p.length := i - 1;
		res.free;

		if p.length > 0 then begin
			p.allocateElements(p.length);
{$IFC AnimationEnabled AND APolygonIntersection }
			if animationFlag[APolygonIntersectionItem] then begin
				p.interactiveFlash;
				p.interactiveHighlight;
				waitForClick(animationFlag[APolygonIntersectionItem]);
				p.interactiveHighlight;
			end;
{$ENDC }
			resList.append(p);
		end
		else
			p.free;
	end;

procedure piYelement.freeAll;
(* emergency procedure, only called when yTable is not empty at the end. *)
	begin
		if res <> nil then begin (* unclosed polygon: save it *)
			res.append(q.duplicate);
			finishPoly(res);
		end;
		shallowFree;
	end;

procedure openResultPoly (l, u: piYelement; p: point2d);
	var
		res: list;
		c: boolean;
	begin
		l.res := nil;
		u.res := nil;
		l.buddy := u;
		u.buddy := l;
		c := l.areaQbelongings;
		if c <> u.areaQbelongings then begin
			res := newEmptyList;
			res.append(p.duplicate);
			if c then begin
				l.buddySign := 1;
				u.buddySign := -1;
				l.res := res;
			end
			else begin
				l.buddySign := -1;
				u.buddySign := 1;
				u.res := res;
			end;
		end
		else begin
			l.buddySign := 0;
			u.buddySign := 0;
		end;
	end;

procedure extendResultPoly (y: piYelement; p: point2d);
	begin
		if y.buddySign > 0 then
			y.res.push(p.duplicate)
		else if y.buddySign < 0 then
			y.buddy.res.append(p.duplicate);
	end;

procedure closeResultPoly (l, u: piYelement; pt: point2d);
	var
		tmp: piYelement;
	begin
		if l.buddySign <> 0 then begin
			if l.buddySign < 0 then begin
				tmp := l;
				l := u;
				u := tmp;
			end;
			u.buddy.res.append(pt.duplicate);
			if u.buddy = l then
				finishPoly(l.res)
			else begin
				u.buddy.res.concatenateDispose(l.res);
				l.buddy.buddy := u.buddy;
				u.buddy.buddy := l.buddy;
			end;
		end
		else if l.buddy <> u then begin
			l.buddy.buddy := u.buddy;
			u.buddy.buddy := l.buddy;
		end;
	end;
{$ENDC }

procedure propagateYElRef (xRef, yRef: reference; xQueue: priorityQueue);
	begin
		if (xRef <> nil) & member(xQueue.getObject(xRef), piBendEvent) then
			piBendEvent(xQueue.getObject(xRef)).change := yRef;
	end;

procedure piStartEvent.processEvent (xQueue: priorityQueue; yTable, intersections: dictionary; function areaQ (s: polSet): boolean);
	var
		l, u: piYelement;
		upper, lower, pred, succ: reference;
		found: boolean;
	begin
		new(l);
		failNil(Handle(l));
		l.p := where;
		l.q := nextLower;
		l.ed := edgeLower;
		l.intersectLower := nil;
		l.intersectUpper := nil;
		lower := yTable.insertObj(l, compareYelement, found);

		new(u);
		failNil(Handle(u));
		u.p := where;
		u.q := nextUpper;
		u.ed := edgeUpper;
		u.intersectLower := nil;
		u.intersectUpper := nil;
		upper := yTable.insert(u, lower, right);

		propagateYElRef(refLower, lower, xQueue);
		propagateYElRef(refUpper, upper, xQueue);

		pred := yTable.next(lower, left);
		succ := yTable.next(upper, right);

		if pred <> nil then
			l.belongings := symmDiff(piYelement(yTable.getObject(pred)).belongings, [polRange(edgeLower mod maxPolys)])
		else
			l.belongings := [polRange(edgeLower mod maxPolys)];
		u.belongings := symmDiff(l.belongings, [polRange(edgeUpper mod maxPolys)]);
{$IFC BelongingsCheck AND ChecksEnabled }
		if succ = nil then
			assert(u.belongings = [], 'belongings1')
		else
			assert((piYelement(yTable.getObject(succ)).belongings = symmDiff(u.belongings, [polRange(piYelement(yTable.getObject(succ)).ed mod maxPolys)])), 'belongings1');
{$ENDC }
		l.areaQbelongings := areaQ(l.belongings);
		u.areaQbelongings := areaQ(u.belongings);

{$IFC ProduceResults }
		openResultPoly(l, u, where);
{$ENDC }

		clearIntersection(pred, succ, yTable, xQueue);
		checkIntersection(pred, lower, yTable, xQueue, intersections);
		checkIntersection(upper, succ, yTable, xQueue, intersections);

		if l.q = u.q then
			checkIntersection(lower, upper, yTable, xQueue, intersections);

{$IFC AnimationEnabled AND APolygonIntersection }
		if animationFlag[APolygonIntersectionItem] then
			waitForClick(animationFlag[APolygonIntersectionItem]);
{$ENDC }
	end;

{$IFC NOT THINK_PASCAL }
{$PUSH }
{$PRAGMAC warn_unusedarg off }
{$ENDC }
procedure piBendEvent.processEvent (xQueue: priorityQueue; yTable, intersections: dictionary; function areaQ (s: polSet): boolean);
	var
		pred, succ: reference;
		yEl: piYelement;
	begin
		pred := yTable.next(change, left);
		succ := yTable.next(change, right);

		clearIntersection(pred, change, yTable, xQueue);
		clearIntersection(change, succ, yTable, xQueue);

		yEl := piYelement(yTable.getObject(change));
		yEl.p := yEl.q;
		yEl.q := next;
		yEl.ed := ed;

		propagateYElRef(ref, change, xQueue);

{$IFC ProduceResults }
		extendResultPoly(yEl, where);
{$ENDC }

		checkIntersection(pred, change, yTable, xQueue, intersections);
		checkIntersection(change, succ, yTable, xQueue, intersections);
{$IFC AnimationEnabled AND APolygonIntersection }
		if animationFlag[APolygonIntersectionItem] then
			waitForClick(animationFlag[APolygonIntersectionItem]);
{$ENDC }
	end;
{$IFC NOT THINK_PASCAL }
{$POP }
{$ENDC }

procedure piIntersectionEvent.processEvent (xQueue: priorityQueue; yTable, intersections: dictionary; function areaQ (s: polSet): boolean);
	var
		pred, succ: reference;
		l, u: piYelement;
		intersection: edgeIntersection;
		c1, c2: boolean;
{$IFC ProduceResults }
		c3, c4: boolean;
		ub: polSet;
{$ENDC }
{$IFC BelongingsCheck AND ChecksEnabled }
		p, s: polSet;
{$ENDC }
		tmp: piYelement;
		tmpS: signType;
		tmpl: list;
	begin
		l := lowerYel;
		u := upperYel;
		l.intersectUpper := nil;
		u.intersectLower := nil;


{$IFC BelongingsCheck AND ChecksEnabled }
		assert(u.belongings = symmDiff(l.belongings, [polRange(u.ed mod maxPolys)]), 'belongings3');
{$ENDC }

		if l.q = u.q then begin
			pred := yTable.next(lower, left);
			succ := yTable.next(upper, right);
{$IFC BelongingsCheck AND ChecksEnabled }
			if pred = nil then
				p := []
			else
				p := piYelement(yTable.getObject(pred)).belongings;
			if succ = nil then
				s := []
			else
				s := symmDiff(piYelement(yTable.getObject(succ)).belongings, [polRange(piYelement(yTable.getObject(succ)).ed mod maxPolys)]);
			assert(p = s, 'belongings2');
{$ENDC }

{$IFC ProduceResults }
			closeResultPoly(l, u, where);
{$ENDC }

			clearIntersection(pred, lower, yTable, xQueue);
			clearIntersection(upper, succ, yTable, xQueue);
			checkIntersection(pred, succ, yTable, xQueue, intersections);
			l.free;
			u.free;
			yTable.delete(lower);
			yTable.delete(upper);
		end
		else begin
			makeIntersection(intersection, l.ed, u.ed);
			intersections.insertNewObj(intersection, compareIntersection);

{$IFC ProduceResults }
			ub := symmDiff(u.belongings, [polRange(l.ed mod maxPolys)]);

			c1 := l.areaQbelongings;
			c2 := u.areaQbelongings;
			c3 := areaQ(ub);
			c4 := areaQ(symmDiff(l.belongings, [polRange(l.ed mod maxPolys)]));

			if (c2 = c4) & (c1 | c3) then (*strongest possible: (c2=c4) & (c1 <> c3); weakest possible: c2=c4 *)
				begin
				closeResultPoly(l, u, where);
				l.belongings := u.belongings;
				l.areaQbelongings := u.areaQbelongings;
				u.belongings := ub;
				u.areaQbelongings := c3;
				openResultPoly(u, l, where);
			end
			else begin
				if c1 = c3 then begin
					if c1 <> c2 then
						extendResultPoly(u, where);
					if c1 <> c4 then
						extendResultPoly(l, where);
				end;
				if l.buddy <> u then begin
					tmp := u.buddy;
					u.buddy := l.buddy;
					l.buddy := tmp;
					u.buddy.buddy := u;
					l.buddy.buddy := l;
				end;
				tmpS := u.buddySign;
				u.buddySign := l.buddySign;
				l.buddySign := tmpS;
				tmpl := u.res;
				u.res := l.res;
				l.res := tmpl;
				l.belongings := u.belongings;
				l.areaQbelongings := u.areaQbelongings;
				u.belongings := ub;
				u.areaQbelongings := c3;
			end;
{$ELSEC }
			l.belongings := u.belongings;
			l.areaQbelongings := u.areaQbelongings;
			u.belongings := symmDiff(u.belongings, [polRange(l.ed mod maxPolys)]);
			u.areaQbelongings := areaQ(u.belongings);
{$ENDC }

			pred := yTable.next(lower, left);
			succ := yTable.next(upper, right);
			clearIntersection(pred, lower, yTable, xQueue);
			clearIntersection(upper, succ, yTable, xQueue);
			yTable.swap(lower, upper);
			checkIntersection(pred, upper, yTable, xQueue, intersections);
			checkIntersection(lower, succ, yTable, xQueue, intersections);
		end;

{$IFC AnimationEnabled AND APolygonIntersection }
		if animationFlag[APolygonIntersectionItem] then
			waitForClick(animationFlag[APolygonIntersectionItem]);
{$ENDC }
	end;

procedure piIntersectionEvent.free;
	begin
		where.free;
		shallowFree;
	end;

function compareYelementPointQuery (a: markedRealPoint; b: piYelement): relationObj;
	begin
		case a.whichSideSign(b.p, b.q) of
			-1: 
				compareYelementPointQuery := greaterThanObj;
			0: 
				compareYelementPointQuery := equalObj;
			1: 
				compareYelementPointQuery := lessThanObj;
		end;
	end;

{$IFC NOT THINK_PASCAL }
{$PUSH }
{$PRAGMAC warn_unusedarg off }
{$ENDC }
procedure piPointQueryEvent.processEvent (xQueue: priorityQueue; yTable, intersections: dictionary; function areaQ (s: polSet): boolean);
	var
		x: reference;
		d: directionType;
		pos: longInt;
	begin
		with markedRealPoint(where).mark do
			if yTable.findPosition(where, compareYelementPointQuery, x, d, pos) then
				value := 'On'
			else if odd(pos) then
				value := 'In'
			else
				value := 'Out';
{$IFC AnimationEnabled AND APolygonIntersection }
		if animationFlag[APolygonIntersectionItem] then begin
			where.interactiveFlash;
			where.interactiveHighlight;
			where.interactiveOutput;
		end;
{$ENDC }
	end;
{$IFC NOT THINK_PASCAL }
{$POP }
{$ENDC }

{$IFC AnimationEnabled AND APolygonIntersection }
function displayOrder (yTable: dictionary): list;
	var
		rank: longInt;
		result: list;

	procedure show (y: piYelement);
		var
			mp: markedRealPoint;
		begin
			rank := rank + 1;
			new(mp);
			failNil(Handle(mp));
			mp.init;
			with y.p do begin
				mp.x := getX;
				mp.y := getY;
			end;
			mp.mark.value := stringOf(rank : 1);
			result.append(mp);
		end;

	begin (* displayOrder *)
		rank := 0;
		result := newEmptyList;
		yTable.forAll(show, left);
		displayOrder := result;
	end;
{$ENDC }

procedure polygon2d.intersection (pols: sequence; function zoneQ (s: polSet): boolean; var result: list; ask: boolean; pointsInPolygon: pointVector);
	var
		xQueue: priorityQueue;
		yTable, intersections: dictionary;
		minRef: reference;
		minObj: piXelement;
		polNo, totalLength, queryLength: longInt;
{$IFC AnimationEnabled AND APolygonIntersection }
		order: list;
{$ENDC }

	procedure addUpLengths (p: polygon2d);
		begin
			totalLength := totalLength + p.length;
		end;

	procedure insertQueryPoints (q: markedRealPoint);
		var
			ev: piPointQueryEvent;
			r: reference;
		begin
			new(ev);
			failNil(Handle(ev));
			ev.where := q;
			r := xQueue.insert(ev, compareEvent);
		end;

	procedure insertPolygon (p: polygon2d);
		type
			refArr = array[1..maxInt] of reference;
		var
			heap: heapPriorityQueue;
			prev, this, next, starting: point2d;
			thisInd, nextInd: seqIndex;
			sEvent: piStartEvent;
			bEvent: piBendEvent;
			edgeNo, prevEdgeNo: edge;
			ev: piXelement;
			refs: ^refArr;
			i, im1, l: longInt;

		procedure getNextPoint;
			begin
				prev := this;
				this := next;
				thisInd := nextInd;
				with p do
					repeat
						nextInd := nextInd + 1;
						if nextInd > length then
							nextInd := 1;
						next := point2d(elements^^[nextInd]);
					until not next.equalPoint(this) | (nextInd = thisInd);
			end;

		procedure setRelationship (i, j: seqIndex);
			var
				xEli, xElj: piXelement;
			begin
				xEli := piXelement(xQueue.getObject(refs^[i]));
				xElj := piXelement(xQueue.getObject(refs^[j]));
				if member(xEli, piStartEvent) then begin
					if xElj.where = piStartEvent(xEli).nextLower then
						piStartEvent(xEli).refLower := refs^[j];
					if xElj.where = piStartEvent(xEli).nextUpper then
						piStartEvent(xEli).refUpper := refs^[j];
				end
				else if member(xEli, piBendEvent) then
					if xElj.where = piBendEvent(xEli).next then
						piBendEvent(xEli).ref := refs^[j];
			end;

		begin {insertPolygon}
			assert(polNo < maxPolys, 'polNo < maxPolys');
			with p do begin {eliminate equal points}
				this := point2d(elements^^[length]);
				next := point2d(elements^^[1]);
				nextInd := 1;
				getNextPoint;
				getNextPoint;
				prevEdgeNo := thisInd * maxPolys + polNo;
				getNextPoint;
				starting := this;
				refs := pointer(NewPtr(length * sizeOf(reference)));
				failNil(Handle(refs));
				l := 0;
				if member(xQueue, heapPriorityQueue) then
					heap := heapPriorityQueue(xQueue);
				repeat
					edgeNo := thisInd * maxPolys + polNo;
					ev := nil;
					if pointLessThanXY(prev, this) then begin
						if pointLessThanXY(this, next) then begin
							new(bEvent);
							failNil(Handle(bEvent));
							bEvent.ref := nil;
							bEvent.next := next;
							bEvent.ed := edgeNo;
							ev := bEvent;
						end; (* else end event *)
					end
					else begin
						if pointLessThanXY(next, this) then begin
							new(bEvent);
							failNil(Handle(bEvent));
							bEvent.ref := nil;
							bEvent.next := prev;
							bEvent.ed := prevEdgeNo;
							ev := bEvent;
						end
						else begin
							new(sEvent);
							failNil(Handle(sEvent));
							sEvent.refLower := nil;
							sEvent.refUpper := nil;
							if next.whichSideSign(this, prev) > 0 then begin
								sEvent.nextLower := next;
								sEvent.edgeLower := edgeNo;
								sEvent.nextUpper := prev;
								sEvent.edgeUpper := prevEdgeNo;
							end
							else begin
								sEvent.nextLower := prev;
								sEvent.edgeLower := prevEdgeNo;
								sEvent.nextUpper := next;
								sEvent.edgeUpper := edgeNo;
							end;
							ev := sEvent;
						end
					end;
					if ev <> nil then begin
						ev.where := this;
						l := l + 1;
						if member(xQueue, heapPriorityQueue) then
							refs^[l] := heap.simpleAppend(ev)
						else
							refs^[l] := xQueue.insert(ev, compareEvent);
					end;
					getNextPoint;
					prevEdgeNo := edgeNo;
				until starting = this;
				if member(xQueue, heapPriorityQueue) then
					heap.heapify(compareEvent);
				if l = 1 then begin
					xQueue.delete(refs^[1], compareEvent);
					if ev <> nil then
						ev.free;
				end
				else begin
					im1 := l;
					for i := 1 to l do begin
						setRelationship(i, im1);
						setRelationship(im1, i);
						im1 := i;
					end;
				end;
{$IFC THINK_PASCAL }
				disposPtr(pointer(refs));
{$ELSEC }
				DisposePtr(pointer(refs));
{$ENDC }
				polNo := polNo + 1;
			end; {with}

{$IFC AnimationEnabled AND APolygonIntersection }
			if animationFlag[APolygonIntersectionItem] then
				p.interactiveHighlight;
{$ENDC }
		end; {insertPolygon}

	begin {polygon2d.intersection}
		totalLength := 0;
		pols.forAll(addUpLengths);
		if pointsInPolygon = nil then
			queryLength := 0
		else
			queryLength := pointsInPolygon.length;
		if ask then begin
			if not getXY(2 * totalLength + queryLength, false, xQueue, totalLength, yTable, false) then
				exit(intersection);
		end
		else begin
			currentVectorLength := 2 * totalLength + queryLength;
			xQueue := priorityQueue(newInitObj(heapPriorityQueueType));
			yTable := dictionary(newInitObj(avlTreeType));
		end;
		intersections := dictionary(newInitObj(avlTreeType));

		polNo := 0;
		pols.forAll(insertPolygon);
		if pointsInPolygon <> nil then
			pointsInPolygon.forAll(insertQueryPoints);

{$IFC ProduceResults }
		resList := result;
{$ENDC }
		while not xQueue.isEmpty do begin
			minRef := xQueue.minimum;
			minObj := piXelement(xQueue.getObject(minRef));
			xQueue.delete(minRef, compareEvent);
{$IFC AnimationEnabled AND APolygonIntersection }
			if animationFlag[APolygonIntersectionItem] then begin
				drawThickVertical(minObj.where.getX);
				minObj.where.interactiveHighlight;
				order := displayOrder(yTable);
				order.interactiveOutput;
			end
			else
				order := nil;
{$ENDC }
			minObj.processEvent(xQueue, yTable, intersections, zoneQ);
{$IFC AnimationEnabled AND APolygonIntersection }
			if animationFlag[APolygonIntersectionItem] then begin
				drawThickVertical(minObj.where.getX);
				minObj.where.interactiveHighlight;
				order.interactiveOutput;
				order.freeAll;
			end
			else if order <> nil then
				order.freeAll;
{$ENDC }
			minObj.free;
{$IFC AnimationEnabled }
			if userAbort then begin
				intersections.freeAll;
				yTable.freeAll;
				xQueue.freeAll;
				resList.freeAll;
				result := nil;
				exit(intersection);
			end;
{$ENDC }
		end; {while}

		displayInfoString2('X', xQueue.infoString, 'Y', yTable.infoString);
		intersections.freeAll;
		assert(yTable.isEmpty, stringOf('yTable.isEmpty (noEl=', yTable.numberOfElements : 1, ')'));
		yTable.freeAll;
		xQueue.free;
	end; {polygon2d.intersection}

end.
