(*	XYZ GeoBench (eXperimental geometrY Zurich):		*)
(*	An object oriented, interactive workbench for		*)
(*	experimental geometric computation.					*)
(*														*)
(*	 1999 by Peter Schorn, peter.schorn@acm.org		*)
(*	Culmannstrasse 77, CH-8006 Zrich, Switzerland		*)
(*	This software may be copied freely for educational	*)
(*	and research purposes only. For commercial use,		*)
(*	please contact the author.							*)
(*														*)
(*	Version 5.0.5: 30-Sep-1999							*)

unit priorityQueue;

interface

uses
	dictionary;

type
	priorityQueue = object(obj)
			numberOfElements: longInt;

			function getObject (where: reference): obj;
(* Retrieves the object referenced by 'where'. *)

			function insert (x: obj; function compare (a, b: obj): relationObj): reference;
(* Inserts 'x' into the priority queue using 'compare' giving a reference for later removal. *)

			procedure delete (x: reference; function compare (a, b: obj): relationObj);
(* Deletes the object referenced by 'x' from the priority queue. 'compare' may be used for restructuring. *)

			function isEmpty: boolean;
(* Tests whether the queue is empty. *)

			function minimum: reference;
(* Retrieves the minimum without deleting it. *)

			function ith (i: longInt): reference;
(*	Returns the i'th element of the priority queue.	(1  'i'  'numberOfElements') *)

			procedure forAll (procedure whatToDo (x: obj));
(* Perform the procedure 'whatToDo' on all elements 'x' of the priority queue. *)

			procedure reset;
(* Reset the priorityqueue to the state after initialization. Dependent objects are not removed! *)

			function infoString: Str255;
(* Computes information about the dictionary (this only makes sense for a 'priorityQueueStatistics'. *)

		end;

	priorityQueueStatistics = object(priorityQueue)
(* Instruments a priority queue. *)
			maxLength,			(* Maximal number of elements in the priority queue.	*)
			insertions,			(* Number of insertions into the priority queue. 			*)
			deletions: longInt;	(* Number of deletions from the priority queue. 			*)
			p: priorityQueue;	(* Actual priority queue that is used. 						*)
			procedure init;
			override;
(* The global variable 'currentPriorityQueueType' determines which kind of		*)
(* dictionary is used for 'p' and initializes it. Note that the initialization of 'p'	*)
(* might require the correct setting of additional global variables. 				*)

			procedure free;
			override;
			procedure freeAll;
			override;
			function getObject (where: reference): obj;
			override;
			function insert (x: obj; function compare (a, b: obj): relationObj): reference;
			override;
			procedure delete (x: reference; function compare (a, b: obj): relationObj);
			override;
			function isEmpty: boolean;
			override;
			function minimum: reference;
			override;
			function ith (i: longInt): reference;
			override;
			procedure forAll (procedure whatToDo (x: obj));
			override;
			procedure reset;
			override;
			function infoString: Str255;
			override;
			procedure resetCounters;
		end;

	dictPriorityQueue = object(priorityQueue)
(* Priority queue implemented with a dictionary featuring an efficient find and next operation.	*)
			d: dictionary;
			minRef: reference;

			procedure init;
			override;
(* The global variable 'currentDictionaryType' determines which kind of dictionary is used		*)
(* for 'd' and initializes it. Note that the initialization of 'd' might require the correct setting		*)
(* of additional global variables. *)

			function find (x: obj; function compare (a, b: obj): relationObj; var where: reference; var direction: directionType): boolean;
(* if 'find' = TRUE then 'where' points to some element 'e' with 'compare(e,x) = equalObj'.		*)
(* If 'find' = FALSE then either 'where' = nil (dictionary empty) or 'x' is a neighbor of				*)
(* 'where' in direction 'direction'. *)

			function next (x: reference; direction: directionType): reference;
(* Find the predecessor ('direction' = left) or the successor ('direction' = right) of 'x'. 			*)

			procedure free;
			override;
			procedure freeAll;
			override;
			procedure interactiveOutput;
			override;
			function getObject (where: reference): obj;
			override;
			function insert (x: obj; function compare (a, b: obj): relationObj): reference;
			override;
			procedure delete (x: reference; function compare (a, b: obj): relationObj);
			override;
			function isEmpty: boolean;
			override;
			function minimum: reference;
			override;
			function ith (i: longInt): reference;
			override;
			procedure forAll (procedure whatToDo (x: obj));
			override;
			procedure reset;
			override;
		end;

	heapPriorityQueue = object(priorityQueue)
(* Implements the priority queue as a heap. *)
			freeList: seqIndex;				(* points to first empty element in 'masters', linked by 'ref' field							*)
			masters: refPairArrayH;		(* contains objects and their position															*)
			vector: seqIndex0ArrayH;	(* the heap																							*)

			procedure init;
			override;
(* The global variable 'currentVectorLength' determines how many elements are allocated. *)
			procedure free;
			override;
			procedure freeAll;
			override;
			procedure interactiveOutput;
			override;
			procedure textOutput;
			override;
			function getObject (where: reference): obj;
			override;
			function insert (x: obj; function compare (a, b: obj): relationObj): reference;
			override;
			procedure delete (x: reference; function compare (a, b: obj): relationObj);
			override;
			function isEmpty: boolean;
			override;
			function minimum: reference;
			override;
			function ith (i: longInt): reference;
			override;
			procedure forAll (procedure whatToDo (x: obj));
			override;
			procedure reset;
			override;

			procedure swap (p, q: seqIndex);	 (* private methods *)
			function ithObj (i: seqIndex): obj;
			procedure pushUp (i: seqIndex; function compare (a, b: obj): relationObj);
			procedure pushDown (i: seqIndex; function compare (a, b: obj): relationObj);
			function isHeap (root: seqIndex; function compare (a, b: obj): relationObj): boolean;
			function append (x: obj): reference;
(* adds 'x' to the end of the vector without maintaining the heap property *)
			function simpleAppend (x: obj): reference;
(* like 'append', but 'freeList' is not updated (must be done using 'heapify' *)
			procedure heapify (function compare (a, b: obj): relationObj);
(* sets 'freeList' to the end of the vector and produces a heap from the structure *)
		end;

procedure initPriorityQueue;

implementation

{$IFC NOT THINK_PASCAL }
{$PUSH }
{$PRAGMAC warn_unusedarg off }
{$ENDC }
function priorityQueue.getObject (where: reference): obj;
	begin
		unimplementedMethod('priorityQueue.getObject');
		getObject := nil;
	end;

function priorityQueue.insert (x: obj; function compare (a, b: obj): relationObj): reference;
	begin
		unimplementedMethod('priorityQueue.insert');
		insert := nil;
	end;

procedure priorityQueue.delete (x: reference; function compare (a, b: obj): relationObj);
	begin
		unimplementedMethod('priorityQueue.delete');
	end;

function priorityQueue.ith (i: longInt): reference;
	begin
		unimplementedMethod('priorityQueue.ith');
		ith := nil;
	end;

procedure priorityQueue.forAll (procedure whatToDo (x: obj));
	begin
		unimplementedMethod('priorityQueue.forAll');
	end;
{$IFC NOT THINK_PASCAL }
{$POP }
{$ENDC }

function priorityQueue.isEmpty: boolean;
	begin
		unimplementedMethod('priorityQueue.isEmpty');
		isEmpty := true;
	end;

function priorityQueue.minimum: reference;
	begin
		unimplementedMethod('priorityQueue.minimum');
		minimum := nil;
	end;

procedure priorityQueue.reset;
	begin
		unimplementedMethod('priorityQueue.reset');
	end;

function priorityQueue.infoString: Str255;
	begin
		infoString := '';
	end;

procedure priorityQueueStatistics.resetCounters;
	begin
		numberOfElements := 0;
		maxLength := 0;
		insertions := 0;
		deletions := 0;
	end;

procedure priorityQueueStatistics.reset;
	begin
		p.reset;
		resetCounters;
	end;

procedure priorityQueueStatistics.init;
	begin
		p := priorityQueue(newInitObj(currentPriorityQueueType));
		resetCounters;
	end;

procedure priorityQueueStatistics.free;
	begin
		p.free;
		shallowFree;
	end;

procedure priorityQueueStatistics.freeAll;
	begin
		p.freeAll;
		shallowFree;
	end;

function priorityQueueStatistics.getObject (where: reference): obj;
	begin
		getObject := p.getObject(where);
	end;

function priorityQueueStatistics.insert (x: obj; function compare (a, b: obj): relationObj): reference;
	begin
		insert := p.insert(x, compare);
		insertions := insertions + 1;
		numberOfElements := numberOfElements + 1;
		if numberOfElements > maxLength then
			maxLength := numberOfElements;
	end;

procedure priorityQueueStatistics.delete (x: reference; function compare (a, b: obj): relationObj);
	begin
		p.delete(x, compare);
		deletions := deletions + 1;
		numberOfElements := numberOfElements - 1;
	end;

function priorityQueueStatistics.isEmpty: boolean;
	begin
		isEmpty := p.isEmpty;
	end;

function priorityQueueStatistics.minimum: reference;
	begin
		minimum := p.minimum;
	end;

function priorityQueueStatistics.ith (i: longInt): reference;
	begin
		ith := p.ith(i);
	end;

procedure priorityQueueStatistics.forAll (procedure whatToDo (x: obj));
	begin
		p.forAll(whatToDo);
	end;

function priorityQueueStatistics.infoString: Str255;
	begin
		infoString := stringOf('(M=', maxLength : 1, ', I=', insertions : 1, ', D=', deletions : 1, ')');
	end;

procedure initPriorityQueue;
	var
		d: dictPriorityQueue;
		h: heapPriorityQueue;
	begin
		new(d);
		failNil(Handle(d));
		prototypes[dictPriorityQueueType] := d;
		new(h);
		failNil(Handle(h));
		prototypes[heapPriorityQueueType] := h;
	end;


end.
