(*	XYZ GeoBench (eXperimental geometrY Zurich):		*)
(*	An object oriented, interactive workbench for		*)
(*	experimental geometric computation.					*)
(*														*)
(*	 1999 by Peter Schorn, peter.schorn@acm.org		*)
(*	Culmannstrasse 77, CH-8006 Zrich, Switzerland		*)
(*	This software may be copied freely for educational	*)
(*	and research purposes only. For commercial use,		*)
(*	please contact the author.							*)
(*														*)
(*	Version 5.0.5: 30-Sep-1999							*)

unit randomizedTriangulation;

interface

uses
	triangulation;


type
	triangleIndex = 1..3;

	simpleTriangle = array[triangleIndex] of point2d;

	seqIndexAsObject = object(obj)
			value: seqIndex;
			function getInfo: obj;
			override;
			function getTextType: Str255;
			override;
		end;

	triangleTreeNode = object(obj)
			p: point2d;
			sons: array[triangleIndex] of obj;
			procedure free;
			override;
			function duplicate: obj;
			override;
			function getInfo: obj;
			override;
			function getTextType: Str255;
			override;
			function height: longInt;
		end;

	searchTriangulation2d = object(triangulation2d)
			boundingTriangle: simpleTriangle;
			root: triangleTreeNode;
			procedure free;
			override;
			procedure freeAll;
			override;
			function duplicate: obj;
			override;
			function getInfo: obj;
			override;
			function getTextType: Str255;
			override;
			function divideTriangle (t4TriangleId: seqIndex; p: point2d): triangleTreeNode;
			procedure find (p: point2d; var node: triangleTreeNode; var sonIndex: triangleIndex);
			procedure initFromPointVector (pv: vector; shuffle: boolean);
		end;

procedure initSearchTriangulation2d;

function delaunayFlipOk (x: graphEdgeTriangulation): boolean;

function delaunayFlipOkReverse (x: graphEdgeTriangulation): boolean;

function justFlipOk (x: graphEdgeTriangulation): boolean;

function inCircle (px, py, qx, qy, rx, ry, sx, sy: extended): boolean;

implementation

{$SETC ATriangulation = TRUE }

function seqIndexAsObject.getTextType: Str255;
	begin
		getTextType := 'Triangle Index';
	end;

function triangleTreeNode.getTextType: Str255;
	begin
		getTextType := 'Triangle Tree Node';
	end;

function searchTriangulation2d.getTextType: Str255;
	begin
		getTextType := concat(inherited getTextType, ' + Search');
	end;

function seqIndexAsObject.getInfo: obj;
	var
		v: vector;
	begin
		v := newVector(1);
		appendObjInfoExtended(v, 'Triangle Index', value, true);
		getInfo := v;
	end;

function triangleTreeNode.getInfo: obj;
	var
		v: vector;
		i: triangleIndex;
	begin
		v := newVector(4);
		appendObjInfoObject(v, 'Cut Point', p, true);
		for i := 1 to 3 do
			appendObjInfoObject(v, StringOf('Son[', i : 1, ']'), sons[i], true);
		getInfo := v;
	end;

function searchTriangulation2d.getInfo: obj;
	var
		v: vector;
		i: triangleIndex;
	begin
		v := vector(inherited getInfo);
		appendObjInfoObject(v, 'Search Tree', root, true);
		for i := 1 to 3 do
			appendObjInfoObject(v, StringOf('Bounding Triangle[', i : 1, ']'), boundingTriangle[i], true);
		getInfo := v;
	end;

function triangleTreeNode.height: longInt;
	var
		max, m: longInt;
		i: triangleIndex;
	begin
		max := 0;
		for i := 1 to 3 do
			if member(sons[i], triangleTreeNode) then begin
				m := triangleTreeNode(sons[i]).height;
				if m > max then
					max := m;
			end;
		height := max + 1;
	end;

function triangleTreeNode.duplicate: obj;
	var
		newNode: triangleTreeNode;
		i: triangleIndex;
	begin
		newNode := triangleTreeNode(shallowDuplicate);
		for i := 1 to 3 do
			newNode.sons[i] := sons[i].duplicate;
		duplicate := newNode;
	end;

function inCircle (px, py, qx, qy, rx, ry, sx, sy: extended): boolean;
	const
		epsilon = 1e-9;
	var
		dxpq, dypq, dxpr, dypr, dxps, dyps, dxqr, dyqr, dxqs, dyqs, dxrs, dyrs, dpqr, np, nq, nr, ns, pos, neg: extended;

	procedure add (x: extended);
		begin
			if x > 0 then
				pos := pos + x
			else
				neg := neg - x;
		end;

	procedure sub (x: extended);
		begin
			if x > 0 then
				neg := neg + x
			else
				pos := pos - x;
		end;

	begin
		dxpq := px - qx;
		dypq := py - qy;
		dxpr := px - rx;
		dypr := py - ry;
		dxps := px - sx;
		dyps := py - sy;
		dxqr := qx - rx;
		dyqr := qy - ry;
		dxqs := qx - sx;
		dyqs := qy - sy;
		dxrs := rx - sx;
		dyrs := ry - sy;
		dpqr := dxpq * dyqr - dypq * dxqr;
		np := sqr(px) + sqr(py);
		nq := sqr(qx) + sqr(qy);
		nr := sqr(rx) + sqr(ry);
		ns := sqr(sx) + sqr(sy);
		pos := 0.0;
		neg := 0.0;
		add(np * dxqr * dyrs);
		sub(np * dyqr * dxrs);
		sub(nq * dxpr * dyrs);
		add(nq * dypr * dxrs);
		add(nr * dxpq * dyqs);
		sub(nr * dypq * dxqs);
		sub(ns * dxpq * dyqr);
		add(ns * dypq * dxqr);
		if dpqr > 0 then
			inCircle := pos > neg * (1.0 + epsilon)
		else if dpqr < 0 then
			inCircle := neg > pos * (1.0 + epsilon)
		else if dxpr * dyrs = dypr * dxrs then
			inCircle := false
		else
			inCircle := ((px < rx) & (rx < qx) | (qx < rx) & (rx < px) | (px = rx) & (rx = qx) & ((py < ry) & (ry < qy) | (qy < ry) & (ry < py))) & (sqr(dxps + dxqs) + sqr(dyps + dyqs) < (sqr(dxpq) + sqr(dypq)) * (1.0 - epsilon))
	end;

function inCircle2d (pTest, p1, p2, p3: point2d): boolean;
	begin
		inCircle2d := inCircle(p1.getX, p1.getY, p2.getX, p2.getY, p3.getX, p3.getY, pTest.getX, pTest.getY);
	end;

function delaunayFlipOk (x: graphEdgeTriangulation): boolean;
	var
		pStart, pEnd, pLeft, pRight: point2d;
	begin
		if x.triangleB > 0 then begin
			pStart := point2d(currentTriangulation.points.elements^^[x.startVertex]);
			pEnd := point2d(currentTriangulation.points.elements^^[x.endVertex]);
			with currentTriangulation.triangles do begin
				triangle(elements^^[x.triangleA]).getThirdPoint(x, obj(pLeft));
				triangle(elements^^[x.triangleB]).getThirdPoint(x, obj(pRight));
			end;
			delaunayFlipOk := inCircle2d(pLeft, pStart, pEnd, pRight) | inCircle2d(pRight, pStart, pEnd, pLeft);
		end
		else
			delaunayFlipOk := false;
	end;

function delaunayFlipOkReverse (x: graphEdgeTriangulation): boolean;
	begin
		delaunayFlipOkReverse := triangulation2d(currentTriangulation).flipOk(x) & not delaunayFlipOk(x);
	end;

function justFlipOk (x: graphEdgeTriangulation): boolean;
	begin
		justFlipOk := triangulation2d(currentTriangulation).flipOk(x);
	end;

procedure swap (var s: seqIndexArray; a, b: seqIndex);
	var
		t: seqIndex;
	begin
		t := s[a];
		s[a] := s[b];
		s[b] := t;
	end;

{$IFC NOT THINK_PASCAL }
{$PUSH }
{$PRAGMAC warn_unusedarg off }
{$ENDC }
function edgeContainer.applicableFlips (function flipOk (x: graphEdgeTriangulation): boolean): seqIndex0;
	begin
		unimplementedMethod('edgeContainer.applicableFlips');
		applicableFlips := 0;
	end;
{$IFC NOT THINK_PASCAL }
{$POP }
{$ENDC }

function stackC.applicableFlips (function flipOk (x: graphEdgeTriangulation): boolean): seqIndex0;
	var
		count, i: seqIndex0;
	begin
		count := 0;
		for i := 1 to stackPointer do
			if flipOk(graphEdgeTriangulation(currentTriangulation.edges.elements^^[stackH^^[i]])) then
				count := count + 1;
		applicableFlips := count;
	end;

function queueC.applicableFlips (function flipOk (x: graphEdgeTriangulation): boolean): seqIndex0;
	var
		countFlips, i, j: seqIndex0;
	begin
		countFlips := 0;
		j := bottom;
		for i := 1 to count do begin
			if flipOk(graphEdgeTriangulation(currentTriangulation.edges.elements^^[stackH^^[j]])) then
				countFlips := countFlips + 1;
			j := j + 1;
			if j > l then
				j := 1;
		end;
		applicableFlips := countFlips;
	end;


procedure edgeContainer.initContainer (length: seqIndex);
	var
		i: seqIndex;
	begin
		stackH := seqIndexArrayH(dynamicArray(length, sizeOf(seqIndex)));
		onStackH := booleanArrayH(dynamicArray(length, sizeOf(boolean)));
		for i := 1 to length do begin
			stackH^^[i] := i;
			onStackH^^[i] := true;
		end;
	end;

procedure stackC.initContainer (length: seqIndex);
	begin
		inherited initContainer(length);
		stackPointer := length;
	end;

procedure queueC.initContainer (length: seqIndex);
	begin
		inherited initContainer(length);
		top := length;
		bottom := 1;
		l := length;
		count := length;
	end;


{$IFC NOT THINK_PASCAL }
{$PUSH }
{$PRAGMAC warn_unusedarg off }
{$ENDC }
procedure edgeContainer.addEdge (e: seqIndex);
	begin
		unimplementedMethod('edgeContainer.addEdge');
	end;
{$IFC NOT THINK_PASCAL }
{$POP }
{$ENDC }

procedure stackC.addEdge (e: seqIndex);
	begin
		stackPointer := stackPointer + 1;
		stackH^^[stackPointer] := e;
	end;

procedure queueC.addEdge (e: seqIndex);
	begin
		top := top + 1;
		if top > l then
			top := 1;
		stackH^^[top] := e;
		count := count + 1;
	end;


procedure edgeContainer.addTriangle (triangleIndex, exceptionEdge: seqIndex);
	begin
		with triangle(currentTriangulation.triangles.elements^^[triangleIndex]) do begin
			if not onStackH^^[edgeA] & (edgeA <> exceptionEdge) then begin
				addEdge(edgeA);
				onStackH^^[edgeA] := true;
			end;
			if not onStackH^^[edgeB] & (edgeB <> exceptionEdge) then begin
				addEdge(edgeB);
				onStackH^^[edgeB] := true;
			end;
			if not onStackH^^[edgeC] & (edgeC <> exceptionEdge) then begin
				addEdge(edgeC);
				onStackH^^[edgeC] := true;
			end;
		end;
	end;


{$IFC NOT THINK_PASCAL }
{$PUSH }
{$PRAGMAC warn_unusedarg off }
{$ENDC }
function edgeContainer.getEdge (function flipOk (x: graphEdgeTriangulation): boolean): graphEdgeTriangulation;
	begin
		unimplementedMethod('edgeContainer.getEdge');
		getEdge := nil;
	end;

function stackC.getEdge (function flipOk (x: graphEdgeTriangulation): boolean): graphEdgeTriangulation;
	var
		result: graphEdgeTriangulation;
	begin
		result := graphEdgeTriangulation(currentTriangulation.edges.elements^^[stackH^^[stackPointer]]);
		stackPointer := stackPointer - 1;
		onStackH^^[result.id] := false;
		getEdge := result;
	end;

function queueC.getEdge (function flipOk (x: graphEdgeTriangulation): boolean): graphEdgeTriangulation;
	var
		result: graphEdgeTriangulation;
	begin
		result := graphEdgeTriangulation(currentTriangulation.edges.elements^^[stackH^^[bottom]]);
		bottom := bottom + 1;
		if bottom > l then
			bottom := 1;
		onStackH^^[result.id] := false;
		getEdge := result;
		count := count - 1;
	end;
{$IFC NOT THINK_PASCAL }
{$POP }
{$ENDC }

function randomC.getEdge (function flipOk (x: graphEdgeTriangulation): boolean): graphEdgeTriangulation;
	begin
{$IFC THINK_PASCAL }
		swap(stackH^^, stackPointer, trunc(randomReal * stackPointer) + 1);
{$ELSEC }
		swap(stackH^^, stackPointer, truncLong(randomReal * stackPointer) + 1);
{$ENDC }
		getEdge := inherited getEdge(flipOk);
	end;

function degreeC.getEdge (function flipOk (x: graphEdgeTriangulation): boolean): graphEdgeTriangulation;
	var
		degrees: seqIndexArray0H;
		i, j, p, q: seqIndex0;
		deltaPotential, minDeltaPotential: longInt;
		a, b, c, d: seqIndex;
		x: graphEdgeTriangulation;

	procedure setDegree (x: graphEdgeTriangulation);
		begin
			with x do begin
				degrees^^[startVertex] := degrees^^[startVertex] + 1;
				degrees^^[endVertex] := degrees^^[endVertex] + 1;
			end;
		end;

	begin
		with currentTriangulation do begin
			degrees := seqIndexArray0H(dynamicArray(points.length, sizeOf(seqIndex)));
			for i := 1 to points.length do
				degrees^^[i] := 0;
			edges.forAll(setDegree);
			minDeltaPotential := 4 * points.length;
			j := 0;
			for i := 1 to stackPointer do begin
				x := graphEdgeTriangulation(edges.elements^^[stackH^^[i]]);
				if flipOk(x) then
					with x do begin
						getOtherPointsId(p, q);
						a := degrees^^[startVertex];
						b := degrees^^[p];
						c := degrees^^[endVertex];
						d := degrees^^[q];
						deltaPotential := abs(a - b - 2) + abs(b - c + 2) + abs(c - d - 2) + abs(d - a + 2) - abs(a - b) - abs(b - c) - abs(c - d) - abs(d - a);
						if deltaPotential < minDeltaPotential then begin
							minDeltaPotential := deltaPotential;
							j := i;
						end;
					end;
			end;
		end;
{$IFC THINK_PASCAL }
		disposHandle(Handle(degrees));
{$ELSEC }
		DisposeHandle(Handle(degrees));
{$ENDC }
		if j = 0 then
			stackPointer := 1
		else
			swap(stackH^^, stackPointer, j);
		getEdge := inherited getEdge(flipOk);
	end;

function minSin (a, b, c, area2: extended): extended;
	var
		result, t: extended;
	begin
		assert((area2 > 0) & (a > 0) & (b > 0) & (c > 0), 'triangle non-degenerate');
		result := 4.0;
		t := area2 / (b * c);
		if t < result then
			result := t;
		t := area2 / (a * c);
		if t < result then
			result := t;
		t := area2 / (a * b);
		if t < result then
			result := t;
		minSin := result;
	end;

function minAngleC.getDelta (a, b, c, d: point2d): extended;
	(*	Before flip: triangles abc and abd. After flip: triangles acd and bcd. *)
	(* Idea: maximize increase in sin() of the smallest angle. *)
	var
		ab, ac, ad, bc, bd, cd, result: extended;
	begin
		ab := a.distanceX(b);
		ac := a.distanceX(c);
		ad := a.distanceX(d);
		bc := b.distanceX(c);
		bd := b.distanceX(d);
		cd := c.distanceX(d);
		result := min(minSin(ac, ad, cd, abs(a.whichSideX(c, d))), minSin(bc, bd, cd, abs(b.whichSideX(c, d)))) - min(minSin(ab, ac, bc, abs(a.whichSideX(b, c))), minSin(ab, ad, bd, abs(a.whichSideX(b, d))));
		getDelta := result;
	end;

function minAngle2C.getDelta (a, b, c, d: point2d): extended;
	(*	Before flip: triangles abc and abd. After flip: triangles acd and bcd. *)
	(* Idea: maximize the smallest resulting angle. *)
	var
		ac, ad, bc, bd, cd: extended;
	begin
		ac := a.distanceX(c);
		ad := a.distanceX(d);
		bc := b.distanceX(c);
		bd := b.distanceX(d);
		cd := c.distanceX(d);
		getDelta := min(minSin(ac, ad, cd, abs(a.whichSideX(c, d))), minSin(bc, bd, cd, abs(b.whichSideX(c, d))));
	end;

function minAngleC.getEdge (function flipOk (x: graphEdgeTriangulation): boolean): graphEdgeTriangulation;
	const
		minInfinity = -1000;
	var
		i, j, p, q: seqIndex0;
		deltaPotential, maxDeltaPotential: extended;
		x: graphEdgeTriangulation;
	begin
		with currentTriangulation do begin
			maxDeltaPotential := minInfinity;
			j := 0;
			for i := 1 to stackPointer do begin
				x := graphEdgeTriangulation(edges.elements^^[stackH^^[i]]);
				if flipOk(x) then begin
					x.getOtherPointsId(p, q);
					deltaPotential := getDelta(point2d(points.elements^^[x.startVertex]), point2d(points.elements^^[x.endVertex]), point2d(points.elements^^[p]), point2d(points.elements^^[q]));
					if deltaPotential > maxDeltaPotential then begin
						maxDeltaPotential := deltaPotential;
						j := i;
					end;
				end;
			end;
		end;
		if j = 0 then
			stackPointer := 1
		else
			swap(stackH^^, stackPointer, j);
		getEdge := inherited getEdge(flipOk);
	end;

function edgeContainer.isNotEmpty: boolean;
	begin
		unimplementedMethod('edgeContainer.isNotEmpty');
		isNotEmpty := false;
	end;

function stackC.isNotEmpty: boolean;
	begin
		isNotEmpty := stackPointer > 0;
	end;

function queueC.isNotEmpty: boolean;
	begin
		isNotEmpty := count > 0;
	end;


procedure edgeContainer.exitContainer;
	begin
{$IFC THINK_PASCAL }
		disposHandle(Handle(onStackH));
		disposHandle(Handle(stackH));
{$ELSEC }
		DisposeHandle(Handle(onStackH));
		DisposeHandle(Handle(stackH));
{$ENDC }
		shallowFree;
	end;

function triangulation.optimize (function flipOk (x: graphEdgeTriangulation): boolean; container: edgeContainer): longInt;
	var
		x: graphEdgeTriangulation;
		flips: longInt;
	begin
		currentTriangulation := self;
		container.initContainer(edges.length);
		flips := 0;
{$IFC AnimationEnabled AND ATriangulation }
		if animationFlag[ATriangulationItem] then
			interactiveHighlight;
{$ENDC }
		while container.isNotEmpty do begin
{$IFC AnimationEnabled }
			if userAbort then begin
				container.exitContainer;
				exit(optimize);
			end;
{$ENDC }
			x := container.getEdge(flipOk);
			if flipOk(x) then begin
{$IFC AnimationEnabled AND ATriangulation }
				if animationFlag[ATriangulationItem] then begin
					displayInfoString(stringOf('Flips possible = ', container.applicableFlips(flipOk) + 1 : 1));
					x.interactiveFlash;
					x.interactiveOutput;
				end;
{$ENDC }
				flip(x);
				flips := flips + 1;
{$IFC AnimationEnabled AND ATriangulation }
				if animationFlag[ATriangulationItem] then begin
					x.interactiveOutput;
					waitForClick(animationFlag[ATriangulationItem]);
				end;
{$ENDC }
				container.addTriangle(x.triangleA, x.id);
				container.addTriangle(x.triangleB, x.id);
			end;
		end;
		container.exitContainer;
		optimize := flips;
	end;

function edgeContainerX.applicableFlips: seqIndex0;
	begin
		applicableFlips := length;
	end;

{$IFC NOT THINK_PASCAL }
{$PUSH }
{$PRAGMAC warn_unusedarg off }
{$ENDC }
procedure edgeContainerX.initContainer (len: seqIndex);
	begin
		length := 0;
	end;
{$IFC NOT THINK_PASCAL }
{$POP }
{$ENDC }

procedure stackCX.initContainer (len: seqIndex);
	var
		i: seqIndex;
	begin
		inherited initContainer(len);
		stackLocationH := seqIndexArray0H(dynamicArray(len, sizeOf(seqIndex0)));
		for i := 1 to len do
			stackLocationH^^[i] := 0;
		stackH := seqIndexArrayH(dynamicArray(len, sizeOf(seqIndex)));
	end;

procedure interactiveCX.initContainer (len: seqIndex);
	begin
		currentTriangulation.interactiveHighlight;
		redrawScreen := true;
		animationStatus := animationFlag[ATriangulationItem];
		animationFlag[ATriangulationItem] := false;
		inherited initContainer(len);
	end;

procedure interactiveFlipCX.initContainer (len: seqIndex);
	var
		i: seqIndex;
	begin
		inherited initContainer(len);
		delaunayEdgeH := booleanArrayH(dynamicArray(len, sizeOf(boolean)));
		for i := 1 to len do
			delaunayEdgeH^^[i] := false;
	end;

{$IFC NOT THINK_PASCAL }
{$PUSH }
{$PRAGMAC warn_unusedarg off }
{$ENDC }
procedure edgeContainerX.updateContainerEdge (e: seqIndex; function flipOk (x: graphEdgeTriangulation): boolean);
	begin
		unimplementedMethod('edgeContainerX.updateContainerEdge');
	end;
{$IFC NOT THINK_PASCAL }
{$POP }
{$ENDC }

procedure stackCX.updateContainerEdge (e: seqIndex; function flipOk (x: graphEdgeTriangulation): boolean);
	var
		x: graphEdgeTriangulation;
		l, stackLocation: seqIndex0;
	begin
		x := graphEdgeTriangulation(currentTriangulation.edges.elements^^[e]);
		stackLocation := stackLocationH^^[e];
		updateGraphicsStatus(e, x);
		if flipOk(x) then begin
			if stackLocation = 0 then begin
				length := length + 1;
				stackH^^[length] := e;
				stackLocationH^^[e] := length;
				updateEdgeGraphics(x);
			end;
		end
		else if stackLocation > 0 then begin
			l := stackH^^[length];
			stackH^^[stackLocation] := l;
			stackLocationH^^[l] := stackLocation;
			stackLocationH^^[e] := 0;
			length := length - 1;
			updateEdgeGraphics(x);
		end;
	end;

procedure showDelaunayStatus (x: graphEdgeTriangulation);
	var
		p: realPoint;
		cx, cy: extended;
	begin
		setCurrentVerticesVector(currentTriangulation.points);
		x.centerOfGravity(cx, cy);
		new(p);
		failNil(Handle(p));
		with p do begin
			setX(cx);
			setY(cy);
			labelPoint('D');
		end;
		p.free;
	end;

procedure interactiveFlipCX.updateGraphicsStatus (e: seqIndex; x: graphEdgeTriangulation);
	begin
		if delaunayFlipOk(x) then begin
			if not delaunayEdgeH^^[e] then begin
				showDelaunayStatus(x);
				delaunayEdgeH^^[e] := true;
			end;
		end
		else if delaunayEdgeH^^[e] then begin
			showDelaunayStatus(x);
			delaunayEdgeH^^[e] := false;
		end;
	end;

{$IFC NOT THINK_PASCAL }
{$PUSH }
{$PRAGMAC warn_unusedarg off }
{$ENDC }
procedure stackCX.deselectEdge (x: graphEdgeTriangulation);
	begin
	end;

procedure stackCX.updateEdgeGraphics (x: graphEdgeTriangulation);
	begin
	end;

procedure stackCX.updateGraphicsStatus (e: seqIndex; x: graphEdgeTriangulation);
	begin
	end;
{$IFC NOT THINK_PASCAL }
{$POP }
{$ENDC }

procedure interactiveFlipCX.deselectEdge (x: graphEdgeTriangulation);
	begin
		if delaunayEdgeH^^[x.id] then begin
			showDelaunayStatus(x);
			delaunayEdgeH^^[x.id] := false;
		end;
	end;

procedure edgeContainerX.update (x: graphEdgeTriangulation; function flipOk (x: graphEdgeTriangulation): boolean);
	begin
		updateTriangle(x.triangleA, x.id, flipOk);
		updateTriangle(x.triangleB, x.id, flipOk);
	end;

procedure edgeContainerX.updateTriangle (triangleIndex, exceptionEdge: seqIndex; function flipOk (x: graphEdgeTriangulation): boolean);
	begin
		with triangle(currentTriangulation.triangles.elements^^[triangleIndex]) do begin
			if edgeA <> exceptionEdge then
				updateContainerEdge(edgeA, flipOk);
			if edgeB <> exceptionEdge then
				updateContainerEdge(edgeB, flipOk);
			if edgeC <> exceptionEdge then
				updateContainerEdge(edgeC, flipOk);
		end;
	end;

{$IFC NOT THINK_PASCAL }
{$PUSH }
{$PRAGMAC warn_unusedarg off }
{$ENDC }
procedure interactiveFlipCX.updateTriangle (triangleIndex, exceptionEdge: seqIndex; function flipOk (x: graphEdgeTriangulation): boolean);
	begin
		with triangle(currentTriangulation.triangles.elements^^[triangleIndex]) do begin
			updateContainerEdge(edgeA, flipOk);
			updateContainerEdge(edgeB, flipOk);
			updateContainerEdge(edgeC, flipOk);
		end;
	end;
{$IFC NOT THINK_PASCAL }
{$POP }
{$ENDC }


function edgeContainerX.getEdge: graphEdgeTriangulation;
	begin
		unimplementedMethod('edgeContainerX.getEdge');
		getEdge := nil;
	end;

function stackCX.getEdge: graphEdgeTriangulation;
	var
		result: graphEdgeTriangulation;
	begin
		if length > 0 then begin
			result := graphEdgeTriangulation(currentTriangulation.edges.elements^^[stackH^^[length]]);
			length := length - 1;
			stackLocationH^^[result.id] := 0;
			getEdge := result;
		end
		else
			getEdge := nil;
	end;

procedure stackCX.exchangeWithTop (newPos: seqIndex);
	begin
		swap(stackH^^, length, newPos);
		stackLocationH^^[stackH^^[newPos]] := newPos;
	end;

function randomCX.getEdge: graphEdgeTriangulation;
	begin
		if length = 0 then
			getEdge := nil
		else begin
{$IFC THINK_PASCAL }
			exchangeWithTop(trunc(randomReal * length) + 1);
{$ELSEC }
			exchangeWithTop(truncLong(randomReal * length) + 1);
{$ENDC }
			getEdge := inherited getEdge;
		end;
	end;

procedure interactiveCX.updateEdgeGraphics (x: graphEdgeTriangulation);
	begin
		x.interactiveHighlight;
	end;

function interactiveCX.getEdge: graphEdgeTriangulation;
	function index (x: seqIndex): seqIndex;
		begin
			index := stackH^^[x];
		end;
		var
			selected: seqIndex0;
			temp: graphEdgeTriangulation;
	begin
		displayInfoString(stringOf('Flips Possible = ', applicableFlips : 1));
		selected := currentTriangulation.edges.interactiveSelect(true, true, index, 1, length);
		if selected = 0 then begin
			getEdge := nil;
			exit(getEdge);
		end;
		temp := graphEdgeTriangulation(currentTriangulation.edges.elements^^[stackH^^[selected]]);
		temp.interactiveOutput;
		temp.interactiveHighlight;
		deselectEdge(temp);
		exchangeWithTop(selected);
		getEdge := inherited getEdge;
	end;

procedure interactiveCX.update (x: graphEdgeTriangulation; function flipOk (x: graphEdgeTriangulation): boolean);
	begin
		x.interactiveOutput;
		inherited update(x, flipOk);
	end;

function edgeContainerX.isNotEmpty: boolean;
	begin
		isNotEmpty := length > 0;
	end;


procedure edgeContainerX.exitContainer;
	begin
		unimplementedMethod('edgeContainerX.exitContainer');
	end;

procedure stackCX.exitContainer;
	begin
{$IFC THINK_PASCAL }
		disposHandle(Handle(stackH));
		disposHandle(Handle(stackLocationH));
{$ELSEC }
		DisposeHandle(Handle(stackH));
		DisposeHandle(Handle(stackLocationH));
{$ENDC }
		shallowFree;
	end;

procedure interactiveCX.exitContainer;
	begin
		doBeep := false;
		animationFlag[ATriangulationItem] := animationStatus;
		inherited exitContainer;
	end;

procedure interactiveFlipCX.exitContainer;
	begin
{$IFC THINK_PASCAL }
		disposHandle(Handle(delaunayEdgeH));
{$ELSEC }
		DisposeHandle(Handle(delaunayEdgeH));
{$ENDC }
		inherited exitContainer;
	end;



function triangulation.optimize2 (function flipOk (x: graphEdgeTriangulation): boolean; container: edgeContainerX): longInt;
	var
		x: graphEdgeTriangulation;
		flips: longInt;
		i: seqIndex0;
	begin
		currentTriangulation := self;
		container.initContainer(edges.length);
		flips := 0;
{$IFC AnimationEnabled AND ATriangulation }
		if animationFlag[ATriangulationItem] then
			interactiveHighlight;
{$ENDC }
		for i := 1 to edges.length do
			container.updateContainerEdge(i, flipOk);
		x := container.getEdge;
		while x <> nil do begin
{$IFC AnimationEnabled }
			if userAbort then begin
				container.exitContainer;
				exit(optimize2);
			end;
{$ENDC }
{$IFC AnimationEnabled AND ATriangulation }
			if animationFlag[ATriangulationItem] then begin
				displayInfoString(stringOf('Flips possible = ', container.applicableFlips + 1 : 1));
				x.interactiveFlash;
				x.interactiveOutput;
			end;
{$ENDC }
			flip(x);
			flips := flips + 1;
{$IFC AnimationEnabled AND ATriangulation }
			if animationFlag[ATriangulationItem] then begin
				x.interactiveOutput;
				waitForClick(animationFlag[ATriangulationItem]);
			end;
{$ENDC }
			container.update(x, flipOk);
			x := container.getEdge;
		end;
		container.exitContainer;
		optimize2 := flips;
	end;

procedure graphEdgeTriangulation.interactiveOutput;
	begin
		setCurrentVerticesVector(currentTriangulation.points);
		inherited interactiveOutput;
	end;

function graphEdgeTriangulation.isSelected (where: obj): boolean;
	begin
		setCurrentVerticesVector(currentTriangulation.points);
		isSelected := inherited isSelected(where);
	end;

procedure triangleTreeNode.free;
	var
		i: triangleIndex;
	begin
		for i := 1 to 3 do
			sons[i].free;
		shallowFree;
	end;

procedure searchTriangulation2d.free;
	var
		i: triangleIndex;
	begin
		root.free;
		for i := 1 to 3 do
			boundingTriangle[i].free;
		inherited free;
	end;

procedure searchTriangulation2d.freeAll;
	var
		i: triangleIndex;
	begin
		root.free;
		for i := 1 to 3 do
			boundingTriangle[i].free;
		inherited freeAll;
	end;

function searchTriangulation2d.duplicate: obj;
	var
		newTree: searchTriangulation2d;
		i: triangleIndex;
	begin
		newTree := searchTriangulation2d(inherited duplicate);
		newTree.root := triangleTreeNode(root.duplicate);
		for i := 1 to 3 do
			boundingTriangle[i] := point2d(boundingTriangle[i].duplicate);
		duplicate := newTree;
	end;

function triangulation2d.createLineSegments: lineSegmentVector;
	var
		udg: simpleUndirectedGraph;
		homogenEdges: homogenVector;
	begin
		new(homogenEdges);
		failNil(Handle(homogenEdges));
		homogenEdges.shallowInitializeFromObject(edges);
		new(udg);
		failNil(Handle(udg));
		udg.vertices := points;
		udg.edges := homogenEdges;
		createLineSegments := udg.getEdges;
		dispose(udg);
		dispose(homogenEdges);
	end;

function triangulation.toUndirectedGraph: undirectedGraph;
	var
		result: undirectedGraph;

	procedure doAddEdge (e: graphEdgeTriangulation);
		begin
			result.edges.append(e.duplicateTypeTransfer(graphEdgeType));
		end;

	begin (* triangulation2d.toUndirectedGraph *)
		new(result);
		failNil(Handle(result));
		result.vertices := vector(points.duplicateAll);
		result.edges := newHomogenVector(edges.length);
		result.adjacency := newHomogenVector(points.length);
		edges.forAll(doAddEdge);
		result.initAdjacency;
		toUndirectedGraph := result;
	end;

function searchTriangulation2d.divideTriangle (t4TriangleId: seqIndex; p: point2d): triangleTreeNode;
(* t4TriangleId is the index of the triangle to be divided. 'currentTriangulation' must have been set to the current triangulation. *)
(*	'p' is the point responsible for the division. *)
	var
		t4Triangle, t5Triangle, t6Triangle: triangle2d;
		t5TriangleId, t6TriangleId, p0, p1, p2, p3: seqIndex;
		e1, e2, e3, e4, e5, e6: graphEdgeTriangulation;
		result: triangleTreeNode;
		r: seqIndexAsObject;
	begin
		new(t5Triangle);
		failNil(Handle(t5Triangle));
		new(t6Triangle);
		failNil(Handle(t6Triangle));
		with triangles do begin
			t4Triangle := triangle2d(elements^^[t4TriangleId]);
			append(t5Triangle);
			t5TriangleId := length;
			append(t6Triangle);
			t6TriangleId := length;
		end;
		with points do begin
			append(p.duplicate);
			p0 := length;
		end;
		with t4Triangle do begin
			getEdges(e1, e2, e3);
			p3 := getThirdPointId(e1);
			p1 := getThirdPointId(e2);
			p2 := getThirdPointId(e3);
		end;
		with currentTriangulation do begin
			e4 := addNewEdge(p1, p0);
			e5 := addNewEdge(p2, p0);
			e6 := addNewEdge(p3, p0);
		end;
{$IFC AnimationEnabled AND ATriangulation }
		if animationFlag[ATriangulationItem] then begin
			p.interactiveFlash;
			waitForClick(animationFlag[ATriangulationItem]);
		end;
		if animationFlag[ATriangulationItem] then begin
			e4.interactiveOutput;
			e5.interactiveOutput;
			e6.interactiveOutput;
			waitForClick(animationFlag[ATriangulationItem]);
		end;
{$ENDC }
		with t4Triangle do begin
			edgeA := e1.id;
			edgeB := e5.id;
			edgeC := e4.id;
		end;
		with t5Triangle do begin
			edgeA := e4.id;
			edgeB := e6.id;
			edgeC := e3.id;
		end;
		with t6Triangle do begin
			edgeA := e5.id;
			edgeB := e2.id;
			edgeC := e6.id;
		end;
		with e2 do
			if triangleA = t4TriangleId then
				triangleA := t6TriangleId
			else
				triangleB := t6TriangleId;
		with e3 do
			if triangleA = t4TriangleId then
				triangleA := t5TriangleId
			else
				triangleB := t5TriangleId;
		with e4 do begin
			triangleA := t4TriangleId;
			triangleB := t5TriangleId;
		end;
		with e5 do begin
			triangleA := t4TriangleId;
			triangleB := t6TriangleId;
		end;
		with e6 do begin
			triangleA := t5TriangleId;
			triangleB := t6TriangleId;
		end;
		new(result);
		failNil(Handle(result));
		result.p := p;
		new(r);
		failNil(Handle(r));
		r.value := t6TriangleId;
		result.sons[1] := r;
		new(r);
		failNil(Handle(r));
		r.value := t5TriangleId;
		result.sons[2] := r;
		new(r);
		failNil(Handle(r));
		r.value := t4TriangleId;
		result.sons[3] := r;
		divideTriangle := result;
	end;

function getBoundingTriangle (pv: vector): simpleTriangle;
	const
		epsilon = 0.1;
	var
		p: point2d;
		i: seqIndex;
		xMin, xMax, yMin, yMax, thisX, thisY: extended;
		result: simpleTriangle;
	begin
		with point2d(pv.elements^^[1]) do begin
			xMin := getX;
			yMin := getY;
		end;
		xMax := xMin;
		yMax := yMin;
		for i := 2 to pv.length do begin
			with point2d(pv.elements^^[i]) do begin
				thisX := getX;
				thisY := getY;
			end;
			if thisX < xMin then
				xMin := thisX
			else if thisX > xMax then
				xMax := thisX;
			if thisY < yMin then
				yMin := thisY
			else if thisY > yMax then
				yMax := thisY;
		end;
		p := point2d(newObj(currentPointType));
		with p do begin
			setX((3 * xMin - xMax) / 2);
			setY((1 - epsilon) * yMin);
		end;
		result[1] := p;
		p := point2d(newObj(currentPointType));
		with p do begin
			setX((3 * xMax - xMin) / 2);
			setY((1 - epsilon) * yMin);
		end;
		result[2] := p;
		p := point2d(newObj(currentPointType));
		with p do begin
			setX((xMax + xMin) / 2);
			setY((1 + epsilon) * (2 * yMax - yMin));
		end;
		result[3] := p;
		getBoundingTriangle := result;
	end;

{$IFC AnimationEnabled AND ATriangulation }
procedure displaySimpleTriangle (t: simpleTriangle; pen: integer);
	begin
		PenSize(pen, pen);
		t[1].drawLine(t[2]);
		t[2].drawLine(t[3]);
		t[3].drawLine(t[1]);
		PenSize(1, 1);
	end;
{$ENDC }

	var
		searchPathLength: longInt;

procedure searchTriangulation2d.find (p: point2d; var node: triangleTreeNode; var sonIndex: triangleIndex);
	var
		currentNode: triangleTreeNode;
		currentTriangle: simpleTriangle;
		index: 0..3;
		ws1, ws2, ws3: signType;
	begin
		currentNode := root;
		currentTriangle := boundingTriangle;
		index := 0;
{$IFC AnimationEnabled AND ATriangulation }
		if animationFlag[ASearchItem] then
			p.interactiveHighlight;
{$ENDC }
		repeat
			searchPathLength := searchPathLength + 1;
			ws1 := p.whichSideSign(currentNode.p, currentTriangle[1]);
			ws2 := p.whichSideSign(currentNode.p, currentTriangle[2]);
			ws3 := p.whichSideSign(currentNode.p, currentTriangle[3]);
{$IFC AnimationEnabled AND ATriangulation }
			if animationFlag[ASearchItem] then begin
				displaySimpleTriangle(currentTriangle, 3);
				waitForClick(animationFlag[ASearchItem]);
				displaySimpleTriangle(currentTriangle, 3);
			end;
{$ENDC }
			if (ws2 <= 0) & (ws3 >= 0) then
				if member(currentNode.sons[1], triangleTreeNode) then begin
					currentTriangle[1] := currentNode.p;
					currentNode := triangleTreeNode(currentNode.sons[1]);
				end
				else
					index := 1
			else if (ws3 <= 0) & (ws1 >= 0) then
				if member(currentNode.sons[2], triangleTreeNode) then begin
					currentTriangle[2] := currentNode.p;
					currentNode := triangleTreeNode(currentNode.sons[2]);
				end
				else
					index := 2
			else if (ws1 <= 0) & (ws2 >= 0) then
				if member(currentNode.sons[3], triangleTreeNode) then begin
					currentTriangle[3] := currentNode.p;
					currentNode := triangleTreeNode(currentNode.sons[3]);
				end
				else
					index := 3
			else
				assert(false, 'Problem in ''searchTriangulation2d.find''');
		until index > 0;
{$IFC AnimationEnabled AND ATriangulation }
		if animationFlag[ASearchItem] then
			p.interactiveHighlight;
{$ENDC }
		sonIndex := index;
		node := currentNode;
	end;

procedure searchTriangulation2d.initFromPointVector (pv: vector; shuffle: boolean);
	var
		p: point2d;
		node, newNode: triangleTreeNode;
		sonIndex: triangleIndex;
		i: seqIndex;
		pvCopy: vector;
		pvCopyFlag: boolean;
	begin
		pvCopyFlag := member(pv, polyLine);
		if pvCopyFlag then
			pvCopy := vector(pv.duplicate)
		else
			pvCopy := pv;
		currentTriangulation := self;
		searchPathLength := 0;
{$IFC AnimationEnabled AND ATriangulation }
		if animationFlag[ATriangulationItem] | animationFlag[ASearchItem] then
			pv.interactiveHighlight;
{$ENDC }
		initialize(pv.length + 3, 3 * pv.length + 3, 2 * pv.length + 1);
		boundingTriangle := getBoundingTriangle(pv);
{$IFC AnimationEnabled AND ATriangulation }
		if animationFlag[ATriangulationItem] then
			displaySimpleTriangle(boundingTriangle, 1);
{$ENDC }
		points.append(boundingTriangle[1].duplicate);
		points.append(boundingTriangle[2].duplicate);
		points.append(boundingTriangle[3].duplicate);
		if shuffle then
			pvCopy.randomShuffleAll;
		addNewTriangle(addNewEdge(1, 2), addNewEdge(2, 3), addNewEdge(3, 1));
		root := divideTriangle(1, point2d(pvCopy.elements^^[1]));
		for i := 2 to pv.length do begin
{$IFC AnimationEnabled }
			if userAbort then begin
				if pvCopyFlag then
					pvCopy.free;
				exit(initFromPointVector);
			end;
{$ENDC }
			p := point2d(pvCopy.elements^^[i]);
			find(p, node, sonIndex);
			newNode := divideTriangle(seqIndexAsObject(node.sons[sonIndex]).value, p);
			node.sons[sonIndex].free;
			node.sons[sonIndex] := newNode;
		end;
		if pvCopyFlag then
			pvCopy.free;
		displayInfoString(StringOf('Total Length Of Search Path = ', searchPathLength : 1, ', Height = ', root.height : 1));
	end;

procedure initSearchTriangulation2d;
	var
		x: searchTriangulation2d;
	begin
		new(x);
		failNil(Handle(x));
		prototypes[searchTriangulation2dType] := x;
	end;


end.
