(*	XYZ GeoBench (eXperimental geometrY Zurich):		*)
(*	An object oriented, interactive workbench for		*)
(*	experimental geometric computation.					*)
(*														*)
(*	 1999 by Peter Schorn, peter.schorn@acm.org		*)
(*	Culmannstrasse 77, CH-8006 Zrich, Switzerland		*)
(*	This software may be copied freely for educational	*)
(*	and research purposes only. For commercial use,		*)
(*	please contact the author.							*)
(*														*)
(*	Version 5.0.5: 30-Sep-1999							*)

unit rectangleContour;

(* Autoren : Peter Lippuner, Peter Skrotzky																*)
(* Datum    : 8.10.90																							*)
(* provides a method to compute the countour of a set of rectangles using the algorithm from	*)
(* the book Computational Geometry by Preparata and Shamos											*)

{$SETC AContourOfRectangle = TRUE }

interface

uses
	rectangleVector;

implementation


type
	statusType = (full, empty, partial); 	(* full  => node full											*)
												(* empty = > node and boths sons empty				*)
												(* partial => node empty, left or right son occupied	*)

	segmentXTree = ^segmentXP;				(* handle *)
	segmentXP = ^segmentX;
	segmentX = record
			start, ending: int;
			count: integer;
			status: statusType;
			leftSon, rightSon: segmentXTree;
		end;

	sideType = (leftSide, rightSide);

	verticalSideType = object(obj)	(* vertical sides of the rectangles *)
			side: sideType;
			abscissae: real;
			bottom, top: integer;				(* index, determines the position in the segment tree *)
			rBottom, rtop: real;
		end;

	realObj = object(realNumber)
			lside, rside: verticalSideType;
		end;

var
	tree: segmentXTree;
	stack: homogenVector;			(* not a  pull-down stack																			*)
	sidesOfContour: list;				(* contains the line segments of the countour (result)										*)
	sideQueue: homogenVector;		(* contains the vertical sides of the rectangles												*)
	indexQueue: homogenVector;	(* makes the relation between the indices in the segment tree and the real y values	*)
	pointQueue: homogenVector;		(* contains a list of points of the result															*)

procedure initSegmentTree (var tree: segmentXTree; begins, ends: integer);
	var
		middle: integer;
		startInt, endingInt: int;
		newSon: segmentXTree;
	begin
		tree := segmentXTree(newHandleFailNil(sizeOf(segmentX)));
		new(startInt);
		failNil(Handle(startInt));
		new(endingInt);
		failNil(Handle(endingInt));
		with tree^^ do begin
			start := startInt;
			ending := endingInt;
			start.value := begins;
			ending.value := ends;
			count := 0;
			status := empty;
			leftSon := nil;
			rightSon := nil;
		end;
		if begins + 1 < ends then begin
			middle := (begins + ends) div 2;
			initSegmentTree(newSon, begins, middle);
			tree^^.leftSon := newSon;
			initSegmentTree(newSon, middle, ends);
			tree^^.rightSon := newSon;
		end;
	end;

procedure deleteSegmentTree (tree: segmentXTree);
	begin
		if tree^^.leftSon <> nil then
			deleteSegmentTree(tree^^.leftSon);
		if tree^^.rightSon <> nil then
			deleteSegmentTree(tree^^.rightSon);
		tree^^.start.free;
		tree^^.ending.free;
{$IFC THINK_PASCAL }
		disposHandle(Handle(tree));
{$ELSEC }
		DisposeHandle(Handle(tree));
{$ENDC }
	end;

procedure updateNode (tree: segmentXTree);
(* fuehrt den Status eines Knotens nach, falls etwas eingetragen oder geloescht wurde *)
	begin
		with tree^^ do
			if count > 0 then
				status := full
			else if (leftSon = nil) | (leftSon^^.status = empty) & (rightSon^^.status = empty) then
				status := empty
			else
				status := partial;
	end;

procedure insertInSegTree (tree: segmentXTree; begins, ends: integer);
	var
		middle: integer;
	begin
		with tree^^ do
			if (begins <= start.value) & (ending.value <= ends) then
				count := count + 1
			else begin
				middle := (start.value + ending.value) div 2;
				if begins < middle then
					insertInSegTree(leftSon, begins, ends);
				if middle < ends then
					insertInSegTree(rightSon, begins, ends);
			end;
		updateNode(tree);
	end;

procedure deleteFromSegTree (tree: segmentXTree; begins, ends: integer);
	var
		middle: integer;
	begin
		with tree^^ do
			if (begins <= start.value) & (ending.value <= ends) then
				count := count - 1
			else begin
				middle := (start.value + ending.value) div 2;
				if begins < middle then
					deleteFromSegTree(leftSon, begins, ends);
				if middle < ends then
					deleteFromSegTree(rightSon, begins, ends);
			end;
		updateNode(tree);
	end;

procedure contribution (tree: segmentXTree; begins, ends: integer);
(* Computes the contribution of a segment and merges contiguous segments using 'stack' *)
	var
		middle: integer;
	begin
		if tree^^.status <> full then
			if (begins <= tree^^.start.value) & (tree^^.ending.value <= ends) & (tree^^.status = empty) then begin (* start, ending is contributeted *)
				if stack.length > 0 then
					if tree^^.start.value = int(stack.elements^^[stack.length]).value then	(* merge contiguous segments *)
						stack.removeLast
					else
						stack.append(tree^^.start)			(* beginning of edge *)
				else
					stack.append(tree^^.start);				(* beginning of edge *)
				stack.append(tree^^.ending);				(* current termination of edge *)
			end
			else begin
				middle := (tree^^.start.value + tree^^.ending.value) div 2;
				if begins < middle then
					contribution(tree^^.leftSon, begins, ends);
				if middle < ends then
					contribution(tree^^.rightSon, begins, ends);
			end
	end;

function insertSide (side: sideType; x, rBottom, rtop: real): verticalSideType;
(* insert the side of a rectangle into the 'sideQueue' *)
	var
		verticalSide: verticalSideType;
	begin
		new(verticalSide);
		failNil(Handle(verticalSide));
		verticalSide.side := side;
		verticalSide.abscissae := x;
		verticalSide.rBottom := rBottom;
		verticalSide.rtop := rtop;
		verticalSide.bottom := 0;
		verticalSide.top := 0;
		sideQueue.append(verticalSide);
		insertSide := verticalSide;
	end;

procedure insertIndex (value: real; lside, rside: verticalSideType);
(* insert a 'realObj' into the 'indexQueue' *)
	var
		ordinate: realObj;
	begin
		new(ordinate);
		failNil(Handle(ordinate));
		ordinate.value := value;
		ordinate.lside := lside;
		ordinate.rside := rside;
		indexQueue.append(ordinate);
	end;

procedure fillIndex (len: integer);
(* update then 'realObj' in the 'indexQueue' with the indices *)
	var
		i: seqIndex0;
	begin
		for i := 1 to len do
			with realObj(indexQueue.elements^^[i]) do begin
				if lside.rBottom = value then begin
					lside.bottom := i;
					rside.bottom := i;
				end
				else begin
					lside.top := i;
					rside.top := i;
				end;
			end;
	end;

function rectangleVector.contourOfUnionOfRectangles: lineSegmentVector; (* Main procedure *)

	var
		result: lineSegmentVector;
		rect: rectangle;
{$IFC AnimationEnabled AND AContourOfRectangle }
		sweepPosAnim: real;
{$ENDC }

	procedure initializeVerticalSideQueue;
		var
			i: seqIndex0;
			lside, rside: verticalSideType;

		function compareSide (a, b: verticalSideType): boolean;
			begin
				if a.abscissae < b.abscissae then
					compareSide := true
				else if (a.abscissae > b.abscissae) | (a.side = rightSide) | (b.side = leftSide) then
					compareSide := false
				else
					compareSide := true;
			end;

		begin (* initializeVerticalSideQueue *)
			sideQueue := newHomogenVector(2 * length);
			indexQueue := newHomogenVector(2 * length);
			for i := 1 to length do begin
				rect := rectangle(elements^^[i]);
				lside := insertSide(leftSide, rect.left, rect.bottom, rect.top);
				rside := insertSide(rightSide, rect.right, rect.bottom, rect.top);
				insertIndex(rect.bottom, lside, rside);
				insertIndex(rect.top, lside, rside);
			end;
			sideQueue.sortAll(compareSide);
			indexQueue.sortAll(lessThanRealNumber);
			fillIndex(2 * length);
		end;

	procedure makeLineSegments (x: real);
(* compute the vertical sides of the contour using the stack and put the points into the 'pointQueue' *)
		var
			ls: lineSegment;
			i: seqIndex0;
		begin
			for i := 1 to stack.length div 2 do begin	(* Bilden einer Seite *)
				new(ls);
				failNil(Handle(ls));
				ls.init;
				ls.p.setX(x);
				ls.q.setX(x);
				ls.p.setY(realObj(indexQueue.elements^^[int(stack.elements^^[2 * i - 1]).value]).value);
				ls.q.setY(realObj(indexQueue.elements^^[int(stack.elements^^[2 * i]).value]).value);
				sidesOfContour.append(ls);
{$IFC AnimationEnabled AND AContourOfRectangle }
				if animationFlag[AContourOfRectanglesItem] then begin
					PenMode(patOr);
					drawPoint(ls.p.getX, ls.p.getY, 4);
					drawPoint(ls.q.getX, ls.q.getY, 4);
					drawThickLineSegment(ls.p.getX, ls.p.getY, ls.q.getX, ls.q.getY);
					PenMode(patXor);
					ls.interactiveFlash;
					waitForClick(animationFlag[AContourOfRectanglesItem]);
				end;
{$ENDC }
				pointQueue.append(ls.p.duplicate);	 (* insert the points into the 'pointQueue' *)
				pointQueue.append(ls.q.duplicate);
			end;
		end;


	procedure processSide (a: verticalSideType);
(* computes the vertical parts of the contour at the position of the sweep line *)
{$IFC AnimationEnabled AND AContourOfRectangle }
		var
			ls: lineSegment;
{$ENDC }
		begin
{$IFC AnimationEnabled }
			if userAbort then begin
				sideQueue.freeAll;
				indexQueue.freeAll;
				sidesOfContour.freeAll;
				deleteSegmentTree(tree);
				pointQueue.freeAll;
				stack.free;
				contourOfUnionOfRectangles := nil;
				exit(contourOfUnionOfRectangles);
			end;
{$ENDC }
{$IFC AnimationEnabled AND AContourOfRectangle }
			if animationFlag[AContourOfRectanglesItem] then begin
				moveSweepX(sweepPosAnim, a.abscissae);
				sweepPosAnim := a.abscissae;
				new(ls);
				failNil(Handle(ls));
				ls.init;
				ls.p.setX(a.abscissae);
				ls.q.setX(a.abscissae);
				ls.p.setY(a.rBottom);
				ls.q.setY(a.rtop);
				ls.interactiveFlash;
				waitForClick(animationFlag[AContourOfRectanglesItem]);
				ls.free;
			end;
{$ENDC }
			if a.side = leftSide then begin
				contribution(tree, a.bottom, a.top);
				makeLineSegments(a.abscissae);
				insertInSegTree(tree, a.bottom, a.top);
			end
			else begin
				deleteFromSegTree(tree, a.bottom, a.top);
				contribution(tree, a.bottom, a.top);
				makeLineSegments(a.abscissae);
			end;
			stack.length := 0;
		end;

	procedure releaseObjects;
		begin
			sideQueue.freeAll;
			indexQueue.freeAll;
			sidesOfContour.free;
			deleteSegmentTree(tree);
			pointQueue.free;
			stack.free;
		end;

	procedure getHorizontalSides;
(* forms the horizontal pieces of the contour using the 'pointQueue' *)
		var
			i: seqIndex0;
			ls: lineSegment;
{$IFC AnimationEnabled AND AContourOfRectangle }
			tempY: extended;
{$ENDC }
		begin
			pointQueue.sortAll(pointLessThanYX);
			for i := 1 to pointQueue.length div 2 do begin
{$IFC AnimationEnabled }
				if userAbort then begin
					sideQueue.freeAll;
					indexQueue.freeAll;
					sidesOfContour.freeAll;
					deleteSegmentTree(tree);
					pointQueue.forSome(freeP, 2 * i - 1, pointQueue.length);
					pointQueue.free;
					stack.free;
					contourOfUnionOfRectangles := nil;
					exit(contourOfUnionOfRectangles);
				end;
{$ENDC }
				new(ls);
				failNil(Handle(ls));
				ls.p := point2d(pointQueue.elements^^[2 * i - 1]);
				ls.q := point2d(pointQueue.elements^^[2 * i]);
				sidesOfContour.append(ls);
{$IFC AnimationEnabled AND AContourOfRectangle }
				if animationFlag[AContourOfRectanglesItem] then begin
					tempY := ls.p.getY;
					moveSweepY(sweepPosAnim, tempY);
					sweepPosAnim := tempY;
					PenMode(patOr);
					drawThickLineSegment(ls.p.getX, tempY, ls.q.getX, tempY);
					PenMode(patXor);
					ls.interactiveFlash;
					waitForClick(animationFlag[AContourOfRectanglesItem]);
				end;
{$ENDC }
			end;
		end;

	procedure reportResults;
		begin
			if sidesOfContour.length = 0 then
				contourOfUnionOfRectangles := nil
			else begin
				new(result);
				failNil(Handle(result));
				result.initFromList(sidesOfContour);
				contourOfUnionOfRectangles := result;
			end;
		end;

	procedure initializeObjects;
		begin
			initSegmentTree(tree, 1, 2 * length);
			initializeVerticalSideQueue;
			sidesOfContour := newEmptyList;
			pointQueue := newHomogenVector(8 * length);
			stack := newHomogenVector(2 * length);
		end;

	begin (* rectangleVector.contourOfUnionOfRectangles *)
		initializeObjects;
{$IFC AnimationEnabled AND AContourOfRectangle }
		if animationFlag[AContourOfRectanglesItem] then
			with currentFrame do
				sweepPosAnim := -dx / mx;
{$ENDC }
		sideQueue.forAll(processSide);
{$IFC AnimationEnabled AND AContourOfRectangle }
		if animationFlag[AContourOfRectanglesItem] then
			with currentFrame do
				sweepPosAnim := -dy / my;
{$ENDC }
		getHorizontalSides;
		reportResults;
		releaseObjects;
	end;

function rectangleVector.boundingBox: rectangle;
	var
		i: seqIndex;
		rect: rectangle;
	begin
		if length > 0 then begin
			rect := rectangle(elements^^[1].duplicate);
			for i := 2 to length do begin
				with rectangle(elements^^[i]) do begin
{$IFC AnimationEnabled }
					if userAbort then begin
						rect.free;
						boundingBox := nil;
						exit(boundingBox);
					end;
{$ENDC }
					if left < rect.left then
						rect.left := left;
					if right > rect.right then
						rect.right := right;
					if top > rect.top then
						rect.top := top;
					if bottom < rect.bottom then
						rect.bottom := bottom
				end
			end;
			boundingBox := rect;
		end
		else
			boundingBox := nil;
	end;

function rectangleVector.memberType (o: obj): boolean;
	begin
		memberType := member(o, rectangleVector) & memberBaseType(o);
	end;

function rectangleVector.equal (o: obj): boolean;
	begin
		equal := (o <> nil) & member(o, rectangleVector) & inherited equal(o);
	end;

function rectangleVector.getType: objectType;
	begin
		getType := rectangleVectorType;
	end;

function rectangleVector.getTextType: Str255;
	begin
		getTextType := stringOf('Rectangle Vector(', length : 1, ')');
	end;

end.
