(*	XYZ GeoBench (eXperimental geometrY Zurich):		*)
(*	An object oriented, interactive workbench for		*)
(*	experimental geometric computation.					*)
(*														*)
(*	 1999 by Peter Schorn, peter.schorn@acm.org		*)
(*	Culmannstrasse 77, CH-8006 Zrich, Switzerland		*)
(*	This software may be copied freely for educational	*)
(*	and research purposes only. For commercial use,		*)
(*	please contact the author.							*)
(*														*)
(*	Version 5.0.5: 30-Sep-1999							*)

unit rectangleVector;

interface

uses
	regularNGon;

type
	rectangleVector = object(homogenVector)
			function equal (o: obj): boolean;
			override;
			procedure describe;
			override;
			procedure execute (item: integer);
			override;
			function memberType (o: obj): boolean;
			override;
			function getType: objectType;
			override;
			function getTextType: Str255;
			override;
			function getBoundingObject: obj;
			override;

			function boundingBox: rectangle;
(* Computes the smallest rectangle that contains all the given rectangles. 	*)

			function contourOfUnionOfRectangles: lineSegmentVector;
(* Computes a set of horizontal and vertical line segments that determines	*)
(* the contour of the given rectangles using a plane sweep algorithm. 			*)

		end;

function newRectangleVector (length: longInt): rectangleVector;

procedure initRectangleVector;


implementation

const
	BoundingBoxC = 1;
	ContRectC = 2;
	RandomPointsC = 3;
	RandomLineSegmentsC = 4;
	RandomRectanglesC = 5;
	RandomCirclesC = 6;
	RandomNGonsC = 7;
	RandomDDimPointsC = 8;
	RandomPolygonsC = 9;
	RandomConvexPolygonsC = 10;
	RandomPolyLinesC = 11;
	RandomConstrainedRectanglesC = 12;
	ContRectDynamicC = 13;
	DirectedGraphC = ContRectDynamicC + 1;
	RandomGraphC = DirectedGraphC + 1;
	EmptyGraphC = RandomGraphC + 1;
	EmptyDirectedGraphC = EmptyGraphC + 1;
	FullGraphC = EmptyDirectedGraphC + 1;
	FullDirectedGraphC = FullGraphC + 1;
{$IFC OptionalCommand }
	MakeRandomSpanningTreeC = FullDirectedGraphC + 1;
	PreviousCommand = MakeRandomSpanningTreeC;
{$ELSEC }
	PreviousCommand = FullDirectedGraphC;
{$ENDC }

function vectorOk (v: vector): boolean;
	begin
		vectorOk := currentArguments1IsHomogenVector & (member(v.ith(1), rectangle) | member(v.ith(1), circle)) | (v.length > 0) & v.memberTypeSet([circleType, rectangleType, regularNGonType]);
	end;

procedure rectangleVector.describe;

	function isRectangleVector (v: vector): boolean;
		begin
			isRectangleVector := (v.length > 1) & (member(v, rectangleVector) | member(v.elements^^[1], rectangle));
		end;

	begin
		with currentArguments do begin
			if (length = 1) & currentArguments1IsHomogenVector & isRectangleVector(vector(ith(1))) | (length > 1) & currentArgumentsSameType & member(first^^.value, rectangle) then begin
				menuEntry('Bounding Box', BoundingBoxC);
				menuEntry('Contour Of Rectangles', ContRectC);
				menuEntry('Contour Of Rectangles Dynamic', ContRectDynamicC);
			end;
			if (length > 0) & (currentArgumentsSameType & (member(ith(1), rectangle) | member(ith(1), circle)) | memberTypeSet([circleType, rectangleType, regularNGonType]) | (length = 1) & member(ith(1), vector) & vectorOk(vector(ith(1)))) then begin
				beginSubMenu('Random Create');
				menuEntry('Points...', RandomPointsC);
				menuEntry('Line Segments...', RandomLineSegmentsC);
				menuEntry('Poly Lines...', RandomPolyLinesC);
				menuEntry('Polygons...', RandomPolygonsC);
				menuEntry('Convex Polygons...', RandomConvexPolygonsC);
				menuEntry('Rectangles...', RandomRectanglesC);
				menuEntry('Constrained Rectangles...', RandomConstrainedRectanglesC);
				menuEntry('Circles...', RandomCirclesC);
				menuEntry('N-Gons...', RandomNGonsC);
				menuEntry('D-Dim Points...', RandomDDimPointsC);
				endSubMenu;
			end;
			if length > 0 then begin
				beginSubMenu('Create Graphs');
				menuEntry('Directed Graph', DirectedGraphC);
				menuEntry('Random Graph...', RandomGraphC);
				menuEntry('Empty Graph', EmptyGraphC);
				menuEntry('Empty Directed Graph', EmptyDirectedGraphC);
				menuEntry('Full Graph', FullGraphC);
				menuEntry('Full Directed Graph', FullDirectedGraphC);
{$IFC OptionalCommand }
				menuEntry('Random Spanning Tree', MakeRandomSpanningTreeC);
{$ENDC }
				endSubMenu;
			end;
		end;
	end;

	var
		sLength: seqIndex0;
		actualWidthPercentage, actualHeightPercentage, lowNumberOfVertices, highNumberOfVertices: extended;

function numberOfObjectsOk (x: extended): boolean;
	begin
		numberOfObjectsOk := (1 <= x) & (x * sLength <= 40000) & (x = trunc(x));
	end;

function numberOfVerticesOk (x: extended): boolean;
	begin
		numberOfVerticesOk := (1 <= x) & (x <= 1024) & (x = trunc(x));
	end;

function lowNumberOfVerticesOk (x: extended): boolean;
	begin
		lowNumberOfVerticesOk := (1 <= x) & (x <= 1024) & (x = trunc(x));
	end;

function highNumberOfVerticesOk (x: extended): boolean;
	begin
		highNumberOfVerticesOk := (1 <= x) & (x <= 1024) & (x = trunc(x)) & (lowNumberOfVertices <= x);
	end;

function doPolycreateRandomObjects (s: sequence; t: objectType; n, lowerBound, upperBound: seqIndex): homogenVector;
	var
		result: homogenVector;

	procedure doCreate (x: obj);
		var
			i: seqIndex;
			o: obj;
		begin
			currentRandomConstraint := x;
			for i := 1 to n do begin
				o := newObj(t);
				vector(o).preset(randomRange(lowerBound, upperBound + 1), currentPointType);
				o.randomChange;
				result.append(o);
			end;
		end;

	begin
		result := newHomogenVector(s.length * n);
		s.forAll(doCreate);
		doPolycreateRandomObjects := result;
	end;

function callCreateRandomObjects (s: sequence; t: objectType; numText: Str255; constrained: boolean): homogenVector;
	begin
		if member(s.ith(1), vector) then
			s := vector(s.ith(1));
		sLength := s.length;
		if (t = polygon2dType) | (t = convexPolygonType) | (t = polyLineType) then
			if getUserParameter3(numText, numberOfObjects, numberOfObjects, @numberOfObjectsOk, 'Low Number of vertices:', lowNumberOfVertices, lowNumberOfVertices, @lowNumberOfVerticesOk, 'High Number of vertices:', highNumberOfVertices, highNumberOfVertices, @highNumberOfVerticesOk) then
				if lowNumberOfVertices = highNumberOfVertices then begin
{$IFC THINK_PASCAL }
					currentVectorLength := trunc(lowNumberOfVertices);
					callCreateRandomObjects := s.createRandomObjects(t, trunc(numberOfObjects));
{$ELSEC }
					currentVectorLength := truncLong(lowNumberOfVertices);
					callCreateRandomObjects := s.createRandomObjects(t, truncLong(numberOfObjects));
{$ENDC }
				end
				else
{$IFC THINK_PASCAL }
					callCreateRandomObjects := doPolycreateRandomObjects(s, t, trunc(numberOfObjects), trunc(lowNumberOfVertices), trunc(highNumberOfVertices))
{$ELSEC }
					callCreateRandomObjects := doPolycreateRandomObjects(s, t, truncLong(numberOfObjects), truncLong(lowNumberOfVertices), truncLong(highNumberOfVertices))
{$ENDC }
			else
				callCreateRandomObjects := nil
		else if constrained then
			if getUserParameter3(numText, numberOfObjects, numberOfObjects, @numberOfObjectsOk, 'Width Percentage', actualWidthPercentage, actualWidthPercentage, @percentageOk, 'Height Percentage', actualHeightPercentage, actualHeightPercentage, @percentageOk) then begin
				currentWidthPercentage := actualWidthPercentage / 100;
				currentHeightPercentage := actualHeightPercentage / 100;
{$IFC THINK_PASCAL }
				callCreateRandomObjects := s.createRandomObjects(t, trunc(numberOfObjects));
{$ELSEC }
				callCreateRandomObjects := s.createRandomObjects(t, truncLong(numberOfObjects));
{$ENDC }
				currentWidthPercentage := 0.0;
				currentHeightPercentage := 0.0;
			end
			else
				callCreateRandomObjects := nil
		else if getUserParameter1(numText, numberOfObjects, numberOfObjects, @numberOfObjectsOk) then
{$IFC THINK_PASCAL }
			callCreateRandomObjects := s.createRandomObjects(t, trunc(numberOfObjects))
{$ELSEC }
			callCreateRandomObjects := s.createRandomObjects(t, truncLong(numberOfObjects))
{$ENDC }
		else
			callCreateRandomObjects := nil;
	end;

function identifyObjects: Str255;

	function classify (o: obj): Str255;
		begin
			if member(o, rectangle) then
				classify := ' per rectangle:'
			else if member(o, regularNGon) then
				classify := ' per n-gon:'
			else
				classify := ' per circle:'
		end;

	begin
		with currentArguments do begin
			if currentArguments.length = 1 then
				if member(ith(1), vector) then
					if currentArguments1IsHomogenVector then
						identifyObjects := classify(vector(ith(1)).ith(1))
					else
						identifyObjects := ' per object:'
				else
					identifyObjects := ':'
			else if currentArgumentsSameType then
				identifyObjects := classify(ith(1))
			else
				identifyObjects := ' per object:';
		end;
	end;

{$IFC OptionalCommand }
function makeRandomSpanningTree (v: vector): spanningTree;
	var
		result: spanningTree;
		e: graphEdge;
		i: seqIndex;
	begin
		if v.length > 1 then begin
			new(result);
			failNil(Handle(result));
			result.vertices := vector(v.duplicateAll);
			result.edges := newHomogenVector(v.length - 1);
			for i := 1 to v.length - 1 do begin
				if userAbort then begin
					result.freeAll;
					makeRandomSpanningTree := nil;
					exit(makeRandomSpanningTree);
				end;
				new(e);
				failNil(Handle(e));
				e.startVertex := i;
				e.endVertex := randomRange(i + 1, v.length + 1);
				result.edges.append(e);
			end;
			makeRandomSpanningTree := result;
		end
		else
			makeRandomSpanningTree := nil;
	end;
{$ENDC }

function createRandomGraph (v: vector; isDirectedGraph, isConnectedGraph: boolean; edgesGivenMethod: integer; edgeProbability: extended; numberOfEdges: seqIndex0): simpleUndirectedGraph;
	var
		result: simpleUndirectedGraph;
		dResult: directedGraph;
	begin
		if isDirectedGraph then begin
			new(dResult);
			result := dResult;
		end
		else
			new(result);
		failNil(Handle(result));
		result.vertices := vector(v.duplicateAll);

		if edgesGivenMethod = givenByNumber then begin
			if isDirectedGraph then
				createRandomGraphByEdgeNumber(numberOfEdges, isDirectedGraph, isConnectedGraph, result)
			else
				createRandomGraphByEdgeNumber(numberOfEdges, isDirectedGraph, isConnectedGraph, result);
		end
		else if edgesGivenMethod = givenByProbability then
			createRandomGraphByProbability(edgeProbability, isDirectedGraph, isConnectedGraph, result);
{$IFC AnimationEnabled }
		if userAbort then begin
			createRandomGraph := nil;
			exit(createRandomGraph);
		end;
{$ENDC }
		if isDirectedGraph then
			with directedGraph(result) do begin
				adjacency := newHomogenVector(v.length);
				initAdjacency;
			end;
		createRandomGraph := result;
	end;
	
const
	isDirectedGraphDefault = false;
	edgesGivenMethodDefault = givenByNumber;
	numberOfEdgesDefault = 20;
	edgeProbabilityDefault = 0.5;
	isConnectedGraphDefault = true;
	stdUseFormula = false;
var
	isDirectedGraph, isConnectedGraph, useFormula: boolean;
	edgesGivenMethod: integer;
	numberOfEdges, edgeProbability: extended;
	psl: longInt;

function defaultFormulaVisible: boolean;
	begin
		defaultFormulaVisible := useFormula;
	end;

function isVisibleNumberOfEdges: boolean;
	begin
		isVisibleNumberOfEdges := edgesGivenMethod = givenByNumber;
	end;

function isVisibleProbability: boolean;
	begin
		isVisibleProbability := edgesGivenMethod = givenByProbability;
	end;

function edgeProbabilityOk (x: extended): boolean;
	begin
		edgeProbabilityOk := (0 <= x) & (x <= 1);
	end;

function defaultNumberOfEdgesOk (x: extended): boolean;
	begin
		defaultNumberOfEdgesOk := (x = trunc(x)) & (isConnectedGraph & (x >= psl - 1) | not isConnectedGraph & (x >= 0)) & ((x <= (psl - 1) * psl div 2) | isDirectedGraph & (x <= (psl - 1) * psl));
	end;

{$IFC AnimationEnabled }
procedure showTrials (trials: longInt);
	begin
		displayInfoString(StringOf('Trials = ', trials : 1));
	end;
{$ENDC }

procedure findRandomObject (function generate: obj; function test (var x: obj; var proof: obj): boolean);
	var
{$IFC AnimationEnabled }
		trials: longInt;
{$ENDC }
		x, proof: obj;
	begin
{$IFC AnimationEnabled }
		trials := 0;
{$ENDC }
		repeat
{$IFC AnimationEnabled }
			trials := trials + 1;
			if trials mod 50 = 0 then
				showTrials(trials);
{$ENDC }
			x := generate;
{$IFC AnimationEnabled }
			if userAbort then begin
				showTrials(trials);
				exit(findRandomObject);
			end;
{$ENDC }
			proof := nil;
			if test(x, proof) then begin
				currentResult.appendNonNil(x);
				currentResult.appendNonNil(proof);
{$IFC AnimationEnabled }
				showTrials(trials);
{$ENDC }
				exit(findRandomObject);
			end
			else begin
				if x <> nil then
					x.freeAll;
				if proof <> nil then
					proof.freeAll;
			end;
		until false;
	end;

procedure callCreateRandomGraph (v: vector);
	var
		h: vector;
		maxEdges: longInt;
		errorPosition: integer;
		dummy: boolean;

	function generateGraph: obj;
		var
			r1: simpleUndirectedGraph;
			r2: undirectedGraph;
		begin
{$IFC THINK_PASCAL }
			r1 := createRandomGraph(v, isDirectedGraph, isConnectedGraph, edgesGivenMethod, edgeProbability, trunc(numberOfEdges));
{$ELSEC }
			r1 := createRandomGraph(v, isDirectedGraph, isConnectedGraph, edgesGivenMethod, edgeProbability, truncLong(numberOfEdges));
{$ENDC }
			if r1 = nil then
				generateGraph := nil
			else if isDirectedGraph then
				generateGraph := r1
			else begin
				new(r2);
				failNil(Handle(r2));
				r2.initializeFromSimpleUndirectedGraph(r1);
				r1.freeAll;
				generateGraph := r2;
			end;
		end;

{$IFC NOT THINK_PASCAL }
{$PUSH }
{$PRAGMAC warn_unusedarg off }
{$ENDC }
	function testGraph (var g: directedGraph; var proof: homogenVector): boolean;
		var
			errorPosition: integer;
		begin
			testGraph := isGraphOk(g, defaultFormula, true, errorPosition);
		end;
{$IFC NOT THINK_PASCAL }
{$POP }
{$ENDC }

	begin
		psl := v.length;
		maxEdges := (psl - 1) * psl;
		if not isDirectedGraph then
			maxEdges := maxEdges div 2;
		if numberOfEdges > maxEdges then
			numberOfEdges := maxEdges;

		h := newVector(9);
		h.append(getBooleanHandler(7, @isDirectedGraph, isDirectedGraphDefault, nil));
		h.append(getBooleanHandler(13, @isConnectedGraph, isConnectedGraphDefault, nil));
		h.append(getEnumerationHandler(4, @edgesGivenMethod, edgesGivenMethodDefault, nil, givenByProbability));
		h.append(getExtendedHandler(8, @edgeProbability, edgeProbabilityDefault, @isVisibleProbability, @edgeProbabilityOk));
		h.append(getExtendedHandler(9, @numberOfEdges, numberOfEdgesDefault, @isVisibleNumberOfEdges, @defaultNumberOfEdgesOk));
		h.append(getStaticTextHandler(10, '', false, @isVisibleNumberOfEdges));
		h.append(getStaticTextHandler(11, '', false, @isVisibleProbability));
		h.append(getBooleanHandler(14, @useFormula, stdUseFormula, nil));
		h.append(getStringHandler(15, @defaultFormula, stdDefaultFormula, @defaultFormulaVisible));

		if processDialogBox(130, 3, 12, 0, alwaysTrue, h) then
			if useFormula then begin
				dummy := isGraphOk(nil, defaultFormula, false, errorPosition);
				if not showErrorInDefaultFormula(errorPosition) then
					findRandomObject(generateGraph, testGraph);
			end
			else
{$IFC THINK_PASCAL }
				currentResult.appendNonNil(createRandomGraph(v, isDirectedGraph, isConnectedGraph, edgesGivenMethod, edgeProbability, trunc(numberOfEdges)));
{$ELSEC }
				currentResult.appendNonNil(createRandomGraph(v, isDirectedGraph, isConnectedGraph, edgesGivenMethod, edgeProbability, truncLong(numberOfEdges)));
{$ENDC }
	end;
	
function computeContourOfUnionOfRectangles (rs: rectangleVector): obj;
	begin
		computeContourOfUnionOfRectangles := rs.contourOfUnionOfRectangles;
	end;

procedure rectangleVector.execute (item: integer);
	var
		rs: rectangleVector;
		s: vector;
	begin
		case item of
			RandomPointsC: 
				currentResult.appendNonNil(callCreateRandomObjects(currentArguments, currentPointType, concat('Number of points', identifyObjects), false));
			RandomLineSegmentsC: 
				currentResult.appendNonNil(callCreateRandomObjects(currentArguments, lineSegmentType, concat('No. of line segments', identifyObjects), false));
			RandomRectanglesC: 
				currentResult.appendNonNil(callCreateRandomObjects(currentArguments, rectangleType, concat('No. of rectangles', identifyObjects), false));
			RandomCirclesC: 
				currentResult.appendNonNil(callCreateRandomObjects(currentArguments, circleType, concat('Number of circles', identifyObjects), false));
			RandomNGonsC: 
				currentResult.appendNonNil(callCreateRandomObjects(currentArguments, regularNGonType, concat('Number of n-gons', identifyObjects), false));
			RandomDDimPointsC: 
				currentResult.appendNonNil(callCreateRandomObjects(currentArguments, dDimPointType, concat('No. of d-dim points', identifyObjects), false));
			RandomPolygonsC: 
				currentResult.appendNonNil(callCreateRandomObjects(currentArguments, polygon2dType, concat('Number of  polygons', identifyObjects), false));
			RandomConvexPolygonsC: 
				currentResult.appendNonNil(callCreateRandomObjects(currentArguments, convexPolygonType, concat('No. of  conv. polygs.', identifyObjects), false));
			RandomPolyLinesC: 
				currentResult.appendNonNil(callCreateRandomObjects(currentArguments, polyLineType, concat('Number of  polylines', identifyObjects), false));
			RandomConstrainedRectanglesC: 
				currentResult.appendNonNil(callCreateRandomObjects(currentArguments, rectangleType, concat('No. of rectangles', identifyObjects), true));
			DirectedGraphC:
				begin
					new(s);
					failNil(Handle(s));
					s.initFromList(currentArguments);
					currentResult.append(s.createGraph(directedGraphType));
				end;
			RandomGraphC:
				begin
					new(s);
					failNil(Handle(s));
					s.initFromList(currentArguments);
					callCreateRandomGraph(s);
				end;
			EmptyGraphC, EmptyDirectedGraphC:
				begin 
					new(s);
					failNil(Handle(s));
					s.initFromList(currentArguments);
					currentResult.appendNonNil(createRandomGraph(s, item = EmptyDirectedGraphC, false, givenByNumber, 0, 0));
				end;
			FullGraphC:
				begin
					new(s);
					failNil(Handle(s));
					s.initFromList(currentArguments);
					currentResult.append(createRandomGraph(s, false, false, givenByNumber, 0, s.length * (s.length - 1) div 2));
				end;
			FullDirectedGraphC:
				begin
					new(s);
					failNil(Handle(s));
					s.initFromList(currentArguments);
					currentResult.append(createRandomGraph(s, true, false, givenByNumber, 0, s.length * (s.length - 1)));
				end;
{$IFC OptionalCommand }
		MakeRandomSpanningTreeC:
				begin
					new(s);
					failNil(Handle(s));
					s.initFromList(currentArguments);
					currentResult.appendNonNil(makeRandomSpanningTree(s));
				end;
{$ENDC }
			otherwise begin
				if currentArguments.length > 1 then begin
					new(rs);
					failNil(Handle(rs));
					rs.initFromList(currentArguments);
				end
				else if member(currentArguments.ith(1), rectangleVector) then
					rs := rectangleVector(currentArguments.ith(1))
				else begin
					new(rs);
					failNil(Handle(rs));
					s := vector(currentArguments.ith(1));
					rs.length := s.length;
					rs.elements := s.elements;
				end;
				case item of
					BoundingBoxC: 
						currentResult.appendNonNil(rs.boundingBox);
					ContRectC: 
						currentResult.appendNonNil(rs.contourOfUnionOfRectangles);
					ContRectDynamicC: 
						rs.dynamicExecutionHighlightShowType(computeContourOfUnionOfRectangles);
				end;
				if currentArguments.length > 1 then
					rs.free
				else if not member(currentArguments.ith(1), rectangleVector) then
					dispose(rs);
			end;
		end;
	end;

function newRectangleVector (length: longInt): rectangleVector;
	var
		v: rectangleVector;
	begin
		currentVectorLength := length;
		new(v);
		failNil(Handle(v));
		v.init;
		newRectangleVector := v;
	end;

procedure initRectangleVector;
	var
		x: rectangleVector;
	begin
		new(x);
		failNil(Handle(x));
		prototypes[rectangleVectorType] := x;
		actualWidthPercentage := 20.0;
		actualHeightPercentage := 10.0;
		lowNumberOfVertices := 7;
		highNumberOfVertices := 9;
		isDirectedGraph := isDirectedGraphDefault;
		isConnectedGraph := isConnectedGraphDefault;
		edgesGivenMethod := edgesGivenMethodDefault;
		numberOfEdges := numberOfEdgesDefault;
		edgeProbability := edgeProbabilityDefault;
		useFormula := stdUseFormula;
	end;


end.
