(*	XYZ GeoBench (eXperimental geometrY Zurich):		*)
(*	An object oriented, interactive workbench for		*)
(*	experimental geometric computation.					*)
(*														*)
(*	 1999 by Peter Schorn, peter.schorn@acm.org		*)
(*	Culmannstrasse 77, CH-8006 Zrich, Switzerland		*)
(*	This software may be copied freely for educational	*)
(*	and research purposes only. For commercial use,		*)
(*	please contact the author.							*)
(*														*)
(*	Version 5.0.5: 30-Sep-1999							*)

unit regularNGon;

interface

uses
	pointVector;

type
	regularNGon = object(circle)
			vertices: integer;	(* Number of vertices *)
			angle: real;			(* Angle by which the n-gon is tilted *)
			alignedNGon: boolean;
			procedure init;
			override;
			procedure describe;
			override;
			procedure execute (item: integer);
			override;
			procedure randomChange;
			override;
	(* 'alignedNGon' is checked. If true, aligned n-gons are created, otherwise the angle is changed as well. *)
			function isSelected (where: obj): boolean;
			override;
			procedure construct (px, py, newX, newY: integer);
			override;
	(*	'alignedNGon' is checked. If true, the angle is takem from 'currentRegularNGonAngle', otherwise the	*)
	(*	user enters the angle via dragging. *)
			procedure showObjectParameter;
			override;
			procedure interactiveInput (px, py: integer);
			override;
			procedure interactiveOutput;
			override;
			procedure windowOutput;
			override;
			procedure textFileInput;
			override;
			procedure textFileOutput;
			override;
			procedure fileInput;
			override;
			procedure fileOutput;
			override;
			procedure rotate (xOrigin, yOrigin, alpha: extended);
			override;
			function area: extended;
			override;
			function perimeter: extended;
			override;
			function equal (o: obj): boolean;
			override;
			function memberType (o: obj): boolean;
			override;
			function getType: objectType;
			override;
			function getTextType: Str255;
			override;
			function genericLessThan (o: obj): boolean;
			override;
			function pointCover (n: longInt): obj;
			override;
			function getInfo: obj;
			override;
			procedure setInfo (o: obj);
			override;
			function convertToConvexPolygon: obj;
			override;
			function convertToConvexPolygonPossible: boolean;
			override;
			function randomPointCover (n: longInt): obj;
			override;
			function drawingComplexity (cutOff: longInt): longInt;
			override;
			function dragVertex (where: obj; mouse: Point; highlight: boolean): boolean;
			override;
			procedure constructWorld (mouse: Point);
			override;

{$IFC GFEnabled }
			function canUseGridfile: BOOLEAN;
			override;
{$ENDC}

			function inside (p: point2d): boolean;
			override;
	(* Determines whether a given point 'p' lies inside the regular n-gon. *)

			function distanceToPerimeterX (p: point2d; var distance: extended): boolean;
			override;

			procedure setRadiusAngle (dx, dy: extended);

		end;

function argument (x, y: extended): extended;
	(* Computes angle of polarcoordinate representation *)

function normalize (alpha: real): real;
	(* Normalizes an angle to the range 0 .. 2* *)

procedure initRegularNGon;

implementation

var
	deg90, deg270, deg360: extended;

procedure regularNGon.interactiveInput (px, py: integer);
	begin
		init;
		inherited interactiveInput(px, py);
	end;

function regularNGon.dragVertex (where: obj; mouse: Point; highlight: boolean): boolean;
	var
		alpha, cosAlpha, sinAlpha, cosSum, newCosSum, sinSum: extended;
		i: integer;

	function drag (px, py: extended): boolean;
		begin
			if (rectangle(where).left <= px) & (px <= rectangle(where).right) & (rectangle(where).bottom <= py) & (py <= rectangle(where).top) then begin
	(*	Cannot use 'with rectangle(where)' here because of conflicts! *)
				drag := true;
				dragVertexWorld(mouse, highlight);
			end
			else
				drag := false;
		end;

	begin (* regularNGon.dragVertex *)
		alpha := deg360 / vertices;
		cosAlpha := cos(alpha);
		sinAlpha := sin(alpha);
		cosSum := cos(angle);
		sinSum := sin(angle);
		for i := 1 to vertices do begin
			if drag(x + radius * cosSum, y + radius * sinSum) then begin
				dragVertex := true;
				exit(dragVertex);
			end;
			newCosSum := cosAlpha * cosSum - sinAlpha * sinSum;
			sinSum := sinAlpha * cosSum + cosAlpha * sinSum;
			cosSum := newCosSum;
		end;
		dragVertex := false;
	end;

{$IFC NOT THINK_PASCAL }
{$PUSH }
{$PRAGMAC warn_unusedarg off }
{$ENDC }
function regularNGon.drawingComplexity (cutOff: longInt): longInt;
	begin
		drawingComplexity := 2 * vertices;
	end;
{$IFC NOT THINK_PASCAL }
{$POP }
{$ENDC }

function regularNGon.area: extended;
	begin
		area := vertices * sqr(radius) * sin(deg360 / vertices) / 2.0;
	end;

function regularNGon.perimeter: extended;
	begin
		perimeter := 2 * vertices * radius * sin(pi / vertices);
	end;

function argument (x, y: extended): extended;
	var
		d: extended;
	begin
		if x = 0 then
			if y >= 0 then
				argument := deg90
			else
				argument := deg270
		else begin
			d := arctan(y / x);
			if x < 0 then
				argument := d + pi
			else if y < 0 then
				argument := d + deg360
			else
				argument := d;
		end;
	end;

function normalize (alpha: real): real;
	begin
		if alpha >= 0 then
			normalize := alpha - deg360 * trunc(alpha / deg360)
		else
			normalize := alpha - deg360 * (trunc(alpha / deg360) - 1);
	end;

procedure regularNGon.init;
	begin
		vertices := currentVertices;
		alignedNGon := currentAlignmentIsOn;
		if alignedNGon then
			angle := normalize(currentRegularNGonAngle)
		else
			angle := 0;
	end;

procedure regularNGon.randomChange;
	begin
		inherited randomChange;
		if not alignedNGon then
			angle := 2 * pi * randomReal;
	end;

function regularNGon.convertToConvexPolygon: obj;
	var
		v: vector;
		p: convexPolygon;
		safe: objectType;
	begin
		safe := currentPointType;
		currentPointType := realPointType;
		v := vector(pointCoverAngle(vertices, angle));
		currentPointType := safe;
		new(p);
		failNil(Handle(p));
		p.elements := v.elements;
		p.length := v.length;
		dispose(v);
		convertToConvexPolygon := p;
	end;

function regularNGon.convertToConvexPolygonPossible: boolean;
	begin
		convertToConvexPolygonPossible := true;
	end;

procedure appendRealPoint (v: vector; x, y: real);
	var
		p: realPoint;
	begin
		new(p);
		p.x := x;
		p.y := y;
		v.append(p);
	end;

function rectangle.convertToConvexPolygon: obj;
	var
		p: convexPolygon;
	begin
		currentVectorLength := 4;
		new(p);
		failNil(Handle(p));
		p.init;
		appendRealPoint(p, left, bottom);
		appendRealPoint(p, right, bottom);
		appendRealPoint(p, right, top);
		appendRealPoint(p, left, top);
		convertToConvexPolygon := p;
	end;

function rectangle.convertToConvexPolygonPossible: boolean;
	begin
		convertToConvexPolygonPossible := true;
	end;

function point2d.convertToConvexPolygon: obj;
	var
		p: convexPolygon;
	begin
		currentVectorLength := 1;
		new(p);
		failNil(Handle(p));
		p.init;
		p.append(duplicate);
		convertToConvexPolygon := p;
	end;

function point2d.convertToConvexPolygonPossible: boolean;
	begin
		convertToConvexPolygonPossible := true;
	end;

function lineSegment.convertToConvexPolygon: obj;
	var
		poly: convexPolygon;
	begin
		currentVectorLength := 2;
		new(poly);
		failNil(Handle(poly));
		poly.init;
		poly.append(p.duplicate);
		poly.append(q.duplicate);
		convertToConvexPolygon := poly;
	end;

function lineSegment.convertToConvexPolygonPossible: boolean;
	begin
		convertToConvexPolygonPossible := true;
	end;

function regularNGon.pointCover (n: longInt): obj;
	var
		p: convexPolygon;
	begin
		p := convexPolygon(convertToConvexPolygon);
		pointCover := p.pointCover(n);
		p.freeAll;
	end;

function regularNGon.randomPointCover (n: longInt): obj;
	var
		p: convexPolygon;
	begin
		p := convexPolygon(convertToConvexPolygon);
		randomPointCover := p.randomPointCover(n);
		p.freeAll;
	end;

function regularNGon.genericLessThan (o: obj): boolean;
	var
		c: regularNGon;
	begin
		c := regularNGon(o);
		genericLessThan := (x < c.x) | (x = c.x) & ((y < c.y) | (y = c.y) & ((radius < c.radius) | (radius = c.radius) & ((vertices < c.vertices) | (vertices = c.vertices) & (angle < c.angle))));
	end;

function regularNGon.isSelected (where: obj): boolean;
	var
		p: convexPolygon;
	begin
		if vertices <= 20 then begin
			p := convexPolygon(convertToConvexPolygon);
			isSelected := p.isSelected(where);
			p.freeAll;
		end
		else
			isSelected := inherited isSelected(where);
	end;

procedure regularNGon.windowOutput;
	var
		p: convexPolygon;
	begin
		p := convexPolygon(convertToConvexPolygon);
		p.windowOutput;
		p.freeAll;
	end;

procedure regularNGon.interactiveOutput;
	var
		p: convexPolygon;
	begin
		if radius > 0 then begin
			p := convexPolygon(convertToConvexPolygon);
			p.interactiveOutput;
			p.freeAll;
		end
		else
			inherited interactiveOutput;
	end;

procedure regularNGon.textFileInput;
	var
		t: integer;
	begin
		inherited textFileInput;
		readTextInteger(vertices);
		readTextReal(angle);
		readTextInteger(t);
		alignedNGon := t = 1;
		angle := normalize(angle);
	end;

procedure regularNGon.textFileOutput;
	var
		p: convexPolygon;
	begin
		case currentTextFileOutputMethod of
			0:  begin
				inherited textFileOutput;
				writeTextInteger(vertices);
				writeTextReal(angle);
				writeTextInteger(ord(alignedNGon));
			end;
			otherwise begin
				p := convexPolygon(convertToConvexPolygon);
				p.textFileOutput;
				p.freeAll;
			end;
		end;
	end;

procedure regularNGon.fileInput;
	var
		b: Byte;
	begin
		inherited fileInput;
		readInteger(vertices);
		readReal(angle);
		angle := normalize(angle);
		if currentVersion >= version9 then begin
			readByte(b);
			alignedNGon := b = 1;
		end
		else
			alignedNGon := false;
	end;

procedure regularNGon.fileOutput;
	begin
		inherited fileOutput;
		writeInteger(vertices);
		writeReal(angle);
		writeByte(ord(alignedNGon));
	end;

procedure regularNGon.rotate (xOrigin, yOrigin, alpha: extended);
	begin
		inherited rotate(xOrigin, yOrigin, alpha);
		angle := normalize(angle + alpha);
	end;

function regularNGon.equal (o: obj): boolean;
	begin
		equal := inherited equal(o) & member(o, regularNGon) & (regularNGon(o).vertices = vertices) & (regularNGon(o).angle = angle);
	end;

function regularNGon.memberType (o: obj): boolean;
	begin
		memberType := member(o, regularNGon);
	end;

function regularNGon.getType: objectType;
	begin
		getType := regularNGonType;
	end;

function regularNGon.getTextType: Str255;
	begin
		if angle = 0.0 then
			getTextType := StringOf('N-Gon(', vertices : 1, ')')
		else
			getTextType := StringOf('N-Gon(', vertices : 1, ',', angle / oneDegree : 4 : 1, ')');
	end;

{$IFC GFEnabled }
function regularNGon.canUseGridfile: BOOLEAN;
	begin
		canUseGridfile := false;
	end; { canUseGridfile }
{$ENDC}

	const
		InsideC = 0;
		AlignToFirstC = 1;

procedure regularNGon.describe;
	begin
		with currentArguments do begin
			if (length = 2) & (member(ith(1), regularNGon) & member(ith(2), point2d) | member(ith(2), regularNGon) & member(ith(1), point2d)) then
				menuEntry('Inside', InsideC);
			if currentArgumentsSameType & (length > 1) & member(ith(1), regularNGon) then
				menuEntry('Align To First', AlignToFirstC);
		end;
	end;

procedure regularNGon.execute (item: integer);

	function callInside (r: regularNGon; p: point2d): markedRealPoint;
		var
			rp: markedRealPoint;
		begin
			new(rp);
			failNil(Handle(rp));
			rp.init;
			rp.x := p.getX;
			rp.y := p.getY;
			rp.mark.value := StringOf(r.inside(p));
			callInside := rp;
		end;

	procedure align (l: list);
		var
			a: real;

		procedure setAngle (o: regularNGon);
			begin
				o.angle := a;
			end;

		begin
			a := regularNGon(l.ith(1)).angle;
			l.forAll(setAngle);
		end;

	begin
		with currentArguments do
			case item of
				InsideC: 
					if member(ith(1), regularNGon) then
						currentResult.append(callInside(regularNGon(ith(1)), point2d(ith(2))))
					else
						currentResult.append(callInside(regularNGon(ith(2)), point2d(ith(1))));
				AlignToFirstC:  begin
					interactiveOutput;
					interactiveHighlight;
					align(currentArguments);
					interactiveOutput;
					interactiveHighlight;
				end;
			end;
	end;

function regularNGon.getInfo: obj;
	var
		v: vector;
	begin
		v := vector(inherited getInfo);
		appendObjInfoExtended(v, 'Vertices', vertices, false);
		appendObjInfoExtended(v, 'Angle', angle, false);
		appendObjInfoString(v, 'Aligned', StringOf(alignedNGon), false);
		getInfo := v;
	end;

procedure regularNGon.setInfo (o: obj);
	var
		v: vector;
		newVertices: extended;
		newAlignedNGonString: Str255;
	begin
		inherited setInfo(o);
		v := vector(o);
		newVertices := objInfoExtended(v.elements^^[4]).e;
		if (trunc(newVertices) = newVertices) & (newVertices > 2) then
{$IFC THINK_PASCAL }
			vertices := trunc(newVertices);
{$ELSEC }
			vertices := truncLong(newVertices);
{$ENDC }
		angle := normalize(objInfoExtended(v.elements^^[5]).e);
		newAlignedNGonString := capString(objInfoString(v.elements^^[6]).s);
		if (newAlignedNGonString = 'F') | (newAlignedNGonString = 'FALSE') then
			alignedNGon := false
		else if (newAlignedNGonString = 'T') | (newAlignedNGonString = 'TRUE') then
			alignedNGon := true;
	end;

function regularNGon.distanceToPerimeterX (p: point2d; var distance: extended): boolean;
	var
		px, py, alpha: extended;
		q, r: point2d;
		i: integer;
	begin
		q := point2d(p.duplicate);
		r := point2d(p.duplicate);
{$IFC THINK_PASCAL }
		i := trunc(vertices * normalize(argument(p.getX - x, p.getY - y) - angle) / deg360);
{$ELSEC }
		i := truncLong(vertices * normalize(argument(p.getX - x, p.getY - y) - angle) / deg360);
{$ENDC }
		alpha := angle + deg360 * i / vertices;
		q.setX(radius * cos(alpha) + x);
		q.setY(radius * sin(alpha) + y);
		alpha := angle + deg360 * (i + 1) / vertices;
		r.setX(radius * cos(alpha) + x);
		r.setY(radius * sin(alpha) + y);
		distanceToPerimeterX := p.whichSideX(q, r) < 0; (* inside if left to the appropriate edge of the n-gon *)
		distance := p.distanceToLineSegmentX(q, r, px, py);
		r.free;
		q.free;
	end;

procedure regularNGon.setRadiusAngle (dx, dy: extended);
	var
		alpha, beta, cosAlpha, sinAlpha, cosBeta, sinBeta, argDxDy: extended;
		i: integer;
	begin
		argDxDy := argument(dx, dy);
		if alignedNGon then begin
			if (dx = 0) & (dy = 0) then
				radius := 0.0
			else begin
{$IFC THINK_PASCAL }
				i := trunc(vertices * normalize(argDxDy - angle) / deg360);
{$ELSEC }
				i := truncLong(vertices * normalize(argDxDy - angle) / deg360);
{$ENDC }
				alpha := angle + deg360 * i / vertices;
				beta := angle + deg360 * (i + 1) / vertices;
				cosAlpha := cos(alpha);
				sinAlpha := sin(alpha);
				cosBeta := cos(beta);
				sinBeta := sin(beta);
				radius := (dx * (sinAlpha - sinBeta) + dy * (cosBeta - cosAlpha)) / (sinAlpha * cosBeta - sinBeta * cosAlpha);
			end;
		end
		else begin
			radius := sqrt(sqr(dx) + sqr(dy));
			angle := argDxDy;
		end;
	end;

procedure regularNGon.construct (px, py, newX, newY: integer);
	begin
		x := px;
		y := py;
		setRadiusAngle(newX - px, newY - py);
	end;

procedure regularNGon.showObjectParameter;
	begin
		displayInfoString(StringOf('Radius = ', radius : 8 : 3, '. Angle = ', angle : 4 : 5, '.'));
	end;

procedure regularNGon.constructWorld (mouse: Point);
	var
		newX, newY: extended;
	begin
		windToWorld(mouse.h, mouse.v, newX, newY);
		setRadiusAngle(newX - x, newY - y);
	end;

function regularNGon.inside (p: point2d): boolean;
	var
		d: extended;
	begin
		d := sqr(p.getX - x) + sqr(p.getY - y);
		if d > sqr(radius) then (* outside of circle -> outside of n-gon *)
			inside := false
		else if d < sqr(radius * cos(pi / vertices)) then (* in inscribed circle -> in n-gon *)
			inside := true
		else
			inside := distanceToPerimeterX(p, d);
	end;

procedure initRegularNGon;
	var
		r: regularNGon;
	begin
		new(r);
		failNil(Handle(r));
		prototypes[regularNGonType] := r;
		deg90 := pi / 2;
		deg270 := 3.0 * pi / 2.0;
		deg360 := 2 * pi;
	end;

end.
