(*	XYZ GeoBench (eXperimental geometrY Zurich):		*)
(*	An object oriented, interactive workbench for		*)
(*	experimental geometric computation.					*)
(*														*)
(*	 1999 by Peter Schorn, peter.schorn@acm.org		*)
(*	Culmannstrasse 77, CH-8006 Zrich, Switzerland		*)
(*	This software may be copied freely for educational	*)
(*	and research purposes only. For commercial use,		*)
(*	please contact the author.							*)
(*														*)
(*	Version 5.0.5: 30-Sep-1999							*)

(*	Implementation of 'shortestPath' by					*)
(*	Peter Bodenmann.									*)

unit shortestPath;

interface

uses
	pointVector;

implementation

{$SETC AShortestPath = TRUE }

type
	bs = object(obj)
			i, j: integer;
			a, b, sa, ca, mx, my: real;
		end;
	qType = object(obj)
			x: real;
			id: 1..3;
		end;
	seg = object(qType)
			i: integer;
		end;
	spmPoint = object(qType)
			y: real;
			b1, b2: bs;
		end;
	endPoint = object(qType)
			y: real;
		end;
	statusType = object(obj)
			b: bs;
			p: shortestPathTreeType;
		end;
	shortestPathTreeType = object(obj)
			index: integer;
			father: shortestPathTreeType;
		end;

function pointVector.shortestPath (pStart, pEnd: point2d): polyLine;

(*	Computes the shortest path from 'pStart' to 'pEnd'. The points in 'self' define vertical obstacles		*)
(*	with the definition that 'self.elements^^[2*i-1]' is the lower end-point and 'self.elements^^[2*i]'	*)
(*	is the upper end-point of the i-th vertical line segment (1  i  'self.length / 2'. The result is the		*)
(*	shortest path connecting the points without hitting the obstacles.											*)

	var
		T: avlTree;
		Q: heapPriorityQueue;
		SPT, spp: shortestPathTreeType;
		event: qType;
		s: seg;
		p1, p2: point2d;
		ep: endPoint;
		x, x1, x2: statusType;
		b, b1, b2: bs;
		i, j, k, h, n: Integer;
		r, r1, r2: reference;
		f, g: boolean;
		d: directionType;
		w: array[0..1000] of real;
		result: polyLine;
		l, freeList: list;
		elem: elementH;

	function evaluateBisector (b: bs; x: real): real;
		var
			c, d, t: real;
		begin
			c := sqr(b.b) * sqr(b.ca) - sqr(b.a) * sqr(b.sa);
			if b.b <> 0 then begin
				t := c + sqr(x - b.mx);
				if t < 0 then
					t := 0.0;
				d := b.a * b.b * sqrt(t);
			end
			else
				d := 0;
			evaluateBisector := b.my + (-(sqr(b.a) + sqr(b.b)) * b.ca * b.sa * (x - b.mx) + d) / c;
		end;

	function compT (x1, x2: statusType): relationObj;
		var
			y1, y2: real;
		begin
			y1 := evaluateBisector(x1.b, event.x);
			y2 := evaluateBisector(x2.b, event.x);
			if y1 < y2 then
				compT := lessThanObj
			else if y1 > y2 then
				compT := greaterThanObj
			else
				compT := equalObj
		end;

	function compQ (e1, e2: qType): relationObj;
		begin
			if e1.x < e2.x then
				compQ := lessThanObj
			else if e1.x > e2.x then
				compQ := greaterThanObj
			else
				compQ := equalObj
		end;

	procedure insertQ (b2, b1: bs);
		var
			spm: spmPoint;
			a, b, c, y1, y2: real;
			r: reference;
		begin
			b := pEnd.getX;
			if evaluateBisector(b1, b) < evaluateBisector(b2, b) then begin
				new(spm);
				failNil(Handle(spm));
				spm.init;
				spm.id := 2;
				spm.b1 := b1;
				spm.b2 := b2;
				a := event.x;
				repeat
					c := (a + b) / 2;
					y1 := evaluateBisector(b1, c);
					y2 := evaluateBisector(b2, c);
					if y1 > y2 then
						a := c
					else
						b := c;
(*    until abs((y1 - y2) / (1 + abs(y1) + abs(y2))) < 1e-4; *)
				until (abs((y1 - y2) / (1 + abs(y1) + abs(y2))) < 1e-4) | (abs((a - b) / (1 + abs(a) + abs(b))) < 1e-4);
				spm.x := a;
				spm.y := evaluateBisector(b1, c);
				r := Q.insert(spm, compQ);
			end;
		end;

	procedure plotBisector (b: bs);
		var
			x: real;
			l: lineSegment;
		begin
			new(l);
			failNil(Handle(l));
			l.init;
			x := round(event.x / 5) * 5;
			l.q.setX(x);
			l.q.setY(evaluateBisector(b, x));
			while x < pEnd.getX do begin
				x := x + 5;
				if x > pEnd.getX then
					x := pEnd.getX;
				l.p.setX(x);
				l.p.setY(evaluateBisector(b, x));
				l.interactiveOutput;
				l.q.setX(l.p.getX);
				l.q.setY(l.p.getY)
			end;
			l.free;
		end;

	function createBisector (i, j: integer): bs;
		var
			b: bs;
			p, q: point2d;
			c: real;
		begin
			new(b);
			failNil(Handle(b));
			b.init;
			if w[i] > w[j] then begin
				b.i := i;
				b.j := j
			end
			else begin
				b.i := j;
				b.j := i
			end;
			if b.i = 0 then
				p := pStart
			else
				p := point2d(self.elements^^[b.i]);
			if b.j = 0 then
				q := pStart
			else
				q := point2d(self.elements^^[b.j]);
			c := p.distanceX(q) / 2;
			b.a := (w[b.i] - w[b.j]) / 2;
			if abs((c - b.a) / (1 + c + b.a)) < 1e-4 then
				b.b := 0
			else
				b.b := sqrt(sqr(c) - sqr(b.a));
			b.sa := (p.getX - q.getX) / 2 / c;
			b.ca := (p.getY - q.getY) / 2 / c;
			b.mx := (p.getX + q.getX) / 2;
			b.my := (p.getY + q.getY) / 2;
			if b.ca < 0 then
				b.a := -b.a;
{$IFC AnimationEnabled AND AShortestPath }
			if animationFlag[AShortestPathItem] then begin
				plotBisector(b);
				waitForClick(animationFlag[AShortestPathItem]);
			end;
{$ENDC }
			createBisector := b
		end;


	function locate (y: real): reference;
		var
			x: statusType;
			f: boolean;
			d: directionType;
			b: bs;
			r: reference;
		begin
			new(x);
			failNil(Handle(x));
			x.init;
			new(b);
			failNil(Handle(b));
			b.init;
			b.my := y;
			b.mx := 0;
			b.a := 0;
			b.sa := 0;
			b.b := 1;
			b.ca := 1;
			x.b := b;
			f := T.find(x, compT, r, d);
			if (not f) and (d = left) then
				r := T.next(r, left);
			locate := r;
			b.free;
			x.free
		end;

	function insertT1 (i, j: integer; p: shortestPathTreeType): reference;
		var
			spp: shortestPathTreeType;
			x: statusType;
		begin
			new(x);
			failNil(Handle(x));
			x.init;
			x.b := createBisector(i, j);
			new(spp);
			failNil(Handle(spp));
			spp.init;
			freeList.append(spp);
			spp.index := j;
			spp.father := p;
			x.p := spp;
			insertT1 := T.insertObj(x, compT, f)
		end;

	function insertT2 (i, j: integer; p: shortestPathTreeType): reference;
		var
			x: statusType;
		begin
			new(x);
			failNil(Handle(x));
			x.init;
			x.b := createBisector(i, j);
			x.p := p;
			insertT2 := T.insertObj(x, compT, f)
		end;

	begin
		new(T);
		failNil(Handle(T));
		T.init;
		n := self.length div 2;
		freeList := newEmptyList;
		new(SPT);
		failNil(Handle(SPT));
		SPT.init;
		freeList.append(SPT);
		SPT.index := 0;
		new(Q);
		failNil(Handle(Q));
		Q.init;
		for i := 2 to n do begin
			new(s);
			failNil(Handle(s));
			s.init;
			s.id := 1;
			s.i := i;
			s.x := point2d(self.elements^^[2 * i]).getX;
			r := Q.insert(s, compQ);
		end;
		new(ep);
		failNil(Handle(ep));
		ep.init;
		ep.id := 3;
		ep.x := pEnd.getX;
		ep.y := pEnd.getY;
		r := Q.insert(ep, compQ);
		new(event);
		failNil(Handle(event));
		event.init;
		event.x := point2d(self.elements^^[1]).getX;
		new(x);
		failNil(Handle(x));
		x.init;
		new(spp);
		failNil(Handle(spp));
		spp.init;
		freeList.append(spp);
		spp.index := 0;
		x.p := spp;
		new(b);
		failNil(Handle(b));
		b.init;
		b.a := 0;
		b.sa := 0;
		b.b := 1;
		b.ca := 1;
		b.my := -1e8;
		b.mx := 0;
		x.b := b;
		r := T.insertObj(x, compT, f);
		w[0] := 0;
		w[1] := pStart.distanceX(point2d(self.elements^^[1]));
		w[2] := pStart.distanceX(point2d(self.elements^^[2]));
		r := insertT1(0, 1, SPT);
		r := insertT1(1, 2, SPT);
		r := insertT2(0, 2, SPT);
		event.free;
		event := qType(Q.getObject(Q.minimum));
		while event.id <> 3 do begin
			if event.id = 1 then begin
				j := 2 * seg(event).i;
				i := j - 1;
				p1 := point2d(self.elements^^[i]);
				p2 := point2d(self.elements^^[j]);
				r1 := locate(p1.getY);
				r2 := locate(p2.getY);
				h := statusType(T.getObject(r1)).p.index;
				if h = 0 then
					w[i] := p1.distanceX(pStart)
				else
					w[i] := p1.distanceX(point2d(self.elements^^[h])) + w[h];
				k := statusType(T.getObject(r2)).p.index;
				if k = 0 then
					w[j] := p2.distanceX(pStart)
				else
					w[j] := p2.distanceX(point2d(self.elements^^[k])) + w[k];
				r := r2;
				spp := statusType(T.getObject(r2)).p;
				b1 := statusType(T.getObject(r1)).b;
				b2 := statusType(T.getObject(r2)).b;
				while r <> r1 do begin
					r2 := T.next(r, left);
					x := statusType(T.getObject(r));
{$IFC AnimationEnabled AND AShortestPath }
					if animationFlag[AShortestPathItem] then begin
						plotBisector(x.b);
						waitForClick(animationFlag[AShortestPathItem]);
					end;
{$ENDC }
					x.b.free;
					x.free;
					T.delete(r);
					r := r2
				end;
				r := insertT1(statusType(T.getObject(r1)).p.index, 2 * seg(event).i - 1, statusType(T.getObject(r1)).p);
				r := insertT1(2 * seg(event).i - 1, 2 * seg(event).i, spp);
				r2 := insertT2(spp.index, 2 * seg(event).i, spp);
				if T.next(r2, right) <> nil then
					r2 := T.next(r2, right);
				while r1 <> r2 do begin
					r := T.next(r1, right);
					insertQ(statusType(T.getObject(r1)).b, statusType(T.getObject(r)).b);
					r1 := r
				end
			end
			else begin
				new(x1);
				failNil(Handle(x1));
				x1.init;
				new(x2);
				failNil(Handle(x2));
				x2.init;
				x1.b := spmPoint(event).b1;
				x2.b := spmPoint(event).b2;
				f := T.find(x1, compT, r1, d);
				g := T.find(x2, compT, r2, d);
				x1.free;
				x2.free;
				if f and g then begin
					b1 := statusType(T.getObject(r1)).b;
					b2 := statusType(T.getObject(r2)).b;
					if b1.i = b2.i then begin
						i := b1.j;
						j := b2.j
					end;
					if b1.i = b2.j then begin
						i := b1.j;
						j := b2.i
					end;
					if b1.j = b2.i then begin
						i := b1.i;
						j := b2.j
					end;
					if b1.j = b2.j then begin
						i := b1.i;
						j := b2.i
					end;
					if compT(statusType(T.getObject(r1)), statusType(T.getObject(r2))) = greaterThanObj then begin
						r := r1;
						r1 := r2;
						r2 := r
					end;
					x := statusType(T.getObject(r1));
{$IFC AnimationEnabled AND AShortestPath }
					if animationFlag[AShortestPathItem] then begin
						plotBisector(x.b);
						waitForClick(animationFlag[AShortestPathItem]);
					end;
{$ENDC }
					x.b.free;
					x.free;
					T.delete(r1);
					b := statusType(T.getObject(r2)).b;
{$IFC AnimationEnabled AND AShortestPath }
					if animationFlag[AShortestPathItem] then begin
						plotBisector(b);
						waitForClick(animationFlag[AShortestPathItem]);
					end;
{$ENDC }
					b.free;
					statusType(T.getObject(r2)).b := createBisector(i, j);
					r1 := T.next(r2, right);
					if r1 <> nil then
						insertQ(statusType(T.getObject(r2)).b, statusType(T.getObject(r1)).b);
					insertQ(statusType(T.getObject(T.next(r2, left))).b, statusType(T.getObject(r2)).b);
				end;
			end;
			event.free;
			Q.delete(Q.minimum, compQ);
			event := qType(Q.getObject(Q.minimum))
		end;
		l := newEmptyList;
		l.append(pEnd.duplicate);
		spp := statusType(T.getObject(locate(endPoint(event).y))).p;
		event.free;
		Q.free;
		while spp.index <> 0 do begin
			l.append(self.elements^^[spp.index].duplicate);
			spp := spp.father
		end;
		l.append(pStart.duplicate);
		l.revert;
		new(result);
		failNil(Handle(result));
		result.initFromList(l);
		l.free;
		shortestPath := result;
		r := T.extreme(left);
		while r <> nil do begin
			x := statusType(T.getObject(r));
			x.b.free;
			x.free;
			r := T.next(r, right)
		end;
		T.free;
		elem := freeList.first;
		while elem <> nil do begin
			elem^^.value.free;
			elem := elem^^.next
		end;
		freeList.free
	end;
end.
