(*	XYZ GeoBench (eXperimental geometrY Zurich):		*)
(*	An object oriented, interactive workbench for		*)
(*	experimental geometric computation.					*)
(*														*)
(*	 1999 by Peter Schorn, peter.schorn@acm.org		*)
(*	Culmannstrasse 77, CH-8006 Zrich, Switzerland		*)
(*	This software may be copied freely for educational	*)
(*	and research purposes only. For commercial use,		*)
(*	please contact the author.							*)
(*														*)
(*	Version 5.0.5: 30-Sep-1999							*)

unit testGraph;

interface

uses
	graph;

const
	stdDefaultFormula = 'E(D=2)';

function isGraphOk (g: directedGraph; formula: Str255; evaluate: boolean; var errorPosition: integer): boolean;

function showErrorInDefaultFormula (errorPosition: integer): boolean;

procedure initTestGraph;

var
defaultFormula: Str255;

implementation

{$IFC NOT THINK_PASCAL }
{$PUSH }
{$PRAGMAC warn_unusedarg off }
{$ENDC }
function graphEdge.drawingComplexity (cutOff: longInt): longInt;
	begin
		drawingComplexity := 2;
	end;

function markedGraphEdge.drawingComplexity (cutOff: longInt): longInt;
	begin
		drawingComplexity := 3;
	end;
{$IFC NOT THINK_PASCAL }
{$POP }
{$ENDC }

function simpleUndirectedGraph.drawingComplexity (cutOff: longInt): longInt;
	begin
		drawingComplexity := drawingComplexity2(vertices, edges, cutOff);
	end;

function showErrorInDefaultFormula (errorPosition: integer): boolean;
	begin
		if errorPosition <= length(defaultFormula) then begin
			SysBeep(1);
			defaultFormula := concat(copy(defaultFormula, 1, errorPosition - 1), '?', copy(defaultFormula, errorPosition, length(defaultFormula)));
			showErrorInDefaultFormula := true;
		end
		else
			showErrorInDefaultFormula := false;
	end;

	const
		maxIterations = 5;

function graphIsRigid (g: simpleUndirectedGraph): boolean;
	var
		reduceDegreeThreeVertex: seqIndex0;
		proof: homogenVector;
		trials: longInt;
	begin
		proof := g.hennebergSequenceMaxTrials(maxIterations, trials, reduceDegreeThreeVertex);
		graphIsRigid := proof.length = g.vertices.length;
		proof.freeAll;
	end;

function graphIsSimpleRigid (g: simpleUndirectedGraph): boolean;
	var
		reduceDegreeThreeVertex: seqIndex0;
		proof: homogenVector;
		trials: longInt;
	begin
		proof := g.hennebergSequenceMaxTrials(maxIterations, trials, reduceDegreeThreeVertex);
		graphIsSimpleRigid := (proof.length = g.vertices.length) & (reduceDegreeThreeVertex = 0);
		proof.freeAll;
	end;

function graphIsTriangulable (g: directedGraph): boolean;
	var
		startEdge: seqIndex0;
	begin
		graphIsTriangulable := g.maxTriangulationOrder(startEdge) = g.vertices.length;
	end;

function isGraphOk (g: directedGraph; formula: Str255; evaluate: boolean; var errorPosition: integer): boolean;
	var
		pos: integer;

	procedure getSym (var pos: integer);
		begin
			repeat
				pos := pos + 1;
			until formula[pos] <> ' ';
		end;

	procedure check (var pos: integer; ch: char);
		begin
			if formula[pos] = ch then
				getSym(pos)
			else begin
				errorPosition := pos;
				exit(isGraphOk);
			end;
		end;

	function simpleExpression (var pos: integer; l: list; evaluate: boolean): boolean;
		var
			result: boolean;

		function simpleTerm (evaluate: boolean): boolean;
			var
				result: boolean;

			function simpleFactor (evaluate: boolean): boolean;
				var
					leftSide: integer;

				function dOrConstant (evaluate: boolean): integer;
					var
						result: integer;
					begin
						case cap(formula[pos]) of
							'0'..'9':  begin
								result := 0;
								repeat
									result := 10 * result + ord(formula[pos]) - ord('0');
									pos := pos + 1;
								until (formula[pos] < '0') | (formula[pos] > '9');
								while formula[pos] = ' ' do
									pos := pos + 1;
								dOrConstant := result;
							end;
							'D':  begin
								if evaluate then
									dOrConstant := l.length
								else
									dOrConstant := 0;
								getSym(pos);
							end;
							otherwise begin
								errorPosition := pos;
								exit(isGraphOk);
							end;
						end;
					end;

				begin (* simpleFactor *)
					case cap(formula[pos]) of
						'(':  begin
							getSym(pos);
							simpleFactor := simpleExpression(pos, l, evaluate);
							check(pos, ')');
						end;
						'~':  begin
							getSym(pos);
							simpleFactor := not simpleFactor(evaluate);
						end;
						'D', '0'..'9':  begin
							leftSide := dOrConstant(evaluate);
							case formula[pos] of
								'>':  begin
									getSym(pos);
									if formula[pos] = '=' then begin
										getSym(pos);
										simpleFactor := evaluate & (leftSide >= dOrConstant(evaluate)) | not evaluate & (dOrConstant(evaluate) = -1);
									end
									else
										simpleFactor := evaluate & (leftSide > dOrConstant(evaluate)) | not evaluate & (dOrConstant(evaluate) = -1);
								end;
								'<':  begin
									getSym(pos);
									if formula[pos] = '=' then begin
										getSym(pos);
										simpleFactor := evaluate & (leftSide <= dOrConstant(evaluate)) | not evaluate & (dOrConstant(evaluate) = -1);
									end
									else if formula[pos] = '>' then begin
										getSym(pos);
										simpleFactor := evaluate & (leftSide <> dOrConstant(evaluate)) | not evaluate & (dOrConstant(evaluate) = -1);
									end
									else
										simpleFactor := evaluate & (leftSide < dOrConstant(evaluate)) | not evaluate & (dOrConstant(evaluate) = -1);
								end;
								'=':  begin
									getSym(pos);
									simpleFactor := evaluate & (leftSide = dOrConstant(evaluate)) | not evaluate & (dOrConstant(evaluate) = -1);
								end;
								otherwise begin
									errorPosition := pos;
									exit(isGraphOk);
								end;
							end;
						end;
					end;
				end;

			begin (* simpleTerm *)
				result := simpleFactor(evaluate);
				while formula[pos] = '&' do begin
					getSym(pos);
					result := simpleFactor(evaluate & result) & result;
				end;
				simpleTerm := result;
			end;

		begin (* simpleExpression *)
			result := simpleTerm(evaluate);
			while formula[pos] = '|' do begin
				getSym(pos);
				result := simpleTerm(evaluate & not result) | result;
			end;
			simpleExpression := result;
		end;

	function simpleExpressionEvaluate (l: list): boolean;
		var
			localPos: integer;
		begin
			localPos := pos;
			simpleExpressionEvaluate := simpleExpression(localPos, l, true);
		end;

	function queryExpression (evaluate: boolean): boolean;
		var
			result: boolean;

		function queryTerm (evaluate: boolean): boolean;
			var
				result: boolean;

			function queryFactor (evaluate: boolean): boolean;
				var
					dummy: boolean;
				begin
					case cap(formula[pos]) of
						'R':  begin
							getSym(pos);
							queryFactor := evaluate & graphIsRigid(g);
						end;
						'S':  begin
							getSym(pos);
							queryFactor := evaluate & graphIsSimpleRigid(g);
						end;
						'T':  begin
							getSym(pos);
							queryFactor := evaluate & graphIsTriangulable(g);
						end;
						'(':  begin
							getSym(pos);
							queryFactor := queryExpression(evaluate);
							check(pos, ')');
						end;
						'~':  begin
							getSym(pos);
							queryFactor := not queryFactor(evaluate);
						end;
						'A':  begin
							getSym(pos);
							check(pos, '(');
							queryFactor := evaluate & g.adjacency.sameProperty(simpleExpressionEvaluate);
							dummy := simpleExpression(pos, nil, false);
							check(pos, ')');
						end;
						'E':  begin
							getSym(pos);
							check(pos, '(');
							queryFactor := evaluate & (g.adjacency.exists(simpleExpressionEvaluate) > 0);
							dummy := simpleExpression(pos, nil, false);
							check(pos, ')');
						end;
						otherwise begin
							errorPosition := pos;
							exit(isGraphOk);
						end;
					end;
				end;

			begin (* queryTerm *)
				result := queryFactor(evaluate);
				while formula[pos] = '&' do begin
					getSym(pos);
					result := queryFactor(evaluate & result) & result;
				end;
				queryTerm := result;
			end;

		begin (* queryExpression *)
			result := queryTerm(evaluate);
			while formula[pos] = '|' do begin
				getSym(pos);
				result := queryTerm(evaluate & not result) | result;
			end;
			queryExpression := result;
		end;

	begin
		isGraphOk := false;
		if length(formula) = 255 then begin
			errorPosition := 1;
			exit(isGraphOk);
		end;
		formula := Concat(formula, '.');
		pos := 0;
		getSym(pos);
		isGraphOk := queryExpression(evaluate);
		errorPosition := pos;
	end;

procedure initTestGraph;
	begin
		defaultFormula := stdDefaultFormula;
	end;

end.
