(*	XYZ GeoBench (eXperimental geometrY Zurich):		*)
(*	An object oriented, interactive workbench for		*)
(*	experimental geometric computation.					*)
(*														*)
(*	 1999 by Peter Schorn, peter.schorn@acm.org		*)
(*	Culmannstrasse 77, CH-8006 Zrich, Switzerland		*)
(*	This software may be copied freely for educational	*)
(*	and research purposes only. For commercial use,		*)
(*	please contact the author.							*)
(*														*)
(*	Version 5.0.5: 30-Sep-1999							*)

unit threeDVision;

interface

const
	nHigh = 4;

type
	indexRangeT = 1..nHigh;
	point3dT = array[1..nHigh] of extended;								(* homogenous coordinates						*)
	trafoMatrT = array[1..nHigh, 1..nHigh] of extended;				(* transformations									*)
	axisT = (xAxis, yAxis, zAxis);

procedure initM (var M: trafoMatrT);
	(* initializes m to the identity matrix																						*)

procedure mvm (M: trafoMatrT; p: point3dT; var res: point3dT);
	(* matrix vector multiplication																								*)

procedure mmm (M, N: trafoMatrT; var res: trafoMatrT);
	(* matrix matrix multiplication																								*)

procedure initViewingTransformation (var eyePoint, proPoint, lightPoint: point3dT; var U: trafoMatrT; var scal: point3dT; var P, V: trafoMatrT);
	(* must be called to set the whole viewing transformation to a defined state											*)

procedure newEyePoint (var oldEyePoint, newEyePoint: point3dT; var U: trafoMatrT; var scal: point3dT; var P, V: trafoMatrT);
	(* sets the viewing transformation for a changing eyepoint																*)
	(* U doesn't change																												*)

procedure newProPoint (var oldProPoint, newProPoint, eyePoint: point3dT; var U: trafoMatrT; var scal: point3dT; var P, V: trafoMatrT);
	(* sets the viewing transformation for a changing propoint																*)
	(* U doesn't change																												*)

procedure translation (delta: point3dT; var eyePoint, proPoint, lightPoint: point3dT; var U: trafoMatrT; var scal: point3dT; var P, V: trafoMatrT);
	(* updates the viewing transformation applying the given translation													*)
	(* P doesn't change																												*)

procedure scaling (s: point3dT; var eyePoint, proPoint, lightPoint: point3dT; var U: trafoMatrT; var scal: point3dT; var P, V: trafoMatrT);
	(* updates the viewing transformation applying the given scaling														*)
	(* P doesn't change																												*)

procedure rotation (a: axisT; angle: extended; var eyePoint, proPoint, lightPoint: point3dT; var U: trafoMatrT; var scal: point3dT; var P, V: trafoMatrT);
	(* updates the viewing transformation applying the given rotation														*)
	(* P doesn't change																												*)


implementation

{$IFC NOT THINK_PASCAL }
uses OSUtils, Sound;
{$ENDC }

const
	eps = 0.0003;

procedure initM (var m: trafoMatrT);
	var
		i, j: indexRangeT;
	begin
		for i := 1 to 4 do
			for j := 1 to 4 do
				if i = j then
					m[i, j] := 1.0
				else
					m[i, j] := 0.0;
	end;

procedure mvm (M: trafoMatrT; p: point3dT; var res: point3dT);
	var
		i, j: indexRangeT;
	begin
		for i := 1 to nHigh do begin
			res[i] := 0;
			for j := 1 to nHigh do
				res[i] := res[i] + M[i, j] * p[j];
		end;
	end;

procedure mmm (M, N: trafoMatrT; var res: trafoMatrT);
	var
		i, j, k: indexRangeT;
	begin
		for i := 1 to nHigh do
			for j := 1 to nHigh do begin
				res[i, j] := 0.0;
				for k := 1 to nHigh do
					res[i, j] := res[i, j] + M[i, k] * N[k, j];
			end;
	end;

procedure initViewingTransformation (var eyePoint, proPoint, lightPoint: point3dT; var U: trafoMatrT; var scal: point3dT; var P, V: trafoMatrT);
	var
		i: indexRangeT;
		T1, T2: trafoMatrT;
		r, s: extended;
	begin
		initM(U);
		initM(P);
		initM(V);
		for i := 1 to 4 do
			scal[i] := 1.0;
		eyePoint[1] := 2000.0;
		eyePoint[2] := 3000.0;
		eyePoint[3] := 1000.0;
		eyePoint[4] := 1.0;
		lightPoint[1] := 2500.0;
		lightPoint[2] := 2500.0;
		lightPoint[3] := 2500.0;
		lightPoint[4] := 1.0;
		for i := 1 to nHigh - 1 do
			proPoint[i] := eyePoint[i] * 0.1;
		proPoint[nHigh] := 1.0;
		initM(T1);
		initM(T2);
		r := sqr(proPoint[1]) + sqr(proPoint[2]);
		s := sqrt(r + sqr(proPoint[3]));
		r := sqrt(r);
		T1[1, 1] := proPoint[2] / r;									(* initializing the projection P	*)
		T1[2, 2] := T1[1, 1];
		T1[1, 2] := -proPoint[1] / r;
		T1[2, 1] := -T1[1, 2];
		T2[2, 2] := r / s;
		T2[3, 3] := T2[2, 2];
		T2[2, 3] := proPoint[3] / s;
		T2[3, 2] := -T2[2, 3];
		mmm(T2, T1, P);
		T1 := P;
		initM(T2);
		T2[2, 2] := 0.0;
		T2[3, 3] := 0.0;
		T2[2, 3] := -1.0;
		T2[3, 2] := -1.0;
		T2[3, 4] := s;
		mmm(T2, T1, P);
		mmm(P, U, V);
	end;

procedure newEyePoint (var oldEyePoint, newEyePoint: point3dT; var U: trafoMatrT; var scal: point3dT; var P, V: trafoMatrT);
	var
		T1, T2, H: trafoMatrT;
		a, adot, o, odot: point3dT;
		t, s: extended;
		i: indexRangeT;
	begin
		mvm(V, newEyePoint, adot);
		initM(T1);
		t := sqr(adot[3]) + sqr(adot[1]);
		s := sqrt(t + sqr(adot[2]));
		t := sqrt(t);
		if (t < eps) | (s < eps) then begin
			SysBeep(2);
			newEyePoint := oldEyePoint;
		end
		else begin
			T1[1, 1] := -adot[3] / t;
			T1[1, 3] := adot[1] / t;
			T1[3, 3] := T1[1, 1];
			T1[3, 1] := -T1[1, 3];
			initM(T2);
			T2[2, 2] := t / s;
			T2[3, 3] := T2[2, 2];
			T2[2, 3] := adot[2] / s;
			T2[3, 2] := -T2[2, 3];
			mmm(T2, T1, H);
			mmm(H, P, T1);
			P := T1;
			a[1] := 0;											(* what was vertically should stay so... *)
			a[2] := 0;
			a[3] := 1;
			a[4] := 1;
			mvm(P, a, adot);
			o[1] := 0;
			o[2] := 0;
			o[3] := 0;
			o[4] := 1;
			mvm(P, o, odot);
			adot[1] := adot[1] - odot[1];
			adot[2] := adot[2] - odot[2];
			adot[3] := adot[3] - odot[3];
			t := sqrt(sqr(adot[1]) + sqr(adot[2]));
			if (t < eps) then begin
				SysBeep(2);
				mmm(P, U, T1);
				initM(T2);
				for i := 1 to 4 do
					T2[i, i] := scal[i];
				mmm(T1, T2, V);
			end
			else begin
				initM(T1);
				T1[1, 1] := -adot[2] / t;
				T1[2, 2] := T1[1, 1];
				T1[1, 2] := adot[1] / t;
				T1[2, 1] := -T1[1, 2];
				mmm(T1, P, T2);
				P := T2;
				mmm(P, U, T1);
				initM(T2);
				for i := 1 to 4 do
					T2[i, i] := scal[i];
				mmm(T1, T2, V);
			end;
		end;
	end;

procedure newProPoint (var oldProPoint, newProPoint, eyePoint: point3dT; var U: trafoMatrT; var scal: point3dT; var P, V: trafoMatrT);
	var
		T1, T2, T3, H: trafoMatrT;
		a, o, adot, pdot, odot: point3dT;
		s, t: extended;
		i: indexRangeT;
	begin
		mvm(V, eyePoint, adot);
		mvm(V, newProPoint, pdot);
		initM(T1);
		T1[1, 4] := -pdot[1];
		T1[2, 4] := -pdot[2];
		T1[3, 4] := -pdot[3];
		initM(T2);
		s := sqr(pdot[3] - adot[3]) + sqr(pdot[1]);
		t := sqrt(s + sqr(pdot[2]));
		s := sqrt(s);
		if (s < eps) | (t < eps) then begin
			SysBeep(2);
			newProPoint := oldProPoint;
		end
		else begin
			T2[1, 1] := (pdot[3] - adot[3]) / s;
			T2[3, 3] := T2[1, 1];
			T2[1, 3] := -pdot[1] / s;
			T2[3, 1] := -T2[1, 3];
			initM(T3);
			T3[2, 2] := s / t;
			T3[3, 3] := T3[2, 2];
			T3[2, 3] := -pdot[2] / t;
			T3[3, 2] := -T3[2, 3];
			mmm(T2, T1, H);
			mmm(T3, H, T1);
			mmm(T1, P, H);
			P := H;
			a[1] := 0;											(* what was vertically should stay so... *)
			a[2] := 0;
			a[3] := 1;
			a[4] := 1;
			mvm(P, a, adot);
			o[1] := 0;
			o[2] := 0;
			o[3] := 0;
			o[4] := 1;
			mvm(P, o, odot);
			adot[1] := adot[1] - odot[1];
			adot[2] := adot[2] - odot[2];
			adot[3] := adot[3] - odot[3];
			t := sqrt(sqr(adot[1]) + sqr(adot[2]));
			if (t < eps) then begin
				SysBeep(2);
				mmm(P, U, T1);
				initM(T2);
				for i := 1 to 4 do
					T2[i, i] := scal[i];
				mmm(T1, T2, V);
			end
			else begin
				initM(T1);
				T1[1, 1] := -adot[2] / t;
				T1[2, 2] := T1[1, 1];
				T1[1, 2] := adot[1] / t;
				T1[2, 1] := -T1[1, 2];
				mmm(T1, P, T2);
				P := T2;
				mmm(P, U, T1);
				initM(T2);
				for i := 1 to 4 do
					T2[i, i] := scal[i];
				mmm(T1, T2, V);
			end;
		end;
	end;

procedure translation (delta: point3dT; var eyePoint, proPoint, lightPoint: point3dT; var U: trafoMatrT; var scal: point3dT; var P, V: trafoMatrT);
	var
		T, T1, T2, TInv: trafoMatrT;
		newEye, newPro, newLight: point3dT;
		i: indexRangeT;
	begin
		initM(T);
		T[1, 4] := delta[1] * scal[1];
		T[2, 4] := delta[2] * scal[2];
		T[3, 4] := delta[3] * scal[3];
		mmm(U, T, T1);
		U := T1;
		initM(TInv);
		TInv[1, 4] := -delta[1] / scal[1];
		TInv[2, 4] := -delta[2] / scal[2];
		TInv[3, 4] := -delta[3] / scal[3];
		for i := 1 to 4 do								(* undo scaling	*)
			begin
			eyePoint[i] := eyePoint[i] * scal[i];
			proPoint[i] := proPoint[i] * scal[i];
			lightPoint[i] := lightPoint[i] * scal[i];
		end;
		mvm(TInv, eyePoint, newEye);
		mvm(TInv, proPoint, newPro);
		mvm(TInv, lightPoint, newLight);
		eyePoint := newEye;
		proPoint := newPro;
		lightPoint := newLight;
		for i := 1 to 4 do begin
			eyePoint[i] := eyePoint[i] / scal[i];		(* scaling		*)
			proPoint[i] := proPoint[i] / scal[i];
			lightPoint[i] := lightPoint[i] / scal[i];
		end;
		mmm(P, U, T1);
		initM(T2);
		for i := 1 to 4 do
			T2[i, i] := scal[i];
		mmm(T1, T2, V);
	end;

procedure scaling (s: point3dT; var eyePoint, proPoint, lightPoint: point3dT; var U: trafoMatrT; var scal: point3dT; var P, V: trafoMatrT);
	var
		T, T1, T2, TInv: trafoMatrT;
		newEye, newPro, newLight: point3dT;
		i: indexRangeT;
	begin
		initM(T);
		for i := 1 to 4 do
			scal[i] := scal[i] * s[i];
		T[1, 1] := s[1];
		T[2, 2] := s[2];
		T[3, 3] := s[3];
		initM(TInv);
		TInv[1, 1] := 1.0 / s[1];
		TInv[2, 2] := 1.0 / s[2];
		TInv[3, 3] := 1.0 / s[3];
		mvm(TInv, eyePoint, newEye);
		mvm(TInv, proPoint, newPro);
		mvm(TInv, lightPoint, newLight);
		eyePoint := newEye;
		proPoint := newPro;
		lightPoint := newLight;
		mmm(P, U, T1);
		initM(T2);
		for i := 1 to 4 do
			T2[i, i] := scal[i];
		mmm(T1, T2, V);
	end;

procedure rotation (a: axisT; angle: extended; var eyePoint, proPoint, lightPoint: point3dT; var U: trafoMatrT; var scal: point3dT; var P, V: trafoMatrT);
	var
		T, T1, T2, TInv: trafoMatrT;
		newEye, newPro, newLight: point3dT;
		i: indexRangeT;
	begin
		initM(T);
		case a of
			xAxis:  begin
				T[2, 2] := cos(angle);
				T[3, 3] := T[2, 2];
				T[2, 3] := sin(angle);
				T[3, 2] := -T[2, 3];
			end;
			yAxis:  begin
				T[1, 1] := cos(angle);
				T[3, 3] := T[1, 1];
				T[1, 3] := -sin(angle);
				T[3, 1] := -T[1, 3];
			end;
			zAxis:  begin
				T[1, 1] := cos(angle);
				T[2, 2] := T[1, 1];
				T[1, 2] := sin(angle);
				T[2, 1] := -T[1, 2];
			end;
		end;
		mmm(U, T, T1);
		U := T1;
		initM(TInv);
		case a of
			xAxis:  begin
				TInv[2, 2] := T[2, 2];
				TInv[3, 3] := T[3, 3];
				TInv[2, 3] := T[3, 2];
				TInv[3, 2] := T[2, 3];
			end;
			yAxis:  begin
				TInv[1, 1] := T[1, 1];
				TInv[3, 3] := T[3, 3];
				TInv[1, 3] := T[3, 1];
				TInv[3, 1] := T[1, 3];
			end;
			zAxis:  begin
				TInv[1, 1] := T[1, 1];
				TInv[2, 2] := T[2, 2];
				TInv[1, 2] := T[2, 1];
				TInv[2, 1] := T[1, 2];
			end;
		end;
		for i := 1 to 4 do								(* undo scaling	*)
			begin
			eyePoint[i] := eyePoint[i] * scal[i];
			proPoint[i] := proPoint[i] * scal[i];
			lightPoint[i] := lightPoint[i] * scal[i];
		end;
		mvm(TInv, eyePoint, newEye);
		mvm(TInv, proPoint, newPro);
		mvm(TInv, lightPoint, newLight);
		eyePoint := newEye;
		proPoint := newPro;
		lightPoint := newLight;
		for i := 1 to 4 do begin
			eyePoint[i] := eyePoint[i] / scal[i];		(* scaling		*)
			proPoint[i] := proPoint[i] / scal[i];
			lightPoint[i] := lightPoint[i] / scal[i];
		end;
		mmm(P, U, T1);
		initM(T2);
		for i := 1 to 4 do
			T2[i, i] := scal[i];
		mmm(T1, T2, V);
	end;

end.
