(*	XYZ GeoBench (eXperimental geometrY Zurich):		*)
(*	An object oriented, interactive workbench for		*)
(*	experimental geometric computation.					*)
(*														*)
(*	 1999 by Peter Schorn, peter.schorn@acm.org		*)
(*	Culmannstrasse 77, CH-8006 Zrich, Switzerland		*)
(*	This software may be copied freely for educational	*)
(*	and research purposes only. For commercial use,		*)
(*	please contact the author.							*)
(*														*)
(*	Version 5.0.5: 30-Sep-1999							*)

unit travelingOptimalBB;

interface

uses
	pointVector;

implementation

(*	By Adrian Bruengger, bruengg@inf.ethz.ch	*)

{$SETC ATravelingSalesman=TRUE }

const
	MaxPts = 501;
	MaxPts1 = MaxPts - 1;
	MaxEdges = longint(MaxPts) * longint(MaxPts1) div 2;
	MaxEdges1 = MaxEdges - 1;

type

	PointT = record
			x, y: real;
			degree: integer;				(* in 1tree	*)
			lambda: real;					(* for iteration	*)
			saveLambda: real;				(* best known lambda *)
			group: integer;					(* to compute 1tree	*)
			inclDeg: integer;				(* degree with respect to included edges	*)
		end;

	PointListT = array[0..MaxPts1] of PointT;
	PointListP = ^PointListT;
	PointListH = ^PointListP;

	PointIndListT = array[0..MaxPts1] of integer;
	PointIndListP = ^PointIndListT;
	PointIndListH = ^PointIndListP;

	EdgeT = record
			fromP, toP: integer;
			length: real;
			updLen: real;
			sortPos: integer;
			count: integer;
			guessLen: real;					(* as long as the expected tour length is < guessLen, the edge is proved to be excluded *)
		end;

	EdgeListT = array[0..MaxEdges1] of EdgeT;
	EdgeListP = ^EdgeListT;
	EdgeListH = ^EdgeListP;

	EdgeIndListT = array[0..MaxEdges1] of integer;
	EdgeIndListP = ^EdgeIndListT;
	EdgeIndListH = ^EdgeIndListP;

	DescT = record
			nrOfPoints, nrOfEdges: integer;
			lastIncl, firstExcl: integer;
			oneTreePoint: integer;
		end;
	DescP = ^DescT;
	DescH = ^DescP;

const
	incrVal = 1.0075;

type
	TSPNodeP = ^TSPNodeT;
	TSPNodeT = record
			son, brother, father: TSPNodeP;
			lowerBound: real;
			myNumber: integer;			(* 1:both edges included						*)
										(* 2:edge1 indcluded, edge2 excluded		*)
										(* 3:fathers edge1 excluded					*)
			edge1, edge2: integer;		(* edges to branch								*)
			notFirst: boolean;				(* the first son is not possible				*)
		end;

procedure newTSPNode(var node: TSPNodeP);
	var
		result: TSPNodeP;
	begin
		new(result);
		failNil(Handle(result));
		result^.son := nil;
		result^.brother := nil;
		result^.father := nil;
		node := result;
	end;

{$IFC AnimationEnabled AND ATravelingSalesman }
procedure showOneTree (desc: DescP; points: PointListP; edges: EdgeListP; oneTree: PointIndListP);
	var
		i: longint;
	begin
		for i := 0 to desc^.nrOfPoints - 1 do
			if edges^[oneTree^[i]].sortPos <= desc^.lastIncl then
				drawThickLineSegment(points^[edges^[oneTree^[i]].fromP].x, points^[edges^[oneTree^[i]].fromP].y, points^[edges^[oneTree^[i]].toP].x, points^[edges^[oneTree^[i]].toP].y)
			else
				drawLineSegment(points^[edges^[oneTree^[i]].fromP].x, points^[edges^[oneTree^[i]].fromP].y, points^[edges^[oneTree^[i]].toP].x, points^[edges^[oneTree^[i]].toP].y);
		drawPoint(points^[desc^.oneTreePoint].x, points^[desc^.oneTreePoint].y, pointRadiusHighlight);
	end;
{$ENDC }

procedure QuickSortEdgesByLength (desc: DescP; edges: EdgeListP; sortedEdges: EdgeIndListP);
	(* QuickSort to sort the edges from l to r according to their updLen *)

	procedure Sort (l, r: integer);
		var
			i, j, k, t: integer;
		begin
			while l < r do begin
				t := sortedEdges^[(l + r) div 2];
				i := l;
				j := r;
				repeat
					while edges^[sortedEdges^[i]].updLen < edges^[t].updLen do
						i := i + 1;
					while edges^[sortedEdges^[j]].updLen > edges^[t].updLen do
						j := j - 1;
					if i <= j then begin (* switch *)
						k := sortedEdges^[i];
						sortedEdges^[i] := sortedEdges^[j];
						sortedEdges^[j] := k;
						edges^[sortedEdges^[i]].sortPos := i;
						edges^[sortedEdges^[j]].sortPos := j;
						i := i + 1;
						j := j - 1;
					end;
				until j < i;
				if j - l < r - i then begin
					Sort(l, j);
					l := i;
				end
				else begin
					Sort(i, r);
					r := j;
				end
			end;
		end;

	begin (* QuickSortEdgesByLength *)
		Sort(desc^.lastIncl + 1, desc^.firstExcl - 1);
	end;

procedure InsertionSortEdgesByLength (desc: DescP; edges: EdgeListP; sortedEdges: EdgeIndListP);
	(* Straight insertion sort to sort the edges from l to r according to their updLen *)
	var
		i, j, theEdge: integer;
	begin
		for i := desc^.lastIncl + 2 to desc^.firstExcl - 1 do begin
			theEdge := sortedEdges^[i];
			j := i - 1;
			while (j > desc^.lastIncl) & (edges^[sortedEdges^[j]].updLen > edges^[theEdge].updLen) do begin
				sortedEdges^[j + 1] := sortedEdges^[j];
				j := j - 1;
			end;
			sortedEdges^[j + 1] := theEdge;
		end;
		for i := desc^.lastIncl + 1 to desc^.firstExcl - 1 do
			edges^[sortedEdges^[i]].sortPos := i;
	end;

procedure QuickSortEdgesByGuessLength (desc: DescP; edges: EdgeListP; sortedEdges: EdgeIndListP);
	(* QuickSort to sort the edges from l to r according to their (len + delta) *)

	procedure Sort (l, r: integer);
		var
			i, j, k, t: integer;
		begin
			while l < r do begin
				t := sortedEdges^[(l + r) div 2];
				i := l;
				j := r;
				repeat
					while edges^[sortedEdges^[i]].guessLen < edges^[t].guessLen do
						i := i + 1;
					while edges^[sortedEdges^[j]].guessLen > edges^[t].guessLen do
						j := j - 1;
					if i <= j then begin (* switch *)
						k := sortedEdges^[i];
						sortedEdges^[i] := sortedEdges^[j];
						sortedEdges^[j] := k;
						edges^[sortedEdges^[i]].sortPos := i;
						edges^[sortedEdges^[j]].sortPos := j;
						i := i + 1;
						j := j - 1;
					end;
				until j < i;
				if j - l < r - i then begin
					Sort(l, j);
					l := i;
				end
				else begin
					Sort(i, r);
					r := j;
				end
			end;
		end;

	begin (* QuickSortEdgesByGuessLength *)
		Sort(desc^.lastIncl + 1, desc^.firstExcl - 1);
	end;

procedure initTSP (nrOfPoints: integer; desc: DescP; points: PointListP; edges: EdgeListP; sortedEdges: EdgeIndListP; oneTree: PointIndListP; tour: PointIndListP; pv: pointVector);
	var
		i, j, h: integer;
	begin
		desc^.nrOfPoints := nrOfPoints;
		desc^.nrOfEdges := desc^.nrOfPoints * (desc^.nrOfPoints - 1) div 2;
		desc^.lastIncl := -1;
		desc^.firstExcl := desc^.nrOfEdges;
		desc^.oneTreePoint := 0;
		for i := 0 to desc^.nrOfPoints - 1 do begin
			points^[i].x := realPoint(pv.elements^^[i + 1]).x;
			points^[i].y := realPoint(pv.elements^^[i + 1]).y;
			points^[i].inclDeg := 0;
			points^[i].degree := 0;
			points^[i].group := i;
			points^[i].lambda := 0.0;
			points^[i].saveLambda := 0;
			oneTree^[i] := i;
			tour^[i] := i;
		end;
		h := 0;
		for i := 0 to desc^.nrOfPoints - 2 do
			for j := (i + 1) to desc^.nrOfPoints - 1 do begin
				edges^[h].fromP := i;
				edges^[h].toP := j;
				edges^[h].length := sqrt((points^[i].x - points^[j].x) * (points^[i].x - points^[j].x) + (points^[i].y - points^[j].y) * (points^[i].y - points^[j].y));
				edges^[h].updLen := edges^[h].length;
				edges^[h].guessLen := 0.0;
				edges^[h].sortPos := h;
				edges^[h].count := 0;
				sortedEdges^[h] := h;
				h := h + 1;
			end;
	end;

function excludeEdge (desc: DescP; edges: EdgeListP; sortedEdges: EdgeIndListP; edgeNumber: integer; var exclCount: integer): boolean;
	var
		i, oldSortPos, fromP, toP: integer;
	begin
		if (edges^[edgeNumber].sortPos > desc^.lastIncl) & (edges^[edgeNumber].sortPos < desc^.firstExcl) & (desc^.firstExcl > desc^.lastIncl) then begin
			fromP := edges^[edgeNumber].fromP;
			toP := edges^[edgeNumber].toP;
			desc^.firstExcl := desc^.firstExcl - 1;
			exclCount := exclCount + 1;
			i := sortedEdges^[desc^.firstExcl];
			oldSortPos := edges^[edgeNumber].sortPos;
			sortedEdges^[desc^.firstExcl] := edgeNumber;
			edges^[edgeNumber].sortPos := desc^.firstExcl;
			sortedEdges^[oldSortPos] := i;
			edges^[i].sortPos := oldSortPos;
			excludeEdge := true;
		end
		else
			excludeEdge := false;
	end;

function includeEdge (desc: DescP; points: PointListP; edges: EdgeListP; sortedEdges: EdgeIndListP; edgeNumber: integer; var inclCount, exclCount: integer): boolean;
	var
		i, oldSortPos, fromP, toP: integer;
		result: boolean;
	begin
		if (edges^[edgeNumber].sortPos > desc^.lastIncl) & (edges^[edgeNumber].sortPos < desc^.firstExcl) & (desc^.firstExcl > desc^.lastIncl) then begin
			desc^.lastIncl := desc^.lastIncl + 1;
			inclCount := inclCount + 1;
			i := sortedEdges^[desc^.lastIncl];
			oldSortPos := edges^[edgeNumber].sortPos;
			sortedEdges^[desc^.lastIncl] := edgeNumber;
			edges^[edgeNumber].sortPos := desc^.lastIncl;
			sortedEdges^[oldSortPos] := i;
			edges^[i].sortPos := oldSortPos;
			result := true;
			fromP := edges^[edgeNumber].fromP;
			toP := edges^[edgeNumber].toP;
			points^[fromP].inclDeg := points^[fromP].inclDeg + 1;
			points^[toP].inclDeg := points^[toP].inclDeg + 1;
			if points^[fromP].inclDeg = 2 then begin
				i := desc^.firstExcl - 1;
				while i > desc^.lastIncl do begin
					if (edges^[sortedEdges^[i]].fromP = fromP) | (edges^[sortedEdges^[i]].toP = fromP) then
						result := result & excludeEdge(desc, edges, sortedEdges, sortedEdges^[i], exclCount);
					i := i - 1;
				end;
			end;
			if points^[toP].inclDeg = 2 then begin
				i := desc^.firstExcl - 1;
				while i > desc^.lastIncl do begin
					if (edges^[sortedEdges^[i]].fromP = toP) | (edges^[sortedEdges^[i]].toP = toP) then
						result := result & excludeEdge(desc, edges, sortedEdges, sortedEdges^[i], exclCount);
					i := i - 1;
				end;
			end;
		end
		else
			result := false;
		includeEdge := result;
	end;

procedure rebuildEdgeList (desc: DescP; points: PointListP; edges: EdgeListP; sortedEdges: EdgeIndListP; var inclCount, exclCount: integer);
	var
		k: integer;
	begin
		for k := (desc^.lastIncl - inclCount + 1) to desc^.lastIncl do begin
			points^[edges^[sortedEdges^[k]].fromP].inclDeg := points^[edges^[sortedEdges^[k]].fromP].inclDeg - 1;
			points^[edges^[sortedEdges^[k]].toP].inclDeg := points^[edges^[sortedEdges^[k]].toP].inclDeg - 1;
		end;
		desc^.lastIncl := desc^.lastIncl - inclCount;
		desc^.firstExcl := desc^.firstExcl + exclCount;
		inclCount := 0;
		exclCount := 0;
		for k := desc^.lastIncl + 1 to desc^.firstExcl - 1 do
			edges^[sortedEdges^[k]].count := 0;
	end;

procedure buildTourFromOneTree (desc: DescP; edges: EdgeListP; oneTree: PointIndListP; tour: PointIndListP);
	var
		thisP, lastP, i, j: integer;
		found: boolean;
	begin
		tour^[0] := edges^[oneTree^[0]].fromP;
		tour^[1] := edges^[oneTree^[0]].toP;
		for i := 2 to desc^.nrOfPoints - 1 do begin
			thisP := tour^[i - 1];
			lastP := tour^[i - 2];
			j := 0;
			found := false;
			while not found do begin
				if (edges^[oneTree^[j]].fromP = thisP) & (edges^[oneTree^[j]].toP <> lastP) then begin
					tour^[i] := edges^[oneTree^[j]].toP;
					found := true;
					lastP := thisP;
					thisP := tour^[i];
				end
				else if (edges^[oneTree^[j]].toP = thisP) & (edges^[oneTree^[j]].fromP <> lastP) then begin
					tour^[i] := edges^[oneTree^[j]].fromP;
					found := true;
					lastP := thisP;
					thisP := tour^[i];
				end;
				j := j + 1;
			end;
		end;
	end;

procedure getBranchingEdges (desc: DescP; points: PointListP; edges: EdgeListP; oneTree: PointIndListP; var edge1, edge2: integer; var notFirst: boolean);
	var
		maxDeg, maxP, i, incEdge: integer;
	begin
		maxDeg := points^[0].degree; (* there must be a node with degree at least 3 with free edges, otherwise the oneTree is a tour *)
		maxP := 0;
		for i := 1 to desc^.nrOfPoints - 1 do
			if points^[i].degree > maxDeg then begin
				maxDeg := points^[i].degree;
				maxP := i;
			end;
		edge1 := -1;
		edge2 := -1;
		incEdge := -1;
		for i := 0 to desc^.nrOfPoints - 1 do
			if (maxP = edges^[oneTree^[i]].fromP) | (maxP = edges^[oneTree^[i]].toP) then begin
				if edges^[oneTree^[i]].sortPos <= desc^.lastIncl then
					incEdge := oneTree^[i]
				else if (edge1 = -1) | ((edges^[oneTree^[i]].count > edges^[edge1].count) & not (edge2 = -1)) then
					edge1 := oneTree^[i]
				else if (edge2 = -1) | (edges^[oneTree^[i]].count > edges^[edge2].count) then
					edge2 := oneTree^[i];
			end;
		notFirst := incEdge <> -1;
		if notFirst then begin
			edge2 := edge1;
			edge1 := incEdge;
		end;
	end;

procedure findOneTree (desc: DescP; points: PointListP; edges: EdgeListP; sortedEdges: EdgeIndListP; oneTree: PointIndListP; var oneTreeLength: real; var tourFound, possible: boolean);
	var
		i, k, theEdge, fromP, toP, fromGroup, toGroup, otDeg: integer;
	begin
		oneTreeLength := 0.0;
		possible := true;
		for i := 0 to desc^.nrOfPoints - 1 do begin
			points^[i].group := i;
			points^[i].degree := 0;
		end;
		i := 0;
		k := 0;
		otDeg := 0;
		while possible & (k < desc^.firstExcl) & (i <> desc^.nrOfPoints) do begin
			theEdge := sortedEdges^[k];
			fromP := edges^[theEdge].fromP;
			toP := edges^[theEdge].toP;
			fromGroup := points^[fromP].group;
			while fromGroup <> points^[fromGroup].group do
				fromGroup := points^[fromGroup].group;
			toGroup := points^[toP].group;
			while toGroup <> points^[toGroup].group do
				toGroup := points^[toGroup].group;
			tourFound := (toP = desc^.oneTreePoint) | (fromP = desc^.oneTreePoint);
			if tourFound & (otDeg < 2) then begin
				oneTree^[i] := theEdge;
				points^[fromP].degree := points^[fromP].degree + 1;
				points^[toP].degree := points^[toP].degree + 1;
				otDeg := otDeg + 1;
				oneTreeLength := oneTreeLength + edges^[theEdge].updLen;
				i := i + 1;
			end
			else if (toGroup <> fromGroup) & not tourFound then begin
				oneTree^[i] := theEdge;
				points^[fromP].degree := points^[fromP].degree + 1;
				points^[toP].degree := points^[toP].degree + 1;
				edges^[theEdge].count := edges^[theEdge].count + 1;
				points^[fromGroup].group := toGroup;
				i := i + 1;
				oneTreeLength := oneTreeLength + edges^[theEdge].updLen;
			end
			else if k <= desc^.lastIncl then
				possible := false;
			k := k + 1;
		end;
		possible := possible & (otDeg = 2) & (i = desc^.nrOfPoints);
		tourFound := true;
		if possible then
			for i := 0 to desc^.nrOfPoints - 1 do
				tourFound := tourFound & (points^[i].degree = 2)
		else
			tourFound := false;
	end;

procedure IterateOneTree (desc: DescP; points: PointListP; edges: EdgeListP; sortedEdges: EdgeIndListP; oneTree: PointIndListP; var oneTreeLength: real; var tourFound, possible: boolean; var best: real; exceed: real; var cancel: boolean);
	const
		minImprove = 0.001;
	var
		q, r, s, t, u, oldBest, bestOneTree: real;
		nrOfIterations, i, theEdge, improveIterations: integer;
	begin
		improveIterations := desc^.nrOfPoints div 3;
(*q := 0.0004 * best; *)
		q := 1.0;
		r := 0.8;
		s := 1.3;
		t := 1.5;
		u := 1.0;
		nrOfIterations := 0;
		tourFound := false;
		possible := true;
		bestOneTree := 0.0;
		for i := 0 to desc^.firstExcl - 1 do begin
			theEdge := sortedEdges^[i];
			edges^[theEdge].updLen := edges^[theEdge].length + points^[edges^[theEdge].fromP].lambda + points^[edges^[theEdge].toP].lambda;
		end;
		repeat
			InsertionSortEdgesByLength(desc, edges, sortedEdges);
			findOneTree(desc, points, edges, sortedEdges, oneTree, oneTreeLength, tourFound, possible);
			if oneTreeLength <= bestOneTree then
				t := t * r
			else begin
				bestOneTree := oneTreeLength;
				for i := 0 to desc^.nrOfPoints - 1 do
					points^[i].saveLambda := points^[i].lambda;
				t := t * s;
				if t < 1.0 then
					t := t * 2.0
				else if t < 2.0 then
					t := 2.0;
			end;
			for i := 0 to desc^.nrOfPoints - 1 do
				points^[i].lambda := points^[i].lambda + t * (points^[i].degree - 2) * q;
			for i := 0 to desc^.firstExcl - 1 do begin
				theEdge := sortedEdges^[i];
				edges^[theEdge].updLen := edges^[theEdge].length + points^[edges^[theEdge].fromP].lambda + points^[edges^[theEdge].toP].lambda;
			end;
			if (nrOfIterations mod improveIterations) = 0 then begin
				if nrOfIterations = 0 then
					oldBest := bestOneTree
				else begin
					u := (bestOneTree / oldBest - 1.0);
					oldBest := bestOneTree;
					cancel := userAbort;
				end;
			end;
			nrOfIterations := nrOfIterations + 1;
		until cancel | tourFound | not possible | (u < minImprove) | (oneTreeLength > best);
		possible := not cancel & possible & not tourFound & (oneTreeLength <= best);
		if possible then begin
			for i := 0 to desc^.nrOfPoints - 1 do
				points^[i].lambda := points^[i].saveLambda;
			for i := 0 to desc^.firstExcl - 1 do
				edges^[sortedEdges^[i]].updLen := edges^[sortedEdges^[i]].length + points^[edges^[sortedEdges^[i]].fromP].lambda + points^[edges^[sortedEdges^[i]].toP].lambda;
			InsertionSortEdgesByLength(desc, edges, sortedEdges);
			findOneTree(desc, points, edges, sortedEdges, oneTree, oneTreeLength, tourFound, possible);
			if oneTreeLength < exceed then
				oneTreeLength := exceed;
		end;
	end;

procedure LowerBounding (desc: DescP; points: PointListP; edges: EdgeListP; sortedEdges: EdgeIndListP; oneTree: PointIndListP; tour: PointIndListP; var oneTreeLength: real; var edge1, edge2: integer; var tourFound, possible, notFirst: boolean; var best: real; exceed: real; var cancel: boolean);
	begin
		IterateOneTree(desc, points, edges, sortedEdges, oneTree, oneTreeLength, tourFound, possible, best, exceed, cancel);
		if tourFound & (oneTreeLength < best) then begin
			buildTourFromOneTree(desc, edges, oneTree, tour);
			best := oneTreeLength;
		end;
		if possible then
			getBranchingEdges(desc, points, edges, oneTree, edge1, edge2, notFirst)
		else begin
			edge1 := -1;
			edge2 := -1;
		end;
{$IFC AnimationEnabled AND ATravelingSalesman }
		if not cancel & animationFlag[ATravelingSalesmanItem] then begin
			showOneTree(desc, points, edges, oneTree);
			displayInfoString(stringOf(oneTreeLength : 12 : 3, ' = LB ', best : 12 : 3, ' = UB ', exceed : 12 : 3, ' = GLB '));
			waitForClick(animationFlag[ATravelingSalesmanItem]);
			showOneTree(desc, points, edges, oneTree);
		end;
{$ENDC }
(*displayInfoString(stringOf(oneTreeLength : 12 : 3, ' = LB ', best : 12 : 3, ' = UB ', exceed : 12 : 3, ' = GLB ')); *)
	end;

procedure InclExcl (desc: DescP; points: PointListP; edges: EdgeListP; sortedEdges: EdgeIndListP; oneTree: PointIndListP; var oneTreeLength: real; var cancel: boolean);
	var
		i, k, l: integer;
		tourFound, possible: boolean;
	begin
		QuickSortEdgesByLength(desc, edges, sortedEdges);
		desc^.lastIncl := 0;
		i := 0;
		while not cancel & (i < desc^.nrOfEdges) do begin
			k := sortedEdges^[i];
			for l := i downto 1 do
				sortedEdges^[l] := sortedEdges^[l - 1];
			sortedEdges^[0] := k;
			findOneTree(desc, points, edges, sortedEdges, oneTree, oneTreeLength, tourFound, possible);
			edges^[k].guessLen := oneTreeLength;
			for l := 0 to i - 1 do
				sortedEdges^[l] := sortedEdges^[l + 1];
			sortedEdges^[i] := k;
			i := i + 1;
			cancel := userAbort;
		end;
		desc^.lastIncl := -1;
		QuickSortEdgesByGuessLength(desc, edges, sortedEdges);
	end;

procedure SetEdgeList (var upperBound: real; desc: DescP; points: PointListP; edges: EdgeListP; sortedEdges: EdgeIndListP);
	var
		i, j, k: integer;
	begin
		upperBound := incrVal * upperBound;
		i := desc^.nrOfEdges - 1;
		j := 0;
		repeat
			k := (i + j) div 2;
			if edges^[sortedEdges^[k]].guessLen <= upperBound then
				j := k + 1
			else
				i := k;
		until i = j;
		desc^.firstExcl := k + 1;
{$IFC AnimationEnabled AND ATravelingSalesman }
		if animationFlag[ATravelingSalesmanItem] then begin
			for i := 0 to desc^.nrOfEdges - 1 do
				drawLineSegment(points^[edges^[sortedEdges^[i]].fromP].x, points^[edges^[sortedEdges^[i]].fromP].y, points^[edges^[sortedEdges^[i]].toP].x, points^[edges^[sortedEdges^[i]].toP].y);
			for i := desc^.firstExcl to desc^.nrOfEdges - 1 do
				drawLineSegment(points^[edges^[sortedEdges^[i]].fromP].x, points^[edges^[sortedEdges^[i]].fromP].y, points^[edges^[sortedEdges^[i]].toP].x, points^[edges^[sortedEdges^[i]].toP].y);
			displayInfoString(stringOf('Upper bound = ', upperBound : 12 : 3, '   =>   ', desc^.firstExcl - 1, ' valid edges'));
			waitForClick(animationFlag[ATravelingSalesmanItem]);
			for i := 0 to desc^.firstExcl - 1 do
				drawLineSegment(points^[edges^[sortedEdges^[i]].fromP].x, points^[edges^[sortedEdges^[i]].fromP].y, points^[edges^[sortedEdges^[i]].toP].x, points^[edges^[sortedEdges^[i]].toP].y);
		end;
{$ENDC }
	end;

function pointVector.TSP1TreeBranchAndBound: polygon2d;
	var
		shPoly, shPoly1: polygon2d;
		points: PointListH;
		edges: EdgeListH;
		sortedEdges: EdgeIndListH;									(* contains current MST, included and excluded edges, free edges	*)
		oneTree, tour: PointIndListH;
		desc: DescH;
		best, oneTreeLength, tourLength: real;
		i, maxOneTP, inclCount, exclCount: integer;
		tourFound, possible: boolean;
		root: TSPNodeP;
		cancel: boolean;

	procedure deleteTree (var node: TSPNodeP);
		begin
			if node^.son <> nil then
				deleteTree(node^.son);
			if node^.brother <> nil then
				deleteTree(node^.brother);
			dispose(node);
			node := nil;
		end;

	procedure expandBest (node: TSPNodeP; var inclCount, exclCount: integer; var cancel: boolean);
		var
			bestSon, curSon: TSPNodeP;
			inclSave, exclSave: integer;
			dummy: boolean;
			oldLow: real;
		begin
			oldLow := node^.lowerBound;
			if node^.son <> nil then begin
				bestSon := node^.son;
				curSon := bestSon;
				while (curSon^.brother <> nil) do begin
					if curSon^.brother^.lowerBound < bestSon^.lowerBound then
						bestSon := curSon^.brother;
					curSon := curSon^.brother;
				end;
				case bestSon^.myNumber of
					1:  begin
						dummy := includeEdge(desc^, points^, edges^, sortedEdges^, node^.edge1, inclCount, exclCount);
						dummy := includeEdge(desc^, points^, edges^, sortedEdges^, node^.edge2, inclCount, exclCount);
					end;
					2:  begin
						dummy := includeEdge(desc^, points^, edges^, sortedEdges^, node^.edge2, inclCount, exclCount);
						if not node^.notFirst then
							dummy := excludeEdge(desc^, edges^, sortedEdges^, node^.edge1, exclCount);
					end;
					3:  begin
						dummy := excludeEdge(desc^, edges^, sortedEdges^, node^.edge2, exclCount);
					end;
				end;
				expandBest(bestSon, inclCount, exclCount, cancel);
				bestSon := node^.son;
				node^.lowerBound := bestSon^.lowerBound;
				bestSon := bestSon^.brother;
				if bestSon^.lowerBound < node^.lowerBound then
					node^.lowerBound := bestSon^.lowerBound;
				bestSon := bestSon^.brother;
				if (bestSon <> nil) & (bestSon^.lowerBound < node^.lowerBound) then
					node^.lowerBound := bestSon^.lowerBound;
			end
			else begin
				inclSave := 0;
				exclSave := 0;
				if (node^.edge1 = -1) then begin
					LowerBounding(desc^, points^, edges^, sortedEdges^, oneTree^, tour^, node^.lowerBound, node^.edge1, node^.edge2, tourFound, possible, node^.notFirst, best, oldLow, cancel);
					if not possible then
						node^.lowerBound := 2.0 * best;
				end
				else if not node^.notFirst then begin
					newTSPNode(node^.son);

					curSon := node^.son;
					curSon^.myNumber := 1;
					curSon^.father := node;
					dummy := includeEdge(desc^, points^, edges^, sortedEdges^, node^.edge1, inclSave, exclSave);
					dummy := includeEdge(desc^, points^, edges^, sortedEdges^, node^.edge2, inclSave, exclSave);
					LowerBounding(desc^, points^, edges^, sortedEdges^, oneTree^, tour^, curSon^.lowerBound, curSon^.edge1, curSon^.edge2, tourFound, possible, curSon^.notFirst, best, oldLow, cancel);
					if not possible then
						curSon^.lowerBound := 2.0 * best;
					rebuildEdgeList(desc^, points^, edges^, sortedEdges^, inclSave, exclSave);
					node^.lowerBound := curSon^.lowerBound;

					newTSPNode(curSon^.brother);

					curSon := curSon^.brother;
					curSon^.myNumber := 2;
					curSon^.father := node;
					dummy := includeEdge(desc^, points^, edges^, sortedEdges^, node^.edge2, inclSave, exclSave);
					dummy := excludeEdge(desc^, edges^, sortedEdges^, node^.edge1, exclSave);
					LowerBounding(desc^, points^, edges^, sortedEdges^, oneTree^, tour^, curSon^.lowerBound, curSon^.edge1, curSon^.edge2, tourFound, possible, curSon^.notFirst, best, oldLow, cancel);
					if not possible then
						curSon^.lowerBound := 2.0 * best;
					rebuildEdgeList(desc^, points^, edges^, sortedEdges^, inclSave, exclSave);
					if curSon^.lowerBound < node^.lowerBound then
						node^.lowerBound := curSon^.lowerBound;

					newTSPNode(curSon^.brother);

					curSon := curSon^.brother;
					curSon^.myNumber := 3;
					curSon^.father := node;
					dummy := excludeEdge(desc^, edges^, sortedEdges^, node^.edge2, exclSave);
					LowerBounding(desc^, points^, edges^, sortedEdges^, oneTree^, tour^, curSon^.lowerBound, curSon^.edge1, curSon^.edge2, tourFound, possible, curSon^.notFirst, best, oldLow, cancel);
					if not possible then
						curSon^.lowerBound := 2.0 * best;
					rebuildEdgeList(desc^, points^, edges^, sortedEdges^, inclSave, exclSave);
					if curSon^.lowerBound < node^.lowerBound then
						node^.lowerBound := curSon^.lowerBound;
				end
				else begin
					newTSPNode(node^.son);

					curSon := node^.son;
					curSon^.myNumber := 2;
					dummy := includeEdge(desc^, points^, edges^, sortedEdges^, node^.edge2, inclSave, exclSave);
					LowerBounding(desc^, points^, edges^, sortedEdges^, oneTree^, tour^, curSon^.lowerBound, curSon^.edge1, curSon^.edge2, tourFound, possible, curSon^.notFirst, best, oldLow, cancel);
					if not possible then
						curSon^.lowerBound := 2.0 * best;
					rebuildEdgeList(desc^, points^, edges^, sortedEdges^, inclSave, exclSave);
					node^.lowerBound := node^.son^.lowerBound;

					newTSPNode(curSon^.brother);
					curSon := curSon^.brother;
					curSon^.myNumber := 3;
					dummy := excludeEdge(desc^, edges^, sortedEdges^, node^.edge2, exclSave);
					LowerBounding(desc^, points^, edges^, sortedEdges^, oneTree^, tour^, curSon^.lowerBound, curSon^.edge1, curSon^.edge2, tourFound, possible, curSon^.notFirst, best, oldLow, cancel);
					if not possible then
						curSon^.lowerBound := 2.0 * best;
					rebuildEdgeList(desc^, points^, edges^, sortedEdges^, inclSave, exclSave);
					if curSon^.lowerBound < node^.lowerBound then
						node^.lowerBound := curSon^.lowerBound;
				end;
			end;
		end;

	procedure propagate (node: TSPNodeP; bound: real);
		begin
			if node <> nil then begin
				propagate(node^.brother, bound);
				propagate(node^.son, bound);
				node^.lowerBound := bound;
			end;
		end;

	begin (* pointVector.TSP1TreeBranchAndBound *)
		if (length > MaxPts) | (length < 3) then begin
			nonFatalError('Too few or too many Points ... sorry.');
			TSP1TreeBranchAndBound := nil;
		end
		else begin
			cancel := false;
			edges := EdgeListH(dynamicArray(length * (length - 1) div 2, sizeOf(EdgeT)));
			sortedEdges := EdgeIndListH(dynamicArray(length * (length - 1) div 2, sizeOf(integer)));
			points := PointListH(dynamicArray(length, sizeOf(PointT)));
			oneTree := PointIndListH(dynamicArray(length, sizeOf(integer)));
			tour := PointIndListH(dynamicArray(length, sizeOf(longint)));
			new(desc);
			failNil(Handle(desc));
			new(desc^);
			failNil(Handle(desc^));
			HLock(Handle(edges));
			HLock(Handle(sortedEdges));
			HLock(Handle(points));
			HLock(Handle(oneTree));
			HLock(Handle(tour));

			shPoly1 := polygon2d(convertToPolygon);
			initTSP(shPoly1.length, desc^, points^, edges^, sortedEdges^, oneTree^, tour^, pointVector(shPoly1));
			QuickSortEdgesByLength(desc^, edges^, sortedEdges^);
			best := 0.0;
			i := 0;
			while not cancel & (i < desc^^.nrOfPoints) do begin
				desc^^.oneTreePoint := i;
				findOneTree(desc^, points^, edges^, sortedEdges^, oneTree^, oneTreeLength, tourFound, possible);
				if oneTreeLength > best then begin
					maxOneTP := i;
					best := oneTreeLength;
				end;
				i := i + 1;
				cancel := userAbort;
			end;
			desc^^.oneTreePoint := maxOneTP;
			best := 2.0 * oneTreeLength + 0.5;
			newTSPNode(root);

			if not cancel then begin
{$IFC AnimationEnabled AND ATravelingSalesman }
				if animationFlag[ATravelingSalesmanItem] then begin
					displayInfoString(stringOf('Iterating 1-Tree, all edges are considered'));
				end;
{$ENDC }
				LowerBounding(desc^, points^, edges^, sortedEdges^, oneTree^, tour^, root^.lowerBound, root^.edge1, root^.edge2, tourFound, possible, root^.notFirst, best, 0.0, cancel);
			end;
			if not cancel & not tourFound then begin
{$IFC AnimationEnabled AND ATravelingSalesman }
				if animationFlag[ATravelingSalesmanItem] then
					displayInfoString(stringOf('Lower bound for each edge is computed... '));
{$ENDC }
				best := root^.lowerBound;
				InclExcl(desc^, points^, edges^, sortedEdges^, oneTree^, oneTreeLength, cancel);
				tour^^[0] := 0;
				tour^^[1] := 0;
				while not cancel & not tourFound do begin
					SetEdgeList(best, desc^, points^, edges^, sortedEdges^);
					while not cancel & (root^.lowerBound < best) do begin
						inclCount := 0;
						exclCount := 0;
						expandBest(root, inclCount, exclCount, cancel);
						rebuildEdgeList(desc^, points^, edges^, sortedEdges^, inclCount, exclCount);
					end;
					tourFound := (tour^^[0] <> 0) | (tour^^[1] <> 0);
					if not tourFound then
						propagate(root, 0.0);
				end;
			end;
			deleteTree(root);
			if not cancel then begin
				tourLength := best;
(* construct the polygon *)
				new(shPoly);
				failNil(Handle(shPoly));
				currentVectorLength := length;
				shPoly.init;
				for i := 0 to length - 1 do begin
					shPoly.append(shPoly1.elements^^[tour^^[i] + 1].duplicateAll);
				end;
				TSP1TreeBranchAndBound := shPoly;
			end
			else
				TSP1TreeBranchAndBound := nil;
			shPoly1.freeAll;
			dispose(desc^);
			dispose(desc);
{$IFC THINK_PASCAL }
			disposHandle(Handle(edges));
			disposHandle(Handle(sortedEdges));
			disposHandle(Handle(points));
			disposHandle(Handle(oneTree));
			disposHandle(Handle(tour));
{$ELSEC }
			DisposeHandle(Handle(edges));
			DisposeHandle(Handle(sortedEdges));
			DisposeHandle(Handle(points));
			DisposeHandle(Handle(oneTree));
			DisposeHandle(Handle(tour));
{$ENDC }
		end;
	end;

end.
