(*	XYZ GeoBench (eXperimental geometrY Zurich):		*)
(*	An object oriented, interactive workbench for		*)
(*	experimental geometric computation.					*)
(*														*)
(*	 1999 by Peter Schorn, peter.schorn@acm.org		*)
(*	Culmannstrasse 77, CH-8006 Zrich, Switzerland		*)
(*	This software may be copied freely for educational	*)
(*	and research purposes only. For commercial use,		*)
(*	please contact the author.							*)
(*														*)
(*	Version 5.0.5: 30-Sep-1999							*)

(* Copyright 1992 by Bert Leukens						*)
(* Informatik, ETH, CH-8092 Zrich, Switzerland			*)
(* Version 1.0:  July-29-92								*)

unit triangulatePlaneSweep;

interface

uses
	triangulation;

{$SETC    ATriangulation = TRUE }

type

	pointType = (closeSector, continueSector, openSector);
	pointInfoRec = record
			p: point2d;									(* reference to original 2D point (source) *)
			typ: pointType;								(* type of this point *)
			poly: seqIndex0;							(* index to polygon which this point belongs to *)
		end;
	pointInfoH = ^pointInfoP;
	pointInfoP = ^pointInfoArray;
	pointInfoArray = array[seqIndex] of pointInfoRec;

	triangulationPolygonInfoDirectory = object(obj)
			t: triangulation;							(* reference to triangulation which has to be created *)
			pInfo: pointInfoH;							(* dynamic array of point information *)
			events: seqIndexArrayH;					(* dynamic array of sorted indices to points *)
			firstInPoly: seqIndexArrayH;			(* dynamic array of indices to first point in every polygon *)
			curPolygon: seqIndex;						(* index to current polygon *)
			maxPolygons: seqIndex0;					(* max. number of polygons *)
			maxVertices: seqIndex0;					(* max. number of vertices *)
			maxEdges: seqIndex0;						(* max. number of edges *)
			maxTriangles: seqIndex0;				(* max. number of triangles *)
			procedure init;
			override;
			procedure free;
			override;

			procedure initialize (tri: triangulation; (* *)
										vertices, outerPolygons, innerPolygons: seqIndex0);
		(* defines size of points, edges and triangles *)
			procedure addNewPoint (p: point2d);
		(* generate new point from source point p *)
			procedure appendPolygon (polygon: polygon2d);
		(* append polygon2d to triangulation *)
			procedure appendPolygons (v: vector);
		(* append vector of polygon2d to triangulation *)
			procedure sort (l, r: seqIndex);
		(* sort events in range l to r *)
			procedure nextPoints (id: seqIndex; var left, right: seqIndex);
		(* returns index of neigbour points in point array left and right of this point (id) on the same polygon *)
			function getEdge (firstVertex, secondVertex: seqIndex): graphEdgeTriangulation;
		(* returns common edge of first and second vertex *)
		end;

	chainElementH = ^chainElementP;
	chainElementP = ^chainElement;
	chainElement = record
			next: array[directionType] of chainElementH;
			edge: graphEdgeTriangulation;
		end;

	triangulationChainPool = object(obj)
			pool: chainElementH;
			procedure init;
			override;
			procedure freeAll;
			override;
			function getChainElement: chainElementH;
			procedure putChainElement (e: chainElementH);
		end;

	triangulationSector = object(obj)
			nextVertex, currentVertex: array[directionType] of seqIndex;
			s: seqIndex;
			chainEnd: array[directionType] of chainElementH;
			start: chainElementH; (* one of the two vertices of the edge has index s *)
		end;

	triangulationSweepLine = object(obj)
			t: triangulation;							(* reference to global or local lists of points, edges and triangles *)
			polygonInfo: triangulationPolygonInfoDirectory;

			sectors: dictionary;
			currentEvent: seqIndex0; 				(* index to current event in event array *)
			currentPoint: seqIndex0;				(* index to current event in point array *)
			chainPool: triangulationChainPool;	(* reference to global chain pool *)

			procedure init;
			override;
			procedure free;
			override;

			procedure initialize (polyInfo: triangulationPolygonInfoDirectory; tri: triangulation);
			function findSector (exactly: boolean; var where: reference; var direction: directionType): boolean;
		(* currentPoint and currentEvent must be set ! *)
			procedure startSweep;
			procedure nextEvent;
			function openSector: triangulationSector;
			procedure continueSector (sector: triangulationSector);
			procedure closeSector (where: reference);
			procedure mergeSectors (lSector, rSector: reference);
			function splitSector (where: reference): triangulationSector;
			function colinearCheck (p, q: seqIndex): boolean;
			procedure updateChain (sector: triangulationSector);
{$IFC AnimationEnabled AND ATriangulation }
			procedure animateNextEvent;
			procedure animateNewEdge;
{$ENDC }
		end;


const
	closeSectorC = closeSector;
	continueSectorC = continueSector;
	openSectorC = openSector;


implementation

procedure triangulationChainPool.init;
	begin
		pool := chainElementH(newHandleFailNil(sizeOf(chainElement)));
		pool^^.next[right] := nil
	end;

procedure triangulationChainPool.freeAll;
	var
		e: chainElementH;
	begin
		while pool <> nil do begin
			e := pool;
			pool := pool^^.next[right];
			DisposeHandle(Handle(e))
		end;
		inherited freeAll;
	end;

function triangulationChainPool.getChainElement: chainElementH;
	begin
		getChainElement := pool;
		pool := pool^^.next[right];
		if pool = nil then
			init;
	end;

procedure triangulationChainPool.putChainElement (e: chainElementH);
	begin
		e^^.next[right] := pool;
		pool := e
	end;

procedure triangulationPolygonInfoDirectory.init;
	var
		vertices: seqIndex0;
	begin
		curPolygon := 1;
		t.points := newVector(maxVertices);
		t.edges := newVector(maxEdges);
		t.triangles := newVector(maxTriangles);

		firstInPoly := seqIndexArrayH(dynamicArray(maxPolygons + 1, sizeOf(seqIndex)));
		failNil(Handle(firstInPoly));
		firstInPoly^^[maxPolygons + 1] := maxVertices + 1;	(* define first polygon points sentinel *)

		pInfo := pointInfoH(dynamicArray(maxVertices + 1, sizeOf(pointInfoRec)));
		failNil(Handle(pInfo));
		with pInfo^^[maxVertices + 1] do begin					(* define point info sentinel *)
			p := nil;
			poly := 0
		end;

		events := seqIndexArrayH(dynamicArray(maxVertices, sizeOf(seqIndex)));
		failNil(Handle(events));
		for vertices := 1 to maxVertices do
			events^^[vertices] := vertices;
	end;

procedure triangulationPolygonInfoDirectory.free;
	begin
		DisposeHandle(Handle(pInfo));
		DisposeHandle(Handle(events));
		DisposeHandle(Handle(firstInPoly));
		inherited free;
	end;

procedure triangulationPolygonInfoDirectory.initialize (tri: triangulation; (* *)
								vertices, outerPolygons, innerPolygons: seqIndex0);
	begin
		maxVertices := vertices;
		maxPolygons := innerPolygons + outerPolygons;		(* total amount of polygons *)
		maxEdges := 3 * (maxVertices + innerPolygons - outerPolygons) - maxVertices;
		maxTriangles := 2 * (maxVertices + innerPolygons - outerPolygons) - maxVertices;

		t := tri;
		init;
	end;

procedure triangulationPolygonInfoDirectory.addNewPoint (p: point2d);
	begin
		t.points.append(p.duplicate);		(* generate new point of triangulation *)
	end;

procedure triangulationPolygonInfoDirectory.appendPolygon (polygon: polygon2d);
	var
		l, offset, rangeMin, rangeMax: seqIndex0;
		e: graphEdgeTriangulation;
		next: boolean;
		signumDx, signumDy: array[boolean] of integer;

	function classifyPoint: pointType;
		begin
			if signumDx[next] = signumDx[not next] then
				classifyPoint := continueSector
			else if (signumDx[next] = 1) & (signumDx[not next] = -1) then
				classifyPoint := openSector
			else if (signumDx[next] = -1) & (signumDx[not next] = 1) then
				classifyPoint := closeSector
			else if (signumDx[not next] = +1) & (signumDy[next] = -1) then
				classifyPoint := closeSector
			else if (signumDx[not next] = -1) & (signumDy[next] = +1) then
				classifyPoint := openSector
			else if (signumDx[next] = -1) & (signumDy[not next] = +1) then
				classifyPoint := closeSector
			else if (signumDx[next] = +1) & (signumDy[not next] = -1) then
				classifyPoint := openSector
			else
				classifyPoint := continueSector;
		end;

	begin
		if curPolygon > maxPolygons then
			fatalError('Polygon overflow');
		offset := t.points.length;
		rangeMin := offset + 1;
		rangeMax := offset + polygon.length;
		firstInPoly^^[curPolygon] := rangeMin;
		for l := rangeMin to rangeMax do begin
			with pInfo^^[l] do begin
				p := point2d(polygon.elements^^[l - offset]);	(* point reference to source of point *)
				poly := curPolygon;
				addNewPoint(p);
				e := t.addNewEdge(l, l + 1);
			end;
{$IFC AnimationEnabled }
			advanceStopWatchHands;
{$ENDC }
		end;
		e.endVertex := firstInPoly^^[curPolygon];
		curPolygon := curPolygon + 1;

		next := true;				(* define all point types *)
		with pInfo^^[rangeMin].p do begin
			signumDx[false] := sign(getX - pInfo^^[rangeMax].p.getX);
			signumDy[false] := sign(getY - pInfo^^[rangeMax].p.getY)
		end;
		for l := rangeMin to rangeMax - 1 do begin
			with pInfo^^[l] do begin
				signumDx[next] := sign(pInfo^^[l + 1].p.getX - p.getX);
				signumDy[next] := sign(pInfo^^[l + 1].p.getY - p.getY);
				typ := classifyPoint;
				next := not next;
			end;
{$IFC AnimationEnabled }
			advanceStopWatchHands;
{$ENDC }
		end;
		with pInfo^^[rangeMax] do begin
			signumDx[next] := sign(pInfo^^[rangeMin].p.getX - p.getX);
			signumDy[next] := sign(pInfo^^[rangeMin].p.getY - p.getY);
			typ := classifyPoint;
		end;
	end;

procedure triangulationPolygonInfoDirectory.appendPolygons (v: vector);
	var
		l: seqIndex;
	begin
		with v do
			for l := 1 to length do
				appendPolygon(polygon2d(elements^^[l]));
	end;

procedure triangulationPolygonInfoDirectory.sort (l, r: seqIndex);
	var
		s: seqIndexArrayH;

	function lessThan (var p, q: seqIndex): boolean;
		var
			qPoint: point2d;
		begin
			with pInfo^^[p].p do begin
				qPoint := pInfo^^[q].p;
				lessThan := xLessThan(qPoint) | (xEqual(qPoint) & yLessThan(qPoint))
			end
		end;

	procedure insertionSort (l: seqIndex; r: seqIndex0);
		var
			i: seqIndex;
			j: seqIndex0;
			y: seqIndex;
		begin
			for i := l + 1 to r do begin	(* insertionsort *)
				j := i;
				y := s^^[j];
				while (j > l) & lessThan(y, s^^[j - 1]) do begin
					s^^[j] := s^^[j - 1];
					j := j - 1;
				end;
				s^^[j] := y;
			end;
		end;

	procedure quickSort (l: seqIndex; r: seqIndex0);
		const
			cutoff = 10;
		var
			i: seqIndex;
			j: seqIndex0;
			x, y: seqIndex;

		begin
			while cutoff < r - l do begin		(* quicksort *)
				i := l;
				j := r;
				x := s^^[(l + r) div 2];
				repeat
					while lessThan(s^^[i], x) do
						i := i + 1;
					while lessThan(x, s^^[j]) do
						j := j - 1;
					if i <= j then begin
						y := s^^[i];
						s^^[i] := s^^[j];
						s^^[j] := y;
						i := i + 1;
						j := j - 1;
					end;
				until i > j;
				if j - l > r - i then begin
					quickSort(i, r);
					r := j;
				end
				else begin
					quickSort(l, j);
					l := i;
				end;
			end;
		end; (* quickSort *)

	begin
		s := events;
		quickSort(l, r);
		insertionSort(l, r)
	end;

procedure triangulationPolygonInfoDirectory.nextPoints (id: seqIndex; var left, right: seqIndex);
	var
		thisPoly, first: seqIndex;
	begin
		thisPoly := pInfo^^[id].poly;
		first := firstInPoly^^[thisPoly];
		right := id + 1;
		if id = first then
			left := firstInPoly^^[thisPoly + 1] - 1
		else if thisPoly <> pInfo^^[right].poly then begin
			left := id - 1;
			right := first
		end
		else
			left := id - 1
	end;


function triangulationPolygonInfoDirectory.getEdge (firstVertex, secondVertex: seqIndex): graphEdgeTriangulation;
	var
		e: graphEdgeTriangulation;
	begin
		with t.edges do begin
			e := graphEdgeTriangulation(elements^^[firstVertex]);
			if (secondVertex <> e.startVertex) & (secondVertex <> e.endVertex) then
				e := graphEdgeTriangulation(elements^^[secondVertex])
		end;
		getEdge := e
	end;



procedure triangulationSweepLine.init;
	begin
		t := nil;
		polygonInfo := nil;
		new(chainPool);
		failNil(Handle(chainPool));
		chainPool.init;
		sectors := dictionary(newInitObj(currentDictionaryType));
		failNil(Handle(sectors));
		sectors.init;
	end;


procedure triangulationSweepLine.free;
	begin
		chainPool.freeAll;
		sectors.freeAll;
		inherited free;
	end;


procedure triangulationSweepLine.initialize (polyInfo: triangulationPolygonInfoDirectory; tri: triangulation);
	begin
		t := tri;
		polygonInfo := polyInfo;
		currentEvent := 0;
	end;

function triangulationSweepLine.findSector (exactly: boolean;(* *)
								var where: reference; var direction: directionType): boolean;
	function strongCompare (a, b: obj): relationObj;
		var
			p: point2d;
			sector: triangulationSector;
		begin
			p := point2d(a);
			sector := triangulationSector(b);
			with polygonInfo, sector do
				if (p = pInfo^^[nextVertex[left]].p) | (p = pInfo^^[nextVertex[right]].p) then
					strongCompare := equalObj
				else if p.whichSideSign(pInfo^^[currentVertex[left]].p, pInfo^^[nextVertex[left]].p) = -1 then
					strongCompare := greaterThanObj
				else
					strongCompare := lessThanObj
		end;

	function weekCompare (a, b: obj): relationObj;
		var
			p: point2d;
			sector: triangulationSector;
		begin
			p := point2d(a);
			sector := triangulationSector(b);
			with polygonInfo, sector do
				if p.whichSideSign(pInfo^^[currentVertex[left]].p, pInfo^^[nextVertex[left]].p) = -1 then
					weekCompare := greaterThanObj
				else if p.whichSideSign(pInfo^^[currentVertex[right]].p, pInfo^^[nextVertex[right]].p) = 1 then
					weekCompare := lessThanObj
				else
					weekCompare := equalObj
		end;

	begin
		if exactly then
			findSector := sectors.find(obj(polygonInfo.pInfo^^[currentPoint].p), strongCompare, where, direction)
		else
			findSector := sectors.find(obj(polygonInfo.pInfo^^[currentPoint].p), weekCompare, where, direction)
	end;

procedure triangulationSweepLine.startSweep;
	var
		where: reference;
		direction: directionType;
	begin
		nextEvent;
		where := sectors.insert(obj(openSector), nil, left);
		nextEvent;
		while currentEvent > 0 do begin
			case polygonInfo.pInfo^^[currentPoint].typ of
				closeSectorC: 
					if findSector(true, where, direction) then
						closeSector(where);
				continueSectorC: 
					if findSector(true, where, direction) then
						continueSector(triangulationSector(sectors.getObject(where)));
				openSectorC: 
					if findSector(false, where, direction) then
						where := sectors.insert(obj(splitSector(where)), where, left)
					else
						where := sectors.insert(obj(openSector), where, direction)
			end;
			nextEvent
		end;
	end;

procedure triangulationSweepLine.nextEvent;
	begin
		if currentEvent < t.points.length then begin
			currentEvent := currentEvent + 1;
			currentPoint := polygonInfo.events^^[currentEvent];
{$IFC AnimationEnabled AND ATriangulation }
			animateNextEvent;
			advanceStopWatchHands;
			if userAbort then
				currentEvent := 0;
{$ENDC }
		end
		else
			currentEvent := 0;
	end;


function triangulationSweepLine.openSector: triangulationSector;
	var
		sector: triangulationSector;
		nextLeft, nextRight: seqIndex;
	begin
		new(sector);
		failNil(Handle(sector));
		sector.init;
		with polygonInfo, sector do begin
			s := currentPoint;
			currentVertex[left] := s;
			currentVertex[right] := s;
			chainEnd[left] := nil;
			start := nil;
			chainEnd[right] := nil;
			nextPoints(s, nextLeft, nextRight);
			if pInfo^^[nextLeft].p.whichSideX(pInfo^^[s].p, pInfo^^[nextRight].p) > 0 then begin
				nextVertex[left] := nextRight;
				nextVertex[right] := nextLeft;
			end
			else begin
				nextVertex[left] := nextLeft;
				nextVertex[right] := nextRight;
			end;
		end;
		openSector := sector;
	end;


procedure triangulationSweepLine.continueSector (sector: triangulationSector);
	var
		e: graphEdgeTriangulation;
		ce: chainElementH;
		dir, otherdir: directionType;
		nextRight: seqIndex;
	begin
		with polygonInfo, sector do begin
			if currentPoint = nextVertex[left] then begin
				dir := left;
				otherdir := right
			end
			else begin
				dir := right;
				otherdir := left
			end;
			e := getEdge(nextVertex[dir], currentVertex[dir]);
			s := currentPoint;
			nextPoints(s, nextVertex[dir], nextRight);
			if nextVertex[dir] = currentVertex[dir] then
				nextVertex[dir] := nextRight;
			currentVertex[dir] := s;
			ce := chainPool.getChainElement;
			with ce^^ do begin
				edge := e;
				next[dir] := nil;
				next[otherdir] := chainEnd[dir]
			end;
			if chainEnd[dir] = nil then
				chainEnd[otherdir] := ce
			else
				chainEnd[dir]^^.next[dir] := ce;
			start := ce;
			chainEnd[dir] := ce;
			updateChain(sector);
		end;
	end;


procedure triangulationSweepLine.closeSector (where: reference);
	var
		e: graphEdgeTriangulation;
		ce: chainElementH;
		sector: triangulationSector;
		other: reference;
	begin
		sector := triangulationSector(sectors.getObject(where));
		with polygonInfo, sector do
			if nextVertex[left] = nextVertex[right] then begin
				ce := chainEnd[right];
				chainEnd[right] := chainEnd[right]^^.next[left];
				if chainEnd[right] <> nil then begin
					chainEnd[right]^^.next[right] := nil;
					continueSector(sector);
					e := chainEnd[left]^^.edge
				end
				else
					e := getEdge(nextVertex[left], currentVertex[left]);
				if self.t.triangles.length < maxTriangles then
					self.t.addNewTriangle(getEdge(nextVertex[right], currentVertex[right]), ce^^.edge, e)
				else
					fatalError('Invalid initialisation! Selection order could be wrong');
				chainPool.putChainElement(ce);
				if chainEnd[right] <> nil then
					chainPool.putChainElement(chainEnd[right]);
				sector.free;
				sectors.delete(where);
			end
			else begin
				if currentPoint = nextVertex[right] then begin
					other := sectors.next(where, left);
					if other = nil then
						fatalError('Sector not found');
					mergeSectors(where, other)
				end
				else begin
					other := sectors.next(where, right);
					if other = nil then
						fatalError('Sector not found');
					mergeSectors(other, where)
				end
			end
	end;


procedure triangulationSweepLine.mergeSectors (lSector, rSector: reference);
	var
		sL, sR: triangulationSector;
	begin
		sL := triangulationSector(sectors.getObject(lSector));
		sR := triangulationSector(sectors.getObject(rSector));
		continueSector(sL);
		continueSector(sR);
		with sL do begin
			start := sR.chainEnd[left];
			start^^.next[left] := chainEnd[right];
			chainEnd[right]^^.next[right] := start;
			chainEnd[right] := sR.chainEnd[right];
			nextVertex[right] := sR.nextVertex[right];
			currentVertex[right] := sR.currentVertex[right];
		end;
		sR.free;
		sectors.delete(rSector);
	end;


function triangulationSweepLine.splitSector (where: reference): triangulationSector;
	var
		sector, newSector: triangulationSector;
		e: graphEdgeTriangulation;
		ce: chainElementH;
		h: seqIndex;
	begin
		sector := triangulationSector(sectors.getObject(where));
		newSector := openSector;
		with sector do begin
			if self.t.edges.length < polygonInfo.maxEdges then
				e := t.addNewEdge(newSector.s, s) (* insert new edge *)
			else
				fatalError('Invalid initialisation! Selection order could be wrong');
{$IFC AnimationEnabled AND ATriangulation }
			animateNewEdge;
{$ENDC }

			if (start = chainEnd[right]) & (s = currentVertex[right]) then
				newSector.start := nil
			else begin
				newSector.start := start;
				newSector.chainEnd[right] := chainEnd[right];
				start := newSector.start^^.next[left]
			end;
			ce := chainPool.getChainElement;
			with ce^^ do begin
				edge := e;
				next[left] := start;
				next[right] := nil
			end;
			if start <> nil then
				start^^.next[right] := ce
			else
				chainEnd[left] := ce;
			chainEnd[right] := ce;
			start := ce;
		end; (* with *)
		with newSector do begin
			ce := chainPool.getChainElement;
			with ce^^ do begin
				edge := e;
				next[right] := start;
				next[left] := nil
			end;
			if start <> nil then
				start^^.next[left] := ce
			else
				chainEnd[right] := ce;
			chainEnd[left] := ce;
			start := ce;
			sector.s := s;
			currentVertex[right] := sector.currentVertex[right];
			sector.currentVertex[right] := s;
			h := nextVertex[right];
			nextVertex[right] := sector.nextVertex[right];
			sector.nextVertex[right] := nextVertex[left];
			nextVertex[left] := h;
		end; (* with *)
		updateChain(sector);
		updateChain(newSector);
		splitSector := newSector
	end;

{$IFC NOT THINK_PASCAL }
{$PUSH }
{$PRAGMAC warn_unusedarg off }
{$ENDC }
function triangulationSweepLine.colinearCheck (p, q: seqIndex): boolean;
	begin
		colinearCheck := false
	end;
{$IFC NOT THINK_PASCAL }
{$POP }
{$ENDC }

procedure triangulationSweepLine.updateChain (sector: triangulationSector);
	var
		e: graphEdgeTriangulation;
		ce: chainElementH;
		dir, otherdir: directionType;
		r: point2d;
		v: array[boolean] of seqIndex;
		toggle: boolean;

	function pointAhead: boolean;
		begin
			if ce = nil then
				pointAhead := false
			else begin
				toggle := not toggle;
				with ce^^.edge do
					if startVertex = v[not toggle] then
						v[toggle] := endVertex
					else
						v[toggle] := startVertex;
				with polygonInfo, pInfo^^[v[toggle]] do
					case r.whichSideSign(pInfo^^[v[not toggle]].p, p) of
						-1: 
							pointAhead := dir = right;
						0: 
							pointAhead := colinearCheck(sector.s, v[toggle]);
						1: 
							pointAhead := dir = left
					end
			end
		end;

	begin
		with polygonInfo, sector do
			if chainEnd[left] <> chainEnd[right] then begin (* there is more than one chain element *)
				if s = currentVertex[left] then begin
					dir := right;
					otherdir := left
				end
				else begin
					dir := left;
					otherdir := right
				end;
				with start^^.edge do
					if startVertex = s then
						v[true] := endVertex
					else
						v[true] := startVertex;
				toggle := true;
				r := pInfo^^[s].p;
				ce := start^^.next[dir];
				while pointAhead do
					with start^^ do begin
						if (self.t.triangles.length < maxTriangles) & (self.t.edges.length < maxEdges) then begin
							e := self.t.addNewEdge(s, v[toggle]);
							self.t.addNewTriangle(e, ce^^.edge, edge)
						end
						else
							fatalError('Invalid initialisation! Selection order could be wrong');
{$IFC AnimationEnabled AND ATriangulation }
						animateNewEdge;
{$ENDC }
						edge := e;
						next[dir] := ce^^.next[dir];
						if next[dir] <> nil then
							next[dir]^^.next[otherdir] := start
						else
							chainEnd[dir] := start;
						chainPool.putChainElement(ce);
						ce := start^^.next[dir]
					end
			end
	end;


{$IFC AnimationEnabled AND ATriangulation }
procedure triangulationSweepLine.animateNextEvent;
	begin
		if animationFlag[ATriangulationItem] then
			with t.points, polygonInfo do begin
				if currentEvent > 1 then
					drawThickVertical(point2d(elements^^[events^^[currentEvent - 1]]).getX);
				elements^^[currentPoint].interactiveFlash;
				drawThickVertical(point2d(elements^^[currentPoint]).getX);
				waitForClick(animationFlag[ATriangulationItem]);
			end;
	end;

procedure triangulationSweepLine.animateNewEdge;
	var
		x1, y1, x2, y2: extended;
		e: graphEdgeTriangulation;
		v: homogenVector;
	begin
		if animationFlag[ATriangulationItem] then
			with t.points do begin
				v := newHomogenVector(2);
				e := graphEdgeTriangulation(t.edges.elements^^[t.edges.length]);
				v.append(elements^^[e.startVertex]);
				v.append(elements^^[e.endVertex]);
				v.interactiveFlash;
				v.free;
				x1 := point2d(elements^^[e.startVertex]).getX;
				y1 := point2d(elements^^[e.startVertex]).getY;
				x2 := point2d(elements^^[e.endVertex]).getX;
				y2 := point2d(elements^^[e.endVertex]).getY;
				drawLineSegment(x1, y1, x2, y2);
				waitForClick(animationFlag[ATriangulationItem]);
			end;
	end;
{$ENDC }



end.
