(*	XYZ GeoBench (eXperimental geometrY Zurich):		*)
(*	An object oriented, interactive workbench for		*)
(*	experimental geometric computation.					*)
(*														*)
(*	 1999 by Peter Schorn, peter.schorn@acm.org		*)
(*	Culmannstrasse 77, CH-8006 Zrich, Switzerland		*)
(*	This software may be copied freely for educational	*)
(*	and research purposes only. For commercial use,		*)
(*	please contact the author.							*)
(*														*)
(*	Version 5.0.5: 30-Sep-1999							*)

unit twoDTree;

interface

uses
	QDPicture, rectangleVector;

type
	binaryTreeNode = object(obj)
			left, right: binaryTreeNode; (* 'left' and 'right' belong to the internal state. *)
			procedure init;
			override;
			procedure free;
			override;
			procedure freeAll;
			override;
			function duplicate: obj;
			override;
			function duplicateAll: obj;
			override;
			procedure fileOutput;
			override;
			procedure fileInput;
			override;
			procedure textFileOutput;
			override;
			procedure textFileInput;
			override;

			procedure freeNode;
(* Frees only the node. *)

			procedure freeNodeAll;
(* Frees only the node and information belonging to the node (used in derived nodes).	*)

			function duplicateNode: obj;
(*	Duplicates only the node. *)

			function duplicateNodeAll: obj;
(*	Duplicates the node and information belonging to the node (used in derived nodes).	*)

			procedure fileInputNode;
(*	Reads in a binary format just the extra information belonging to the node from		*)
(*	'currentStream'. The default implementation raises an error condition.				*)

			procedure fileOutputNode;
(*	Writes in a binary format just the extra information belonging to the node to			*)
(* 'currentStream'. The default implementation raises an error condition.				*)

			procedure textFileInputNode;
(*	Reads in a textual format just the extra information belonging to the node from		*)
(*	'currentStream'. The default implementation raises an error condition.				*)

			procedure textFileOutputNode;
(*	Writes in a textual format just the extra information belonging to the node to		*)
(* 'currentStream'. The default implementation raises an error condition.				*)

			function nodeCompare (x: binaryTreeNode): relationObj;
(*	Defines a canonical ordering just on the extra information belonging to the node.		*)
(*	The default implementation raises an error condition.										*)

		end;

	binaryTree = object(obj)
			root: binaryTreeNode;		(* The root of the binary tree *)
			origin: realPoint;			(* The point where the triangle depicting the tree is drawn *)
			procedure init;
			override;
			procedure free;
			override;
			procedure freeAll;
			override;
			function duplicate: obj;
			override;
			function duplicateAll: obj;
			override;
			procedure fileOutput;
			override;
			procedure fileInput;
			override;
			procedure textFileOutput;
			override;
			procedure textFileInput;
			override;
			function isSelected (where: obj): boolean;
			override;
			procedure interactiveOutput;
			override;
			procedure centerOfGravity (var cx, cy: extended);
			override;
			function equal (o: obj): boolean;
			override;
			function memberType (o: obj): boolean;
			override;
			procedure translate (dx, dy: extended);
			override;
			procedure rotate (xOrigin, yOrigin, alpha: extended);
			override;
			procedure scale (lambda: extended);
			override;
			procedure reflect (a, b, c: extended);
			override;
			function genericLessThan (o: obj): boolean;
			override;
			function drawingComplexity (cutOff: longInt): longInt;
			override;
			function makeTriangle: polygon2d;
(*	Creates the triangle that serves as an icon for the binary tree. *)
		end;

	twoDTreeNode = object(binaryTreeNode)
			p: point2d;	(* The point associated with the 2-d tree; belongs to the internal state. *)
			procedure freeNodeAll;
			override;
			function duplicateNodeAll: obj;
			override;
			procedure fileInputNode;
			override;
			procedure fileOutputNode;
			override;
			procedure textFileInputNode;
			override;
			procedure textFileOutputNode;
			override;
			function getType: objectType;
			override;
			function getTextType: Str255;
			override;
			function nodeCompare (x: binaryTreeNode): relationObj;
			override;
			function getInfo: obj;
			override;
			function memberType (o: obj): boolean;
			override;

			procedure boundingBox (var rLeft, rBottom, rRight, rTop: extended; var length: longInt);
			procedure getRectangles (rLeft, rBottom, rRight, rTop: extended; vertical: boolean; result: rectangleVector);
			procedure drawRectangles (rLeft, rBottom, rRight, rTop: extended; vertical: boolean);
		end;

	twoDTree = object(binaryTree)
			procedure describe;
			override;
			procedure execute (item: integer);
			override;
			function getType: objectType;
			override;
			function getTextType: Str255;
			override;
			function getInfo: obj;
			override;
			function getBoundingObject: obj;
			override;
			procedure insert (p: point2d);
(* Inserts 'p' into the 2-d tree. *)
			function rangeQuery (r: rectangle): pointVector;
(*	Returns as a result all the points of the 2-d tree that lie inside the rectangle 'r'. *)
			procedure boundingBox (var rLeft, rBottom, rRight, rTop: extended; var length: longInt);
(* Computes the bounding box of the points in the 2-d tree and the number of elements in the tree *)
			function getRectangles: rectangleVector;
(*	Computes the partition of the bounding rectangle of the point set induced by the 2-d tree *)
(*	as a vector of rectangles *)
			procedure drawRectangles;
(*	Draws the partition of the bounding rectangle of the point set induced by the 2-d tree. *)
		end;

procedure initTwoDTree;

implementation

function twoDTreeNode.getTextType: Str255;
	begin
		getTextType := '2-d Node';
	end;

function twoDTreeNode.memberType (o: obj): boolean;
	begin
		memberType := member(o, twoDTreeNode);
	end;

function twoDTreeNode.getInfo: obj;
	var
		v: vector;
	begin
		v := newVector(1 + ord(left <> nil) + ord(right <> nil));
		appendObjInfoObject(v, 'Point', p, true);
		if left <> nil then
			appendObjInfoObject(v, 'Left', left, true);
		if right <> nil then
			appendObjInfoObject(v, 'Right', right, true);
		getInfo := v;
	end;

function twoDTree.getInfo: obj;
	var
		v: vector;
	begin
		v := newVector(2);
		appendObjInfoObject(v, 'Origin', origin, false);
		appendObjInfoObject(v, 'Root', root, true);
		getInfo := v;
	end;

function treeCompare (a, b: binaryTreeNode): relationObj;
	var
		t: relationObj;
	begin
		if a = nil then
			if b = nil then
				treeCompare := equalObj
			else
				treeCompare := lessThanObj
		else if b = nil then
			treeCompare := greaterThanObj
		else begin
			t := treeCompare(a.left, b.left);
			if t = equalObj then begin
				t := a.nodeCompare(b);
				if t = equalObj then
					treeCompare := treeCompare(a.right, b.right)
				else
					treeCompare := t;
			end
			else
				treeCompare := t;
		end;
	end;

function binaryTree.genericLessThan (o: obj): boolean;
	var
		t: binaryTree;
	begin
		t := binaryTree(o);
		genericLessThan := origin.genericLessThan(t.origin) | origin.equal(t.origin) & (treeCompare(root, t.root) = lessThanObj);
	end;

function binaryTree.equal (o: obj): boolean;
	var
		t: binaryTree;
	begin
		if (o <> nil) & member(o, binaryTree) then begin
			t := binaryTree(o);
			equal := origin.equal(t.origin) & (treeCompare(root, t.root) = equalObj);
		end
		else
			equal := false;
	end;

procedure binaryTree.centerOfGravity (var cx, cy: extended);
	begin
		origin.centerOfGravity(cx, cy);
	end;

procedure binaryTree.fileOutput;
	begin
		origin.fileOutput;
		writeByte(ord(root = nil));
		if root <> nil then
			root.fileOutputTag;
	end;

procedure binaryTree.textFileOutput;
	begin
		origin.textFileOutput;
		writeTextLongInt(ord(root = nil));
		if root <> nil then
			writeTextObj(root);
	end;

function getTreeNode: binaryTreeNode;
	var
		n: binaryTreeNode;
		b: Byte;
	begin
		readByte(b);
		if b = 0 then begin
			n := binaryTreeNode(newObj(fileGetType));
			n.fileInput;
			getTreeNode := n;
		end
		else
			getTreeNode := nil;
	end;

procedure binaryTree.fileInput;
	var
		r: realPoint;
	begin
		new(r);
		failNil(Handle(r));
		r.fileInput;
		origin := r;
		root := getTreeNode;
	end;

function textGetTreeNode: binaryTreeNode;
	var
		i: longInt;
	begin
		readTextLongInt(i);
		if i = 0 then
			textGetTreeNode := binaryTreeNode(readTextObj)
		else
			textGetTreeNode := nil;
	end;

procedure binaryTree.textFileInput;
	var
		r: realPoint;
	begin
		new(r);
		failNil(Handle(r));
		r.textFileInput;
		origin := r;
		root := textGetTreeNode;
	end;

procedure binaryTreeNode.fileInputNode;
	begin
		unimplementedMethod('binaryTreeNode.fileInputNode');
	end;

procedure binaryTreeNode.fileOutputNode;
	begin
		unimplementedMethod('binaryTreeNode.fileOutputNode');
	end;

procedure binaryTreeNode.textFileInputNode;
	begin
		unimplementedMethod('binaryTreeNode.fileInputNode');
	end;

procedure binaryTreeNode.textFileOutputNode;
	begin
		unimplementedMethod('binaryTreeNode.fileOutputNode');
	end;

{$IFC NOT THINK_PASCAL }
{$PUSH }
{$PRAGMAC warn_unusedarg off }
{$ENDC }
function binaryTreeNode.nodeCompare (x: binaryTreeNode): relationObj;
	begin
		unimplementedMethod('binaryTreeNode.nodeCompare');
		nodeCompare := lessThanObj;
	end;
{$IFC NOT THINK_PASCAL }
{$POP }
{$ENDC }

function twoDTreeNode.nodeCompare (x: binaryTreeNode): relationObj;
	begin
		if p.genericLessThan(twoDTreeNode(x).p) then
			nodeCompare := lessThanObj
		else if p.equal(twoDTreeNode(x).p) then
			nodeCompare := equalObj
		else
			nodeCompare := greaterThanObj;
	end;

procedure twoDTreeNode.textFileInputNode;
	var
		r: realPoint;
	begin
		new(r);
		failNil(Handle(r));
		r.textFileInput;
		p := r;
	end;

procedure twoDTreeNode.textFileOutputNode;
	begin
		p.textFileOutput;
	end;

procedure twoDTreeNode.fileInputNode;
	var
		r: realPoint;
	begin
		new(r);
		failNil(Handle(r));
		r.fileInput;
		p := r;
	end;

procedure twoDTreeNode.fileOutputNode;
	begin
		p.fileOutput;
	end;

procedure binaryTreeNode.fileOutput;
	begin
		fileOutputNode;
		writeByte(ord(left = nil));
		if left <> nil then
			left.fileOutputTag;
		writeByte(ord(right = nil));
		if right <> nil then
			right.fileOutputTag;
	end;

procedure binaryTreeNode.fileInput;
	begin
		fileInputNode;
		left := getTreeNode;
		right := getTreeNode;
	end;

procedure binaryTreeNode.textFileOutput;
	begin
		textFileOutputNode;
		writeTextLongInt(ord(left = nil));
		if left <> nil then
			writeTextObj(left);
		writeTextLongInt(ord(right = nil));
		if right <> nil then
			writeTextObj(right);
	end;

procedure binaryTreeNode.textFileInput;
	begin
		textFileInputNode;
		left := textGetTreeNode;
		right := textGetTreeNode;
	end;

procedure binaryTreeNode.init;
	begin
		left := nil;
		right := nil;
	end;

procedure binaryTreeNode.freeNode;
	begin
		shallowFree;
	end;

procedure binaryTreeNode.freeNodeAll;
	begin
		shallowFree;
	end;

procedure binaryTreeNode.free;
	begin
		if left <> nil then
			left.free;
		if right <> nil then
			right.free;
		freeNode;
	end;

procedure binaryTreeNode.freeAll;
	begin
		if left <> nil then
			left.freeAll;
		if right <> nil then
			right.freeAll;
		freeNodeAll;
	end;

function binaryTreeNode.duplicateNode: obj;
	begin
		duplicateNode := shallowDuplicate;
	end;

function binaryTreeNode.duplicateNodeAll: obj;
	begin
		duplicateNodeAll := shallowDuplicate;
	end;

function binaryTreeNode.duplicate: obj;
	var
		r: binaryTreeNode;
	begin
		r := binaryTreeNode(duplicateNode);
		if left <> nil then
			r.left := binaryTreeNode(left.duplicate);
		if right <> nil then
			r.right := binaryTreeNode(right.duplicate);
		duplicate := r;
	end;

function binaryTreeNode.duplicateAll: obj;
	var
		r: binaryTreeNode;
	begin
		r := binaryTreeNode(duplicateNodeAll);
		if left <> nil then
			r.left := binaryTreeNode(left.duplicateAll);
		if right <> nil then
			r.right := binaryTreeNode(right.duplicateAll);
		duplicateAll := r;
	end;

procedure twoDTreeNode.freeNodeAll;
	begin
		p.free;
		shallowFree;
	end;

function twoDTreeNode.duplicateNodeAll: obj;
	var
		o: twoDTreeNode;
	begin
		o := twoDTreeNode(shallowDuplicate);
		o.p := point2d(p.duplicate);
		duplicateNodeAll := o;
	end;

procedure binaryTree.init;
	var
		o: realPoint;
	begin
		root := nil;
		new(o);
		failNil(Handle(o));
		o.init;
		o.x := 50;
		o.y := 50;
		origin := o;
	end;

procedure binaryTree.free;
	begin
		root.free;
		origin.free;
		shallowFree;
	end;

procedure binaryTree.freeAll;
	begin
		root.freeAll;
		origin.free;
		shallowFree;
	end;

function binaryTree.duplicate: obj;
	var
		b: binaryTree;
	begin
		b := binaryTree(shallowDuplicate);
		b.root := binaryTreeNode(root.duplicate);
		b.origin := realPoint(origin.duplicate);
		duplicate := b;
	end;

function binaryTree.duplicateAll: obj;
	var
		b: binaryTree;
	begin
		b := binaryTree(shallowDuplicate);
		b.root := binaryTreeNode(root.duplicateAll);
		b.origin := realPoint(origin.duplicate);
		duplicateAll := b;
	end;

function binaryTree.makeTriangle: polygon2d;
	var
		poly: polygon2d;
		p: realPoint;
	begin
		poly := newPolygon2d(3);
		poly.append(origin.duplicate);
		new(p);
		failNil(Handle(p));
		p.x := origin.x + 30;
		p.y := origin.y;
		poly.append(p);
		new(p);
		failNil(Handle(p));
		p.x := origin.x + 15;
		p.y := origin.y - 15;
		poly.append(p);
		makeTriangle := poly;
	end;

function binaryTree.isSelected (where: obj): boolean;
	var
		p: polygon2d;
	begin
		p := makeTriangle;
		isSelected := p.isSelected(where);
		p.freeAll;
	end;

procedure binaryTree.interactiveOutput;
	var
		p: polygon2d;
	begin
		p := makeTriangle;
		p.interactiveOutput;
		p.freeAll;
	end;

{$IFC NOT THINK_PASCAL }
{$PUSH }
{$PRAGMAC warn_unusedarg off }
{$ENDC }
function binaryTree.drawingComplexity (cutOff: longInt): longInt;
	begin
		drawingComplexity := 3;
	end;
{$IFC NOT THINK_PASCAL }
{$POP }
{$ENDC }

function binaryTree.memberType (o: obj): boolean;
	begin
		memberType := (o <> nil) & member(o, binaryTree);
	end;

procedure binaryTree.translate (dx, dy: extended);
	begin
		origin.translate(dx, dy);
	end;

procedure binaryTree.rotate (xOrigin, yOrigin, alpha: extended);
	begin
		origin.rotate(xOrigin, yOrigin, alpha);
	end;

procedure binaryTree.scale (lambda: extended);
	begin
		origin.scale(lambda);
	end;

procedure binaryTree.reflect (a, b, c: extended);
	begin
		origin.reflect(a, b, c);
	end;

function twoDTree.getType: objectType;
	begin
		getType := twoDTreeType;
	end;

function twoDTreeNode.getType: objectType;
	begin
		getType := twoDTreeNodeType;
	end;

function twoDTree.getTextType: Str255;
	begin
		getTextType := '2-d Tree';
	end;

procedure twoDTree.drawRectangles;
	var
		rLeft, rBottom, rRight, rTop: extended;
		size: longInt;
	begin
		boundingBox(rLeft, rBottom, rRight, rTop, size);
		if size > 0 then begin
			drawRectangle2d(rLeft, rBottom, rRight, rTop);
			twoDTreeNode(root).drawRectangles(rLeft, rBottom, rRight, rTop, true);
		end;
	end;

function twoDTree.getRectangles: rectangleVector;
	var
		rLeft, rBottom, rRight, rTop: extended;
		result: rectangleVector;
		size: longInt;
	begin
		boundingBox(rLeft, rBottom, rRight, rTop, size);
		if size > 0 then begin
			result := newRectangleVector(size);
			twoDTreeNode(root).getRectangles(rLeft, rBottom, rRight, rTop, true, result);
			if result.length > 0 then
				getRectangles := result
			else begin
				result.freeAll;
				getRectangles := nil;
			end;
		end
		else
			getRectangles := nil;
	end;

procedure twoDTree.boundingBox (var rLeft, rBottom, rRight, rTop: extended; var length: longInt);
	begin
		length := 0;
		if root <> nil then begin
			with twoDTreeNode(root) do begin
				rLeft := p.getX;
				rRight := rLeft;
				rBottom := p.getY;
				rTop := rBottom;
			end;
			twoDTreeNode(root).boundingBox(rLeft, rBottom, rRight, rTop, length);
		end;
	end;

procedure twoDTreeNode.boundingBox (var rLeft, rBottom, rRight, rTop: extended; var length: longInt);
	var
		px, py: extended;
	begin
		length := length + 1;
		px := p.getX;
		py := p.getY;
		if px < rLeft then
			rLeft := px
		else if px > rRight then
			rRight := px;
		if py < rBottom then
			rBottom := py
		else if py > rTop then
			rTop := py;
		if left <> nil then
			twoDTreeNode(left).boundingBox(rLeft, rBottom, rRight, rTop, length);
		if right <> nil then
			twoDTreeNode(right).boundingBox(rLeft, rBottom, rRight, rTop, length);
	end;

procedure twoDTreeNode.drawRectangles (rLeft, rBottom, rRight, rTop: extended; vertical: boolean);
	var
		px, py: extended;
	begin
		px := p.getX;
		py := p.getY;
		if vertical then begin
			if (py <> rBottom) & (py <> rTop) then
				drawLineSegment(rLeft, py, rRight, py);
			if left <> nil then
				twoDTreeNode(left).drawRectangles(rLeft, rBottom, rRight, py, not vertical);
			if right <> nil then
				twoDTreeNode(right).drawRectangles(rLeft, py, rRight, rTop, not vertical)
		end
		else begin
			if (px <> rLeft) & (px <> rRight) then
				drawLineSegment(px, rBottom, px, rTop);
			if left <> nil then
				twoDTreeNode(left).drawRectangles(rLeft, rBottom, px, rTop, not vertical);
			if right <> nil then
				twoDTreeNode(right).drawRectangles(px, rBottom, rRight, rTop, not vertical);
		end;
	end;

procedure twoDTreeNode.getRectangles (rLeft, rBottom, rRight, rTop: extended; vertical: boolean; result: rectangleVector);
	var
		px, py: extended;
		r: rectangle;
	begin
		px := p.getX;
		py := p.getY;
		if left = nil then begin
			r := rectangle(newInitObj(rectangleType));
			r.left := rLeft;
			r.bottom := rBottom;
			if vertical then begin
				r.right := rRight;
				r.top := py;
			end
			else begin
				r.right := px;
				r.top := rTop;
			end;
			if (r.right <> r.left) & (r.bottom <> r.top) then
				result.append(r)
			else
				r.free;
		end
		else if vertical then
			twoDTreeNode(left).getRectangles(rLeft, rBottom, rRight, py, not vertical, result)
		else
			twoDTreeNode(left).getRectangles(rLeft, rBottom, px, rTop, not vertical, result);
		if right = nil then begin
			r := rectangle(newInitObj(rectangleType));
			if vertical then begin
				r.left := rLeft;
				r.bottom := py;
			end
			else begin
				r.left := px;
				r.bottom := rBottom;
			end;
			r.right := rRight;
			r.top := rTop;
			if (r.right <> r.left) & (r.bottom <> r.top) then
				result.append(r)
			else
				r.free;
		end
		else if vertical then
			twoDTreeNode(right).getRectangles(rLeft, py, rRight, rTop, not vertical, result)
		else
			twoDTreeNode(right).getRectangles(px, rBottom, rRight, rTop, not vertical, result);
	end;

	const
		InsertIntoTreeC = 1;
		InsertIntoNewTreeC = 2;
		RangeQueryC = 3;
		ShowPartitionC = 4;
		CreateRegionC = 5;

procedure twoDTree.describe;
	begin
		with currentArguments do begin
			if (length > 0) & currentArgumentsSameType & member(ith(1), point2d) | (length = 1) & currentArguments1IsHomogenVector & member(vector(ith(1)).elements^^[1], point2d) then
				menuEntry('Insert Into New 2-D Tree', InsertIntoNewTreeC)
			else if (length = 2) & (currentArguments1IsHomogenVector & member(vector(ith(1)).elements^^[1], point2d) & member(ith(2), twoDTree) | currentArguments2IsHomogenVector & member(vector(ith(2)).elements^^[1], point2d) & member(ith(1), twoDTree)) then
				menuEntry('Insert Into 2-D Tree', InsertIntoTreeC)
			else if (length = 2) & (member(ith(1), rectangle) & member(ith(2), twoDTree) | member(ith(2), rectangle) & member(ith(1), twoDTree)) then
				menuEntry('Range Query In 2-D Tree', RangeQueryC)
			else if (length = 1) & member(ith(1), twoDTree) then begin
				menuEntry('Show Partition', ShowPartitionC);
				menuEntry('Create Regions', CreateRegionC);
			end;
		end;
	end;

procedure twoDTree.execute (item: integer);

	procedure callGetRectangles (t: twoDTree);
		var
			rLeft, rBottom, rRight, rTop: extended;
			rl, rb, rr, rt: integer;
			length: longInt;
		begin
			t.boundingBox(rLeft, rBottom, rRight, rTop, length);
			if (length > 0) & worldToWindow(rLeft, rBottom, rl, rb) & worldToWindow(rRight, rTop, rr, rt) then begin
				OpenQDPicture(rl - 1, rb - 1, rr + 1, rt + 1);
				t.drawRectangles;
				currentResult.append(CloseQDPicture(false));
			end;
		end;

	procedure callInsertIntoTree (t: twoDTree; h: homogenVector);
		procedure insertP (x: point2d);
			begin
				t.insert(x);
			end;
		begin
			h.forAll(insertP);
		end;

	function callInsertIntoNewTree (l: sequence): twoDTree;
		var
			t: twoDTree;
		procedure insertP (x: point2d);
			begin
				t.insert(x);
			end;
		begin
			new(t);
			failNil(Handle(t));
			t.init;
			l.forAll(insertP);
			callInsertIntoNewTree := t;
		end;

	begin
		with currentArguments do
			case item of
				InsertIntoTreeC: 
					if member(ith(1), twoDTree) then
						callInsertIntoTree(twoDTree(ith(1)), homogenVector(ith(2)))
					else
						callInsertIntoTree(twoDTree(ith(2)), homogenVector(ith(1)));
				InsertIntoNewTreeC: 
					if member(ith(1), vector) then
						currentResult.append(callInsertIntoNewTree(vector(ith(1))))
					else
						currentResult.append(callInsertIntoNewTree(currentArguments));
				RangeQueryC: 
					if member(ith(1), twoDTree) then
						currentResult.appendNonNil(twoDTree(ith(1)).rangeQuery(rectangle(ith(2))))
					else
						currentResult.appendNonNil(twoDTree(ith(2)).rangeQuery(rectangle(ith(1))));
				ShowPartitionC: 
					callGetRectangles(twoDTree(ith(1)));
				CreateRegionC: 
					currentResult.appendNonNil(twoDTree(ith(1)).getRectangles);
			end;
	end;

procedure nodeInsert (var r: twoDTreeNode; vertical: boolean; n: twoDTreeNode);
	begin
		if r = nil then
			r := n
		else if (vertical & (n.p.getY < r.p.getY)) | (not vertical & (n.p.getX < r.p.getX)) then
			nodeInsert(twoDTreeNode(r.left), not vertical, n)
		else
			nodeInsert(twoDTreeNode(r.right), not vertical, n)
	end;

procedure twoDTree.insert (p: point2d);
	var
		n: twoDTreeNode;
	begin
		new(n);
		failNil(Handle(n));
		n.init;
		n.p := point2d(p.duplicate);
		nodeInsert(twoDTreeNode(root), true, n);
	end;

	var
		extractCalls: longInt;

procedure extract (r: twoDTreeNode; vertical: boolean; rec: rectangle; result: list);
	begin
		if r <> nil then begin
			extractCalls := extractCalls + 1;
			if (rec.left <= r.p.getX) & (r.p.getX <= rec.right) & (rec.bottom <= r.p.getY) & (r.p.getY <= rec.top) then
				result.append(r.p.duplicate);
			if (vertical & (rec.bottom <= r.p.getY)) | (not vertical & (rec.left <= r.p.getX)) then
				extract(twoDTreeNode(r.left), not vertical, rec, result);
			if (vertical & (rec.top >= r.p.getY)) | (not vertical & (rec.right >= r.p.getX)) then
				extract(twoDTreeNode(r.right), not vertical, rec, result);
		end;
	end;

function twoDTree.rangeQuery (r: rectangle): pointVector;
	var
		l: list;
		pv: pointVector;
	begin
		l := newEmptyList;
		extractCalls := 0;
		extract(twoDTreeNode(root), true, r, l);
		displayInfoString(stringOf('# extracts = ', extractCalls : 1));
		if l.length > 0 then begin
			new(pv);
			failNil(Handle(pv));
			pv.initFromList(l);
			rangeQuery := pv;
		end
		else
			rangeQuery := nil;
		l.free;
	end;

procedure initTwoDTree;
	var
		t: twoDTree;
		n: twoDTreeNode;
	begin
		new(t);
		failNil(Handle(t));
		prototypes[twoDTreeType] := t;
		new(n);
		failNil(Handle(n));
		prototypes[twoDTreeNodeType] := n;
	end;

end.
