(*	XYZ GeoBench (eXperimental geometrY Zurich):		*)
(*	An object oriented, interactive workbench for		*)
(*	experimental geometric computation.					*)
(*														*)
(*	 1999 by Peter Schorn, peter.schorn@acm.org		*)
(*	Culmannstrasse 77, CH-8006 Zrich, Switzerland		*)
(*	This software may be copied freely for educational	*)
(*	and research purposes only. For commercial use,		*)
(*	please contact the author.							*)
(*														*)
(*	Version 5.0.5: 30-Sep-1999							*)

unit unionFind;

interface

uses
	graph;

function createUnionFindSimple (n: seqIndex): unionFindSimple;

function createUnionFindCount (n: seqIndex): unionFindCount;

procedure createRandomGraphByProbability (p: extended; isDirectedGraph, connected: boolean; result: simpleUndirectedGraph);

procedure createRandomGraphByEdgeNumber (m: seqIndex0; isDirectedGraph, connected: boolean; result: simpleUndirectedGraph);

procedure randomSubset (n, k: longInt; procedure generate (x: longInt));

implementation

{$SETC DEBUG = false }
{$SETC Instrument = false }

type
	integerPair = object(obj)
			index, value: longInt;
		end;

{$IFC Instrument }
var
	loopCount: longInt;
{$ENDC }

procedure lookup (l: list; var e: elementH; i: longInt; var ip: integerPair);
	begin
		e := l.first;
		while e^^.next <> nil do begin
{$IFC Instrument }
			loopCount := loopCount + 1;
{$ENDC }
			ip := integerPair(e^^.next^^.value);
			if ip.index = i then
				exit(lookup);
			e := e^^.next;
		end;
		ip := nil;
	end;

procedure randomSubset (n, k: longInt; procedure generate (x: longInt));
	var
		hashTable: homogenVector; (* of list of integerPair *)
		l: list;
		i, j: longInt;
		ip, ip2: integerPair;
		e: elementH;
	begin
		assert(n > 0, 'n > 0');
{$IFC DEBUG }
		showText;
		writeLn;
{$ENDC }
{$IFC Instrument }
		loopCount := 0;
{$ENDC }
		hashTable := newHomogenVector(k);
		for i := 1 to k do begin
{$IFC AnimationEnabled }
			if userAbort then begin
				hashTable.freeAll;
				exit(randomSubset);
			end;
{$ENDC }
			l := newEmptyList;
			new(ip);
			failNil(Handle(ip));
			ip.index := 0; (* sentinel *)
			l.append(ip);
			hashTable.append(l);
		end;
		for i := 1 to k do begin
{$IFC AnimationEnabled }
			if userAbort then begin
				hashTable.freeAll;
				exit(randomSubset);
			end;
{$ENDC }
{$IFC THINK_PASCAL }
			j := trunc(randomReal * (n - i + 1)) + i;
{$ELSEC }
			j := truncLong(randomReal * (n - i + 1)) + i;
{$ENDC }
			l := list(hashTable.elements^^[(j - 1) mod k + 1]);
			lookup(l, e, j, ip);
			if ip = nil then begin
{$IFC DEBUG }
				writeLn(j);
{$ENDC }
				generate(j);
				new(ip);
				failNil(Handle(ip));
				ip.index := j;
				l.append(ip);
			end
			else begin
{$IFC DEBUG }
				writeLn(ip.value);
{$ENDC }
				generate(ip.value);
			end;
			l := list(hashTable.elements^^[i]);
			lookup(l, e, i, ip2);
			if ip2 = nil then
				ip.value := i
			else begin
				ip.value := ip2.value;
				l.delete(e);
				ip2.free;
			end;
		end;
		hashTable.freeAll;
{$IFC Instrument }
		displayInfoString(StringOf('Work = ', loopCount : 1));
{$ENDC }
	end;

procedure sequence.forRandomSubset (cardinality: seqIndex0; procedure whatToDo (x: obj));

	procedure doIt (i: longInt);
		begin
			whatToDo(ith(i));
		end;

	begin (* forRandomSubset *)
		if (cardinality > 0) & (length > 0) then
			if cardinality >= length then
				forAll(whatToDo)
			else
				randomSubset(length, cardinality, doIt);
	end;

procedure unionFindSimple.init;
	begin
		n := currentVectorLength;
		uf := seqIndexArrayH(dynamicArray(currentVectorLength, sizeOf(seqIndex)));
		reset;
	end;

procedure unionFindCount.init;
	begin
		count := seqIndexArrayH(dynamicArray(currentVectorLength, sizeOf(seqIndex)));
		inherited init;
	end;

procedure unionFindSimple.reset;
	var
		i: seqIndex;
	begin
		for i := 1 to n do begin
			uf^^[i] := i;
		end;
	end;

procedure unionFindCount.reset;
	var
		i: seqIndex;
	begin
		for i := 1 to n do begin
			count^^[i] := 1;
		end;
		inherited reset;
	end;

procedure unionFindSimple.free;
	begin
{$IFC THINK_PASCAL }
		disposHandle(Handle(uf));
{$ELSEC }
		DisposeHandle(Handle(uf));
{$ENDC }
		shallowFree;
	end;

procedure unionFindCount.free;
	begin
{$IFC THINK_PASCAL }
		disposHandle(Handle(count));
{$ELSEC }
		DisposeHandle(Handle(count));
{$ENDC }
		inherited free;
	end;

function unionFindSimple.find (startVertex: seqIndex): seqIndex;
	var
		s, next, t: seqIndex;
	begin
		s := startVertex;
		next := uf^^[s];
		while next <> s do begin
			s := next;
			next := uf^^[s];
		end;
		t := startVertex;
		next := uf^^[t];
		while next <> t do begin
			uf^^[t] := s;
			t := next;
			next := uf^^[t];
		end;
		find := s;
	end;

procedure unionFindSimple.compressAll;
	var
		visitedH: booleanArrayH;
		i, s, t, next: seqIndex;
	begin
		visitedH := booleanArrayH(dynamicArray(n, sizeOf(boolean)));
		for i := 1 to n do
			visitedH^^[i] := false;
		for i := 1 to n do
			if not visitedH^^[i] then begin
				s := i;
				next := uf^^[s];
				while next <> s do begin
					s := next;
					next := uf^^[s];
				end;
				t := i;
				next := uf^^[t];
				while (next <> t) & not visitedH^^[t] do begin
					uf^^[t] := s;
					visitedH^^[t] := true;
					t := next;
					next := uf^^[t];
				end;
				visitedH^^[i] := true;
			end;
{$IFC THINK_PASCAL }
		disposHandle(Handle(visitedH));
{$ELSEC }
		DisposeHandle(Handle(visitedH));
{$ENDC }
	end;

{$IFC NOT THINK_PASCAL }
{$PUSH }
{$PRAGMAC warn_unusedarg off }
{$ENDC }
procedure unionFindSimple.addParts (s, e: seqIndex);
	begin
	end;
{$IFC NOT THINK_PASCAL }
{$POP }
{$ENDC }

procedure unionFindCount.addParts (s, e: seqIndex);
	begin
		count^^[e] := count^^[e] + count^^[s];
	end;

procedure unionFindSimple.union (startVertex, endVertex: seqIndex);
	var
		s, e: seqIndex;
	begin
		s := find(startVertex);
		e := find(endVertex);
		if s <> e then begin
			uf^^[s] := e;
			addParts(s, e);
		end;
	end;

function unionFindCount.isOneComponent: boolean;
	begin
		isOneComponent := cardinality(1) = n;
	end;

function unionFindCount.cardinality (s: seqIndex): seqIndex;
	begin
		cardinality := count^^[find(s)];
	end;

function createUnionFindSimple (n: seqIndex): unionFindSimple;
	var
		result: unionFindSimple;
	begin
		new(result);
		failNil(Handle(result));
		currentVectorLength := n;
		result.init;
		createUnionFindSimple := result;
	end;

function createUnionFindCount (n: seqIndex): unionFindCount;
	var
		result: unionFindCount;
	begin
		new(result);
		failNil(Handle(result));
		currentVectorLength := n;
		result.init;
		createUnionFindCount := result;
	end;

function getEdge (freeEdgeList: list): graphEdge;
	var
		e: graphEdge;
	begin
		if freeEdgeList.length = 0 then begin
			new(e);
			failNil(Handle(e));
			getEdge := e;
		end
		else begin
			getEdge := graphEdge(freeEdgeList.first^^.value);
			freeEdgeList.pop;
		end;
	end;

function isConnected (uf: unionFindCount; freeEdgeList: list; edges: vector): boolean;
	var
		i: seqIndex0;
	begin
		if uf.isOneComponent then
			isConnected := true
		else begin
			isConnected := false;
			uf.reset;
			for i := 1 to edges.length do
				freeEdgeList.append(edges.elements^^[i]);
			edges.length := 0;
		end;
	end;

procedure assign (x, n: longInt; var a, b: seqIndex);
	var
		y: extended;
		truncY: longInt;
	begin
		y := (sqrt(8 * x - 3) - 1) / 2;
{$IFC THINK_PASCAL }
		truncY := trunc(y);
{$ELSEC }
		truncY := truncLong(y);
{$ENDC }
		a := n - 1 - truncY;
		b := n + 1 - x + truncY * (truncY + 1) div 2;
	end;


procedure setEdge (x, n, limit: longInt; ge: graphEdge);
	begin
		with ge do
			if x > limit then
				assign(x - limit, n, endVertex, startVertex)
			else
				assign(x, n, startVertex, endVertex);
	end;

{$IFC AnimationEnabled }
procedure destroyResult (result: simpleUndirectedGraph);
	begin
		result.edges.freeAll;
		result.vertices.freeAll;
		result.shallowFree;
	end;
{$ENDC }


procedure createRandomGraphByEdgeNumber (m: seqIndex0; isDirectedGraph, connected: boolean; result: simpleUndirectedGraph);
	var
		n: seqIndex0;
		uf: unionFindCount;
		freeEdgeList: list;
		sizeOfUniverse, limit: longInt;

	procedure generateConnectedGraphEdge (x: longInt);
		var
			ge: graphEdge;
		begin
			ge := getEdge(freeEdgeList);
			setEdge(x, n, limit, ge);
			result.edges.append(ge);
			uf.union(ge.startVertex, ge.endVertex);
		end;

	procedure generateGraphEdge (x: longInt);
		var
			ge: graphEdge;
		begin
			new(ge);
			failNil(Handle(ge));
			setEdge(x, n, limit, ge);
			result.edges.append(ge);
		end;

	begin
		result.edges := newHomogenVector(m);
		n := result.vertices.length;
		sizeOfUniverse := n * (n - 1);
		limit := sizeOfUniverse div 2;
		if not isDirectedGraph then
			sizeOfUniverse := limit;
		if connected then begin
			uf := createUnionFindCount(n);
			freeEdgeList := newEmptyList;
			repeat
				randomSubset(sizeOfUniverse, m, generateConnectedGraphEdge);
{$IFC AnimationEnabled }
				if userAbort then begin
					destroyResult(result);
					leave;
				end;
{$ENDC }
			until isConnected(uf, freeEdgeList, result.edges);
			uf.free;
			freeEdgeList.freeAll;
		end
		else begin
			randomSubset(sizeOfUniverse, m, generateGraphEdge);
{$IFC AnimationEnabled }
			if userAbort then
				destroyResult(result);
{$ENDC }
		end;
	end;

procedure createRandomGraphByProbability (p: extended; isDirectedGraph, connected: boolean; result: simpleUndirectedGraph);
	var
		n, m, i, j: seqIndex0;
		uf: unionFindCount;
		freeEdgeList: list;
		edge: graphEdge;
	begin
		n := result.vertices.length;
		m := n * (n - 1);
		if not isDirectedGraph then
			m := m div 2;
		result.edges := newHomogenVector(m);
		if connected then begin
			uf := createUnionFindCount(n);
			freeEdgeList := newEmptyList;
			repeat
				for i := 1 to n do
					for j := ord(not isDirectedGraph) * i + 1 to n do begin
{$IFC AnimationEnabled }
						if userAbort then begin
							destroyResult(result);
							uf.free;
							freeEdgeList.freeAll;
							exit(createRandomGraphByProbability);
						end;
{$ENDC }
						if (i <> j) & (randomReal < p) then begin
							uf.union(i, j);
							edge := getEdge(freeEdgeList);
							edge.startVertex := i;
							edge.endVertex := j;
							result.edges.append(edge);
						end;
					end;
			until isConnected(uf, freeEdgeList, result.edges);
			uf.free;
			freeEdgeList.freeAll;
		end
		else
			for i := 1 to n do
				for j := ord(not isDirectedGraph) * i + 1 to n do begin
{$IFC AnimationEnabled }
					if userAbort then begin
						destroyResult(result);
						exit(createRandomGraphByProbability);
					end;
{$ENDC }
					if (i <> j) & (randomReal < p) then
						result.addNewEdge(i, j);
				end;
	end;


end.
