(*	XYZ GeoBench (eXperimental geometrY Zurich):		*)
(*	An object oriented, interactive workbench for		*)
(*	experimental geometric computation.					*)
(*														*)
(*	 1999 by Peter Schorn, peter.schorn@acm.org		*)
(*	Culmannstrasse 77, CH-8006 Zrich, Switzerland		*)
(*	This software may be copied freely for educational	*)
(*	and research purposes only. For commercial use,		*)
(*	please contact the author.							*)
(*														*)
(*	Version 5.0.5: 30-Sep-1999							*)

unit upperMonotonePolygon;

interface

uses
	pointVector;

type
	upperMonotonePolygon = object(polygon2d)

			procedure describe;
			override;
			procedure execute (item: integer);
			override;
			function memberType (o: obj): boolean;
			override;
			function getType: objectType;
			override;
			function getTextType: Str255;
			override;
			function equal (o: obj): boolean;
			override;

			function isMonotoneChain: boolean;
			procedure convexHullGS;
			function triangulationGS: homogenVector; (* of polygons which are triangles *)
			function tangentLog (p: point2d): lineSegment;
			function pointInUpperMonotonePolygon (p: point2d): boolean;
		end;

procedure initUpperMonotonePolygon;

implementation

procedure initUpperMonotonePolygon;
	var
		x: upperMonotonePolygon;
	begin
		new(x);
		failNil(Handle(x));
		prototypes[upperMonotonePolygonType] := x;
	end;

function upperMonotonePolygon.memberType (o: obj): boolean;
	begin
		memberType := member(o, upperMonotonePolygon) & memberBaseType(o);
	end;

function upperMonotonePolygon.getType: objectType;
	begin
		getType := upperMonotonePolygonType;
	end;

function upperMonotonePolygon.getTextType: Str255;
	begin
		getTextType := concat('Upper Monotone Polygon', getTextBaseType);
	end;

function upperMonotonePolygon.equal (o: obj): boolean;
	begin
		equal := (o <> nil) & member(o, upperMonotonePolygon) & inherited equal(o);
	end;

	const
		IsMonotoneC = 1;
		ConvexHullGSC = 2;
		TriangulationGSC = 3;
		TangentC = 4;
		pointInUpperMonotonePolygonC = 5;
		TangentDynamicC = 6;
		pointInUpperMonotonePolygonDynamicC = 7;

procedure upperMonotonePolygon.describe;
	begin
{$IFC OptionalCommand }
		with currentArguments do begin
			if (length = 1) & member(ith(1), polygon2d) & (polygon2d(ith(1)).length > 2) then begin
				beginSubMenu('Monotone Chain');
				menuEntry('Is Monotone', IsMonotoneC);
				menuEntry('Convex Hull', ConvexHullGSC);
				menuEntry('Triangulation', TriangulationGSC);
				endSubMenu;
			end;
			if (length = 2) & member(ith(1), polygon2d) & (polygon2d(ith(1)).length > 2) & member(ith(2), point2d) then begin
				beginSubMenu('Monotone Chain');
				menuEntry('Tangent', TangentC);
				menuEntry('Tangent Dynamic', TangentDynamicC);
				menuEntry('Point In Polygon', pointInUpperMonotonePolygonC);
				menuEntry('Point In Polygon Dynamic', pointInUpperMonotonePolygonDynamicC);
				endSubMenu;
			end;
		end;
{$ENDC }
	end;

procedure computepointInUpperMonotonePolygon (p: upperMonotonePolygon; t: point2d; var mp: markedRealPoint);
	begin
		new(mp);
		failNil(Handle(mp));
		mp.init;
		mp.x := t.getX;
		mp.y := t.getY;
		mp.mark.value := stringOf(p.pointInUpperMonotonePolygon(t));
	end;

procedure callpointInUpperMonotonePolygonDynamic (p: polygon2d; t: point2d);
	var
		m: upperMonotonePolygon;
	begin
		m := upperMonotonePolygon(p.duplicateTypeTransfer(upperMonotonePolygonType));
		dynamicExecution2(m, t, computepointInUpperMonotonePolygon, false, true);
		m.free;
	end;

procedure computeTangent (p: upperMonotonePolygon; t: point2d; var l: lineSegment);
	begin
		l := p.tangentLog(t);
		if l = nil then begin
			new(l);
			failNil(Handle(l));
			l.p := point2d(t.duplicate);
			l.q := point2d(t.duplicate);
		end;
	end;

procedure callTangentDynamic (p: polygon2d; t: point2d);
	var
		m: upperMonotonePolygon;
	begin
		m := upperMonotonePolygon(p.duplicateTypeTransfer(upperMonotonePolygonType));
		dynamicExecution2(m, t, computeTangent, false, true);
		m.free;
	end;

function callIsMonotone (p: polygon2d): markedRealPoint;
	var
		mrp: markedRealPoint;
		m: upperMonotonePolygon;
	begin
		if p.length > 0 then begin
			m := upperMonotonePolygon(p.duplicateTypeTransfer(upperMonotonePolygonType));
			new(mrp);
			failNil(Handle(mrp));
			mrp.init;
			with point2d(p.elements^^[1]) do begin
				mrp.setX(getX);
				mrp.setY(getY);
			end;
			mrp.mark.value := stringOf(m.isMonotoneChain);
			m.free;
			callIsMonotone := mrp;
		end
		else
			callIsMonotone := nil;
	end;

function callConvexHullGS (p: polygon2d): upperMonotonePolygon;
	var
		m: upperMonotonePolygon;
	begin
		m := upperMonotonePolygon(p.duplicateTypeTransfer(upperMonotonePolygonType));
		if m.length > 2 then
			m.convexHullGS;
		callConvexHullGS := upperMonotonePolygon(m.duplicateAll);
		m.free;
	end;

function callTriangulationGS (p: polygon2d): homogenVector;
	var
		m: upperMonotonePolygon;
	begin
		m := upperMonotonePolygon(p.duplicateTypeTransfer(upperMonotonePolygonType));
		callTriangulationGS := m.triangulationGS;
		m.free;
	end;

function callTangent (p: polygon2d; t: point2d): lineSegment;
	var
		m: upperMonotonePolygon;
	begin
		m := upperMonotonePolygon(p.duplicateTypeTransfer(upperMonotonePolygonType));
		callTangent := m.tangentLog(t);
		m.free;
	end;

function callpointInUpperMonotonePolygon (p: polygon2d; t: point2d): markedRealPoint;
	var
		mrp: markedRealPoint;
		m: upperMonotonePolygon;
	begin
		m := upperMonotonePolygon(p.duplicateTypeTransfer(upperMonotonePolygonType));
		new(mrp);
		failNil(Handle(mrp));
		mrp.init;
		mrp.x := t.getX;
		mrp.y := t.getY;
		mrp.mark.value := stringOf(m.pointInUpperMonotonePolygon(t));
		m.free;
		callpointInUpperMonotonePolygon := mrp;
	end;

procedure upperMonotonePolygon.execute (item: integer);
	begin
		with currentArguments do
			case item of
				IsMonotoneC: 
					currentResult.appendNonNil(callIsMonotone(polygon2d(ith(1))));
				ConvexHullGSC: 
					currentResult.appendNonNil(callConvexHullGS(polygon2d(ith(1))));
				TriangulationGSC: 
					currentResult.appendNonNil(callTriangulationGS(polygon2d(ith(1))));
				TangentC: 
					currentResult.appendNonNil(callTangent(polygon2d(ith(1)), point2d(ith(2))));
				pointInUpperMonotonePolygonC: 
					currentResult.appendNonNil(callpointInUpperMonotonePolygon(polygon2d(ith(1)), point2d(ith(2))));
				pointInUpperMonotonePolygonDynamicC: 
					callpointInUpperMonotonePolygonDynamic(polygon2d(ith(1)), point2d(ith(2)));
				TangentDynamicC: 
					callTangentDynamic(polygon2d(ith(1)), point2d(ith(2)));
			end;
	end;

function upperMonotonePolygon.isMonotoneChain: boolean;
	var
		p, q: point2d;
		i: seqIndex;
	begin
		i := 1;
		while (i < length) & (point2d(elements^^[i]).getX < point2d(elements^^[i + 1]).getX) do
			i := i + 1;
		if i = length then begin
			p := point2d(elements^^[1]);
			q := point2d(elements^^[length]);
			i := 2;
			while (i < length) & (point2d(elements^^[i]).whichSideX(p, q) < 0) do
				i := i + 1;
			isMonotoneChain := i = length;
		end
		else
			isMonotoneChain := false;
	end;

procedure upperMonotonePolygon.convexHullGS;
	var
		i: seqIndex1;
		j: seqIndex0;
	begin
		i := 2;
		for j := 3 to length do begin
			while (i > 1) & (point2d(elements^^[j]).whichSideX(point2d(elements^^[i - 1]), point2d(elements^^[i])) <= 0) do
				i := i - 1;
			i := i + 1;
			elements^^[i] := elements^^[j];
		end;
		length := i;
	end;

procedure emitTriangle (p, q, r: point2d; result: homogenVector);
	var
		triangle: polygon2d;
	begin
		triangle := newPolygon2d(3);
		triangle.append(p.duplicate);
		triangle.append(q.duplicate);
		triangle.append(r.duplicate);
		result.append(triangle);
	end;

function upperMonotonePolygon.triangulationGS: homogenVector;
	var
		i: seqIndex1;
		j: seqIndex0;
		result: homogenVector;
	begin
		result := newHomogenVector(length - 2);
		i := 2;
		for j := 3 to length do begin
			while (i > 1) & (point2d(elements^^[j]).whichSideX(point2d(elements^^[i - 1]), point2d(elements^^[i])) >= 0) do begin
				emitTriangle(point2d(elements^^[i - 1]), point2d(elements^^[i]), point2d(elements^^[j]), result);
				i := i - 1;
			end;
			i := i + 1;
			elements^^[i] := elements^^[j];
		end;
		triangulationGS := result;
	end;

function upperMonotonePolygon.tangentLog (p: point2d): lineSegment;
	var
		lower, upper, middle: seqIndex01;
		wsMinus, wsPlus: extended;
		m: point2d;
		l: lineSegment;
		found: boolean;
	begin
		if pointInUpperMonotonePolygon(p) then
			tangentLog := nil
		else begin
			if p.whichSideSign(point2d(elements^^[length]), point2d(elements^^[1])) <> p.whichSideSign(point2d(elements^^[1]), point2d(elements^^[2])) then
				m := point2d(elements^^[1])
			else if p.whichSideSign(point2d(elements^^[length - 1]), point2d(elements^^[length])) <> p.whichSideSign(point2d(elements^^[length]), point2d(elements^^[1])) then
				m := point2d(elements^^[length])
			else begin
				lower := 1;
				upper := length;
				found := false;
				repeat
					middle := (lower + upper) div 2;
					m := point2d(elements^^[middle]);
					assert((middle <> 1) & (middle <> length) & (lower <= upper), '(middle <> 1) & (middle <> length) & (lower <= upper)');
					wsMinus := p.whichSideX(point2d(elements^^[middle - 1]), m);
					wsPlus := p.whichSideX(m, point2d(elements^^[middle + 1]));
					if (wsMinus > 0) & (wsPlus > 0) then
						if p.getX > m.getX then
							lower := middle + 1
						else
							upper := middle - 1
					else if (wsMinus < 0) & (wsPlus < 0) then
						if p.getX > m.getX then
							upper := middle - 1
						else
							lower := middle + 1
					else
						found := true;
				until found;
			end;
			new(l);
			failNil(Handle(l));
			l.p := point2d(p.duplicate);
			l.q := point2d(m.duplicate);
			tangentLog := l;
		end;
	end;

function upperMonotonePolygon.pointInUpperMonotonePolygon (p: point2d): boolean;
	var
		q, r: point2d;
		lower, upper, middle: seqIndex;
	begin
		q := point2d(elements^^[1]);
		r := point2d(elements^^[length]);
		if (p.getX < q.getX) | (p.getX > r.getX) | (p.whichSideX(q, r) > 0) then
			pointInUpperMonotonePolygon := false
		else begin
			lower := 1;
			upper := length;
			repeat
				middle := (lower + upper) div 2;
				if point2d(elements^^[middle]).getX < p.getX then
					lower := middle
				else
					upper := middle;
			until upper - lower = 1;
			pointInUpperMonotonePolygon := p.whichSideX(point2d(elements^^[lower]), point2d(elements^^[upper])) > 0
		end;
	end;

end.
