(*	XYZ GeoBench (eXperimental geometrY Zurich):		*)
(*	An object oriented, interactive workbench for		*)
(*	experimental geometric computation.					*)
(*														*)
(*	 1999 by Peter Schorn, peter.schorn@acm.org		*)
(*	Culmannstrasse 77, CH-8006 Zrich, Switzerland		*)
(*	This software may be copied freely for educational	*)
(*	and research purposes only. For commercial use,		*)
(*	please contact the author.							*)
(*														*)
(*	Version 5.0.5: 30-Sep-1999							*)

unit vector;

interface

uses
	basicGeometry;

procedure initVector;

procedure exitVector;

function getSequence (s: sequence): sequence;

function binaryFind (v: homogenVector; value: real): seqIndex;

implementation

const
	stdPerturbationWidth = 1;
	stdPerturbationHeight = 1;
	stdPerturbationRadius = 1;

var
	perturbationWidth, perturbationHeight, perturbationRadius: extended;
	globalP: realPoint;
	globalR: rectangle;
	globalC: circle;

procedure vector.compact (mayRearrange: boolean);
	var
		i, j: seqIndex01;
		t: objArrayH;
	begin
		if mayRearrange then begin
			i := 1;
			while i <= length do begin
				while (i <= length) & (elements^^[length] = nil) do
					length := length - 1;
				if (i < length) & (elements^^[i] = nil) then begin
					elements^^[i] := elements^^[length];
					length := length - 1;
				end;
				i := i + 1;
			end;
		end
		else begin
			t := objArrayH(dynamicArray(length + 1, sizeOf(obj)));
			j := 0;
			for i := 1 to length do begin
				if elements^^[i] <> nil then begin
					j := j + 1;
					t^^[j] := elements^^[i];
				end;
			end;
{$IFC THINK_PASCAL }
			disposHandle(Handle(elements));
{$ELSEC }
			DisposeHandle(Handle(elements));
{$ENDC }
			elements := t;
			length := j;
		end;
	end;

function binaryFind (v: homogenVector; value: real): seqIndex;
	var
		l: seqIndex;
		r, m: seqIndex0;
	begin
		l := 1;
		r := v.length;
		repeat
			m := (l + r) div 2;
			if realNumber(v.elements^^[m]).value > value then
				r := m
			else
				l := m + 1
		until l = r;
		binaryFind := l;
	end;

	var
		defaultNewFirst, defaultCover, defaultObjectsPerLine: extended;

procedure list.flatten;
	var
		e, t: elementH;
		l: list;
		o: obj;
		v: vector;

	procedure doAppend (x: obj);
		begin
			l.append(x);
		end;

	begin (* list.flatten *)
		l := newEmptyList;
		e := first;
		while e <> nil do begin
			o := e^^.value;
			if member(o, sequence) then begin
				sequence(o).forAll(doAppend);
				o.free;
			end
			else if member(o, point2d) then
				l.append(o)
			else if o.convertToConvexPolygonPossible then begin
				v := vector(o.convertToConvexPolygon);
				v.forAll(doAppend);
				v.free;
				o.free;
			end
			else
				l.append(o);
			t := e;
			e := e^^.next;
			disposeElement(t);
		end;
		first := l.first;
		last := l.last;
		length := l.length;
		dispose(l);
	end;

function vector.existsObject (function property (x: obj): boolean): obj;
	var
		i: seqIndex0;
	begin
		i := exists(property);
		if i = 0 then
			existsObject := nil
		else
			existsObject := elements^^[i];
	end;

function sequence.arrange (objectsPerLine: longInt): vector;
	var
		result: vector;
		xCount: longInt;
		maxWidth, maxHeight, xPos, yPos: extended;

	procedure findMaxWidthHeight (o: obj);
		var
			r: obj;
		begin
			r := o.getBoundingObject;
			if r <> nil then begin
				if member(r, rectangle) then
					with rectangle(r) do begin
						if right - left > maxWidth then
							maxWidth := right - left;
						if top - bottom > maxHeight then
							maxHeight := top - bottom;
					end;
				r.free;
			end;
		end;

	procedure positionObject (o: obj);
		var
			r: obj;
		begin
			r := o.getBoundingObject;
			if r <> nil then begin
				if member(r, rectangle) then
					with rectangle(r) do
						o.translate(xPos + (maxWidth - (left + right)) / 2, yPos + (maxHeight - (bottom + top)) / 2);
				r.free;
			end;
			xCount := xCount + 1;
			xPos := xPos + maxWidth;
			if xCount = objectsPerLine then begin
				xCount := 0;
				xPos := 0;
				yPos := yPos + maxHeight;
			end;
		end;

	begin (* sequence.arrange *)
		assert(member(self, list) | member(self, vector), 'list or vector expected');
		if member(self, list) then begin
			new(result);
			failNil(Handle(result));
			result.initFromList(list(self));
			free;
		end
		else
			result := vector(self);
		maxWidth := 0;
		maxHeight := 0;
		result.forAll(findMaxWidthHeight);
		maxWidth := maxWidth * 1.05;
		maxHeight := maxHeight * 1.05;
		xPos := 0;
		yPos := 0;
		xCount := 0;
		result.forAll(positionObject);
		arrange := result;
	end;

function getSequence (s: sequence): sequence;
	begin
		with s do
			if (length > 1) | (length = 1) & not member(ith(1), sequence) then
				getSequence := s
			else
				getSequence := sequence(ith(1));
	end;

function nonTrivialCurrentArgumentsIndex (x: extended): boolean;
	begin
		nonTrivialCurrentArgumentsIndex := (x = trunc(x)) & (1 < x) & (x <= getSequence(currentArguments).length);
	end;

function callPointCover (l: obj; isRandom: boolean): obj;
	begin
		if getUserParameter1('Number of points per line primitive:', defaultCover, defaultCover, @nonTrivialSeqIndex) then
			if isRandom then
{$IFC THINK_PASCAL }
				callPointCover := l.randomPointCover(trunc(defaultCover))
			else
				callPointCover := l.pointCover(trunc(defaultCover))
{$ELSEC }
				callPointCover := l.randomPointCover(truncLong(defaultCover))
			else
				callPointCover := l.pointCover(truncLong(defaultCover))
{$ENDC }
		else
			callPointCover := nil;
	end;

procedure getCenterOfGravity (x: obj);
	var
		p: point2d;
		cx, cy: extended;
	begin
		p := point2d(newInitObj(currentPointType));
		x.centerOfGravity(cx, cy);
		p.setXY(cx, cy);
		currentResult.append(p);
	end;

procedure ArrangeObjectsSameCenterOfGravity (s: sequence);
	var
		cx, cy: extended;

	procedure doTranslate (x: obj);
		var
			myCx, myCy: extended;
			copyX: obj;
		begin
			copyX := x.duplicateAll;
			copyX.centerOfGravity(myCx, myCy);
			copyX.translate(cx - myCx, cy - myCy);
			currentResult.append(copyX);
		end;

	begin
		if s.length > 1 then begin
			s.ith(1).centerOfGravity(cx, cy);
			s.forSome(doTranslate, 2, s.length);
		end;
	end;

procedure appendConvexPolygon (x: obj);
	begin
		currentResult.appendNonNil(x.convertToConvexPolygon)
	end;

procedure appendPolygon (x: obj);
	begin
		currentResult.appendNonNil(x.convertToPolygon)
	end;

procedure appendConvertToReal (x: obj);
	begin
		currentResult.appendNonNil(x.convertToType(realPointType));
	end;

procedure appendConvertToFloat (x: obj);
	begin
		currentResult.appendNonNil(x.convertToType(floatPointType));
	end;

procedure appendConvertToLongInt (x: obj);
	begin
		currentResult.appendNonNil(x.convertToType(longIntPointType));
	end;

procedure doPerturb (x: obj);
	var
		y: obj;
	begin
		globalP.randomChange;
		y := x.duplicateAll;
		y.translate(globalP.x, globalP.y);
		currentResult.append(y);
	end;

procedure perturbInRectangle (s: sequence; perturbationWidth, perturbationHeight: extended);
	begin
		if not (s.getType in [convexPolygonType, splinegonType]) then begin
			globalR.right := perturbationWidth / 2;
			globalR.left := -globalR.right;
			globalR.top := perturbationHeight / 2;
			globalR.bottom := -globalR.top;
			currentRandomConstraint := globalR;
			s.forAll(doPerturb);
		end;
	end;

procedure perturbInCircle (s: sequence; perturbationRadius: extended);
	begin
		if not (s.getType in [convexPolygonType, splinegonType]) then begin
			globalC.radius := perturbationRadius;
			globalC.x := 0;
			globalC.y := 0;
			currentRandomConstraint := globalC;
			s.forAll(doPerturb);
		end;
	end;

function duplicateAllFromSequence (t: objectType; s: sequence): vector;
	var
		v: vector;

	procedure doAppend (x: obj);
		begin
			v.append(x.duplicateAll);
		end;

	begin (* duplicateAllFromSequence *)
		currentVectorLength := s.length;
		v := vector(newInitObj(t));
		s.forAll(doAppend);
		duplicateAllFromSequence := v;
	end;

function isPointVector (v: vector): boolean;
	begin
		isPointVector := (v.length > 0) & member(v.ith(1), point2d);
	end;

	const
		AreaC = 0;
		ShowNumberingC = 1;
		RevertC = 2;
		CylicShiftC = 3;
		EliminateDuplicatesC = 4;
		RandomShuffleC = 5;
		RandomChangeC = 6;
		ToPointVectorC = 7;
		PointCoverC = 8;
		EliminateRepeatedElementsC = 9;
		ArrangeObjectsC = 10;
		PerimeterC = 11;
		CenterOfGravityC = 12;
		ToConvexPolygonC = 13;
		ToRealC = 14;
		ToFloatC = 15;
		ToLongIntC = 16;
		ArrangeObjectsSameCenterOfGravityC = 17;
		ToPolygonC = 18;
		RandomPointCoverC = 19;
		PerturbInRectangleC = 20;
		PerturbInCircleC = 21;

procedure sequence.describe;
	begin
		with currentArguments do
			if length > 0 then begin
				menuEntry('Point Cover...', PointCoverC);
				menuEntry('Random Point Cover...', RandomPointCoverC);
				beginSubMenu('Universal Geometry');
				menuEntry('Perimeter', PerimeterC);
				menuEntry('Area', AreaC);
				endSubMenu;
				beginSubMenu('Random');
				menuEntry('Perturb In Rectangle...', PerturbInRectangleC);
				menuEntry('Perturb In Circle...', PerturbInCircleC);
				menuEntry('Shuffle', RandomShuffleC);
				menuEntry('Change', RandomChangeC);
				endSubMenu;
				beginSubMenu('Conversion');
				menuEntry('To Center Of Gravity', CenterOfGravityC);
				menuEntry('To Polygon', ToPolygonC);
				menuEntry('To Convex Polygon', ToConvexPolygonC);
				menuEntry('To Real', ToRealC);
				menuEntry('To Float', ToFloatC);
				menuEntry('To Longint', ToLongIntC);
				if (length = 1) & currentArguments1IsHomogenVector & isPointVector(vector(ith(1))) | (length > 1) & currentArgumentsSameType & member(ith(1), point2d) then
					menuEntry('To Point Vector', ToPointVectorC);
				endSubMenu;
				beginSubMenu('Numbering');
				menuEntry('Show', ShowNumberingC);
				menuEntry('Revert', RevertC);
				menuEntry('Cyclic Shift...', CylicShiftC);
				endSubMenu;
				beginSubMenu('Arrange Objects');
				menuEntry('On Box Grid...', ArrangeObjectsC);
				menuEntry('With Same Center Of Gravity', ArrangeObjectsSameCenterOfGravityC);
				endSubMenu;
				beginSubMenu('Eliminate');
				menuEntry('Repeated Elements', EliminateRepeatedElementsC);
				menuEntry('Duplicates', EliminateDuplicatesC);
				endSubMenu;
			end;
	end;

procedure sequence.execute (item: integer);
	var
		n: seqIndex0;
		v: vector;
		s, t: sequence;
	begin
		with currentArguments do
			case item of
				PerturbInRectangleC: 
					if getUserParameter2('Width', 'Height', perturbationWidth, perturbationHeight, perturbationWidth, perturbationHeight, @isPositiveOrZero, @isPositiveOrZero) then
						perturbInRectangle(getSequence(currentArguments), perturbationWidth, perturbationHeight);
				PerturbInCircleC: 
					if getUserParameter1('Radius', perturbationRadius, perturbationRadius, @isPositiveOrZero) then
						perturbInCircle(getSequence(currentArguments), perturbationRadius);
				CenterOfGravityC: 
					getSequence(currentArguments).forAll(getCenterOfGravity);
				PerimeterC: 
					currentResult.append(getSequence(currentArguments).perimeterMP);
				ToRealC: 
					currentArguments.forAll(appendConvertToReal);
				ToFloatC: 
					currentArguments.forAll(appendConvertToFloat);
				ToLongIntC: 
					currentArguments.forAll(appendConvertToLongInt);
				AreaC: 
					currentResult.append(getSequence(currentArguments).areaMP);
				ToPolygonC: 
					currentArguments.forAll(appendPolygon);
				ToConvexPolygonC: 
					currentArguments.forAll(appendConvexPolygon);
				PointCoverC, RandomPointCoverC: 
					currentResult.appendNonNil(callPointCover(currentArguments, item = RandomPointCoverC));
				ToPointVectorC: 
					currentResult.append(duplicateAllFromSequence(pointVectorType, getSequence(currentArguments)));
				EliminateDuplicatesC:  begin
					v := duplicateAllFromSequence(vectorType, getSequence(currentArguments));
					n := v.length;
					v.eliminateDuplicates;
					if v.length = n then
						v.freeAll
					else
						currentResult.append(v);
				end;
				RandomShuffleC:  begin
					t := getSequence(currentArguments);
					if not (t.getType in [convexPolygonType, splinegonType]) then
						with t do begin
							interactiveOutput;
							interactiveHighlight;
							randomShuffleAll;
							interactiveOutput;
							interactiveHighlight;
							rebuildOperationsMenu := true;
						end;
				end;
				RandomChangeC: 
					with getSequence(currentArguments) do begin
						interactiveOutput;
						interactiveHighlight;
						randomChange;
						interactiveOutput;
						interactiveHighlight;
						rebuildOperationsMenu := true;
					end;
				ShowNumberingC: 
					currentResult.append(getSequence(currentArguments).showNumbering);
				RevertC:  begin
					getSequence(currentArguments).revert;
					rebuildOperationsMenu := true;
				end;
				EliminateRepeatedElementsC:  begin
					t := getSequence(currentArguments);
					s := t.removeRepeatedElements;
					if s.length = t.length then
						s.freeAll
					else if member(s, list) then
						currentResult.concatenateDispose(list(s))
					else
						currentResult.append(s);
				end;
				ArrangeObjectsSameCenterOfGravityC: 
					ArrangeObjectsSameCenterOfGravity(getSequence(currentArguments));
				ArrangeObjectsC: 
					if getUserParameter1('Objects Per Line', defaultObjectsPerLine, defaultObjectsPerLine, @isNatural) then
{$IFC THINK_PASCAL }
						currentResult.append(sequence(getSequence(currentArguments).duplicateAll).arrange(trunc(defaultObjectsPerLine)));
{$ELSEC }
						currentResult.append(sequence(getSequence(currentArguments).duplicateAll).arrange(truncLong(defaultObjectsPerLine)));
{$ENDC }
				CylicShiftC:  begin
					s := getSequence(currentArguments);
					if defaultNewFirst > s.length then
						defaultNewFirst := 2;
					if getUserParameter1('Index of new first element', defaultNewFirst, defaultNewFirst, @nonTrivialCurrentArgumentsIndex) then begin
						if s.getType = polyLineType then begin
							interactiveOutput;
							interactiveHighlight;
{$IFC THINK_PASCAL }
							s.cyclicShift(trunc(defaultNewFirst));
{$ELSEC }
							s.cyclicShift(truncLong(defaultNewFirst));
{$ENDC }
							interactiveOutput;
							interactiveHighlight;
						end
						else
{$IFC THINK_PASCAL }
							s.cyclicShift(trunc(defaultNewFirst));
{$ELSEC }
							s.cyclicShift(truncLong(defaultNewFirst));
{$ENDC }
						rebuildOperationsMenu := true;
					end;
				end;
			end;
	end;

procedure vector.cyclicShift (newFirstElement: seqIndex);
	var
		elementsMoved, cycleStart, delta, current, currentNext: longInt;
		safe: obj;
	begin
		assert((1 <= newFirstElement) & (newFirstElement <= length), 'newFirstElement in range');
		if newFirstElement > 1 then begin
			elementsMoved := 0;
			cycleStart := 0;
			delta := newFirstElement - 1;
			repeat
				cycleStart := cycleStart + 1;
				safe := elements^^[cycleStart];
				current := cycleStart;
				currentNext := current + delta;
				if currentNext > length then
					currentNext := currentNext - length;
				repeat
					elements^^[current] := elements^^[currentNext];
					elementsMoved := elementsMoved + 1;
					current := currentNext;
					currentNext := current + delta;
					if currentNext > length then
						currentNext := currentNext - length;
				until currentNext = cycleStart;
				elements^^[current] := safe;
				elementsMoved := elementsMoved + 1;
			until elementsMoved = length;
		end;
	end;

procedure vector.partialInteractiveFlash (l: seqIndex; r: seqIndex0);
	var
		i: seqIndex0;
		v: vector;
	begin
		if l = 1 then begin
			i := length;
			length := r;
			interactiveFlash;
			length := i;
		end
		else begin
			v := newVector(r - l + 1);
			for i := l to r do
				v.append(elements^^[i]);
			v.interactiveFlash;
			v.free;
		end;
	end;

procedure vector.allocateElements (number: seqIndex0);
	var
		t, tt: objArrayH;
		i: seqIndex0;
	begin
		if number > 0 then
			t := objArrayH(dynamicArray(number + 1, sizeOf(obj)))
		else
			t := nil;
		if elements <> nil then begin
			if length > number then
				length := number;
			for i := 1 to length do
				t^^[i] := elements^^[i];
			tt := elements;
{$IFC THINK_PASCAL }
			disposHandle(Handle(tt));
{$ELSEC }
			DisposeHandle(Handle(tt));
{$ENDC }
		end;
		elements := t;
	end;

function vector.ith (i: seqIndex): obj;
	begin
		ith := elements^^[i];
	end;

procedure vector.deleteElement (i: seqIndex);
	begin
		swap(i, length);
		length := length - 1;
	end;

procedure vector.deleteElementFree (i: seqIndex);
	begin
		swap(i, length);
		elements^^[length].free;
		length := length - 1;
	end;

procedure vector.deleteElementFreeAll (i: seqIndex);
	begin
		swap(i, length);
		elements^^[length].freeAll;
		length := length - 1;
	end;

procedure vector.init;
	begin
		length := 0;
		elements := nil;
		allocateElements(currentVectorLength);
	end;

procedure vector.free;
	begin
		if elements <> nil then
{$IFC THINK_PASCAL }
			disposHandle(Handle(elements));
{$ELSEC }
			DisposeHandle(Handle(elements));
{$ENDC }
		shallowFree;
	end;

function vector.duplicate: obj;
	var
		t: vector;
		i: seqIndex0;
	begin
		t := vector(shallowDuplicate);
		t.elements := nil;
		t.allocateElements(length);
		for i := 1 to length do
			t.elements^^[i] := elements^^[i];
		duplicate := t;
	end;

function vector.duplicateAll: obj;
	var
		t: vector;
		i: seqIndex0;
	begin
		t := vector(shallowDuplicate);
		t.elements := nil;
		t.allocateElements(length);
		for i := 1 to length do
			t.elements^^[i] := elements^^[i].duplicateAll;
		duplicateAll := t;
	end;

procedure vector.readLength;
	var
		l: integer;
		l4: longInt;
	begin
		if currentVersion < version4 then begin
			readInteger(l);
			length := l;
		end
		else begin
			readLongInt(l4);
			length := l4;
		end;
	end;

procedure vector.textFileInput;
	var
		i: seqIndex0;
		l: longInt;
	begin
		readTextLongInt(l);
		length := l;
		elements := nil;
		allocateElements(length);
		for i := 1 to length do begin
{$IFC AnimationEnabled }
			advanceStopWatchHandsNoMem;
{$ENDC }
			elements^^[i] := readTextObj;
		end;
	end;

procedure vector.fileInput;
	var
		i: seqIndex0;
		x: obj;
	begin
		readLength;
		elements := nil;
		allocateElements(length);
		for i := 1 to length do begin
{$IFC AnimationEnabled }
			advanceStopWatchHandsNoMem;
{$ENDC }
			x := newObj(fileGetType);
			x.fileInput;
			elements^^[i] := x;
		end;
	end;

function vector.memberType (o: obj): boolean;
	begin
		memberType := member(o, vector);
	end;

function vector.getType: objectType;
	begin
		getType := vectorType;
	end;

function vector.getTextType: Str255;
	begin
		getTextType := stringOf('Vector(', length : 1, ')');
	end;

procedure vector.initFromList (l: list);
	begin
		initFromListPartial(l, l.length);
	end;

procedure vector.initFromListPartial (l: list; len: seqIndex0);
	var
		e: elementH;
		i: seqIndex0;
	begin
		elements := nil;
		allocateElements(len);
		e := l.first;
		i := 0;
		while (e <> nil) & (i < len) do begin
			i := i + 1;
			elements^^[i] := e^^.value;
			e := e^^.next;
		end;
		length := i;
	end;

procedure vector.forSome (procedure whatToDo (x: obj); l: seqIndex; r: seqIndex0);
	var
		i: seqIndex0;
	begin
		for i := l to r do
			whatToDo(elements^^[i]);
	end;

procedure vector.forAllDown (procedure whatToDo (x: obj));
	var
		i: seqIndex0;
	begin
		for i := length downto 1 do begin
			whatToDo(elements^^[i]);
		end;
	end;

function vector.sameProperty (function property (x: obj): boolean): boolean;
	var
		i: seqIndex1;
	begin
		i := 1;
		while (i <= length) & property(elements^^[i]) do
			i := i + 1;
		sameProperty := i > length;
	end;

function vector.exists (function property (x: obj): boolean): seqIndex0;
	var
		i: seqIndex1;
	begin
		i := 1;
		while (i <= length) & not property(elements^^[i]) do
			i := i + 1;
		if i > length then
			exists := 0
		else
			exists := i;
	end;

procedure vector.forAllPermutations (n: seqIndex0; procedure whatToDo (v: vector));
	var
		i: seqIndex0;
	begin
		if n = 1 then
			whatToDo(self)
		else
			for i := 1 to n do begin
				swap(i, n);
				forAllPermutations(n - 1, whatToDo);
				swap(i, n);
			end;
	end;

function vector.findExtreme (function lessThan (p, q: obj): boolean): seqIndex0;
	var
		t: obj;
		i, ti: seqIndex;
	begin
		if length > 0 then begin
			t := elements^^[1];
			ti := 1;
			for i := 2 to length do
				if lessThan(elements^^[i], t) then begin
					t := elements^^[i];
					ti := i;
				end;
			findExtreme := ti;
		end
		else
			findExtreme := 0;
	end;

procedure vector.sort (function lessThan (p, q: obj): boolean; l: seqIndex; r: seqIndex0);
	const
		cutoff = 10;

	procedure quickSort (function lessThan (p, q: obj): boolean; l: seqIndex; r: seqIndex0; var s: objArray);
		var
			i: seqIndex;
			j: seqIndex0;
			x, t: obj;
		begin
			while cutoff < r - l do begin
				i := l;
				j := r;
				x := s[(l + r) div 2];
				repeat
					while lessThan(s[i], x) do
						i := i + 1;
					while lessThan(x, s[j]) do
						j := j - 1;
					if i <= j then begin
						t := s[i];
						s[i] := s[j];
						s[j] := t;
						i := i + 1;
						j := j - 1;
					end;
				until i > j;
				if j - l > r - i then begin
					quickSort(lessThan, i, r, s);
					r := j;
				end
				else begin
					quickSort(lessThan, l, j, s);
					l := i;
				end;
			end;
		end; (* quickSort *)

	procedure insertionSort (function lessThan (p, q: obj): boolean; l: seqIndex; r: seqIndex0; var s: objArray);
		var
			i, j: seqIndex0;
			t: obj;
		begin
			for i := l + 1 to r do begin
				j := i;
				t := s[j];
				while (j > l) & lessThan(t, s[j - 1]) do begin
					s[j] := s[j - 1];
					j := j - 1;
				end;
				s[j] := t;
			end;
		end;

	begin
		quickSort(lessThan, l, r, elements^^);
		insertionSort(lessThan, l, r, elements^^);
	end;

procedure vector.sortAll (function lessThan (p, q: obj): boolean);
	begin
		sort(lessThan, 1, length);
	end;

procedure vector.append (x: obj);
	begin
		length := length + 1;
		elements^^[length] := x;
	end;

procedure vector.appendSafe (x: obj; increment: seqIndex0);
	var
		allocated: seqIndex0;
	begin
		allocated := allocatedElements;
		if length = allocated then
			if increment = 0 then
				fatalError('Length out of range in vector.appendSafe')
			else begin
				allocateElements(allocated + increment);
				append(x);
			end
		else
			append(x);
	end;

function vector.allocatedElements: seqIndex0;
	begin
		if elements = nil then
			allocatedElements := 0
		else
			allocatedElements := GetHandleSize(Handle(elements)) div sizeOf(obj) - 1;
	end;

procedure vector.swap (i, j: seqIndex);
	var
		t: obj;
	begin
		t := elements^^[i];
		elements^^[i] := elements^^[j];
		elements^^[j] := t;
	end;

procedure vector.revert;
	var
		i: seqIndex;
		j: seqIndex0;
	begin
		i := 1;
		j := length;
		while i < j do begin
			swap(i, j);
			i := i + 1;
			j := j - 1;
		end;
	end;

procedure vector.randomShuffle (n, l: seqIndex0);
	var
		i: seqIndex0;
	begin
		assert(n <= l, 'n  l');
		for i := 1 to n do
			swap(i, randomLongInt mod (l - i + 1) + i);
	end;

procedure list.randomShuffle (n, l: seqIndex0);
	var
		v: vector;
		i: seqIndex0;
		e: elementH;
		t: obj;
	begin
		assert(n <= l, 'n  l');
		if (n = 2) & (l = 2) & (length = 2) then begin
			if randomReal <= 0.5 then begin
				t := first^^.value;
				first^^.value := last^^.value;
				last^^.value := t;
			end;
		end
		else if l > 1 then begin
			new(v);
			failNil(Handle(v));
			v.initFromListPartial(self, l);
			v.randomShuffle(n, l);
			e := first;
			for i := 1 to l do begin
				e^^.value := v.elements^^[i];
				e := e^^.next;
			end;
			v.free;
		end;
	end;

procedure vector.preset (number: seqIndex0; t: objectType);
	var
		i: seqIndex;
		h: objArrayH;
	begin
		if number > 0 then begin
			if elements <> nil then begin
				h := elements;
{$IFC THINK_PASCAL }
				disposHandle(Handle(h));
{$ELSEC }
				DisposeHandle(Handle(h));
{$ENDC }
				elements := nil;
			end;
			allocateElements(number);
			length := number;
			if (t = polygon2dType) | (t = convexPolygonType) | (t = polyLineType) then
				for i := 1 to length do begin
					elements^^[i] := newInitObj(t);
					vector(elements^^[i]).preset(currentVectorLength, currentPointType);
				end
			else
				for i := 1 to length do
					elements^^[i] := newInitObj(t);
		end;
	end;

function vector.sameType: boolean;
	var
		first: obj;
		i: seqIndex1;
	begin
		if length <= 1 then
			sameType := true
		else begin
			first := elements^^[1];
			i := 2;
			while (i <= length) & sameType2(first, elements^^[i]) do
				i := i + 1;
			sameType := i > length;
		end;
	end;

function sequence.convertToType (t: pointTypeRange): obj;
	var
		v: vector;
		l: list;

	procedure appendConverted (x: obj);
		begin
			l.appendNonNil(x.convertToType(t));
		end;

	begin (* sequence.convertToType *)
		l := newEmptyList;
		forAll(appendConverted);
		if member(self, vector) then begin
			v := vector(shallowDuplicate);
			v.initFromList(l);
			l.free;
			convertToType := v;
		end
		else
			convertToType := l;
	end;

function vector.binarySearch (x: obj; function compare (a, b: obj): relationObj; l: seqIndex; r: seqIndex0): seqIndex0;
	var
		rel: relationObj;
		k: seqIndex;
	begin
		if l > r then
			binarySearch := 0
		else begin
			repeat
				k := (l + r) div 2;
				rel := compare(x, elements^^[k]);
				if rel = greaterThanObj then
					l := k + 1
				else
					r := k - 1;
			until (rel = equalObj) | (l > r);
			if rel = equalObj then
				binarySearch := k
			else
				binarySearch := 0;
		end;
	end;

function globalLessThan (a, b: obj): boolean;
	begin
		globalLessThan := (a.getType < b.getType) | (a.getType = b.getType) & a.genericLessThan(b);
	end;

function equalObjects (a, b: obj): boolean;
	begin
		equalObjects := (a.getType = b.getType) & a.equal(b);
	end;

function list.genericLessThan (o: obj): boolean;
	var
		i, j: elementH;
	begin
		i := first;
		j := list(o).first;
		while (i <> nil) & (j <> nil) & equalObjects(i^^.value, j^^.value) do begin
			i := i^^.next;
			j := j^^.next;
		end;
		genericLessThan := (j <> nil) & ((i = nil) | (i <> nil) & globalLessThan(i^^.value, j^^.value));
	end;

function vector.genericLessThan (o: obj): boolean;
	var
		i: seqIndex1;
		v: vector;
	begin
		i := 1;
		v := vector(o);
		while (i <= length) & (i <= v.length) & equalObjects(elements^^[i], v.elements^^[i]) do
			i := i + 1;
		genericLessThan := (i <= v.length) & ((i > length) | (i <= length) & globalLessThan(elements^^[i], v.elements^^[i]));
	end;

function list.equal (o: obj): boolean;
	var
		e1, e2: elementH;
	begin
		if (o <> nil) & member(o, list) & (list(o).length = length) then begin
			e1 := first;
			e2 := list(o).first;
			while (e1 <> nil) & equalObjects(e1^^.value, e2^^.value) do begin
				e1 := e1^^.next;
				e2 := e2^^.next;
			end;
			equal := e1 = nil;
		end
		else
			equal := false;
	end;

function vector.equal (o: obj): boolean;
	var
		i: seqIndex1;
		v: vector;
	begin
		if (o <> nil) & member(o, vector) & (vector(o).length = length) then begin
			i := 1;
			v := vector(o);
			while (i <= length) & equalObjects(elements^^[i], v.elements^^[i]) do
				i := i + 1;
			equal := i > length;
		end
		else
			equal := false;
	end;

procedure vector.eliminateDuplicates;
	var
		v: vector;
		a, b: obj;
		i: seqIndex;
	begin
		if length > 1 then begin
			sortAll(globalLessThan);
			v := newVector(length);
			a := elements^^[1];
			v.append(a);
			for i := 2 to length do begin
				b := elements^^[i];
				if equalObjects(a, b) then
					b.freeAll
				else begin
					a := b;
					v.append(a);
				end;
			end;
			if length = v.length then
				v.free
			else begin
				v.allocateElements(v.length);
{$IFC THINK_PASCAL }
				disposHandle(Handle(elements));
{$ELSEC }
				DisposeHandle(Handle(elements));
{$ENDC }
				elements := v.elements;
				length := v.length;
				dispose(v);
			end
		end;
	end;

procedure homogenVector.fileInput;
	var
		i: seqIndex;
		x: obj;
		t: objectType;
	begin
		readLength;
		elements := nil;
		allocateElements(length);
		if length > 0 then begin
			t := fileGetType;
			for i := 1 to length do begin
				x := newObj(t);
				x.fileInput;
				elements^^[i] := x;
			end;
		end;
	end;

procedure homogenVector.fileOutput;
	var
		i: seqIndex;
	begin
		writeLongInt(length);
		if length > 0 then begin
			elements^^[1].fileOutputTag;
			for i := 2 to length do
				elements^^[i].fileOutput;
		end;
	end;

function homogenVector.equal (o: obj): boolean;
	begin
		equal := (o <> nil) & member(o, homogenVector) & inherited equal(o);
	end;

function homogenVector.memberType (o: obj): boolean;
	begin
		memberType := member(o, homogenVector);
	end;

function homogenVector.getType: objectType;
	begin
		getType := homogenVectorType;
	end;

function stringLength (x: Str255): integer;
	begin
		stringLength := length(x);
	end;

function homogenVector.getTextBaseType: Str255;
	const
		cutoff = 50;
	var
		firstElementType: Str255;
	begin
		if length > 0 then begin
			firstElementType := elements^^[1].getTextType;
			if stringLength(firstElementType) > cutoff then
				firstElementType := concat(copy(firstElementType, 1, cutoff), '...');
			getTextBaseType := stringOf('(', firstElementType, ',', length : 1, ')');
		end
		else
			getTextBaseType := '(0)';
	end;

function homogenVector.drawingComplexity (cutOff: longInt): longInt;
	var
		c: longInt;
	begin
		if length > 0 then begin
			c := ith(1).drawingComplexity(cutOff);
			if c < cutOff then
				drawingComplexity := length * c
			else
				drawingComplexity := c;
		end
		else
			drawingComplexity := 0;
	end;

function homogenVector.memberBaseType (o: obj): boolean;
	begin
		memberBaseType := (length = 0) | (homogenVector(o).length = 0) | elements^^[1].memberType(homogenVector(o).elements^^[1])
	end;

function homogenVector.getTextType: Str255;
	begin
		getTextType := concat('Homogene Vector', getTextBaseType);
	end;

function homogenVector.memberTypeSet (types: objectTypeSet): boolean;
	begin
		memberTypeSet := (length = 0) | (elements^^[1].getType in types);
	end;

function homogenVector.sameType: boolean;
	begin
		sameType := true;
	end;

procedure vector.heapify (function lessThan (p, q: obj): boolean);
	var
		l: seqIndex;
	begin
		l := length div 2 + 1;
		while l > 1 do begin
			l := l - 1;
			sift(lessThan, l, length);
		end;
	end;

procedure vector.sift (function lessThan (p, q: obj): boolean; l, r: seqIndex);
	var
		i, j: seqIndex;
		x: obj;
	begin
		i := l;
		j := 2 * i;
		x := elements^^[i];
		while j <= r do begin
			if j < r then
				if lessThan(elements^^[j + 1], elements^^[j]) then
					j := j + 1;
			if not lessThan(elements^^[j], x) then
				leave;
			elements^^[i] := elements^^[j];
			i := j;
			j := 2 * i;
		end;
		elements^^[i] := x;
	end;

procedure vector.nextMin (function lessThan (p, q: obj): boolean);
	begin
		if length > 0 then begin
			swap(1, length);
			length := length - 1;
			if length > 1 then
				sift(lessThan, 1, length);
		end;
	end;

procedure vector.insert (x: obj; function lessThan (p, q: obj): boolean);
	var
		i, j: seqIndex;
	begin
		append(x);
		i := length;
		while (i > 1) & lessThan(elements^^[i], elements^^[i div 2]) do begin
			j := i div 2;
			swap(i, j);
			i := j;
		end;
	end;

procedure vector.heapSort (function greaterThan (p, q: obj): boolean);
	var
		r: seqIndex0;
	begin
		heapify(greaterThan);
		r := length;
		while r > 1 do begin
			swap(r, 1);
			r := r - 1;
			sift(greaterThan, 1, r);
		end;
	end;

function vector.isHeap (root: seqIndex; function lessThan (p, q: obj): boolean): boolean;
	var
		left, right: longInt;
	begin
		left := 2 * root;
		right := left + 1;
		isHeap := ((left > length) | (not lessThan(elements^^[left], elements^^[root])) & isHeap(left, lessThan)) & ((right > length) | (not lessThan(elements^^[right], elements^^[root])) & isHeap(right, lessThan));
	end;

procedure initVector;
	begin
		defaultNewFirst := 2;
		defaultCover := 10;
		defaultObjectsPerLine := 4;
		perturbationWidth := stdPerturbationWidth;
		perturbationHeight := stdPerturbationHeight;
		perturbationRadius := stdPerturbationRadius;
		new(globalP);
		failNil(Handle(globalP));
		new(globalR);
		failNil(Handle(globalR));
		new(globalC);
		failNil(Handle(globalC));
	end;

procedure exitVector;
	begin
		dispose(globalP);
		dispose(globalR);
		dispose(globalC);
	end;

end.
