(*	XYZ GeoBench (eXperimental geometrY Zurich):		*)
(*	An object oriented, interactive workbench for		*)
(*	experimental geometric computation.					*)
(*														*)
(*	 1999 by Peter Schorn, peter.schorn@acm.org		*)
(*	Culmannstrasse 77, CH-8006 Zrich, Switzerland		*)
(*	This software may be copied freely for educational	*)
(*	and research purposes only. For commercial use,		*)
(*	please contact the author.							*)
(*														*)
(*	Version 5.0.5: 30-Sep-1999							*)

unit voronoiDiagram;

interface

uses
	pointVector;

implementation

{$SETC AEMST = TRUE }
var
	currentPointVector: pointVector; (* vector of 'idRealPoint' of currently read in Voronoi-diagram *)

const
	DelaunayTriangulationC = 1;
	EMSTC = DelaunayTriangulationC + 1;
	AllANNC = EMSTC + 1;
	VoronoiPolygonAreaC = AllANNC + 1;
	GetFiniteVoronoiEdgesC = VoronoiPolygonAreaC + 1;
	GetVoronoiVerticesC = GetFiniteVoronoiEdgesC + 1;

(* begin Marco Ciot *)
	ModifiableDTC = GetVoronoiVerticesC + 1;
	GabrielGraphC = ModifiableDTC + 1;
	RNGfalseC = GabrielGraphC + 1;
(* end Marco Ciot *)

procedure voronoiDiagram.describe;
	begin
		with currentArguments do
			if (length = 1) & member(ith(1), voronoiDiagram) & (voronoiDiagram(ith(1)).edges.length > 0) then begin
				menuEntry('All Nearest Neighbors', AllANNC);
				menuEntry('Delaunay Triangulation', DelaunayTriangulationC);

(* begin Marco Ciot *)
				menuEntry('Modifiable Delaunay Triangulation', ModifiableDTC);
				menuEntry('Gabriel Graph', GabrielGraphC);
				menuEntry('Quick Relative Neighborhood Graph False', RNGfalseC);
(* end Marco Ciot *)

				menuEntry('Euclidean Minimum Spanning Tree', EMSTC);
				menuEntry('Voronoi Polygon Area', VoronoiPolygonAreaC);
				menuEntry('Get Voronoi Vertices', GetVoronoiVerticesC);
				menuEntry('Get Finite Voronoi Edges', GetFiniteVoronoiEdgesC);
			end;
	end;

function voronoiDiagram.drawingComplexity (cutOff: longInt): longInt;
	begin
		drawingComplexity := drawingComplexity2(points, edges, cutOff);
	end;

function voronoiDiagram.pointCover (n: longInt): obj;
	var
		lsv: lineSegmentVector;
		v: vector;
	begin
		lsv := getFiniteEdges;
		v := vector(lsv.pointCover(n));
		lsv.freeAll;
		if v <> nil then
			v.eliminateDuplicates;
		pointCover := v;
	end;

function doVertex (a, b, c: idRealPoint): realPoint;
	var
		d: realPoint;
		cx, cy: extended;
	begin
		a.circleCenterX(b, c, cx, cy);
		new(d);
		failNil(Handle(d));
		d.x := cx;
		d.y := cy;
		doVertex := d;
	end;

function computeVertex (a, b, c: idRealPoint): realPoint;
	begin
		if a.id < b.id then
			if b.id < c.id then
				computeVertex := doVertex(a, b, c)
			else if a.id < c.id then
				computeVertex := doVertex(a, c, b)
			else
				computeVertex := doVertex(c, a, b)
		else if a.id < c.id then
			computeVertex := doVertex(b, a, c)
		else if b.id < c.id then
			computeVertex := doVertex(b, c, a)
		else
			computeVertex := doVertex(c, b, a);
	end;

function voronoiDiagram.getFiniteEdges: lineSegmentVector;
	var
		l: list;
		lsv: lineSegmentVector;

	procedure getEdge (e: voronoiEdge);
		var
			ls: lineSegment;
		begin
			with e do begin
				if eType = segment then begin
					new(ls);
					failNil(Handle(ls));
					ls.p := computeVertex(idRealPoint(p1), idRealPoint(p2), idRealPoint(v1));
					ls.q := computeVertex(idRealPoint(p1), idRealPoint(p2), idRealPoint(v2));
					l.append(ls);
				end;
			end;
		end;

	begin (* voronoiDiagram.getFiniteEdges *)
		l := newEmptyList;
		edges.forAll(getEdge);
		new(lsv);
		failNil(Handle(lsv));
		lsv.initFromList(l);
		l.free;
		getFiniteEdges := lsv;
	end;


function voronoiDiagram.getVertices: pointVector;
	var
		result: pointVector;

	procedure getVertex (e: voronoiEdge);
		begin
			if e.eType = ray then
				result.append(computeVertex(idRealPoint(e.p1), idRealPoint(e.p2), idRealPoint(e.v1)));
			if e.eType = segment then
				result.append(computeVertex(idRealPoint(e.p1), idRealPoint(e.p2), idRealPoint(e.v2)));
		end;

	begin
		result := newPointVector(2 * edges.length);
		edges.forAll(getVertex);
		result.eliminateDuplicates;
		getVertices := result;
	end;

function callVoronoiPolygonArea (v: voronoiDiagram): homogenVector;
	var
		r, rv: homogenVector;
		mp: markedRealPoint;
		x: real;
		i: seqIndex0;
	begin
		rv := v.polygonArea;
		r := newHomogenVector(v.points.length);
		r.length := v.points.length;
		for i := 1 to v.points.length do begin
			new(mp);
			failNil(Handle(mp));
			mp.init;
			with realPoint(v.points.elements^^[i]) do begin
				mp.x := x;
				mp.y := y;
			end;
			x := realNumber(rv.elements^^[i]).value;
			if x = -1.0 then
				mp.mark.value := ''
			else
				mp.mark.value := stringOf(x : 8 : 3);
			r.elements^^[i] := mp;
		end;
		rv.freeAll;
		callVoronoiPolygonArea := r;
	end;


(* begin Marco Ciot *)

function executeModifiablesDTGG (vd: voronoiDiagram; item: integer): obj;
	begin
{$IFC AnimationEnabled }
		if (animationFlag[ARelativeNeighborhoodItem]) then begin
			vd.interactiveHighlight;	(* unhighlight voronoiDiagram: precondition for methods! *)
		end;
{$ENDC }
		case item of
			ModifiableDTC: 
				executeModifiablesDTGG := vd.modifiableDT;
			GabrielGraphC: 
				executeModifiablesDTGG := vd.gabrielGraph;
			RNGfalseC: 
				executeModifiablesDTGG := vd.RNGfalse;
		end;
	end;
(* end Marco Ciot *)

procedure voronoiDiagram.execute (item: integer);
	var
		uf: unionFindSimple;
	begin
		with currentArguments do
			case item of
				DelaunayTriangulationC: 
					currentResult.append(voronoiDiagram(ith(1)).delaunayTriangulation);

(* begin Marco Ciot *)
				ModifiableDTC, GabrielGraphC, RNGfalseC: 
					currentResult.appendNonNil(executeModifiablesDTGG(voronoiDiagram(ith(1)), item));
(* end Marco Ciot *)

				EMSTC: 
					currentResult.appendNonNil(voronoiDiagram(ith(1)).EMST(1, uf));
				AllANNC: 
					currentResult.appendNonNil(voronoiDiagram(ith(1)).nearestNeighborsVoronoi);
				VoronoiPolygonAreaC: 
					currentResult.append(callVoronoiPolygonArea(voronoiDiagram(ith(1))));
				GetFiniteVoronoiEdgesC: 
					currentResult.append(voronoiDiagram(ith(1)).getFiniteEdges);
				GetVoronoiVerticesC: 
					currentResult.append(voronoiDiagram(ith(1)).getVertices);
			end;
	end;

function voronoiDiagram.getInfo: obj;
	var
		v: vector;
	begin
		v := newVector(2);
		appendObjInfoObject(v, 'Points', points, true);
		appendObjInfoObject(v, 'Edges', edges, true);
		getInfo := v;
	end;

function voronoiDiagram.equal (o: obj): boolean;
	begin
		equal := (o <> nil) & member(o, voronoiDiagram) & points.equal(voronoiDiagram(o).points) & edges.equal(voronoiDiagram(o).edges);
	end;

procedure voronoiDiagram.centerOfGravity (var cx, cy: extended);
	begin
		points.centerOfGravity(cx, cy);
	end;

function voronoiDiagram.genericLessThan (o: obj): boolean;
	var
		v: voronoiDiagram;
	begin
		v := voronoiDiagram(o);
		genericLessThan := points.genericLessThan(v.points) | (points.equal(v.points) & (edges.genericLessThan(v.edges) | edges.equal(v.edges)));
	end;

procedure voronoiDiagram.textFileInput;
	begin
		points := pointVector(readTextObj);
		currentPointVector := points;
		edges := list(readTextObj);
	end;

procedure voronoiDiagram.textFileOutput;
	begin
		writeTextObj(points);
		writeTextObj(edges);
	end;

procedure voronoiDiagram.fileInput;
	var
		l: list;
		p: pointVector;
	begin
		assert(currentVersion >= version7, 'Correct version for Voronoi-Diagram');
		p := pointVector(newObj(fileGetType));
		points := p;
		points.fileInput;
		currentPointVector := p;
		new(l);
		failNil(Handle(l));
		edges := l;
		edges.fileInput;
	end;

procedure voronoiDiagram.fileOutput;
	begin
		points.fileOutputTag;
		edges.fileOutput;
	end;

function voronoiDiagram.duplicate: obj;
	var
		v: voronoiDiagram;
	begin
		new(v);
		failNil(Handle(v));
		v.points := pointVector(points.duplicate);
		v.edges := list(edges.duplicateAll);
		duplicate := v;
	end;

function voronoiDiagram.duplicateAll: obj;
	var
		v: voronoiDiagram;

	procedure fixReferences (e: voronoiEdge);
		begin
			with e do begin
				p1 := idRealPoint(v.points.elements^^[idRealPoint(p1).id]);
				p2 := idRealPoint(v.points.elements^^[idRealPoint(p2).id]);
				if eType >= ray then
					v1 := idRealPoint(v.points.elements^^[idRealPoint(v1).id]);
				if eType = segment then
					v2 := idRealPoint(v.points.elements^^[idRealPoint(v2).id]);
			end;
		end;

	begin
		new(v);
		failNil(Handle(v));
		v.points := pointVector(points.duplicateAll);
		v.edges := list(edges.duplicateAll);
		v.edges.forAll(fixReferences);
		duplicateAll := v;
	end;

procedure voronoiDiagram.free;
	begin
		points.free;
		edges.freeAll;
		shallowFree;
	end;

procedure voronoiDiagram.freeAll;
	begin
		points.freeAll;
		edges.freeAll;
		shallowFree;
	end;

function voronoiDiagram.isSelected (where: obj): boolean;
	begin
		isSelected := points.isSelected(where);
	end;

procedure voronoiDiagram.translate (dx, dy: extended);
	begin
		points.translate(dx, dy);
	end;

procedure voronoiDiagram.rotate (xOrigin, yOrigin, alpha: extended);
	begin
		points.rotate(xOrigin, yOrigin, alpha);
	end;

procedure voronoiDiagram.scale (lambda: extended);
	begin
		points.scale(lambda);
	end;

procedure voronoiDiagram.reflect (a, b, c: extended);
	begin
		points.reflect(a, b, c);
	end;

procedure voronoiDiagram.interactiveOutput;
	const
		safeX = 50;
		safeY = 50;
	var
		save: clipper;
		d: boolean;
		xmin, ymin, xmax, ymax: extended;
	begin
		points.interactiveOutput;
		save := clippingRectangle;
{$IFC THINK_PASCAL }
		if thePort^.pnVis < 0 then begin (* if not drawing to a window, but e.g. to the scrap *)
{$ELSEC }
		if qd.thePort^.pnVis < 0 then begin (* if not drawing to a window, but e.g. to the scrap *)
{$ENDC }
(* determine the bounding rectangle of the points *)
			with points do begin
				xmin := point2d(elements^^[findExtreme(pointLessThanXY)]).getX;
				ymin := point2d(elements^^[findExtreme(pointLessThanYX)]).getY;
				xmax := point2d(elements^^[findExtreme(pointGreaterThanXY)]).getX;
				ymax := point2d(elements^^[findExtreme(pointGreaterThanYX)]).getY;
			end;
			with clippingRectangle do begin
				d := worldToWindow(xmin, ymin, left, bottom);
				d := worldToWindow(xmax, ymax, right, top);
				left := left - safeX;
				bottom := bottom - safeY;
				right := right + safeX;
				top := top + safeY;
			end;
(* if the window rectangle was bigger when the user selected 'Copy': extend it *)
{$IFC THINK_PASCAL }
			with thePort^.portRect do begin
{$ELSEC }
			with qd.thePort^.portRect do begin
{$ENDC }
				if left < clippingRectangle.left then
					clippingRectangle.left := left;
				if right > clippingRectangle.right then
					clippingRectangle.right := right;
				if bottom > clippingRectangle.top then
					clippingRectangle.top := bottom;
				if top < clippingRectangle.bottom then
					clippingRectangle.bottom := top;
			end;
		end;
		edges.interactiveOutput;
		clippingRectangle := save;
	end;

procedure voronoiDiagram.interactiveHighlight;
	begin
		points.interactiveHighlight;
	end;

function voronoiDiagram.memberType (o: obj): boolean;
	begin
		memberType := member(o, voronoiDiagram);
	end;

function voronoiDiagram.getType: objectType;
	begin
		getType := voronoiDiagramType;
	end;

function voronoiDiagram.getTextType: Str255;
	begin
		getTextType := stringOf('Voronoi Diagram(V=', points.length : 1, ', E=', edges.length : 1, ')');
	end;

procedure updateArea (ap1, ap2: realNumber; e: voronoiEdge);
	var
		p, q: realPoint;
	begin
		e.toPoints(p, q, 0);
		if ap1.value <> -1.0 then
			ap1.value := ap1.value + abs(e.p1.whichSideX(p, q)) / 2.0;
		if ap2.value <> -1.0 then
			ap2.value := ap2.value + abs(e.p2.whichSideX(p, q)) / 2.0;
		p.free;
		q.free;
	end;

function voronoiDiagram.polygonArea: homogenVector;
	var
		r: realNumber;
		rv: homogenVector;
		i: seqIndex0;

	procedure processEdge (e: voronoiEdge);
		var
			ap1, ap2: realNumber;
		begin
			ap1 := realNumber(rv.elements^^[idRealPoint(e.p1).id]);
			ap2 := realNumber(rv.elements^^[idRealPoint(e.p2).id]);
			if e.eType = segment then
				updateArea(ap1, ap2, e)
			else begin
				ap1.value := -1.0;
				ap2.value := -1.0;
			end;
		end;

	begin
		rv := newHomogenVector(points.length);
		rv.length := points.length;
		for i := 1 to points.length do begin
			new(r);
			failNil(Handle(r));
			r.value := 0.0;
			rv.elements^^[i] := r;
		end;
		edges.forAll(processEdge);
		polygonArea := rv;
	end;

function voronoiDiagram.delaunayTriangulation: simpleUndirectedGraph;
	var
		sug: simpleUndirectedGraph;

	procedure createGraphEdge (x: voronoiEdge);
		var
			ge: graphEdge;
		begin
			new(ge);
			failNil(Handle(ge));
			ge.startVertex := idRealPoint(x.p1).id;
			ge.endVertex := idRealPoint(x.p2).id;
			sug.edges.append(ge);
		end;

	begin
		new(sug);
		failNil(Handle(sug));
		sug.vertices := vector(points.duplicateAll);
		sug.edges := newHomogenVector(edges.length);
		edges.forAll(createGraphEdge);
		delaunayTriangulation := sug;
	end;

function voronoiDiagram.EMST (k: seqIndex; var uf: unionFindSimple): spanningTree; (* Euclidean minimum spanning tree *)
	var
		g: simpleUndirectedGraph;

	function edgeLessThan (e1, e2: graphEdge): boolean;
		begin
			edgeLessThan := realPoint(g.vertices.elements^^[e1.startVertex]).squaredDistanceX(realPoint(g.vertices.elements^^[e1.endVertex])) < realPoint(g.vertices.elements^^[e2.startVertex]).squaredDistanceX(realPoint(g.vertices.elements^^[e2.endVertex]));
		end;

	begin
		g := delaunayTriangulation;
		g.edges.sortAll(edgeLessThan);
{$IFC AnimationEnabled AND AEMST }
		if animationFlag[AMinimumSpanningTreeItem] then
			edges.interactiveOutput;
{$ENDC }
		EMST := g.minimumSpanningTree(k, uf);
		g.freeAll;
	end;

function voronoiDiagram.nearestNeighborsVoronoi: lineSegmentVector;
	var
		h: lineSegmentVector;
		l: lineSegment;
		pp, qq: realPoint;
		e: elementH;
		d: extended;
	begin
		h := points.initNearestNeighbors(allNeighbors);
		e := edges.first;
		while e <> nil do begin
{$IFC AnimationEnabled }
			if userAbort then begin
				h.nilQ;
				h.freeAll;
				nearestNeighborsVoronoi := nil;
				exit(nearestNeighborsVoronoi);
			end;
{$ENDC }
			with voronoiEdge(e^^.value) do begin
				pp := realPoint(points.elements^^[idRealPoint(p1).id]);
				qq := realPoint(points.elements^^[idRealPoint(p2).id]);
				l := lineSegment(h.elements^^[idRealPoint(p1).id]);
				d := qq.distanceX(pp);
				if (l.q = nil) | (d < l.length) then
					l.q := qq;
				l := lineSegment(h.elements^^[idRealPoint(p2).id]);
				if (l.q = nil) | (d < l.length) then
					l.q := pp;
			end;
			e := e^^.next;
		end;
		h.finishNearestNeighbors(allNeighbors);
		nearestNeighborsVoronoi := h;
	end;

	const
		maxCoord = 100000; (* for calculating *)
		maxCoordDrawing = maxInt; (* for drawing *)

function voronoiEdge.getTextType: Str255;
	begin
		getTextType := 'Voronoi Edge';
	end;

function voronoiEdge.getInfo: obj;
	var
		v: vector;
	begin
		v := newVector(3 + ord(eType));
{$IFC THINK_PASCAL }
		appendObjInfoString(v, 'Edge Type', stringOf(eType), true);
{$ELSEC }
		appendObjInfoString(v, 'Edge Type', stringOf(ord(eType)), true);
{$ENDC }
		appendObjInfoObject(v, 'P1', p1, true);
		appendObjInfoObject(v, 'P2', p2, true);
		if eType >= ray then
			appendObjInfoObject(v, 'V1', v1, true);
		if eType = segment then
			appendObjInfoObject(v, 'V2', v2, true);
		getInfo := v;
	end;

{$IFC NOT THINK_PASCAL }
{$PUSH }
{$PRAGMAC warn_unusedarg off }
{$ENDC }
function voronoiEdge.drawingComplexity (cutOff: longInt): longInt;
	begin
		drawingComplexity := 3;
	end;
{$IFC NOT THINK_PASCAL }
{$POP }
{$ENDC }

function voronoiEdge.equal (o: obj): boolean;
	var
		e: voronoiEdge;
	begin
		if (o <> nil) & member(o, voronoiEdge) then begin
			e := voronoiEdge(o);
			equal := (eType = e.eType) & p1.equal(e.p1) & p2.equal(e.p2) & ((eType < ray) | (eType >= ray) & v1.equal(e.v1) & ((eType <> segment) | (eType = segment) & v2.equal(e.v2)));
		end
		else
			equal := false;
	end;

function voronoiEdge.genericLessThan (o: obj): boolean;
	var
		e: voronoiEdge;
	begin
		e := voronoiEdge(o);
		genericLessThan := (eType < e.eType) | (eType = e.eType) & (p1.genericLessThan(e.p1) | p1.equal(e.p1) & (p2.genericLessThan(e.p2) | (eType >= ray) & p2.equal(e.p2) & (v1.genericLessThan(e.v1) | (eType = segment) & v1.equal(e.v1) & v2.genericLessThan(e.v2))));
	end;

procedure voronoiEdge.textOutput;
	begin
{$IFC THINK_PASCAL }
		write(eType, ' ');
{$ELSEC }
		write(ord(eType), ' ');
{$ENDC }
		idRealPoint(p1).textOutput;
		p2.textOutput;
		if eType >= ray then
			v1.textOutput;
		if eType = segment then
			v2.textOutput;
	end;

procedure voronoiEdge.textFileInput;
	var
		x: longInt;
	begin
		readTextLongInt(x);
		eType := voronoiEdgeT(x);
		readTextLongInt(x);
		p1 := idRealPoint(currentPointVector.elements^^[x]);
		readTextLongInt(x);
		p2 := idRealPoint(currentPointVector.elements^^[x]);
		if eType >= ray then begin
			readTextLongInt(x);
			v1 := idRealPoint(currentPointVector.elements^^[x]);
		end;
		if eType = segment then begin
			readTextLongInt(x);
			v2 := idRealPoint(currentPointVector.elements^^[x]);
		end;
	end;

procedure voronoiEdge.textFileOutput;
	begin
		writeTextLongInt(ord(eType));
		writeTextLongInt(idRealPoint(p1).id);
		writeTextLongInt(idRealPoint(p2).id);
		if eType >= ray then
			writeTextLongInt(idRealPoint(v1).id);
		if eType = segment then
			writeTextLongInt(idRealPoint(v2).id);
	end;

procedure voronoiEdge.fileInput;
	var
		b: Byte;
		x: longInt;
	begin
		readByte(b);
		eType := voronoiEdgeT(b);
		readLongInt(x);
		p1 := idRealPoint(currentPointVector.elements^^[x]);
		readLongInt(x);
		p2 := idRealPoint(currentPointVector.elements^^[x]);
		if eType >= ray then begin
			readLongInt(x);
			v1 := idRealPoint(currentPointVector.elements^^[x]);
		end;
		if eType = segment then begin
			readLongInt(x);
			v2 := idRealPoint(currentPointVector.elements^^[x]);
		end;
	end;

procedure voronoiEdge.fileOutput;
	begin
		writeByte(Byte(eType));
		writeLongInt(idRealPoint(p1).id);
		writeLongInt(idRealPoint(p2).id);
		if eType >= ray then
			writeLongInt(idRealPoint(v1).id);
		if eType = segment then
			writeLongInt(idRealPoint(v2).id);
	end;

function voronoiEdge.getType: objectType;
	begin
		getType := voronoiEdgeType;
	end;

function voronoiEdge.memberType (o: obj): boolean;
	begin
		memberType := member(o, voronoiEdge);
	end;

function voronoiEdge.definingPoints (p, q: realPoint): boolean;
	begin
		definingPoints := (p1 = p) & (p2 = q) | (p1 = q) & (p2 = p)
	end;

function voronoiEdge.bisecX (y: real): real;
	begin
		if p1.x = p2.x then
			bisecX := maxCoord
		else
			bisecX := ((((p1.x + p2.x) * (p1.x - p2.x) / 2.0) + ((p1.y + p2.y) * (p1.y - p2.y) / 2.0)) - (y * (p1.y - p2.y))) / (p1.x - p2.x);
	end;

function voronoiEdge.bisecY (x: real): real;
	begin
		if p1.y = p2.y then
			bisecY := maxCoord
		else
			bisecY := ((((p1.x + p2.x) * (p1.x - p2.x) / 2.0) + ((p1.y + p2.y) * (p1.y - p2.y) / 2.0)) - (x * (p1.x - p2.x))) / (p1.y - p2.y);
	end;


(*************************************************************	*)
(****** Silke Schweifer, Universitt Mnster, 27.3.93 ************************	*)
(*************************************************************	*)
(******* Kommentar zur Vernderung der Prozedur 'voronoiEdge.toPoints' ***********	*)
(*																													*)
(* Problem: Colineare Verteilung von n Punkten 															*)
(*																													*)
(* Bei einer colinearen Verteilung von n Datenpunkten besteht das zugehrige Voronoi- 			*)
(* Diagramm aus n-1 Geraden, die die Mittelsenkrechten zwischen jeweils zwei benachbarten 	*)
(* Datenpunkten auf der Linie darstellen. Die Kanten sind also insbesondere unbegrenzt und 		*)
(* liegen parallel.																								*)
(*																													*)
(* Auf Grund von Rundungsfehlern kann es nun aber passieren, da sich im Verlauf des Algo-		*)
(* rithmus die (bzw. einige) transformierte Bisektoren schneiden, und demnach bei den ent-		*)
(* sprechenden Voronoi Kanten Halbgerade, aber auch Segmente vorkommen knnen. Da die		*)
(* flschlicherweise entstandenen Knoten aber in (sehr) groer Entfernung von den Daten-	 	*)
(* punkten liegen, wirken zum einen die gezeichneten Bisektoren trotzdem parallel, zum 			*)
(* anderen, wenn der Algorithmus korrekt wre, wrden auch alle ntigen Kanten vorhanden 	*)
(* sein, so da diese Rundungsfehler nicht von Bedeutung wren.										*)
(*																													*)
(* Liegen nun gengend viele Datenpunkte colinear und hinreichend dicht beieinander, werden 		*)
(* zwar einige Kanten an den Seiten gezeichnet, die Kanten zwischen den mittleren Daten-			*)
(* punkten werden aber nicht dargestellt. Diese Kanten sind (auf Grund des Rundungsfehlers) 		*)
(* vom Typ 'Segment',	haben also an beiden Seiten einen Endpunkt.										*)
(*																													*)
(* In dem Algorithmus wird ein rechnerischer Bildausschnitt (geg. durch 'maxCoord') und ein,	*)
(* etwas grerer, zeichnerischer Bildausschnitt (geg. durch 'maxCoordDrawing') zugrunde 		*)
(* gelegt. In dem letztgenannten Ausschnit erscheint spter das erzeugte VoronoiDiagramm.		*)
(*																													*)
(* Betrachten wir nun den Fall, da fr jeweils zwei collineare Datenpunkte p1 und p2 gilt:		*)
(*                abs(p1.x - p2.x) > abs(p1.y - p2.y)      (das Analoge gilt fr den anderen Fall)		*)
(*																													*)
(* e sei eine Kante, die flschlicherweise vom Typ 'segment' ist. Dieses Segment liegt "fast"		*)
(* parallel zu seinen Nachbarkanten, d.h. die Endpunkte p und q liegen weit auerhalb des (rela-	*)
(* tiv kleinen) Bildausschnitts, der fr die Gleichnungen vorgegeben ist.								*)
(*																													*)
(* Es gilt also  abs(p.y) >> maxCoordDrawing         (A)													*)
(*				  abs(q.y) >> maxCoordDrawing																*)
(*																													*)
(* In der Ausgabeprozedur 'interactiveOutput' werden, falls die jeweilige Kante ein Segment 		*)
(* ist, lediglich die zugehrigen Endpunkte p und q berechnet (in der Prozedur 'toPoints').			*)
(* Wegen (A) ist die 'if'- Bedingung in der Prozedur 'interactivOutput' nicht erfllt und das		*)
(* Segment wird nicht gezeichnet, obwohl es eine notwendige Voronoi Kante ist.						*)
(*																													*)
(* Um dies zu unterbinden wurde in der Prozedur 'toPoints' eine weitere Fallunterscheidung 		*)
(* eingebaut:																									*)
(*																													*)
(* Im Fall der rein colinearen Verteilung von Punkten ist ein Segment nur von Bedeutung, wenn	*)
(* die Endpunkte auf verschiedenen Seiten der "Linie" liegen, d.h. liegen beide Endpunkte auf		*)
(* einer Seite, luft der Algorithmus wie gewohnt weiter.												*)
(*																													*)
(* Also: Abfrage, ob die Endpunkte (eines Segments) auf verschiedenen Seiten liegen.				*)
(*																													*)
(* Wenn ja, und nur dann, werden die Punkte des entsprechenden Bisektors an der maxCoord -	*)
(* Grenze berechnet (mittels 'bisecX' und 'bisecY'). Da maxCoord < maxCoordDrawing gengen	*)
(* diese "neuen" Endpunkte der 'if'- Bedingung in der Ausgabeprozedur und die erforderliche 		*)
(* Voronoi Kante wird korrekt gezeichnet.																	*)
(*																													*)
(*************************************************************	*)

procedure voronoiEdge.toPoints (var p, q: realPoint; maxCoord: real);
	var
		flat: boolean;
		int1x, int1y, int2x, int2y, intX, intY: extended;
		c: realPoint;
	begin
		new(p);
		failNil(Handle(p));
		new(q);
		failNil(Handle(q));
		if not p1.equalPoint(p2) then begin
			if eType = segment then begin
				v1.circleCenterX(p1, p2, int1x, int1y);
				p.x := int1x;
				p.y := int1y;
				v2.circleCenterX(p1, p2, int2x, int2y);
				q.x := int2x;
				q.y := int2y;

(************************************************************************	*)
(****** Silke Schweifer, Universitt Mnster, 27.3.93 ***********************************	*)
(************************************************************************	*)
				flat := abs(p1.x - p2.x) < abs(p1.y - p2.y);
				if flat & (abs(p.x) > maxCoord) & (abs(q.x) > maxCoord) then begin
					if (p.x < -maxCoord) & (q.x > maxCoord) then begin
						p.x := -maxCoord;
						q.x := maxCoord
					end
					else if (p.x > maxCoord) & (q.x < -maxCoord) then begin
						p.x := maxCoord;
						q.x := -maxCoord
					end;
					p.y := bisecY(p.x);
					q.y := bisecY(q.x);
				end;
				if not flat & (abs(p.y) > maxCoord) & (abs(q.y) > maxCoord) then begin
					if (p.y < -maxCoord) & (q.y > maxCoord) then begin
						p.y := -maxCoord;
						q.y := maxCoord
					end
					else if (p.x > maxCoord) & (q.x < -maxCoord) then begin
						p.y := maxCoord;
						q.y := -maxCoord
					end;
					p.x := bisecX(p.y);
					q.x := bisecX(q.y);
				end;
(************************************************************** *)

			end
			else begin
				flat := abs(p1.x - p2.x) < abs(p1.y - p2.y);
				if flat then begin
					int1x := -maxCoord;
					int1y := bisecY(int1x);
					int2x := maxCoord;
					int2y := bisecY(int2x);
				end
				else begin
					int1y := -maxCoord;
					int1x := bisecX(int1y);
					int2y := maxCoord;
					int2x := bisecX(int2y);
				end;
				if eType = bisector then begin
					p.x := int1x;
					p.y := int1y;
					q.x := int2x;
					q.y := int2y;
				end
				else  (* =>  eType = ray *)
					begin
					new(c);
					failNil(Handle(c));
					c.x := int1x;
					c.y := int1y;
					v1.circleCenterX(p1, p2, intX, intY);
					p.x := intX;
					p.y := intY;
					if sign(v1.whichSideX(p1, p2)) = sign(c.whichSideX(p1, p2)) then begin
						q.x := int2x;
						q.y := int2y;
					end
					else begin
						q.x := int1x;
						q.y := int1y;
					end;
					dispose(c);
				end;
			end;
		end;
	end;

function valueInRange (val: real; range: integer): real;
	begin
		if val > range then
			valueInRange := range
		else if val < -range then
			valueInRange := -range
		else
			valueInRange := val
	end;

procedure voronoiEdge.interactiveOutput;
	var
		p, q: realPoint;
	begin
		toPoints(p, q, 10000);
		if (abs(p.x) <= maxCoordDrawing) & (abs(p.y) <= maxCoordDrawing) | (abs(q.x) <= maxCoordDrawing) & (abs(q.y) <= maxCoordDrawing) then begin
			while (abs(p.x) > maxCoordDrawing) | (abs(p.y) > maxCoordDrawing) do begin
				p.x := (p.x + q.x) / 2;
				p.y := (p.y + q.y) / 2;
			end;
			while (abs(q.x) > maxCoordDrawing) | (abs(q.y) > maxCoordDrawing) do begin
				q.x := (p.x + q.x) / 2;
				q.y := (p.y + q.y) / 2;
			end;
			p.drawLine(q);
		end;
		dispose(p);
		dispose(q);
	end;

procedure voronoiEdge.commonPoints (e: voronoiEdge; var singlePoint1, doublePoint, singlePoint2: realPoint; var ok: boolean);
	begin
		ok := true;
		if p1 = e.p1 then begin
			singlePoint1 := p2;
			doublePoint := p1;
			singlePoint2 := e.p2;
		end
		else if p1 = e.p2 then begin
			singlePoint1 := p2;
			doublePoint := p1;
			singlePoint2 := e.p1;
		end
		else if p2 = e.p1 then begin
			singlePoint1 := p1;
			doublePoint := p2;
			singlePoint2 := e.p2;
		end
		else if p2 = e.p2 then begin
			singlePoint1 := p1;
			doublePoint := p2;
			singlePoint2 := e.p1;
		end
		else
			ok := false;
	end;

function voronoiEdge.otherPointOfEdge (p: realPoint): realPoint;
	begin
		if p1 = p then
			otherPointOfEdge := p2
		else
			otherPointOfEdge := p1
	end;

end.
