(*	XYZ GeoBench (eXperimental geometrY Zurich):		*)
(*	An object oriented, interactive workbench for		*)
(*	experimental geometric computation.					*)
(*														*)
(*	 1999 by Peter Schorn, peter.schorn@acm.org		*)
(*	Culmannstrasse 77, CH-8006 Zrich, Switzerland		*)
(*	This software may be copied freely for educational	*)
(*	and research purposes only. For commercial use,		*)
(*	please contact the author.							*)
(*														*)
(*	Version 5.0.5: 30-Sep-1999							*)

(* Silke Schweifer, Universitt Mnster, 29.3.93		*)

unit voronoiDiagramSLSeidel;

(* A topological sweepline algorithm for computing the Voronoi Diagram *)


{$SETC AvoronoiDiagramSL = TRUE }

interface

uses
	pointVector;

implementation

const
	absTol = 0.0001;
	Tneg = -10000.0;		(* Konstanten, die den Anfangs- bzw. Endpunkt der topologischen Wellenfront im		*)
	Tpos = 10000.0;			(* Unendlichen darstellen.															*)

type
	edgeSLSeidel = object(voronoiEdge)
		end;

	activeType = (wave, newPoint);
	activeElSeidel = object(obj)
			aType: activeType;
			p: realPoint;      			(*	ist aType =  newPoint, dann ist p der Datenpunkt, der das Wellenstck definiert *)
			side: boolean;          	(* side=true <=> e.p1 ist der Datenpunkt, der die Welle definiert.					*)
			r, q, s: realPoint; 		(* nur wenn aType=vertex, (r,q,s) entspricht dem Tripel aus L, welches den  		*)
										(* Knoten definiert 																*)
			e: edgeSLSeidel;			(* nur wenn aType=wave, e gibt den entsprechenden Bisektoren und die 				*)
										(* zugehoerigen Voronoi-Knoten wieder														*)
			procedure textOutput;
			override;
		end;

	queueType = (site, vertex);
	queueElSeidel = object(obj)
			qType: queueType;
			r, q, s: realPoint;			(* qType = site => nur q belegt, mit q=Datenpunkt 										*)
										(* qType = vertex => Tripel (r,q,s) in L definiert den Knoten 							*)
			v: realPoint;				(* qType = vertex:  Mittelpunkt des Kreises durch (r,q,s )= Knoten und Endpunkt  		*)
										(* der Bisektoren Brq, Bqs und Brs 														*)
			yTransf: real;				(* qType = vertex: y-Koordinate der "obersten Spitze" der Kreises C(v)					*)
			zq: reference;  			(* qType = vertex: verweist auf die Welle in L, die durch den Datenpunkt q definiert	*)
			procedure free;				(* ist	*)
			override;
			procedure freeAll;
			override;
			procedure textOutput;		(* implementation correct if queue is avlTree or sortedList *)
			override;
			procedure interactiveOutput;
			override;
		end;

procedure activeElSeidel.textOutput;
	begin
{$IFC THINK_PASCAL }
		write(atype, ' ');
{$ELSEC }
		write(ord(aType), ' ');
{$ENDC }
		if aType = wave then begin
			e.textOutput;
			p.textOutput;
			write(side);
			r.textOutput;
			q.textOutput;
			s.textOutput;
		end
		else
			p.textOutput;
	end;

procedure queueElSeidel.textOutput;
	begin
{$IFC THINK_PASCAL }
		write(qType, ' ');
{$ELSEC }
		write(ord(qType), ' ');
{$ENDC }
		q.textOutput;
		if qType = vertex then begin
			r.textOutput;
			s.textOutput;
			v.textOutput;
			if zq <> nil then
				avlNodeH(zq)^^.value.textOutput;
			write(yTransf : 4 : 1, ' ');
		end;
	end;

procedure queueElSeidel.free;
	begin
		shallowFree;
	end;

procedure queueElSeidel.freeAll;
	begin
		if qType = vertex then
			dispose(v);
		shallowFree;
	end;

procedure queueElSeidel.interactiveOutput;
	begin
		q.interactiveOutput;
		if qType = vertex then begin
			r.interactiveOutput;
			s.interactiveOutput;
			v.interactiveOutput;
		end;
	end;

procedure updateIds (p: pointVector);
(* weist den lexikographisch geordneten Datenpunkten eine "integer" Zahl als Identifikator zu.		*)
	var
		i: seqIndex0;
	begin
		for i := 1 to p.length do
			with idRealPoint(p.elements^^[i]) do
				id := i;
	end;

function pointVector.voronoiDiagramSweepLineSeidel (ask: boolean): voronoiDiagram;
	var
		active: dictionary;
		queue: dictPriorityQueue;
		yActive: real;
		siteIndex: seqIndex0;
		firstReg: realPoint;
		vor: voronoiDiagram;
		regLinear: reference;

	function compareQueueInsert (a, b: queueElSeidel): relationObj;
(* Vergleichsprozedur zum Einfuegen eines Elementes in die queue Q. 								*)
(* Sortiert wird lexicographisch, bei lexicogr. gleichen Punkten wird nach lexicogr. Ordnung des 	*)
(* mittleren Punktes im Knoten-definierenden Tripel  (r,q,s) sortiert.                   			*)
		begin
			if a.yTransf < b.yTransf then
				compareQueueInsert := lessThanObj
			else if a.yTransf > b.yTransf then
				compareQueueInsert := greaterThanObj
			else if a.q.y < b.q.y then
				compareQueueInsert := lessThanObj
			else if a.q.y > b.q.y then
				compareQueueInsert := greaterThanObj
			else
				compareQueueInsert := equalObj;
		end; (* compareQueueInsert *)

	function compareQueueFind (a, b: queueElSeidel): relationObj;
(* Vergleichsprozedur, um ein Element in der queue Q zu finden.									 	*)
		begin
			if a.yTransf < b.yTransf - absTol then
				compareQueueFind := lessThanObj
			else if a.yTransf > b.yTransf + absTol then
				compareQueueFind := greaterThanObj
			else
				compareQueueFind := equalObj;
		end; (* compareQueueFind *)

	function Parabelschnittpunkt (a, b: realPoint): real;
(* Berechnet x-Koordinate xa des Parabelschnittpunktes, der durch die Parabeln mit Focus a bzw. 	*)
(* b und der Sweepline als Leitlinie gegeben ist,													*)
(* (= x-Koordinate des Schnittpunktes der Wellen a und b).											*)
		var
			da, db, pa, pb: double;
			xa, AA, BB, T1, T2: double;
		begin
			if a = b then
				Parabelschnittpunkt := a.x
			else begin
				if a.y = b.y then
					Parabelschnittpunkt := 0.5 * (a.x + b.x)
				else begin
					pa := yActive - a.y;
					da := a.y + 0.5 * pa;
					pb := yActive - b.y;
					db := b.y + 0.5 * pb;
					AA := (pb * a.x - pa * b.x) / (pb - pa);
					BB := (2 * pa * pb * (da - db) - pb * sqr(a.x) + pa * sqr(b.x)) / (pb - pa);
					xa := sqrt(BB + sqr(AA));
					T1 := AA - xa;
					T2 := AA + xa;
(* Zwei Parabeln mit Focus a bzw. b, mit a.y<>b.y, schneiden sich genau zweimal (in den Punkten	*)
(* T1 bzw. T2). Es ist nun jeweils der richtige Schnittpunkt zu suchen.							*)
					case (a.y < b.y) of
						true: 
							if (T1 < b.x) then
								Parabelschnittpunkt := T1
							else
								Parabelschnittpunkt := T2;
						false: 
							if a.x < T2 then
								Parabelschnittpunkt := T2
							else
								Parabelschnittpunkt := T1;
					end; (* case *)
				end; (* else *)
			end; (* else *)
		end;

	function definedPoint (w: activeElSeidel): realPoint;
(* Gibt den Punkt an, der die Welle definiert.							*)
(* Ist 'w.side'=true, dann ist 'definiedPoint'='w.e.p1',				*)
(* Ist 'w.side'=false, dann ist 'definiedPoint'='w.e.p2'.				*)
		begin
			if w.aType = newPoint then
				definedPoint := w.p
			else begin
				if w.side = true then
					definedPoint := w.e.p1
				else
					definedPoint := w.e.p2;
			end;
		end; (* definedPoint *)

	function otherPoint (w: activeElSeidel): realPoint;
(* Gibt den Punkt an, der die in L rechts benachbarte Welle definiert.	*)
(* Ist 'w.side'=true, dann ist 'definiedPoint'='w.e.p2',				*)
(* Ist 'w.side'=false, dann ist 'definiedPoint'='w.e.p1'.				*)
		begin
			if w.aType = newPoint then
				otherPoint := w.p
			else begin
				if w.side = true then
					otherPoint := w.e.p2
				else
					otherPoint := w.e.p1;
			end;
		end; (* definedPoint *)

	function compareTx (a, b: activeElSeidel): relationObj;
(* Vergleichsfunktion zum Einfgen eines Elementes in die active-list L.                           	*)
(* Hier werden nicht die Elemente der active-list selbst verglichen, sondern die            		*)
(* Tangentialpunkte, die durch das Element selbst (definedPoint(a) bzw. definedPoint(b)), seinen	*)
(* rechten Nachbarn (otherPoint(a) bzw. otherPoint(b)) und der Sweepline als Tangente gegeben ist.	*)
		var
			xa, xb: double;
		begin
			xa := Parabelschnittpunkt(definedPoint(a), otherPoint(a));
			xb := Parabelschnittpunkt(definedPoint(b), otherPoint(b));
			if (xa < xb - absTol) then
				compareTx := lessThanObj
			else if (abs(xa - xb) <= absTol) then
				compareTx := equalObj
			else
				compareTx := greaterThanObj;
		end;  (* compareTx *)

	function extractMinQueue: queueElSeidel;
(* Berechnet das minimale Element der queue Q und der noch nicht verarbeiteten Datenpunkten.		*)
(* Das Minimum wird anhand der  y-Koordinate des Datenpunktes bzw. der "obersten Spitze" des		*)
(* Kreises, der durch das Knotentripel (r,q,s) definiert ist ermittelt.								*)
(* Fallen auf einen Datenpunkt Duplikate von Knoten, werden zuerst die Knotenduplikate abge-  		*)
(* arbeitet, erst  ganz zum Schluss wird der Datenpunkt bearbeitet. Auf Grund moeglicher      	 	*)
(* Rundungsfehler, muss der Algorithmus  auch die Faelle korrekt behandeln, bei denen zuerst der	*)
(* Datenpunkt und dann ein Knotenduplikat bearbeitet wird.											*)
		var
			vertexMin: reference;
			qEl: queueElSeidel;
		begin
			vertexMin := queue.minimum;
			if (vertexMin = nil) | (siteIndex <= vor.points.length) & (realPoint(vor.points.elements^^[siteIndex]).y <= queueElSeidel(queue.getObject(vertexMin)).yTransf) then begin
				new(qEl);
				failNil(Handle(qEl));
				qEl.qType := site;
				qEl.q := realPoint(vor.points.elements^^[siteIndex]);
				siteIndex := siteIndex + 1;
(* Sind weitere Datenpunkte im Umkreis mit Radius "absTol", so werden diese zur Berechneung  	*)
(* des Voronoi-Diagramms nicht in Betracht gezogen. Die "absTol" Grenze ist so klein gewaehlt,	*)
(* dass das menschliche Auge dies nicht bemerkt.												*)
				while (siteIndex <= vor.points.length) & (abs(realPoint(vor.points.elements^^[siteIndex]).y - qEl.q.y) <= absTol) & (abs(realPoint(vor.points.elements^^[siteIndex]).x - qEl.q.x) <= absTol) do
					siteIndex := siteIndex + 1;
				extractMinQueue := qEl;
			end
			else begin
				extractMinQueue := queueElSeidel(queue.getObject(vertexMin));
				queue.delete(vertexMin, compareQueueFind);
			end;
		end; (* extractMinQueue *)

	function findWave (p: realPoint; var regHandle: reference; var found: boolean): realPoint;
(* Der Zeiger 'regHandle' verweist auf den aktuell zu bearbeitenden Datenpunkt p (vom Typ 				*)
(* 'newPoint' ), der in der active-list L hinter der Welle q eingefuegt wird, die von der entarteten 	*)
(* Welle p geschnitten wird. 'findwave' gibt dabei die Datenpunkt an, der die Welle definiert, die 		*)
(* von der Welle p geschnitten wird.																	*)
(*        found=true <=> Vertikale durch p schneidet den Schnittpunkt zweier benachbarter Wellen 		*)
		var
			np: activeElSeidel;

		function createNewPoint (p: realPoint): activeElSeidel;
(* Convertiert den aktuell zu bearbeitenden Datenpunkt p in ein Element der active-list. Erst dann		*)
(* kann p in die Liste L eingefuegt werden.																*)
			var
				np: activeElSeidel;
			begin
				new(np);
				failNil(Handle(np));
				np.aType := newPoint;
				np.p := p;
				createNewPoint := np;
			end; (* createNewPoint *)

		begin (* findWave *)
			np := createNewPoint(p);
			regHandle := active.insertObj(np, compareTx, found);
			if siteIndex = 3 then
				findWave := firstReg
			else begin
				if active.next(regHandle, left) = nil then
					findWave := definedPoint(activeElSeidel(active.getObject(active.next(regHandle, right))))
				else
					findWave := otherPoint(activeElSeidel(active.getObject(active.next(regHandle, left))));
			end;
		end; (* findWave *)

	function createWave (p, q: realPoint): activeElSeidel;
(* Wird eine neue Welle in die active-list L eingefuegt, muss der entsprechenden 	Bisektor	*)
(* erzeugt werden, d.h. die zugehoerige Kante e wird belegt mit:  e.eType=bisecktor, e.p1=a 	*)
(* und e.p2=b, falls a<b, sonst p1=b und p2=a.													*)
		var
			e: edgeSLSeidel;
			b: activeElSeidel;
		begin
			new(e);
			failNil(Handle(e));
			e.init;
			new(b);
			failNil(Handle(b));
			e.eType := bisector;
			if pointLessThanYX(p, q) then begin
				e.p1 := p;
				e.p2 := q;
				b.side := true;
			end
			else begin
				e.p1 := q;
				e.p2 := p;
				b.side := false;
			end;
			vor.edges.append(e);
			b.aType := wave;
			b.e := e;
			createWave := b
		end; (* createWave *)

	procedure createWave1 (p, q: realPoint; var bLeft, bRight: activeElSeidel);
(* Erzeugt die Welle bzw. den Bisektor, der beim Einfuegen eines neuen Datenpunktes in die 		*)
(* active-list L entsteht, siehe Funktion 'createWave'.											*)
		begin
			bRight := createWave(p, q);
			new(bLeft);
			failNil(Handle(bLeft));
			bLeft.aType := bRight.aType;
			bLeft.e := bRight.e;
			bLeft.side := true;
		end; (* createWave1 *)

	procedure createWave2 (lp, p, q: realPoint; regHandle: reference; var bLeft, bRight: activeElSeidel);
(* Erzeugt den Bisektor, der neu entsteht, wenn eine Welle aus der active-list L entfernt wird,	*)
(* siehe Funtkion 'createWave'.																	*)
		begin
			lp := definedPoint(activeElSeidel(active.getObject(active.next(regHandle, left))));
			bRight := createWave(p, q);
			bLeft := createWave(lp, p);
		end; (* createWave2 *)

	procedure updateLlinear (p: realPoint);
(* Liegen die ersten k Datenpunkte auf einer horizontalen Geraden, sind die zugehoerigen 		*)
(* Wellen entartet, d.h. sie bestehen nur aus nach unten offenen vertikalen Halbgeraden mit 	*)
(* Endpunkten auf der Sweepline. Da sich diese entarteten Wellen nicht schneiden, koennen 		*)
(* weitere colineare Datenpunkte nicht mittels der Vergleichsprocedur 'compareTx' in die Liste	*)
(* L eingefuegt werden. Fuer diesen Fall wird also eine besondere Einfuegeprozedur benoetigt,	*)
(* in der die neu eingefuegt Welle automatisch am Ende der list L gesetzt wird.					*)
		var
			np: activeElSeidel;
			help: realPoint;
		begin
			if siteIndex = 3 then begin
				np := createWave(firstReg, p);
				regLinear := active.insert(np, nil, left);
			end
			else begin
				help := otherPoint(activeElSeidel(active.getObject(regLinear)));
				np := createWave(help, p);
				regLinear := active.insert(np, regLinear, right);
			end;
		end; (*updateLlinear *)

	procedure updateL1 (regHandle: reference; bRecLeft, bRecRight: activeElSeidel; var bLeft, bRight: reference);
(* p ist ein Datenpunkt und  die Vertikale durch p verlaeuft nicht durch den Schnittpunkt zweier 	*)
(* Wellen .               																			*)
(* Aktualisiert active-list: 	- statt ...a ...  setze  ...a p a... 								*)
		begin
			active.getObject(regHandle).free;
			bRight := regHandle;
			active.setObject(bRight, bRecRight);
			bLeft := active.insert(bRecLeft, bRight, left);
		end; (* updateL1 *)

	procedure updateL3 (zq: reference; neu: activeElSeidel);
(* p ist ein Knoten, definiert durch das Tripel (a,b,c). Der Zeiger 'zq' verweist auf die mittlere	*)
(* Welle b in der active-list L, die geloescht werden muss.											*)
(* Aktualisiert active-list: 	- statt   ...a b c...  setze ...a c...								*)
		begin
			active.getObject(zq).free;
			active.getObject(active.next(zq, left)).free;
			active.setObject(active.next(zq, left), neu);
		end; (*updateL3 *)

	procedure deleteFromQueue (Z1, Z2: reference);
(* Loescht von der queue Q  den Knoten, der durch das Tripel (a1,a2,a3) in L, welches durch	*)
(* die Zeiger Z1 und  Z2 gegeben ist, falls der Knoten in Q existiert.						*)
		var
			xpointer: reference;
			v, x: queueElSeidel;
			deleted: boolean;
			a1, a2, a3, c: realPoint;
			dir: directionType;
			cx, cy: extended;
		begin
			if (Z1 <> nil) & (Z2 <> nil) then begin
				a1 := definedPoint(activeElSeidel(active.getObject(Z1)));
				a2 := definedPoint(activeElSeidel(active.getObject(Z2)));
				a3 := otherPoint(activeElSeidel(active.getObject(Z2)));
				if a1.whichSideX(a2, a3) <> 0 then begin
					new(c);
					failNil(Handle(c));
					a1.circleCenterX(a2, a3, cx, cy);
					c.x := cx;
					c.y := cy;
					deleted := false;     (* deleted=true <=> Objekt zum Lschen ist gefunden *)
					new(x);
					failNil(Handle(x));
					x.qType := vertex;
					x.yTransf := c.y + sqrt(c.squaredDistanceX(a1));
					x.r := a1;
					x.q := a2;
					x.s := a3;
					x.v := c;           (* Suchobjekt wird aufgebaut *)
					if queue.find(x, compareQueueFind, xpointer, dir) then begin
						while (queue.next(xpointer, left) <> nil) & (compareQueueFind(queueElSeidel(queue.getObject(queue.next(xpointer, left))), queueElSeidel(queue.getObject(xpointer))) > lessThanObj) do
							xpointer := queue.next(xpointer, left);
						while not deleted & (xpointer <> nil) do begin
							v := queueElSeidel(queue.getObject(xpointer));
							if (v.r = x.r) & (v.q = x.q) & (v.s = x.s) then begin
								queue.getObject(xpointer).freeAll;
								queue.delete(xpointer, compareQueueFind);
								deleted := true;
							end
							else begin
								xpointer := queue.next(xpointer, right);
								if (xpointer <> nil) & (compareQueueFind(x, queueElSeidel(queue.getObject(xpointer))) = lessThanObj) then
									xpointer := nil;
							end;
						end;
					end;
					dispose(c);
					dispose(x);
				end;
			end;
		end; (* deleteFromQueueIfExist *)

	procedure insertIntoQueueIfExist (Z1, Z2: reference);
(* Ist die "oberste Spitze" des Kreises, definiert durch das Punktetripel (Z1^,Z2^,rechter Nachbar von		*)
(* Z2^ in L) lexicogr. grer oder gleich p (Gleich, falls mehr als drei Sites cocircular liegen) 			*)
(* und liegen Z1^, Z2^ und der rechte Nachbar von Z2^ in L gegen den Uhrzeigersinn auf dem diesem Kreis, 	*)
(* d.h.  Z2^ liegt rechts von der gerichteten Geraden (Z1^, rechter Nachbar von Z2^ in L), dann wird v		*)
(* als Knoten in Q eingefgt.        																		*)
		var
			a1, a2, a3, c: realPoint;   (* c= Knoten= Mittelpunkt des Kreises gegeben durch a1, a2 und a3;	*)
			cx, cy: extended;

		procedure insertIntoQ (r, q, s, v: realPoint; Z2: reference);
			var
				qEl: queueElSeidel;
				newQueueEl: reference;
			begin
				new(qEl);
				failNil(Handle(qEl));
				qEl.qType := vertex;
				qEl.v := v;
				qEl.yTransf := v.y + sqrt(v.squaredDistanceX(r));
				qEl.r := r;
				qEl.q := q;
				qEl.s := s;
				qEl.zq := Z2;
				newQueueEl := queue.insert(qEl, compareQueueInsert);
			end; (* insertVertexIntoQueue *)

		begin (* insertIntoQueueIfExist *)
			if (Z1 <> nil) & (Z2 <> nil) then begin
				a1 := definedPoint(activeElSeidel(active.getObject(Z1)));
				a2 := definedPoint(activeElSeidel(active.getObject(Z2)));
				a3 := otherPoint(activeElSeidel(active.getObject(Z2)));
				if a2.whichSideX(a1, a3) > 0 then begin
					new(c);
					failNil(Handle(c));
					a1.circleCenterX(a2, a3, cx, cy);
					c.x := cx;
					c.y := cy;
					insertIntoQ(a1, a2, a3, c, Z2);
				end;
			end;
		end; (* insertIntoQueueIfExist *)


	procedure markEndPoint (p: realPoint; e: edgeSLSeidel);
(* Markiert Punkt p als Endpunkt der Kante e.	*)
		begin
			if e.eType = bisector then begin
				e.eType := ray;
				e.v1 := p;
			end
			else if e.eType = ray then begin
				e.eType := segment;
				e.v2 := p;
			end;
		end; (* markEndPoint *)


(************* Animation**************** **** *************  ** *)
	procedure voronoiDiagramAnimationEdge (eList: list);
		const
			maxY = 1000;

		procedure output (x: obj);
			var
				e: edgeSLSeidel;
			begin
				e := edgeSLSeidel(x);
				e.interactiveOutput;
			end;

		begin
			if yActive <= maxY then
				eList.forAll(output);
		end; (* voronoiDiagramAnimationEdge *)

{$IFC AnimationEnabled AND AvoronoiDiagramSL }

	procedure voronoiDiagramAnimation (active: dictionary);
		const
			minX = -1000;
			maxX = 1000;
			minY = 0;
			maxY = 1000;
			incX = 5;
			incY = 5;
			maxYOut = 5000;
			over = 20;
		var
			aDef, aOth, oldDef, oldOth: realPoint;
			lWave, rWave: real;
			lastType: activeType;

		function parabelX (y, d, p: real): real;
(* Berechnet die x-Koordinate, die zum Punkt auf der Parabel mit der vorgegebenen	*)
(* y-Koordinate gehoert.															*)
			var
				w: real;
			begin
				w := -2 * p * (y - d);
				parabelX := sqrt(w);
			end;

		function parabelY (x, c, d, p: real): real;
(* Berechnet die y-Koordinate, die zum Punkt auf der Parabel mit der vorgegebenen	*)
(* x-Koordinate gehoert.															*)
			begin
				parabelY := d - (x - c) * (x - c) / (2 * p);
			end;

		procedure output (x: obj);
			var
				a: activeElSeidel;
				c, d, p, yp, w: real;
				p1, p2, p3, p4: realPoint;
			begin
				a := activeElSeidel(x);
				if a.aType = newPoint then begin
					new(p2);
					failNil(Handle(p2));
					c := oldDef.x;
					p := yActive - oldDef.y;
					d := oldDef.y + 0.5 * p;
					yp := parabelY(a.p.x, c, d, p);
					p2.x := a.p.x;
					p2.y := yp - over;  (* 'over' ist eine Konstante, die angibt, wieviel ber einen Schnittpunkt heraus gezeichnet wird. *)
					a.p.drawLine(p2);
					dispose(p2);
				end
				else begin
					aDef := definedPoint(a);
					aOth := otherPoint(a);
					if abs(aDef.y - yActive) <= 0.5 then begin
						new(p2);
						failNil(Handle(p2));
						c := oldDef.x;
						p := yActive - oldDef.y;
						d := oldDef.y + 0.5 * p;
						yp := parabelY(aDef.x, c, d, p);
						p2.x := aDef.x;
						p2.y := yp - over;  (* 'over' ist eine Konstante, die angibt, wieviel ber einen Schnittpunkt heraus gezeichnet wird. *)
						aDef.drawLine(p2);
						dispose(p2);
					end
					else  (*a.atype <> newPoint *)
						begin
						rWave := Parabelschnittpunkt(aDef, aOth);
						c := aDef.x;
						p := yActive - aDef.y;
						d := aDef.y + 0.5 * p;
						new(p1);
						failNil(Handle(p1));
						new(p2);
						failNil(Handle(p2));
						if c > rWave then begin
							p1.x := rWave + over;
							p1.y := parabelY(p1.x, c, d, p);
							p2.y := p1.y - incY;
							p2.x := c - parabelX(p2.y, d, p);
							while (p2.x >= lWave - over) & (p2.x >= minX) & (p2.y >= minY) do begin
								p1.drawLine(p2);
								dispose(p1);
								p1 := p2;
								new(p2);
								failNil(Handle(p2));
								p2.y := p1.y - incY;
								p2.x := c - parabelX(p2.y, d, p);
							end;
						end  (* c > rWave  *)
						else  (* c <= rWave *)
							begin
							if c < lWave then begin
								p1.x := lWave - over;
								p1.y := parabelY(p1.x, c, d, p);
								p2.y := p1.y - incY;
								p2.x := c + parabelX(p2.y, d, p);
								while (p2.x <= rWave + over) & (p2.x <= maxX) & (p2.y >= minY) do begin
									p1.drawLine(p2);
									dispose(p1);
									p1 := p2;
									new(p2);
									failNil(Handle(p2));
									p2.y := p1.y - incY;
									p2.x := c + parabelX(p2.y, d, p);
								end;
							end  (*  c < lWave *)
							else   (* lWave <= c <= rWave *)
								begin
								p1.x := c;
								p1.y := d;
								p2.x := p1.x + incX;
								p2.y := parabelY(p2.x, c, d, p);
								new(p3);
								failNil(Handle(p3));
								new(p4);
								failNil(Handle(p4));
								p3.x := c;
								p3.y := d;
								p4.x := p3.x - incX;  (* im 1ten Schritt wird, unabhngig von K, x inkrementiert *)
								p4.y := parabelY(p4.x, c, d, p);
								while (p2.y >= minY) & ((p2.x <= rWave + over) & (p2.x <= maxX) | (p4.x >= lWave - over) & (p4.x >= minX)) do begin
									if (p2.x <= rWave + over) & (p2.x <= maxX) then
										p1.drawLine(p2);
									if (p4.x >= lWave - over) & (p4.x >= minX) then
										p3.drawLine(p4);
									dispose(p1);
									p1 := p2;
									new(p2);
									failNil(Handle(p2));
									p2.y := p1.y - incY;
									w := parabelX(p2.y, d, p);
									p2.x := c + w;
									dispose(p3);
									p3 := p4;
									new(p4);
									failNil(Handle(p4));
									p4.y := p3.y - incY;
									p4.x := c - w;
								end;
								dispose(p3);
								dispose(p4);
							end;  (*    lWave <= c <= rWave *)
						end;
						dispose(p1);
						dispose(p2);
					end; (* else -  a.atype <> newPoint *)
					lWave := rWave;
					oldDef := aDef;
					oldOth := aOth;
				end;
				lastType := a.aType;
			end; (* output *)

		procedure lastOutput;
(* Da mit der obigen Ausgabeprozedur die letzte Welle in der Wellenfront nicht			*)
(* erfasst wird, muss diese mittels einer eigenen Ausgabeprozedur gezeichnet werden.	*)
			var
				c, d, p, w: real;
				p1, p2, p3, p4: realPoint;
			begin
				c := oldOth.x;
				p := yActive - oldOth.y;
				d := oldOth.y + 0.5 * p;
				new(p1);
				failNil(Handle(p1));
				new(p2);
				failNil(Handle(p2));
				if c < lWave then begin
					p1.x := lWave - over;
					p1.y := parabelY(p1.x, c, d, p);
					p2.y := p1.y - incY;
					p2.x := c + parabelX(p2.y, d, p);
					while (p2.x <= Tpos + over) & (p2.x <= maxX) & (p2.y >= minY) do begin
						p1.drawLine(p2);
						dispose(p1);
						p1 := p2;
						new(p2);
						failNil(Handle(p2));
						p2.y := p1.y - incY;
						p2.x := c + parabelX(p2.y, d, p);
					end;
				end  (*	c < lWave			*)
				else (*	lWave <= c <= rWave	*)
					begin
					p1.x := c;
					p1.y := d;
					p2.x := p1.x + incX;
					p2.y := parabelY(p2.x, c, d, p);
					new(p3);
					failNil(Handle(p3));
					new(p4);
					failNil(Handle(p4));
					p3.x := c;
					p3.y := d;
					p4.x := p3.x - incX;  (* im 1ten Schritt wird, unabhngig von K, x inkrementiert *)
					p4.y := parabelY(p4.x, c, d, p);
					while (p2.y >= minY) & ((p2.x <= Tpos + over) & (p2.x <= maxX) | (p4.x >= lWave - over) & (p4.x >= minX)) do begin
						if (p2.x <= Tpos + over) & (p2.x <= maxX) then
							p1.drawLine(p2);
						if (p4.x >= lWave - over) & (p4.x >= minX) then
							p3.drawLine(p4);
						dispose(p1);
						p1 := p2;
						new(p2);
						failNil(Handle(p2));
						p2.y := p1.y - incY;
						w := parabelX(p2.y, d, p);
						p2.x := c + w;
						dispose(p3);
						p3 := p4;
						new(p4);
						failNil(Handle(p4));
						p4.y := p3.y - incY;
						p4.x := c - w;
					end;
					dispose(p3);
					dispose(p4);
				end;  (*    lWave <= c <= rWave *)
				dispose(p1);
				dispose(p2);
			end; (* lastOutput *)

		begin
			if yActive <= maxY then begin
				lWave := Tneg; (* die erste Welle beginnt im Unendlichen bzgl. der negativen x-Richtung. *)
				oldDef := firstReg;
				active.forAll(output, left);
				if siteIndex > 3 then
					lastOutput;
			end;
		end; (* voronoiDiagramAnimation *)
{$ENDC }


	procedure voronoiDiagramAnimationLinear (active: dictionary);
(* Bei colinearer Verteilung der k ersten Datenpunkte besteht die Wellenfront aus unten offenen		*)
(* vertikalen Halbgeraden, deren Endpunkte die jeweiligen Datenpunkte sind, insbesondere ist die	*)
(* Wellenfront nicht zusammenhaengend. Die allgemeine Ausgabeprozedur erfaSSt diesen Fall somit		*)
(* nicht. 	 																						*)
		const
			minY = 0;
			maxYOut = 5000;
			maxY = 1000;

		procedure output (x: obj);
			var
				a: activeElSeidel;
				p2: realPoint;
				aDef: realPoint;
			begin
				a := activeElSeidel(x);
				aDef := definedPoint(a);
				new(p2);
				failNil(Handle(p2));
				p2.x := aDef.x;
				p2.y := minY;
				realPoint(aDef).drawLine(p2);
				dispose(p2);
			end;

		begin (* voronoiDiagramAnimationLinear *)
			if yActive <= maxY then
				active.forAll(output, left);
		end; (* voronoiDiagramAnimationLinear *)



 (* voronoiDiagramSLSeidel *)
		var
			p: queueElSeidel;
			regHandle, bLeft, bRight: reference;
			neu, bRecLeft, bRecRight: activeElSeidel;
			found: boolean;
			r: pointVector;
			ap: realPoint;
			yalt: real;
			linear, line: boolean;
{$IFC AnimationEnabled AND AvoronoiDiagramSL }
			sweepLineP, sweepLineQ: realPoint;
{$ENDC }

	begin (* voronoiDiagramSLSeidel *)
		if self.length > 0 then begin
			if ask then begin
				if not getXY(length, true, priorityQueue(queue), 2 * length, active, false) then begin
					voronoiDiagramSweepLineSeidel := nil;
					exit(voronoiDiagramSweepLineSeidel);
				end;
			end
			else begin
				currentDictionaryType := avlTreeType;
				queue := dictPriorityQueue(newInitObj(dictPriorityQueueType));
				active := dictionary(newInitObj(avlTreeType));
			end;
{$IFC AnimationEnabled AND AvoronoiDiagramSL }
			new(sweepLineP);
			failNil(Handle(sweepLineP));
			new(sweepLineQ);
			failNil(Handle(sweepLineQ));
			sweepLineP.x := -1000;
			sweepLineQ.x := +1000;
{$ENDC }
			new(vor);
			failNil(Handle(vor));
			vor.init;
			r := makeIdPoints;
			r.sortAll(pointLessThanYX);
			updateIds(r);
			vor.points := r;
			new(vor.edges);
			failNil(Handle(vor.edges));
			vor.edges.init;
			firstReg := realPoint(vor.points.elements^^[1]);
			yalt := firstReg.y;
			siteIndex := 2;
			linear := true;
			line := true;
			while not queue.isEmpty | (siteIndex <= vor.points.length) do begin
{$IFC AnimationEnabled }
				if userAbort then begin
					queue.freeAll;
					active.freeAll;
					vor.freeAll;
					voronoiDiagramSweepLineSeidel := nil;
{$IFC AvoronoiDiagramSL }
					dispose(sweepLineP);
					dispose(sweepLineQ);
{$ENDC }
					exit(voronoiDiagramSweepLineSeidel);
				end;
{$ENDC }
				p := extractMinQueue;
				case p.qType of
					site:  begin
						yActive := p.q.y;
						if siteIndex > 2 then begin			(* Ueberpruefung, ob die ersten k Datenpunkte colinear liegen.	*)
							if line = true then begin
								linear := (yalt = yActive);
								if linear = false then
									line := false;
							end;
						end;
						case linear of
							true:  begin
								updateLlinear(p.q);
(*  AnimationOutput *)
{$IFC AnimationEnabled AND AvoronoiDiagramSL }
								if animationFlag[AVoronoiDiagramItem] & (yActive <= 2000) then begin
									sweepLineP.y := yActive;
									sweepLineQ.y := yActive;
									sweepLineP.drawLine(sweepLineQ);
									p.interactiveOutput;
									voronoiDiagramAnimationLinear(active);
									waitForClick(animationFlag[AVoronoiDiagramItem]);
									voronoiDiagramAnimationEdge(vor.edges);
									if animationFlag[AVoronoiDiagramItem] then
										waitForClick(animationFlag[AVoronoiDiagramItem]);
									sweepLineP.drawLine(sweepLineQ);
									p.interactiveOutput;
									voronoiDiagramAnimationEdge(vor.edges);
									voronoiDiagramAnimationLinear(active);
								end;
{$ENDC }
							end;
							false:  begin
								ap := findWave(p.q, regHandle, found);
(* - 'findwave' fgt den convertierten Punkt p (=ap) in die active-list ein				*)
(* - 'regHandle' ist Zeiger auf p														*)
(* - 'found'=true <=> die Vertikale durch p geht durch den Schnittpunkt der links 		*)
(*						 bzw rechts in der active-list benachbarten Wellen von p 		*)
(* In der active-list steht p genau hinter der Welle (a), die von der Vertikalen durch 	*)
(* p auf jeden Fall geschnitten wird. 													*)
(*  AnimationOutput *)
{$IFC AnimationEnabled AND AvoronoiDiagramSL }
								if animationFlag[AVoronoiDiagramItem] & (yActive <= 2000) then begin
									sweepLineP.y := yActive;
									sweepLineQ.y := yActive;
									sweepLineP.drawLine(sweepLineQ);
									p.interactiveOutput;
									voronoiDiagramAnimation(active);
									waitForClick(animationFlag[AVoronoiDiagramItem]);
									voronoiDiagramAnimationEdge(vor.edges);
									if animationFlag[AVoronoiDiagramItem] then
										waitForClick(animationFlag[AVoronoiDiagramItem]);
									sweepLineP.drawLine(sweepLineQ);
									p.interactiveOutput;
									voronoiDiagramAnimation(active);
									voronoiDiagramAnimationEdge(vor.edges);
								end;
{$ENDC }
								createWave1(p.q, ap, bRecLeft, bRecRight);
								deleteFromQueue(active.next(regHandle, left), active.next(regHandle, right));
					(* lscht aus Q den Knoten v*(-,a,-), falls dieser in Q existiert *)
								updateL1(regHandle, bRecLeft, bRecRight, bLeft, bRight);	(* aktualisiert L: statt ... p,a...  steht  ...a, p, a...  *)
								insertIntoQueueIfExist(active.next(bLeft, left), bLeft);
					(* fgt in Q den Knoten v*(p,b,-) ein, falls dieser lexicographisch grer oder gleich p ist *)
								insertIntoQueueIfExist(bRight, active.next(bRight, right));
					(* fgt in Q den Knoten v*(-,a,p) ein, falls dieser lexicographisch grer oder gleich p ist *)
							end;
						end;
						yalt := yActive;
					end; (* site *)

					vertex: 	(*   p=v*(a,b,c)	*)
						begin
						yActive := p.yTransf;
(*  AnimationOutput *)
{$IFC AnimationEnabled AND AvoronoiDiagramSL }
						if animationFlag[AVoronoiDiagramItem] & (yActive <= 2000) then begin
							sweepLineP.y := yActive;
							sweepLineQ.y := yActive;
							sweepLineP.drawLine(sweepLineQ);
							p.interactiveOutput;
							voronoiDiagramAnimation(active);
							waitForClick(animationFlag[AVoronoiDiagramItem]);
							voronoiDiagramAnimationEdge(vor.edges);
							if animationFlag[AVoronoiDiagramItem] then
								waitForClick(animationFlag[AVoronoiDiagramItem]);
							sweepLineP.drawLine(sweepLineQ);
							p.interactiveOutput;
							voronoiDiagramAnimation(active);
							voronoiDiagramAnimationEdge(vor.edges);
						end;
{$ENDC }
						neu := createWave(p.r, p.s);
				   (* p wird in die active-list in dem Tripel hinter dem mittleren Punkt - b -  eingefgt.	*)
				   (* regHandle verweist auf p.																*)
						deleteFromQueue(active.next(active.next(p.zq, left), left), active.next(p.zq, left));
			        (* lscht aus Q den Knoten v*(-,a,b), falls dieser in Q existiert *)
						deleteFromQueue(p.zq, active.next(p.zq, right));
				  (* lscht aus Q den Knoten v*(b,c,-), falls dieser in Q existiert *)
						markEndPoint(p.s, activeElSeidel(active.getObject(active.next(p.zq, left))).e);
			       (* markiert Bisektor Bab mit c *)
						markEndPoint(p.r, activeElSeidel(active.getObject(p.zq)).e);
				  (* markiert Bisektor Bbc mit a *)
						markEndPoint(p.q, neu.e);
				  (* markiert Bisektor Bac mit b *)
						updateL3(p.zq, neu);
						insertIntoQueueIfExist(active.next(active.next(p.zq, left), left), active.next(p.zq, left));
				  (* fgt in Q den Knoten v*(-,a,c) ein, falls dieser lexicographisch grer oder gleich p ist *)
						insertIntoQueueIfExist(active.next(p.zq, left), active.next(p.zq, right));
				  (* fgt in Q den Knoten v*(a,c,-) ein, falls dieser lexicographisch grer oder gleich p ist *)
						active.delete(p.zq);
					end;   (* vertex *)

				end; (* case *)
				p.freeAll;
			end; (* while queueRoot <> nil *)

			if active.infoString <> '' then
				displayInfoString(concat('Active ', active.infoString));
			queue.free;
			active.freeAll;
			voronoiDiagramSweepLineSeidel := vor;
{$IFC AnimationEnabled AND AvoronoiDiagramSL }
			dispose(sweepLineP);
			dispose(sweepLineQ);
{$ENDC }
		end (* if vor.points.n > 0 *)
		else
			voronoiDiagramSweepLineSeidel := nil;
	end; (* pointVector.voronoiDiagramSweepLineSeidel *)

end. (* voronoiDiagramSLSeidel *)
